/*
 Copyright (C) 2006 Pedro Felzenszwalb

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include "apply_stereo.h"

int ITER; // number of BP iterations at each scale
int LEVELS; // number of scales

float DISC_K; // truncation of discontinuity cost
float DATA_K; // truncation of data cost
float LAMBDA; // weighting of data cost

float INF; // large cost
int VALUES; // number of possible disparities
int SCALE; // scaling from disparity to graylevel in output

double SIGMA; // amount to smooth the input images


// dt of 1d function
void dt(float *f) {
	for (int q = 1; q < VALUES; q++) {
		float prev = f[q - 1] + 1.0F;
		if (prev < f[q])
			f[q] = prev;
	}
	for (int q = VALUES - 2; q >= 0; q--) {
		float prev = f[q + 1] + 1.0F;
		if (prev < f[q])
			f[q] = prev;
	}
}

// compute message
void msg(float *s1, float *s2, float *s3, float *s4, float *dst) {
	float val;

	// aggregate and find min
	float minimum = INF;
	for (int value = 0; value < VALUES; value++) {
		dst[value] = s1[value] + s2[value] + s3[value] + s4[value];
		if (dst[value] < minimum)
			minimum = dst[value];
	}

	// dt
	dt(dst);

	// truncate
	minimum += DISC_K;
	for (int value = 0; value < VALUES; value++)
		if (minimum < dst[value])
			dst[value] = minimum;

	// normalize
	val = 0;
	for (int value = 0; value < VALUES; value++)
		val += dst[value];

	val /= VALUES;
	for (int value = 0; value < VALUES; value++)
		dst[value] -= val;
}

// computation of data costs
image<float*> *comp_data(image<uchar> *img1, image<uchar> *img2) {
	int width = img1->width();
	int height = img1->height();
	image<float*> *data = initialize_image(new image<float*> (width, height));

	image<float> *sm1, *sm2;
	if (SIGMA >= 0.1) {
		sm1 = smooth(img1, SIGMA);
		sm2 = smooth(img2, SIGMA);
	} else {
		sm1 = imageUCHARtoFLOAT(img1);
		sm2 = imageUCHARtoFLOAT(img2);
	}

	for (int y = 0; y < height; y++) {
		for (int x = VALUES - 1; x < width; x++) {
			for (int value = 0; value < VALUES; value++) {
				float val = abs(imRef(sm1, x, y) - imRef(sm2, x - value, y));
				imRef(data, x, y)[value] = LAMBDA * std::min(val, DATA_K);
			}
		}
	}

	delete sm1;
	delete sm2;
	return data;
}

// generate output from current messages
image<uchar> *output(image<float*> *u, image<float*> *d, image<float*> *l,
		image<float*> *r, image<float*> *data) {
	int width = data->width();
	int height = data->height();
	image<uchar> *out = new image<uchar> (width, height);

	for (int y = 1; y < height - 1; y++) {
		for (int x = 1; x < width - 1; x++) {
			// keep track of best value for current pixel
			int best = 0;
			float best_val = INF;
			for (int value = 0; value < VALUES; value++) {
				float val = imRef(u, x, y + 1)[value]
						+ imRef(d, x, y - 1)[value] + imRef(l, x + 1, y)[value]
						+ imRef(r, x - 1, y)[value] + imRef(data, x, y)[value];
				if (val < best_val) {
					best_val = val;
					best = value;
				}
			}
			imRef(out, x, y) = best * SCALE;
		}
	}

	return out;
}

// belief propagation using checkerboard update scheme
void bp_cb(image<float*> *u, image<float*> *d, image<float*> *l,
		image<float*> *r, image<float*> *data) {
	int width = data->width();
	int height = data->height();

	for (int t = 0; t < ITER; t++) {
		std::cout << "iter " << t << "\n";

		for (int y = 1; y < height - 1; y++) {
			for (int x = ((y + t) % 2) + 1; x < width - 1; x += 2) {

				msg(imRef(u, x, y + 1), imRef(l, x + 1, y), imRef(r, x - 1, y),
						imRef(data, x, y), imRef(u, x, y));

				msg(imRef(d, x, y - 1), imRef(l, x + 1, y), imRef(r, x - 1, y),
						imRef(data, x, y), imRef(d, x, y));

				msg(imRef(u, x, y + 1), imRef(d, x, y - 1), imRef(r, x - 1, y),
						imRef(data, x, y), imRef(r, x, y));

				msg(imRef(u, x, y + 1), imRef(d, x, y - 1), imRef(l, x + 1, y),
						imRef(data, x, y), imRef(l, x, y));

			}
		}
	}
}

// multiscale belief propagation for image restoration
image<uchar> *stereo_ms(image<uchar> *img1, image<uchar> *img2, int iterations,
		int scales, float trunk_disc_cost, float trunk_data_cost,
		float weight_data_cost, float large_cost, int poss_disparities,
		int scaling, double smoothing) {

	ITER = iterations;
	LEVELS = scales;
	DISC_K = trunk_disc_cost;
	DATA_K = trunk_data_cost;
	LAMBDA = weight_data_cost;
	INF = large_cost;
	VALUES = poss_disparities;
	SCALE = scaling;
	SIGMA = smoothing;

	image<float*> **u = new image<float *> *[LEVELS];
	image<float*> **d = new image<float *> *[LEVELS];
	image<float*> **l = new image<float *> *[LEVELS];
	image<float*> **r = new image<float *> *[LEVELS];
	image<float*> **data = new image<float *> *[LEVELS];

	std::cout << "1" << std::endl;
	// data costs
	data[0] = comp_data(img1, img2);

	std::cout << "2" << std::endl;
	// data pyramid
	for (int i = 1; i < LEVELS; i++) {
		int old_width = data[i - 1]->width();
		int old_height = data[i - 1]->height();
		int new_width = (int) ceil(old_width / 2.0);
		int new_height = (int) ceil(old_height / 2.0);

		assert(new_width >= 1);
		assert(new_height >= 1);

		data[i] = initialize_image(new image<float*> (new_width, new_height));
		for (int y = 0; y < old_height; y++) {
			for (int x = 0; x < old_width; x++) {
				for (int value = 0; value < VALUES; value++) {
					imRef(data[i], x / 2, y / 2)[value] += imRef(data[i - 1],
							x, y)[value];
				}
			}
		}
	}
	std::cout << "3" << std::endl;

	// run bp from coarse to fine
	for (int i = LEVELS - 1; i >= 0; i--) {
		int width = data[i]->width();
		int height = data[i]->height();

		// allocate & init memory for messages
		if (i == LEVELS - 1) {
			// in the coarsest level messages are initialized to zero
			u[i] = initialize_image(new image<float*> (width, height));
			d[i] = initialize_image(new image<float*> (width, height));
			l[i] = initialize_image(new image<float*> (width, height));
			r[i] = initialize_image(new image<float*> (width, height));
		} else {
			// initialize messages from values of previous level
			u[i] = initialize_image(new image<float*> (width, height, false));
			d[i] = initialize_image(new image<float*> (width, height, false));
			l[i] = initialize_image(new image<float*> (width, height, false));
			r[i] = initialize_image(new image<float*> (width, height, false));

			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					for (int value = 0; value < VALUES; value++) {
						imRef(u[i], x, y)[value]
								= imRef(u[i + 1], x / 2, y / 2)[value];
						imRef(d[i], x, y)[value]
								= imRef(d[i + 1], x / 2, y / 2)[value];
						imRef(l[i], x, y)[value]
								= imRef(l[i + 1], x / 2, y / 2)[value];
						imRef(r[i], x, y)[value]
								= imRef(r[i + 1], x / 2, y / 2)[value];
					}
				}
			}
			// delete old messages and data
			delete destroy_image(u[i + 1]);
			delete destroy_image(d[i + 1]);
			delete destroy_image(l[i + 1]);
			delete destroy_image(r[i + 1]);
			delete destroy_image(data[i + 1]);
		}

		// BP
		bp_cb(u[i], d[i], l[i], r[i], data[i]);

	}
	std::cout << "4" << std::endl;
	image<uchar> *out = output(u[0], d[0], l[0], r[0], data[0]);

	delete destroy_image(u[0]);
	delete destroy_image(d[0]);
	delete destroy_image(l[0]);
	delete destroy_image(r[0]);
	delete destroy_image(data[0]);

	return out;
}

image<float *> * initialize_image(image<float *> *im) {
	for (int x = 0; x < im->width(); ++x) {
		for (int y = 0; y < im->height(); ++y) {
			float * f = new float[VALUES];
			for (int i = 0; i < VALUES; ++i) {
				f[i] = 0.f;
			}
			im->data[y * im->width() + x] = f;
		}
	}
	return im;
}

image<float *> * destroy_image(image<float *> *im) {
	for (int x = 0; x < im->width(); ++x) {
		for (int y = 0; y < im->height(); ++y) {
			delete[] im->data[y * im->width() + x];
		}
	}
	return im;
}
