

/*  Copyright (C) 2009 2010 Mirko Schmidt

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file adjustbase.hxx
 *  Implementation of the parameter class adjustbase.
 *  \author Mirko Schmidt
 *  \date 13.04.2010
 */

#ifndef _ADJUSTBASE_HXX_
#define _ADJUSTBASE_HXX_

#include <hekate/AdjustBase.h>

template <typename T>
adjustbase<T>::adjustbase(const std::string& name) :
		TemplatedParameteredObject<T>("adjustbase", name,
			"This module generates a stereo image from a given image and its depth map. It also adjusts the baseline of a given stereo image pair, using a depth map computed before. " 
			"Use the parameters alphaL and alphaR to specify the baseline between left and right image.")
{
	ParameteredObject::_addParameter (alpha_l, "alphaL", "Amount of how strong new stereo view is shifted in the left image", 1.0f);
	ParameteredObject::_addParameter (alpha_r, "alphaR", "Amount of how strong the stereo view is shifted in the right image", 1.0f);
	ParameteredObject::_addParameter (interpolation_style, "interpolation_style", "interpolation style for inpainting: 1 no interpolation, 2 linear", 2, "int");

	
	ParameteredObject::_addInputSlot(inputImagesL, "imageSequenceL", "image input sequence left", "CImgList<T>"); 
	ParameteredObject::_addInputSlot(inputImagesR, "imageSequenceR", "image input sequence right", "CImgList<T>"); 
	ParameteredObject::_addInputSlot(image_disparity_map, "disparity_map", "disparity map", "CImgList<T>"); 
	
	ParameteredObject::_addOutputSlot(outputImagesL, "outputSequenceL", "Synthesized stereo images left", "CImgList<T>"); 
	ParameteredObject::_addOutputSlot(outputImagesR, "outputSequenceR", "Synthesized stereo images right", "CImgList<T>"); 

	ParameteredObject::_addOutputSlot(outputSlotThis, "outputSlotThis", "pointer to instance of this plugin", "adjustbase<T>*");

	outputSlotThis = this;
	called = false;

	ParameteredObject::_setTags("Hekate") ;
}

template <typename T>
void adjustbase<T>::execute() {
	
	if( this->inputImagesL().size() != this->inputImagesR().size() )
		throw( std::range_error("Number of left and right images are not equal.") );
	else if( this->inputImagesL().size() != this->image_disparity_map().size() )
		throw( std::range_error("Number of left and disparity images are not equal.") );

	std::cout << "\tAdjustBase called with alpha_l = "
				<< alpha_l << " alpha_r = " << alpha_r << std::endl;

	for( size_t i = 0; i < this->inputImagesL().size(); i++ )
	{
	
		cimg_library::CImg<T> image_il, image_ir, image_ol, image_or, image_disparity;

		image_il.assign(this->inputImagesL()[i]);
		image_ir.assign(this->inputImagesR()[i]);

		image_ol.assign(image_il.width(),image_il.height(), 1, 3);
		//image_ol.assign(input_image_left()[0]);

		image_or.assign(image_ir.width(),image_ir.height(), 1, 3);
		//image_or.assign(input_image_right()[0]);

		image_disparity.assign(image_disparity_map()[i]);

	//	int lastNonZeroX=-1;

	// the following is only necessary if we have real depth maps
		//float p;	// [pixels]		disparity = displacement of features;	p = l + r
		//float bd;	// [m^2]		bd = b * d = baseline * focal length;	p = bd / z		// correction by pixel pitch necessary
		//bd = (float) 0.065 * 0.03; // typical: b = 65mm, f = 30mm 
		// bd has to be estimated from data! We need the real values here, or (simplier) we have to use the disparity map p.

		// here, the depth map is a disparity map!

		if (   image_il.width() != image_ol.width() || image_il.height() != image_ol.height()
			|| image_ir.width() != image_or.width() || image_ir.height() != image_or.height() 
			|| image_ir.width() != image_il.width() || image_ir.height() != image_il.height()
			|| image_disparity.width() != image_il.width() || image_disparity.height() != image_il.height())
		{
			throw "AdjustBase: Input images (left, right or disparity) are not of equal size!";
		}

		// computing output images, assuming left and right in and out images have same sizes
		float d; // pixel shift (original position -> new position)
		T* p_i = image_il.data();
		T* p_o = image_ol.data();
		T* p_disp = image_disparity.data();
		long pos_i = 0;
		long pos_o = 0;
		int width = image_ol.width();
		long im_len = image_ol.width() * image_ol.height();
		int x = 0;
		

		switch (this->interpolation_style)
		{
			case 0:
			case 1: // no interpolation, leave holes	# pointer based	########################
				p_i = image_il.data();
				p_o = image_ol.data();
				p_disp = image_disparity.data();
				pos_i = 0;
				pos_o = 0;
				width = image_ol.width();
				im_len = image_ol.width() * image_ol.height();
				x = 0;
			 /* So, a color image with dim=3 and depth=1, will be stored in memory as :
				R1R2R3R4R5R6......G1G2G3G4G5G6.......B1B2B3B4B5B6.... (i.e following a 'planar' structure)
				and *not* as R1G1B1R2G2B2R3G3B3... (interleaved channels), where R1 = img(0,0,0,0) is the first upper-left pixel of the red component of the image, R2 is img(1,0,0,0), G1 = img(0,0,0,1), G2 = img(1,0,0,1), B1 = img(0,0,0,2), and so on...
				Another example, a (1x5x1x1) CImg<T> (column vector A) will be stored as : A1A2A3A4A5 where A1 = img(0,0), A2 = img(0,1), ... , A5 = img(0,4).
				- taken from http://cimg.sourceforge.net/reference/group__cimg__storage.html */

				while (pos_o < image_ol.width() * image_ol.height() * image_ol.spectrum())		// left image
				{
					d = *p_disp++ / 2 * (this->alpha_l - 1);	// disparity shift d
					d *= -1;									// stereocv uses negative definition???
					if (floor(d) == d && (x + d) > 0 && (x + d) < width)
					{
						p_o[pos_o] = p_i[pos_i + (long)d];
						pos_i++;
						pos_o++;
					}
					else
					{
						p_o[pos_o] = 0;
						pos_i++;
						pos_o++;
					}
					x++;
				
					if (x == width) x = 0;// new line
					if (pos_o % im_len == 0)  p_disp = image_disparity.data(); // use only first channel of disparity image
				}


				p_i = image_ir.data();
				p_o = image_or.data();
				p_disp = image_disparity.data();
				pos_i = 0;
				pos_o = 0;
				width = image_or.width();
				im_len = image_or.width() * image_or.height();
				x = 0;
				while (pos_o < image_or.width() * image_or.height() * image_or.spectrum())		// right image
				{
					d = *p_disp++ / 2 * (this->alpha_r - 1);	// disparity shift d
					d *= -1;									// stereocv uses negative definition???

					if (floor(d) == d && (x - d) > 0 && (x - d) < width)
					{
						p_o[pos_o] = p_i[pos_i - (long)d];
						pos_i++;
						pos_o++;
					}
					else
					{
						p_o[pos_o] = 0;
						pos_i++;
						pos_o++;
					}
					x++;
				
					if (x == width) x = 0;// new line
					if (pos_o % im_len == 0)  p_disp = image_disparity.data(); // use only first channel of disparity image
				}

				break;
		
			case 2: // linear interpolation				# pointer based	############################
				p_i = image_il.data();
				p_o = image_ol.data();
				p_disp = image_disparity.data();
				pos_i = 0;
				pos_o = 0;
				width = image_ol.width();
				im_len = image_ol.width() * image_ol.height();
				x = 0;

				while (pos_o < image_ol.width() * image_ol.height() * image_ol.spectrum())		// left image
				{
					d = *p_disp++ / 2 * (this->alpha_l - 1);	// disparity shift d
					d *= -1;									// stereocv uses negative definition???

					if ((x + d) > 0 && (x + d) < width)	// if in image (border check)
					{
						p_o[pos_o] = p_i[pos_i + (long) floor(d)] + ((d - floor(d)) * (p_i[pos_i + (long) ceil(d)] - p_i[pos_i + (long) floor(d)])); // linear interpolation

						pos_i++;
						pos_o++;
					}
					else
					{
						p_o[pos_o] = 0;
						pos_i++;
						pos_o++;
					}
					x++;
				
					if (x == width) x = 0;// new line
					if (pos_o % im_len == 0)  p_disp = image_disparity.data(); // use only first channel of disparity image
				}

				p_i = image_ir.data();
				p_o = image_or.data();
				p_disp = image_disparity.data();
				pos_i = 0;
				pos_o = 0;
				width = image_or.width();
				im_len = image_or.width() * image_or.height();
				x = 0;

				while (pos_o < image_or.width() * image_or.height() * image_or.spectrum())		// right image
				{
					d = *p_disp++ / 2 * (this->alpha_r - 1);	// disparity shift d
					d *= -1;									// stereocv uses negative definition???

					if ((x - d) > 0 && (x - d) < width)	// if in image (border check)
					{
						p_o[pos_o] = p_i[pos_i - (long) ceil(d)] + ((ceil(d) - (d)) * (p_i[pos_i - (long) floor(d)] - p_i[pos_i - (long) ceil(d)])); // linear interpolation

						pos_i++;
						pos_o++;
					}
					else
					{
						p_o[pos_o] = 0;
						pos_i++;
						pos_o++;
					}
					x++;
				
					if (x == width) x = 0;// new line
					if (pos_o % im_len == 0)  p_disp = image_disparity.data(); // use only first channel of disparity image
				}


				break;

			case 11: // no interpolation, leave holes	# cimg based			##########################
				for(int c = 0; c < 3; c++) // color
				{
					for(int y = 0; y < image_il.height(); y++)
					{
						for(int x = 0; x < image_il.width(); x++)
						{
							d = *p_disp++ / 2 * (this->alpha_l - 1);	// disparity shift d
							d *= -1;									// stereocv uses negative definition???

							if (d == floor(d) && (x + d) > 0 && (x + d) < image_il.width())
							{
								image_ol(x,y,0,c) = image_il((x + d), y, 0, c);
							} 
							else 
							{
								image_ol(x,y,0,c) = 0;
							}

							d = *p_disp++ / 2 * (this->alpha_r - 1);	// disparity shift d
							d *= -1;									// stereocv uses negative definition???
							if (d == floor(d) && (x - d) > 0 && (x - d) < image_il.width())
							{
								image_or(x,y,0,c) = image_ir((x - d), y, 0, c);
							} 
							else 
							{
								image_or(x,y,0,c) = 0;
							}
						
						}
					}
				}
				break;
			case 12: // linear interpolation	# cimg based		###########################
				for(int c = 0; c < 3; c++) // color
				{
					for(int y = 0; y < image_il.height(); y++)
					{
						for(int x = 0; x < image_il.width(); x++)
						{
							//p = image_disparity(x,y); // if conversion depth_map -> disparity_map: p =  bd / image_disparity(x,y);
							//d *= -1;									// stereocv uses negative definition???							###### debug here: perhaps correct ..x + -> ...x - and vice versa
							image_ol(x,y,0,c) = image_il.linear_atX(x + (image_disparity(x,y)/2*(this->alpha_l - 1)), y, 0, c);
							image_or(x,y,0,c) = image_ir.linear_atX(x - (image_disparity(x,y)/2*(this->alpha_r - 1)), y, 0, c);
						}
					}
				}
				break;
		}

		this->outputImagesL().push_back( image_ol );
		this->outputImagesR().push_back( image_or );

	}

}

#endif /* _ADJUSTBASE_HXX_ */
