/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file BackProject.hxx
 *  Implementation of the parameter class BackProject.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 27.01.2013
 */

#ifndef _BACKPROJECT_HXX_
#define _BACKPROJECT_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/BackProject.h>
#include <ImgTool.hxx>

template <typename T>
BackProject<T>::BackProject(const std::string& name) :
		TemplatedParameteredObject<T>(
			"BackProject", name,
			"<h2>Back projects 3D points by cameras</h2><br>"
			"Back projects 3D points by cameras"
			), correspondencesIn(true,false)
{
	ParameteredObject::_addInputSlot(
		points3D, "points3D",
		"points3D",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(cameraMatrices, "cameraMatrices", 
	"cameraMatrices", 
	"IndexedMatrixList"); 
	ParameteredObject::_addInputSlot(
		correspondencesIn, "correspondencesIn",
		"Correspondences to check, which 3d points are visible in the according 2d frames",
		"CorrespondenceList");
	
	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");
	
	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void BackProject<T>::execute() {

	
	const IndexedMatrixList &cameras = this->cameraMatrices();
	size_t nCams = cameras.size();
	const CorrespondenceList<T> &points = this->points3D();
	CorrespondenceList<T> &corrs = this->correspondences();
	
	if( !this->correspondencesIn.connected() )
		sout << "ATTENTION: No correspondences connected. Points will "
			<< "be backprojected to every image no matter if the 2d point "
			<< "can natively can be seen or not" << std::endl;

	sout << "points.size() " << points.size() << std::endl;
	sout << "points.sizePoints(0) " << points.sizePoints(0) << std::endl;
	
	sout << "Backprojection of 3D to 2D points:" << std::endl;


	int nInlier = 0;
	int number = 0;
	
	for(size_t i = 0; i < nCams; i++)
	{

		sout << "Camera " << cameras[i].getIndex1() << std::endl;
		if( cameras[i].getIndex1() != i )
			sout << "Current camera index != current loop iteration:\n"
				<< "\ti = " << i << " / cameras[i].getIndex1() = " << cameras[i].getIndex1() << std::endl;

		ImgTool::printInfo( sout, cameras[i], "camera");

		sout << "Back project points of Image " << i << std::endl;
		int numberOfIt = 0;
		int notOnPlane = 0;

		auto it = points.begin() ;
		auto itEnd = points.end() ;
		for(; it != itEnd ; it++)
		{

			// Check whether the 3d point is visible in the current image
			// If the 3D point is not visible, continue without back projection
			if( this->correspondencesIn.connected() )
			{
				if( !this->correspondencesIn().exist( cameras[i].getIndex1(), it->pointIndex ) )
					continue;
			}

			// Get backprojected point
			CImg<double> X, x;
			X = (CImg<double>) it->pt3d();
			x =  cameras[i] * X;
			if(x[2] != 0) x /= x[2];

			corrs.push_back( Correspondence<T>(x[0], x[1], 0, x[2], it->pointIndex, cameras[i].getIndex1(), it->outlier, it->R, it->G, it->B) );
			
			number++;
		}

		sout << "\t" << notOnPlane << " points not on plane" << std::endl;
		sout << "Number of points backprojected: " << numberOfIt << std::endl;

	}

	sout << number << " features backprojected" << std::endl;
}


#endif /* _BACKPROJECT_HXX_ */
