/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file BackProjection3D2D.hxx
 *  Implementation of the parameter class BackProjection3D2D.
 *  \author Moritz Becker
 *  \date 16.02.2011
 */

#ifndef _BACKPROJECTION3D2D_HXX_
#define _BACKPROJECTION3D2D_HXX_

#include <hekate/BackProjection3D2D.h>
#include <charon-utils/ImgTool.hxx>

template <typename T>
BackProjection3D2D<T>::BackProjection3D2D(const std::string& name) :
		TemplatedParameteredObject<T>(
				"BackProjection3D2D", name,
				"Projects 3D points X to 2D points x,x' via Camera Matrices P,P'. (x = P * X). "
				"The module also removes bad points and cameras."
		)
			,cameraMatrices(false, false) 
			,points3D(false, false) 
			
			
{
	
	ParameteredObject::_addParameter< bool >(
		removeWholeCorrespondences, "removeWholeCorrespondences",
		"Remove whole correspondence if one point of it is declared as outlier",
		false, "bool");
	ParameteredObject::_addParameter< float >(
		maxBackProjectionError, "maxBackProjectionError",
		"Maximal back projection error. The module also removes points with backprojection errors taller than threshold 'maxBackProjectionError'.",
		2.0f, "float");
	ParameteredObject::_addParameter< int >(
		min3d2dCorrs, "min3d2dCorrs",
		"Minimum number of 3d-2d point correspondences used to retrieve views."
		"If there are less correspondences, the camera is set to degenerated.",
		50, "int");

	ParameteredObject::_addInputSlot(cameraMatrices, "cameraMatrices", 
	"cameraMatrices", 
	"IndexedMatrixList"); 
	ParameteredObject::_addInputSlot(points3D, "points3D", 
	"points3D", 
	"CorrespondenceList"); 
	ParameteredObject::_addInputSlot(correspondences, "correspondences", 
	"correspondences", 
	"CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(correspondencesBack, "correspondencesBack", 
	"correspondencesBack", 
	"CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(correspondencesInliers, "correspondencesInliers", 
	"Corrected input correspondences where 2d points whose 3d point"
	" has a a taller back projection error as defined are marked as outliers", 
	"CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(pointCloudInliers, "pointCloudInliers", 
	"3D points whos backprojection is smaller than a defined threshold in all views.", 
	"CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(cameraMatricesOut, "cameraMatricesOut", 
	"cameraMatricesOut", 
	"IndexedMatrixList"); 
	
	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void BackProjection3D2D<T>::execute() {
	
	const IndexedMatrixList &cameras = this->cameraMatrices();
	size_t nCams = cameras.size();
	const CorrespondenceList<T> &points = this->points3D();
	CorrespondenceList<T> &corrs = this->correspondencesBack();
	CorrespondenceList<T> corrsInput = this->correspondences();

	this->pointCloudInliers().clear();

	sout << "Backprojection of 3D to 2D points:" << std::endl;

	int nInlier = 0;

	std::vector<int> corrs3d2d;
	corrs3d2d.resize(nCams);

	auto it = points.begin();
	auto itEnd = points.end();
	for(;it != itEnd ; it++)
	{

		if(it->outlier)
			continue;
	
		bool inlier = true;
		int inlierCount = 0;

		for(size_t i = 0; i < nCams; i++)
		{
				
			size_t imageId = cameras[i].getIndex1();

			try{

				// Try to get correspondence
				Correspondence<T> &correspondence = corrsInput(imageId, it->pointIndex);
				CImg<float> x2 = correspondence.pt2d();

				// Get camera 
				CImg<double> cam = cameras[i];
		
				// Get backprojected point
				CImg<double> X, x;
				X = (CImg<double>) it->pt3d();
				x = cam * X;
				if(x[2] != 0) x /= x[2];

				// Continue if outlier
				if(correspondence.outlier)
					continue;

				// Get distance
				double dist = getDist(x, x2);
				if( dist > this->maxBackProjectionError() )
				{
					inlier = false;
					correspondence.outlier = true;
				}
				else
				{
					inlierCount++;
					corrs3d2d[i]++;
					corrs.push_back( Correspondence<T>(x[0], x[1], 0, x[3], it->pointIndex, imageId, false, it->R, it->G, it->B) );
				}
			}
			catch(...){}
		
		}

		if(inlierCount > 1)
		{
			this->pointCloudInliers().push_back(*it);
			nInlier++;
		}
		
	}

	// Check, if there are enough 3d-2d correspondences for every cam
	this->cameraMatricesOut() = cameras;
	for(size_t i = 0; i < nCams; i++)
	{
		if(corrs3d2d[i] < this->min3d2dCorrs())
		{
			// Set cam to degenerated
			this->cameraMatricesOut()[i].degenerated = true;
			// Write information to log, if cameras has not been set to degenerated before
			if(!cameras[i].degenerated)
			{
			sout << "Camera P" << this->cameraMatricesOut()[i].getIndex1() << " set to degenerated"
				<< " due to poor 3d-2d correspondences (" << corrs3d2d[i] << ")" << std::endl;
			}
		}
	}

	int nPoint = points.size();
	sout << nPoint - nInlier << " outliers of " << nPoint << " points "
		<< "removed due to bad back projections. (error > " 
		<< this->maxBackProjectionError() << ")" << std::endl;
	sout << this->pointCloudInliers().size() <<  " 3d inliers have been put to output slot." << std::endl;

	corrsInput.removeOutlierCorrespondences();
	this->correspondencesInliers() = corrsInput;

}

double getDist(const CImg<float> &pt1, const CImg<float> &pt2)
{
	return std::sqrt( pow((double)(pt1[0]-pt2[0]),2.0) + pow((double)(pt1[1]-pt2[1]),2.0) );		
}

#endif /* _BACKPROJECTION3D2D_HXX_ */



