/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CImgListFromPath.hxx
 *  Implementation of the parameter class CImgListFromPath.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.10.2012
 */

#ifndef _CIMGLISTFROMPATH_HXX_
#define _CIMGLISTFROMPATH_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/CImgListFromPath.h>

template <typename T>
CImgListFromPath<T>::CImgListFromPath(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CImgListFromPath", name,
			"<h2>Reads in images from path</h2><br>"
			"Reads in CImgList using a given path vector."
		)
{
	ParameteredObject::_addInputSlot(
		paths, "paths",
		"paths",
		"vector<string>");
	ParameteredObject::_addOutputSlot(
		images, "images",
		"images",
		"CImgList<T>");
			
	ParameteredObject::_addParameter< bool >(
		readFlo, "readFloFiles", "Read .flo files from string list. U and V of flow are save in C-axis.", false);

	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void CImgListFromPath<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
		
	// Clear output slot (for the case the workflow is run more than once)
	this->images().clear();

	int _w = 0;
	int _h = 0;	
	
	// Read images
	if(!this->readFlo())
	{
		// Get image dimensions
		for( size_t i = 0; i < this->paths().size(); i++){
			if( this->paths()[i] != "" ){
				cimg_library::CImg<T> img( this->paths()[i].c_str() );
				_w = img.width();
				_h = img.height();
			}
		}

		sout << "Read in " << this->paths().size() << " images" << std::endl;

		for( size_t i = 0; i < this->paths().size(); i++){

			if( this->paths()[i] != "" ){
				cimg_library::CImg<T> img( this->paths()[i].c_str() );
				this->images().push_back( img );
			}
			else{
				this->images().push_back( cimg_library::CImg<T>(_w,_h,1,1,0) );
			}

		}

	}
	// Read flo files
	else
	{
		for( size_t i = 0; i < this->paths().size(); i++){

			if( this->paths()[i] != "" ){
			// Code from "FloReader.hxx"
			// u and v are save in C-axis
				try {
					const char *fn = this->paths()[i].c_str();
					FILE *stream = fopen(fn, "rb");

					if (!stream) {
						ParameteredObject::raise("Failed to open file!");
					}

					int width, height;
					float tag;
					size_t read;


					read = fread(&tag,    sizeof(float), 1, stream);
					if (read != 1) {
						throw std::runtime_error("error reading flo tag");
					}
					read = fread(&width,  sizeof(int),   1, stream);
					if (read != 1) {
						throw std::runtime_error("error reading flo width");
					}
					read = fread(&height, sizeof(int),   1, stream);
					if (read != 1) {
						throw std::runtime_error("error reading flo height");
					}

					int nBands = 2;
					cimg_library::CImg<T> o(width, height, 1, 2);
					for (int y = 0; y < height; y++) {
						for (int x = 0; x < nBands*width; x++) {
							float fv;
							read = fread(&fv, sizeof(float), 1, stream);
							if (read != 1) {
								throw std::runtime_error("error reading flo values");
							}
							o.atXYZ(x/nBands, y, 1, x%nBands) = T(fv);

						}
					}
					this->images().push_back( o );
					fclose(stream);
				}
				catch (const cimg_library::CImgException& err) {
					ParameteredObject::raise(
								std::string("Could not read file\n\t") + err.what());
				}

			}
			// no file given with current id
			else{
				this->images().push_back( cimg_library::CImg<T>(_w,_h,2,1,0) );
			}

		}

	}

}

#endif /* _CIMGLISTFROMPATH_HXX_ */
