/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file calibrationMatrices.hxx
 *  Implementation of the parameter class calibrationMatrices.
 *  \author Moritz Becker
 *  \date 14.01.2011
 */

#ifndef _CALIBRATIONMATRICES_HXX_
#define _CALIBRATIONMATRICES_HXX_

#include <hekate/CalibrationMatrices.h>

template <typename T>
calibrationMatrices<T>::calibrationMatrices(const std::string& name) :
		TemplatedParameteredObject<T>(
				"calibrationMatrices", name,
				"Create two calibration matrices saved in a CImgList. Those calibration "
				"matrices can used to reconstruct scenes where the calibration of used "
				"cameras is known but also to estimate calibrations by the image size "
				"of an optional related image."
		)
			
		,calibration1()
		,calibration2()
		,image(true,false)
	
{	

	ParameteredObject::_addParameter (calibration1, "calibration1", 
	"Calibration parameters of camera 1." );
	ParameteredObject::_addParameter (calibration2, "calibration2", 
	"Calibration parameters of camera 2." );
	
	ParameteredObject::_addInputSlot(image, "image", "Optional image from which the calibration is estimated from.", 
	"CImgList<T>"); 

	ParameteredObject::_addOutputSlot(calibration, "calibration", 
	"Output calibration matrices.", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(invertedCalibration, "invertedCalibration", 
	"Inverted output calibration matrices.", 
	"CImgList<T>");

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void calibrationMatrices<T>::execute() {

	//estimate calibration from images
	if(this->image.connected()){

		double iwidth = this->image()[0].width();
		double iheight = this->image()[0].height();

		// fill estimated parameters from image to calibration matrix
		cimg_library::CImg<T> cal1(3,3);
		cal1.fill(0);
		cal1(0,0) = iwidth + iheight;
		cal1(2,0) = iwidth / 2. ;
		cal1(1,1) = iwidth + iheight;
		cal1(2,1) = iheight / 2. ;
		cal1(2,2) = 1.;

		// push back the estimated calibration matrix to output vectors
		this->calibration().push_back(cal1);
		this->invertedCalibration().push_back(cal1.get_invert());

		//if second image available
		if(this->image().size() == 2){

			double iwidth2 = this->image()[1].width();
			double iheight2 = this->image()[1].height();

			cal1.fill(0);
			cal1(0,0) = iwidth2 + iheight2;
			cal1(2,0) = iwidth2 / 2. ;
			cal1(1,1) = iwidth2 + iheight2;
			cal1(2,1) = iheight2 / 2. ;
			cal1(2,2) = 1. ;

		}

		// push back the second calibration matrix to output vectors
		this->calibration().push_back(cal1);
		this->invertedCalibration().push_back(cal1.get_invert());

	}

	//take parameter input of plugin as calibration
	else{
		if(this->calibration1().size() == 9){
			// read out parameter list
			cimg_library::CImg<T> cal1(3,3);
			cal1(0,0)=this->calibration1()[0];
			cal1(1,0)=this->calibration1()[1];
			cal1(2,0)=this->calibration1()[2];
			cal1(0,1)=this->calibration1()[3];
			cal1(1,1)=this->calibration1()[4];
			cal1(2,1)=this->calibration1()[5];
			cal1(0,2)=this->calibration1()[6];
			cal1(1,2)=this->calibration1()[7];
			cal1(2,2)=this->calibration1()[8];

			// push back the calibration matrix to output vectors
			this->calibration().push_back(cal1);
			this->invertedCalibration().push_back(cal1.get_invert());

			// If not valid parameters for the second calibration matrix
			// are given, take the first calibration matrix twice.
			if(this->calibration2().size() != 9){
				sout << "No valid parameters for second calibration. Taking parameters of calibration 1." << std::endl;
				this->calibration().push_back(cal1);
				this->invertedCalibration().push_back(cal1.get_invert());
			}
		}
		if(this->calibration2().size() == 9){
			// read out parameter list
			cimg_library::CImg<T> cal2(3,3);
			cal2(0,0)=this->calibration2()[0];
			cal2(1,0)=this->calibration2()[1];
			cal2(2,0)=this->calibration2()[2];
			cal2(0,1)=this->calibration2()[3];
			cal2(1,1)=this->calibration2()[4];
			cal2(2,1)=this->calibration2()[5];
			cal2(0,2)=this->calibration2()[6];
			cal2(1,2)=this->calibration2()[7];
			cal2(2,2)=this->calibration2()[8];

			// push back the calibration matrix to output vectors
			this->calibration().push_back(cal2);
			this->invertedCalibration().push_back(cal2.get_invert());
		
			// If not valid parameters for the first calibration matrix
			// are given, take the second calibration matrix twice.
			if(this->calibration1().size() != 9){
				sout << "No valid parameters of first calibration. Taking parameters of calibration 2." << std::endl;
				this->calibration().push_back(cal2);
				this->invertedCalibration().push_back(cal2.get_invert());
			}
		}
		// If no calibration parameters are given,
		// write an attention message to the log.
		else if(this->calibration1().size() != 9 &&
			this->calibration2().size() == 9){
			sout << "ATTENTION: No valid calibration parameters given!" << std::endl;
		}
	}

}

#endif /* _CALIBRATIONMATRICES_HXX_ */



