/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CalibrationMatrixGenerator.hxx
 *  Implementation of the parameter class CalibrationMatrixGenerator.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 15.01.2014
 */

#ifndef _CALIBRATIONMATRIXGENERATOR_HXX_
#define _CALIBRATIONMATRIXGENERATOR_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CalibrationMatrixGenerator.h"

template <typename T>
CalibrationMatrixGenerator<T>::CalibrationMatrixGenerator(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CalibrationMatrixGenerator", name,
			"<h2>Generates a calibration matrix</h2><br>"
			"Generates a calibration matrix using given focal length and "
			"pricipal point. The focal length can be given in pixel unit or "
			"in milimeters. For the latter one, you also need to know the "
			"sensor size in milimeters which can be checked on "
			"http://www.dpreview.com/glossary/camera-system/sensor-sizes "
			"and the image width."
		)
{
	ParameteredObject::_addOutputSlot(
		calibration, "calibration",
		"Calibration matrix",
		"CImgList<T>");

	ParameteredObject::_addParameter< float >(
		f, "focalLengthPixels",
		"Focal length in pixel units. If you do not know this, use the entries focalLengthMM, "
		"sensorWidthMM and imageWidth.",
		3000, "float");
	ParameteredObject::_addParameter< int >(
		px, "px",
		"Principal point X coordinate. The principle point can be assumed as"
		"the middle of the image",
		960, "int");
	ParameteredObject::_addParameter< int >(
		py, "py",
		"Principal point Y coordinate. The principle point can be assumed as"
		"the middle of the image",
		540, "int");
	ParameteredObject::_addParameter< bool >(
		useMilimeterValues, "useMilimeterValues",
		"Use the given milimeter values (imageWidth, focalLengthMM, sensorWidthMM) "
		"to determine the focal length",
		false, "bool");
	ParameteredObject::_addParameter< int >(
		imageWidth, "imageWidth",
		"width of the image",
		1920, "int");
	ParameteredObject::_addParameter< int >(
		focalLengthMM, "focalLengthMM",
		"focal length in milimeters",
		50, "int");
	ParameteredObject::_addParameter< float >(
		sensorWidthMM, "sensorWidthMM",
		"Sensor width in milimeters",
		23, "float");

	

}

template <typename T>
void CalibrationMatrixGenerator<T>::execute() {

	cimg_library::CImg<float> cali(3,3,1,1,0);

	float fl = this->f();

	// determine focal length with given milimeter values
	if( this->useMilimeterValues() ){
		fl = this->focalLengthMM() * this->imageWidth() / this->sensorWidthMM() ;
	}
	
	// focal length
	cali(0,0) = fl;
	cali(1,1) = fl;

	// principal point
	cali(2,0) = this->px();
	cali(2,1) = this->py();

	cali(2,2) = 1.0;
	
	this->calibration() = cali;

}

#endif /* _CALIBRATIONMATRIXGENERATOR_HXX_ */
