
/*  Copyright (C) 2009 SasiKanth

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file cameramatrixgenerator.hxx
 *  Implementation of the parameter class cameramatrixgenerator.
 *  \author SasiKanth
 *  \date 02.06.2010
 */

#ifndef _CAMERAMATRIXGENERATOR_HXX_
#define _CAMERAMATRIXGENERATOR_HXX_

#include <hekate/CameraMatrixGenerator.h>
#include <charon-utils/ImgTool.hxx>

template <typename T>
cameramatrixgenerator<T>::cameramatrixgenerator(const std::string& name) :
	TemplatedParameteredObject<T>(
		"cameramatrixgenerator", name,
		"This plugin generates a set of camera matrices. "
		"The cameras are projective, if only a fundamental"
		"matrix without calibration is known or metric, if an "
		"essential matrix (fundamental matrix with "
		"calibration) is known."
	)
		,calibration(true, false)
		,f_matrix(false, false)
{
	
	ParameteredObject::_addParameter (
		useEssentialMatrix, "useEssentialMatrix",
		"Flag to use the input matrix as essential matrix for metric cameras.",
		true);
	ParameteredObject::_addParameter (
		transformCam, "transformCam",
		"Try to transform subsequent cams to same coordinate system. If "
		"this is activated, the 'useReferenceCam' flag in triangulation plugin "
		"has to be deactivated.",
		false);
	ParameteredObject::_addInputSlot(f_matrix, "f_matrix",
		"The fundamental matrix.",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(p_matrices, "p_matrices",
		"The two output camera matrices. ",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(correspondences, "correspondences",
		"Correspondences to determine depth of points to check which of the calculated cameras is the correct one.",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(calibration, "calibration", "Calibration matrix which is necessary for metric cameras.", "CImgList<T>");

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void cameramatrixgenerator<T>::execute() {

	//const std::vector<std::vector<std::vector<T> > > corrs = this->correspondences().getVectorForm();


	sout << std::endl << "---------------------- Camera Matrices--------------------------" << std::endl;

	// Generate general projective cameras (without calibration).
	if(!this->useEssentialMatrix()){

		cimg_library::CImg<T> P1(4,3), P2(4,3), fmatrix(3,3);
		fmatrix = this->f_matrix()[0];

		// Generate camera at origin as first camera.
		P1.fill(0.);
		P1(0,0) = 1;
		P1(1,1) = 1;
		P1(2,2) = 1;
	/*
		sout << "Calculating projective camera matrices taking fundamental matrix." << std::endl;

		bool positiveDepth = false;
		bool secondRun = false;

		// rerun code until depth of points is positive for both cameras
		while(!positiveDepth){

			cimg_library::CImg<T> U, S, V, U2, S2, V2, e_vec(1,3), e_mat(3,3), R(3,3);

			// Solving (F.transpose) * e' = 0 for vector e'
			fmatrix.get_transpose().SVD(U,S,V);
			fmatrix.SVD(U2,S2,V2);

			//Assigning the values from the found V matrix
			e_vec = V.crop(2,0,2,2);
			ImgTool::printInfo(sout,e_vec,"epipole at: ");
			ImgTool::printInfo(sout,U2 ,"epipole at: ");

			//normalize e_vec to max(e_vec) = 1
			double maxE = e_vec[0];
			if(std::abs(maxE)<std::abs(e_vec[1])) maxE=e_vec[1];
			if(std::abs(maxE)<std::abs(e_vec[2])) maxE=e_vec[2];
			e_vec /= maxE;

			ImgTool::printInfo(sout,e_vec,"epipole at: ");

			//Generating the e_matrix from the e_vector
			e_mat(0,0) = 0;				e_mat(1,0) = -e_vec(0,2);	e_mat(2,0) = e_vec(0,1);
			e_mat(0,1) = e_vec(0,2);	e_mat(1,1) = 0;				e_mat(2,1) = -e_vec(0,0);
			e_mat(0,2) = -e_vec(0,1);	e_mat(1,2) = e_vec(0,0);	e_mat(2,2) = 0;

			//Generating the rotation matrix
			sout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << std::endl;

			R = e_mat.get_transpose()*fmatrix;

			//Generating the P matrices according to Hartley Result 9.14, pg. 256
			cimg_library::CImg<T> R_corrected(3,3), vec(1,3);

			R_corrected = e_vec * e_vec.get_transpose();

			//correct R so R has rank 3 (rank(P)=3 not sufficient)
			R += R_corrected;

			//Loading the values of P2
			for(int i=0; i<3; i++)
			{
				for(int j=0; j<3; j++)
				{
					P2(i,j) = R(i,j);
				}
			}

			P2(3,0) = e_vec(0,0);	P2(3,1) = e_vec(0,1);	P2(3,2) = e_vec(0,2);

			double maxP2=P2[0];
			for(int i=1; i<12; i++){
				if(std::abs(maxP2)<std::abs(P2[i])) maxP2=P2[i];
			}
			sout << "max P2: " << maxP2 << std::endl;
			ImgTool::printInfo(sout,P2,"P2");

			//determine depths
			int numberOfOutliers=0;
			bool depthDirection = true;

			if(corrs[0].size() < 8)
			{	throw std::runtime_error("Not enough correspondences, must be >= 8") ;	}
			for(int i=0; i<8; i++){

				cimg_library::CImg<T> pt1(1,3), pt2(1,3);
				pt1(0,0) = corrs[0][i][0];
				pt1(0,1) = corrs[0][i][1];
				pt1(0,2) = corrs[0][i][2];
				pt2(0,0) = corrs[1][i][0];
				pt2(0,1) = corrs[1][i][1];
				pt2(0,2) = corrs[1][i][2];

				// triangulate points
				cimg_library::CImg<T> X(1,4);
				X = getTriangulation(P1,P2,pt1,pt2);

				// calculate depth between X and principle plane
				// of P1 / P2
				double depth1 = getDepth(P1, X);
				double depth2 = getDepth(P2, X);

				if((depth1>=0 && depth2<0) || (depth1<0 && depth2>=0)) numberOfOutliers += 1;
				else (depth1 >= 0 ? depthDirection = true : depthDirection = false);

			}

			if(numberOfOutliers==0){
				positiveDepth = true;
				sout << "PMatrix: Depth of 3D points in front of cameras." << std::endl;
				//if(depthDirection) P2(3,2) += 10;
				//else P2(3,2) -= 10;
			}
			else if(secondRun){
				positiveDepth = true;
				sout << "Couldn't determine positive depths!" << std::endl;
			}
			else{
				numberOfOutliers = 0;
				sout << "PMatrix: recalculating P2 with f=-f -------------------" << std::endl << std::endl;
				fmatrix *= -1;
				secondRun = true;
			}

		}

		p_matrices().push_back(P1);
		p_matrices().push_back(P2);

		ImgTool::printInfo(sout,P2,"");
		*/
	}


	// Generate metric cameras (with calibration).
	// (see Hartley p.258)
	else{

		sout << "Calculating camera matrices by essential matrix." << std::endl;

		// check if calibration matrices are connected
		if(!this->calibration.connected()){
			sout << "No calibration connected!" << std::endl;
			throw std::runtime_error("No calibration matrix connected. Cannot create essential matrix.");
		}

		// read in calibration matrices
		const cimg_library::CImg<T> cal1 = this->calibration()[0];
		
		// Generate camera at origin as first camera.
		cimg_library::CImg<T> P1(4,3);
		P1.fill(0.);
		P1(0,0) = 1;
		P1(1,1) = 1;
		P1(2,2) = 1;

		// Iterate for every fundamental matrix
		for(size_t i=0; i<this->f_matrix().size(); i++){
		
			cimg_library::CImg<T> P2(4,3), fmatrix(3,3);

			// Convert fmatrix to essential matrix
			fmatrix = this->f_matrix()[i];
			//ImgTool::printInfo(sout,fmatrix,"fmatrix");
			fmatrix = cal1.get_transpose()*fmatrix*cal1;
			//ImgTool::printInfo(sout,fmatrix,"cal1^T*fmatrix*cal1");

			// Calculated SVD of the fundamental matrix.
			cimg_library::CImg<T> U, S, V, W(3,3);
			fmatrix.SVD(U,S,V);

			W.fill(T(0.));
			W(1,0) = (T)-1;
			W(0,1) = (T)1;
			W(2,2) = (T)1;

			// Print SVD results to log.
			//ImgTool::printInfo(sout,U,"U");
			//ImgTool::printInfo(sout,S,"S");
			//ImgTool::printInfo(sout,V,"V");
			//ImgTool::printInfo(sout,fmatrix,"ematrix");
			//ImgTool::printInfo(sout,U*S.diagonal()*V.get_transpose(),"Udiag(1,1,0)V^T");
			//ImgTool::printInfo(sout,W,"W");

			// Declare possible cameras
			cimg_library::CImg<T> P2_1, P2_2, P2_3, P2_4, u(1,3);
			u=U.get_crop(2,0,2,2);

			// Possible cam 1
			P2_1 = U*W*V.get_transpose();
			P2_1.append(u);

			// Possible cam 2
			P2_2 = U*W*V.get_transpose();
			P2_2.append(-u);

			// Possible cam 3
			P2_3 = U*W.get_transpose()*V.get_transpose();
			P2_3.append(u);

			// Possible cam 4
			P2_4 = U*W.get_transpose()*V.get_transpose();
			P2_4.append(-u);

			// Select camera where point correspondences are reconstructed
			// with a positive depth in fron of the image plane. Only
			// one camera (of the above four) will satisfiy this constraint.
			cimg_library::CImg<T> pt1(1,3), pt2(1,3), X1, X2, X3, X4;

			const CorrespondenceList<T> &list = correspondences();

			// Get correspondence inlier to check cameras
			auto it3 = list.get<imgIndex>().lower_bound(i) ;
			auto it3End = list.get<imgIndex>().upper_bound(i) ;
			for(int j=0; it3 != it3End ; it3++, j++){
				try{
					const Correspondence<T> &pt2Corr = list(i+1, it3->pointIndex);
					if(!it3->outlier && !pt2Corr.outlier)
					{
						//sout << "Point index: " << it3->pointIndex << std::endl;
						//sout << *it3 << std::endl;
						//sout << pt2Corr << std::endl;
						pt1(0,0) = (*it3).x();
						pt1(0,1) = (*it3).y();
						pt1(0,2) = (*it3).w();
						pt2(0,0) = pt2Corr.x();
						pt2(0,1) = pt2Corr.y();
						pt2(0,2) = pt2Corr.w();
						break;
					}
				}
				catch(...){}
			}
			// give out selected corrspondence
			//ImgTool::printInfo(sout,pt1,"Selected pt1 for depth estimation");
			//ImgTool::printInfo(sout,pt2,"Selected pt2 for depth estimation");

			pt1 = cal1.get_invert() * pt1;
			pt2 = cal1.get_invert() * pt2;

			ImgTool::printInfo(sout,pt2.get_transpose()*fmatrix*pt1,"e2^TEe1==0?");

			// triangulate points with camera 1 to receive depth
			X1 = getTriangulation(P1, P2_1, pt1, pt2);
			//sout << "Case 1: " << std::endl;
			//sout << "Depth 1: " << getDepth(P1, X1) << std::endl;
			//sout << "Depth 1: " << getDepth(P2_1, X1) << std::endl;

			// triangulate points with camera 2 to receive depth
			X2 = getTriangulation(P1, P2_2, pt1, pt2);
			//sout << "Case 2: " << std::endl;
			//sout << "Depth 2: " << getDepth(P1, X2) << std::endl;
			//sout << "Depth 2: " << getDepth(P2_2, X2) << std::endl;

			// triangulate points with camera 3 to receive depth
			X3 = getTriangulation(P1, P2_3, pt1, pt2);
			//sout << "Case 3: " << std::endl;
			//sout << "Depth 3: " << getDepth(P1, X3) << std::endl;
			//sout << "Depth 3: " << getDepth(P2_3, X3) << std::endl;

			// triangulate points with camera 4 to receive depth
			X4 = getTriangulation(P1, P2_4, pt1, pt2);
			//sout << "Case 4: " << std::endl;
			//sout << "Depth 4: " << getDepth(P1, X4) << std::endl;
			//sout << "Depth 4: " << getDepth(P2_4, X4) << std::endl;
			
			cimg_library::CImg<T> Pselected;
			// Select correct camera.
			if(getDepth(P1, X1) > 0. && getDepth(P2_1, X1) > 0.){
				sout << "Taking case 1: P2=[UWV^T|u3]" << std::endl;
				Pselected = P2_1;
			}
			else if(getDepth(P1, X2) > 0. && getDepth(P2_2, X2) > 0.){
				sout << "Taking case 2: P2=[UWV^T|-u3]" << std::endl;
				Pselected = P2_2;
			}
			else if(getDepth(P1, X3) > 0. && getDepth(P2_3, X3) > 0.){
				sout << "Taking case 3: P2=[UW^TV^T|u3]" << std::endl;
				Pselected = P2_3;
			}
			else if(getDepth(P1, X4) > 0. && getDepth(P2_4, X4) > 0.){
				sout << "Taking case 4: P2=[UW^TV^T|-u3]" << std::endl;
				Pselected = P2_4;
			}
			else{
				throw std::runtime_error("It could not be found a camera which reconstructs positive depths");
			}
			
			// push back unit camera (at origin) to the output vector
			if(i==0)
			{
				p_matrices().push_back(P1);
				p_matrices().push_back(Pselected);
				ImgTool::printInfo(sout,P1,"P_0");
				ImgTool::printInfo(sout,Pselected,"P_i");
			}
			// Calculate correct camera position using last iteration cam
			// as reference (and not the unique camera matrix)
			else
			{
				if(this->transformCam())
				{
					sout << "Transforming camera using last iteration" << std::endl;
					cimg_library::CImg<T> _M, _t, _M2, _t2, H1, H2, Cam, v(4,1);
					v.fill(T(0)); 
					v[3] = T(1);
					H1 = p_matrices()[i] ;
					H1.append(v,'y');
					H2 = Pselected;
					H2.append(v,'y');
					ImgTool::printInfo(sout,p_matrices()[i],"P_last");
					ImgTool::printInfo(sout,Pselected,"	P_new");
					ImgTool::printInfo(sout,H1,"	H1");
					ImgTool::printInfo(sout,H2,"	H2");
					
					Cam = (H1*H2).get_crop(0,0,3,2);
					ImgTool::printInfo(sout,Cam,"	Cam");
					p_matrices().push_back(Cam);

					_M = p_matrices()[i].get_crop(0,0,2,2);
					_t = p_matrices()[i].get_crop(3,0,3,2);
					_M2 = _M * Pselected.get_crop(0,0,2,2) ;
					//sout << "[R2|t2] = [R1*R2'|-R1*t2+t1]" << std::endl;
					_t2 = _M * Pselected.get_crop(3,0,3,2) + _t;
					_M2.append(_t2,'x');
					//p_matrices().push_back(_M2);
					ImgTool::printInfo(sout,_M2,"Cam other method");
				
					ImgTool::printInfo(sout,Cam.get_crop(0,0,2,2)*Cam.get_crop(0,0,2,2).transpose(),"	RR^T == 1?");
				}
				else
					p_matrices().push_back(Pselected);
			}


			//sout << "Denormalizing P2 -> K2 * P2" << std::endl;
			//sout << "Normalized camera P'=[R|t]:" << std::endl;
			//ImgTool::printInfo(sout,(p_matrices()[1]),"P'");

			// Multipy camera matrix so determinant is > 0
			//if(p_matrices()[i+1].get_crop(0,0,2,2).det() < 0)
			//	p_matrices()[i+1] = p_matrices()[i+1] * -1;
			//sout << "DET: " << p_matrices()[1].get_crop(0,0,2,2).det() << std::endl;

			sout << "Principal point:" << std::endl;
			ImgTool::printInfo(sout,(p_matrices()[i].get_crop(0,0,2,2)*p_matrices()[i].get_crop(0,2,2,2).transpose()),"principal point");

			/*
			sout << "Denormalized cameras P=[K1|0] P'=K1[R|t]:" << std::endl;
			ImgTool::printInfo(sout,(p_matrices()[0]),"P");
			ImgTool::printInfo(sout,(p_matrices()[1]),"P'");

			ImgTool::printInfo(sout,pt2.get_transpose()*fmatrix*pt1,"x2Fx1");
			*/
			/*
			for(size_t j=0; j<corrs[0].size(); j++){
				cimg_library::CImg<T> vec(1,3), vec2(1,3);
				vec[0]=corrs[0][j][0];
				vec[1]=corrs[0][j][1];
				vec[2]=corrs[0][j][2];
				vec2[0]=corrs[1][j][0];
				vec2[1]=corrs[1][j][1];
				vec2[2]=corrs[1][j][2];
				vec = cal1.get_invert() * vec;
				vec2 = cal2.get_invert() * vec2;
				//sout << "pt" << j << ": " << (vec2.get_transpose()*fmatrix*vec)[0] << std::endl;
			}
			*/
			//sout << "ATTENTION: If using an essential matrix based on a rough calibration, in the following algorithms P sometimes is supposed to be P=[1|0]. Be aware that P for now is P=K[1|0]!" << std::endl;

		}
		
		// denormalize P_cam -> P = K * P_cam and P' = K * P'_cam
		// since the determined projection matrix has the form [R|t]
		for(size_t i=0; i<p_matrices().size(); i++)
		{
			p_matrices()[i] = cal1 * p_matrices()[i];
			ImgTool::printInfo(sout,(p_matrices()[i]),"\tP");
		}
		

		//ImgTool::printInfo(sout,cal1,"calibrationMatrix");
		//ImgTool::printInfo(sout,cal2,"calibrationMatrix2");

	}
}

/// Triangulate 2D points pt1 and pt2 with cameras P1 and P2 to retrieve the 3D point.
template <typename T>
cimg_library::CImg<T> cameramatrixgenerator<T>::getTriangulation(
	cimg_library::CImg<T> P1, cimg_library::CImg<T> P2,
	cimg_library::CImg<T> pt1, cimg_library::CImg<T> pt2){


	// define matrix A = x cross X and x' cross X
	cimg_library::CImg<T> A(4,4);
	//first row of A
	A(0,0) = pt1(0,0)*P1(0,2) - P1(0,0);
	A(1,0) = pt1(0,0)*P1(1,2) - P1(1,0);
	A(2,0) = pt1(0,0)*P1(2,2) - P1(2,0);
	A(3,0) = pt1(0,0)*P1(3,2) - P1(3,0);
	//second row of A
	A(0,1) = pt1(0,1)*P1(0,2) - P1(0,1);
	A(1,1) = pt1(0,1)*P1(1,2) - P1(1,1);
	A(2,1) = pt1(0,1)*P1(2,2) - P1(2,1);
	A(3,1) = pt1(0,1)*P1(3,2) - P1(3,1);
	//third row of A
	A(0,2) = pt2(0,0)*P2(0,2) - P2(0,0);
	A(1,2) = pt2(0,0)*P2(1,2) - P2(1,0);
	A(2,2) = pt2(0,0)*P2(2,2) - P2(2,0);
	A(3,2) = pt2(0,0)*P2(3,2) - P2(3,0);
	//fourth row of A
	A(0,3) = pt2(0,1)*P2(0,2) - P2(0,1);
	A(1,3) = pt2(0,1)*P2(1,2) - P2(1,1);
	A(2,3) = pt2(0,1)*P2(2,2) - P2(2,1);
	A(3,3) = pt2(0,1)*P2(3,2) - P2(3,1);


	// determine 3D point via SVD of A (Hartley p.592)
	cimg_library::CImg<T> X(1,4), U, S, V;
	A.SVD(U, S, V);

	X = V.get_crop(3,0,3,3);
	X /= X(0,3);

	return X;

}

/// Calculate depth between x and principle plane of P.
template <typename T>
double cameramatrixgenerator<T>::getDepth(const cimg_library::CImg<T> P,
		const cimg_library::CImg<T> X) {
	int signP;
	(P.get_crop(0,0,2,2).det() < 0 ? signP = -1 : signP = 1);

	double P_m3_abs = P(0,2)*P(0,2)+P(1,2)*P(1,2)+P(2,2)*P(2,2);
	P_m3_abs = pow(P_m3_abs,.5);

	double depth;
	depth = signP * (P*X)(0,2);
	if(P_m3_abs!=0) depth /= P_m3_abs;

	return depth;

}



#endif /* _CAMERAMATRIXGENERATOR_HXX_ */
