/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CannyEdges.hxx
 *  Implementation of the parameter class CannyEdges.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 17.09.2012
 */

#ifndef _CANNYEDGES_HXX_
#define _CANNYEDGES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CannyEdges.h"
#include "CannyByProbability.h"

template <typename T>
CannyEdges<T>::CannyEdges(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CannyEdges", name,
			"<h2>Detects edges by Canny algorithm</h2><br>"
			"Detects edges in an image using the canny edge detector. "
			"The sensitivity can be adjusted by an upper and a lower threshold, defining the edge strenght detected."
		), 
		probability( true, false )
{
	ParameteredObject::_addInputSlot(
		inputImages, "inputImages",
		"List of input images",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		probability, "probability",
		"Input path of optional probability images. Probabilities adjust thresholds pixel wise. A edge is detected, if one pixel "
		"satisfies an upper threshold. The edge is tracked until a lower threshold is not satisfied anymore.",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		edgeImages, "edgeImages",
		"Edge images",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		threshold1, "threshold1",
		"First threshold for the hysteresis procedure. It defines the threshold to find putative edges.",
		200, "int");
	ParameteredObject::_addParameter< int >(
		threshold2, "threshold2",
		"Second threshold for the hysteresis procedure. "
		"It defines the lower threshold (pixelwise) which an edge has to satisfiy, to continue detection.",
		200, "int");
	ParameteredObject::_addParameter< int >(
		apertureSize, "apertureSize",
		"Aperture size for the Sobel() operator.",
		3, "int");
	ParameteredObject::_addParameter< bool >(
		l2gradient, "l2gradient",
		"Flag indicating whether a more accurate L_2 norm =\sqrt{(dI/dx)^2 +"
		"(dI/dy)^2} should be used to compute the image gradient magnitude ("
		"L2gradient=true ), or a faster default L_1 norm =|dI/dx|+|dI/dy| is"
		"enough ( L2gradient=false ).",
		0, "bool");
	
	ParameteredObject::_addParameter< bool >(H, "H", "Activate hue channel for edge detection (HSV).", true, "bool");	
	ParameteredObject::_addParameter< bool >(S, "S", "Activate saturation channel for edge detection (HSV).", true, "bool");	
	ParameteredObject::_addParameter< bool >(V, "V", "Activate brightness channel for edge detection (HSV).", true, "bool");	
	ParameteredObject::_addParameter< bool >(R, "R", "Activate red channel for edge detection (RGB)", true, "bool");	
	ParameteredObject::_addParameter< bool >(G, "G", "Activate green channel for edge detection (RGB)", true, "bool");	
	ParameteredObject::_addParameter< bool >(B, "B", "Activate blue channel for edge detection (RGB)", true, "bool");	
	
	ParameteredObject::_addParameter< bool >(discrete, "discrete", 
		"Sum different channels to discrete edge value. Otherwise the resulting edge images can "
		"obtain seven different pixel values depending in how many channels per pixel an edge was detected.", true, "bool");	

	ParameteredObject::_addParameter< float >(
		blur, "blur",
		"Blur before canny to avoid noise artefacts.",
		0, "float");
	
	ParameteredObject::_addParameter< float >(
		sigma1, "sigmaProbability1",
		"This factor sets the amount of the probability effectness on threshold 1, "
		"which is the threshold to start looking for an edge. Only affects edge detection, if a probability image is given.",
		0.5, "float");
	ParameteredObject::_addParameter< float >(
		sigma2, "sigmaProbability2",
		"This factor sets the amount of the probability effectness on threshold 2, "
		"which is the threshold to add pixels to an existing edge. Only affects edge detection, if a probability image is given.",
		0.5, "float");
	
	ParameteredObject::_addParameter< int >(
		inFrame, "inFrame",
		"Start frame",
		0, "int");

	ParameteredObject::_addParameter< int >(
		outFrame, "outFrame",
		"End frame",
		-1, "int");

	ParameteredObject::_setTags("Hekate;Edges") ;

}

template <typename T>
void CannyEdges<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	// Check, if canny edges shall be modified by probability image
	bool useProb = false;
	if( this->probability.connected() )
	{
		if( this->inputImages().size() != this->probability().size() )
			throw( std::range_error("CannyEdges<T>::execute() number of images != probabilities") );
		else
			useProb = true;
	}

	int nc = 0;
	if( H() ) nc++;
	if( S() ) nc++;
	if( V() ) nc++;
	if( R() ) nc++;
	if( G() ) nc++;
	if( B() ) nc++;
	nc *= 2;	// for normal and equalized image

	// Define sliding window ----------------------------------------
	
	const cimg_library::CImgList<T> &_input = this->inputImages();
	cimg_library::CImgList<T> &_output = this->edgeImages();
	
	const int _w = _input[0].width();
	const int _h = _input[0].height();
	const int _seq_length = _input.size();

	_output.clear();
	_output.assign( _seq_length, _w, _h, 1, 1, T(0) );

	int _id_start = 0 ;
	if( inFrame() > 0 && inFrame() < _seq_length )
		_id_start = inFrame();

	int _id_end = _seq_length -1 ;
	if( outFrame() >= _id_start && outFrame() < _seq_length )
		_id_end = outFrame();

	sout << "Calculating canny edges for frame " << _id_start << " - " << _id_end 
		<< " of " << _seq_length << " frames in total" << std::endl;


	// iterate over all images in list
	for(int i = _id_start; i <= _id_end; i++){
			
		CImg< double > edges_ = _input[i];
		edges_.normalize(0,255).blur( this->blur() );

		CImg< double > sum(_w,_h,1,1);
		
		for( size_t k = 0; k < 2; k++ ){

			if( k == 1 )
				 edges_.get_equalize(256);

			CImgList<float> channels(6);

			// iterate over all channels HSVRGB
			#pragma omp parallel for
			for( int j = 0; j < 6; j++ )
			{

				// read channel
				CImg<double> edges = edges_;

				if( j == 0 && H() )
				{
					edges.RGBtoHSV();
					edges.channel(0);
				}
				else if ( j == 1 && S() )
				{
					edges.RGBtoHSV();
					edges.channel(1);
				}
				else if ( j == 2 && V() )
				{
					edges.RGBtoHSV();
					edges.channel(2);
				}
				else if ( j == 3 && R() )
				{
					edges.channel(0);
				}
				else if ( j == 4 && G() )
				{
					edges.channel(1);
				}
				else if ( j == 5 && B() )
				{
					edges.channel(2);
				}
				else
					continue;

				Mat cvEdges, cvEdgesOut;
				cvEdges = CImg2Mat(edges.get_normalize(0,255), CV_8U) ;
		
				/*	
					Use normal Canny edge detector from open cv:
					Canny(	InputArray image, OutputArray edges, 
							double threshold1, double threshold2, 
							int apertureSize=3, bool L2gradient=false 
							)
				*/		
				if( !useProb || this->probability()[i].sum() == 0 )
				{
					std::string msg("\tCalculating edges in mask");
					msg += std::to_string(long double( i ));
					msg += "\n" ;
					sout << msg;

					Canny(cvEdges, cvEdgesOut, this->threshold1(), this->threshold2(), 
						this->apertureSize(), this->l2gradient() );
				}
				// Use modified open cv canny detector using probabilities
				else
				{
					std::string msg("\tCalculating edges by probability in mask");
					msg += std::to_string(long double( i ));
					msg += "\n" ;
					sout << msg;

					// read in probability
					CImg< float > prob = this->probability()[i];
					if( edges.width() != prob.width() || edges.height() != prob.height() )
						throw( std::runtime_error("CannyEdges<T>::execute(): Inconsistent image sizes") );
					prob.channel(0).slice(0).normalize(0,1);

					// check given sigmas
					float sigma1 = std::abs( this->sigma1() );
					float sigma2 = std::abs( this->sigma2() );
					if( sigma2 < sigma1)
					{
						float sigma_swap = sigma1;
						sigma1 = sigma2;
						sigma2 = sigma_swap;				
					}
					if( sigma2 > 1 )
					{
						sigma1 /= sigma2;
						sigma2 = 1;
					}

					// calculate canny by given probability
					CannyByProb(sigma1, sigma2, prob, cvEdges, cvEdgesOut, this->threshold1(), this->threshold2(), 
						this->apertureSize(), this->l2gradient() );
				}
		
				channels[j] = Mat2CImg<double>(cvEdgesOut) ;
				channels[j].quantize(2).normalize(0,255) ;

			}

			if( discrete() )
				for( size_t k = 0; k < 6; k++ )
					sum.max(channels[k]);
			else
				for( size_t k = 0; k < 6; k++ )
					sum += channels[k];

		}

		if( !discrete() )
			sum /= nc ;

		_output[i] = sum;

	}

}

#endif /* _CANNYEDGES_HXX_ */
