/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CompleteEdges.hxx
 *  Implementation of the parameter class CompleteEdges.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 06.04.2013
 */

#ifndef _COMPLETEEDGES_HXX_
#define _COMPLETEEDGES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CompleteEdges.h"

using namespace cimg_library;

template <typename T>
CompleteEdges<T>::CompleteEdges(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CompleteEdges", name,
			"<h2>This module takes an edge image and extends edges pixelwise</h2><br>"
			"Edges are extended pixelwise where pixels not on an edge are added to an edge, if enough neighbour "
			"pixels are already on an edge."
		)
{
	ParameteredObject::_addInputSlot(
		images, "images",
		"Edge images.",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		imagesOut, "imagesOut",
		"Corrected edge images",
		"CImgList<T>");

	ParameteredObject::_addParameter< bool >(
		removeSinglePixels, "removeSinglePixels",
		"Remove pixels without nearest neighbour edge pixels. In this way, single pixels are removed.", true);

	ParameteredObject::_addParameter< int >(
		numberOfNearestNeighbours, "numberOfNearestNeighbours",
		"The number of nearest neighbour edge pixels to define a none edge pixel as an"
		"edge pixel.",
		5, "int");

	ParameteredObject::_addParameter< int >(
		numberOfNearestNeighbours2, "numberOfNearestNeighbours2",
		"The number of nearest neighbours to define an edge pixel as a"
		"none edge pixel.",
		1, "int");

	ParameteredObject::_addParameter< int >(
		numberOfIterations, "numberOfIterations",
		"Number of iterations per image.",
		1, "int");

	ParameteredObject::_addParameter< float >(
		threshold, "threshold",
		"Threshold which defines an edge (for non discrete 2 value images)", 0.5);
	
	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void CompleteEdges<T>::execute() {

	this->imagesOut() = this->images();
	
	for( size_t i = 0; i < this->imagesOut().size(); i++ )
	{

		sout << "Process image " << i << std::endl;

		for( size_t j = 0; j < numberOfIterations; j++)
		{

			sout << "\titeration " << j ;
	
			int nc = 0;
			CImg<T> img = this->imagesOut()[i].get_normalize(0,1);
			img.threshold( this->threshold() );
			cimg_forXYC( img, x, y, c )
			{
				if( img(x,y,0,c) == 0 )
				{
					
					CImg<T> crop = img.get_crop(x-1,y-1,0,c,x+1,y+1,0,c);
					
					if( crop.sum() >= this->numberOfNearestNeighbours() )
					{

						nc++;

						int nnn = 0;
						float sum = 0;
						cimg_forXY(crop, x2, y2)
						{
							if( crop(x2,y2,0,c) > 0 )
							{
								nnn++;
								sum += this->imagesOut()[i](x-1+x2,y-1+y2,0,c);
							}
						}
						if(nnn != 0) sum /= nnn;
						this->imagesOut()[i](x,y,0,c) = sum;

					}

				}
				else if( this->removeSinglePixels() && img.get_crop(x-1,y-1,0,c,x+1,y+1,0,c).sum() <= this->numberOfNearestNeighbours2() )
				{
					nc++;
					this->imagesOut()[i](x,y,0,c) = 0;
				}

			}

			sout << ": corrected " << nc << " pixels" << std::endl;

		}
	
	}


}

#endif /* _COMPLETEEDGES_HXX_ */
