/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CompleteMask.hxx
 *  Implementation of the parameter class CompleteMask.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 06.09.2013
 */

#ifndef _COMPLETEMASK_HXX_
#define _COMPLETEMASK_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CompleteMask.h"
#include "CompleteMaskSegment.hxx"

template <typename T>
CompleteMask<T>::CompleteMask(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CompleteMask", name,
			"<h2>Repairs a broken mask where a consisten edge is not "
			"available</h2><br>"
			"Repairs a broken mask where a consisten edge is not available. "
			"A maximal distance which defines the maximal length of an "
			"repair edge has to be defined."
		)
{
	ParameteredObject::_addInputSlot(
		masks, "masks",
		"Mask images which are putatively broken.",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		masksCompleted, "masksCompleted",
		"The repaired masks",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		maxDistance, "maxDistance",
		"Maximal distance between two broken segments.",
		10, "int");

	ParameteredObject::_addParameter< int >(
		minSize, "minSize",
		"Minimal allowed number of pixels of a sub edge "
		"to be considered for completion.",
		10, "int");

	ParameteredObject::_addParameter< int >(
		maxSizeBlack, "maxSizeBlackHoles",
		"Maximal allowed number of pixels of black hole "
		"segments to be removed.",
		20, "int");

}

template <typename T>
void CompleteMask<T>::execute() {

	sout << "Processing " << this->masks().size() << " edge images" << std::endl;

	this->masksCompleted().assign( this->masks().size() );
	
	for( size_t index = 0; index < this->masks().size(); index++ )
	{

		sout << "Processing frame " << index << std::endl;

		const CImg<int> &mask	= this->masks()[index];
		const unsigned int w	= mask.width();
		const unsigned int h	= mask.height();
		
		// Get min/max value of pixels
		const unsigned int max = mask.max();

		// Get histogram
		// starting from 1 skips the black background (?)
		CImg<int> hist = mask.get_histogram( max+1 );
		const unsigned int n_levels = hist.width();
		
		CImgList<int> mask_list(n_levels);
		
		//sout << "Found " << max << " different edge labels" << std::endl;
		//sout << "Found " << n_levels << " different edge labels" << std::endl;

		/* 
			All pixels which belong to a specific color/grey value will be handled as pixels
			for one and the same edge type (e.g. depth edges).
			The next loop iterates over these different edge types and processes the
			edge type pixels seperately.
			(1) First, connection labels are build where all pixels connected to a cluster are 
			uniquely labeled.
			(2) Afterwards, black holes (small black clusters) are removed and filled in 
			the mask image, if they are smaller	than the maximum allowed cluster size for 
			black holes. Also clusters smaller than the minimum allowed size are removed for 
			further processing. In this way, noisy pixels/clusters are handled. 
			Both parameters (minimum,maximum cluster size) are set in the plugin parameters.
			(3) Segments are reassigned according to modified mask image
			(4) Black background segments are removed
			(5) Nearest neighbours between segments are searched
		*/

		// Iteration over every mask
		unsigned int n_masks = 0;
		#pragma omp parallel for
		for( int i = 2; i < n_levels; i++ )
		{

			// if there is no mask with value i, continue
			if( hist(i,0) == 0 )
				continue;

			sout << "Completing mask " << i << " (" << hist(i,0) << "pixels)" << std::endl;

			n_masks++;	// increase number of masks

			// Get mask with value i
			CImg<int> mask_i(w, h, 1, 1, 0);
			cimg_forXY( mask_i, x, y )
			{
				if( mask(x,y) == i )
					mask_i(x,y) = 1 ; // DEBUG i or 1 ?
			}

			// Add pixels with 6-pixel-neighbourhood to segment
			// (Pixels with 6 or more neighbours are surounded pixels
			// which can be set to 1 to close such holes)
			cimg_forXY( mask_i, x, y )
			{
				if( mask_i(x,y) == 0 )
					if( mask_i.get_crop(x-1,y-1,x+1,y+1).sum() >= 7 )
						mask_i(x,y) = 1 ;
			}
			
			// Painting color
			const unsigned char color[] = { 1 } ;

			//sout << "Get labels" << std::endl;

			// (1)
			/* 
			Get labels of this mask
			The labels CImg contains pixels whose values
			are the label ids
			*/
			CImg<int> labels		= mask_i.get_label(true);
			unsigned int n_labels	= labels.max() + 1 ;

			sout << "Generate " << n_labels << " segments" << std::endl;

			std::vector< hekate::CompleteMaskSegment > segments;
			getSegments(segments, labels, n_labels );
			
			// (2)
			/* 
			Remove black segments smaller than min value and fill pixels with 1, 
			remove segments smaller than min value and fill pixels with 0,
			and remove black segments at all 
			*/
			removeSmallClusters(mask_i, segments, n_labels, this->maxSizeBlack(), this->minSize() );

			// (3) Get new labels according to modified mask_i
			labels		= mask_i.get_label(true);
			n_labels	= labels.max() + 1 ;
			getSegments(segments, labels, n_labels );

			// (4) Remove segments representing black background
			removeBackgroundClusters(mask_i, segments, n_labels, this->maxSizeBlack(), this->minSize() );
			
			CImg<int> initial_labels = labels;
			CImg<int> final_mask = mask_i;
						
			// (5) Iterate over segments/labels until completion is done
			unsigned int it				= 0;
			const unsigned int max_it	= 2 * segments.size();
			//while( 1 ) //segments.size() > 1 && it < max_it )
			//{
			//	it++; // iteration number for break condition

				//sout << "n_labels = " << n_labels << std::endl;

				// iterate through current segment vector
				for( unsigned int i_seg = 0; i_seg < segments.size(); i_seg++ )
				{

					bool nnn = false; // no nearest neighbours

					while(!nnn)
					{
					
						unsigned int id1, id2;	// id of nearest neighbour segment
						unsigned int x1, y1, x2, y2;// coordinates of distance vector

						// Get minimum distance
						//sout << "Get distance" << std::endl;
						bool has_min_dist = hekate::getDistance( segments, i_seg, this->minSize(), 
											this->maxDistance(), id2, x1, y1, x2, y2, mask_i, labels );

						// If minimum distance found, merge segments and start again					
						if( has_min_dist )	
						{
							
							// Get id1 belonging to segment 1
							hekate::CompleteMaskSegment &seg1 = segments[ i_seg ];
							id1 = seg1.id;

							// Get segment 2 belonging to id2
							unsigned int i_seg2 = 0;
							bool has_segment = false;
							for( size_t i_id2 = 0; i_id2 < segments.size(); i_id2++ )
							{
								if( segments[i_id2].id == id2 ){
									i_seg2 = i_id2;
									has_segment = true;
									break;
								}
							}
							if( !has_segment ){
								sout << "Segment with label " << id2 << " is not in vector anymore!" << std::endl;
								/*sout << "Segment1 label = " << id1 << std::endl;
								sout << "Segments.size = " << segments.size() << std::endl;
								sout << "\tx1=" << x1 << ",y1=" << y1 << "/x2=" << x2 << ",y2=" << y2 << std::endl;
								sout << "\tmask_i(x1,y1)=" << mask_i(x1,y1) << std::endl;
								sout << "\tlabels(x1,y1)=" << labels(x1,y1) << std::endl;
								sout << "\tmask_i(x2,y2)=" << mask_i(x2,y2) << std::endl;
								sout << "\tlabels(x2,y2)=" << labels(x2,y2) << std::endl;
								sout << "\tinitial_mask_i(x1,y1)=" << initial_mask_i(x1,y1) << std::endl;
								sout << "\tinitial_labels(x1,y1)=" << initial_labels(x1,y1) << std::endl;
								sout << "\tinitial_mask_i(x2,y2)=" << initial_mask_i(x2,y2) << std::endl;
								sout << "\tinitial_labels(x2,y2)=" << initial_labels(x2,y2) << std::endl;*/
								
								//for(size_t kkk = 0; kkk < segments.size(); kkk++)
								//{
								//	sout << segments[kkk].id << std::endl;
								//}

								throw(std::range_error("CompleteMask:execute:Segment of nearest distance not in vector"));
							}

							int count2 = 0;
							cimg_forXY( labels, _x, _y)
							{
								if(labels(_x,_y) == id1 )
									count2++;
							}
							if( count2 != seg1.size())
							sout << "labels_count = " << count2 << " / segment.size = " << seg1.size() << " / segment.coord.size = " << seg1.coordinates.size() << std::endl;

							// Reference to segment 2
							hekate::CompleteMaskSegment &seg2 = segments[ i_seg2 ];

							// add coordinates of seg2 to seg1
							seg1.addCoordinates( seg2 );

							// Change labels in label image from id2 to id1
							for( unsigned int ii_seg2 = 0; ii_seg2 < seg2.coordinates.size(); ii_seg2++){
								labels(seg2.coordinates[ii_seg2].first, seg2.coordinates[ii_seg2].second) = id1;
							}

							final_mask.draw_line( x1, y1, x2, y2, color );

							/*
							// change labels of seg2 to label id of seg1
							mask_i.draw_line( x1, y1, x2, y2, color );
						
							const unsigned char color_id[] = { id1 } ;
							labels.draw_line( x1, y1, x2, y2, color_id );
						
							std::pair<unsigned int, unsigned int> x_coord = std::minmax( x1, x2 );
							std::pair<unsigned int, unsigned int> y_coord = std::minmax( y1, y2 );
							// Add drawn line pixels to segment also
							for(unsigned int itx = x_coord.first; itx <= x_coord.second; itx++)
								for(unsigned int ity = y_coord.first; ity <= y_coord.second; ity++)
									if( itx != x_coord.first && ity != y_coord.first )
										if( itx != x_coord.second && ity != y_coord.second )
											if( labels( itx, ity ) == id1 )
												seg1.addCoordinate( itx, ity, true );
							*/
							// Erase segment2 since pixels have been merged with segment1
							//sout << "erase" << std::endl;
							//sout << "i_seg2 = " << i_seg2 << " | segments.size= " << segments.size() <<  std::endl;

							int count = 0;
							cimg_forXY( labels, _x, _y)
							{
								if(labels(_x,_y) == id2 ){
									sout << "connected x1=" << x1 << ",y1=" << y1 << "/x2=" << x2 << ",y2=" << y2 << std::endl;
									sout << "point " << _x << " / " << _y << " still has id2" << std::endl;
								}
								else if(labels(_x,_y) == id1 )
									count++;
							}

							if( count != seg1.size())
							sout << "labels_count = " << count << " / segment.size = " << seg1.size() << " / segment.coord.size = " << seg1.coordinates.size() << std::endl;
							
							//sout << "\tSegment " << id2 << " is removed due to merging with " << id1 << std::endl;
							segments.erase( segments.begin() + i_seg2, segments.begin() + i_seg2 +1 ); 

							// if segment 2 was in front of segment 1, reduce iterator. Thus
							// next iteration continues with segment 1
							if( i_seg2 < i_seg )
								i_seg --;

							//sout << "erase2" << std::endl;
							//break;
						}
						// Else, continue with next segment
						else{
							sout << "\t" << segments.size() << " segments remaining" << std::endl;
							nnn = false;
							break;
						//	// erase segment since no further segments can be nn
						//	if( segments[i_seg].id == 453 ) sout << "Removed as first segment" << std::endl;
						//	segments.erase( segments.begin() + i_seg, segments.begin() + i_seg +1 );
						//	break;
						}
					
					}

				}

			//}

			cimg_forXY( final_mask, x, y){
				if( final_mask(x,y) == 0 && final_mask.get_crop(x-1,y-1,x+1,y+1).sum() >= 5 ){
					final_mask(x,y) = 1;
				}
			}
			
			mask_list[i] = final_mask ;
			
		}		

		// Write masks
		sout << "Found " << n_masks << " different masks" << std::endl;
		CImg<int> mask_complete(w,h,1,1,0);
		for( size_t i = 0; i < mask_list.size(); i++ )
		{
			if(!mask_list[i].is_empty())
				mask_complete += mask_list[i];
		}
		mask_complete.cut(0,1).normalize(0,255);
		this->masksCompleted()[index] = mask_complete;

	}
	/*
    void removeDuplicates(std::vector<T>& vec)
    {
        std::sort(vec.begin(), vec.end());
        vec.erase(std::unique(vec.begin(), vec.end()), vec.end());
    }
	*/

}

// Add all label pixels to segments
void getSegments(
			std::vector< hekate::CompleteMaskSegment > &segments,
			const CImg<int> &labels,
			const unsigned int &n_labels ){

	segments.clear();
	segments.resize(n_labels);

	cimg_forXY( labels, x, y )
	{ 
		// add segment coordinates x,y to segment of label id labels(x,y)
		if( labels(x,y) < n_labels )
		{
			segments[ labels(x,y) ].id = labels(x,y);
			segments[ labels(x,y) ].addCoordinate(x,y); 
		} 
		else
		{
			sout << "segments.size < labels(x,y)" << std::endl;
			throw(std::range_error("CompleteMask::execute: segments.size < labels(x,y)"));
		}
	} 

}

/// Remove black holes and small clusters in mask
/* 
	Attention: This method does not modify the segments.
	Segments have to be reassigned afterwards, if modification
	of mask should be taken into account
*/
void removeSmallClusters(
			CImg<int> &mask_i,
			const std::vector< hekate::CompleteMaskSegment > &segments,
			const unsigned int &n_labels,
			const int maxSizeBlack,
			const int minSize ){

			
	//sout << "Added coordinates to segments" << std::endl;
	//for( size_t j = 0; j < n_labels; j++ ){ sout << "segments["<<j<<"].size = " << segments[j].size() << std::endl;	}
			
	/* 
	Remove black segments smaller than min value and fill pixels with 1, 
	remove segments smaller than min value and fill pixels with 0,
	and remove black segments at all 
	*/
	int sb=0, sw=0;
	int lb=0, lw=0;
	for( size_t k = 0; k < n_labels; k++ ){

		const hekate::CompleteMaskSegment &seg = segments[k]; // ref to segments
				
		const unsigned int x = seg.coordinates[0].first;
		const unsigned int y = seg.coordinates[0].second;

		// Condition for a black hole segment
		if(mask_i(x,y) == 0 && seg.size() <= maxSizeBlack )
		{
			lb++;
			// fill black pixels
			for( size_t kk = 0; kk < seg.coordinates.size(); kk++ )
			{
				mask_i( seg.coordinates[kk].first, seg.coordinates[kk].second ) = 1;	
				sb++;
			}
		}

		else if( mask_i(x,y) != 0 && seg.size() <= minSize ){
			lw++;
			// fill pixels with 0
			for( size_t kk = 0; kk < seg.coordinates.size(); kk++ )
			{
				mask_i( seg.coordinates[kk].first, seg.coordinates[kk].second ) = 0;	
				//mask_i_final( seg.coordinates[kk].first, seg.coordinates[kk].second ) = 0;	
				// decrease segments size by 1
				sw++;
			}
		}
	}

	sout << "Removed black holes and segments which are too small" << std::endl;
	sout << "\tRemoved " << lb << " black wholes and made them white (" << sb << "pixels) " << std::endl;
	sout << "\tRemoved " << lw << " small segments (" << sw << "pixels) " << std::endl;

}

/// Remove black background clusters
void removeBackgroundClusters(
			const CImg<int> &mask_i,
			std::vector< hekate::CompleteMaskSegment > &segments,
			const unsigned int &n_labels,
			const int maxSizeBlack,
			const int minSize ){
	
	sout << segments.size() << " segments before background removement" << std::endl;
				
	int lbb=0;
	for( size_t k = 0; k < segments.size(); k++ ){

		hekate::CompleteMaskSegment &seg = segments[k]; // ref to segments
				
		const unsigned int x = seg.coordinates[0].first;
		const unsigned int y = seg.coordinates[0].second;

		// If pixels are black but do not belong to a small enough segment
		// remove it from further processsing
		if( mask_i(x,y) == 0 )
		{
			for( size_t kk = 0; kk < seg.size(); kk++ ){if(mask_i(seg.coordinates[kk].first,seg.coordinates[kk].second)!=0)throw(std::range_error("DIFFERENT LABELS!!"));}
			sout << "removed label " << seg.id << std::endl;
			lbb++;
			segments.erase( segments.begin() + k, segments.begin() + k + 1 );
			k--;
		}

	}
		
	sout << "Removed " << lbb << " black background segments" << std::endl;
	sout << segments.size() << " segments remaining" << std::endl;
}


#endif /* _COMPLETEMASK_HXX_ */
