/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CompleteMaskDijkstra.h
 *  Declaration of the parameter class CompleteMaskDijkstra.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 11.12.2013
 */

#ifndef _COMPLETEMASKDIJKSTRA_H_
#define _COMPLETEMASKDIJKSTRA_H_

#ifdef _MSC_VER
#ifdef completemaskdijkstra_EXPORTS
/// Visual C++ specific code
#define completemaskdijkstra_DECLDIR __declspec(dllexport)
#else
#define completemaskdijkstra_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else
/// Not needed without MSVC
#define completemaskdijkstra_DECLDIR
#endif

#include <charon-core/ParameteredObject.h>
#include <charon-utils/CImg.h>
#include <list>

using namespace cimg_library;

/// Endpoint class with x,y,id,cost as members
struct Endpoint;

/// Connects segments using dijkstra and a weight mask
/** Connects segments using dijkstra and a weight mask for the costs (higher
 *  values mean a lower cost, 0 means infinite cost --> no connection possible).
 */
template <typename T>
class completemaskdijkstra_DECLDIR CompleteMaskDijkstra :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	CompleteMaskDijkstra(const std::string& name = "");

	/// Mask segments to connect. The segments are supposed to be saved as an
	/// integer cimg where segments have values taller than zero.
	InputSlot< cimg_library::CImgList<T> > masks;
	/// A weight mask which weights the possible connection paths. The input
	/// values are considered to be high for a low cost and zero if the cost
	/// is indefinite which means segments cannot be connected by this path.
	InputSlot< cimg_library::CImgList<T> > weights;

	/// Connected mask segments
	OutputSlot< cimg_library::CImgList<T> > maskConnected;

	/// Maximal allowed distance of the connection path
	Parameter< int > maxDistance;
	/// Minimal allowed number of pixels of a segment to be considered for completion
	Parameter< int > minSize;
	/// Minimum allowed size of black holes. If a black segment is smaller
	/// than this value, the hole is filled with its surrounding mask values
	Parameter< int > minBlackHoleSize;
	/// Number of iterations removing black holes surrounded by at least 5 pixels
	Parameter< int > singlePixelCorrection;
	/// Lower threshold for the weight mask; Pixels with values below
	/// won't be considered for connections
	Parameter< float > weightThreshold;

	/// Start calculation from here
	Parameter< int > inFrame;
	/// End calculation here
	Parameter< int > outFrame;


protected:
	/// Update object.
	virtual void execute();

	/// Get ids for different segments of mask value i
	/** 
	Get labels/ids for different segments of mask value i
	The CImg function get_label returns an integer instance
	containing ids for the different segments.
	Remove bad segments such as segments smaller than min size
	or black holes smaller than min size
	*/
	void getSegmentLabels( CImg<int> &mask_i, CImg<int> &labels, CImg<int> &cluster_size );

	/// Connect clusters using Dijkstra
	void connectSegmentLabels( CImg<int> &mask_i, 
					CImg<int> &labels, const CImg<float> &weight, CImg<int> &cluster_size );

	bool find_path( std::list< std::pair<int,int> > &stack, std::vector< Endpoint > &endpoints, CImg<int> &mask_i,
					const CImg<float> &weight, CImg<float> &weight_i, CImg<float> &cost_i, CImg<int> &labels,
					const int id, const int maxDist, CImg<int> &hist );
};

#endif // _COMPLETEMASKDIJKSTRA_H_
