/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CompleteMaskDijkstra.hxx
 *  Implementation of the parameter class CompleteMaskDijkstra.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 11.12.2013
 */

#ifndef _COMPLETEMASKDIJKSTRA_HXX_
#define _COMPLETEMASKDIJKSTRA_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CompleteMaskDijkstra.h"
#include "CompleteMaskSegment.hxx"
#include <cfloat>
#include <list>

template <typename T>
CompleteMaskDijkstra<T>::CompleteMaskDijkstra(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CompleteMaskDijkstra", name,
			"<h2>Connects segments using dijkstra and a weight mask</h2><br>"
			"Connects segments using dijkstra and a weight mask for the "
			"costs (higher values means a lower cost, 0 means infinite cost --> no connection possible). "
			"Only segments of the same value are connected."
		)
{
	ParameteredObject::_addInputSlot(
		masks, "masks",
		"Mask segments to connect. The segments are supposed to be saved as"
		"an integer cimg where segments have values taller than zero.",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		weights, "weights",
		"A weight mask which weights the possible connection paths. The input"
		"values are considered to be high for a low cost and zero if the cost"
		"is indefinite which means segments cannot be connected by this path.",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		maskConnected, "maskConnected",
		"Connected mask segments",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		maxDistance, "maxDistance",
		"Maximal allowed distance of the connection path",
		25, "int");
	ParameteredObject::_addParameter< int >(
		minSize, "minSize",
		"Minimal allowed number of pixels of a segment "
		"to be considered for completion.",
		5, "int");
	ParameteredObject::_addParameter< int >(
		minBlackHoleSize, "minBlackHoleSize",
		"Minimum allowed size of black holes. If a black segment is smaller"
		"than this value, the hole is filled with its surrounding mask values",
		25, "int");
	ParameteredObject::_addParameter< int >(
		singlePixelCorrection, "singlePixelCorrectionStrength",
		"Number of iterations removing black holes surrounded by at least 5 pixels",
		10, "int");
	ParameteredObject::_addParameter< float >(
		weightThreshold, "weightThreshold",
		"Lower threshold for the weight mask; Pixels with values below"
		" won't be considered for connections",
		0.01, "float");

	ParameteredObject::_addParameter< int >(
		inFrame, "inFrame",
		"Start calculation from here",
		0, "int");
	ParameteredObject::_addParameter< int >(
		outFrame, "outFrame",
		"End calculation here",
		-1, "int");

}

template <typename T>
void CompleteMaskDijkstra<T>::execute() {
			
	// Painting color
	const unsigned char color[] = { 1 } ;

	const int in_size = this->masks().size();

	float threshold_weight = std::min(float(0.99), float(this->weightThreshold()));
	threshold_weight = std::max( float(threshold_weight), float(0) );
	sout << "Lower weight threshold: " << threshold_weight << std::endl;

	if(this->weights().size() != in_size )
		throw(std::range_error("this->weights().size() != in_size"));

	int i_start = this->inFrame();
	int i_end = this->outFrame();

	i_start = std::max(0, i_start);
	i_end = ( i_end == -1 ? in_size : std::min(in_size,i_end+1) );

	sout << "Processing " << in_size << " images" << std::endl;
	
	CImgList<T> &out = this->maskConnected();
	out.assign(in_size);
	
	for( size_t index = 0; index < in_size; index++ )
	{
		
		sout << "---" << std::endl;
		sout << "Processing frame " << index << std::endl;

		const CImg<int> &mask	= this->masks()[index];
		const int w	= mask.width();
		const int h	= mask.height();

		// ACHTUNG KEI NTHRESHOLD HERE!!!
		CImg<float> weight = this->weights()[index].get_channel(0).normalize(0,1);
		cimg_forXY(weight,x,y){
			if(weight(x,y) < threshold_weight)
				weight(x,y) = FLT_MAX;
			else{
				// weight should not be 0!
				weight(x,y) = weight(x,y) * (-1) +1 +0.1;
			}
		}
		
		out[index].resize(w,h,1,1,0);
				
		// Get min/max value of pixels
		const int max = mask.max();

		// Get histogram to determine different values of the given mask
		// If the mask only contains two different values (background and
		// foreground) the next iteration only runs once. Otherwise there is
		// one iteration for every segment/mask value.
		// starting from 1 skips the black background (?)
		CImg<int> hist = mask.get_histogram( max+1 );

		// Number of histrogram values. This is the highest value
		// of the given mask +1 an so the maximum possible number
		// of different mask values. Though, if a mask is given with
		// two different values 0 and 255, the algorithm at first considers
		// 256 different mask values and will skip all beside one of them
		// by the first condition inside of the loop
		const int n_levels = hist.width();
		
		CImgList<int> mask_list(n_levels);

		/* 
			All pixels which belong to a specific color/grey value will be handled as pixels
			for one and the same edge type (e.g. depth edges).
			The next loop iterates over these different edge types and processes the
			edge type pixels seperately.
			(1) First, connection labels are build where all pixels connected to a cluster are 
			uniquely labeled.
			(2) Afterwards, black holes (small black clusters) are removed and filled in 
			the mask image, if they are smaller	than the maximum allowed cluster size for 
			black holes. Also clusters smaller than the minimum allowed size are removed for 
			further processing. In this way, noisy pixels/clusters are handled. 
			Both parameters (minimum,maximum cluster size) are set in the plugin parameters.
			(3) Segments are reassigned according to modified mask image
			(4) Black background segments are removed
			(5) Nearest neighbours between segments are searched
		*/

		// Iteration over every possible mask value i
		// Start with i == 1 to skip the black background
		for( int i = 1; i < n_levels; i++ )
		{

			const clock_t begin_time = clock();

			// check if there are pixels with value i
			// if there is no mask with value i, continue
			if( hist(i,0) == 0 )
				continue;

			sout << "Completing mask segments with value " << i << " (" 
					<< hist(i,0) << "pixels)" << std::endl;


			// Get 0||1 mask of pixels with the former value i
			CImg<int> mask_i(w, h, 1, 1, 0);
			cimg_forXY( mask_i, x, y )
			{
				if( mask(x,y) == i )
					mask_i(x,y) = 1 ;
				// Also add non-segment pixels to segments which have 6 or more
				// surrounding segment neighbour pixels
				else if( mask(x,y) == 0 && neighbours( mask, x, y, i ) >= 5 )
					mask_i(x,y) = 1 ;
			}

			// (1)
			/* 
			Get cluster ids for segments of mask value i.
			The function getSegmentLabels thereby returns a label image
			where pixels of interest have a cluster id != 0
			*/
			CImg<int> labels, cluster_size;
			getSegmentLabels( mask_i, labels, cluster_size );

			sout << "Took " << float( clock () - begin_time ) / CLOCKS_PER_SEC << "sseconds for preprocessing " << std::endl;

			// connect clusters using the cluster ids and the weight field
			connectSegmentLabels( mask_i, labels, weight, cluster_size );
			
			for( int k = 0; k < this->singlePixelCorrection(); k++ ){
				cimg_forXY( mask_i, x, y ){
					if( mask_i(x,y) == 0 && neighbours( mask_i, x, y, 1 ) >= 5 )
						mask_i(x,y) = 1 ;
				}
			}

			out[index] += mask_i;

		}

		out[index].cut(0,1).normalize(0,255);

	}

}


/// Calculate offset to given mask
/*
	Slow but only run once
*/
void get_offset( CImg<int> &img, const int distance ){

	int count = 0;
	CImg<int> temp(img.width(),img.height(),1,1,0);
	cimg_forXY( img, x, y ){
		if( img(x, y) != 0 ){
			
			// paint box around pixel
			for( int i = std::max(0,x-distance); i <= std::min(img.width()-1, x+distance); i++ ){
				for( int j = std::max(0,y-distance); j <= std::min(img.height()-1, y+distance); j++ ){
					temp(i,j) = 1;
					count++;
				}
			}
		}
	}
	
	img = temp;

}


/// Return number of neighbours pixels with value 'val' which is 1 by default
inline const int neighbours(const CImg<int> &img, const int x, const int y, const int val = 1){

	int count = 0;

	// Get search window
	int minX = std::max(0,x-1);
	int maxX = std::min(img.width()-1,x+1);
	int minY = std::max(0,y-1);
	int maxY = std::min(img.height()-1,y+1);

	// count neighbour pixels
	for(int i = minX; i <= maxX; i++ ){
		for(int j = minY; j <= maxY; j++ ){
			if( img(i,j) == val && !(i==x && j==y) )
				count ++;
		}
	}

	return count;

}


/// Return minimum value of cost from neighbour pixels with value 'val' of img
const float neighboursMin(const CImg<int> &img, const CImg<float> &cost, const int x, const int y, const int val){

	float min = FLT_MAX;

	// Get search window
	int minX = std::max(0,x-1);
	int maxX = std::min(img.width()-1,x+1);
	int minY = std::max(0,y-1);
	int maxY = std::min(img.height()-1,y+1);

	// count neighbour pixels
	for(int i = minX; i <= maxX; i++ ){
		for(int j = minY; j <= maxY; j++ ){
			if( (img(i,j) == val || img(i,j) == 0) && !(i==x && j==y) && cost(i,j) < min )
				min = cost(i,j);
		}
	}

	return min;

}

/// Get ids for different segments of mask value i
/** 
Get labels/ids for different segments of mask value i
The CImg function get_label returns an integer instance
containing ids for the different segments.
Remove bad segments such as segments smaller than min size
or black holes smaller than min size
*/
template<typename T>
void CompleteMaskDijkstra<T>::getSegmentLabels( CImg<int> &mask_i, CImg<int> &labels, CImg<int> &cluster_size ){

	// At first, remove small black hole clusters -------------------------------------------
		
	const clock_t begin_time2 = clock();
	sout << "Took " << float( clock () - begin_time2 ) / CLOCKS_PER_SEC << " seconds to remove black holes " << std::endl;

	// get unique labels for every cluster
	labels = mask_i.get_label(true);

	// Get number of different clusters.
	int n_labels	= labels.max() + 1 ;

	// Get histogram of label image to determine sizes of clusters
	cluster_size = labels.get_histogram( n_labels, 0, n_labels -1 );

	cimg_forXY( labels, x, y ){
		// if pixel of mask == 0 and belongs to a small cluster, 
		// transform it to a segment pixel
		if( mask_i(x,y) == 0 && cluster_size( labels(x,y), 0 ) < this->minBlackHoleSize() ){
				mask_i(x,y) = 1;
		}
	}
	
	// Then set label id of pixels not if interest to 0 -------------------------------------------

	// get unique labels for every cluster
	labels = mask_i.get_label(true);

	// Add 1 to every single pixel, thus 0 values are
	// unused and can be taken to mark non-segment pixels
	// afterwards. (for example for the background pixels). Otherwise
	// it is not known, which pixels did get the id = 0
	labels += 1;

	// Get number of different clusters. This also includes
	// segments of pixels with value 0
	n_labels	= labels.max() + 1 ;

	// Get histogram of label image to determine sizes of clusters

	cluster_size = labels.get_histogram( n_labels, 0, n_labels -1 );
		
	cimg_forXY( labels, x, y ){
		// if pixel of mask == 1 and belongs to a small cluster, 
		// set its id to zero, aswell as if pixel of mask == 0
		if( mask_i(x,y) == 0 ){
			// !! first remove custersize of label, then change label
			cluster_size( labels(x,y) , 0 ) = 0;
			labels(x,y) = 0;
		}
		else if(cluster_size( labels(x,y), 0 ) < this->minSize() ){
			// !! first remove custersize of label, then change label
			cluster_size( labels(x,y) , 0 ) = 0;
			mask_i(x,y) = 0;
			labels(x,y) = 0;
		}
	}

	/*
	We now should have a label image where small clusters and background pixels
	where set to zero and otherwise pixels have a unique cluster id
	*/
	
	sout << "Took " << float( clock () - begin_time2 ) / CLOCKS_PER_SEC << " seconds to remove black holes " << std::endl;

}

/// Endpoints class which defines points with coordinates and ids
struct Endpoint{

public:

	int x, y, id;
	float cost;

	explicit Endpoint( int _x, int _y, int _id, float _cost ):
		x(_x), y(_y), id(_id), cost(_cost) {}



	virtual ~Endpoint() {}

};

/// Connect clusters using Dijkstra algorithm
template<typename T>
void CompleteMaskDijkstra<T>::connectSegmentLabels( CImg<int> &mask_i, 
						CImg<int> &labels, const CImg<float> &weight, CImg<int> &cluster_size ){
									
	const int w = mask_i.width();
	const int h = mask_i.height();
	const float cs = cluster_size.width();

	// Cost image for cluster i
	CImg<float> cost_i( w, h, 1, 1, FLT_MAX );

	// blur area mask to define offset
	CImg<int> area_i(w,h,1,1);

	CImg<float> weight_i(w,h,1,1);
					
	// Start vectors are neighbourvectors of the current cluster
	std::list< std::pair<int,int> > stack;

	// List of possible ending points
	std::vector< Endpoint > endpoints;

	sout << "Process up to " << cs << " clusters" << std::endl;
		
	sout << "0%" << std::endl ;

	// iterate over all cluster ids
	for( int i = 1; i < cs; i++ ){
		
		//sout << float(i)/cluster_size.width() * 100 << "%" << std::endl;
		if(i%10 == 0)
				sout << int(float(i)/float(cs)*100) << "%" << std::endl;

		stack.clear();
		endpoints.clear();
		endpoints.assign(cs, Endpoint(0,0,0,FLT_MAX) );

		// if no pixels with id 'i', continue
		if( cluster_size(i,0) == 0 )
			continue;
		
		// Set cost value of cluster i pixels to 0
		// Set neighbourpixels of cluster i to weight
		// select one of the neighbour pixels as a start point
		// for dijkstra
		cimg_forXY( labels, x, y ){
			if(labels(x,y) == i){
				cost_i(x,y) = 0;
				area_i(x,y) = 1;
				if(neighbours(labels,x,y,0) > 0){
					stack.push_back( std::pair<int,int>(x,y) );
				}
			}
			else{
				cost_i(x,y) = FLT_MAX;
				area_i(x,y) = 0;
			}
		}

		if( stack.size() == 0 ){
			sout << "Stack == 0, why ist that???" << std::endl;
			sout << "cluster_size(i,0) = " << cluster_size(i,0) << std::endl;
		cimg_forXY( labels, x, y ){
			if(labels(x,y) == i){
				sout << x << ", " << y << std::endl;
			}
		}
			continue;
		}

		get_offset( area_i, this->maxDistance() );

		cimg_forXY( weight_i, x, y ){
			if( area_i(x,y) == 1 )
				weight_i(x,y) = weight(x,y);
			else
				weight_i(x,y) = FLT_MAX;
		}		
	
		//const clock_t begin_time2 = clock();
		// Find paths to unknown end points by calculating the cost field
		while( find_path( stack, endpoints, mask_i, weight, weight_i, cost_i, labels, i, this->maxDistance(), cluster_size ) ){
			// loop while neighbour clusters are found
			//sout << "\t...Found neighbour, continue to find some more..." << std::endl;
			if(i%10 == 0)
				sout << int(float(i)/float(cs)*100) << "%" << std::endl;
		}
			
		//sout << "\tTook " << float( clock () - begin_time2 ) / CLOCKS_PER_SEC << "s to connect neighbour clusters " << std::endl;
	


	}

}


/// Check whether position has a smaller distance as prior values
inline void addPathPos(const int x, const int y, std::list< std::pair<int,int> > &stack, CImg<float> &distances, 
				const CImg<float> &weights, float weight_factor, 
				const CImg<int> &labels, std::vector<Endpoint> &endpoints, const int id)
{
	
	// Check point if it is inside of the image, inside of the allowed area
	// and not belonging to the source cluster id for which the weight is set to 0
	if( distances.containsXYZC(x, y) && weights(x,y) < FLT_MAX )
	{

		// label id of current pixel
		const int id2 = labels(x,y);

		// Dont process pixels of source cluster
		if( id == id2 )
			return;

		float oldDist = distances(x, y);
		//float newDist = currentDist + weights(x, y) ;
		float newDist = neighboursMin(labels,distances,x,y,id) + weight_factor * weights(x, y) ;

		// Update minimum cost for an end point
		if( id2 != 0 ){
			//sout << "d2" << std::endl;
			if( endpoints[id2].cost > newDist )
				endpoints[id2] = Endpoint(x,y,id2,newDist);
			return;
		}
		// Update cost for a point between clusters
		else if(newDist < oldDist){
			//sout << "d3" << std::endl;
			distances(x, y) = newDist;
			stack.push_back( std::pair<int,int>(x,y) );
		}
	}
}


/// Sort function sorting endpoints by id
bool SortEndpointsId( const Endpoint &p1, const Endpoint &p2 ){
	return ( p1.id < p2.id );
}


// /// Sort function sorting endpoints by distance
// bool SortEndpointsDist( const Endpoint &p1, const Endpoint &p2 ){
// 	return ( p1.cost < p2.cost );
// }

/// Sort function sorting endpoints by distance
bool SortEndpointsDist( const Endpoint* p1, const Endpoint* p2 ){
	return ( p1->cost < p2->cost );
}

/// Sort function sorting endpoints by distance
bool HasUniqueId( const Endpoint &p1, const Endpoint &p2 ){
	return ( p1.id==p2.id );
}

/// Sort function sorting endpoints by distance
bool IsEqual( const Endpoint &p1, const Endpoint &p2 ){
	return ( p1.cost==p2.cost && p1.id==p2.id && p1.x==p2.x && p1.y==p2.y );
}


/// Find path to end points within the search area_i
template<typename T>
bool CompleteMaskDijkstra<T>::find_path( std::list< std::pair<int,int> > &stack, std::vector< Endpoint > &endpoints,
	CImg<int> &mask_i,
	const CImg<float> &weight, CImg<float> &weight_i, CImg<float> &cost_i, CImg<int> &labels,
	const int id, const int maxDist, CImg<int> &hist ){
		
	const int w = labels.width();
	const int h = labels.height();
	const int dist = this->maxDistance();
		
	//const clock_t begin_time2 = clock();
	
	//sout << "\t\tCreate cost field for " << area_size << " pixels" << std::endl;

	float diagonal_weight = std::sqrt(float(2));

	int ccount = 0;
	//this->maskConnected().push_back( cost_i.get_cut(0,10).normalize(0,255) );
	//this->maskConnected().push_back( weight_i.get_cut(0,2).normalize(0,255) );
	while(stack.size() > 0 )
	{
		std::pair<int,int> p = stack.front();
		int _x = p.first;
		int _y = p.second;
		stack.pop_front(); 
		ccount++;
		
		//add 8-neighborhood to stack and update weights
		float currentDist = cost_i(_x, _y);
		addPathPos(_x-1, _y,	stack, cost_i, weight_i, 1, labels, endpoints, id);
		addPathPos(_x-1, _y-1,	stack, cost_i, weight_i, diagonal_weight, labels, endpoints, id);
		addPathPos(_x,	 _y-1,	stack, cost_i, weight_i, 1, labels, endpoints, id);
		addPathPos(_x+1, _y-1,	stack, cost_i, weight_i, diagonal_weight, labels, endpoints, id);
		addPathPos(_x+1, _y,	stack, cost_i, weight_i, 1, labels, endpoints, id);
		addPathPos(_x+1, _y+1,	stack, cost_i, weight_i, diagonal_weight, labels, endpoints, id);
		addPathPos(_x,	 _y+1,	stack, cost_i, weight_i, 1, labels, endpoints, id);
		addPathPos(_x-1, _y+1,	stack, cost_i, weight_i, diagonal_weight, labels, endpoints, id);
	}

	//sout << "\tTook " << float( clock () - begin_time2 ) / CLOCKS_PER_SEC << "s to create cost field " << std::endl;
	//sout << "\tNo possible connections found" << std::endl;

	// DEBUG DEBUG DEBUG
	//this->maskConnected().push_back( cost_i.get_cut(0,5).normalize(0,255) );
	
	std::list<Endpoint*> end_sort;
	for( auto it = endpoints.begin(); it != endpoints.end(); it++ ){
		end_sort.push_back( &(*it) );
	}

	end_sort.sort(SortEndpointsDist);

	// Back tracing using the cost image -----------------------------------------

	// Iterate through all possible endpoints	
	for( auto it = end_sort.begin(); it != end_sort.end(); it++ ){
				
		
		// Skip pixels which belong to current cluster or no cluster
		// as a end point
		if( (*it)->id == 0 || (*it)->id == id )
			continue;

		// Back tracing for end points of other clusters
		if( (*it)->cost < FLT_MAX ){
			
			// temporary cost image for back tracing
			CImg<float> cost_i_temp;

			// path for back tracing
			std::vector<std::pair<int,int>> path;
			
			int x = (*it)->x;
			int y = (*it)->y;
			int id2 = (*it)->id;

			cost_i_temp = cost_i;

			path.clear();
			// take end point as first point for the path
			path.push_back(std::pair<int,int>(x,y));
			cost_i_temp(x,y) = FLT_MAX; //remove point from temporary cost matrix
			int count =0;
			
			// do until start point is arrived
			while(labels(x,y) != id)
			{
				int minX =-1;
				int minY =-1;
				float minDist = FLT_MAX;
				for(int dy =-1; dy <=1; dy++)
				{
					for(int dx =-1; dx <=1; dx++)
					{
						//only look at other pixels inside distance image
						if(cost_i_temp.containsXYZC(x+dx, y+dy) && !(dx == 0 && dy == 0)) 
						{
							float v =cost_i_temp(x+dx, y+dy);
							// weight diagnal tracks
							if( dx!=0 && dy!=0 )
								v *= std::sqrt(float(2));
							if(v <= minDist)
							{
								minDist =v;
								minX =x+dx;
								minY =y+dy;
							}
						}
					}
				}

				if(minX == -1){
					sout << "!Not connection found for clusters" << id2 
						<< " and " << id << std::endl;
					// set id to 0 so we know this cluster is not merged with cluster 'id'
					break;
				}
				cost_i_temp(minX, minY) = FLT_MAX; //remove point from distance matrix
				path.push_back(std::pair<int,int>(minX,minY));
				x = minX;
				y = minY;
			}

			// If path is complete, add last pixel belongs to cluster 'id'
			if( labels(x,y) == id ){

				//#pragma omp critical
				//{
				
					// Set histogram value of cluster 'id2' to zero
					// since those values are transfered to cluster 'id'
					hist(id2,0) = 0;
					
					// Add path to cluster 'id'
					for(int k = 0; k < path.size(); k++ ){
						mask_i(path[k].first,path[k].second) = 1;
						labels(path[k].first,path[k].second) = id;
					}
					
					// temporary search area determined in next loop
					CImg<int> area( labels.width(), labels.height(), 1, 1, 0 );

					// Merge clusters, update weight
					cimg_forXY( labels, __x, __y ){
						if( labels(__x,__y) == id2 ){
							labels(__x,__y) = id;
						}
						if(labels(__x,__y) == id){
							// set costs to zero for just added cluster pixels
							cost_i(__x,__y) = 0;
							// set weight to zero for just added cluster pixels
							weight_i(__x,__y) = 0;
							// reset starting points for merged cluster
							if(neighbours(labels,__x,__y,0) > 0){
								stack.push_back( std::pair<int,int>(__x,__y) );
							}
							// determine new search area ---------------
							// (same as in get_offset(....)
							int left = std::max(0,__x-dist);
							int right = std::min(w-1, __x+dist);
							int top = std::max(0,__y-dist);
							int bottom = std::min(h-1, __y+dist);
							for( int i = left; i <= right; i++ ){
								for( int j = top; j <= bottom; j++ ){
									area(i,j) = 1;
								}
							}
						}
					}
										
					// Update weight according to new search area
					cimg_forXY( labels, __x, __y ){
						// set weight for all pixels inside of the search area
						// and outside of cluster
						if(area(__x,__y) == 1){
							weight_i(__x,__y) = weight(__x,__y);
						}
						// set weight to max for all pixels outside 
						// of the search area
						else{
							weight_i(__x,__y) = FLT_MAX;
						}
					}

					// Set cost of ending point to zero for next iteration
					endpoints[id2].cost = FLT_MAX;

					//this->maskConnected().push_back( weight_i.get_cut(0,1).normalize(0,255) );
					
					//sout << "\tMerged cluster " << id << " with " << (*it)->id << std::endl;
					
					// return true, since an endpoint was found
					return true;

				//}

			}

		}

	}

	return false;

}

#endif /* _COMPLETEMASKDIJKSTRA_HXX_ */
