#ifndef _COMPLETEMASKSEGMENT_HXX_
#define _COMPLETEMASKSEGMENT_HXX_

namespace hekate {

	/// Container class which includes one ore more random forests, its image ids, 
	class CompleteMaskSegment
	{
	public:
		
		/// left x coordinate
		unsigned int x_min;
		/// right x coordinate
		unsigned int x_max;
		/// top y coordinate
		unsigned int y_min;
		/// bottom y coordinate
		unsigned int y_max;

		/// id of segment
		unsigned int id;

		/// vector of pixel coordinates
		std::vector< std::pair<unsigned int, unsigned int> > coordinates;

		/// default constructor
		explicit CompleteMaskSegment(){
			_size = 0;
			_used = false;
		}
		
		/*
		/// constructor
		explicit CompleteMaskSegment( unsigned int id, unsigned int size, unsigned int x_min,
			unsigned int x_max, unsigned int y_min, unsigned int y_max) :
			id(id), size(size), x_min(x_min), x_max(x_max), y_min(y_min), y_max(y_max)
		{		
			_used = false;
		}
		*/

		unsigned int size() const { return _size;}

		void decreaseSize() { _size-- ; }

		void setEmpty() { _size = 0; }
		
		void setMinX( const unsigned int x ){ x_min = std::min( x, x_min ); }
		void setMinY( const unsigned int y ){ y_min = std::min( y, y_min ); }
		void setMaxX( const unsigned int x ){ x_max = std::max( x, x_max ); }
		void setMaxY( const unsigned int y ){ y_max = std::max( y, y_max ); }

		/// Add a pixel coordinate to this segment
		/** if exists flag is set, method checks, if point already exists */
		void addCoordinate( const unsigned int x, const unsigned int y, bool exists = false )
		{
			if(exists)
				for(size_t i = 0; i < coordinates.size(); i++)
				{
					if( coordinates[i].first == x && coordinates[i].second == y)
						return;
				}

			// add coordinate
			coordinates.push_back( std::pair<unsigned int, unsigned int>( x, y ) );
			// increase size
			_size++;
			// set min max
			if(_used){
				setMinX(x);
				setMinY(y);
				setMaxX(x);
				setMaxY(y);
			}
			else{
				x_min = x;
				y_min = y;
				x_max = x;
				y_max = y;
			}

			// set object to used
			_used = true;
		}

		/// Add pixels coordinates of another segment
		void addCoordinates( const CompleteMaskSegment &segment ){

			// add coordinates
			for( size_t i = 0; i < segment.coordinates.size(); i++ )
			{
				// add coordinate
				coordinates.push_back( segment.coordinates[i] );
				// increase size value
				_size++;
			}

			// set new min max
			if(_used){
				setMinX(segment.x_min);
				setMinY(segment.y_min);
				setMaxX(segment.x_max);
				setMaxY(segment.y_max);
			}
			else{
				x_min = segment.x_min;
				y_min = segment.y_min;
				x_max = segment.x_max;
				y_max = segment.y_max;
			}

			// set object to used
			_used = true;
		}

		/// returns coordinates of patch with additional border conditions
		void getCoordinates( int &x_min_, int &y_min_, int &x_max_, int &y_max_, const unsigned int border_width = 0 ) const
		{
			x_min_ = std::max( int(0), int(x_min) - int(border_width) );
			y_min_ = std::max( int(0), int(y_min) - int(border_width) );
			x_max_ = x_max + border_width;
			y_max_ = y_max + border_width;
		}

		/// checks if coordinate x,y exists
		bool hasCoordinate( unsigned int x, unsigned int y ) const
		{
			for( size_t i=0; i < coordinates.size(); i++ )
			{
				if(coordinates[i].first == x && coordinates[i].second == y)
					return true;
			}
			return false;
		}

		/// checks if coordinate of id i lies inside of the given patch
		bool hasCoordinateInPatch( const unsigned int _id, const int &x_min_, const int &y_min_, 
					const int &x_max_, const int &y_max_ ) const
		{
			if( coordinates[_id].first >= x_min_ && coordinates[_id].first <= x_max_ &&
				coordinates[_id].second >= y_min_ && coordinates[_id].second <= y_max_ )
				return true;
			else
				return false;
		}
		
		unsigned int width()	{ return x_max - x_min; }
		unsigned int height()	{ return y_max - y_min; }
			
		virtual ~CompleteMaskSegment() {}

	private:
		bool _used; /// flag if a coordinate has been set

		/// Number of pixels of this segment
		unsigned int _size;

	} ;	

	/// Calculates distance between two coordinates
	inline float getDistance(const unsigned int x1, const unsigned int y1, const unsigned int x2, const unsigned int y2 )
	{
		float dx = float(x1) - float(x2);
		float dy = float(y1) - float(y2);
		return std::sqrt( dx*dx + dy*dy );
	}

	/// Checks pixel position for another label and its distance.
	/** If distance is smaller than reference, replace according position */
	inline bool getDistance(
			const CImg<int> &mask, 
			const CImg<int> &labels, 
			const unsigned int id,
			const unsigned int x1,
			const unsigned int y1,
			const unsigned int x2,
			const unsigned int y2,
			float &distance,
			unsigned int &_id2,
			unsigned int &_x1, 
			unsigned int &_y1, 
			unsigned int &_x2, 
			unsigned int &_y2 )
	{
		if( mask.atXY(x2,y2,0,0,0) > 0 ){	// check if mask != 0					
			// check if pixel belongs to another label
			if( labels(x2,y2) != id ){	
				float dist = getDistance(x1,y1,x2,y2);
				if( dist < distance ){
					distance = dist;
					_x1 = x1, _y1 = y1, _x2 = x2, _y2 = y2;
					_id2 = labels(x2,y2);
					return true;
				}
			}
		}
		return false;
	}

	/// Determine shortest distance between segment of 'id' and other segments
	/** Returns id2 which is the id of the nearest segment and the coordinates for the distance vector*/
	bool getDistance( 
			std::vector<CompleteMaskSegment> &segments, 
			const unsigned int _id_vec,
			const unsigned int min_size, 
			const unsigned int max_distance, 
			unsigned int &_id2,
			unsigned int &_x1, unsigned int &_y1, unsigned int &_x2, unsigned int &_y2, 
			const CImg<int> &mask, 
			const CImg<int> &labels
			)
	{
			
		bool _d = false;	// flag for minimal distance

		float distance = float(max_distance) +0.1 ; // initial distance

		// Segment of interest
		const CompleteMaskSegment &seg = segments[_id_vec];
		const unsigned int id1 = seg.id;

		// Iterate over segment pixels and look for nearest neighbour if pixel
		// is at the border
		for( size_t j = 0; j < seg.size(); j++ )
		{
			const unsigned int x = seg.coordinates[j].first;
			const unsigned int y = seg.coordinates[j].second;

			// Continue, if pixel does not ly at a relevant border position
			// ( If there are 5 or more neighbour pixels (sum()-1), the pixel does not ly
			// at an relevant border position )
			if( mask.get_crop(x-1, y-1, x+1, y+1).normalize(0,1).sum() >= 6 )
				continue;

			// Search in max_distance area for nearest neighbours
			// iterates until d is true (smaller distance found) or
			// until no nn was found in nn area
			bool d = false;	// flag for new minimal distance
			for( size_t border = 2; border <= max_distance && !d; border++)
			{

				// Search in ROWS and COLUMNS with a width/height of 2*distance+1 ---------------------------
				for( size_t border_i = 0; border_i < border * 2 + 1; border_i++)
				{
					// rows

					const unsigned int xrr = x+border_i;	// x coord row right
					const unsigned int xrl = x-border_i;	// x coord row left
					const unsigned int yrt = y-border;		// y coord row top
					const unsigned int yrb = y+border;		// y coord row bottom

					// from top center to right
					bool d1 = getDistance( mask, labels, id1, x, y, xrr, yrt, distance, _id2, _x1, _y1, _x2, _y2 );
					// from top center to left
					bool d2 = getDistance( mask, labels, id1, x, y, xrl, yrt, distance, _id2, _x1, _y1, _x2, _y2 );
					// from bottom center to right
					bool d3 = getDistance( mask, labels, id1, x, y, xrr, yrb, distance, _id2, _x1, _y1, _x2, _y2 );
					// from bottom center to left
					bool d4 = getDistance( mask, labels, id1, x, y, xrl, yrb, distance, _id2, _x1, _y1, _x2, _y2 );

					// columns

					const unsigned int xcr = x+border;		// x coord column right
					const unsigned int xcl = x-border;		// x coord column left
					const unsigned int yct = y-border_i;	// y coord column top
					const unsigned int ycb = y+border_i;	// y coord column bottom

					// from top center to right
					bool d5 = getDistance( mask, labels, id1, x, y, xcr, yct, distance, _id2, _x1, _y1, _x2, _y2 );
					// from top center to left
					bool d6 = getDistance( mask, labels, id1, x, y, xcl, yct, distance, _id2, _x1, _y1, _x2, _y2 );
					// from bottom center to right
					bool d7 = getDistance( mask, labels, id1, x, y, xcr, ycb, distance, _id2, _x1, _y1, _x2, _y2 );
					// from bottom center to left
					bool d8 = getDistance( mask, labels, id1, x, y, xcl, ycb, distance, _id2, _x1, _y1, _x2, _y2 );
					
					if( d1 || d2 || d3 || d4 || d5 || d6 || d7 || d8 ){ // if new minimal distance found
						d = true;
						_d = true;
						break;
					}
				}
			}
		}

		if( _d )
			return true;
		else
			return false;

	}


} ; /*namespace hekate */


#endif /* #ifndef _COMPLETEMASKSEGMENT_HXX_ */