/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file ConvertVecToCImg.hxx
 *  Implementation of the parameter class ConvertVecToCImg.
 *  \author Moritz Becker
 *  \date 02.03.2011
 */

#ifndef _CONVERTVECTOCIMG_HXX_
#define _CONVERTVECTOCIMG_HXX_

#include <hekate/ConvertVecToCImg.h>
#include <charon-utils/ImgTool.hxx>

template <typename T>
ConvertVecToCImg<T>::ConvertVecToCImg(const std::string& name) :
		TemplatedParameteredObject<T>(
				"ConvertVecToCImg", name,
				"Converts 3D and 2D vector containers into CImgLists and the other way around."
		)
			,points3DVec(true, false) 
			,points3DCImg(true, false) 
			,points2DVec(true, false) 
			,points2DCImg(true, false) 
			,pointsStereo(true, false) 
			
			
{

	ParameteredObject::_addInputSlot(points3DVec, "points3DVec", 
	"3D vector which will be converted to a 3D CImgList", 
	"vector<vector<T> >"); 
	ParameteredObject::_addInputSlot(points3DCImg, "points3DCImg", 
	"3D CImgList which will be converted to a 3D vector", 
	"CImgList<T>"); 
	ParameteredObject::_addInputSlot(points2DVec, "points2DVec", 
	"2D vector which will be converted to a 2D CImgList", 
	"vector<vector<vector<T> > >"); 
	ParameteredObject::_addInputSlot(points2DCImg, "points2DCImg", 
	"2D CImgList which will be converted to a 2D vector", 
	"CImgList<T>");  
	ParameteredObject::_addInputSlot(pointsStereo, "pointsStereo", 
	"Stereo disp image which contains rectified correspondences as CImg", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(points3DCImgOut, "points3DCImgOut", 
	"Converted 3D CImgList from the 3D input vector", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(points3DVecOut, "points3DVecOut", 
	"Converted 3D vector from the 3D input CImgList", 
	"vector<vector<T> >"); 
	ParameteredObject::_addOutputSlot(points2DCImgOut, "points2DCImgOut", 
	"Converted 2D CImgList from the 2D input vector", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(points2DVecOut, "points2DVecOut", 
	"Converted 2D vector from the 3D input CImgList", 
	"vector<vector<vector<T> > >"); 
	ParameteredObject::_addOutputSlot(pointsStereoOut, "pointsStereoOut", 
	"Stereo disp image which contains rectified correspondences as std::vector", 
	"vector<vector<vector<T> > >");

	ParameteredObject::_setTags("Hekate") ;
	
}

template <typename T>
void ConvertVecToCImg<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	//convert stereo cimg of rectified correspondences to vectors
	if(this->pointsStereo.connected()){

		const cimg_library::CImg<T> &ptStereo = this->pointsStereo()[0];

		sout << "Converting Stereo CImgList to std::vector<...>:" << std::endl;
		sout << "Loaded correspondence list consist of " <<  ptStereo.size() <<
			" correspondences." << std::endl;

		this->pointsStereoOut().resize(2);
		
		for(int i=0; i<ptStereo.width(); i++){
			for(int j=0; j<ptStereo.height(); j++){
				std::vector<T> pt1;
				pt1.push_back(T(i));
				pt1.push_back(T(j));
				pt1.push_back(T(1));
				std::vector<T> pt2;
				pt2.push_back(T(i)+ptStereo(i,j));
				pt2.push_back(T(j));
				pt2.push_back(T(1));
				this->pointsStereoOut()[0].push_back(pt1);
				this->pointsStereoOut()[1].push_back(pt2);
			}
		}

		sout << "Stereo correspondence size: " << this->pointsStereoOut().size() << std::endl;

	}

	//convert cimg list 2d correspondences to vectors
	if(this->points2DCImg.connected()){

		const cimg_library::CImgList<T> &pt2DCImg = this->points2DCImg();

		sout << "Converting 2D CImgList correspondences to std::vector<...>:" << std::endl;
		sout << "Loaded correspondence list consist of " <<  pt2DCImg.size() <<
			" feature lists with each of " << pt2DCImg[0].width() <<
			" points" << std::endl;

		this->points2DVecOut().resize(pt2DCImg.size());

		for(unsigned int i=0; i<pt2DCImg.size(); i++){
			for(int j=0; j<pt2DCImg[i].width(); j++){
				std::vector<T> pt;
				pt.push_back(pt2DCImg[i](j,0));
				pt.push_back(pt2DCImg[i](j,1));
				pt.push_back(pt2DCImg[i](j,2));
				this->points2DVecOut()[i].push_back(pt);
			}
		}
	}

	//convert cimg list 3d points to vectors
	if(this->points3DCImg.connected()){
		sout << "Converting 3D CImgList points to std::vector." << std::endl;
		const cimg_library::CImgList<T> &pt3DCImg = this->points3DCImg();
		sout << pt3DCImg.size() << " points were found." << std::endl;
		for(size_t i=0; i<pt3DCImg.size(); i++){
			std::vector<T> pt1;
			pt1.push_back(pt3DCImg[i][0]);
			pt1.push_back(pt3DCImg[i][1]);
			pt1.push_back(pt3DCImg[i][2]);
			this->points3DVecOut().push_back(pt1);
		}
	}

	//convert 2d vector correspondences to cimg
	if(this->points2DVec.connected()){

		sout << "Converting 2D correspondence vectors to CImgList." << std::endl;

		// Iterate for all feature lists
		for(size_t i=0; i<this->points2DVec().size(); i++){

			typename std::vector< std::vector<T> >::const_iterator pointIt =
				this->points2DVec()[i].begin();
			const typename std::vector< std::vector<T> >::const_iterator pointEnd =
				this->points2DVec()[i].end();

			cimg_library::CImg<T>featureList;
			for(; pointIt != pointEnd; pointIt++){
				cimg_library::CImg<T> pt(1,3);
				pt[0] = (*pointIt)[0];
				pt[1] = (*pointIt)[1];
				pt[2] = (*pointIt)[2];
				featureList.append(pt);
			}
			
			this->points2DCImgOut().push_back(featureList);
		}
	}

	//convert 3d vector points to cimg
	if(this->points3DVec.connected()){

		sout << "Converting 3D points vectors to CImgList." << std::endl;

		typename std::vector< std::vector<T> >::const_iterator pointIt =
			this->points3DVec().begin();
		const typename std::vector< std::vector<T> >::const_iterator pointEnd =
			this->points3DVec().end();

		for(; pointIt != pointEnd; pointIt++){
			cimg_library::CImg<T> pt(1,3);
			pt[0] = (*pointIt)[0];
			pt[1] = (*pointIt)[1];
			pt[2] = (*pointIt)[2];
			this->points3DCImgOut().push_back(pt);
			//ImgTool::printInfo(sout,pt,"pt");
		}
	}

}

#endif /* _CONVERTVECTOCIMG_HXX_ */



