/*  Copyright (C) 2011 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _CORRVIEW_H_
#define _CORRVIEW_H_

#include <charon-utils/CImg.h>
#include <vector>
#include <libsvm/svm.h>

/// Correspondence view which can be used to learn correspondences
/** 
	The CorrView class can learn a correspondence view from 
	a reference feature list and a putative corresponding list.
*/
template <typename T>
class CorrView{

public:
	/// Creates a correspondence view object
	CorrView();
	/// Creates a correspondence view object by two feature lists
	/** Creates a correspondence view object by two feature lists
		and learns a the correspondece view base on both lists.
		Then the standard deviation between training data and 
		learned data is calculated
	*/
	CorrView(const std::vector<std::vector<T> > &corrRef,
		const std::vector<std::vector<T> > &corrPutative);
	~CorrView();

	/// learns model by reference and traning correspondence
	bool learnView(const std::vector<std::vector<T> > &corrRef,
		const std::vector<std::vector<T> > &corrPutative);

	/// Predicts a correspondence view for given feature points
	std::vector<std::vector<T> > predictView(const std::vector<std::vector<T> > &corrPredict);

	/// Returns the standard deviation in x direction
	double sigmaX();
	/// Returns the standard deviation in y direction
	double sigmaY();
	/// Returns the standard deviation as  vector
	std::vector<double> sigma();
	/// Returns number of feature points of that view
	int size();

	/// Returns reference feature list
	std::vector<std::vector<T> > corrRef();
	/// Returns putative corresponding list
	std::vector<std::vector<T> > corrPutative();
	/// Returns learned correspondence list
	std::vector<std::vector<T> > corrLearned();
	
private:
	double sigX;
	double sigY;
	size_t sizeCorr;
	std::vector<std::vector<T> > cRef;
	std::vector<std::vector<T> > cPutative;
	std::vector<std::vector<T> > cLearned;
	bool learned;

	// Calculates the standard deviation of the error function between
	// putative correspondences and the predicted correspondences
	// of the correspondence function
	bool getSigma();
	
	/// Creates an SVM model
	void createSVMModelX();
	
	/// Creates an SVM model
	void createSVMModelY();

	/// parameters for the learning algorithm
	struct svm_parameter param;
	
	struct svm_problem probX;
	struct svm_node *x_spaceX;
	struct svm_problem probY;
	struct svm_node *x_spaceY;

	/// Correspondence view model
	struct svm_model *modelX;
	struct svm_model *modelY;
	

};



#endif // _CORRVIEW_H_
