/*  Copyright (C) 2011 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <hekate/CorrView.h>

template <typename T>
CorrView<T>::CorrView():learned(0),sigX(0),sigY(0),sizeCorr(0){

	// default values for svm parameters
	param.svm_type = EPSILON_SVR; // OR NU_SVR
	param.kernel_type = RBF;
	param.degree = 3;
	param.gamma = 0.00000009765625;	// 1/num_features
	param.coef0 = 0;
	param.nu = 0.5;
	param.cache_size = 100;
	param.C = 512;
	param.eps = 1e-3;
	param.p = 0.25;
	param.shrinking = 1;
	param.probability = 0;
	param.nr_weight = 0;
	param.weight_label = NULL;
	param.weight = NULL;

}

template <typename T>
CorrView<T>::CorrView(const std::vector<std::vector<T> > &corrRef,
		const std::vector<std::vector<T> > &corrPutative) :
			sigX(0),sigY(0),sizeCorr(0),learned(0) {
	// default values for svm parameters
	param.svm_type = EPSILON_SVR; // OR NU_SVR
	param.kernel_type = RBF;
	param.degree = 3;
	param.gamma = 0.00000009765625;	// 1/num_features
	param.coef0 = 0;
	param.nu = 0.5;
	param.cache_size = 100;
	param.C = 512;
	param.eps = 1e-3;
	param.p = 0.25;
	param.shrinking = 1;
	param.probability = 0;
	param.nr_weight = 0;
	param.weight_label = NULL;
	param.weight = NULL;

	learnView(corrRef, corrPutative);
}

template <typename T>
bool CorrView<T>::learnView(const std::vector<std::vector<T> > &corrRef,
						 const std::vector<std::vector<T> > &corrPutative){
							 
	// Clear corrRef and corrPutative
	this->cRef.clear();
	this->cPutative.clear();
	this->cLearned.clear();

	// Read in new putative correspondence
	this->cRef = corrRef;
	this->cPutative = corrPutative;
	
	this->sizeCorr = corrRef.size();
	if(sizeCorr != corrPutative.size())
		throw(std::runtime_error("Inconsistent sizes of correspondence vectors."));
	

	// Train models -----------------------------------------
	createSVMModelX();
	createSVMModelY();
	

	// Predict correspondences ------------------------------
	this->cLearned = predictView(corrRef);	


	// If view could have been learned
	if(getSigma()){
		this->learned = true;
		return 1;
	}
	else{
		sout << "Sigma could not be calculated" << std::endl;
		return 0;
	}
	
}


template <typename T>
std::vector<std::vector<T> > CorrView<T>::predictView(const std::vector<std::vector<T> > &corrPredict){
	

	size_t sizeCorr = corrPredict.size();
	std::vector<std::vector<T> > corrOutput;
	
	// Predict correspondences
	struct svm_node *feat;
	feat = (struct svm_node*) malloc(3*sizeof(struct svm_node));


	for(size_t i=0; i<sizeCorr; i++){

		// Read in correspondence
		feat[0].index = 0;
		feat[0].value = corrPredict[i][0];
		feat[0+1].index = 1;
		feat[0+1].value = corrPredict[i][1];
		feat[0+2].index = -1;

		
		// Predict correspondence i
		double predictX = svm_predict(modelX, feat);
		double predictY = svm_predict(modelY, feat);


		//sout << feat[0].value << ", " << feat[0+1].value << " --> ";
		//sout << predictX << ", " << predictY << std::endl;
		

		// Output predicted values
		std::vector<T> featPredict(3);
		featPredict[0] = T(predictX);
		featPredict[1] = T(predictY);
		featPredict[2] = T(1);
		corrOutput.push_back(featPredict);

	}
			
	free(feat);

	return corrOutput;

}


template <typename T>
bool CorrView<T>::getSigma(){
	
	if(this->cPutative.size() != this->cLearned.size()){
		throw(std::runtime_error("Inconsistent sizes of putative and learned correspondences."));
		return 0;
	}
	else if(this->cLearned.size()==0){
		throw(std::runtime_error("No learned correspondences available."));
		return 0;
	}
	else{

		double sigma1=0, sigma2=0;
		int n = this->cLearned.size();

		if(n == 0)
			return 0;

		// Iterate over all correspondences
		for( int i=0; i<n; i++ ){
			sigma1 += pow(double(this->cPutative[i][0]-this->cLearned[i][0]),2.0);
			sigma2 += pow(double(this->cPutative[i][1]-this->cLearned[i][1]),2.0);
		}

		sigma1 /= double(n);
		sigma2 /= double(n);

		sigma1 = pow(sigma1,0.5);
		sigma2 = pow(sigma2,0.5);

		this->sigX = sigma1;
		this->sigY = sigma2;

		return 1;
	}
}

template <typename T>
double CorrView<T>::sigmaX(){
	if(this->learned)
		return this->sigX;
	else
		return 0;
}

template <typename T>
double CorrView<T>::sigmaY(){
	if(this->learned)
		return this->sigY;
	else
		return 0;
}

template <typename T>
std::vector<double> CorrView<T>::sigma(){
	std::vector<double> sig;
	if(this->learned){
		sig.push_back(sigX);
		sig.push_back(sigY);
		return sig;
	}
	else{
		sig.push_back(T(0));
		sig.push_back(T(0));
		return sig;
	}
}

template <typename T>
int CorrView<T>::size(){
	return sizeCorr;
}

template <typename T>
std::vector<std::vector<T> > CorrView<T>::corrRef(){
	return this->cRef;
}

template <typename T>
std::vector<std::vector<T> > CorrView<T>::corrPutative(){
	return this->cPutative;
}

template <typename T>
std::vector<std::vector<T> > CorrView<T>::corrLearned(){

	// Throw error if the correspondence view has not been learned
	if(!this->learned){
		throw(std::runtime_error("Correspondences have to "
			"be learned before you can access to them."));
	}

	return this->cLearned;
}

template <typename T>
void CorrView<T>::createSVMModelX(){

	//sout << "Train x model (x,y) -> x'" << std::endl;

	// Free space of existing models
	if(this->learned){
		//sout << "Clear old model." << std::endl;
		svm_free_model_content(this->modelX);
		free(this->probX.y);
		free(this->probX.x);
		free(this->x_spaceX);
	}


	// Number of correspondence sets
	this->probX.l = cRef.size();

	// Number of features from which the model is created
	// (number of features + 1) * number of feature sets.
	const int elements = this->probX.l * 3;
	

	// Allocate prob
	this->probX.x = (struct svm_node **) malloc((this->probX.l) * sizeof(struct svm_node *));
	this->probX.y = (double *) malloc((this->probX.l) * sizeof(double));
	this->x_spaceX = (struct svm_node *) malloc(elements*sizeof(struct svm_node));
	
	// Read in problem
	for(int i=0; i<this->probX.l; i++){

		// target value
		this->probX.y[i] = cPutative[i][0];

		// training vector
		this->probX.x[i] = &this->x_spaceX[i*3];


		// create svm_node of training vector i
		this->x_spaceX[i*3].index = 0; // feature 1 index
		this->x_spaceX[i*3].value = cRef[i][0]; // feature 1 value
		this->x_spaceX[i*3+1].index = 1; // feature 2 index
		this->x_spaceX[i*3+1].value = cRef[i][1]; // feature 2 value
		this->x_spaceX[i*3+2].index = -1; // end of training vector
		
		//sout << this->x_spaceX[i*3].value << ", " << this->x_spaceX[i*3+1].value;
		//sout << " --> " << this->probX.y[i] << std::endl;
		

	}
		
	this->modelX = svm_train(&this->probX,&param);	

	const char *error_msg;
	error_msg = svm_check_parameter(&this->probX,&param);
	if(error_msg)
	{
		sout << printf ("Error: %s\n",error_msg) << std::endl;
	}
		
}

template <typename T>
void CorrView<T>::createSVMModelY(){

	//sout << "Train y model (x,y) -> y'" << std::endl;

	// Free space of existing models
	if(this->learned){
		//sout << "Clear old model." << std::endl;
		svm_free_model_content(this->modelY);
		free(this->probY.y);
		free(this->probY.x);
		free(this->x_spaceY);
	}


	// Number of correspondence sets
	this->probY.l = cRef.size();

	// Number of features from which the model is created
	// (number of features + 1) * number of feature sets.
	const int elements = this->probY.l * 3;
	

	// Allocate prob
	this->probY.x = (struct svm_node **) malloc((this->probY.l) * sizeof(struct svm_node *));
	this->probY.y = (double *) malloc((this->probY.l) * sizeof(double));
	this->x_spaceY = (struct svm_node *) malloc(elements*sizeof(struct svm_node));
	
	// Read in problem
	for(int i=0; i<this->probY.l; i++){

		// target value
		this->probY.y[i] = cPutative[i][1];

		// training vector
		this->probY.x[i] = &this->x_spaceY[i*3];


		// create svm_node of training vector i
		this->x_spaceY[i*3].index = 0; // feature 1 index
		this->x_spaceY[i*3].value = cRef[i][0]; // feature 1 value
		this->x_spaceY[i*3+1].index = 1; // feature 2 index
		this->x_spaceY[i*3+1].value = cRef[i][1]; // feature 2 value
		this->x_spaceY[i*3+2].index = -1; // end of training vector
		
		//sout << this->x_spaceY[i*3].value << ", " << this->x_spaceY[i*3+1].value;
		//sout << " --> " << this->probY.y[i] << std::endl;
		

	}
		
	this->modelY = svm_train(&this->probY,&param);	

	const char *error_msg;
	error_msg = svm_check_parameter(&this->probY,&param);
	if(error_msg)
	{
		sout << printf ("Error: %s\n",error_msg) << std::endl;
	}
		
}


/// Remove outlier from correspondence set
template <typename T>
void removeOutliers(std::vector<std::vector<T> > &corr1,
	std::vector<std::vector<T> > &corr2){

	std::vector<std::vector<T> > c1, c2;
	
	for(size_t i=0; i<corr1.size(); i++){
		if(checkPt(corr1[i]) && checkPt(corr2[i]))
			c1.push_back(corr1[i]);
			c2.push_back(corr2[i]);
	}
	
	corr1.clear();
	corr2.clear();
	corr1 = c1;
	corr2 = c2;

}


/// Checks if point is an outlier, if yes the function returns false
template <typename T>
bool checkPt(std::vector<T> point){
	if(point[0] == 0 && point[1] == 0 && point[2] == 0)
		return false;
	else
		return true;
}


template <typename T>
CorrView<T>::~CorrView(){

	//sout << "Destroying CorrView." << std::endl;

	svm_free_and_destroy_model(&modelX);
	svm_free_and_destroy_model(&modelY);
	svm_destroy_param(&param);
	
	free(probX.x);
	free(probX.y);
	free(probY.x);
	free(probY.y);
	free(x_spaceX);
	free(x_spaceY);

}

