/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file correctCorrespondences.hxx
 *  Implementation of the parameter class correctCorrespondences.
 *  \author Moritz Becker
 *  \date 22.10.2010
 */

#ifndef _CORRECTCORRESPONDENCES_HXX_
#define _CORRECTCORRESPONDENCES_HXX_

#include <hekate/CorrectCorrespondences.h>
#include <charon-utils/ImgTool.hxx> 
#include <vigra/polynomial.hxx>

template <typename T>
correctCorrespondences<T>::correctCorrespondences(const std::string& name) :
		TemplatedParameteredObject<T>(
				"correctCorrespondences", name,
				"Corrects correspondences by the epipolar constraint; The points are shifted to the nearest point of the epipolar line. (See Hartley p 318)"
		)
			,inputCorrespondences(false, false) 
			
			
{
	ParameteredObject::_addParameter (maxDistance, "maxDistance", 
	"Maximal allowed distance of the corrected point from the original point.", T(1));
	ParameteredObject::_addInputSlot(inputCorrespondences, "inputCorrespondences", 
	"Input correspondences", 
	"CorrespondenceList"); 
	ParameteredObject::_addInputSlot(fmatrix, "fundamentalMatrices", "Fudamental matrices.", "IndexedMatrixList"); 
	ParameteredObject::_addOutputSlot(outputCorrespondences, "outputCorrespondences", 
	"Output correspondences", 
	"CorrespondenceList"); 

	ParameteredObject::_setTags("Hekate") ;
	
}

template <typename T>
void correctCorrespondences<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	const CorrespondenceList<T> corrs = this->inputCorrespondences();
	size_t corrSizeImg = corrs.sizeImages() ;
	size_t corrSizePt = corrs.sizePoints() ;
	sout << "Processing " << corrSizeImg << " images overall" << std::endl;
	sout << "Processing " << corrSizePt << " points overall" << std::endl;
	
	// Read in fundamental matrices
	const IndexedMatrixList &fm = this->fmatrix();
	
	int itCount = 0;
	int count = 0;
	int countOut = 0;
	int countNoDist = 0;
	size_t countStatus = 0;

	// Loop over all correspondences
	auto it = corrs.begin() ;
	auto itEnd = corrs.end() ;
	for(; it != itEnd; it++)
	{

		itCount++;
		if( itCount % 100 == 0 )
			sout << "Processing point " << itCount << std::endl;
		
		// Get vector of points with current point index
		const std::vector<const Correspondence<T>* > pointlist = corrs.getListByPointId(it->pointIndex) ;
						
		for(size_t i=0; i < pointlist.size(); i++)
		{

			//sout << "Checking points in image " << i << std::endl;

			for(size_t j=i; j < pointlist.size(); j++)
			{
				
				// Read in fundamental matrix
				CImg<double> F;
				try{
					if(fm(pointlist[i]->imgIndex, pointlist[j]->imgIndex).degenerated)
					{
						continue;
					}
					else
					{
						F = fm(pointlist[i]->imgIndex, pointlist[j]->imgIndex);
					}
				}
				catch(...)
				{
					continue;
				}
								
				// If one of both points is an outlier, 
				// continue with next interation
				if(pointlist[i]->outlier || pointlist[j]->outlier)
				{
					continue;
				}

				count++;
				
				// Read in correspondence
				CImg<T> pt1 = pointlist[i]->pt2d() ;
				CImg<T> pt2 = pointlist[j]->pt2d() ;
				pt1 /= pt1[2];
				pt2 /= pt2[2];

				// Run algorithm from Hartley p. 318

				// corrected coordinates
				CImg<double> _x1 = pt1, _x2 = pt2 ;

				//transformation matrix T
				CImg<T> T1(3,3), T2(3,3);
				T1.fill(0);  
				T1(0,0)=1;						T1(2,0)=-pt1(0,0);
								T1(1,1)=1;		T1(2,1)=-pt1(0,1);	
												T1(2,2)=1;
		
				//transformation matrix T'
				T2.fill(0);  
				T2(0,0)=1;						T2(2,0)=-pt2(0,0);	
								T2(1,1)=1;		T2(2,1)=-pt2(0,1);
												T2(2,2)=1;

				
				// transform f-matrix -> _F2 = T2^(-T)*f-matrix*T1^(-1)
				F = T2.get_transpose().invert() * F * T1.get_invert() ;

				//calculate epipoles
				cimg_library::CImg<double> _e1(1,3), _e2(1,3), _e12d, _e22d, fe1, e2tf;			
				getEpipoles(F, _e1) ;
				getEpipoles(F.get_transpose(), _e2) ;
				
				//_e12d = _e1.get_crop(0,0,0,1);
				//_e22d = _e2.get_crop(0,0,0,1);

				//fe1 = F*_e1;
				//e2tf = _e2.get_transpose() * F;

				//ImgTool::printInfo(sout, F*_e1," F*_e1------------");
				//ImgTool::printInfo(sout, _e2.get_transpose()*F," _e2.get_transpose()*F------------");
				//sout << "F*_e1 = " << fe1[0]+fe1[1]+fe1[2] << std::endl;
				//sout << "_e2^T*F = " << e2tf[0]+e2tf[1]+e2tf[2] << std::endl;
				//sout << _e12d.dot(_e12d) << std::endl;
				//sout << _e22d.dot(_e22d) << std::endl;

				
				//check if epipoles can be determined
				if((_e1[0] == 0 && _e1[1]==0) || (_e2[0] == 0 && _e2[1]==0)){		
					sout << "epipole wrong" << std::endl; 
					countOut++;
					continue;
				}

				//determine rotation matrix _R1
				cimg_library::CImg<double> _R1(3,3);
				_R1(0,0)=_e1[0];			_R1(1,0)=_e1[1];		_R1(2,0)=0;
				_R1(0,1)=_e1[1]*-1;			_R1(1,1)=_e1[0];		_R1(2,1)=0;
				_R1(0,2)=0;					_R1(1,2)=0;				_R1(2,2)=1;

				//determine rotation matrix _R2
				cimg_library::CImg<double> _R2(3,3);
				_R2(0,0)=_e2[0];			_R2(1,0)=_e2[1];		_R2(2,0)=0;
				_R2(0,1)=_e2[1]*-1;			_R2(1,1)=_e2[0];		_R2(2,1)=0;
				_R2(0,2)=0;					_R2(1,2)=0;				_R2(2,2)=1;
				
				cimg_library::CImg<double> F2(3,3);
				F2 = _R2 * F * _R1.get_transpose() ;

				//ImgTool::printInfo(sout,F2," F2------------");

				F2(0,0)=_e1[2]*_e2[2]*F2(2,2);
				F2(1,0)=-_e2[2]*F2(1,2);
				F2(2,0)=-_e2[2]*F2(2,2);
				F2(0,1)=-_e1[2]*F2(2,1);
				F2(0,2)=-_e1[2]*F2(2,2);
				
				//ImgTool::printInfo(sout,F2," F2------------");

				// solve polynomial g(t) = 0
				std::vector<double> _t ;			
						

				bool minDist = getT(_t, F2, _e1, _e2);
				if(!minDist)
				{				
					//sout << "Couldn't find a minimum distance" << std::endl;
					countNoDist++;
					continue;
				} 
				else
				{
					// determine t with s(t)=min(s(t))

					double _tmin = _t[0] ;
					
					for(size_t j=1; j<_t.size(); j++){
						if(getS(_t[j],F2,_e1,_e2) < getS(_t[j-1],F2,_e1,_e2)) _tmin = _t[j] ;
					}
					
					// left coordinate closest to l=(t*e1,1,-t) and l'=_F2*(0,t,1)^t
					// closest point to origin on a line (x,y,k) is (-xk,-yk,x^2+y^2)
					pt1[0] = _tmin * _tmin * _e1[2] ;
					pt1[1] = _tmin ;
					pt1[2] = _tmin * _tmin * _e1[2] * _e1[2] +1 ;
					if(pt1(0,2) != 0 ) pt1 /= pt1(0,2);
					
					// right corrdinate closest to l'
					cimg_library::CImg<double> lHelp(1,3), lHelp2(1,3);
					lHelp(0,0) = 0;
					lHelp(0,1) = _tmin;
					lHelp(0,2) = 1;
					lHelp2 = F2 * lHelp;
					pt2(0,0) = - lHelp2[0] * lHelp2[2] ;
					pt2(0,1) = - lHelp2[1] * lHelp2[2] ;
					pt2(0,2) =   lHelp2[0] * lHelp2[0] + lHelp2[1] * lHelp2[1];
					if(pt2(0,2) != 0 ) pt2 /= pt2(0,2);
									
					// transfer coordinates back to original ones
					pt1 = T1.get_invert() * _R1.get_transpose() * pt1 ;
					pt2 = T2.get_invert() * _R2.get_transpose() * pt2 ;

					pt1 /= pt1[2];
					pt2 /= pt2[2];

					double dist1 = sqrt((double)((pt1[0]-_x1[0])*(pt1[0]-_x1[0])+(pt1[1]-_x1[1])*(pt1[1]-_x1[1])));
					double dist2 = sqrt((double)((pt2[0]-_x2[0])*(pt2[0]-_x2[0])+(pt2[1]-_x2[1])*(pt2[1]-_x2[1])));
					/*
					sout << "new(" << pt1(0,0) << ", " 	<< pt1(0,1) << ") / old(" 
					<< _x1(0,0) << ", " << _x1(0,1) << ") distance: " << dist1 << std::endl;

					sout << "new(" << pt2(0,0) << ", " 	<< pt2(0,1) << ") / old(" 
					<< _x2(0,0) << ", " << _x2(0,1) << ") distance: " << dist2 << std::endl;
					*/
					// Check if distance to original point is less than max distance.
					if(dist1 >= maxDistance() || dist2 >= maxDistance){
						//sout << "Distance of new points taller than max Distance -> OUTLIER!" << std::endl;
						//const_cast<Correspondence<T>* >(pointlist[i])->outlier = true; 
						//const_cast<Correspondence<T>* >(pointlist[j])->outlier = true; 
						countOut ++;
					}
					else{
						pointlist[i]->x() = pt1[0];
						pointlist[i]->y() = pt1[1];
						pointlist[i]->w() = pt1[2];
						pointlist[j]->x() = pt2[0];
						pointlist[j]->y() = pt2[1];
						pointlist[j]->w() = pt2[2];
						//const_cast<Correspondence<T>* >(pointlist[i])->outlier = false; 
						//const_cast<Correspondence<T>* >(pointlist[j])->outlier = false; 
					}
											
				}

			}

		}

	}
						
	this->outputCorrespondences() = corrs ;
	sout << count << " matches have been tried to correct" << std::endl;
	sout << countOut << " matches seems to be bad" << std::endl;
	sout << countNoDist << " matches has not been optimized since "
		<< "minimum distances could not be found." << std::endl;

}

/// Calculate total squared distance S defined in Hartley et al, Multiple View Geometry, 317.
template <typename T>
double correctCorrespondences<T>::getS(double &t, CImg<double> &F, 
			CImg<double> &e1, CImg<double> &e2){

	double _s;
	_s = t*t / (1+e1[2]*e1[2]*t*t) + pow(F(1,2)*t+F(2,2), 2.) 
		/ (pow(F(1,1)*t+F(2,1), 2.)+e2[2]*e2[2]*pow(F(1,2)*t+F(2,2), 2.)) ;
	return _s ;

}

/// Calculates t_min which delivers the minimum of the cost function g(t).
template <typename T>
bool correctCorrespondences<T>::getT(std::vector<double> &t, CImg<double> &F, 
			CImg<double> &_e1, CImg<double> &_e2){
							
	vigra::Polynomial<double> poly(6);

	double a=F(1,1); double b=F(2,1); 
	double c=F(1,2); double d=F(2,2);
	double e1=_e1[2]; double e2=_e2[2];
	
	double a2=pow(a,2.), a3=pow(a,3.), a4=pow(a,4.);
	double b2=pow(b,2.), b3=pow(b,3.), b4=pow(b,4.);
	double c2=pow(c,2.), c3=pow(c,3.), c4=pow(c,4.);
	double d2=pow(d,2.), d3=pow(d,3.), d4=pow(d,4.);
	double e12=pow(e1,2.), e13=pow(e1,3.), e14=pow(e1,4.);
	double e22=pow(e2,2.), e23=pow(e2,3.), e24=pow(e2,4.);

	// A[6]*t^6
	poly[6] = a*c*(b*c-a*d)*e14 ;
		
	// A[5]*t^5
	poly[5] = a4+b2*c2*e14+c4*e24+a2*(-d2*e14+2*c2*e22);

	poly[4] = 4*a3*b
		+2*a*c*(b*c-a*d)*e12
		+b*d*(b*c-a*d)*e14
		+4*a*b*c2*e22
		+4*a2*c*d*e22
		+4*c3*d*e24;

	poly[3] = 2*(4*a*b*c*d*e22+a2*(3*b2+d2*(-e12+e22))+c2*(3*d2*e24+b2*(e12+e22))) ;

	poly[2] = -a2*c*d+a*b*(4*b2+c2-2*d2*(e12-2*e22))+2*c*d*(2*d2*e24+b2*(e12+2*e22)) ;

	poly[1] = b4-a2*d2+d4*e24+b2*(c2+2*d2*e22);
	
	// A[0]*t^0
	poly[0] = b*d*(b*c-a*d) ;


    vigra::ArrayVector<double> roots;
    vigra::polynomialRealRoots(poly, roots);

	if(roots.size() == 0)
		return 0;

	for (size_t i = 0; i <  roots.size(); i++) {
		t.push_back(roots[i]) ;
		/*
		double _t = roots[i] ;
		double _p6 = pow(_t,6.) * poly[6] ;
		double _p5 = pow(_t,5.) * poly[5] ;
		double _p4 = pow(_t,4.) * poly[4] ;
		double _p3 = pow(_t,3.) * poly[3] ;
		double _p2 = pow(_t,2.) * poly[2] ;
		double _p1 = pow(_t,1.) * poly[1] ;
		double _p0 = poly[0] ;
		
		sout << "Root " << i <<": g(root) = " << _p6+_p5+_p4+_p3+_p2+_p1+_p0 << std::endl;
		*/
	}

	return 1;

}
 
/// Calculate epipoles from fundamental matrix.
template <typename T>
void correctCorrespondences<T>::getEpipoles(
		const CImg<double>& fmatrix,
		CImg<double>& epipole){

	// SVD methode
	CImg<double> U2, D2, V2;
	fmatrix.SVD(U2, D2, V2);

	// solution is right column of V
	epipole=V2.get_crop(2,0,2,2);
	if(epipole[2]!=0) epipole/=epipole[2];
			
	if(epipole[0]!=0 || epipole[1]!=0){
		double c = pow(pow((double)epipole[0],2.) + pow((double)epipole[1],2.), -0.5) ;
		epipole *= c;
	}
	else sout << "ATTENTION - Epipole equal (0, 0, "
		<< epipole[2] << ")" << std::endl;
	
}

#endif /* _CORRECTCORRESPONDENCES_HXX_ */



