/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CorrectCorrespondencesOpenCV.hxx
 *  Implementation of the parameter class CorrectCorrespondencesOpenCV.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.08.2012
 */

#ifndef _CORRECTCORRESPONDENCESOPENCV_HXX_
#define _CORRECTCORRESPONDENCESOPENCV_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CorrectCorrespondencesOpenCV.h"
#include <opencv2/core/core.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
#include <hekate/openCvUtil.hxx>

template <typename T>
CorrectCorrespondencesOpenCV<T>::CorrectCorrespondencesOpenCV(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CorrectCorrespondencesOpenCV", name,
			"<h2>Corrects correspondences by epipolar geometry</h2><br>"
			"Corrects correspondences by epipolar geometry using the "
			"fundamental matrices."
		), initialFrames(1,0)
{
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"Correspondences to correct.",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		fundamentalMatrix, "fundamentalMatrix",
		"Fundamental Matrices for epipolar geometry",
		"IndexedMatrixList");
	ParameteredObject::_addInputSlot(
		initialFrames, "initialFrames",
		"Initial frames (optional)",
		"vector<int>");

	ParameteredObject::_addOutputSlot(
		correctedCorrespondences, "correctedCorrespondences",
		"Corrected correspondences",
		"CorrespondenceList");

	ParameteredObject::_addParameter< int >(
		initFrame1, "initFrame1",
		"initFrame1",
		0, "int");
	ParameteredObject::_addParameter< int >(
		initFrame2, "initFrame2",
		"initFrame2",
		1, "int");
	ParameteredObject::_addParameter< int >(
		initFrame3, "initFrame3",
		"initFrame3",
		2, "int");

	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void CorrectCorrespondencesOpenCV<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	const hekate::CorrespondenceList<T> &corrs = correspondences();

	sout << "Correcting " << corrs.size() 
		<< " correspondences by OpenCV::correctMatches(...)" << std::endl;

	// Two possibilities:
	// Correct all correspondences and at the end again the initial frames (since 
	//		different correction iteration can modify already corrected matches
	//		and initial frame matches are more important than other ones
	// OR only correct matches of initial frames.

	// Convert CorrespondenceList to point vector for OCV

	for(size_t i=0; i < fundamentalMatrix().size(); i++){

		// get fundamental matrix and according image ids
		cv::Mat fmatrix(3,3,CV_32FC1);
		fmatrix = openCvUtil::CImg2Mat<float>(fundamentalMatrix()[i]);
		int frameId1 = fundamentalMatrix()[i].getIndex1() ;
		int frameId2 = fundamentalMatrix()[i].getIndex2() ;

		// create cv vector with correspondences
		std::vector< cv::Point2d > corr1, corr2, corr1out, corr2out;
		std::vector< int > corrId;
		std::vector< Correspondence<T> > corr1src, corr2src;

		auto it = corrs.get<imgIndex>().lower_bound(frameId1) ;
		auto itEnd = corrs.get<imgIndex>().upper_bound(frameId1) ;
		for(; it != itEnd ; it++ ){

			const Correspondence<T> &pt1 = *it;

			if( pt1.outlier )
				continue;

			try{

				// Try to find correspondence for pt1
				const Correspondence<T> &pt2 = corrs( frameId2, pt1.pointIndex );

				if( pt2.outlier)
					continue;

				// Save correspondences in cv vector
				corr1.push_back( cv::Point2d( pt1.x(), pt1.y() ) );
				corr2.push_back( cv::Point2d( pt2.x(), pt2.y() ) );
				corrId.push_back( pt1.pointIndex );
				corr1src.push_back( pt1 );
				corr2src.push_back( pt2 );

			}

			catch(...){
				continue;
			}

		}

		// data format for CV function is still wrong
		cv::correctMatches(fmatrix, corr1, corr2, corr1out, corr2out );

		for(size_t j = 0; j < corr1out.size(); j++){
			
			Correspondence<T> c1 = corr1src[j];
			c1.point[0] = corr1out[j].x;
			c1.point[1] = corr1out[j].y;
			Correspondence<T> c2 = corr2src[j];
			c2.point[0] = corr2out[j].x;
			c2.point[1] = corr2out[j].y;
			
			correctedCorrespondences().push_back( c1 );
			correctedCorrespondences().push_back( c2 );

		}

		correctedCorrespondences().removeDoubles();

	}

	sout << correctedCorrespondences().size() << " points corrected" << std::endl;

}

#endif /* _CORRECTCORRESPONDENCESOPENCV_HXX_ */
