/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CorrectDepthEdgeProbability.h
 *  Declaration of the parameter class CorrectDepthEdgeProbability.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 11.12.2012
 */

#ifndef _CORRECTDEPTHEDGEPROBABILITY_H_
#define _CORRECTDEPTHEDGEPROBABILITY_H_

#ifdef _MSC_VER
#ifdef correctdepthedgeprobability_EXPORTS
/// Visual C++ specific code
#define correctdepthedgeprobability_DECLDIR __declspec(dllexport)
#else
#define correctdepthedgeprobability_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else
/// Not needed without MSVC
#define correctdepthedgeprobability_DECLDIR
#endif

#include <charon-core/ParameteredObject.h>
#include <charon-utils/CImg.h>

using namespace cimg_library;

/// Modifies probabilities of depth edges by optical flow and segmentation
/** Modifies probabilities of depth edges by optical flow and segmentation
 */
template <typename T>
class correctdepthedgeprobability_DECLDIR CorrectDepthEdgeProbability :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	CorrectDepthEdgeProbability(const std::string& name = "");

	/// probabilities of depth edges
	InputSlot< cimg_library::CImgList<T> > edgesProbability;
	/// Optical flow of image sequence
	InputSlot< cimg_library::CImgList<T> > opticalFlow;
	/// Masks depth edges of edges CImgList
	InputSlot< cimg_library::CImgList<T> > edgesMask;

	/// Adjusted input probabilities
	OutputSlot< cimg_library::CImgList<T> > adjustedProbability;
	/// warpedMasks
	OutputSlot< cimg_library::CImgList<T> > warpedMasks;

	/// The edgeMasks need an index according to which frame inside the edges
	/// sequence they belong
	ParameterList< int > edgeMaskIndices;
	
	/// Percental probability threshold. 
	/** All pixels with values smaller than the threshold,
		will be considered for decreasing probability correction.
	*/
	Parameter< float > probThreshold ;
	
	/// Percental probability threshold.
	/** All pixels with values taller than the threshold, 
		will be considered for increasing probability correction. 
	*/
	Parameter< float > probThreshold2 ;
	
	/// Probability threshold which defines a depth edge
	Parameter< float > correctionFactor ;

	/// Use all pixels which are no depth edges to warp the probability
	Parameter< bool > onlyDecreaseProb;

	/// Use all pixels which are no depth edges to warp the probability
	Parameter< bool > isForwardFlow;

	/// Use all pixels which are no depth edges to warp the probability
	Parameter< unsigned int > edgeBorder;


protected:
	/// Update object.
	virtual void execute();

private:
	/// Initialize and check input consistency
	void init();

	/// Get a list of depth edges where only the entries for whose index a mask is given are calculated
	CImgList<T> getWarpMask() ;
	
	/// Warps and image by a given flow, if forward is false, the flow is taken in the opposite direction
	CImg<float> getWarp(const CImg<float> &img, const CImg<float> &flow, bool forward = true);
	
	// Calculate a new probability image using the given mask
	void correctProbability(CImg<float> &probability, const CImg<float> &mask);

};

#endif // _CORRECTDEPTHEDGEPROBABILITY_H_
