/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CorrectDepthEdgeProbability.hxx
 *  Implementation of the parameter class CorrectDepthEdgeProbability.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 11.12.2012
 */

#ifndef _CORRECTDEPTHEDGEPROBABILITY_HXX_
#define _CORRECTDEPTHEDGEPROBABILITY_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CorrectDepthEdgeProbability.h"
#include <charon-utils/ImgTool.hxx>

template <typename T>
CorrectDepthEdgeProbability<T>::CorrectDepthEdgeProbability(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CorrectDepthEdgeProbability", name,
			"<h2>Modifies probabilities of depth edges by optical flow and "
			"segmentation</h2><br>"
			"This module modifies depth edge images using optical flow and ground truth. "
			"Ground truth values are warped through the image sequence and adjust depth edge probabilities pixelwise."
		)
{
	ParameteredObject::_addInputSlot(
		edgesProbability, "edgesProbability",
		"Probabilities images of depth edges",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		edgesMask, "edgesMask",
		"Given ground truth for depth edges. These images are warped by optical flow and modify values of 'edgesProbability'.",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		opticalFlow, "opticalFlow",
		"Optical flow of image sequence",
		"CImgList<T>");
	
	ParameteredObject::_addOutputSlot(
		adjustedProbability, "adjustedProbability",
		"Adjusted input images",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		warpedMasks, "warpedMasks",
		"Warped ground truth images.",
		"CImgList<T>");

	ParameteredObject::_addParameter(
		edgeMaskIndices, "edgeMaskIndices",
		"Indices of the given ground truth according to the image sequence. Since ground truth normally is only given for a few frames, "
		"the indices have to be given to know the according frames of the sequence.");

	ParameteredObject::_addParameter< float >(
		probThreshold, "probThreshold",
		"Percental probability threshold. All pixels with values smaller than the threshold, "
		"will be considered for a decreasing probability correction.",
		0.6, "float");

	ParameteredObject::_addParameter< float >(
		probThreshold2, "probThreshold2",
		"Percental probability threshold. All pixels with values taller than the threshold, "
		"will be considered for increasing probability correction.",
		0.4, "float");

	ParameteredObject::_addParameter< float >(
		correctionFactor, "correctionFactor",
		"Amount of how strong the correction will effect the result.",
		0.5, "float");

	ParameteredObject::_addParameter< unsigned int >(
		edgeBorder, "edgeBorder",
		"Size of border around depth edge, which won't be considered for probability decreasing.", 2);

	ParameteredObject::_addParameter< bool >(
		onlyDecreaseProb, "onlyDecreaseProb",
		"Do not increase the probability for a depth edge.",
		true, "bool");

	ParameteredObject::_addParameter< bool >(
		isForwardFlow, "isForwardFlow",
		"The given flow is a forward flow.",
		true, "bool");

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void CorrectDepthEdgeProbability<T>::execute() {
	
	// Check input consistency
	init();	
	
	CImgList<float> probability = this->edgesProbability();

	const int w = probability[0].width();
	const int h = probability[0].height();

	// Get mask for warping procedure
	/* 
		At first, warping maks are calculated for those frames, which have a given corresponding depth edge mask.
		During the algorithm, the warp masks for the other frames are calculated by warping.
		The pixels of the warping masks are set to 1, if they shall negatively effect the probability, otherwise they are 0.
	*/
	CImgList<float> warpMask = getWarpMask();

	// Iterate over number of given maks indices
	for( size_t j=0; j < this->edgeMaskIndices().size() ; j++)
	{

		const int index = this->edgeMaskIndices()[j] ;

		sout << "Correct with warp mask " << index << std::endl;
		
		/* 
		
			1) Lese masken und kanten bilder, bestimme texturemasken bzw !depthEdges
			2) warpe die initial masks mit optischem fluss auf die anderen bilder. wenn mehr als 1 maske vorhanden ist, warpen aus verschiedenen richtungen
			3) beeinflusse wahrscheinlichkeitsbilder aufgrund der gewarpten wahrscheinlichkeiten.
		
		*/

		// Adjust probability edges of current frame
		correctProbability(probability[index],warpMask[index]);
			
		// Warp texture edges through image sequence in positive frame direction
		// Begin at key frame index and stop, if new keyframe is reached
		for( size_t i=index+1; i < probability.size() ; i++ )
		{

			// Stop iteration, if next key frame - 1 is reached
			if( i == this->edgeMaskIndices()[std::min(int(j+1),int(this->edgeMaskIndices().size()-1))] - 1 )
				break;

			sout << "\tprocess image " << i << " (forward flow)" << std::endl;
			
			// Optical flow
			const CImg<float> &flow = this->opticalFlow()[i-1];
			assert(flow.width() == w && flow.height() == h );

			// probability of depth edges
			assert(probability[i].width() == w && probability[i].height() == h );

			// Get warp mask[ i+1 ]		
			CImg<float> warp = getWarp(warpMask[i-1], flow, this->isForwardFlow() );

			// Save warped mask, if next frame is no key frame
			warpMask[i] = warp ;
			sout << warp.sum() << std::endl;
			sout << warp.max() << std::endl;

			// Adjust probability edges of next frame by warped edges
			correctProbability(probability[i],warp);
			
		}
		
		// Warp texture edges through image sequence in negative frame direction
		// Begin at key frame index and stop, if another keyframe is reached
		for( size_t i=index; i > 0 ; i-- )
		{
			
			// Stop iteration, if next key frame + 1 is reached
			if( i == this->edgeMaskIndices()[std::max(int(j-1),int(0))] + 1 )
				break;

			sout << "\tprocess image " << i << " (forward flow)" << std::endl;
			
			// Optical flow
			const CImg<float> &flow = this->opticalFlow()[i+1];
			assert(flow.width() == w && flow.height() == h );

			// probability of depth edges
			assert(probability[i].width() == w && probability[i].height() == h );
			
			// Get warp mask[ i-1 ]	
			CImg<float> warp = getWarp(warpMask[i], flow, !this->isForwardFlow() );

			// Save warped mask, if next frame is no key frame
			warpMask[i-1] = warp ;
			
			// Adjust probability edges of next frame by warped edges 
			correctProbability(probability[i-1],warp);

		}

	}

	this->adjustedProbability() = probability;
	//this->warpedMasks() = warpMask;

}

// Get a list of depth edges where only the entries for whose index a mask is given are calculated
template <typename T>
CImgList<T> CorrectDepthEdgeProbability<T>::getWarpMask(){

	CImgList<T> depthEdges( this->edgesProbability().size() );

	const unsigned int size = this->edgeBorder();

	for( size_t i = 0; i < this->edgesProbability().size(); i++)
	{

		// Check if mask with index i is given
		for( size_t j = 0; j < this->edgeMaskIndices().size() ; j++)
		{
			if( this->edgeMaskIndices()[j] == i )
			{
				const CImg<double> &depthEdge	= this->edgesMask()[j];
				CImg<double> sum(depthEdge.width(),depthEdge.height());
				sum.fill(0);

				// mask all pixels without the depth edges
				cimg_forXY(sum,x,y)
				{
					if( depthEdge.get_crop(x-size,y-size,x+size,y+size).sum() == 0 )
						sum(x,y) = 1;
				}

				depthEdges[i] = sum ;
			}
			else
			{
				depthEdges[i] = CImg<T>() ;
			}
		}
	}

	return depthEdges;

}


/// warp function, stolen from the Warp module of charon utils
template<typename T>
CImg<float> CorrectDepthEdgeProbability<T>::getWarp(const CImg<float> &img, const CImg<float> &flow, bool forward = true){
	
	// output image
	CImg<float> out(img.width(),img.height(),1,1);
	out.fill(0);
	
	sout << "flow.spectrum() = " << flow.spectrum() << std::endl;
	sout << "flow.depth() = " << flow.depth() << std::endl;

	const float l = ( forward ? -1 : 1 );
	float xn, yn ;
	
	float val = 0;
	cimg_forXY(flow,x,y) { // for each position within volume
		xn = (float)x+l*flow(x,y,0,0);
		yn = (float)y+l*flow(x,y,1,0);
		if(xn>=0 && xn < img.width() && yn >= 0 && yn < img.height() )
		{
			out(x,y,0,0) = img.linear_atXY(xn,yn);
			val += out(x,y,0,0);
		}
	}
	sout << "warped overall value of " << val << std::endl;
	this->warpedMasks().push_back( out.get_normalize(0,255) );
	return out;
}

template <typename T>
void CorrectDepthEdgeProbability<T>::init() {

	if( this->edgesMask().size() < this->edgeMaskIndices().size() )
		throw(std::range_error("CorrectDepthEdgeProbability<T>::init(): Number of mask indices != masks"));
	if( this->opticalFlow.connected() )
		if( this->opticalFlow().size() < this->edgesProbability().size() - 1 )
			throw(std::range_error("CorrectDepthEdgeProbability<T>::init(): Number of opticalFlow images != probabilities"));


}


template <typename T>
void CorrectDepthEdgeProbability<T>::correctProbability(CImg<float> &prob, const CImg<float> &mask)
{
	// Adjust probability edges of next frame by warped edges -------------------
	// Decrease probabilities for non depth edge pixels
	int dec = 0;
	cimg_forXY(prob,x,y)
	{
		// Only decrease value, if it is already smaller than the defined threshold
		if( prob(x,y) < this->probThreshold() * 255 )
		{
			prob(x,y) *= ( 1 - this->correctionFactor() * mask(x,y) ) ;		
			dec ++;
		}				
	}
	sout << "Number of decreased prob pixels: " << dec << std::endl;	
	// Increase probabilities for depth edge pixels
	if(!this->onlyDecreaseProb())
	{
		int inc = 0;
		CImg<float> warpInv = mask.get_normalize(0,1);
		warpInv -= 1;
		warpInv.normalize(0,1);
		cimg_forXY(prob,x,y)
		{
			// Only increase value, if it is already taller than the defined threshold
			if( prob(x,y) > this->probThreshold2() * 255 )
			{
				prob(x,y) *= ( 1 + this->correctionFactor() * warpInv(x,y) ) ;				
				inc ++;
			}
		}
		sout << "Number of increased prob pixels: " << inc << std::endl;
	}
	prob.cut(0,255);
}

#endif /* _CORRECTDEPTHEDGEPROBABILITY_HXX_ */
