#ifndef _CORRESPONDENCE_HXX_
#define _CORRESPONDENCE_HXX_

#include <algorithm>
#include <boost/multi_index_container.hpp>
#include <boost/multi_index/composite_key.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/identity.hpp>
#include <boost/multi_index/member.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/multi_index/random_access_index.hpp>

#include <CImg.h>
#include <stdexcept>
#include <iostream>
#include <iomanip>

namespace hekate {

	/// class for describing 2D or 3D Correspondences between images
	template <typename T>
	class Correspondence
	{
	public:
		/// point or pixel coordinates
		mutable cimg_library::CImg<T> point ;

		/// the Correspondences for the same point in different images should have the same pointIndex
		/*! the combination of pointIndex and imgIndex should be unique */
		//const size_t pointIndex ;
		size_t pointIndex ;

		/// denotes which image this Correspondence belongs to
		/*! the combination of pointIndex and imgIndex should be unique */
		//const size_t imgIndex ;
		size_t imgIndex ;

		/// marks vector as outlier
		bool outlier;

		/// RGB values for this point
		float R, G, B;

		/// vector of corresponding 2d point ids
		//std::vector< size_t > correspondences;

		/// default constructor
		explicit Correspondence(T x = 0.0 , T y = 0.0 , T z = 0.0 , T w = 1.0, size_t pointIndex = 0, size_t imgIndex = 0, bool outlier=false,
			float rgb_r = 0, float rgb_g = 0, float rgb_b = 0) : 
			point(4,1,1,1,x,y,z,w), pointIndex(pointIndex), imgIndex(imgIndex), outlier(outlier), 
				R(rgb_r), G(rgb_g), B(rgb_b)
		{}
			
		/// construct point from an CImg, 
		/*! @param vec this->point will be initialized with vec
			@param pointIdx new point Index
			@param imgIdx new Image Index */
		template <typename T2>
		explicit Correspondence(cimg_library::CImg<T2> vec, size_t pointIdx = 0, size_t imgIdx = 0, bool outlier=false, 
			float rgb_r = 0, float rgb_g = 0, float rgb_b = 0) :
			point(vec), pointIndex(pointIdx), imgIndex(imgIdx), outlier(outlier), 
				R(rgb_r), G(rgb_g), B(rgb_b)
		{}

		template <typename T2>
		explicit Correspondence(std::vector<T2> vec, size_t pointIdx = 0, size_t imgIdx = 0, bool outlier=false, 
			float rgb_r = 0, float rgb_g = 0, float rgb_b = 0) :
			point(vec.size(),1,1,1), pointIndex(pointIdx), imgIndex(imgIdx), outlier(outlier), 
				R(rgb_r), G(rgb_g), B(rgb_b)
		{
			std::copy(vec.begin(), vec.end(), point.begin()) ;
		}

		virtual ~Correspondence() {} ;

		/// save coordinates to a std::vector of size 4
		template <typename T2>
		std::vector<T2> toVector() const
		{
			std::vector<T2> result(point.size(),0) ;
			std::copy(point.begin(), point.end(), result.begin()) ;
			return result ;
		}

		T& x() const
		{
			if(point.size() < 1)
				throw std::range_error("Correspondence Dimenision is zero!") ;
			else
				return point[0] ;
		}

		T& y() const
		{
			if(point.size() < 2)
				throw std::range_error("Correspondence Dimenision is zero!") ;
			else
				return point[1] ;
		}

		T& z() const
		{
			if(point.size() < 3)
				throw std::range_error("Correspondence Dimenision is zero!") ;
			else
				return point[2] ;
		}

		T& w() const
		{
			if(point.size() < 4)
				throw std::range_error("Correspondence Dimenision is zero!") ;
			else
				return point[3] ;
		}
		
		const bool& isOutlier() const
		{
			return outlier;
		}

		cimg_library::CImg<T> pt2d() const
		{
			cimg_library::CImg<T> pt(1,3);
			pt[0] = this->x();
			pt[1] = this->y();
			pt[2] = this->w();
			if(pt[2] != 0) pt /= pt[2];
			return pt;
		}

		cimg_library::CImg<T> pt3d() const
		{
			cimg_library::CImg<T> pt(1,4);
			pt[0] = this->x();
			pt[1] = this->y();
			pt[2] = this->z();
			pt[3] = this->w();
			if(pt[3] != 0) pt /= pt[3];
			return pt;
		}

		//friend std::ostream &operator<<(std::ostream &stream, const typename Correspondence<T>& obj);
		//friend std::istream &operator>>(std::istream &stream, typename Correspondence<T> &obj);

	} ;

	template <typename T>
	std::ostream& operator<< (std::ostream& stream, const Correspondence<T>& obj)
	{
		if(obj.point.size() < 4)
		{	throw std::range_error("Illegal Correspondence; point has to few dimensions") ;	}
			
		std::ios_base::fmtflags flgs = stream.flags() ;
		stream.precision(4) ;
		stream	<< std::scientific << obj.point[0] << "\t" << obj.point[1] << "\t" 
				<< obj.point[2] << "\t" <<  obj.point[3] << "\t" 
				<< obj.pointIndex << "\t" << obj.imgIndex << "\t" << obj.outlier << "\t"
				<< obj.R << "\t" << obj.G << "\t" << obj.B;
		stream.flags(flgs) ;
		return stream;
	}

	//-------------------------------------------------------------------------
	//This Section contains some rather freaky template code and you shouldn't have to use it directly

	namespace bmi = boost::multi_index ;
	using namespace bmi ;

	//named structs for simple access to boost::multi_index_container indices
	struct pointIndex {} ;
	struct imgIndex {} ;
	struct pointImgIndex {} ;
	struct imgPointIndex {} ;
	struct random {} ;

	//struct which derives from the actual index_set specification
	//we use this as a template parameter for the acutall multi_index_container to shorten the type name
	//this way the error and warning messages for this type get much shorter and are easier readable
	template<typename T>
	struct CorrespondenceIndexSet :
			indexed_by<
				random_access<boost::multi_index::tag<random> >,
				ordered_non_unique<boost::multi_index::tag<pointIndex>, member<Correspondence<T>, const size_t, &Correspondence<T>::pointIndex> >, 
				ordered_non_unique<boost::multi_index::tag<imgIndex>, member<Correspondence<T>, const size_t, &Correspondence<T>::imgIndex> >,
				ordered_non_unique<boost::multi_index::tag<pointImgIndex>,
									composite_key<Correspondence<T>,
													member<Correspondence<T>, const size_t, &Correspondence<T>::pointIndex>, 
													member<Correspondence<T>, const size_t, &Correspondence<T>::imgIndex>
									>
				>,
				ordered_non_unique<boost::multi_index::tag<imgPointIndex>,
									composite_key<Correspondence<T>,
													member<Correspondence<T>, const size_t, &Correspondence<T>::imgIndex>, 
													member<Correspondence<T>, const size_t, &Correspondence<T>::pointIndex>
									>
				>
			> {};

	
	///Helper struct for template instanciation, typdef can not be used in combination with the template keyword 
	template <typename T>
	struct CorrespondenceBaseList {
	
		typedef multi_index_container<Correspondence<T>, CorrespondenceIndexSet<T> > List ;
	};

	typedef CorrespondenceBaseList<double>::List DCorList ;
	typedef CorrespondenceBaseList<float>::List FCorList ;
	
	//-------------------------------------------------------------------------
	//Here starts the class you should actually use

	/// List which contains Correspondence<T> elements
	template <typename T>
	struct CorrespondenceList : public CorrespondenceBaseList<T>::List
	{
	public:
		CorrespondenceList();
		CorrespondenceList(std::vector<std::vector<std::vector<T> > > &);
		/// reference to correspondence of imgIndex and pointIndex
		Correspondence<T>& operator()(size_t imgIndex,size_t pointIndex);
		const Correspondence<T>& operator()(size_t imgIndex,size_t pointIndex) const;

		//convenience types to access the list by various indices
		typedef typename CorrespondenceBaseList<T>::List::template index<pointIndex>::type ListByIndex ;
		typedef typename CorrespondenceBaseList<T>::List::template index<imgIndex>::type ListByImg ;
		typedef typename CorrespondenceBaseList<T>::List::template index<random>::type ListAsVec ;
		//typedef typename nth_index<CorrespondenceBaseList<T>::List,3>::type ListByPointThenImg ;
		typedef typename CorrespondenceBaseList<T>::List::template nth_index<3>::type ListByPointThenImg ;
		typedef typename CorrespondenceBaseList<T>::List::template nth_index<4>::type ListByImgThenPoint ;

		const size_t size() const;
		const size_t sizeImages() const;
		const size_t sizePoints() const;
		const size_t sizePoints(const int i) const;
		const size_t getMaxPointIndex() const;
		
		bool exist(size_t imgIndex,size_t pointIndex) const;
		bool existPtIndex(size_t pointIndex) const;
		bool existImgIndex(size_t imgIndex) const;

		Correspondence<T> getPointByIndex(size_t pointIndex) const;

		std::vector< const Correspondence<T>*> getListByPointId(const size_t pointIndex) const;
		std::vector< const Correspondence<T>*> getListByPointId(const size_t pointIndex);
		std::vector< const Correspondence<T>*> getListByImgId(const size_t imageIndex) const;
		std::vector< const Correspondence<T>*> getListByImgId(const size_t imageIndex) ;

		void normalize();
		/// marks correspondences with doubles of imgIndex AND pointIndex 
		int removeDoubles();
		std::vector<std::vector<std::vector<T> > > getVectorForm() const;
		CorrespondenceList<T> removeOutliers();
		CorrespondenceList<T> removeOutlierCorrespondences();

	};

	/// Constructor of CorrespondenceList which takes a correspondence vector list as input
	template <typename T>
	CorrespondenceList<T>::CorrespondenceList(){}

	/// Constructor of CorrespondenceList which takes a correspondence vector list as input
	template <typename T>
	CorrespondenceList<T>::CorrespondenceList(std::vector<std::vector<std::vector<T> > > &corr){
		for(size_t i=0; i<corr.size(); i++){
			for(size_t j=0; j<corr[i].size(); j++){
				if(corr[i][j][0] != 0 || corr[i][j][1] != 0 || corr[i][j][2] != 0)
				{
					this->push_back(Correspondence<T>(corr[i][j], j, i, false, 0, 0, 0)) ;
				}
			}
		}
	}

	/// Remove single outlier points
	template <typename T>
	CorrespondenceList<T> CorrespondenceList<T>::removeOutliers(){
		CorrespondenceList<T> listOut;
		auto it = this->begin();
		auto itEnd = this->end();
		for(; it != itEnd; it++){
			if(!(it->outlier))
				listOut.push_back( *it );
		}
		return listOut;
	}

	/// Remove complete correspondences (points of same point id)
	template <typename T>
	CorrespondenceList<T> CorrespondenceList<T>::removeOutlierCorrespondences(){
		CorrespondenceList<T> listOut;
		int nOut = 0;
		for( size_t i = 0; i < this->getMaxPointIndex(); i++)
		{
			try{
				const std::vector< const Correspondence<T>*> list = this->getListByPointId(i);
				bool outliers = false;
				for(size_t j = 0; j < list.size(); j++)
				{
					if(list[j]->outlier){
						outliers = true;
						nOut++;
						break;
					}
				}
				if(!outliers)
				{
					for(size_t j = 0; j < list.size(); j++)
					{
						listOut.push_back( *list[j] );
					}
				}
			}
			catch(...){}
		}
		sout << "Removed " << nOut << " whole correspondences" << std::endl;
		return listOut;
	}
		
	/// Create vector with pointers to all correspondences with point id i
	template <typename T>
	std::vector< const Correspondence<T>*> CorrespondenceList<T>::getListByPointId(const size_t ptID) const{
		std::vector< const Correspondence<T>*> vec;
		ListByPointThenImg::iterator it = this->get<3>().lower_bound(ptID) ;
		for(; it != this->get<3>().upper_bound(ptID) ; it++)
		{
			vec.push_back(&(*it));
		}
		return vec;
	}
		
	/// Create vector with pointers to all correspondences with point id i
	template <typename T>
	std::vector< const Correspondence<T>*> CorrespondenceList<T>::getListByPointId(const size_t ptID) {
		std::vector< const Correspondence<T>*> vec;
		ListByPointThenImg::iterator it = this->get<3>().lower_bound(ptID) ;
		for(; it != this->get<3>().upper_bound(ptID) ; it++)
		{
			vec.push_back(&(*it));
		}
		return vec;
	}
		
	/// Create vector with pointers to all correspondences with img id i
	template <typename T>
	std::vector<const Correspondence<T>*> CorrespondenceList<T>::getListByImgId(const size_t imgID) const{
		std::vector<const Correspondence<T>*> vec;
		ListByImgThenPoint::iterator it = this->get<4>().lower_bound(imgID) ;
		for(; it != this->get<4>().upper_bound(imgID) ; it++)
		{
			vec.push_back(&(*it));
		}
		return vec;
	}
		
	/// Create vector with pointers to all correspondences with img id i
	template <typename T>
	std::vector< const Correspondence<T>*> CorrespondenceList<T>::getListByImgId(const size_t imgID) {
		std::vector<const Correspondence<T>*> vec;
		ListByImgThenPoint::iterator it = this->get<4>().lower_bound(imgID) ;
		for(; it != this->get<4>().upper_bound(imgID) ; it++)
		{
			vec.push_back(&(*it));
		}
		return vec;
	}
	
	/// returns the correspondencelist as a vector (former correspondence format)
	template <typename T>
	std::vector<std::vector<std::vector<T> > > CorrespondenceList<T>::getVectorForm() const
	{

		size_t sizeImg = this->sizeImages();
		size_t sizePt = this->size();
		size_t maxIndex = this->getMaxPointIndex();
		std::vector<std::vector<std::vector<T> > > _pointsList;

		sout << "Converting CorrespondenceList&ltT&gt to vector&ltvector&ltvector&ltT&gt&gt&gt" << std::endl;
		sout << "Found " << sizeImg << " different image ids, " << sizePt
			<< " different points with a maximal point index of " << maxIndex << std::endl;

		// Define outlier in vector format
		std::vector<T> outlier;
		outlier.push_back(T(0));
		outlier.push_back(T(0));
		outlier.push_back(T(0));
		outlier.push_back(T(0));

		// Resize pointsList to number of images
		_pointsList.resize(sizeImg);

		for(size_t i=0; i<sizeImg; i++)
		{

			if(!this->existImgIndex(i))
				throw(std::range_error("CorrespondenceList<T>::getVectorForm(): Image index not available"));

			// Define vector list of one image and fill it with outliers
			std::vector<std::vector<T> > _points;
			for(size_t j=0; j<maxIndex+1; j++) _points.push_back(outlier);
			
			auto it = this->get<imgIndex>().lower_bound(i) ;
			auto itEnd = this->get<imgIndex>().upper_bound(i) ;
			for(; it != itEnd ; it++)
			{

				if(!it->outlier)
				{
					std::vector<T> pt;
					pt.push_back(it->x());
					pt.push_back(it->y());
					pt.push_back(it->w());
					_points[it->pointIndex] = pt;
				}

			}

			_pointsList[i] = _points ;

		}

		return _pointsList;
	}
	

	/// Normalize all correspondences by w!=1
	template <typename T>
	void CorrespondenceList<T>::normalize()
	{
		ListAsVec::iterator it = this.begin() ;
		for(; it != corrOut.end(); it++)
		{
			it->pt.x() /= it->pt.w();
			it->pt.y() /= it->pt.w();
			it->pt.z() /= it->pt.w();
		}
	}

	/// Get number of all Correspondence<T> elements 
	template <typename T>
	const size_t CorrespondenceList<T>::size() const
	{			
		size_t sizeAll = 0;
		ListAsVec::iterator it = this->begin() ;
		for(; it != this->end(); it++){sizeAll++;}
		return sizeAll;
	}

	/// Get maximal point index
	template <typename T>
	const size_t CorrespondenceList<T>::getMaxPointIndex() const
	{			
		size_t maxIndex = 0;
		auto it = this->begin() ;
		auto itEnd = this->end();
		for(; it != itEnd ; it++)
		{
			if(maxIndex < it->pointIndex)
				maxIndex = it->pointIndex;
		}
		return maxIndex;
	}

	/// Get number of different images indices of the Correspondence<T> elements 
	template <typename T>
	const size_t CorrespondenceList<T>::sizeImages() const
	{
		std::vector<size_t> imageIndices;
		auto it = this->begin() ;
		for(; it != this->end(); it++)
		{
			// Push back first image index found
			if(imageIndices.size() == 0)
				imageIndices.push_back(it->imgIndex);

			// Check if current img index is equal with an existing one
			else
			{
				bool exists = false;
				for(size_t i=0; i<imageIndices.size(); i++)
				{
					if(imageIndices[i] == it->imgIndex)
					{
						exists = true;
						break;
					}
				}
				// if img index is not known yet, push back
				if(!exists)
					imageIndices.push_back(it->imgIndex);
			}
		}
		return imageIndices.size();
	}
	
	/// Get number of different point indices of the Correspondence elements 
	template <typename T>
	const size_t CorrespondenceList<T>::sizePoints(const int i) const
	{
		// out commented since if index is not available, method returns 0
		// so the throw is not needed?
		//if(i >= this->sizeImages())
		//	throw(std::range_error("CorrespondenceList<T>::sizePoints(const int i)"));
		size_t sizePt = 0;
		ListByImg::iterator it = this->get<imgIndex>().lower_bound(i) ;
		for(; it != this->get<imgIndex>().upper_bound(i) ; it++ )
		{
			sizePt++;
		}
		return sizePt;
	}
	
	/// Get number of different point indices of the Correspondence<T> elements 
	template <typename T>
	const size_t CorrespondenceList<T>::sizePoints() const
	{

		std::vector<size_t> pointIndices;
		auto it = this->begin() ;
		for(; it != this->end(); it++)
		{
			// Push back first image index found
			if(pointIndices.size() == 0)
				pointIndices.push_back(it->pointIndex);

			// Check if current img index is equal with an existing one
			else
			{
				bool exists = false;
				for(size_t i=0; i<pointIndices.size(); i++)
				{
					if(pointIndices[i] == it->pointIndex)
					{
						exists = true;
						break;
					}
				}
				// if img index is not known yet, push back
				if(!exists)
					pointIndices.push_back(it->pointIndex);
			}
		}
		return pointIndices.size();

	}



	/// Bracket operator to access a specific Correspondence<T>
	template <typename T>
	Correspondence<T>& CorrespondenceList<T>::operator()(size_t imgId, size_t pointId)
	{

		ListByIndex::iterator it = this->get<pointIndex>().lower_bound(pointId) ;
		for(; it != this->get<pointIndex>().upper_bound(pointId) ; it++)
		{
			if((*it).imgIndex == imgId) return const_cast<Correspondence<T>& >(*it);
		}

		char buffer [80];
		sprintf (buffer, "CorrespondenceList(%d,%d) -> Point could not be found", imgId, pointId);
		throw(std::range_error( buffer ));

	}
	
	/// Bracket operator to access a specific Correspondence<T>
	template <typename T>
	const Correspondence<T>& CorrespondenceList<T>::operator()(size_t imgId, size_t pointId) const
	{

		ListByIndex::iterator it = this->get<pointIndex>().lower_bound(pointId) ;
		for(; it != this->get<pointIndex>().upper_bound(pointId) ; it++)
		{
			if((*it).imgIndex == imgId) return const_cast<Correspondence<T>& >(*it);
		}

		throw(std::range_error("CorrespondenceList(i,j) -> Point could not be found"));

	}

	/// Checks if point (i,j) exists
	template <typename T>
	bool CorrespondenceList<T>::exist(size_t imgId, size_t pointId) const
	{

		ListByIndex::iterator it = this->get<pointIndex>().lower_bound(pointId) ;
		for(; it != this->get<pointIndex>().upper_bound(pointId) ; it++)
		{
			if(it->imgIndex == imgId) return 1;
		}

		return 0;

	}

	/// Checks if point with point index 'pointId' exists
	template <typename T>
	bool CorrespondenceList<T>::existPtIndex(size_t pointId) const
	{

		ListByIndex::iterator it = this->get<pointIndex>().lower_bound(pointId) ;
		for(; it != this->get<pointIndex>().upper_bound(pointId) ; it++)
		{
			return 1;
		}

		return 0;

	}

	template <typename T>
	Correspondence<T> CorrespondenceList<T>::getPointByIndex(size_t pointIndex) const
	{
		typename ListAsVec::iterator it = this->begin() ;
		for(; it != this->end() ; it++)
		{
			if(it->pointIndex == pointIndex) 
				return *it;
		}

		throw std::runtime_error("CorrespondenceList: Point Index does not exist");		
	}

	/// Checks if point with image index 'imgId' exists
	template <typename T>
	bool CorrespondenceList<T>::existImgIndex(size_t imgId) const
	{

		ListByImg::iterator it = this->get<imgIndex>().lower_bound(imgId) ;
		for(; it != this->get<imgIndex>().upper_bound(imgId) ; it++)
		{
			sout << "CorrespondenceList<T>::existImgIndex: New version of this function" << std::endl;
			return 1;
		}

		return 0;

	}


	//stream operator for CorrespondenceList, makes cout << list, possible
	//using CorrespondenceList<T>::List as the type for the second parameter does not work, we need the full description
	template <typename T>
	std::ostream &operator<<(std::ostream &stream, 
		const CorrespondenceList<T> & list
	)
	{
		for(CorrespondenceList<T>::index<random>::type::const_iterator it= list.begin() ; it != list.end() ; it++)
		{
			stream << *it << "\n" ;
		}
		return stream ;
	}
	
	template <typename T>
	std::istream &operator>>(std::istream &stream, 
		CorrespondenceList<T> & list
	)
	{
		T x,y,z,w ;
		float R, G, B;
		size_t pointIndex, imgIndex ;
		bool outlier ;
		while(stream.good())
		{
			stream >> x >> y >> z >> w >> pointIndex >> imgIndex >> outlier >> R >> G >> B ;
			Correspondence<T> cor(x,y,z,w,pointIndex,imgIndex, outlier, R, G, B) ;
			list.push_back(cor) ;
		}
		return stream ;
	}




	/// marks correspondences with doubles of imgIndex AND pointIndex as outliers
	template <typename T>
	int CorrespondenceList<T>::removeDoubles()
	{

		int nDoubles = 0;
		size_t nImg = this->sizeImages();

		auto it2 = this->begin() ;
		auto it2End = this->end() ;
		for(; it2 != it2End ; it2++)
		{

			if(it2->outlier)
				continue;

			std::vector< Correspondence<T>* > vec;// = this->getListByPointId(j);
			auto it = this->get<3>().lower_bound(it2->pointIndex) ;
			for(; it != this->get<3>().upper_bound(it2->pointIndex) ; it++)
			{
				if(it->outlier)
					continue;
				vec.push_back(const_cast<Correspondence<T>*>(&(*it)));
			}

			bool hasDoubles = false;
				
			// Check vector for double indices
			for(size_t m = 0; m < vec.size(); m++)
			{
				for(size_t k = m+1; k < vec.size(); k++)
				{
					if(vec[m]->pointIndex == vec[k]->pointIndex
						&& vec[m]->imgIndex == vec[k]->imgIndex)
					{
						if(vec[m]->point != vec[k]->point)
						{
							hasDoubles = true;
							break;
						}
					}
				}
				if(hasDoubles) break;
			}

			// If vector has doubles, mark all points of it as outliers
			if(hasDoubles)
			{
				nDoubles++;
				//sout << "Correspondences with point id " << vec[0]->pointIndex
				//	<< " has doubles. Marked doubles as outliers." << std::endl;
				for(size_t k = 0; k < vec.size(); k++)
				{
					vec[k]->outlier = true;
					//sout << *(vec[k]) << std::endl;
				}
			}


		}

		return nDoubles;

	}


} ; /*namespace hekate */

/*
	Usage examples:

	//create a new list
	using namespace hekate ;
	CorrespondenceList<T>::List list ;

	//use it (almost) like a std::vector
	list.push_back(Correspondence<T>()) ;
	list.push_back(Correspondence<T>()) ;
	list.push_back(Correspondence<T>(1.0,2.0,-5.0, 2, 3)) ;
	size_t s = list.size() ; // = 2
	const Correspondence<T>& cor = list[0] ;
	sout << cor.y ;
	cor.x = 2.0 ;
	
	Correspondence<T> newCor ;
	//list[5] = newCor ; //<- does not work
	list.replace(list.begin() + 5,newCor) ; //<- use this instead

	//iterate over all Correspondences whose point index is 0
	CorrespondenceList<T>::ListByIndex::iterator it = list.get<pointIndex>().lower_bound(0) ;
	for(; it != list.get<pointIndex>().upper_bound(0) ; it++)
	{
		it->x = 2.0 ;
	}

	//iterate over all Correspondences
	CorrespondenceList<T>::ListAsVec::iterator it = corrOut.begin() ;
	for(; it != corrOut.end(); it++)
	{
	}

	//find a Correspondence with the Image Index 2
	CorrespondenceList<T>::ListByImg::iterator it2 = list.get<imgIndex>().find(2) ;
	const Correspondence<T>& anotherCor = *it ;
	sout << anotherCor.imgIndex ; //prints 2


*/

#endif /* #ifndef _CORRESPONDENCE_HXX_ */