/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Correspondence2FlowMap.hxx
 *  Implementation of the parameter class Correspondence2FlowMap.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 07.03.2013
 */

#ifndef _CORRESPONDENCE2FLOWMAP_HXX_
#define _CORRESPONDENCE2FLOWMAP_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/Correspondence2FlowMap.h>

template <typename T>
Correspondence2FlowMap<T>::Correspondence2FlowMap(const std::string& name) :
		TemplatedParameteredObject<T>(
			"Correspondence2FlowMap", name,
			"<h2>Writes correspondences to a images which can be used as an "
			"initial guess for optical flow</h2><br>"
			"This plugin converts correspondences to:<br>"
			"1) a mask where all pixels which are part of the correspondence "
			"are 1, otherwise 0<br>"
			"2) two images with the x and y shift value for the masked "
			"pixels to shift the point to its correspondence in the "
			"subsequent frame"
		)
{
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"Correspondences to generate the mask and images",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		images, "images",
		"Source images to read out width and heigth of every image",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		masks, "masks",
		"Masks which mask correspondence pixels",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		xTranslation, "xTranslation",
		"Translation in x direction of the correspondence",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		yTranslation, "yTranslation",
		"Translation in y direction of the correspondence",
		"CImgList<T>");

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void Correspondence2FlowMap<T>::execute() {
	
	if( this->images().size() < this->correspondences().sizeImages() )
		throw( std::range_error("Number of images smaller than number of given correspondences"));

	const CorrespondenceList<T> &corr = this->correspondences();

	for( size_t i = 0; i < this->images().size() -1; i++ )
	{

		sout << "Convert correspondences between frame " << i << " and " << i+1 << std::endl;
		
		CImg<T> src(this->images()[i].c_str());
		size_t w = src.width();
		size_t h = src.height();
		CImg<int> mask( w, h );			// image with masked correspondence pixels
		CImg< float > shiftX( w, h );	// image with x shift to subsequent point in sequence
		CImg< float > shiftY( w, h );	// image with y shift to subsequent point in sequence
		mask.fill(0);
		shiftX.fill(0);
		shiftY.fill(0);
		
		int xxx = 0;

		auto it = corr.get<imgIndex>().lower_bound(i) ;
		auto itEnd = corr.get<imgIndex>().upper_bound(i) ;
		for( ; it != itEnd; it++)
		{

			if( it->outlier )
				continue;

			try{
				// Get corresponding point of next frame
				const Correspondence<T> &corrNext = corr( i+1, it->pointIndex );

				if( corrNext.outlier )
					continue;
				
				// calculate integer value of current point
				unsigned int x	= int( it->pt2d()[0] + 0.5 );
				unsigned int y	= int( it->pt2d()[1] + 0.5 );

				// Write mask
				mask(x,y)		= 1;

				// Write difference between x and y values of current and next frame point
				shiftX(x,y)		= corrNext.pt2d()[0] - it->pt2d()[0];
				shiftY(x,y)		= corrNext.pt2d()[1] - it->pt2d()[1];
			
			}
			catch( ... ){}
		
		}

		this->masks().push_back( mask );
		this->xTranslation().push_back( shiftX );
		this->yTranslation().push_back( shiftY );

	}

}

#endif /* _CORRESPONDENCE2FLOWMAP_HXX_ */
