/*  Copyright (C) 2011 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.

	-------------------------------------------------------------------------
	This plugin uses the LIBSVM:
	Copyright (c) 2000-2011 Chih-Chung Chang and Chih-Jen Lin
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:

	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.

	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	3. Neither name of copyright holders nor the names of its contributors
	may be used to endorse or promote products derived from this software
	without specific prior written permission.


	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
	``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
	LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
	A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	-----------------------------------------------------------------------

*/
/** \file CorrespondenceOutlierElimination.hxx
 *  Implementation of the parameter class CorrespondenceOutlierElimination.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.09.2011
 */

#ifndef _CORRESPONDENCEOUTLIERELIMINATION_HXX_
#define _CORRESPONDENCEOUTLIERELIMINATION_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/CorrespondenceOutlierElimination.h>
#include <libsvm/svm.h>
#include <hekate/CorrView.h>
#include <charon-utils/ImgTool.hxx>

using namespace hekate;

template <typename T>
CorrespondenceOutlierElimination<T>::CorrespondenceOutlierElimination(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CorrespondenceOutlierElimination", name,
			"<h2>Eliminates correspondence outliers by SVM regression "
			"method</h2><br>"
			"Eliminates correspondence outliers by SVM regression method. "
			"This plugin uses <a "
			"href=\"http://www.csie.ntu.edu.tw/~cjlin/libsvm/\">LIBSVM</a>."
		)
{
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");

	ParameteredObject::_addOutputSlot(
		correspondencesOut, "correspondencesOut",
		"correspondencesOut",
		"CorrespondenceList");

	ParameteredObject::_addParameter< float >(
		threshold1, "subsetThreshold",
		"Confidence level for the standard deviation (stdev) of"
		" sub set correspondence views (CV). The ratio between"
		" stdev of initial CV and stdev of sub set CV has "
		"to be taller than subsetThreshold.",
		0.1, "float");
	ParameteredObject::_addParameter< float >(
		threshold2, "outlierThreshold",
		"Confidence level for outliers:<br>"
		" To detect an outlier, Err(x)^2*Stdev(x)^2+Err(y)^2*Stdev(y)^2 "
		" has to be taller than outlierThreshold.",
		5, "float");
	ParameteredObject::_addParameter< float >(
		threshold3, "influentalThreshold",
		"Confidence level for influentals. The error of influentals"
		" has to be taller than influentalThreshold*sigma where as "
		" sigma is calculated automatically.",
		5, "float");
	ParameteredObject::_addParameter< float >(
		threshold4, "maxInfluentals",
		"Maximal relative amount of influentals which can be removed "
		"to estimate a correspondence view.",
		0.9, "float");
	
	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void CorrespondenceOutlierElimination<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();


	// Read in correspondences
	std::vector<std::vector<std::vector<T> > > corr = this->correspondences().getVectorForm();
	std::vector<std::vector<bool> > outlierMask = getOutlierMask(corr.size(), corr[0].size());
	
	//int initialSize = this->correspondences().sizePoints();
	//sout << "correspondence<T> point size = " << initialSize << std::endl;
	

	// 1) Normalize correspondences ----------------------------------------------------
	const cimg_library::CImg<T> trafo = normalizeCorrespondences(corr);
	

	/*
	Iterate over all possible correspondence pairs -------------------------------------
	(At least one pair has to be available but also more
	than one pair can be processed (Such as correspondences
	of the FeatureKLTSequence module))
	*/
	std::vector<int> nOuts;
	for(size_t i = 0; i<corr.size()-1; i++){

		sout << std::endl << "Looking for outliers of correspondence between "
			<< "image " << i << " and " << i+1 << ":" << std::endl;

		std::vector<std::vector<T> > & corr1 = corr[i];
		std::vector<std::vector<T> > & corr2 = corr[i+1];

		/* 
			corr[0] = Reference feature points of initial frame
			*corrIt = Iterated feature points of additional frames
			corrlearned = Estimated feature points from learned function		
		*/


		// 2) Learn f(u,v) and f'(u',v') from complete set S
		sout << "Learn correspondence views S, S2:" << std::endl;
		CorrView<T> S(corr1,corr2);
		CorrView<T> S2(corr2,corr1);


		// Remove influentals from both sets
		sout << "Remove influental sub sets to learn a better " 
			<< "correspondence view --------------" << std::endl;
		int infl1 = removeInfluentals(S, this->threshold1(), this->threshold3(), this->threshold4());
		if(infl1 == 0)
		{
			sout << "Could not calculate good correspondence views between "
				<< "image " << i << " and " << i+1 << std::endl;
			sout << "Skip outlier estimation between those images." << std::endl;
			nOuts.push_back(0);
			continue;
		}

		int infl2 = removeInfluentals(S2, this->threshold1(), this->threshold3(), this->threshold4());
		if(infl2 == 0)
		{
			sout << "Could not calculate good correspondence views between "
				<< "image " << i+1 << " and " << i << std::endl;
			sout << "Skip outlier estimation between those images." << std::endl;
			nOuts.push_back(0);
			continue;
		}

		sout << infl1 << " influentals have been removed from"
			<< " correspondence view f(u,v)." << std::endl;
		sout << infl2 << " influentals have been removed from"
			<< " correspondence view f'(u',v')." << std::endl;		
		sout << "Use corrected correspondence views to compare "
			<< "learned and predicted correspondences..." << std::endl;



		// Detect and mark (zero vector) outliers by comparing found 
		// correspondences with predicted ones.
		// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
		// An dieser Stelle nochmal aufpassen, welche Outliers wirklich geloescht werden und ob danach noch alles synchron ist!!!!!
		// Extra function um outlier zu detektieren, detectOutliers() ist eigentlich fuer influentals
		int c2 = detectOutliers(corr1,corr2, i, i+1, S, outlierMask, this->threshold2());
		sout << c2 << " outliers detected in correspondence view f(u,v)!" << std::endl;
		int c3 = detectOutliers(corr2,corr1, i+1, i, S2, outlierMask, this->threshold2());
		sout << c3 << " outliers detected in correspondence view f'(u',v')!" << std::endl;
		
		nOuts.push_back(c2+c3);

		
	}



	// Reject all outliers
	// This is done after all outliers of all correspondence sets
	// were determined, to ensure synchronisation between
	// all frames
	// (removes complete correspondence if at least one is an outlier)
	// (for workflows where outliers are no problems, do not perform
	// such this method since you will get much less points)
	//int out = rejectOutliers(corr); 

	//sout << std::endl;
	//sout << "------------------------------------------" << std::endl;
	//sout << out << " of " << initialSize << " outliers removed." << std::endl;
	//sout << initialSize - out << " correspondences remaining." << std::endl;
	//sout << "------------------------------------------" << std::endl;
	//sout << std::endl;


	// Denormalize correspondences
	// Not needed anymore, since we use the input correspondences now
	// and just mark the ones, where the according outlier mask is true
	// denormalizeCorrespondences(corr, trafo);		
	
	for(size_t i=0; i<corr.size()-1; i++)
	{

		sout << "Removed " << nOuts[i] << " outliers between images " << i 
			<< " and " << i+1 << std::endl;

	}
	
	// Mark correspondence as outlier
	this->correspondencesOut() = this->correspondences();
	auto it = this->correspondencesOut().begin();
	auto itEnd = this->correspondencesOut().end();
	for(; it != itEnd; it++)
	{

		if( outlierMask[it->imgIndex][it->pointIndex] == true )
		{
			const_cast<Correspondence<T> &>(*it).outlier = true;
		}
		
	}

}


/// Removes influentals from correspondence view (CV)
/*	The function creates sub sets of correspondences by removing 
	influentals, calculates new correspondence views from it and 
	compares the standard deviations between putative and learned 
	correspondences. If the standard deviation of initial CV and 
	sub set CV differs distinctly, the influentals are also
	removed from initial CV.<br>
	epsilon = confidence level for the standard deviation of
	sub set CVs.<br>
	tau = confidence level for influentals. the error of influentals
	has to be taller than tau*sigma where as sigma is calculated
	automatically.
	*/
template <typename T>
int removeInfluentals(CorrView<T> &S, const float epsilon = 0.1, const float tau = 2, const float maxInf = 0.9){

	sout << "Influental parameters: epsilon="<< epsilon << "; tau=" 
		<< tau << "; maxInf=" << maxInf <<";" << std::endl;

	std::vector<std::vector<T> > corrSubRef, corrSubPutative, corrSubLearned;
	
	// Repeat until the new chosen subset does not 
	// influent the learning function anymore
	int influentals = 0;
	int c1 = 0;

	while(1){

		c1++;

		/*
		3) Select suspect influental subset -----------------------------------------
		(see page 6, "Rejecting Mismatches by Correspondence Function", Xiangru Li et. al)
		*/
		corrSubRef.clear();
		corrSubPutative.clear();
		corrSubLearned.clear();

		corrSubRef = S.corrRef();
		corrSubPutative = S.corrPutative();
		corrSubLearned = S.corrLearned();

		sout << "Number of correspondences in current correspondence view ("
			<< "iteration " << c1 << "):" << std::endl;
		sout << "	N(correspondences)         = " << corrSubRef.size() << std::endl;

		/*
		double nMax = 0;
		for(size_t i=0; i<corrSubPutative.size(); i++)
		{
			if(checkPoint(corrSubPutative[i]))
				nMax++;
		}
		*/

		int outs = chooseSubset(corrSubRef, corrSubPutative, corrSubLearned,S.sigma(), tau);
		influentals += outs;

		if(outs > 0)
		{
			sout << "	N(influentals)             = " << outs << std::endl;
			/*
			double nInf = (double) outs / nMax;
			double nInf2 = (double) outs / nMax;
			if(  nInf > maxInf )
			{
				sout << "Too many influentals (n(inf)/n(All):" << nInf << " > maxInfluentals). "
					<< "Stop searching." << std::endl;
				break;
			}
			if( nInf2 < (1-maxInf) )
			{
				sout << "Too many influentals (n(inf)/n(All):" << nInf2 << " < (1-maxInfluentals)). "
					<< "Stop searching." << std::endl;
				break;
			}
			*/
			sout << "Check if removing the influentals leads to an improvement "
				<< "of the correspondence view:" << std::endl;

			CorrView<T> S2(corrSubRef,corrSubPutative);

			sout << "	Stdev(with influentals)    = " << S.sigma()[0] << ", " << S.sigma()[1] << std::endl;
			sout << "	Stdev(without influentals) = " << S2.sigma()[0] << ", " << S2.sigma()[1] << std::endl;

			// Check influence
			if( checkInf( S.sigma() , S2.sigma(), epsilon ) ){
				sout << "-->	Ratio between both standard deviations taller than "
					<< "epsilon = " << epsilon << ":" << std::endl;
				sout << "-->	Removing " << outs << " influentals to improve "
					<< "correspondence view." << std::endl;
				S.learnView(corrSubRef,corrSubPutative);

			}
			else{
				sout << "-->	Ratio between both standard deviations smaller than "
					<< "epsilon = " << epsilon << ":" << std::endl;
				sout << "-->	Removing influentals would not lead to an improvement "
					<< "of the correspondence view." << std::endl;
				sout << "-->	Stop searching for further influentals." << std::endl;
				break;
			}

		}
		else{
			sout << "No further influentals could be found." << std::endl;
			break;
		}

	}
	
	sout << std::endl;
	return influentals;

}


/// Calculate influence of subset
/**
	Compares the standard deviations between putative
	and learned correspondences of two sets. If the 
	standard deviation of the second set is smaller
	than the threshold "epsilon", the function returns false.
*/
bool checkInf(const std::vector<double> & sigma1, 
	const std::vector<double> & sigma2,
	const float epsilon=0.1){

	if(sigma1[0] == 0 || sigma1[1]==0){
		sout << "Standard deviation is zero!" << std::endl;
		return 0;
	}
		
	double infl1 = std::abs( sigma1[0]*sigma1[0] - sigma2[0]*sigma2[0] );		
	double infl11 = infl1 / (sigma1[0]*sigma1[0]);	
		
	double infl2 = std::abs( sigma1[1]*sigma1[1] - sigma2[1]*sigma2[1] );		
	double infl22 = infl2 / (sigma1[1]*sigma1[1]);

	if(infl1 == 0 || infl2 == 0)
		return 0;

	//sout << "Influental value 1 = " << infl11 << std::endl;
	//sout << "Influental value 2 = " << infl22 << std::endl;
	
	if(infl11 > epsilon || infl22 > epsilon)
		return 1;
	else
		return 0;

}

// Choose subset with influentals
template <typename T>
int chooseSubset(std::vector<std::vector<T> > & corrSubRef, 
					 std::vector<std::vector<T> > & corrSubPutative,
					 std::vector<std::vector<T> > & corrSubLearned,
					 const std::vector<double> sigma,
					 const float tau = 2
					 ){

	int c2 = detectInfluentals(corrSubRef,corrSubPutative,corrSubLearned, sigma, tau);
	
	// If outliers could be found, reject them from subset
	if(c2){
		int out = rejectInfluentals(corrSubRef, corrSubPutative);
		return out;
	}
	else
		return 0;

}

// Detect and mark (zero vector) outliers by comparing found 
// correspondences with estimated ones from the learned function.
template <typename T>
int detectInfluentals(std::vector<std::vector<T> > & corrSubRef, 
					std::vector<std::vector<T> > & corrSubPutative,
					std::vector<std::vector<T> > & corrSubLearned,
					const std::vector<double> & sigma,
					const float tau = 2){
	
	
	int infl = 0;				
	int corrSize = corrSubPutative.size();
	std::vector<T> out(3);
	out[0] = T(0), out[1] = T(0), out[2] = T(0);

	for(int i=0; i<corrSize; i++){
		// If outlier is detected, make points a zero vector
		if(!checkForInlier(corrSubPutative[i], corrSubLearned[i], sigma, tau)){
			corrSubRef[i] = out;
			corrSubPutative[i] = out;
			corrSubLearned[i] = out;
			infl++;
		}
	}
	
	return infl;
}


// Compares the predicted value (x or y) with the putative correspondence value
template <typename T>
bool checkForInlier(const std::vector<T> & corrPutative, 
					const std::vector<T> & corrLearned, 
					const std::vector<double> & sigma, 
					const float tau = 2){

	// Calculate absolute error
	const double error1 = std::abs(double(corrPutative[0]-corrLearned[0]));
	const double error2 = std::abs(double(corrPutative[1]-corrLearned[1]));

	const double threshold1 = tau * sigma[0];
	const double threshold2 = tau * sigma[1];
	
	//sout << "error1 = " << error1 << " / threshold1 = " << threshold1 << std::endl;
	//sout << "error2 = " << error2 << " / threshold2 = " << threshold2 << std::endl;
	
	if(error1 > threshold1 || error2 > threshold2){
		return false;
	}
	else
		return true;

}


// Normalizes 2D feature correspondences so the points are centered at origin and have a mean length of sqrt(2).
template <typename T>
cimg_library::CImg<T> normalizeCorrespondences(std::vector<std::vector<std::vector<T> > > & corr){

	sout << "Normalize Correspondences" << std::endl;

	// Get number of correspondences (there are 3 values per point)
	const size_t dimN = corr.size();
	const size_t dimM = corr[0].size();
	sout << "dimN=" << dimN << std::endl;
	sout << "dimM=" << dimM << std::endl;

	// Get centroids of the two feature lists
	const std::vector<T> centroid = getCentroid(corr);
	sout << "Centroid at = (" << centroid[0] << ", " << centroid[1] << ", " << centroid[2] << ")" << std::endl;

	// Translating correspondences
	for(size_t i=0; i<dimN; i++){
		
		if(dimM != corr[i].size())
			throw(std::runtime_error("normalizeCorrespondences(): different number of points per image!"));

		for(size_t j=0; j<dimM; j++){
			if(checkPoint(corr[i][j])){
			corr[i][j][0] -= centroid[0];
			corr[i][j][1] -= centroid[1];
			}
		}
	}

	// Get scale factor from RMS distance
	T scale = getRMS(corr);
	sout << "RMS = " << scale << std::endl;
		
	// Apply scale factor
	if(scale != 0){
		for(size_t i=0; i<dimN; i++){
			for(size_t j=0; j<dimM; j++){
				if(checkPoint(corr[i][j]))
				{
					corr[i][j][0] *= 1/scale;
					corr[i][j][1] *= 1/scale;
					//sout << "Pt(" << corr[i][j][0] << "," << corr[i][j][1] << ")" << "\timgIndex=" << i << ";\tpointIndex=" << j << std::endl;
				}
			}
		}
	}
	else{
		throw std::runtime_error("normalizeCorrespondences(): Division by zero.") ;
	}


	//Generation of the translation matrices ------------------------
	cimg_library::CImg<T> tr1(3,3), sc1(3,3);

	//Translation matrix
	tr1(0,0) = (T)1; tr1(1,0) = (T)0; tr1(2,0) = -centroid[0];
	tr1(0,1) = (T)0; tr1(1,1) = (T)1; tr1(2,1) = -centroid[1];
	tr1(0,2) = (T)0; tr1(1,2) = (T)0; tr1(2,2) = (T)1;

	//Scaling matrix of Img 1
	sc1(0,0) = 1/scale;	sc1(1,0) = 0;		sc1(2,0) = 0;
	sc1(0,1) = 0;		sc1(1,1) = 1/scale;	sc1(2,1) = 0;
	sc1(0,2) = 0;		sc1(1,2) = 0;		sc1(2,2) = 1;

	//Generation of Transformation Matrices (combination of scaling and translation)
	cimg_library::CImg<T> trafo;
	trafo = sc1*tr1;

	return trafo;

}

// Calculate centroid of feature list
template <typename T>
std::vector<T> getCentroid(const std::vector<std::vector<std::vector<T> > > & corr){

	// Get number of features
	const size_t dimN = corr.size();
	const size_t dimM = corr[0].size();

	std::vector<T> centroid;
	centroid.resize(3);
	centroid[0] = T(0);
	centroid[1] = T(0);
	centroid[2] = T(1);
	
	int dimNInlier = 0;

	for(size_t i=0; i<dimN; i++){

		for(size_t j=0; j<dimM; j++){

			// Check points for outliers
			if(checkPoint(corr[i][j])){
				centroid[0] += corr[i][j][0];
				centroid[1] += corr[i][j][1];
				dimNInlier++;
			}
		}

	}

	centroid[0] /= T(dimNInlier);
	centroid[1] /= T(dimNInlier);

	return centroid;

}

// Get RMS distance
template <typename T>
T getRMS(const std::vector<std::vector<std::vector<T> > > & corr){

	// Get number of features
	const size_t dimN = corr.size();
	const size_t dimM = corr[0].size();

	T rms=0;

	int dimNInlier = 0;

	for(size_t i=0; i<dimN; i++){

		for(size_t j=0; j<dimM; j++){

			// Check points for outliers
			if(checkPoint(corr[i][j])){
				double x=corr[i][j][0];
				double y=corr[i][j][1];
				rms += T( pow( x*x+y*y , 0.5) );
				dimNInlier++;
			}
		}

	}
			
	rms /= T(dimNInlier);

	T s1 = rms/T(sqrt(2.00));
	return s1;
}


// Denormalizes 2D feature correspondences
template <typename T>
void denormalizeCorrespondences(std::vector<std::vector<std::vector<T> > > & corr, const cimg_library::CImg<T> & trafo){

	const size_t dimN = corr.size();
	const size_t dimM = corr[0].size();

	// Translating correspondences
	for(size_t i=0; i<dimN; i++){
		for(size_t j=0; j<dimM; j++){
			if(checkPoint(corr[i][j])){
				cimg_library::CImg<T> vector(1,3);
				vector[0] = corr[i][j][0];
				vector[1] = corr[i][j][1];
				vector[2] = corr[i][j][2];
				vector = trafo.get_invert() * vector;
				vector /= vector[2];
				corr[i][j][0] = vector[0];
				corr[i][j][1] = vector[1];
				corr[i][j][2] = vector[2];
				
			}
		}
	}
	
}


/// Reject outliers from complete correspondence list
template <typename T>
int rejectOutliers(std::vector<std::vector<std::vector<T> > > & corr){

	bool completed = false;
	int out = 0;

	// Check correspondences for outliers. If an outlier is found,
	// erase point in all correspondence vectors and restart
	// outlier search.
	while(!completed){

		const size_t dimN = corr.size();
		const size_t dimM = corr[0].size();

		completed = true;
		for(size_t i=0; i<dimN; i++){

			for(size_t j=0; j<dimM; j++){
				if(!checkPoint(corr[i][j])){
					eraseCorr(corr,j);
					completed = false;
					out++;
					break;
				}
			}

			if(!completed) break;
		}
	}

	return out;

}


/// Reject influentals from a subset
template <typename T>
int rejectInfluentals(std::vector<std::vector<T> > & corr1, std::vector<std::vector<T> > & corr2){

	bool completed = false;
	int out = 0;

	// Check correspondences for outliers. If an outlier is found,
	// erase point in all correspondence vectors and restart
	// outlier search.
	
	while(!completed){

		const size_t dimM = corr1.size();

		completed = true;

		for(size_t j=0; j<dimM; j++){
			if(!checkPoint(corr1[j])){
				corr1.erase(corr1.begin()+j);
				corr2.erase(corr2.begin()+j);
				completed = false;
				out++;
				break;
			}
		}
	}

	return out;

}


// Erase all features assiciated with point i
template <typename T>
void eraseCorr(std::vector<std::vector<std::vector<T> > > & corr, const int position){
	
	const size_t dimN = corr.size();
	for(size_t i=0; i<dimN; i++){
		corr[i].erase(corr[i].begin()+position);
	}

}


// Detect outliers of two correspondence views
template <typename T>
int detectOutliers(std::vector<std::vector<T> > & corrRef, 
	std::vector<std::vector<T> > &corrPutative, size_t id1, size_t id2,
	CorrView<T> & S, std::vector<std::vector<bool> > &_outliermask,
	const double epsilon = 2.0
	){
	
	int outliers = 0;
	sout << "Confidence level for outliers = " << epsilon << std::endl;

	//	Calculate predicted views
	std::vector<std::vector<T> > corrPredicted;
	corrPredicted = S.predictView(corrRef);

	// Calculate cost vector according to the paper p. 20
	cimg_library::CImg<T> D(2,2);
	D.fill(T(0));
	D(0,0) = T( S.sigmaX() * S.sigmaX() );
	D(1,1) = T( S.sigmaY() * S.sigmaY() );

	int corrSize = corrPredicted.size();
	for(int i=0; i<corrSize; i++){

		cimg_library::CImg<T> e(2,1);	
		e[0] = corrPredicted[i][0] - corrPutative[i][0];
		e[1] = corrPredicted[i][1] - corrPutative[i][1];

		double c;
		c = (double)( e * D.get_invert() * e.get_transpose() )[0];

		if(c > epsilon){
			outliers++;
			_outliermask[id1][i] = true;
			_outliermask[id2][i] = true;
			//sout << "Outlier detected:" << std::endl;
			//sout << "c = " << c << std::endl;
		}
		else
		{
			_outliermask[id1][i] = false;
			_outliermask[id2][i] = false;
		}

	}

	return outliers;

}


/// Checks if point is an outlier, if yes the function returns false
template <typename T>
bool checkPoint(std::vector<T> point){
	if(point[0] == 0 && point[1] == 0 && point[2] == 0)
		return false;
	else
		return true;
}


std::vector<std::vector<bool> > getOutlierMask(size_t nImg, size_t nPoints)
{

	std::vector<std::vector<bool> > outliermask;
	outliermask.resize(nImg);

	for(size_t i=0; i<nImg; i++)
	{
		for(size_t j=0; j<nPoints; j++)
		{
			outliermask[i].push_back(false);
		}
	}

	return outliermask;

}


#endif /* _CORRESPONDENCEOUTLIERELIMINATION_HXX_ */
