/*  Copyright (C) 2011 Stephan Meister

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CorrespondenceReadWrite.hxx
 *  Implementation of the parameter class CorrespondenceReadWrite.
 *  \author <a href="mailto:stephan.meister@iwr.uni-heidelberg.de">
 *      Stephan Meister</a>
 *  \date 23.11.2011
 */

#ifndef _CORRESPONDENCEREADWRITE_HXX_
#define _CORRESPONDENCEREADWRITE_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/CorrespondenceReadWrite.h>
#include <fstream>

#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>

class comment_exception : public std::exception
{
public:
		
	comment_exception(const std::string& comment) : _comment(comment)
	{	;	}

	virtual const char* what() const throw()
	{	return _comment.c_str() ;	}

private:

	std::string _comment ;
} ;
	
template <typename Iterator, typename T>
typename hekate::Correspondence<T> parseCorrespondence(Iterator first, Iterator last)
{
	using boost::spirit::qi::double_;
	using boost::spirit::qi::float_ ;
	using boost::spirit::qi::char_ ;
	using boost::spirit::qi::uint_ ;
	using boost::spirit::qi::int_ ;
	using boost::spirit::qi::bool_ ;
	using boost::spirit::qi::lit ;
	using boost::spirit::qi::_1;
	using boost::spirit::qi::phrase_parse;
	using boost::spirit::ascii::space;
	namespace phx = boost::phoenix ;
	using boost::phoenix::ref;


	double x,y,z,w ; x = y = z = w = 0.0 ;
	size_t pointIndex = 0 ;
	size_t imgIndex = 0 ;
	int outlier = false ;
	bool isComment = false;
	std::string comment ;
	float R,G,B ; R = G = B = 0.0 ;
	std::vector< size_t > correspondences;
	//use boost.spirit to parse Correspondence file
	bool r = phrase_parse(first, last,

		//  Begin grammar
		(
			(
				(char_('#')[ref(isComment) = true] >> *(char_[boost::phoenix::ref(comment) += _1])) || //match line beginning with #, interpret as comment
				(double_[ref(x) = _1] >> double_[ref(y) = _1] >> 
				double_[ref(z) = _1] >> double_[ref(w) = _1] >> 
				uint_[ref(pointIndex) = _1] >> uint_[ref(imgIndex) = _1] >> //match 4 doubles and 2 uint as minimum requirement (x,y,z,w,indx,index)
				-(int_[ref(outlier) = _1]) >> -(float_[ref(R) = _1] >> float_[ref(G) = _1] >> float_[ref(B) = _1])) //optional 0 or 1 as bool (outlier) and 3 float (r,g,b)
			)
		),
		//  End grammar

		space); //delimiter symbol

	if (!r || first != last) // fail if we did not get a full match
		throw std::invalid_argument("") ;
	if(isComment)
		throw comment_exception(comment) ; //throw comment line as exception
	hekate::Correspondence<T> c(x,y,z,w,pointIndex,imgIndex,bool(outlier), R,G,B) ;
	return c ;
}


template <typename T>
CorrespondenceReadWrite<T>::CorrespondenceReadWrite(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CorrespondenceReadWrite", name,
			"<h2>read or write list Correspondences from file</h2><br>"
			"File structure (ASCII):<br>"
			"(double double double double uint uint [0||1] [float float float]) || (# string)}<br>"
			"x y z w pointIndex imgIndex outlier R G B<br>"
			"One Correspondence per line"
			"lines starting with # are interpreted as comments, empty lines are ignored"
			"<br>When the input slot is conected, "
			"this plugin will act as an output plugin "
			" and can pass the input through to the output slot<br>"
			"Otherwise it will act as an input plugin and read "
			"correspondences from file<br><br>"
		),
		_in(true, true)
{
	ParameteredObject::_addParameter (_filename, "filename",
		"filename to read from or write to", "fileopen");

	ParameteredObject::_addParameter<bool> (_verbose, "verbose",
		"write correspondences to log",
		false, "bool") ;

	ParameteredObject::_addParameter(_pass, "passThrough",
		"concatenate multiple input list and pass then to the output slot<br>"
		"switching this option off can save memory",
		false, "bool") ;
	
	ParameteredObject::_addInputSlot(
		_in, "in",
		"input correspondences<br>"
		"when multiple list are connected, they will all be appended to the file in "
		"an arbitrary order! "
		"Reading the same file again will create a single correspondence list!",
		"CorrespondenceList");
	
	ParameteredObject::_addOutputSlot(
		_out, "out",
		"output correspondences<br>"
		"when an the input slot is connected, this output slot will contain"
		"a concatenated list of all input lists",
		"CorrespondenceList");

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void CorrespondenceReadWrite<T>::execute() {

	using namespace hekate ;
	
	CorrespondenceList<T> list ;
	
	//using namespace std;
	if(_filename().empty())
	{	raise("Filename is empty!") ;	}

	if(_in.connected())
	{
		std::ofstream file(_filename()) ;
		if(!file || !file.good())
		{
			raise(std::string("Could not open \"")+ _filename() + "\" for writing!") ;
		}
		for(size_t ii = 0 ; ii < _in.size() ; ii++)
		{
			const CorrespondenceList<T>& list = _in[ii] ;
			if(_verbose)
			{	sout	<< "Writing " << size_t(_in[ii].size()) << " correspondences:\n"
						<< "x\ty\tz\tw\tindex1\tindex2\toutlier\tR\tG\tB\n"
						<< list << "\n" ;
			}
				
			if(!(file << list << "\n")) //<-writing happens here!
			{	raise(std::string("Error while writing to file \"")+ _filename() + "\" !") ;
			}
			if(_pass)
			{
				CorrespondenceList<T>& outputList = _out() ;
				outputList.reserve(outputList.size() + list.size()) ;
				CorrespondenceList<T>::ListAsVec::const_iterator it = list.begin() ;
				for(;it != list.end() ; it++)
				{
					outputList.push_back(*it) ;
				}
			}
		}
	}
	else //read from file
	{
		std::ifstream file(_filename()) ;
		if(!file || !file.good())
		{
			raise(std::string("Could not open \"")+ _filename() + "\" for reading!") ;
		}
		size_t linenumber = 1 ;
		while(file.good())
		{
			std::string line ; 
			getline(file, line) ;
			Correspondence<T> c ;
			if(line.empty())
			{
				linenumber++ ;
				continue ;
			}
			try{
				_out().push_back(parseCorrespondence<std::string::iterator, T>(line.begin(), line.end())) ;
				if(_verbose)
				{	sout << _out().back() << std::endl ;	}
			}
			catch(comment_exception& err)
			{
				sout << "Comment in line " << linenumber << ": " << err.what() << std::endl ;
			}
			catch(std::invalid_argument&)
			{
				sout << "Parse Error in line: " << linenumber << std::endl ;
			}
			linenumber++ ;
		}
		
		/*
		file >> _out() ;
		if(_verbose)
		{
			sout << "x\ty\tz\tw\tindex1\tindex2\toutlier\tRed\tGreen\tBlue\n" ;
			sout << _out() ;
		}
		*/
		_sizeAll = _out().size() ;
		sout << "Read " << _sizeAll << " correspondences" << std::endl ;

	}
	
}

#endif /* _CORRESPONDENCEREADWRITE_HXX_ */
