/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CorrespondenceReader.hxx
 *  Implementation of the parameter class CorrespondenceReader.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 13.01.2014
 */

#ifndef _CORRESPONDENCEREADER_HXX_
#define _CORRESPONDENCEREADER_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "CorrespondenceReader.h"
#include <fstream>

#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>

class comment_exception : public std::exception
{
public:
		
	comment_exception(const std::string& comment) : _comment(comment)
	{	;	}

	virtual const char* what() const throw()
	{	return _comment.c_str() ;	}

private:

	std::string _comment ;
} ;
	
template <typename Iterator, typename T>
typename hekate::Correspondence<T> parseCorrespondence(Iterator first, Iterator last)
{
	using boost::spirit::qi::double_;
	using boost::spirit::qi::float_ ;
	using boost::spirit::qi::char_ ;
	using boost::spirit::qi::uint_ ;
	using boost::spirit::qi::int_ ;
	using boost::spirit::qi::bool_ ;
	using boost::spirit::qi::lit ;
	using boost::spirit::qi::_1;
	using boost::spirit::qi::phrase_parse;
	using boost::spirit::ascii::space;
	namespace phx = boost::phoenix ;
	using boost::phoenix::ref;


	double x,y,z,w ; x = y = z = w = 0.0 ;
	size_t pointIndex = 0 ;
	size_t imgIndex = 0 ;
	int outlier = false ;
	bool isComment = false;
	std::string comment ;
	float R,G,B ; R = G = B = 0.0 ;
	std::vector< size_t > correspondences;
	//use boost.spirit to parse Correspondence file
	bool r = phrase_parse(first, last,

		//  Begin grammar
		(
			(
				(char_('#')[ref(isComment) = true] >> *(char_[boost::phoenix::ref(comment) += _1])) || //match line beginning with #, interpret as comment
				(double_[ref(x) = _1] >> double_[ref(y) = _1] >> 
				double_[ref(z) = _1] >> double_[ref(w) = _1] >> 
				uint_[ref(pointIndex) = _1] >> uint_[ref(imgIndex) = _1] >> //match 4 doubles and 2 uint as minimum requirement (x,y,z,w,indx,index)
				-(int_[ref(outlier) = _1]) >> -(float_[ref(R) = _1] >> float_[ref(G) = _1] >> float_[ref(B) = _1])) //optional 0 or 1 as bool (outlier) and 3 float (r,g,b)
			)
		),
		//  End grammar

		space); //delimiter symbol

	if (!r || first != last) // fail if we did not get a full match
		throw std::invalid_argument("") ;
	if(isComment)
		throw comment_exception(comment) ; //throw comment line as exception
	hekate::Correspondence<T> c(x,y,z,w,pointIndex,imgIndex,bool(outlier), R,G,B) ;
	return c ;
}

template <typename T>
CorrespondenceReader<T>::CorrespondenceReader(const std::string& name) :
		TemplatedParameteredObject<T>(
			"CorrespondenceReader", name,
			"<h2>Read from correspondence file</h2><br>"
			"Read from correspondence file"
		)
{
	ParameteredObject::_addOutputSlot(
		correspondence, "correspondence",
		"Read correspondence",
		"CorrespondenceList");

	ParameteredObject::_addParameter(
		filename, "filename",
		"File to read from",
		"FileOpen");

}

template <typename T>
void CorrespondenceReader<T>::execute() {

	using namespace hekate ;
	
	CorrespondenceList<T> &out = this->correspondence(); ;

	std::ifstream file(this->filename()) ;
	if(!file || !file.good())
	{
		raise(std::string("Could not open \"")+ this->filename() + "\" for reading!") ;
	}
	size_t linenumber = 1 ;
	while(file.good())
	{
		std::string line ; 
		getline(file, line) ;
		Correspondence<T> c ;
		if(line.empty())
		{
			linenumber++ ;
			continue ;
		}
		try{
			out.push_back(parseCorrespondence<std::string::iterator, T>(line.begin(), line.end())) ;
		}
		catch(comment_exception& err)
		{
			sout << "Comment in line " << linenumber << ": " << err.what() << std::endl ;
		}
		catch(std::invalid_argument&)
		{
			sout << "Parse Error in line: " << linenumber << std::endl ;
		}
		linenumber++ ;
	}

}

#endif /* _CORRESPONDENCEREADER_HXX_ */
