/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Create3DMultiPointCloud.h
 *  Declaration of the parameter class Create3DMultiPointCloud.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 09.02.2012
 */

#ifndef _CREATE3DMULTIPOINTCLOUD_H_
#define _CREATE3DMULTIPOINTCLOUD_H_

#ifdef _MSC_VER
	#ifdef create3dmultipointcloud_EXPORTS
		/// Visual C++ specific code
		#define create3dmultipointcloud_DECLDIR __declspec(dllexport)
	#else
		#define create3dmultipointcloud_DECLDIR __declspec(dllimport)
	#endif /*Export or import*/
#else
/// Not needed without MSVC
	#define create3dmultipointcloud_DECLDIR
#endif

#include <charon-core/ParameteredObject.h>
#include <charon-utils/CImg.h>
#include <hekate/Correspondence.hxx>
#include <hekate/IndexedMatrix.h>
#include <hekate/ParallelBundleAdjustment.h>
#include <hekate/BackProjection3D2D.h>
#include <iostream>

#ifdef USE_PCL
	#ifdef Success
		#undef Success
	#endif
	#include <pcl/visualization/cloud_viewer.h>
	#include <pcl/io/io.h>
	#include <pcl/io/pcd_io.h>
#endif

#include "hekate/Create3DPointCloud_plugins.h"	

using namespace hekate;
using namespace cimg_library;

struct InitialFrameSet{
	unsigned int frames[3];
};

/// Creates a 3D point cloud by correspondencs and an initial fundamental
/// matrix
/** Creates a 3D point cloud by correspondencs and an initial fundamental
 *  matrix.
 *  Step 1: The fundamental matrix between the initial frame pair (default is frame 0 and
 *  frame 1) is used to determine two camera matrices (P0 and P1). 
 *  Step 2: With P0 and P1 the correspondences are triangulated and an initial 3D
 *  point could is achieved.
 *  Step 3: An initial bundle adjustment is performed on the cams and 3d points
 *  Step 4: 3d points are refined by adjusted cams
 *  Step 5: The 3D-2D correspondences between the initial 3d point cloud and the 2d 
 *  points of the following
 *  frames are used to retrieve a camera matrices of the outstanding frames 
 *  Step 6: A bundle adjustment is performed for every new retrieved camera
 *  Step 7: 2d points of the new view which do not have a 3d correspondence 
 *  are triangulated and corresponding 3d points are achieved
 *  Step 8: All 3d points are refined by all cameras
 *  Step 9: A bundle adjustment is performed on all cameras
 *  Step 10: Repeat step 5 - step 9 until all frames were used for point
 *  generation.
 */
template <typename T>
class create3dmultipointcloud_DECLDIR Create3DMultiPointCloud :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	Create3DMultiPointCloud(const std::string& name = "");
	~Create3DMultiPointCloud();

	/// correspondences
	InputSlot< CorrespondenceList<T> > correspondencesIn;
	/// fundamentalMatrices
	InputSlot< IndexedMatrixList > fundamentalMatricesIn;
	/// calibration matrix
	InputSlot< CImgList<T> > calibrationMatrixIn;
	/// optional initial cameras
	InputSlot< IndexedMatrixList > initialCamerasIn;
	/// initial frames (optional)
	InputSlot< std::vector<int> > initialFramesIn;

	/// cameraMatrices
	OutputSlot< std::vector<IndexedMatrixList> > cameraMatricesOut;
	/// pointCloud
	OutputSlot< std::vector<CorrespondenceList<T> > > pointCloudOut;
	/// calibrationMatrixOut
	//OutputSlot< CImgList<T> > calibrationMatrixOut;
	/// correspondences where more outliers could be detected
	/// compared to the input correspondences
	//OutputSlot< CorrespondenceList<T> > correspondencesOut;

	/// Numer of maximal frames which will be considered
	Parameter< int > maxFrames ;

	/// Show extended log
	Parameter< bool > verbose ;

	/// Show extended log
	Parameter< bool > useInitialCameras ;

	/// useCUDA
	Parameter< bool > useCUDA ;

	/// Assume fixed focal length
	Parameter< bool > fixedFocalLength ;

	/// First initial frame from which the first point cloud is calculated.
	Parameter< unsigned int > initialFrame1;

	/// Second initial frame from which the first point cloud is calculated.
	Parameter< unsigned int > initialFrame2;

	/// Second initial frame from which the first point cloud is calculated.
	Parameter< unsigned int > initialFrame3;

	/// Minimum number of views, a point has to be visible in
	Parameter < int > minPtViews; 	

	/// Minimum number of 3d-2d point correspondences used to retrieve views
	Parameter < int > min3d2dCorrs; 	

	/// Sort out 3d points by back projection errors, while calculating new 3d points.
	Parameter< bool > sortOutByBackProjection ;
	
	/// "Maximal back projection error of a 3d inlier.
	Parameter< float > maxBackProjectionError ;

#ifdef USE_PCL
	/// Shows 3d point cloud in pcl cloud viewer
	Parameter< bool > showPointCloudViewer ;

	/// highFrequencyViewerUpdates
	Parameter< bool > highFrequencyViewerUpdates ;
#endif

	/// Draws outliers in pcl cloud viewer
	Parameter< bool > drawOutliers ;

	/// Perform bundle adjustment for every new camera
	Parameter< bool > pbaAfterEveryIteration ;	

	/// Recover cams by OpenCV solvePnPRANSAC
	Parameter< bool > recoverCamByRANSAC ;

	/// Sort out 3D outliers detected by solvePnPRANSAC
	Parameter< bool > sortOutOutliersByPnPRANSAC ;	
	
	/// Draws outliers in pcl cloud viewer
	Parameter< bool > keepCalibration ;

	/// Use all frames as starting frames for Point Cloud calculation
	Parameter< bool > useAllFramesAsStart ;

	/// Use set frames as starting frames and do not detect them automaticly
	Parameter< bool > forceFixedStartingFrames ;

	/// Number of correspondences between frame 1 and 2
	Parameter< int > nCorrs1;
	/// Number of correspondences between frame 1 and 3
	Parameter< int > nCorrs2;
	/// Number of correspondences between frame 2 and 3
	Parameter< int > nCorrs3;
	/// Maximum number of points (in percent) that may get lost while adding a new camera
	Parameter < double > maxPointLoss;
	/// Update object.
	virtual void execute();
	
#ifdef USE_PCL

	// Show Point Cloud in PCL CloudViewer
	pcl::visualization::CloudViewer *viewer;
	
	/// Update PCL Point cloud and show it in PCL Viewer
	void updatePCLViewer(pcl::visualization::CloudViewer *viewer);

#endif

private:

	bool isInitialized;

	//bool addNewInitialFrames(unsigned int frame1, unsigned int frame2, unsigned int frame3, const IndexedMatrixList &fundamentalMatrices, std::vector<InitialFrameSet> &frameSet);

	void executeTriangulation	(InitialFrameSet startFrames, IndexedMatrixList &cameras, 
								CorrespondenceList<T> &corrs, CorrespondenceList<T> &originalInput, 
								CImgList<T> &calibrations, size_t nIterations, 
								CorrespondenceList<T> &pointCloud, 
								const IndexedMatrixList &fundamentalMatrices, 
								std::vector<InitialFrameSet> &frameSet);

	void initCam(	IndexedMatrixList &cameras, const CImgList<T> &calibrations, 
					const IndexedMatrixList &fundamentalMatrices, CorrespondenceList<T> &originalInput, 
					const unsigned int initFrame1, const unsigned int initFrame2);

	CImg<T> recoverCamByFMatrix( 	IndexedMatrixList &cameras, const CImgList<T> &calibrations, 
									const IndexedMatrixList &fundamentalMatrices, 
									CorrespondenceList<T> &originalInput, unsigned int frame1, 
									unsigned int frame2);

	void triangulatePoints(	CorrespondenceList<T> &corrs, IndexedMatrixList &cameras, 
							CorrespondenceList<T> &pointCloud);

	bool triangulatePoints(	const size_t index, CorrespondenceList<T> &corrs, IndexedMatrixList &cameras, 
							CorrespondenceList<T> &pointCloud);

	CImg<T> getTriangulation(const CImgList<double> &cameraList, const CImgList<double> &pointList);

	CImg<T> getTriangulation(const CImg<T> &P1, const CImg<T> &P2, const CImg<T> &pt1, const CImg<T> &pt2);

	double getDepth(const CImg<T> &P, const CImg<T> &X);

	double calculateBackProjection(	const IndexedMatrixList &cameras, CorrespondenceList<T> &pointCloud, 
									CorrespondenceList<T> &originalInput, 
									const IndexedMatrixList &fundamentalMatrices, 
									std::vector<InitialFrameSet> &frameSet);

	void selectInitialFrames(	size_t nIterations, unsigned int &frame1, 
								unsigned int &frame2, unsigned int &frame3);
/*
#ifdef USE_PCL
	void updatePCLViewer(pcl::visualization::CloudViewer *viewer, CorrespondenceList<T> &pointCloud);
#endif
*/
	void refinePoints(	CorrespondenceList<T> &pointCloud, CorrespondenceList<T> &originalInput, 
						const IndexedMatrixList &cameras);

	bool recoverCameras(	CorrespondenceList<T> &pointCloud, IndexedMatrixList &cameras, 
							CorrespondenceList<T> &originalInput, CorrespondenceList<T> &corrs, 
							CImgList<T> &calibrations, size_t index, bool first, bool sortOut, 
							const IndexedMatrixList &fundamentalMatrices, 
							std::vector<InitialFrameSet> &frameSet);

	
	bool getBestInitialFrames(	CorrespondenceList<T> &corss, unsigned int &frame1, 
								unsigned int &frame2, unsigned &frame3, bool keepFirstFrame, 
								const IndexedMatrixList &fundamentalMatrices, size_t nIterations);
	

	// PBA slots
	OutputSlot< IndexedMatrixList > pbaCam;
	OutputSlot< CImgList<T> > 		pbaCalibration;
	
	// Back projection slots
	OutputSlot< IndexedMatrixList > bpCam;
	
	// Pseudo output slots for internal plugins
	OutputSlot< CorrespondenceList<T> > op_2d, op_3d;

};

#endif // _CREATE3DMULTIPOINTCLOUD_H_
