/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Create3DMultiPointCloud.hxx
 *  Implementation of the parameter class Create3DMultiPointCloud.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 09.02.2012
 */

#ifndef _CREATE3DMULTIPOINTCLOUD_HXX_
#define _CREATE3DMULTIPOINTCLOUD_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/Create3DMultiPointCloud.h>
#include <charon-utils/ImgTool.hxx>

#include <algorithm>
//#define AVOID_SVD
//#define DONT_USE_PLUGINS

#ifdef AVOID_SVD
	#include "hekate/totalLstsqMinimize.h"
#endif

#include <limits>

template <typename T>
Create3DMultiPointCloud<T>::Create3DMultiPointCloud(const std::string& name) :
		TemplatedParameteredObject<T>(
			"Create3DMultiPointCloud", name,
			"<h2>Determines a 3D point cloud and camera matrices by correspondences and an initial "
			"fundamental matrix</h2><br>"
			"Determines a 3D point cloud and camera matrices by correspondences and an initial "
			"fundamental matrix.<br>"
			"Step 1: The fundamental matrix between the first frame pair "
			"(frame1 and frame2) is used to determine two camera matrices "
			"(cam1 and cam2). <br>"
			"Step 2: With cam1 and cam2 the correspondences are triangulated "
			"and a 3D point could is achieved.<br>"
			"Step 3: The 3D point cloud is used together with 2D points of "
			"the next frame (frame3) to retrieve a third camera matrix "
			"(cam3). <br>"
			"Step 4: Repeat step 2 and step 3 until all frames were used for "
			"point generation."
		), 
		isInitialized(0), 
		initialFramesIn(1,0), 
		initialCamerasIn(1,0)
/*
#ifdef USE_PCL
		, viewer(0)
#endif*/
{
	ParameteredObject::_addParameter< int >(
		maxFrames, "maxFrames",
		"Numer of maximal frames which will be considered. (0 = all frames)",
		0, "int");
	ParameteredObject::_addParameter< unsigned int > (initialFrame1, "initialFrame1", 
		"First initial frame from which the first point cloud is calculated.", 0, "int");
	ParameteredObject::_addParameter< unsigned int > (initialFrame2, "initialFrame2", 
		"Second initial frame from which the first point cloud is calculated.", 1, "int");
	ParameteredObject::_addParameter< unsigned int > (initialFrame3, "initialFrame3", 
		"Third initial frame from which the first point cloud is calculated.", 2, "int");

	ParameteredObject::_addParameter< bool >(
		verbose, "verbose",
		"Show extended log",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		useInitialCameras, "useInitialCameras",
		"useInitialCameras",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		useCUDA, "useCUDA",
		"useCUDA",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		fixedFocalLength, "fixedFocalLength",
		"Assume fixed focal length.",
		false, "bool");
	ParameteredObject::_addParameter< int >(
		minPtViews, "minPtViews",
		"Minimum number of views, a point is visible in",
		3, "int");
	ParameteredObject::_addParameter< int >(
		min3d2dCorrs, "min3d2dCorrs",
		"Minimum number of 3d-2d point correspondences used to retrieve views.",
		50, "int");
	
	ParameteredObject::_addParameter< bool >(
		sortOutByBackProjection, "sortOutByBackProjection",
		"Sort out 3d points by back projection errors, while calculating new 3d points.",
		true, "bool");
	ParameteredObject::_addParameter< float >(
		maxBackProjectionError, "maxBackProjectionError",
		"Maximal back projection error of a 3d inlier.",
		2.0, "float");
/*
#ifdef USE_PCL
	ParameteredObject::_addParameter< bool >(
		showPointCloudViewer, "showPointCloudViewer",
		"Shows 3d point cloud in pcl cloud viewer.",
		true, "bool");
	ParameteredObject::_addParameter< bool >(
		highFrequencyViewerUpdates, "highFrequencyViewerUpdates",
		"highFrequencyViewerUpdates",
		false, "bool");
#endif
*/
	ParameteredObject::_addParameter< bool >(
		drawOutliers, "drawOutliers",
		"Draws outliers in pcl cloud viewer.",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		pbaAfterEveryIteration, "pbaAfterEveryIteration",
		"Perform bundle adjustment for every new camera",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		recoverCamByRANSAC, "recoverCamByRANSAC",
		"Recover cams by OpenCV method solvePnPRANSAC",
		false, "bool");	
	ParameteredObject::_addParameter< bool >(
		sortOutOutliersByPnPRANSAC, "sortOutOutliersByPnPRANSAC",
		"Sort out 3D outliers detected by solvePnPRANSAC",
		true, "bool");	
	ParameteredObject::_addParameter< bool >(
		keepCalibration, "keepCalibration",
		"Keep calibration matrix during each pointcloud calculation",
		false, "bool");	
	ParameteredObject::_addParameter< bool >(
		useAllFramesAsStart, "useAllFramesAsStart",
		"Keep calibration matrix during each pointcloud calculation",
		false, "bool");	
	ParameteredObject::_addParameter< bool >(
		forceFixedStartingFrames, "forceFixedStartingFrames",
		"Use set frames as starting frames and do not detect them automaticly",
		false, "bool");	

	ParameteredObject::_addParameter< int >(
		nCorrs1, "nCorrs1",
		"Number of correspondences between frame 1 and 2",
		400, "int");
	ParameteredObject::_addParameter< int >(
		nCorrs2, "nCorrs2",
		"Number of correspondences between frame 1 and 3",
		300, "int");
	ParameteredObject::_addParameter< int >(
		nCorrs3, "nCorrs3",
		"Number of correspondences between frame 2 and 3",
		200, "int");
	ParameteredObject::_addParameter<double>(
		maxPointLoss, "maxPointLoss",
		"Maximum number of points (in percent) that may get lost while adding a new camera",
		0.15, "double");


	ParameteredObject::_addInputSlot(
		correspondencesIn, "correspondences",
		"correspondences",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		fundamentalMatricesIn, "fundamentalMatrices",
		"fundamentalMatrices",
		"IndexedMatrixList");
	ParameteredObject::_addInputSlot(
		calibrationMatrixIn, "calibrationMatrix",
		"optional initial cameras",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		initialCamerasIn, "initialCameras",
		"initialCameras",
		"IndexedMatrixList");
	ParameteredObject::_addInputSlot(
		initialFramesIn, "initialFrames",
		"Initial frames (optional)",
		"vector<int>");

	ParameteredObject::_addOutputSlot(
		cameraMatricesOut, "cameraMatrices",
		"cameraMatrices",
		"vector<IndexedMatrixList>");
	ParameteredObject::_addOutputSlot(
		pointCloudOut, "pointClouds",
		"pointClouds",
		"vector<CorrespondenceList>");
	/*
	ParameteredObject::_addOutputSlot(
		calibrationMatrixOut, "calibrationMatrixOut",
		"calibrationMatrixOut",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		correspondencesOut, "correspondencesOut",
		"correspondencesOut",
		"CorrespondenceList");
	*/

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void Create3DMultiPointCloud<T>::execute() {
	/*
	#ifdef USE_PCL
		if(!viewer)
			viewer = new pcl::visualization::CloudViewer("3D Point Cloud Viewer");
	#endif
	*/
	// Init Plugins
	if(!isInitialized){

		op_2d.init(this, 			"Pseudo2DCorrespondences", 	"CorrespondenceList");
		op_3d.init(this, 			"Pseudo3DCloud", 			"CorrespondenceList");
		bpCam.init(this, 			"bpcam", 					"IndexedMatrixList");
		pbaCam.init(this, 			"pbacam", 					"IndexedMatrixList");
		pbaCalibration.init(this, 	"pbacal", 					"CImgList<T>");

		isInitialized = true;
	}
	
	/* 	Save input correspondences two times:
		One copy is altered during the process by marking outliers and
		the other copy is used for the final triangulation with all correspondences
	*/
	CorrespondenceList<T> input  		= this->correspondencesIn();
	CorrespondenceList<T> originalInput = this->correspondencesIn();

	CImgList<T> startCalibration;

	// If number of known calibrations is not equal number of images
	// take first calibration as default for every image
	if(this->calibrationMatrixIn().size() < originalInput.sizeImages()){

		for(size_t i = 0; i < originalInput.sizeImages(); i++){
			startCalibration.push_back(this->calibrationMatrixIn()[0]);
		}
	}
	// Read in calibrations
	else{
		startCalibration = this->calibrationMatrixIn();
	}

	// Select
	size_t nIterations;
	nIterations = maxFrames() > 2 ? std::min((size_t)maxFrames(), originalInput.sizeImages()) : originalInput.sizeImages(); 

	std::vector<InitialFrameSet>		initialFrames;
	std::vector<CorrespondenceList<T> > pointClouds;
	std::vector<IndexedMatrixList> 		cameras;
	
	
	// Select starting Frames:
	InitialFrameSet startFrames;
	
	if(this->useAllFramesAsStart()){

		// Use every frames as starting frame for a point cloud calculation
		for(int i = 0; i < nIterations; i++){

			startFrames.frames[0] = i;
			
			// Try to find 2 additional images to create a set of 3 starting frames
			if (getBestInitialFrames(input, startFrames.frames[0], startFrames.frames[1], startFrames.frames[2], true, this->fundamentalMatricesIn(), nIterations)){
				sout 	<< "Initial Frames: " << startFrames.frames[0] << ", " << startFrames.frames[1] << ", "
						<<  startFrames.frames[2] << std::endl;
			
				initialFrames.push_back(startFrames);
			}
		}

	}else{

		// Try to find 3 good starting frames
		bool foundStaringFrames = true;

		if(!this->forceFixedStartingFrames()){

			if (getBestInitialFrames(input, startFrames.frames[0], startFrames.frames[1], startFrames.frames[2], false, this->fundamentalMatricesIn(), nIterations)){
				initialFrames.push_back(startFrames);
				sout 	<< "Initial Frames: " << startFrames.frames[0] << ", " << startFrames.frames[1] << ", "
						<<  startFrames.frames[2] << std::endl;
			}else{
				sout << "Could not find any good starting frames, using set ones" << std::endl;
				foundStaringFrames = false;
			}

		}

		// Fallback to the specified images if we could not find a goot starting frame or
		// if the user wants to select them.
		if(!foundStaringFrames || this->forceFixedStartingFrames()){

			selectInitialFrames(nIterations, startFrames.frames[0], startFrames.frames[1], startFrames.frames[2]);
			
			initialFrames.push_back(startFrames);
			sout 	<< "Initial Frames: " << startFrames.frames[0] << ", " << startFrames.frames[1] << ", "
					<<  startFrames.frames[2] << std::endl;	
		}

	}

	CImgList<T> calibration;

	// contains all frames and their overall status
	std::vector<size_t> processedFrames;

	calibration = startCalibration;

	//IndexedMatrixList cameras;
	for(int i = 0; i < initialFrames.size(); i++){

		input  			= this->correspondencesIn();
		originalInput 	= this->correspondencesIn();
		
		pointClouds.push_back(CorrespondenceList<T>());
		cameras.push_back(IndexedMatrixList());
		
		if(!keepCalibration){
			calibration = startCalibration;
		}

		try{
			executeTriangulation(initialFrames[i], cameras.back(), input, originalInput, calibration, nIterations, pointClouds.back(), this->fundamentalMatricesIn(), initialFrames);
		}catch(...){
			sout << "Failed to calculate PointCloud for " << initialFrames[i].frames[0] << ", " << initialFrames[i].frames[1] << ", "
					<<  initialFrames[i].frames[2] << std::endl;

			// Remove the failed calculation from the output list
			pointClouds.pop_back();
			cameras.pop_back();

			continue;
		}

		if (pointClouds.back().size() == 0){

			// Also remove empty point clouds
			pointClouds.pop_back();
			cameras.pop_back();

			continue;			
		}

		for(size_t frame = 0; frame < nIterations; frame++){
				
			if (std::find(processedFrames.begin(), processedFrames.end(), frame) == processedFrames.end()){
				processedFrames.push_back(frame);

				// Add as new initial frames, if the cameras do not exist or are degenerated
				if(!cameras.back().exists(frame) || cameras.back()(frame).degenerated){
					startFrames.frames[0] = frame;

					if (getBestInitialFrames(input, startFrames.frames[0], startFrames.frames[1], startFrames.frames[2], true, this->fundamentalMatricesIn(), nIterations)){
							
							// Check if we already added this set of frames (only check first and second image)
							bool found = false;

							/* MSVC 2010 does not support range based loops
							for (InitialFrameSet &frameSet : initialFrames){
							*/
							for (auto frameSetIt = initialFrames.begin(); frameSetIt != initialFrames.end(); frameSetIt++){

								InitialFrameSet &frameSet = *frameSetIt;

								if (frameSet.frames[0] == startFrames.frames[0] && frameSet.frames[1] == startFrames.frames[1]){
									found = true;
									break;
								}

							}

							if (!found){
								sout 	<< "Add new frames: " << startFrames.frames[0] << ", " << startFrames.frames[1] << ", "
										<<  startFrames.frames[2] << std::endl;

								initialFrames.push_back(startFrames);
							}
					}
				}
			}

		}
	}

	for(int i = 0; i < pointClouds.size(); i++){
		sout << "PointCloud " << i << " size: " << pointClouds[i].size() << " " << cameras[i].size() << std::endl;

	}

	//this->correspondencesOut() 		= originalInput.removeOutliers();
	//this->calibrationMatrixOut() 	= calibration;
	this->pointCloudOut()			= pointClouds;
	this->cameraMatricesOut()		= cameras;

};

template <typename T>
void Create3DMultiPointCloud<T>::executeTriangulation(	InitialFrameSet startFrames, 	IndexedMatrixList &cameras,
														CorrespondenceList<T> &corrs,	CorrespondenceList<T> &originalInput,	
														CImgList<T> &calibrations, size_t nIterations,
														CorrespondenceList<T> &pointCloud,
														const IndexedMatrixList &fundamentalMatrices,
														std::vector<InitialFrameSet> &frameSet){

	sout << "Calculate initial camera pair" << std::endl;
	initCam( cameras, calibrations, fundamentalMatrices, originalInput, startFrames.frames[0], startFrames.frames[1] );

	// Triangulate with initial camera pair
	triangulatePoints( startFrames.frames[0], corrs, cameras, pointCloud );

	// Calculate back projection errors
	double meanBPError = calculateBackProjection(cameras, pointCloud, originalInput, fundamentalMatrices, frameSet);

	// Perform back projection and sort out bad 3d points
	CorrespondenceList<T> backprojections;

	if( this->sortOutByBackProjection() ){
		pluginBackProjection3D2D<T>(cameras, pointCloud, backprojections, originalInput, this->maxBackProjectionError()/*meanBPError*/, this->min3d2dCorrs());	
	}

	IndexedMatrixList 		pbacameraMatricesOut;
	CorrespondenceList<T> 	pbapointCloudOut;
	CImgList<T> 			pbacalibrationMatrixOut;
	
	ParallelBundleAdjustmentPlugin( cameras, calibrations, pointCloud, originalInput, pbacameraMatricesOut, 
									pbapointCloudOut, pbacalibrationMatrixOut, 2, this->fixedFocalLength(), this->useCUDA());

	cameras 		= pbacameraMatricesOut;
	calibrations 	= pbacalibrationMatrixOut;
	pointCloud 		= pbapointCloudOut;
	
	// Calculate back projection errors
	meanBPError = calculateBackProjection(cameras, pointCloud, originalInput, fundamentalMatrices, frameSet);

	// Refine all 3D points by all available views
	refinePoints(pointCloud, originalInput, cameras);

	/*
	// Update viewer
	#ifdef USE_PCL
	if( this->showPointCloudViewer() && this->highFrequencyViewerUpdates() ) 
		updatePCLViewer(viewer);
	#endif
	*/

	recoverCameras(pointCloud, cameras, originalInput, corrs, calibrations, startFrames.frames[2], true, true, fundamentalMatrices, frameSet);

	int badCounter = 0;
	for(size_t i = 0; i < nIterations; i++)
	{

		if( i == startFrames.frames[0] || i == startFrames.frames[1] || i == startFrames.frames[2])
			continue;
		
		sout << "Trying to estimate " << nIterations - i << " more camera views." << std::endl;

		if(!recoverCameras(pointCloud, cameras, originalInput, corrs, calibrations, i, false, sortOutOutliersByPnPRANSAC(), fundamentalMatrices, frameSet)){
			badCounter++;
		}
		/*
		if (badCounter > 10){
			sout << "---> Stopping calculation of this point cloud as we got more than 10 bad cameras" << std::endl;
			break;
		}
		*/
	}

	// Update viewer
	/*
	#ifdef USE_PCL
	if( this->showPointCloudViewer() && this->highFrequencyViewerUpdates() ) 
		updatePCLViewer(viewer);
	#endif
	*/
	
	// Perform sparse bundle adjustment
	pbacameraMatricesOut.clear();
	pbapointCloudOut.clear();
	pbacalibrationMatrixOut.clear();

	ParallelBundleAdjustmentPlugin( cameras, calibrations, pointCloud, originalInput, pbacameraMatricesOut, 
									pbapointCloudOut, pbacalibrationMatrixOut, this->minPtViews(), this->fixedFocalLength(), this->useCUDA());

	cameras 		= pbacameraMatricesOut;
	calibrations 	= pbacalibrationMatrixOut;
	pointCloud 		= pbapointCloudOut;
	
	if(sortOutOutliersByPnPRANSAC()){
	
		pointCloud.removeOutliers();

		triangulatePoints( corrs, cameras, pointCloud );
		//triangulatePoints( originalInput, cameras, pointCloud );
		backprojections.clear();
		
		if( this->sortOutByBackProjection() ){
			pluginBackProjection3D2D<T>(cameras, pointCloud, backprojections, originalInput, maxBackProjectionError(), this->min3d2dCorrs());
		}

		ParallelBundleAdjustmentPlugin( cameras, calibrations, pointCloud, originalInput, pbacameraMatricesOut, 
										pbapointCloudOut, pbacalibrationMatrixOut, this->minPtViews(), this->fixedFocalLength(), this->useCUDA(), false, true);
		
		cameras 		= pbacameraMatricesOut;
		calibrations 	= pbacalibrationMatrixOut;
		pointCloud 		= pbapointCloudOut;		
	}

	//if(this->verbose())
	//	calculateBackProjection(cameras, pointCloud, originalInput, fundamentalMatrices, initialFrames);


}


// Calculate initial camera pair
template <typename T>
void Create3DMultiPointCloud<T>::initCam(IndexedMatrixList &cameras, 	const CImgList<T> &calibrations,
										const IndexedMatrixList &fundamentalMatrices,
										CorrespondenceList<T> &originalInput, 
										const unsigned int initFrame1, 	const unsigned int initFrame2)
{

	if( this->initialCamerasIn.connected() && this->useInitialCameras() ){		
		cameras.push_back( this->initialCamerasIn()[0] );
		cameras.push_back( this->initialCamerasIn()[1] );
	}else{

		// read in calibration matrices
		cimg_library::CImg<T> cal1 = calibrations[initFrame1];
		
		// Generate camera at origin as first camera.
		CImg<double> CP1(4,3);
		IndexedMatrix P1, P2;

		CP1.fill(0.); CP1(0,0) = 1; CP1(1,1) = 1; CP1(2,2) = 1;
		P1 = IndexedMatrix(cal1 * CP1);
		P1.setIndex( initFrame1 );
		cameras.push_back(P1);
	
		// Get and save second camera P2
		P2 = IndexedMatrix(recoverCamByFMatrix(cameras, calibrations, fundamentalMatrices, originalInput, initFrame1, initFrame2 ));
		P2.setIndex( initFrame2 );
		cameras.push_back(P2);

	}

}

// Recover camera matrices by x, x', F
template <typename T>
CImg<T> Create3DMultiPointCloud<T>::recoverCamByFMatrix( IndexedMatrixList &cameras, const CImgList<T> &calibrations,
													const IndexedMatrixList &fundamentalMatrices, 
													CorrespondenceList<T> &originalInput,
													unsigned int frame1, unsigned int frame2){
	
	sout << "Get initial cameras P" << frame1 << " and P" << frame2
		<< " by fundamental matrix." << std::endl;

	// read in calibration matrices
	cimg_library::CImg<T> cal1, cal2;
	cal1 = calibrations[frame1];
	cal2 = calibrations[frame2];
	
	ImgTool::printInfo(sout,cal2,"	initial calibration 2: ");
		
	// Get identity camera matrix (at origin)
	CImg<T> P1(4,3);
	P1.fill(0.); P1(0,0) = 1; P1(1,1) = 1; P1(2,2) = 1;

	// Calculate second camera matrix
	CImg<T> fmatrix(3,3);
	fmatrix = fundamentalMatrices(frame1,frame2);

	// Convert fmatrix to essential matrix
	fmatrix = cal2.get_transpose()*fmatrix*cal1;

	// Calculated SVD of the fundamental matrix.
	CImg<T> U, S, V, W(3,3);
	fmatrix.SVD(U,S,V);

	W.fill(T(0.));
	W(1,0) = (T)-1;
	W(0,1) = (T)1;
	W(2,2) = (T)1;

	// Declare possible cameras
	CImg<T> P2_1, P2_2, P2_3, P2_4, u(1,3);
	u=U.get_crop(2,0,2,2);

	// Possible cam 1
	P2_1 = U*W*V.get_transpose();
	P2_1.append(u);

	// Possible cam 2
	P2_2 = U*W*V.get_transpose();
	P2_2.append(-u);

	// Possible cam 3
	P2_3 = U*W.get_transpose()*V.get_transpose();
	P2_3.append(u);

	// Possible cam 4
	P2_4 = U*W.get_transpose()*V.get_transpose();
	P2_4.append(-u);

	// Select camera where point correspondences are reconstructed
	// with a positive depth in front of the image plane. Only
	// one camera (of the above four) will satisfiy this constraint.
	CImg<T> pt1(1,3), pt2(1,3), X1, X2, X3, X4;


	// Get correspondence inlier to check cameras
	auto it3 	= originalInput.get<imgIndex>().lower_bound(frame1) ;
	auto it3End = originalInput.get<imgIndex>().upper_bound(frame2) ;
	
	for(int j = 0; it3 != it3End ; it3++, j++){
		try{
			const Correspondence<T> &pt2Corr = originalInput(frame2, it3->pointIndex);
			if(!it3->outlier && !pt2Corr.outlier)
			{
				pt1 = it3->pt2d();
				pt2 = pt2Corr.pt2d();
				break;
			}
		}
		catch(...){}
	}

	// normalize points
	pt1 = cal1.get_invert() * pt1;
	pt2 = cal2.get_invert() * pt2;

	//ImgTool::printInfo(sout,pt2.get_transpose()*fmatrix*pt1,"e2^TEe1==0?");


	// triangulate points to receive depth
	X1 = getTriangulation(P1, P2_1, pt1, pt2);
	X2 = getTriangulation(P1, P2_2, pt1, pt2);
	X3 = getTriangulation(P1, P2_3, pt1, pt2);
	X4 = getTriangulation(P1, P2_4, pt1, pt2);
			

	CImg<T> P2;
	// Select correct camera.
	if(			getDepth(P1, X1) > 0. && getDepth(P2_1, X1) > 0.){
		P2 = P2_1;
	}else if(	getDepth(P1, X2) > 0. && getDepth(P2_2, X2) > 0.){
		P2 = P2_2;
	}else if(	getDepth(P1, X3) > 0. && getDepth(P2_3, X3) > 0.){
		P2 = P2_3;
	}else if(	getDepth(P1, X4) > 0. && getDepth(P2_4, X4) > 0.){
		P2 = P2_4;
	}else{
		throw std::runtime_error("It could not be found a camera which reconstructs positive depths");
	}


	// denormalize P_cam -> P = K * P_cam and P' = K' * P'_cam
	// since the determined projection matrix has the form [R|t]
	P2 = cal2 * P2;			
	P1 = cal1 * P1;
	
	sout << "Calculated P"<< frame1 << " and P"<< frame2 <<":" << std::endl;
	ImgTool::printInfo(sout, P1, "P"+frame1);
	ImgTool::printInfo(sout, P2, "P"+frame2);

	// Calculate principal point for evaluation purposes
	sout << "Principal points:" << std::endl;

	CImg<T> pp1 = P1.get_crop(0,0,2,2)*P1.get_crop(0,2,2,2).transpose() ;
	CImg<T> pp2 = P2.get_crop(0,0,2,2)*P2.get_crop(0,2,2,2).transpose() ;

	sout 	<< "Principal Point of P"<< frame1 << ": (" << pp1[0] << "," 
			<< pp1[1] << "," << pp1[2] << ")" << std::endl;
	sout 	<< "Principal Point of P"<< frame2 << ": (" << pp2[0] << "," 
			<< pp2[1] << "," << pp2[2] << ")" << std::endl;

	return P2;

}

template <typename T>
void Create3DMultiPointCloud<T>::triangulatePoints(CorrespondenceList<T> &corrs, 
											  IndexedMatrixList &cameras, CorrespondenceList<T> &pointCloud)
{
	for(size_t i = 0; i < cameras.size(); i++)
	{
		triangulatePoints( cameras[i].getIndex1(), corrs, cameras, pointCloud);
	}

}

template <typename T>
bool Create3DMultiPointCloud<T>::triangulatePoints(const size_t index, CorrespondenceList<T> &corrs, 
											  IndexedMatrixList &cameras, CorrespondenceList<T> &pointCloud)
{
	sout << "Triangulate points..." << std::endl;

	int n1=0, n2=0, n3=0, n4=0, n5=0, n6=0;
	
	// 1) Iterate over every 2d point per camera
	// 2) Check 2d point is an outlier -> continue with next iteration
	// 3) If the corresponding 3D point already exists -> continue with next iteration
	// 4) Check if there are 2d correspondences with a known camera
	// 5) If there is at least one correspondence, try to triangulate


	// If bad camera, continue with next one
	if(cameras(index).degenerated)
	{
		sout << "Degererated camera -> won't triangulate points." << std::endl;
		return 0;
	}

	auto it 	= corrs.get<imgIndex>().lower_bound(index);
	auto itEnd 	= corrs.get<imgIndex>().upper_bound(index);
	for(; it != itEnd ; it++ )
	{
		
		n1++;
			
		// Continue with next iteration, if point is an outlier
		if(it->outlier){
			n6++;
			continue;
		}

		// If point index already exists in 3d point cloud, continue
		// with next iteration
		if(pointCloud.existPtIndex(it->pointIndex))
		{
			n3++;
			continue;
		}

		std::vector< const Correspondence<T>*> corrPointList = corrs.getListByPointId(it->pointIndex);
		
		CImgList<double> cameraList, pointList;

		for(size_t j = 0; j < corrPointList.size(); j++)
		{

			if( corrPointList[j]->outlier ) 
				continue;

			CImg<double> camTemp;

			// If bad camera, continue
			try{ camTemp = cameras(corrPointList[j]->imgIndex); }
			catch(...){ continue; }

			if( cameras(corrPointList[j]->imgIndex).degenerated )
				continue;

			// Save correspondence in pointList
			pointList.push_back( corrPointList[j]->pt2d() );

			// Save camera in cameraList
			cameraList.push_back( camTemp );

		}

		// If no correspondence could be found
		// continue with next point
		if(pointList.size() < 2){
			n4++;
			continue;
		}
		
		CImg<T> X;
		try{
			X = getTriangulation( cameraList, pointList );
		}catch(...){
			sout << "Point could not be triangulated" << std::endl;
		}

		// check whether the triangulated point lies before or behind the priciple planes
		// the principle plane is the plane the camera center lies on and which is
		// intersected of by the priciple ray perpendicularly.
		bool positiveDepth = true;
		for(size_t j = 0; j < cameraList.size(); j++)
		{
			double depth = getDepth( cameraList[j], X );
			
			//push back 3D corrdinates
			if(depth < 0 ){
				positiveDepth = false;
				n5++;
				break;
			}
		}
			
		// If triangulated point lies in front of all camera pricipal planes
		// save it in 3D point cloud
		if(positiveDepth)
		{
			Correspondence<T> point3D( X, it->pointIndex, cameras(index).getIndex1(), false, it->R, it->G, it->B );
			pointCloud.push_back(point3D);
			n2 ++;
		}

	}

	sout << n2 << " / " << n1 << " points of frame " << index << " could be triangulated" << std::endl;
	sout << n6 << " points in frame " << index << " were outliers." << std::endl;
	sout << n4 << " 2d points do not have inlier correspondences with known cameras." << std::endl;
	sout << n5 << " 3d points do not ly in front of all cameras which project them." << std::endl;
	sout << n3 << " 2d points have already a corresponding 3d point." << std::endl;

	return 1;

}



/// Triangulate 2D points pt1 and pt2 with cameras P1 and P2 to retrieve the 3D point. (Hartley p.592)
template <typename T>
CImg<T> Create3DMultiPointCloud<T>::getTriangulation(
	const CImgList<double> &cameraList, const CImgList<double> &pointList){
				
	if( pointList.size() < 2 || cameraList.size() < 2 )
		throw( std::range_error("Create3DMultiPointCloud<T>::getTriangulation: No correspondence or camera")) ;

	// define matrix A = x cross X and x' cross X
	CImg<T> A( 4, cameraList.size() * 2 ) ;
	
	for(size_t i = 0; i < pointList.size(); i++)
	{
		//first row of A
		A(0,i*2) = pointList[i](0,0)*cameraList[i](0,2) - cameraList[i](0,0);
		A(1,i*2) = pointList[i](0,0)*cameraList[i](1,2) - cameraList[i](1,0);
		A(2,i*2) = pointList[i](0,0)*cameraList[i](2,2) - cameraList[i](2,0);
		A(3,i*2) = pointList[i](0,0)*cameraList[i](3,2) - cameraList[i](3,0);
		//second row of A
		A(0,i*2+1) = pointList[i](0,1)*cameraList[i](0,2) - cameraList[i](0,1);
		A(1,i*2+1) = pointList[i](0,1)*cameraList[i](1,2) - cameraList[i](1,1);
		A(2,i*2+1) = pointList[i](0,1)*cameraList[i](2,2) - cameraList[i](2,1);
		A(3,i*2+1) = pointList[i](0,1)*cameraList[i](3,2) - cameraList[i](3,1);
	}

	// determine 3D point via SVD of A (Hartley p.592)

	#ifdef AVOID_SVD

		T result[4];
		totalLstsqMinimizeImg(A, (T*)&result, false);

		cimg_library::CImg<T> X(1,4);

		X(0, 0) = result[0] / result[3]; 				
		X(0, 1) = result[1] / result[3]; 
		X(0, 2) = result[2] / result[3]; 
		X(0, 3) = 1; 

	#else

		cimg_library::CImg<T> X(1,4), U, S, V;
		A.SVD(U, S, V);

		X = V.get_crop(3,0,3,3);
		X /= X(0,3);

	#endif

	return X;

}

/// Triangulate 2D points pt1 and pt2 with cameras P1 and P2 to retrieve the 3D point. (Hartley p.592)
template <typename T>
CImg<T> Create3DMultiPointCloud<T>::getTriangulation(
		const CImg<T> &P1, const CImg<T> &P2, 
		const CImg<T> &pt1, const CImg<T> &pt2){
		
	CImg<T> X;
	CImgList<double> cameraList, pointList;

	cameraList.push_back(P1);
	cameraList.push_back(P2);
	
	pointList.push_back(pt1);
	pointList.push_back(pt2);

	X = getTriangulation(cameraList, pointList);

	return X;

}

/// Calculate depth between x and principle plane of P.
template <typename T>
double Create3DMultiPointCloud<T>::getDepth(const CImg<T> &P, const CImg<T> &X) {

	int signP = P.get_crop(0,0,2,2).det() < 0 ? -1 : 1;

	double P_m3_abs = P(0,2)*P(0,2)+P(1,2)*P(1,2)+P(2,2)*P(2,2);
	P_m3_abs = pow(P_m3_abs,.5);

	double depth;
	depth = signP * (P*X)(0,2);
	
	if(P_m3_abs!=0) 
		depth /= P_m3_abs;

	return depth;

}

/// Calculate back projection error in every image
template <typename T>
double Create3DMultiPointCloud<T>::calculateBackProjection(	const IndexedMatrixList &cameras, 
														CorrespondenceList<T> &pointCloud,
														CorrespondenceList<T> &originalInput,
														const IndexedMatrixList &fundamentalMatrices, 
														std::vector<InitialFrameSet> &frameSet){

	sout << "Calculate quality of reconstruction by back projection" << std::endl;
		
	// list of 3d points
	const CorrespondenceList<T> &list3d = pointCloud;

	double meanMean = 0, meanStdDev = 0;
		
	// Iterate over all cameras available
	for(size_t i = 0; i < cameras.size(); i++)
	{

		const IndexedMatrix &P = cameras[i] ;
		if( P.degenerated )
			continue;

		std::vector<double> distVec;
		double distAbs = 0;

		auto it 	= list3d.begin() ;
		auto itEnd 	= list3d.end();
		for(; it != itEnd ; it++ )
		{
			
			if(it->outlier)
				continue;

			try
			{

				// Calculate distance between 2d feature point
				// and back projected 3d point

				// Get inhomogeneous 2D point
				const Correspondence<T> &corr2D = originalInput(P.getIndex1(), it->pointIndex);

				if(corr2D.outlier)
					continue;

				CImg<double> pt2D = (CImg<double>) corr2D.pt2d() ;
				pt2D.crop(0,0,0,1);

				//ImgTool::printInfo(sout, pt2D, "pt2D");

				// Get inhomogeneous backprojected 2D point
				const Correspondence<T> &corr3D = *it ;
				CImg<double> pt3D = (CImg<double>) corr3D.pt3d() ;

				CImg<double> pt2D_back = (CImg<double>) (P * pt3D) ;

				if( pt2D_back[2] == 0 ) 
					continue;

				pt2D_back /= pt2D_back[2];
				pt2D_back.crop(0,0,0,1);
				
				//ImgTool::printInfo(sout, pt2D_back, "pt2D_back");
				
				// Calculate distance pt2D - pt2D_back
				CImg<double> dist = pt2D - pt2D_back ;
				//ImgTool::printInfo(sout, dist, "dist");

				double distVal = std::sqrt(dist.dot(dist)) ;

				distVec.push_back(distVal) ;
				distAbs += distVal ;

			}

			catch(...)
			{
				continue;
			}

		}

		if( verbose() ){
			for(size_t i = 0; i < distVec.size(); i++)
				sout << "distVal = " << distVec[i] << std::endl;
		}


		// Continue with next iteration, if no corresponding
		// 2D - 3D points could be found
		if(distVec.size() == 0)
		{
			sout << "No correspondences between 2d and 3d points "
				<< "in frame " << i << ". Continue with next frame." << std::endl;
			continue;
		}

		double distMean = distAbs / distVec.size() ;

		double stdev = 0;
		for(size_t j = 0; j < distVec.size(); j++)
		{
			stdev += (distVec[j]-distMean)*(distVec[j]-distMean) ;
		}
		stdev = std::sqrt( stdev/distVec.size() );

		sout << "Image[" << P.getIndex1() << "]: mean = " << distMean
			<< "; stdev = " << stdev << ";" << std::endl;

		meanMean 	+= distMean;
		meanStdDev 	+= stdev;

	}

	meanMean 	/= cameras.size();
	meanStdDev 	/= cameras.size();
	/*
	if( isinf(meanMean+meanStdDev) )
	{
		sout << "Error while calculating stdev (=inf) " <<  std::endl;
		return 1;
	}
	*/
	if( meanMean+meanStdDev > 1 && meanMean+meanStdDev < 200 )
		return meanMean+meanStdDev;
	else
	{
		sout << "No proper stdev:\nmeanMean=" << meanMean << "\nmeanStdev="<< meanStdDev << std::endl;
		return 1;
	}

}

/// Refine 3d points by (PM - m = 0) for all views
template <typename T>
void Create3DMultiPointCloud<T>::refinePoints(CorrespondenceList<T> &pointCloud, CorrespondenceList<T> &originalInput, const IndexedMatrixList &cameras)
{
	sout << "-------------------------------" << std::endl;
	sout << "Refine 3D points by all existing views..." << std::endl;

	int nCount = 0, n3D = 0, nOutlier = 0, nLess = 0, nA = 0;
	
	auto it 	= pointCloud.begin();
	auto itEnd 	= pointCloud.end();
	for(; it != itEnd; it++)
	{

		n3D++;

		if(it->outlier){
			nOutlier++;
			continue;
		}

		std::vector< const Correspondence<T>*> ptList =	originalInput.getListByPointId(it->pointIndex);

		// Get number of corresponding inliers
		size_t nPt = 0;
		for(size_t i = 0; i < ptList.size(); i++)
		{	
			if(!ptList[i]->outlier)
				nPt++;
		}


		// Continue if less than two 2d inlier points correspond with 3d point
		if(nPt < 2)
		{
			nLess++;
			continue;
		}

		// Create refinement matrix (Pollefeys, Visual 3D modelling from Images, p.46)
		CImg<double> X = (CImg<double>) it->pt3d();
		//ImgTool::printInfo(sout, X , "X");


		CImg<double> A;
		for(size_t i = 0; i < ptList.size(); i++)
		{	
			
			if(!ptList[i]->outlier)
			{

				try
				{
					// Get camera matrix
					CImg<double> P = cameras(ptList[i]->imgIndex);
					CImg<double> P1 = P.get_crop(0,0,3,0);
					CImg<double> P2 = P.get_crop(0,1,3,1);
					CImg<double> P3 = P.get_crop(0,2,3,2);


					// get 2d coordinates
					double x = ptList[i]->x(), y = ptList[i]->y();

					//sout << "x=" << x << ", y=" << y << std::endl;

					double c = P3.dot(X);
					A.append( (x * P3 - P1) / c, 'y' );
					A.append( (y * P3 - P2) / c, 'y' );

				}
				catch(...){}
			}
		}

		if(A.height() < 4)
		{
			//sout << "Refinement matrix too small" << std::endl;
			//ImgTool::printInfo(sout, A , "A");
			nA++;
			continue;
		}

		#ifdef AVOID_SVD

			T result[4];
			totalLstsqMinimizeImg(A, (T*)&result, false);
			it->x() = result[0] / result[3];
			it->y() = result[1] / result[3];
			it->z() = result[2] / result[3];
			it->w() = 1;

		#else

			CImg<double> S, V, U;
			A.SVD(U, S, V);
			
			CImg<double> Xnew = V.get_crop(3,0,3,3);

			bool isInf = false;
			if(Xnew[3] == 0)
				isInf = true;

			Xnew /= Xnew[3];
			
			it->x() = Xnew[0];
			it->y() = Xnew[1];
			it->z() = Xnew[2];
			it->w() = Xnew[3];
			//it->outlier = true;

		#endif

		nCount++;

		//ImgTool::printInfo(sout, Xnew , "Xnew");

	}

	sout << nCount << "/" << n3D << " 3d points have been refined." << std::endl;
	if(nOutlier > 0 ) sout << nOutlier << " 3D points were outliers." << std::endl;
	if(nLess > 0 ) sout << nLess << " 2D correspondences had too less points." << std::endl;
	if(nA > 0 ) sout << nA << " 3D points had too less views to refine them." << std::endl;
	if(nOutlier + nLess + nA > 0 ) sout << nOutlier + nLess + nA << " 3D points could not be used for refinement." << std::endl;
	sout << "-------------------------------" << std::endl;

}

string convertInt(int number)
{
   stringstream ss;//create a stringstream
   ss << number;//add number to the stream
   return ss.str();//return a string with the contents of the stream
}

template <typename T>
bool Create3DMultiPointCloud<T>::recoverCameras(CorrespondenceList<T> &pointCloud, 		IndexedMatrixList &cameras, 
												CorrespondenceList<T> &originalInput, 	CorrespondenceList<T> &corrs, 
												CImgList<T> &calibrations,
												size_t index, bool first, bool sortOut, const IndexedMatrixList &fundamentalMatrices,
												std::vector<InitialFrameSet> &frameSet){
	
	if (cameras.exists(index))
		return true;

	auto backupCloud 	= pointCloud;
	auto backupCameras 	= cameras;
	bool result			= true;

	sout << std::endl << "------------------------------------" << std::endl;
	sout << "Calculate camera P" << index << std::endl;

	try
	{
		
		IndexedMatrix P, P2;
		
		if( recoverCamByRANSAC() ){
			P = recoverCamRANSAC(index, calibrations, originalInput, pointCloud, 300, this->maxBackProjectionError(), 
				20, 0, IndexedMatrixList(),	sortOut);
		}else{
			P = recoverCam(index, calibrations, originalInput, pointCloud);
		}

		P.setIndex(index);
		P.f = calibrations[index][0];
		cameras.push_back(P);
	
		if(this->pbaAfterEveryIteration()){
				IndexedMatrixList 		pbacameraMatricesOut;
				CorrespondenceList<T> 	pbapointCloudOut;
				CImgList<T> 			pbacalibrationMatrixOut;

				ParallelBundleAdjustmentPlugin( cameras, calibrations, pointCloud, originalInput, pbacameraMatricesOut, 
												pbapointCloudOut, pbacalibrationMatrixOut, 2, this->fixedFocalLength(), this->useCUDA());

				cameras 		= pbacameraMatricesOut;
				calibrations 	= pbacalibrationMatrixOut;
				pointCloud 		= pbapointCloudOut;
		}

		// Triangulate missing points by P_i and all prior cameras
		if( !recoverCamByRANSAC() || !first){
			triangulatePoints(index, originalInput, cameras, pointCloud);
		}
		// Refine all 3D points by all available views
		refinePoints(pointCloud, originalInput, cameras);

		// Sort out 3d points whose back projection
		// has an error taller than this->maxBackProjectionError()
		if(this->sortOutByBackProjection()) // sortOut ?
		{
			CorrespondenceList<T> backprojections;
			pluginBackProjection3D2D<T>(cameras, pointCloud, backprojections, originalInput, this->maxBackProjectionError(), this->min3d2dCorrs());

			if(cameras(index).degenerated){
				result	= false;
				throw(std::runtime_error("Camera degenerated."));
			}
		}
	
		// Update viewer
		/*
		#ifdef USE_PCL
		if( this->showPointCloudViewer() && this->highFrequencyViewerUpdates() ) 
			updatePCLViewer(viewer);
		#endif
		*/

		// Calculate back projection errors
		if(this->verbose())
			calculateBackProjection(cameras, pointCloud, originalInput, fundamentalMatrices, frameSet);

	}
	catch(...)
	{
		sout 	<< "Cam " << index << " could not be reconstructed "
				<< "due to too less matches between already known 3D "
				<< "points and features in image " << index << "." << std::endl;
	}
	
	if (!result || pointCloud.size() < (backupCloud.size() - (backupCloud.size() * this->maxPointLoss()) ) ){

		pointCloud 	= backupCloud;
		cameras		= backupCameras;
		result		= false;
		sout 	<< "Removed camera as it's degenerated or the pointCloud would be more than " 
				<< this->maxPointLoss() * 100 << "%% smaller" << std::endl;
	}

	return result;	
}

// Select initial frames
template <typename T>
void Create3DMultiPointCloud<T>::selectInitialFrames(	size_t nIterations, unsigned int &frame1, 
														unsigned  int &frame2, unsigned int &frame3){

	// Select initial frames according to given initial cameras
	if( this->initialCamerasIn.connected() && this->initialCamerasIn().size() == 2 && this->useInitialCameras() ){

			sout << "Take initial frames from initial cameras." << std::endl;
			frame1 = this->initialCamerasIn()[0].getIndex1();
			frame2 = this->initialCamerasIn()[1].getIndex1();
			frame3 = 3;
			/*
			sout 	<< "frame 1 = " << frame1 << " / "
					<< "frame 2 = " << frame2 << " / "
					<< "frame 3 = " << frame3 << std::endl;
			*/
	}
	// Select initial frames automatically from input slot
	else if( this->initialFramesIn.connected()
			&& this->initialFramesIn().size() == 3 ){

		frame1 = this->initialFramesIn()[0];
		frame2 = this->initialFramesIn()[1];
		frame3 = this->initialFramesIn()[2];

		// Swap index 1 and 2 if 1 > 2
		if(frame1 > frame2)
			std::swap(frame1, frame2);

		/*
		sout << "Take initial Frames given by input vector:\n\t" 
			<< "frame 1 = " << frame1 << " / "
			<< "frame 2 = " << frame2 << " / "
			<< "frame 3 = " << frame3 << std::endl;
		*/
	}

	// If not initial frame vector is connected, take plugin parameters
	else{

		if( this->initialFramesIn.connected()
				&& this->initialFramesIn().size() != 3 )
				sout << "InitialFrame vector size != 3" << std::endl;

		sout << "Taking initial frames from parameter values" << std::endl;

		frame1 = this->initialFrame1(); 
		frame2 = this->initialFrame2();
		frame3 = this->initialFrame3();

		// Swap index 1 and 2 if 1 > 2
		if(frame1 > frame2)
			std::swap(frame1, frame2);
		
		/*	 
		sout 	<< "frame 1 = " << frame1 << " / "
				<< "frame 2 = " << frame2 << " / "
				<< "frame 3 = " << frame3 << std::endl;
		*/
	}

	// Revert initial frames to default value, if value > nIt
	if(frame1 > nIterations || frame2 > nIterations || frame3 > nIterations)
	{
		sout << "Bad initial frames, take default frames 1, 2 and 3." << std::endl;
		frame1 = 0;
		frame2 = 1;
		frame3 = 2;
	}

	//Check if a fundamental matrix for initial frames is available
	try{
		sout << "Looking for fundamental matrices between initial frames " 
			<< frame1 << ", " << frame2 << std::endl;
		this->fundamentalMatricesIn()(frame1, frame2);
	}
	catch(...){
		sout << "Initial frame pair has no fundamental matrix!" << std::endl;
		bool foundNewInitialFrames = false;
		for(size_t i = 0; i < nIterations; i++)
		{
			for(size_t j = i+1; j < nIterations; j++)
			{
				try
				{
					this->fundamentalMatricesIn()(i,j);
					frame1 = i;
					frame2 = j;
					foundNewInitialFrames = true;
					sout << "Choose new initial frame pair ("
						<< i << "," << j << ")" << std::endl;
					break;
				}
				catch(...){}
			}

			if(foundNewInitialFrames)
				break;
		}

		if(!foundNewInitialFrames)
		{
			sout << "Could not find fundamental matrix "
				<< " to generate an initial point cloud." << std::endl;
			throw( std::runtime_error("Could not find fundamental matrix to generate an initial point cloud.") );
		}

	}
}
/*
#ifdef USE_PCL

void setPointSize(pcl::visualization::PCLVisualizer &viewer)
{
	viewer.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 4.0);
}

// Update PCL Point cloud and show it in PCL Viewer
template <typename T>
void Create3DMultiPointCloud<T>::updatePCLViewer(pcl::visualization::CloudViewer *viewer, CorrespondenceList<T> &pointCloud)
{

	viewer->runOnVisualizationThreadOnce(setPointSize);

	sout << "Updating Point Cloud Viewer" << std::endl;

	pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZRGBA>);
	
	auto it 	= pointCloud.begin();
	auto itEnd 	= pointCloud.end();

	for(; it != itEnd; it++)
	{
		if(!it->outlier)
		{
			pcl::PointXYZRGBA pt;
			pt.x = it->x();
			pt.y = it->y();
			pt.z = it->z();
			pt.r = it->R;
			pt.g = it->G;
			pt.b = it->B;
			cloud->push_back(pt);
		}
		else if( this->drawOutliers() )
		{
			pcl::PointXYZRGBA pt;
			pt.x = it->x();
			pt.y = it->y();
			pt.z = it->z();
			pt.r = 255;
			pt.g = 0;
			pt.b = 0;
			cloud->push_back(pt);
		}
	}
	sout << "Drawing " << cloud->size() << " points in 3d point cloud viewer" << std::endl;
	viewer->showCloud(cloud);
}

#endif
*/

template <typename T>
bool Create3DMultiPointCloud<T>::getBestInitialFrames(CorrespondenceList<T> &corrs, unsigned int &frame1, unsigned int &frame2, unsigned &frame3, bool keepFirstFrame, const IndexedMatrixList &fundamentalMatrices, size_t nIterations){
	
	size_t nImg =  nIterations; //corrs.sizeImages();
	
	int nC1 = this->nCorrs1();
	int nC2 = this->nCorrs2();
	int nC3 = this->nCorrs3();

	cimg_library::CImg<int>		nCorrs		(nImg,nImg);
	cimg_library::CImg<float>	nDisparity	(nImg,nImg);
	
	nCorrs.fill(0);
	nDisparity.fill(0.);

	// Iterate over images
	for(size_t i = 0; i < nImg; i++ ){

		// Iterate over points of current image
		auto itPoint1 		= corrs.get<imgIndex>().lower_bound(i);
		auto itPoint1End 	= corrs.get<imgIndex>().upper_bound(i);
		
		for(; itPoint1 != itPoint1End; itPoint1++ ){

			// Iterate over points with current point id
			const size_t pointId = itPoint1->pointIndex;

			auto itPoint2 		= corrs.get<pointIndex>().lower_bound(pointId);
			auto itPoint2End 	= corrs.get<pointIndex>().upper_bound(pointId);
			
			for(; itPoint2 != itPoint2End; itPoint2++){

				if(itPoint2->imgIndex >= nImg)
					continue;

				if(itPoint1->imgIndex != itPoint2->imgIndex){

					nCorrs(itPoint2->imgIndex, i) += 1;

					// calculate euclidean distance for disparity
					float dx = itPoint1->x() - itPoint2->x();
					float dy = itPoint1->y() - itPoint2->y();

					float d = std::sqrt(dx*dx+dy*dy);
					
					nDisparity(itPoint2->imgIndex, i) += d;
				}
			}

		}

	}

	// Normalize disparity by number of corrs
	for(size_t i = 0; i < nImg; i++){
		for(size_t j = 0; j < nImg; j++){			
			if(nCorrs(i,j) != 0)
				nDisparity(i,j) /= nCorrs(i,j);
		}
	}

	bool initialFrameFound = false;

	// Iterate over every image until initial frames are found
	size_t endFrame1;

	if(keepFirstFrame){
		endFrame1 = frame1 + 1;
	}else{
		frame1 		= 0;
		endFrame1 	= nImg;		
	}

	for(; frame1 < endFrame1; frame1++){

		cimg_library::CImg<float> 	dispSorted;
		cimg_library::CImg<int> 	permutation;

		dispSorted = nDisparity.get_crop(0,0,nImg-1,0).sort(permutation,0);

		// Iterate over remaining images and look for second image with maximum(disp)
		for(size_t j = 0; j < nImg; j++){

			frame2 = permutation(j,0);

			if(frame1 > frame2){
				if(!fundamentalMatrices.exists(frame2, frame1)){ 
					continue;
				}
			}else{
				if(!fundamentalMatrices.exists(frame1, frame2)){ 
					continue;
				}
			}	

			// Do not compare frame with itself
			if(frame2 == frame1)
				continue;

			// Continue if current image pair has too less correspondences
			if( nCorrs(frame2, frame1) < nC1 )
				continue;

			// Continue if disparity between selected first and second frame is 0
			if(nDisparity(frame2, frame1) == 0)
				continue;

			// Image with energy values regarding number of correspondences between
			// image 3 and 2 | image 3 and 1
			cimg_library::CImg<float> energyThirdFrame(nImg,1);
			energyThirdFrame.fill(0);
		
			// Iterate over remaining images and look for a third image where
			// the disparity between the third image and image 1 AND 2 is as big as possible
			for(frame3 = 0; frame3 < nImg; frame3++){
			
				// Do not compare frames with eachother
				if(frame3 == frame1 || frame3 == frame2)
					continue;
			
				// Continue if third selected frame does not have enough correspondences
				// with frame 1 and 2
				if( nCorrs(frame3,frame1) < nC2 || nCorrs(frame3,frame2) < nC3 )
					continue;

				energyThirdFrame(frame3,0) = nDisparity(frame3,frame1) * nDisparity(frame3,frame2);

			}

			// Sort energyThirdFrame and evaluate entries with high value first
			cimg_library::CImg<int> permutation2;
			energyThirdFrame.sort(permutation2,0);
			frame3 = permutation2(0,0);

			// Select initial frames if disparities > 0 and nCorrs > min
			if(energyThirdFrame(0,0) >= 0.001){
				//sout << "Choose image " << frame3 << " as third initial frame" << std::endl;
				
				if(frame1 > frame2)
					std::swap(frame1, frame2);

				return true;

			}
			else
				sout << "Disparities between 1-3 and 1-2 are too small...";

		}

	}


	//Did not found any suitable frames
	return false;

}


template <typename T>
Create3DMultiPointCloud<T>::~Create3DMultiPointCloud(){


#ifdef USE_PCL
	if(viewer)
		delete viewer;
	
#endif

}



#endif