/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Create3DPointCloud.hxx
 *  Implementation of the parameter class Create3DPointCloud.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 09.02.2012
 */

#ifndef _CREATE3DPOINTCLOUD_HXX_
#define _CREATE3DPOINTCLOUD_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/Create3DPointCloud.h>
#include <charon-utils/ImgTool.hxx>

//#define AVOID_SVD
//#define DONT_USE_PLUGINS

#ifdef AVOID_SVD
	#include "hekate/totalLstsqMinimize.h"
#endif

#include <limits>

template <typename T>
Create3DPointCloud<T>::Create3DPointCloud(const std::string& name) :
		TemplatedParameteredObject<T>(
			"Create3DPointCloud", name,
			"<h2>Creates a 3D point cloud by correspondencs and an initial "
			"fundamental matrix</h2><br>"
			"Creates a 3D point cloud by correspondencs and an initial "
			"fundamental matrix.<br>"
			"Step 1: The fundamental matrix between the first frame pair "
			"(frame1 and frame2) is used to determine two camera matrices "
			"(cam1 and cam2). <br>"
			"Step 2: With cam1 and cam2 the correspondences are triangulated "
			"and a 3D point could is achieved.<br>"
			"Step 3: The 3D point cloud is used together with 2D points of "
			"the next frame (frame3) to retrieve a third camera matrix "
			"(cam3). <br>"
			"Step 4: Repeat step 2 and step 3 until all frames were used for "
			"point generation."
		), 
		isInitialized(0), 
		initialFrames(1,0), 
		initialCameras(1,0)
#ifdef USE_PCL
		, viewer(0)
#endif
{
	ParameteredObject::_addParameter< int >(
		maxFrames, "maxFrames",
		"Numer of maximal frames which will be considered. (0 = all frames)",
		0, "int");
	ParameteredObject::_addParameter< unsigned int > (initialFrame1, "initialFrame1", 
		"First initial frame from which the first point cloud is calculated.", 0, "int");
	ParameteredObject::_addParameter< unsigned int > (initialFrame2, "initialFrame2", 
		"Second initial frame from which the first point cloud is calculated.", 1, "int");
	ParameteredObject::_addParameter< unsigned int > (initialFrame3, "initialFrame3", 
		"Third initial frame from which the first point cloud is calculated.", 2, "int");

	ParameteredObject::_addParameter< bool >(
		verbose, "verbose",
		"Show extended log",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		useInitialCameras, "useInitialCameras",
		"useInitialCameras",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		useCUDA, "useCUDA",
		"useCUDA",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		fixedFocalLength, "fixedFocalLength",
		"Assume fixed focal length.",
		false, "bool");
	ParameteredObject::_addParameter< int >(
		minPtViews, "minPtViews",
		"Minimum number of views, a point is visible in",
		3, "int");
	ParameteredObject::_addParameter< int >(
		min3d2dCorrs, "min3d2dCorrs",
		"Minimum number of 3d-2d point correspondences used to retrieve views.",
		50, "int");
	
	ParameteredObject::_addParameter< bool >(
		sortOutByBackProjection, "sortOutByBackProjection",
		"Sort out 3d points by back projection errors, while calculating new 3d points.",
		true, "bool");
	ParameteredObject::_addParameter< float >(
		maxBackProjectionError, "maxBackProjectionError",
		"Maximal back projection error of a 3d inlier.",
		2.0, "float");
#ifdef USE_PCL
	ParameteredObject::_addParameter< bool >(
		showPointCloudViewer, "showPointCloudViewer",
		"Shows 3d point cloud in pcl cloud viewer.",
		true, "bool");
	ParameteredObject::_addParameter< bool >(
		highFrequencyViewerUpdates, "highFrequencyViewerUpdates",
		"highFrequencyViewerUpdates",
		false, "bool");
#endif
	ParameteredObject::_addParameter< bool >(
		drawOutliers, "drawOutliers",
		"Draws outliers in pcl cloud viewer.",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		pbaAfterEveryIteration, "pbaAfterEveryIteration",
		"Perform bundle adjustment for every new camera",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		recoverCamByRANSAC, "recoverCamByRANSAC",
		"Recover cams by OpenCV method solvePnPRANSAC",
		false, "bool");	
	ParameteredObject::_addParameter< bool >(
		sortOutOutliersByPnPRANSAC, "sortOutOutliersByPnPRANSAC",
		"Sort out 3D outliers detected by solvePnPRANSAC",
		true, "bool");	

	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		fundamentalMatrices, "fundamentalMatrices",
		"fundamentalMatrices",
		"IndexedMatrixList");
	ParameteredObject::_addInputSlot(
		calibrationMatrix, "calibrationMatrix",
		"optional initial cameras",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		initialCameras, "initialCameras",
		"initialCameras",
		"IndexedMatrixList");
	ParameteredObject::_addInputSlot(
		initialFrames, "initialFrames",
		"Initial frames (optional)",
		"vector<int>");

	ParameteredObject::_addOutputSlot(
		cameraMatrices, "cameraMatrices",
		"cameraMatrices",
		"IndexedMatrixList");
	ParameteredObject::_addOutputSlot(
		pointCloud, "pointCloud",
		"pointCloud",
		"CorrespondenceList");
	ParameteredObject::_addOutputSlot(
		calibrationMatrixOut, "calibrationMatrixOut",
		"calibrationMatrixOut",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		correspondencesOut, "correspondencesOut",
		"correspondencesOut",
		"CorrespondenceList");

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void Create3DPointCloud<T>::execute() {
	
	#ifdef USE_PCL
	if(!viewer)
		viewer = new pcl::visualization::CloudViewer("3D Point Cloud Viewer");
	#endif

	if(!isInitialized){
		initSlots();
		isInitialized = true;
	}
	
	// list of correspondences
	list = this->correspondences();

	cameras.clear();
	backprojections.clear();
	
	// Calculate number of frames considered in reconstruction
	nImg = list.sizeImages();
	nFM  = this->fundamentalMatrices().size();
	nCal = this->calibrationMatrix().size();

	sout << "Found:\n\t" << nImg << "\tdifferent img ids in correspondence list\n\t"
		<< nFM << "\tfundamental matrices\n\t" << nCal << "\tcalibration matrices" << std::endl;

	// If number of known calibrations is not equal number of images
	// take first calibration as default for every image
	if(this->calibrationMatrix().size() < nImg){
		for(size_t i=0; i<nImg; i++){
			calibrations.push_back(this->calibrationMatrix()[0]);
		}
	}
	// Read in calibrations
	else{
		calibrations = this->calibrationMatrix();
	}

	// Determine number of frames considered for calculation
	size_t nIterations ;
	if(maxFrames() > 2 && maxFrames() < nImg)
		nIterations = maxFrames();
	else
		nIterations = nImg ;
			
	// Select initial frames
	sout << "Select initial Frames" << std::endl;
	selectInitialFrames(nIterations);

	// Get initial camera pair
	sout << "Calculate initial camera pair" << std::endl;
	initCam( initFrame1, initFrame2 );
	
	// Triangulate with initial camera pair
	triangulatePoints( initFrame1 );

	// Calculate back projection errors
	double meanBPError = calculateBackProjection(cameras);
	
	// Perform back projection and sort out bad 3d points
	pluginBackProjection3D2D<T>(cameras, this->pointCloud(), backprojections, list, meanBPError, this->min3d2dCorrs());	
	
	IndexedMatrixList pbacameraMatricesOut2;
	CorrespondenceList<T> pbapointCloudOut2;
	CImgList<T> pbacalibrationMatrixOut2;
	ParallelBundleAdjustmentPlugin( cameras, calibrations, this->pointCloud(), list, pbacameraMatricesOut2, 
										pbapointCloudOut2, pbacalibrationMatrixOut2, 2, this->fixedFocalLength(), this->useCUDA());
	cameras = pbacameraMatricesOut2;
	calibrations = pbacalibrationMatrixOut2;
	this->pointCloud() = pbapointCloudOut2;

	// Calculate back projection errors
	meanBPError = calculateBackProjection(cameras);

	// Refine all 3D points by all available views
	refinePoints();

	sout << "this->maxBackProjectionError() = " << this->maxBackProjectionError() << std::endl;

	// Update viewer
	#ifdef USE_PCL
	if( this->showPointCloudViewer() && this->highFrequencyViewerUpdates() ) 
		updatePCLViewer(viewer);
	#endif
	
	// Iterate over (nIterations - 2) frames
	bool firstIteration = true;
	bool sortOutOutliersByInitialFrame3 = true;
	bool sortOutOutliersBySolvePnPRansac = recoverCamByRANSAC();
	for(size_t i = 0; i < nIterations; i++)
	{

		if(firstIteration)
			i = initFrame3;
		else if( i == initFrame1 || i == initFrame2 )
			continue;
		
		sout << "Trying to estimate " << nIterations - i << " more camera views." << std::endl;

		// Check, if camera already exists
		try
		{
			cameras(i);
			sout << "Camera with index " << i << " already exists." << std::endl;
			continue;		// Camera exists, continue with next iteration
		}

		// If camera does not exist, calculate it
		catch(...){
		
			sout << std::endl << "------------------------------------" << std::endl;
			sout << "Calculate camera P" << i << std::endl;

			try
			{
				IndexedMatrix P, P2;
				if( recoverCamByRANSAC() && firstIteration && sortOutOutliersByInitialFrame3 )
					P = (CImg<double>) recoverCamRANSAC(i, calibrations, list, this->pointCloud(), 300, this->maxBackProjectionError(), 
						20, 0, IndexedMatrixList(),	true);
				else if( recoverCamByRANSAC() )
					P = (CImg<double>) recoverCamRANSAC(i, calibrations, list, this->pointCloud(), 300, this->maxBackProjectionError(),
						20, 0, IndexedMatrixList(), sortOutOutliersBySolvePnPRansac);
				else 
					P = (CImg<double>) recoverCam(i, calibrations, list, this->pointCloud());
				P.setIndex(i);
				cameras.push_back(P);
			
				if(this->pbaAfterEveryIteration()){
						IndexedMatrixList pbacameraMatricesOut;
						CorrespondenceList<T> pbapointCloudOut;
						CImgList<T> pbacalibrationMatrixOut;

						ParallelBundleAdjustmentPlugin( cameras, calibrations, this->pointCloud(), list, pbacameraMatricesOut, 
															pbapointCloudOut, pbacalibrationMatrixOut, 2, this->fixedFocalLength(), this->useCUDA());
						cameras = pbacameraMatricesOut;
						calibrations = pbacalibrationMatrixOut;
						this->pointCloud() = pbapointCloudOut;
				}

				// Triangulate missing points by P_i and all prior cameras
				if( !recoverCamByRANSAC() || !firstIteration || !sortOutOutliersByInitialFrame3 )
				triangulatePoints( i );
				
				// Refine all 3D points by all available views
				refinePoints();

				// Sort out 3d points whose back projection
				// has an error taller than this->maxBackProjectionError()
				if(this->sortOutByBackProjection())
				{
					backprojections.clear();
					pluginBackProjection3D2D<T>(cameras, this->pointCloud(), backprojections, list, maxBackProjectionError(), this->min3d2dCorrs());
					if(cameras(i).degenerated)
						throw(std::runtime_error("Camera degenerated."));
				}
			
				// Update viewer
				#ifdef USE_PCL
				if( this->showPointCloudViewer() && this->highFrequencyViewerUpdates() ) 
					updatePCLViewer(viewer);
				#endif
		
				// Calculate back projection errors
				if(this->verbose())
					calculateBackProjection(cameras);

			}
			catch(...)
			{
				sout << "Cam " << i << " could not be reconstructed "
					<< "due to too less matches between already known 3D "
					<< "points and features in image " << i << "." << std::endl;
			}
			
		}
		
		if ( firstIteration ){
			i = -1;
			firstIteration = false;
		}

	}

	sout << std::endl << "------------------------------------" << std::endl;
	sout << "Reconstructed cameras:\t\t\t" << cameras.size() << std::endl;

	// Perform sparse bundle adjustment
	IndexedMatrixList pbacameraMatricesOut;
	CorrespondenceList<T> pbapointCloudOut;
	CImgList<T> pbacalibrationMatrixOut;

	ParallelBundleAdjustmentPlugin( cameras, calibrations, this->pointCloud(), list, pbacameraMatricesOut, 
										pbapointCloudOut, pbacalibrationMatrixOut, this->minPtViews(), this->fixedFocalLength(), this->useCUDA());

	cameras = pbacameraMatricesOut;
	calibrations = pbacalibrationMatrixOut;
	this->pointCloud() = pbapointCloudOut;

	if(sortOutOutliersBySolvePnPRansac){
	
		this->pointCloud().removeOutliers();
		triangulatePoints();
		backprojections.clear();
		pluginBackProjection3D2D<T>(cameras, this->pointCloud(), backprojections, list, maxBackProjectionError(), this->min3d2dCorrs());
		ParallelBundleAdjustmentPlugin( cameras, calibrations, this->pointCloud(), list, pbacameraMatricesOut, 
										pbapointCloudOut, pbacalibrationMatrixOut, this->minPtViews(), this->fixedFocalLength(), this->useCUDA());
		cameras = pbacameraMatricesOut;
		calibrations = pbacalibrationMatrixOut;
		this->pointCloud() = pbapointCloudOut;		
	
	}

	// Update viewer
	#ifdef USE_PCL
	if( this->showPointCloudViewer() ) 
		updatePCLViewer(viewer);
	#endif

	// Save calculated cameras in output slot
	this->cameraMatrices() = cameras;
	
	// Calculate back projection error
	if(this->verbose())
		calculateBackProjection(cameras);

	this->correspondencesOut() = list.removeOutliers();

	this->calibrationMatrixOut() = calibrations;
		
	
}


// Select initial frames
template <typename T>
void Create3DPointCloud<T>::selectInitialFrames(size_t nIterations){

	// Select initial frames according to given initial cameras
	if( this->initialCameras.connected() && this->initialCameras().size() == 2 && this->useInitialCameras() ){

			sout << "Take initial frames from initial cameras." << std::endl;
			initFrame1 = this->initialCameras()[0].getIndex1();
			initFrame2 = this->initialCameras()[1].getIndex1();
			initFrame3 = 3;
			 
			sout << "frame 1 = " << initFrame1 << " / "
			<< "frame 2 = " << initFrame2 << " / "
			<< "frame 3 = " << initFrame3 << std::endl;

	}
	// Select initial frames automatically from input slot
	else if( this->initialFrames.connected()
			&& this->initialFrames().size() == 3 ){

		initFrame1 = this->initialFrames()[0];
		initFrame2 = this->initialFrames()[1];
		initFrame3 = this->initialFrames()[2];

		// Swap index 1 and 2 if 1 > 2
		if(initFrame1 > initFrame2){
			int if1= initFrame1;
			initFrame1 = initFrame2;
			initFrame2 = if1;
		}

		sout << "Take initial Frames given by input vector:\n\t" 
			<< "frame 1 = " << initFrame1 << " / "
			<< "frame 2 = " << initFrame2 << " / "
			<< "frame 3 = " << initFrame3 << std::endl;

	}

	// If not initial frame vector is connected, take plugin parameters
	else{

		if( this->initialFrames.connected()
				&& this->initialFrames().size() != 3 )
				sout << "InitialFrame vector size != 3" << std::endl;

		sout << "Taking initial frames from parameter values" << std::endl;

		initFrame1 = this->initialFrame1(); 
		initFrame2 = this->initialFrame2();
		initFrame3 = this->initialFrame3();

		// Swap index 1 and 2 if 1 > 2
		if(initFrame1 > initFrame2){
			int if1= initFrame1;
			initFrame1 = initFrame2;
			initFrame2 = if1;
		}
			 
		sout << "frame 1 = " << initFrame1 << " / "
		<< "frame 2 = " << initFrame2 << " / "
		<< "frame 3 = " << initFrame3 << std::endl;

	}

	// Revert initial frames to default value, if value > nIt
	if(initFrame1 > nIterations || initFrame2 > nIterations || initFrame3 > nIterations)
	{
		sout << "Bad initial frames, take default frames 1, 2 and 3." << std::endl;
		initFrame1 = 0;
		initFrame2 = 1;
		initFrame3 = 2;
	}
	//Check if a fundamental matrix for initial frames is available
	try{
		sout << "Looking for fundamental matrices between initial frames " 
			<< initFrame1 << ", " << initFrame2 << std::endl;
		this->fundamentalMatrices()(initFrame1,initFrame2);
	}
	catch(...){
		sout << "Initial frame pair has no fundamental matrix!" << std::endl;
		bool foundNewInitialFrames = false;
		for(size_t i = 0; i < nIterations; i++)
		{
			for(size_t j = i+1; j < nIterations; j++)
			{
				try
				{
					this->fundamentalMatrices()(i,j);
					initFrame1 = i;
					initFrame2 = j;
					foundNewInitialFrames = true;
					sout << "Choose new initial frame pair ("
						<< i << "," << j << ")" << std::endl;
					break;
				}
				catch(...){}
			}

			if(foundNewInitialFrames)
				break;
		}

		if(!foundNewInitialFrames)
		{
			sout << "Could not find fundamental matrix "
				<< " to generate an initial point cloud." << std::endl;
			throw( std::runtime_error("Could not find fundamental matrix to generate an initial point cloud.") );
		}

	}
}

#ifdef USE_PCL

void setPointSize(pcl::visualization::PCLVisualizer &viewer)
{
	viewer.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 4.0);
}

// Update PCL Point cloud and show it in PCL Viewer
template <typename T>
void Create3DPointCloud<T>::updatePCLViewer(pcl::visualization::CloudViewer *viewer)
{

	viewer->runOnVisualizationThreadOnce(setPointSize);

	sout << "Updating Point Cloud Viewer" << std::endl;
	pcl::PointCloud<pcl::PointXYZRGBA>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZRGBA>);
	auto it = this->pointCloud().begin();
	auto itEnd = this->pointCloud().end();
	for(; it != itEnd; it++)
	{
		if(!it->outlier)
		{
			pcl::PointXYZRGBA pt;
			pt.x = it->x() ;
			pt.y = it->y() ;
			pt.z = it->z() ;
			pt.r = it->R;
			pt.g = it->G;
			pt.b = it->B;
			cloud->push_back(pt);
		}
		else if( this->drawOutliers() )
		{
			pcl::PointXYZRGBA pt;
			pt.x = it->x() ;
			pt.y = it->y() ;
			pt.z = it->z() ;
			pt.r = 255;
			pt.g = 0;
			pt.b = 0;
			cloud->push_back(pt);
		}
	}
	sout << "Drawing " << cloud->size() << " points in 3d point cloud viewer" << std::endl;
	viewer->showCloud(cloud);
}

#endif

// Initiate back projection
template <typename T>
void Create3DPointCloud<T>::initBackProjection(BackProjection3D2D<T> &bp, const double meanBPError)
{
	
	bp.resetExecuted();

	bpCam = cameras;
	op_2d = list;
	op_3d = this->pointCloud();

	bp.cameraMatrices.connect(bpCam);
	bp.correspondences.connect(op_2d);
	bp.points3D.connect(op_3d);

	double maxError = std::max(meanBPError,(double)this->maxBackProjectionError());
	sout << "Using maxError = " << maxError << " as maximal back projection error." << std::endl;
	bp.maxBackProjectionError = maxError;
	bp.min3d2dCorrs = this->min3d2dCorrs();

}


// Initiate back projection
template <typename T>
void  Create3DPointCloud<T>::updateBackProjection(BackProjection3D2D<T> &bp)
{
	
	bp.resetExecuted();

	bpCam = cameras;
	op_2d = list;
	op_3d = this->pointCloud();

	bp.cameraMatrices.connect(bpCam);
	bp.correspondences.connect(op_2d);
	bp.points3D.connect(op_3d);

}


// Calculate initial camera pair
template <typename T>
void Create3DPointCloud<T>::initCam(const unsigned int initFrame1, const unsigned int initFrame2)
{

	if( this->initialCameras.connected() ){
		if( this->useInitialCameras() ){
		
		cameras.push_back( this->initialCameras()[0] );
		cameras.push_back( this->initialCameras()[1] );
	
		}
	}
	else{

		// read in calibration matrices
		cimg_library::CImg<T> cal1;
		cal1 = calibrations[initFrame1];
	
		ImgTool::printInfo(sout,cal1,"	calibration: ");
			
		// Generate camera at origin as first camera.
		CImg<double> P1(4,3), P2;
		P1.fill(0.); P1(0,0) = 1; P1(1,1) = 1; P1(2,2) = 1;
		P1 = cal1 * P1;
	
		// Save first camera P1
		IndexedMatrix P1i, P2i;
		P1i = P1;
		P1i.setIndex( initFrame1 );
		cameras.push_back(P1i);
	
		// Get and save second camera P2
		P2 = recoverCamByFMatrix( initFrame1, initFrame2 );
		P2i = P2;
		P2i.setIndex( initFrame2 );
		cameras.push_back(P2i);

	}

}


/// Refine 3d points by (PM - m = 0) for all views
template <typename T>
void Create3DPointCloud<T>::refinePoints()
{
	sout << "-------------------------------" << std::endl;
	sout << "Refine 3D points by all existing views..." << std::endl;

	int nCount = 0, n3D = 0, nOutlier = 0, nLess = 0, nA = 0;
	
	auto it = this->pointCloud().begin();
	auto itEnd = this->pointCloud().end();
	for(; it!=itEnd; it++)
	{

		n3D++;

		if(it->outlier){
			nOutlier++;
			continue;
		}

		std::vector< const Correspondence<T>*> ptList = 
			list.getListByPointId(it->pointIndex);

		// Get number of corresponding inliers
		size_t nPt = 0;
		for(size_t i=0; i<ptList.size(); i++)
		{	
			if(!ptList[i]->outlier)
				nPt++;
		}


		// Continue if less than two 2d inlier points correspond with 3d point
		if(nPt < 2)
		{
			nLess++;
			continue;
		}

		// Create refinement matrix (Pollefeys, Visual 3D modelling from Images, p.46)
		CImg<double> X = (CImg<double>) it->pt3d();
		//ImgTool::printInfo(sout, X , "X");


		CImg<double> A;
		for(size_t i=0; i<ptList.size(); i++)
		{	
			
			if(!ptList[i]->outlier)
			{

				try
				{
					// Get camera matrix
					CImg<double> P = cameras(ptList[i]->imgIndex);
					CImg<double> P1 = P.get_crop(0,0,3,0);
					CImg<double> P2 = P.get_crop(0,1,3,1);
					CImg<double> P3 = P.get_crop(0,2,3,2);


					// get 2d coordinates
					double x = ptList[i]->x(), y = ptList[i]->y();

					//sout << "x=" << x << ", y=" << y << std::endl;

					double c = P3.dot(X);
					A.append( (x * P3 - P1) / c, 'y' );
					A.append( (y * P3 - P2) / c, 'y' );

				}
				catch(...){}
			}
		}

		if(A.height() < 4)
		{
			//sout << "Refinement matrix too small" << std::endl;
			//ImgTool::printInfo(sout, A , "A");
			nA++;
			continue;
		}

		#ifdef AVOID_SVD

		T result[4];
		totalLstsqMinimizeImg(A, (T*)&result, false);
		it->x() = result[0] / result[3];
		it->y() = result[1] / result[3];
		it->z() = result[2] / result[3];
		it->w() = 1;

		#else

		CImg<double> S, V, U;
		A.SVD(U, S, V);
		
		CImg<double> Xnew = V.get_crop(3,0,3,3);

		bool isInf = false;
		if(Xnew[3] == 0)
			isInf = true;

		Xnew /= Xnew[3];
		
		it->x() = Xnew[0];
		it->y() = Xnew[1];
		it->z() = Xnew[2];
		it->w() = Xnew[3];
		//it->outlier = true;

		#endif

		nCount++;

		//ImgTool::printInfo(sout, Xnew , "Xnew");

	}

	sout << nCount << "/" << n3D << " 3d points have been refined." << std::endl;
	if(nOutlier > 0 ) sout << nOutlier << " 3D points were outliers." << std::endl;
	if(nLess > 0 ) sout << nLess << " 2D correspondences had too less points." << std::endl;
	if(nA > 0 ) sout << nA << " 3D points had too less views to refine them." << std::endl;
	if(nOutlier + nLess + nA > 0 ) sout << nOutlier + nLess + nA << " 3D points could not be used for refinement." << std::endl;
	sout << "-------------------------------" << std::endl;

}


/// Calculate back projection error in every image
template <typename T>
double Create3DPointCloud<T>::calculateBackProjection(const IndexedMatrixList &cameras){

	sout << "Calculate quality of reconstruction by back projection" << std::endl;
		
	// list of 3d points
	const CorrespondenceList<T> &list3d = pointCloud();

	double meanMean = 0, meanStdDev = 0;
		
	// Iterate over all cameras available
	for(size_t i = 0; i < cameras.size(); i++)
	{

		const IndexedMatrix &P = cameras[i] ;
		if( P.degenerated )
			continue;

		std::vector<double> distVec;
		double distAbs = 0;

		auto it = list3d.begin() ;
		auto itEnd = list3d.end();
		for(; it != itEnd ; it++ )
		{
			
			if(it->outlier)
				continue;

			try
			{

				// Calculate distance between 2d feature point
				// and back projected 3d point

				// Get inhomogeneous 2D point
				const Correspondence<T> &corr2D = list(P.getIndex1(), it->pointIndex) ;

				if(corr2D.outlier)
					continue;

				CImg<double> pt2D = (CImg<double>) corr2D.pt2d() ;
				pt2D.crop(0,0,0,1);

				//ImgTool::printInfo(sout, pt2D, "pt2D");

				// Get inhomogeneous backprojected 2D point
				const Correspondence<T> &corr3D = *it ;
				CImg<double> pt3D = (CImg<double>) corr3D.pt3d() ;

				CImg<double> pt2D_back = (CImg<double>) (P * pt3D) ;

				if(pt2D_back[2] == 0) continue;

				pt2D_back /= pt2D_back[2];
				pt2D_back.crop(0,0,0,1);
				
				//ImgTool::printInfo(sout, pt2D_back, "pt2D_back");
				
				// Calculate distance pt2D - pt2D_back
				CImg<double> dist = pt2D - pt2D_back ;
				//ImgTool::printInfo(sout, dist, "dist");

				double distVal = std::sqrt(dist.dot(dist)) ;

				distVec.push_back(distVal) ;
				distAbs += distVal ;

			}

			catch(...)
			{
				continue;
			}

		}

		if( verbose() ){
			for(size_t i=0; i<distVec.size(); i++)
				sout << "distVal = " << distVec[i] << std::endl;
		}


		// Continue with next iteration, if no corresponding
		// 2D - 3D points could be found
		if(distVec.size() == 0)
		{
			sout << "No correspondences between 2d and 3d points "
				<< "in frame " << i << ". Continue with next frame." << std::endl;
			continue;
		}

		double distMean = distAbs / distVec.size() ;

		double stdev = 0;
		for(size_t j = 0; j < distVec.size(); j++)
		{
			stdev += (distVec[j]-distMean)*(distVec[j]-distMean) ;
		}
		stdev = std::sqrt( stdev/distVec.size() );

		sout << "Image[" << P.getIndex1() << "]: mean = " << distMean
			<< "; stdev = " << stdev << ";" << std::endl;

		meanMean += distMean;
		meanStdDev += stdev;

	}

	meanMean /= cameras.size();
	meanStdDev /= cameras.size();
	/*
	if( isinf(meanMean+meanStdDev) )
	{
		sout << "Error while calculating stdev (=inf) " <<  std::endl;
		return 1;
	}
	*/
	if( meanMean+meanStdDev > 1 && meanMean+meanStdDev < 200 )
		return meanMean+meanStdDev;
	else
	{
		sout << "No proper stdev:\nmeanMean=" << meanMean << "\nmeanStdev="<< meanStdDev << std::endl;
		return 1;
	}

}


/// Triangulate all correspondences between frame i and i+1 by P1 and P2
template <typename T>
void Create3DPointCloud<T>::triangulatePoints()
{

	for(size_t i = 0; i < cameras.size(); i++)
	{
		triangulatePoints( cameras[i].getIndex1() );
	}

}



/// Triangulate all correspondences between frame i and i+1 by P1 and P2
template <typename T>
bool Create3DPointCloud<T>::triangulatePoints(const size_t i)
{
	sout << "Triangulate points..." << std::endl;

	const CorrespondenceList<T> &corrs = list;
	int n1=0, n2=0, n3=0, n4=0, n5=0, n6=0;
	
	// 1) Iterate over every 2d point per camera
	// 2) Check 2d point is an outlier -> continue with next iteration
	// 3) If the corresponding 3D point already exists -> continue with next iteration
	// 4) Check if there are 2d correspondences with a known camera
	// 5) If there is at least one correspondence, try to triangulate


	// If bad camera, continue with next one
	if(cameras(i).degenerated)
	{
		sout << "Degererated camera -> won't triangulate points." << std::endl;
		return 0;
	}

	auto it = corrs.get<imgIndex>().lower_bound(i) ;
	auto itEnd = corrs.get<imgIndex>().upper_bound(i);
	for(; it != itEnd ; it++ )
	{
		
		n1++;
			
		// Continue with next iteration, if point is an outlier
		if(it->outlier){
			n6++;
			continue;
		}

		// If point index already exists in 3d point cloud, continue
		// with next iteration
		if(this->pointCloud().existPtIndex(it->pointIndex))
		{
			//sout << "Point already exists" << std::endl;
			n3++;
			continue;
		}

		std::vector< const Correspondence<T>*> corrPointList = corrs.getListByPointId(it->pointIndex);
		
		CImgList<double> cameraList, pointList;
		//pointList.push_back( it->pt2d() );
		//cameraList.push_back( (CImg<double>) cameras(i) );


		for(size_t j = 0; j < corrPointList.size(); j++)
		{

			if( corrPointList[j]->outlier ) continue;

			CImg<double> camTemp;

			// If bad camera, continue
			try{ camTemp = cameras(corrPointList[j]->imgIndex); }
			catch(...){ continue; }

			if( cameras(corrPointList[j]->imgIndex).degenerated )
				continue;

			// Save correspondence in pointList
			pointList.push_back( corrPointList[j]->pt2d() );

			// Save camera in cameraList
			cameraList.push_back( camTemp );

		}

		// If no correspondence could be found
		// continue with next point
		if(pointList.size() < 2){
			n4++;
			continue;
		}
		
		CImg<T> X;
		try{
			X = getTriangulation( cameraList, pointList );
		}
		catch(...){
			sout << "Point could not be triangulated" << std::endl;
		}

		// check whether the triangulated point lies before or behind the priciple planes
		// the principle plane is the plane the camera center lies on and which is
		// intersected of by the priciple ray perpendicularly.
		bool positiveDepth = true;
		for(size_t j = 0; j < cameraList.size(); j++)
		{
			double depth = getDepth( cameraList[j], X );
			
			//push back 3D corrdinates
			if(depth < 0 ){
				positiveDepth = false;
				n5++;
				break;
			}
		}
			
		// If triangulated point lies in front of all camera pricipal planes
		// save it in 3D point cloud
		if(positiveDepth)
		{
			Correspondence<T> point3D( X, it->pointIndex, cameras(i).getIndex1(), false, it->R, it->G, it->B );
			this->pointCloud().push_back(point3D);
			n2 ++;
		}

	}

	sout << n2 << " / " << n1 << " points of frame " << i << " could be triangulated" << std::endl;
	sout << n6 << " points in frame " << i << " were outliers." << std::endl;
	sout << n4 << " 2d points do not have inlier correspondences with known cameras." << std::endl;
	sout << n5 << " 3d points do not ly in front of all cameras which project them." << std::endl;
	sout << n3 << " 2d points have already a corresponding 3d point." << std::endl;

	return 1;

}



/// Triangulate 2D points pt1 and pt2 with cameras P1 and P2 to retrieve the 3D point. (Hartley p.592)
template <typename T>
CImg<T> Create3DPointCloud<T>::getTriangulation(
	const CImgList<double> &cameraList, const CImgList<double> &pointList){
				
	if( pointList.size() < 2 || cameraList.size() < 2 )
		throw( std::range_error("Create3DPointCloud<T>::getTriangulation: No correspondence or camera")) ;

	// define matrix A = x cross X and x' cross X
	CImg<T> A( 4, cameraList.size() * 2 ) ;
	
	for(size_t i = 0; i < pointList.size(); i++)
	{
		//first row of A
		A(0,i*2) = pointList[i](0,0)*cameraList[i](0,2) - cameraList[i](0,0);
		A(1,i*2) = pointList[i](0,0)*cameraList[i](1,2) - cameraList[i](1,0);
		A(2,i*2) = pointList[i](0,0)*cameraList[i](2,2) - cameraList[i](2,0);
		A(3,i*2) = pointList[i](0,0)*cameraList[i](3,2) - cameraList[i](3,0);
		//second row of A
		A(0,i*2+1) = pointList[i](0,1)*cameraList[i](0,2) - cameraList[i](0,1);
		A(1,i*2+1) = pointList[i](0,1)*cameraList[i](1,2) - cameraList[i](1,1);
		A(2,i*2+1) = pointList[i](0,1)*cameraList[i](2,2) - cameraList[i](2,1);
		A(3,i*2+1) = pointList[i](0,1)*cameraList[i](3,2) - cameraList[i](3,1);
	}

	// determine 3D point via SVD of A (Hartley p.592)

	#ifdef AVOID_SVD

	T result[4];
	totalLstsqMinimizeImg(A, (T*)&result, false);

	cimg_library::CImg<T> X(1,4);

	X(0, 0) = result[0] / result[3]; 				
	X(0, 1) = result[1] / result[3]; 
	X(0, 2) = result[2] / result[3]; 
	X(0, 3) = 1; 

	#else

	cimg_library::CImg<T> X(1,4), U, S, V;
	A.SVD(U, S, V);

	X = V.get_crop(3,0,3,3);
	X /= X(0,3);
	#endif

	return X;

}


/// Triangulate 2D points pt1 and pt2 with cameras P1 and P2 to retrieve the 3D point. (Hartley p.592)
template <typename T>
CImg<T> Create3DPointCloud<T>::getTriangulation(
		const CImg<T> &P1, const CImg<T> &P2, 
		const CImg<T> &pt1, const CImg<T> &pt2){
		
	CImg<T> X;
	CImgList<double> cameraList, pointList;

	cameraList.push_back(P1);
	cameraList.push_back(P2);
	
	pointList.push_back(pt1);
	pointList.push_back(pt2);

	X = getTriangulation(cameraList, pointList);

	return X;

}

/// Calculate depth between x and principle plane of P.
template <typename T>
double Create3DPointCloud<T>::getDepth(const CImg<T> &P, const CImg<T> &X) {

	int signP;
	(P.get_crop(0,0,2,2).det() < 0 ? signP = -1 : signP = 1);

	double P_m3_abs = P(0,2)*P(0,2)+P(1,2)*P(1,2)+P(2,2)*P(2,2);
	P_m3_abs = pow(P_m3_abs,.5);

	double depth;
	depth = signP * (P*X)(0,2);
	if(P_m3_abs!=0) depth /= P_m3_abs;

	return depth;

}


// Recover camera matrices by x, x', F
template <typename T>
CImg<T> Create3DPointCloud<T>::recoverCamByFMatrix(int id1, int id2){
	
	sout << "Get initial cameras P" << id1 << " and P" << id2
		<< " by fundamental matrix." << std::endl;

	// read in calibration matrices
	cimg_library::CImg<T> cal1, cal2;
	cal1 = calibrations[id1];
	cal2 = calibrations[id2];
	
	ImgTool::printInfo(sout,cal2,"	initial calibration 2: ");
		
	// Get identity camera matrix (at origin)
	CImg<T> P1(4,3);
	P1.fill(0.); P1(0,0) = 1; P1(1,1) = 1; P1(2,2) = 1;

	// Calculate second camera matrix
	CImg<T> fmatrix(3,3);
	fmatrix = this->fundamentalMatrices()(id1,id2);

	// Convert fmatrix to essential matrix
	fmatrix = cal2.get_transpose()*fmatrix*cal1;

	// Calculated SVD of the fundamental matrix.
	CImg<T> U, S, V, W(3,3);
	fmatrix.SVD(U,S,V);

	W.fill(T(0.));
	W(1,0) = (T)-1;
	W(0,1) = (T)1;
	W(2,2) = (T)1;

	// Declare possible cameras
	CImg<T> P2_1, P2_2, P2_3, P2_4, u(1,3);
	u=U.get_crop(2,0,2,2);

	// Possible cam 1
	P2_1 = U*W*V.get_transpose();
	P2_1.append(u);

	// Possible cam 2
	P2_2 = U*W*V.get_transpose();
	P2_2.append(-u);

	// Possible cam 3
	P2_3 = U*W.get_transpose()*V.get_transpose();
	P2_3.append(u);

	// Possible cam 4
	P2_4 = U*W.get_transpose()*V.get_transpose();
	P2_4.append(-u);

	// Select camera where point correspondences are reconstructed
	// with a positive depth in front of the image plane. Only
	// one camera (of the above four) will satisfiy this constraint.
	CImg<T> pt1(1,3), pt2(1,3), X1, X2, X3, X4;


	// Get correspondence inlier to check cameras
	auto it3 = list.get<imgIndex>().lower_bound(id1) ;
	auto it3End = list.get<imgIndex>().upper_bound(id1) ;
	for(int j=0; it3 != it3End ; it3++, j++){
		try{
			const Correspondence<T> &pt2Corr = list(id2, it3->pointIndex);
			if(!it3->outlier && !pt2Corr.outlier)
			{
				pt1 = it3->pt2d();
				pt2 = pt2Corr.pt2d();
				break;
			}
		}
		catch(...){}
	}

	// normalize points
	pt1 = cal1.get_invert() * pt1;
	pt2 = cal2.get_invert() * pt2;

	//ImgTool::printInfo(sout,pt2.get_transpose()*fmatrix*pt1,"e2^TEe1==0?");


	// triangulate points to receive depth
	X1 = getTriangulation(P1, P2_1, pt1, pt2);
	X2 = getTriangulation(P1, P2_2, pt1, pt2);
	X3 = getTriangulation(P1, P2_3, pt1, pt2);
	X4 = getTriangulation(P1, P2_4, pt1, pt2);
			

	CImg<T> P2;

	// Select correct camera.
	if(getDepth(P1, X1) > 0. && getDepth(P2_1, X1) > 0.){
		//sout << "Taking case 1: P2=[UWV^T|u3]" << std::endl;
		P2 = P2_1;
	}
	else if(getDepth(P1, X2) > 0. && getDepth(P2_2, X2) > 0.){
		//sout << "Taking case 2: P2=[UWV^T|-u3]" << std::endl;
		P2 = P2_2;
	}
	else if(getDepth(P1, X3) > 0. && getDepth(P2_3, X3) > 0.){
		//sout << "Taking case 3: P2=[UW^TV^T|u3]" << std::endl;
		P2 = P2_3;
	}
	else if(getDepth(P1, X4) > 0. && getDepth(P2_4, X4) > 0.){
		//sout << "Taking case 4: P2=[UW^TV^T|-u3]" << std::endl;
		P2 = P2_4;
	}
	else{
		throw std::runtime_error("It could not be found a camera which reconstructs positive depths");
	}


	// denormalize P_cam -> P = K * P_cam and P' = K' * P'_cam
	// since the determined projection matrix has the form [R|t]
	P2 = cal2 * P2;			
	P1 = cal1 * P1;
	
	sout << "Calculated P"<<id1<< " and P"<<id2<<":" << std::endl;
	ImgTool::printInfo(sout,P1,"P"+id1);
	ImgTool::printInfo(sout,P2,"P"+id2);

	// Calculate principal point for evaluation purposes
	sout << "Principal points:" << std::endl;
	CImg<T> pp1 = P1.get_crop(0,0,2,2)*P1.get_crop(0,2,2,2).transpose() ;
	CImg<T> pp2 = P2.get_crop(0,0,2,2)*P2.get_crop(0,2,2,2).transpose() ;
	sout << "Principal Point of P"<<id1<< ": (" << pp1[0] << "," 
		<< pp1[1] << "," << pp1[2] << ")" << std::endl;
	sout << "Principal Point of P"<<id2<< ": (" << pp2[0] << "," 
		<< pp2[1] << "," << pp2[2] << ")" << std::endl;

	return P2;

}

/*
// Recover camera matrices by x, x', X
template <typename T>
CImg<T> Create3DPointCloud<T>::recoverCam(int i){
	
	// read in calibration matrices
	cimg_library::CImg<T> cal1;
	cal1 = calibrations[i];

	// Create matrix A to solve A.SVD to recover camera
	CImg<double> A;

	int s1=0, s2=0, s3=0;

	auto it = list.get<imgIndex>().lower_bound(i) ;
	auto itEnd = list.get<imgIndex>().upper_bound(i) ;
	for(; it != itEnd ; it++){
				
		auto it2 = this->pointCloud().begin() ;
		for(; it2 != this->pointCloud().end(); it2++)
		{

			if(it->pointIndex == it2->pointIndex)
			{

				const Correspondence<T> pt3d = *it2 ;
				if(!it->outlier && !pt3d.outlier)
				{

					// Read in 3D point X and 2d point x.
					cimg_library::CImg<double> X, x(1,3);
				
					X = pt3d.point;

					// Read in j-th 2D point from frame i
					x[0] = it->x();
					x[1] = it->y();
					x[2] = it->w();
					x /= x[2] ;

					// Create rowA where which contains two
					// constraints for every point and is appended
					// to matrix A afterwards.
					cimg_library::CImg<T> rowA(12,2);

					// Reset row
					rowA.fill(0);

					// Fill A as on Wang et al., page 92
					rowA(0,0) = X[0];
					rowA(1,0) = X[1];
					rowA(2,0) = X[2];
					rowA(3,0) = X[3];
					rowA(8,0) = -x[0]*X[0];
					rowA(9,0) = -x[0]*X[1];
					rowA(10,0) = -x[0]*X[2];
					rowA(11,0) = -x[0]*X[3];
					rowA(4,1) = X[0];
					rowA(5,1) = X[1];
					rowA(6,1) = X[2];
					rowA(7,1) = X[3];
					rowA(8,1) = -x[1]*X[0];
					rowA(9,1) = -x[1]*X[1];
					rowA(10,1) = -x[1]*X[2];
					rowA(11,1) = -x[1]*X[3];

					// Append rowA to A
					const char direction = 'y';
					A.append(rowA,direction);

				}

				break;

			}

		}

	}
	
	if(A.size() == 0)
	{
		sout << "Matrix A is empty." << std::endl;
		throw(std::runtime_error("A is singularity"));
	}

	//ImgTool::printInfo(sout,A,"A");

	// Create SVD of A to determine the projective matrix i;
	cimg_library::CImg<double> U, D, V;
	A.SVD(U,D,V);

	// Choose entries of A which deliver minimal A*X results
	// Only those entries of A which result in such a minimal
	// A*X will be applied to a SVD of A2
	// This is done to prevent outliers in A
	std::vector<double> results, results_index;
	for(size_t i=0; i < A.height()/2; i++)
	{
		CImg<double> A2_rows = A.get_crop(0,i*2,11,i*2+1);
		CImg<double> A2_res = A2_rows * V.get_crop(11,0,11,11);
		double res = std::abs( A2_res[0] + A2_res[1] ) ;
		results.push_back(res);
		results_index.push_back(i);
	}

	if(this->min3d2dCorrs() > results.size())
		throw(std::runtime_error("Too less 3d-2d correspondences to retrieve good camera."));

	// Simple increasing sorting of results vector
	for(size_t i=0; i < results.size(); i++)
	{
		for(size_t j=i+1; j<results.size(); j++)
		{
			if(results[j] < results[i])
			{
				// Swap values
				double d1 = results[j];
				results[j] = results[i];
				results[i] = d1;
				// Swap indices
				int idx = results_index[j];
				results_index[j] = results_index[i];
				results_index[i] = idx;
			}
		}
	}

	float c = 0.5;
	int c2 = c * results.size();
	sout << "Using " << c2 << " of " << results.size()
		<< " 3d-2d correspondences to estimate new camera." << std::endl;
	
	CImg<double> A2;
	for(size_t i=0; i < c2; i++)
	{
		A2.append(A.get_crop(0,results_index[i]*2,11,results_index[i]*2+1),'y');
	}

	// Calculate second SVD of A2
	A2.SVD(U,D,V);

	// Generate projective matrix i from V
	CImg<T> P(4,3);
	P[0] = V(11,0);
	P[1] = V(11,1);
	P[2] = V(11,2);
	P[3] = V(11,3);
	P[4] = V(11,4);
	P[5] = V(11,5);
	P[6] = V(11,6);
	P[7] = V(11,7);
	P[8] = V(11,8);
	P[9] = V(11,9);
	P[10] = V(11,10);
	P[11] = V(11,11);

	// Normalize new projection matrix by focal length
	// to get comparable values for all matrices
	P *= cal1[0]/P[0] ;

	ImgTool::printInfo(sout, P, "Pi");
	CImg<T> pp = P.get_crop(0,0,2,2)*P.get_crop(0,2,2,2).transpose() ;
	pp /= pp[2] ;
	sout << "Principal Point of P" << i << ": (" << pp[0] << "," 
		<< pp[1] << "," << pp[2] << ")" << std::endl;

	// Save projective matrix which is the last row of V in CImgList
	return P;

}
*/

// requires #include <limits>
template<typename T>
inline bool isinf(T value)
{
return std::numeric_limits<T>::has_infinity &&
value == std::numeric_limits<T>::infinity();
}


template <typename T>
void Create3DPointCloud<T>::initSlots(){
	op_2d.init(this, "Pseudo2DCorrespondences", "CorrespondenceList");
	op_3d.init(this, "Pseudo3DCloud", "CorrespondenceList");
	bpCam.init(this, "bpcam", "IndexedMatrixList");
	pbaCam.init(this, "pbacam", "IndexedMatrixList");
	pbaCalibration.init(this, "pbacal", "CImgList<T>");
}

template <typename T>
Create3DPointCloud<T>::~Create3DPointCloud(){
	
#ifdef USE_PCL
	if(viewer)
		delete viewer;
	
#endif
}

#endif /* _CREATE3DPOINTCLOUD_HXX_ */
