#ifndef CREATE3DPOINT_PLUGINS_H
#define CREATE3DPOINT_PLUGINS_H

#include <charon-utils/CImg.h>
#include <hekate/Correspondence.hxx>
#include <hekate/IndexedMatrix.h>
#include <pba/pba.h>
#include <pba/util.h>

#include <opencv2/core/core.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/highgui/highgui.hpp"
#include <opencv2/opencv.hpp>

#include <hekate/openCvUtil.hxx>	
#include <charon-utils/ImgTool.hxx>

using namespace cimg_library;
using namespace hekate;

/* 
	---- Recover Cam by RANSAC
*/

/// Recover camera matrices by OpenCV solvePnPRansac
template <typename T>
IndexedMatrix recoverCamRANSAC(int index, const CImgList<T> &calibrations, const CorrespondenceList<T> &list,
						CorrespondenceList<T> &pointCloud, const int _nOIt, const float _reprojError, 
						const float _minInlierCount, const int _flags, const IndexedMatrixList &cams,
						bool removeOutliers){

	sout << "--------------------------" << std::endl;
	sout << "Pose estimation for camera " << index << std::endl;

	if(_flags > 2 || _flags < 0 ) 
		throw(std::range_error("recoverCamRANSAC: Wrong value for flags!"));

	// read in calibration matrices
	CImg<T> cal;
	if( index > calibrations.size() - 1 )
		cal = calibrations[0];
	else
		cal = calibrations[index];

	CImgList<T> list3D, list2D;
			 
	// Read 3d and 2d points into open cv vectors
	std::vector< cv::Point3f > cv_3d;
	std::vector< cv::Point2f > cv_2d;
	std::vector< std::pair<int,int> > corrLUT;

	auto it2d = list.get<imgIndex>().lower_bound(index) ;
	auto it2dEnd = list.get<imgIndex>().upper_bound(index) ;
	for(; it2d != it2dEnd ; it2d++){
				
		auto it3d = pointCloud.begin() ;
		auto it3dEnd = pointCloud.end() ;
		for(; it3d != it3dEnd; it3d++)
		{

			if(it2d->pointIndex == it3d->pointIndex)
			{

				if(!it2d->outlier && !it3d->outlier)
				{
					
					float w = it3d->w() ;
					cv::Point3f cv_3dpt( it3d->x()/w , it3d->y()/w , it3d->z()/w );
					cv_3d.push_back( cv_3dpt );
					list3D.push_back( it3d->point );

					float w2 = it2d->w() ;
					cv::Point2f cv_2dpt( it2d->x()/w2 , it2d->y()/w2 );
					cv_2d.push_back( cv_2dpt );
					list2D.push_back( it2d->point );

					corrLUT.push_back( std::pair<int,int>(it3d->imgIndex,it3d->pointIndex) );
					
					// Only for evaluation purposes
					if(cams.size() != 0){
						CImg<T> test = cams[index] * it3d->point;
						test /= test[2];
						float s = test[0] - it2d->point[0] + test[1] - it2d->point[1] ;
						//sout << s << std::endl;
						if(index != cams[index].getIndex1())
							sout << "WRONGGGGG" << std::endl;
					}

				}					
			}
		}
	}

	if(cv_3d.size() == 0)
		throw(std::runtime_error("No 3D-2D correspondences for this view."));
	
	// Estimate camera pose by Open CV RANSAC ---------------------------------

	// inlier points
	vector< int > inlierIndices;

	bool useExtrinsicGuess	= false;
	int minInliersCount		= _minInlierCount; 
	if( minInliersCount >= cv_3d.size() ){
		minInliersCount = 16;
		sout << "Choosing minimal inlier count: " << minInliersCount << std::endl;
	}
	sout << "minInliersCount: " << minInliersCount << std::endl;
	sout << "reprojError: " << _reprojError << std::endl;
	sout << "Number of iterations: " << _nOIt << std::endl;

	cv::Mat io_cv_3d2(cv_3d, CV_32FC1);
	cv::Mat io_cv_2d2(cv_2d, CV_32FC1);
	
	double _cm[9] = {cal[0], cal[1], cal[2], cal[3], cal[4], cal[5], cal[6], cal[7], cal[8] }; 
	//double _cm[9] = {1,0,0,0,1,0,0,0,1 }; 
	cv::Mat io_cali(3, 3, CV_32FC1);
	for (int i = 0; i < 3; i++)
		for (int j = 0; j < 3; j++)
			io_cali.at<float>(i, j) = cal[j+3*i];

	cv::Mat io_rvec, io_tvec;
	// #define CV_ITERATIVE 0
	cv::solvePnPRansac(cv_3d, cv_2d, io_cali, cv::Mat(1, 8, CV_32F, cv::Scalar::all(0)),
			io_rvec, io_tvec, useExtrinsicGuess, _nOIt, _reprojError, minInliersCount, inlierIndices) ;
	
	// Remove all 3D Points which are not on inlier list ------------------------

	// Create vector mask with size of number of 3D-2D corrs (0=outlier,1=inlier)
	std::vector< int > inlierList(cv_3d.size(),0);
	for(size_t i = 0; i < inlierIndices.size(); i++){
		inlierList[ inlierIndices[i] ] = 1 ;
	}

	// Mark outliers in point cloud
	if( inlierIndices.size() != cv_3d.size() && inlierIndices.size() >= minInliersCount 
			&& removeOutliers  ){

		sout << "Number of points in initial 3d point cloud = " 
			<< pointCloud.size() << std::endl;
		sout << "Removing 3D outliers by solvePnP result..." << std::endl;

		for(size_t i=0; i<cv_3d.size(); i++){
			if( inlierList[i] == 0 ){
				pointCloud( corrLUT[i].first, corrLUT[i].second ).outlier = true;
				//sout << inlierList[i] << " " << corrLUT[i].first
				//	<< " " << corrLUT[i].second << std::endl;
			}
		}

		//pointCloud.removeOutliers();
		sout << "Number of points in corrected point cloud = " 
			<< pointCloud.size() << std::endl;

	}

	// Attention: Output rmat maybe has data format CV_64F
	// access over rmat.at<double>(i,j)

	cv::Mat rmat;
	cv::Rodrigues(io_rvec, rmat);
	//CImg<double> R = openCvUtil::Mat2CImg<double>(rmat) ;
	CImg<double> t = openCvUtil::Mat2CImg<double>(io_tvec) ;
	//CImg<double> Rrodriges = openCvUtil::Mat2CImg<double>(io_rvec) ;
	//ImgTool::printInfo(sout,Rrodriges,"Rrodriges");
	CImg<T> cam(4,3); cam.fill(0);
	cam(0,0) = rmat.at<double>(0,0);
	cam(1,0) = rmat.at<double>(0,1);
	cam(2,0) = rmat.at<double>(0,2);
	cam(3,0) = t[0];
	cam(0,1) = rmat.at<double>(1,0);
	cam(1,1) = rmat.at<double>(1,1);
	cam(2,1) = rmat.at<double>(1,2);
	cam(3,1) = t[1];
	cam(0,2) = rmat.at<double>(2,0);
	cam(1,2) = rmat.at<double>(2,1);
	cam(2,2) = rmat.at<double>(2,2);
	cam(3,2) = t[2];
	
	// P = K *(R|t)
	cam = cal * cam;

	char buffer[3];
	sprintf(buffer, "P%d", index);
	ImgTool::printInfo(sout, cam, buffer);

	/*
	int out = 0;
	for(size_t i = 0; i < list3D.size(); i++){
		CImg<T> pt = cam * list3D[i];
		pt /= pt[2];
		//ImgTool::printInfo(sout,pt,"pt");
		//ImgTool::printInfo(sout,list2D[i],"list2D[i]");
		CImg<T> diff = list2D[i]/list2D[i][3] - pt;
		double dist = std::sqrt( double(diff[0]*diff[0] + diff[1]*diff[1]) );
		if(dist > _reprojError ){
			//sout << dist << std::endl;
			out++;
		}
	}	

	sout << "Number of outliers (BPE>" << _reprojError << ") = " << out << std::endl;
	*/
	return IndexedMatrix(cam);

}


/// Recover camera matrices by OpenCV solvePnPRansac
template <typename T>
IndexedMatrix recoverCamRANSAC(	int index, 
							const CImgList<T> &calibrations, 
							const CorrespondenceList<T> &list,
							CorrespondenceList<T> &pointCloud){

	int nOIt = 300;
	float reprojError = 4;
	int minInlierCount = 20;
	int flags = 0;
	IndexedMatrixList cams;

	return recoverCamRANSAC( index, calibrations, list,
			pointCloud, nOIt, reprojError, minInlierCount, flags, cams, false );

}



// Recover camera matrices by x, x', X
template <typename T>
IndexedMatrix recoverCam(	int i, 
					const CImgList<T> &calibrations,
					const hekate::CorrespondenceList<T> &corr2D,
					const hekate::CorrespondenceList<T> &corr3D	){

	const CImg<T> &cal1 = calibrations[i];

	// Create matrix A to solve A.SVD to recover camera
	CImg<double> A;

	int s1=0, s2=0, s3=0;

	auto it = corr2D.get<imgIndex>().lower_bound(i) ;
	auto itEnd = corr2D.get<imgIndex>().upper_bound(i) ;
	for(; it != itEnd ; it++){
				
		auto it2 = corr3D.begin() ;
		for(; it2 != corr3D.end(); it2++)
		{

			if(it->pointIndex == it2->pointIndex)
			{
				
				const Correspondence<T> pt2d = *it ;
				const Correspondence<T> pt3d = *it2 ;
				if(!pt2d.outlier && !pt3d.outlier)
				{

					// Read in 3D point X and 2d point x.
					cimg_library::CImg<double> X, x(1,3);
				
					X = pt3d.pt3d();

					// Read in j-th 2D point from frame i
					x = pt2d.pt2d();

					// Create rowA where which contains two
					// constraints for every point and is appended
					// to matrix A afterwards.
					cimg_library::CImg<T> rowA(12,2);

					// Reset row
					rowA.fill(0);

					// Fill A as on Wang et al., page 92
					rowA(0,0) = X[0];
					rowA(1,0) = X[1];
					rowA(2,0) = X[2];
					rowA(3,0) = X[3];
					rowA(8,0) = -x[0]*X[0];
					rowA(9,0) = -x[0]*X[1];
					rowA(10,0) = -x[0]*X[2];
					rowA(11,0) = -x[0]*X[3];
					rowA(4,1) = X[0];
					rowA(5,1) = X[1];
					rowA(6,1) = X[2];
					rowA(7,1) = X[3];
					rowA(8,1) = -x[1]*X[0];
					rowA(9,1) = -x[1]*X[1];
					rowA(10,1) = -x[1]*X[2];
					rowA(11,1) = -x[1]*X[3];

					// Append rowA to A
					const char direction = 'y';
					A.append(rowA,direction);

				}

				break;

			}

		}

	}
	
	if(A.size() == 0)
	{
		sout << "Matrix A is empty." << std::endl;
		throw(std::runtime_error("A is singularity"));
	}

	//ImgTool::printInfo(sout,A,"A");

	// Create SVD of A to determine the projective matrix i;
	cimg_library::CImg<double> U, D, V;
	A.SVD(U,D,V);
	
	/*
	// Choose entries of A which deliver minimal A*X results
	// Only those entries of A which result in such a minimal
	// A*X will be applied to a SVD of A2
	// This is done to prevent outliers in A
	std::vector<double> results, results_index;
	for(size_t i=0; i < A.height()/2; i++)
	{
		CImg<double> A2_rows = A.get_crop(0,i*2,11,i*2+1);
		CImg<double> A2_res = A2_rows * V.get_crop(11,0,11,11);
		double res = std::abs( A2_res[0] + A2_res[1] ) ;
		results.push_back(res);
		results_index.push_back(i);
	}

	if(this->min3d2dCorrs() > results.size())
		throw(std::runtime_error("Too less 3d-2d correspondences to retrieve good camera."));

	// Simple increasing sorting of results vector
	for(size_t i=0; i < results.size(); i++)
	{
		for(size_t j=i+1; j<results.size(); j++)
		{
			if(results[j] < results[i])
			{
				// Swap values
				double d1 = results[j];
				results[j] = results[i];
				results[i] = d1;
				// Swap indices
				int idx = results_index[j];
				results_index[j] = results_index[i];
				results_index[i] = idx;
			}
		}
	}

	float c = 0.5;
	int c2 = c * results.size();
	sout << "Using " << c2 << " of " << results.size()
		<< " 3d-2d correspondences to estimate new camera." << std::endl;	
	CImg<double> A2;
	for(size_t i=0; i < c2; i++)
	{
		A2.append(A.get_crop(0,results_index[i]*2,11,results_index[i]*2+1),'y');
	}
	// Calculate second SVD of A2
	A2.SVD(U,D,V);
	*/
	
	// Generate projective matrix i from V
	IndexedMatrix P(4,3);
	P[0] = V(11,0);
	P[1] = V(11,1);
	P[2] = V(11,2);
	P[3] = V(11,3);
	P[4] = V(11,4);
	P[5] = V(11,5);
	P[6] = V(11,6);
	P[7] = V(11,7);
	P[8] = V(11,8);
	P[9] = V(11,9);
	P[10] = V(11,10);
	P[11] = V(11,11);

	// Normalize new projection matrix by focal length
	// to get comparable values for all matrices
	P *= cal1[0]/P.sum() ;

	char buffer[10];
	sprintf(buffer,"P%d",i);
	ImgTool::printInfo(sout, P, buffer);
	CImg<T> pp = P.get_crop(0,0,2,2)*P.get_crop(0,2,2,2).transpose() ;
	pp /= pp[2] ;
	sout << "Principal Point of P" << i << ": (" << pp[0] << "," 
		<< pp[1] << "," << pp[2] << ")" << std::endl;

	// Save projective matrix which is the last row of V in CImgList
	return P;

}


/*
	---- BackProjection3D2D ----
*/

inline double getDist(const CImg<float> &pt1, const CImg<float> &pt2)
{
	return std::sqrt( pow((double)(pt1[0]-pt2[0]),2.0) + pow((double)(pt1[1]-pt2[1]),2.0) );		
}


template <typename T>
void pluginBackProjection3D2D(IndexedMatrixList &cameras, CorrespondenceList<T> &points3D,
	 						  CorrespondenceList<T> &corrsBack, CorrespondenceList<T> &corrsInput,
							  float maxBackProjectionError, int min3d2dCorrs){

	int nPoint = points3D.size();
	size_t nCams = cameras.size();

	sout << "3D Point cloud size: " << nPoint << std::endl;

	int nInlier = 0;

	std::vector<int> corrs3d2d(nCams);

	auto it = points3D.begin();
	auto itEnd = points3D.end();
	while(it != itEnd)
	{

		if(it->outlier){
			it++;
			continue;
		}
	
		bool inlier = true;
		int inlierCount = 0;

		for(size_t i = 0; i < nCams; i++)
		{
				
			size_t imageId = cameras[i].getIndex1();

			try{

				// Try to get correspondence
				Correspondence<T> &correspondence = corrsInput(imageId, it->pointIndex);
			
				// Continue if outlier
				if(correspondence.outlier)
					continue;

				CImg<float> x2 = correspondence.pt2d();

				// Get camera 
				CImg<double> cam = cameras[i];
		
				// Get backprojected point
				CImg<double> X, x;
				X = (CImg<double>) it->pt3d();
				x = cam * X;

				if(x[2] != 0)
					x /= x[2];
				else
					continue;

				// Get distance
				double dist = getDist(x, x2);
				if( dist > maxBackProjectionError )
				{
					inlier = false;
					correspondence.outlier = true;
				}
				else
				{
					inlierCount++;
					corrs3d2d[i]++;
					corrsBack.push_back( Correspondence<T>(x[0], x[1], 0, 1, it->pointIndex, imageId, false, it->R, it->G, it->B) );
				}
			}
			catch(...){}
		
		}

		if(inlierCount <= 0)
		{
			it = points3D.erase(it);
		}
		else{
			nInlier++;
			it++;
		}
		
	}

	// Check, if there are enough 3d-2d correspondences for every cam
	for(size_t i = 0; i < nCams; i++)
	{
		if(corrs3d2d[i] < std::min(min3d2dCorrs,nPoint))
		{
			// Write information to log, if cameras has not been set to degenerated before
			if(!cameras[i].degenerated)
			{
			sout << "Camera P" << cameras[i].getIndex1() << " set to degenerated"
				<< " due to poor 3d-2d correspondences (" << corrs3d2d[i] << ")" << std::endl;
			}
			// Set cam to degenerated
			cameras[i].degenerated = true;

		}
	}

	sout << nPoint - nInlier << " outliers of " << nPoint << " points "
		<< "removed due to bad back projections. (error > " 
		<< maxBackProjectionError << ")" << std::endl;
	sout << points3D.size() <<  " 3d inliers have been put to output slot." << std::endl;
	sout << "Real Size: " << points3D.size() << std::endl;
}

/*
	---- ParallelBundleAdjustment ----
*/

/// convert 3d points, cameras and 2d correspondences to PBA conform types
/** 
	Set takeOverlap, if only the overlap of given data should be forwarded to pba
*/
template <typename T>
bool loadPBAData(vector<CameraT> &camera_data, vector<Point3D> &point_data, 
	vector<Point2D> &measurements, vector<int> &camidx, vector<int> &pointidx, vector<int> &point3didx,
	const IndexedMatrixList &cameras, const CorrespondenceList<T> &correspondences, const CorrespondenceList<T> &ptcloud, 
	const CImgList<T> &calibrationMatrix, const int minPtViews, const bool takeOverlap = false, bool SetConstantCamera = false )
{
	
	//const IndexedMatrixList &cameras = cameraMatrices().getNonDegeneratedList();
	//const CorrespondenceList<T> &ptcloud = this->pointCloud();
	size_t nCam = cameras.sizeNonDegenerated();
	size_t nCamAll = cameras.size();
	size_t nPoints = ptcloud.size();
	size_t nPointsImages = correspondences.sizeImages();
	
	sout << "PBA.nCam = " << nCam << std::endl;
	sout << "PBA.nCamAll = " << nCamAll << std::endl;
	sout << "PBA.nPoints = " << nPoints << std::endl;
	sout << "correspondences.sizeImages = " << nPointsImages << std::endl;	
	
	// Exit if none or only one camera is available
	// or if less than 8 points are available
	if(nCam < std::min(minPtViews,2) ){
		sout << "Only " << nCam << " camera is available!" << std::endl;
		return 0;
	}
	else if(nPoints < 8){
		sout << "Only " << nPoints << " are available" << std::endl;
		return 0;
	}
	
	
	// Iterate over all cameras	
	sout << "Read in " << nCam << " non degenerated cameras of "
		<< nCamAll << " overall cameras to PBA" << std::endl;
	camera_data.clear();
	camera_data.resize(nCam);
	int nCamCount = 0;
	for(size_t i = 0; i < nCam; ++i)
	{

		// ATTENTION: Since this model requires K and R separately
		// P = K[R|t] first has to be decomposed to R and t where
		// K is given as initial guess

		// Get parameters
		const CImg<double> cam = (CImg<double>) cameras(cameras[i].getIndex1());
		double f;
		if (calibrationMatrix.size() > 0){
			f = (double) calibrationMatrix[cameras[i].getIndex1()][0];
		}else{
			f = cameras[i].f;
		}
		sout << "load f = " << f << std::endl;

		/*
		CImg<double> calibration = (CImg<double>) calibrationMatrix[0];

		CImg<double> calibrationNew(3,3); 
		calibrationNew.fill(0); calibrationNew(0,0) = f; calibrationNew(1,1) = f; calibrationNew(2,2) = 1;
		CImg<double> R = (calibrationNew.get_invert() * cam).crop(0,0,2,2);
		CImg<double> X = (calibrationNew.get_invert() * cam).crop(3,0,3,2);


		//double q[9] = {R(0,0), R(0,1), R(0,2), R(1,0), R(1,1), R(1,2), R(2,0), R(2,1), R(2,2) };
		//double q[9] = {R[0], R[3], R[6], R[1], R[4], R[7], R[2], R[5], R[8]};
		double q[9] = {R[0], R[1], R[2], R[3], R[4], R[5], R[6], R[7], R[8]};
		double d[2] = {1, 1};
		double translation[3] = {X(0,0), X(0,1), X(0,2)}; 

		*/

		double q[9] = {cam(0,0) / f, cam(1,0) / f, cam(2,0) / f, cam(0,1) / f, cam(1,1) / f, cam(2,1) / f, cam(0,2) , cam(1,2), cam(2,2)};
		double translation[3] = {cam(3,0) / f, cam(3,1) / f, cam(3,2)};


		// Define camera i
		camera_data[i].SetFocalLength(f);
		camera_data[i].SetTranslation(translation);
		camera_data[i].SetMatrixRotation(q);

		// Set camera constant, if flag is set
		if(SetConstantCamera)
			camera_data[i].SetConstantCamera();

		//camera_data[i].SetNormalizedMeasurementDistortion(d[0]); 

		nCamCount++;		

	}
			
	// Get 3D point locations and save it with new
	// index i, since this loop could deliver
	// random order of point indices and we need ascending
	// point indices for PBA
	sout << "Try to read in " << nPoints << " 3D points to PBA" << std::endl;
	int inconsistentPoints = 0;
	auto it = ptcloud.begin();
	auto itEnd = ptcloud.end();
	
	for(int i = 0; it < itEnd; it++)
	{

		// Only consider point, if it is an inlier
		if(it->outlier)
			continue;
		
		// Read in all corresponding 2d points into temporary
		// vectors. If enough points are available, the temporary
		// vector will be assigned to the actual vector
		vector<Point2D> measurements_temp;
		vector<int> camidx_temp, pointidx_temp;
		std::vector<int> camIds;
		int k = 0;
		//auto it2 = correspondences().get<pointIndex>().lower_bound(it->pointIndex) ;
		//auto it2End = correspondences().get<pointIndex>().upper_bound(it->pointIndex) ;
		bool hasUniquePoints = true;
		auto it2 = correspondences.get<3>().lower_bound(it->pointIndex) ;
		auto it2End = correspondences.get<3>().upper_bound(it->pointIndex) ;
		for(; it2 != it2End ; it2++)
		{

			// Only consider inliers
			if(it2->outlier)
				continue;

			// Check if 2d point has camera
			try{ cameras(it2->imgIndex); }
			catch(...){ continue; }

			// Increase number of points considered
			// for the bundle adjustment
			k++;

			// Check if imgId/pointId pair already exists
			for(size_t j = 0; j<camIds.size(); j++)
			{
				if(it2->imgIndex == camIds[j])
				{
					//sout << "2D point is not unique!" << std::endl;
					hasUniquePoints = false;
					break;
				}
			}
			if(!hasUniquePoints) break;

			size_t camIdOfPosI = 0;
			for(size_t l = 0; l < cameras.size(); l++)
			{
				if(cameras[l].getIndex1() == it2->imgIndex){
					camIdOfPosI = l;
					break;				
				}
			}
			camidx_temp.push_back(camIdOfPosI);
			camIds.push_back(it2->imgIndex);

			// Get point index which is i and so consistent with
			// the new 3d point index
			pointidx_temp.push_back(i);

			// Get 2d point coordinates
			float imx = it2->x();
			float imy = it2->y();
			measurements_temp.push_back( Point2D(imx , imy) );

		}


		// If at least two correspondences are available, the set of
		// 2d points and 3d point will be assigned to the actual vectors
		if(k > minPtViews -1 && hasUniquePoints)
		{

			//sout << *it << std::endl;

			// Add 3d point
			Point3D pt;
			pt.SetPoint((float) it->x(), (float) it->y(), (float) it->z());
			point_data.push_back(pt);
			//sout << "Pt3d: " << pt[0] << ", " << pt[1] << ", " << pt[2] << ";" << std::endl;
			//sout << "iteration " << i << ": " << *it << std::endl;

			point3didx.push_back(it->pointIndex);

			// Add 2d points
			measurements.insert(measurements.end(), measurements_temp.begin(), measurements_temp.end() );

			// Add camids
			camidx.insert(camidx.end(), camidx_temp.begin(), camidx_temp.end() );

			// Add pointids
			pointidx.insert(pointidx.end(), pointidx_temp.begin(), pointidx_temp.end() );
			
			/*
			for(int l = 0; l < k; l++)
			{
			sout << "\tCamIdx: " << camidx_temp[l] << "; ptIdx: " << pointidx_temp[l]
				<< "; Pt2d: " << measurements_temp[l].x << ", " << measurements_temp[l].y << ";" << std::endl;
			}
			*/

			i++;

		}

		if(!hasUniquePoints)
		{
			inconsistentPoints++;
		}


	}

	sout << "Finished read in process" << std::endl;
	sout << inconsistentPoints << " inconsistent correspondence trajectories have "
		<< "been found! (imgId/pointId pairs not unique)" << std::endl;
	return 1;

}



/// Save PBA data types to hekate data type output slots
template <typename T>
bool savePBAData(const vector<CameraT> &camera_data, const vector<Point3D> &point_data, 
		const vector<int> &point3didx,
		const CorrespondenceList<T> &pointCloud, const CImgList<T> &calibrationMatrix, const IndexedMatrixList &cameraMatrices,
		CorrespondenceList<T> &ptOut, IndexedMatrixList &cameras, CImgList<T> &calibrationMatrixOut,
		const bool fixedFocalLength)
{
	
	//CorrespondenceList<T> &ptOut = this->pointCloudOut();
	//CorrespondenceList<T> ptNew = this->pointCloud();
	//IndexedMatrixList &cameras = this->cameraMatricesOut();
	cameras.clear();
	ptOut.clear();
	calibrationMatrixOut = calibrationMatrix;
	//this->calibrationMatrixOut().clear();

	sout << "Save results..." << std::endl;

	// Save bundle adjusted cameras
	for(size_t i = 0; i < camera_data.size(); i++)
	{

		IndexedMatrix cam;

		float m[9], t[3], f;
		camera_data[i].GetMatrixRotation(m);
		camera_data[i].GetTranslation(t);
		f = camera_data[i].GetFocalLength();

		if (calibrationMatrix.size() > 0){
			calibrationMatrixOut[ cameraMatrices[i].getIndex1() ](0,0) = (double) f;
			calibrationMatrixOut[ cameraMatrices[i].getIndex1() ](1,1) = (double) f;
		}

		/*
		CImg<double> calibrationNew(3,3); 
		calibrationNew.fill(0); calibrationNew(0,0) = f; calibrationNew(1,1) = f; calibrationNew(2,2) = 1;

		CImg<double> R(3,3);
		R(0,0) = m[0];	R(1,0) = m[1]; 	R(2,0) = m[2];
		R(0,1) = m[3];	R(1,1) = m[4]; 	R(2,1) = m[5];
		R(0,2) = m[6];	R(1,2) = m[7]; 	R(2,2) = m[8];
		//R(0,0) = m[0];	R(1,0) = m[3]; 	R(2,0) = m[6];
		//R(0,1) = m[1];	R(1,1) = m[4]; 	R(2,1) = m[7];
		//R(0,2) = m[2];	R(1,2) = m[5]; 	R(2,2) = m[8];

		CImg<double> X(1,3);
		X[0] = t[0];	X[1] = t[1];	X[2] = t[2];

		cam = (IndexedMatrix) calibrationNew * (R.append(X));
		*/

		cam = IndexedMatrix(4,3);

		cam(0,0) = m[0] * f;	cam(1,0) = m[1] * f; 	cam(2,0) = m[2] * f; 	cam(3,0) = t[0] * f;
		cam(0,1) = m[3] * f;	cam(1,1) = m[4] * f; 	cam(2,1) = m[5] * f; 	cam(3,1) = t[1] * f;
		cam(0,2) = m[6];		cam(1,2) = m[7]; 		cam(2,2) = m[8];		cam(3,2) = t[2];

		cam.setIndex( cameraMatrices[i].getIndex1() );
		cam.f = f;
		
		// Check NaN
		bool isNan = false;
		for(size_t iii = 0; iii < 12; iii++)
		{
			if( cam[iii] != cam[iii] )
			isNan = true;
		}
		if(isNan){
			sout << "Matrix is NaN -> degenerated" << std::endl;
			cam.degenerated = true;
		}

		//ImgTool::printInfo(sout, cam);
		if(!fixedFocalLength)
			sout << "New focal length f = " << f << std::endl;
		cameras.push_back(cam);

	}

	// Create a copy of input 3D points and replace point by
	// the adjusted 3D points (not all points will be replaced,
	// only the ones, which satisfy the PBA criterion)
	size_t nPointOut = point3didx.size();
	size_t readout_count = 0;
	float readout_percent = 0;
	sout << "Save " << nPointOut << " adjusted 3D points" << std::endl;
	if(nPointOut != point_data.size())
		throw(std::range_error("ParallelBundleAdjustment<T>::saveData: point3didx.size() != point_data.size()"));
	auto it = pointCloud.begin();
	auto itEnd = pointCloud.end();
	for(; it != itEnd; it++)
	{

		for(size_t i = 0; i < nPointOut; i++)
		{

			if(point3didx[i] == it->pointIndex)
			{
	
				float pt[3];
				point_data[i].GetPoint(pt);
				ptOut.push_back(Correspondence<T>(pt[0], pt[1], pt[2], 1, 
					it->pointIndex, it->imgIndex, false, it->R, it->G, it->B));
				readout_count++;

				if( float(readout_count) / float(nPointOut) > readout_percent && readout_percent <= 1)
				{
					sout << int( float(readout_count) / float(nPointOut) * 100) << "%... ";
					readout_percent += 0.2;
				}

				break;

			}

		}

	}

	sout << std::endl;

	// Save list ptNew to output slot ptOut
	//ptOut = ptNew ;

	return 1;

}

template <typename T>
void ParallelBundleAdjustmentPlugin( const IndexedMatrixList &cameraMatrices, const CImgList<T> &calibrationMatrix, const CorrespondenceList<T> &pointCloud,
									const CorrespondenceList<T> &correspondences, IndexedMatrixList &cameraMatricesOut, 
									CorrespondenceList<T> &pointCloudOut, CImgList<T> &calibrationMatrixOut, int minPtViews, 
									bool fixedFocalLength, bool useCUDA, bool takeDataOverlap = false, bool SetConstantCamera = false ) {    

    //CameraT, Point3D, Point2D are defined in src/pba/DataInterface.h
    vector<CameraT>        camera_data;    //camera (input/ouput)
    vector<Point3D>        point_data;     //3D point(input/output)
    vector<Point2D>        measurements;   //measurment/projection vector
    vector<int>            camidx, ptidx, pt3didx;  //index of camera/point/3dpoint for each projection
	

	// load cameras, 3d and 2d points to pba compatible classes
	if(	
		loadPBAData( camera_data, point_data, measurements, camidx, ptidx, pt3didx, cameraMatrices.getNonDegeneratedList(), 
					 correspondences, pointCloud, calibrationMatrix, minPtViews, takeDataOverlap, SetConstantCamera ) 
		)
	{
				
		// Define PBA device once
		ParallelBA::DeviceT device = ParallelBA::PBA_CPU_DOUBLE;
		if(useCUDA){
			sout << "Use CUDA PBA Device." << std::endl;
			device = ParallelBA::PBA_CUDA_DEVICE_DEFAULT;
		}
		ParallelBA pba(device);          //You should reusing the same object for all new data

		/////////////////////////////////////////////////////////
		//Parameters can be changed before every call of RunBundleAdjustment
		//But do not change them from another thread when it is running BA.
		//pba.ParseParam(argc, argv);      //indirect parameter tuning from commandline
		//pba.SetFocalLengthFixed(true); //if your focal lengths are calibrated. 
		//           equivalent to pba.GetInternalConfig()->__fixed_focallength = true;
		//pba.EnableRadialDistortion(*); // if you want to enable radial distortion
		//           equivalent to pba.GetInternalConfig()->__use_radial_distortion = *;
		//check src/pba/ConfigBA.h for more details on the parameter system
		////////////////////////////////////
		//Tweaks before bundle adjustment
		//1. For each camera, you can call CameraT::SetConstantCamera() to keep it unchanged.
		//2. pba.SetNextBundleMode(ParallelBA::BUNDLE_ONLY_MOTION) //chose a truncated mode?	
		////////////////////////////////////////////////////////////////

		cout << "Initialize ParallelBA object by read in data" << std::endl;
		pba.SetCameraData(camera_data.size(),  	&camera_data[0]);                        //set camera parameters
		pba.SetPointData(point_data.size(), 	&point_data[0]);                         //set 3D point data
		pba.SetProjection(measurements.size(), 	&measurements[0], &ptidx[0], &camidx[0]);//set the projections

		//pba.EnableRadialDistortion( ParallelBA::PBA_MEASUREMENT_DISTORTION ); //(default 0, 1 for projection distortion, -1 for measurement distortion)

		//pba.SetNextBundleMode(ParallelBA::BUNDLE_ONLY_MOTION);

		// Set up focal length handling
		if(fixedFocalLength)
		{
			pba.SetFocalLengthFixed(1);
			sout << "	ParallelBundleAdjustment assumes a fixed "
				<< "focal length for all cameras." << std::endl;
		}

		sout << "	cameras: " << camera_data.size() << std::endl;
		sout << "	3D points: " << point_data.size() << std::endl;
		sout << "	2D points: " << measurements.size() << std::endl;
		sout << "	2D point indices: " << ptidx.size() << std::endl;
		sout << "	camera indices: " << camidx.size() << std::endl;
    
		//WARNING: measumrents must be stored in correct order
		//all measurements (in different images) for a single 3D point must be stored continously, 
		//Currently, there is no order verification internally.
		//Basically, ptidx should be non-decreasing 
    
		//pba.SetTimeBudget(10);      //use at most 10 seconds?
		sout << "Run bundle adjustment" << std::endl;
		pba.RunBundleAdjustment();    //run bundle adjustment, and camera_data/point_data will be modified


		// Save adjusted data
		savePBAData(camera_data, point_data, pt3didx, pointCloud, calibrationMatrix, cameraMatrices, pointCloudOut,
					cameraMatricesOut, calibrationMatrixOut, fixedFocalLength);

		sout << "Corrected cameras: " << camera_data.size() << std::endl;		
		if(cameraMatrices.size() != cameraMatricesOut.size()){
			sout << "\t" << cameraMatrices.size() - cameraMatricesOut.size()
				<< " cameras removed by bundle adjustment." << std::endl;
		}
		sout << pointCloudOut.size() << " of " << pointCloud.size()
			<< " 3d points have been optimized by PBA." << std::endl;
		sout << "\t--> " << pointCloud.size() - pointCloudOut.size() 
			<< " points removed" << std::endl;
		
		if(calibrationMatrix.size() != calibrationMatrixOut.size()){
			sout << "\t" << calibrationMatrix.size() - calibrationMatrixOut.size() 
				<< "calibrations removed by bundle adjustment." << std::endl;
		}

	}
	// If data could not be read in
	else
	{
		sout << "Could not read in appropriate data. Skip PBA..." << std::endl;
		throw( std::runtime_error("ParallelBundleAdjustment<T>::execute(): Bad PBA data") );
	}


}

/// Calculate overlap between camera indices, 2d-point indices and 3d point indices
template <typename T>
void GetDataOverlap( IndexedMatrixList &cams, CImgList<T> &cali, const CorrespondenceList<T> &points2D, const CorrespondenceList<T> &points3D )
{
	IndexedMatrixList camsOut;
	CImgList<T> caliOut;

	sout << "Calculate data overlap for " << cams.size() << " cameras." << std::endl;
	
	for( size_t i = 0; i < cams.size(); i++ )
	{
		const size_t indexImg = cams[i].getIndex1();
		if( points2D.existImgIndex(indexImg) )
		{
							
			// Save camera and calibration matrix
			camsOut.push_back( cams[i] );
			caliOut.push_back( cali[i] );
		
		}

	}
	

	cams = camsOut;
	cali = caliOut;

	
	sout << "Found " << cams.size() << " camera matrices with corresponding 2D points" << std::endl;
		
}

#endif

