/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file CreateTrackingMatrix.hxx
 *  Implementation of the parameter class CreateTrackingMatrix.
 *  \author Moritz Becker
 *  \date 29.08.2011
 */

#ifndef _CREATETRACKINGMATRIX_HXX_
#define _CREATETRACKINGMATRIX_HXX_

#include <hekate/CreateTrackingMatrix.h>
#include <charon-utils/ImgTool.hxx>
#include <cmath>

template <typename T>
CreateTrackingMatrix<T>::CreateTrackingMatrix(const std::string& name) :
		TemplatedParameteredObject<T>(
				"CreateTrackingMatrix", name,
				"Compares subsequent feature coordinates and uses them to create a tracking matrix."
		)
			,correspondences(false, false) 
			
			
{
	ParameteredObject::_addParameter (filename, "filename", 
	"Path where the tracking matrix is saved.", "filewrite"
	);
	
	
	ParameteredObject::_addOutputSlot(trackingMatrixOut, "trackingMatrixOut", 
	"Optimized tracking matrix.", 
	"CImgList<T>"); 
	

	ParameteredObject::_addInputSlot(correspondences, "correspondences", 
	"Input correspondences.", 
	"vector<vector<vector<T> > >"); 

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void CreateTrackingMatrix<T>::execute() {
	
	std::string inFile = this->filename() ;

	// Check if input string already contains file type .dat
	bool saveFile = false;
	if(inFile!="" && inFile.substr(inFile.size()-9,9)!=".cimg"){
		saveFile = true;
		inFile.append(".cimg");
	}

	sout << "Read in tracking matrix." << std::endl;

	cimg_library::CImg<T> W, W2;
	
	int heightW = this->correspondences().size();

	for(int corIt=0; corIt<heightW; corIt++){
		// Create new lines which will be appended to the existing image
		cimg_library::CImg<T> newLine;

		typename std::vector<std::vector<T> >::const_iterator ptIt1 =
			this->correspondences()[corIt].begin();
		const typename std::vector<std::vector<T> >::const_iterator ptEnd1 =
			this->correspondences()[corIt].end();

		for(; ptIt1 != ptEnd1 ; ptIt1++){

			cimg_library::CImg<T> pt1(1,3);	

			pt1[0]=(T)(*ptIt1)[0];
			pt1[1]=(T)(*ptIt1)[1];
			pt1[2]=(T)(*ptIt1)[2];

			//pt1.noise(0.01);

			newLine.append(pt1);

		}
		
		W.append(newLine,'y');
		//ImgTool::printInfo(sout,newLine,"newLine");
		//ImgTool::printInfo(sout,W,"W");
		
	}


	// Check tracking matrix for outliers
	int numberOfPoints = 0;
	for(int i=0; i<W.width(); i++){
		bool isValid = true;
		for(int j=0; j<W.height(); j=j+3){
			if(!checkPoint(W.get_crop(i,j,i,j+2))){
				isValid = false;
				break;
			}
		}
		if(isValid) numberOfPoints++;
	}
	sout << "Number of points, which exist in every "
			<< "view: " << numberOfPoints << std::endl;

	
	if(W.width()!=0){
		ImgTool::printInfo(sout,W,"W");
		if(saveFile)
			W.save(inFile.c_str());
		this->trackingMatrixOut().push_back(W);
	}
	else{
		sout << "Tracking matrix is empty." << std::endl;
		throw(std::runtime_error("Tracking matrix is empty."));
	}

}

double Round(double figure, unsigned int positions)
{
    figure *= pow(10., (double)positions);
    if (figure >= 0)
        figure = floor(figure + 0.5);
    else
        figure = ceil(figure - 0.5);
    figure /= pow(10., (double)positions);
    return figure;
}

/// Checks if point is an outlier, if yes the function returns false
template <typename T>
bool checkPoint(cimg_library::CImg<T> pt){
	if(pt[0] == 0 && pt[1] == 0 && pt[2] == 0)
		return false;
	else
		return true;
}


#endif /* _CREATETRACKINGMATRIX_HXX_ */



