/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeEvaluation.hxx
 *  Implementation of the parameter class DepthEdgeEvaluation.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 27.03.2013
 */

#ifndef _DEPTHEDGEEVALUATION_HXX_
#define _DEPTHEDGEEVALUATION_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/DepthEdgeEvaluation.h>

template <typename T>
DepthEdgeEvaluation<T>::DepthEdgeEvaluation(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthEdgeEvaluation", name,
			"<h2>Compares predicted depth edges with a given ground truth to "
			"evaluate results</h2><br>"
			"The module determines which pixels are predicted correctly and "
			"which pixels are wrongly identified as depth edges."
		), mask(true, false)
{
	ParameteredObject::_addInputSlot(
		groundTruth, "groundTruth",
		"Ground truth images.",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		prediction, "prediction",
		"Predicted pixels. "
		"Continuous probabilities are possible, but a threshold defining "
		"depth edges has to be given by the parameter 'edgeThreshold.'",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		mask, "mask",
		"Masks to limit ground truth pixels and predicted pixels to the masked pixels. "
		"Number of given masks has to be equal to number of given predictions. "
		"Maks are useful, since a manually given ground truth and the predicted pixels can "
		"be reduced to pixels, lying on an edge. The drawback is, that the edge masks not only "
		"reduces given values to edges but also may remove values from actual edges, when the mask "
		"is wrong. Check the quality of the mask to classify the evaluation.",
		"CImgList<T>");
	
	ParameteredObject::_addOutputSlot(
		predictionEvaluated, "predictionEvaluated",
		"Evaluated pixels separated in RGB channels (R=false negative,G=positive,B=false positive)",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		statistic, "statistic",
		"Image with statistical values. Columns as follows:"
		"<ol>"
		"<li>Total pixels</li>"
		"<li>Total ground truth pixels</li>"
		"<li>positive predicted pixels</li>"
		"<li>false positive predicted pixels</li>"
		"<li>False positive pixels with a distance to ground truth taller than 1px</li>"
		"<li>False positive pixels with a distance to ground truth taller than 2px</li>"
		"<li>false negative predicted pixels</li>"
		"</ol>"
		"Every image of the input list is saved to one row.",
		"CImgList<T>");

	ParameteredObject::_addParameter< float >(
		edgeThreshold, "edgeThreshold",
		"If continuous probability images are given, a threshold for depth edges "
		"has to be defined. (0-1)", 0.5);

	/// Use index list to assign indices to ground truth images
	/** If no index list is given, image pairs are assigned in increasing order */
	ParameteredObject::_addParameter< bool >(
		useGtIndices, "useGtIndices",
		"Use index list to assign indices to ground truth images. "
		"If no index list is given, image pairs are assigned in increasing order.", true); 

	ParameteredObject::_addParameter(
		indices, "indices",
		"List of ground truth indices, according to given ground truth list.");
	
	ParameteredObject::_setTags("Hekate;Edges") ;

}

template <typename T>
void DepthEdgeEvaluation<T>::execute() {

	const unsigned int size = std::min( int(this->prediction().size()), int(this->groundTruth().size() ) );
	
	sout << "Comparing " << size << " ground truth / prediction pairs" << std::endl;
	if( int(this->prediction().size()) != int(this->groundTruth().size() ) )
	{
		sout << "Ensure, that the image pairs correspond to each other, since the number of "
			<< "ground truth images and the number of prediction images differ! \nCheck the index list "
			<< "to ensure the correct image pairs are matched." << std::endl;
	}
	// If masks are given, there have to be as many masks as prediction images
	if( this->mask.connected() )
		if( this->mask().size() != this->prediction().size() )
			throw( std::range_error("Number of masks and predictions inconsistent.") );

	if( this->useGtIndices() )
		if( this->indices().size() != size )
			throw( std::range_error("Not enough indices for given ground truth images") );

	// Statistical image
	/* Colums as follows:
	   - ground truth pixels
	   - predicted as depth edge
	   - positive predicted pixels
	   - false positive predicted pixels
	   - False positive pixels with a distance to ground truth taller than 1p
	   - False positive pixels with a distance to ground truth taller than 2px
	   - false negative predicted pixels
	*/
	CImg<float> stats(10, size, 1, 1);

	for( size_t i = 0; i < size; i++ )
	{

		unsigned int index = i;
		if( this->useGtIndices() )
			index = this->indices[i];

		sout << "Evaluating prediction[" << index << "] and groundTruth[" << i << "]" << std::endl;

		// Get depth edges
		CImg<float> edge = this->prediction()[index].get_channel(0).normalize(0,1).threshold( this->edgeThreshold() );

		// Get ground truth
		CImg<float> gt = this->groundTruth()[i].get_channel(0).quantize(2).normalize(0,1);

		// Check dimensions
		assert( edge.size() == gt.size() );

		// Reduce to masked pixels
		if( this->mask.connected() )
		{
			sout << "Use mask[" << index << "] to reduce depth edges." << std::endl;
			CImg<float> masked = this->mask()[index].get_channel(0).normalize(0,1);
			assert( edge.size() == masked.size() );
			edge.min( masked );
			gt.min( masked );
		}

		// Get positive prediction (intersection with ground truth)
		CImg<float> positive = gt.get_min( edge );
		
		// Get false possitive prediction (max(ground truth,prediction) - ground truth)
		CImg<float> falsePositive = gt.get_max( edge ) - gt;
		
		// Get missed prediction (ground truth - intersection)
		CImg<float> falseNegative = gt - positive;
		// If a mask is connected, reduce false negative pixels to masked pixels

		// Evaluate distance between falsePositive pixels and ground truth pixels
		// Therefore a histogram is generated
		sout << "Calculate distances for false positive pixels" << std::endl;
		CImg<float> dist = getDistHistogram( falsePositive, gt );

		// Save stats -----------------------------------------------------------
		float predictedOverall = positive.sum() + falsePositive.sum() - dist[0] + falseNegative.sum();
		
		// Get positive prediction (intersection with ground truth)
		stats(0,i) = gt.size();
		stats(1,i) = gt.sum();
		stats(2,i) = positive.sum();
		stats(3,i) = int(positive.sum() / gt.sum() * 100 + 0.5);
		stats(4,i) = falsePositive.sum();
		stats(5,i) = falsePositive.sum() - dist[0];
		stats(6,i) = int((falsePositive.sum() - dist[0]) / predictedOverall * 100 + 0.5);
		stats(7,i) = falsePositive.sum() - dist[0] - dist[1];
		stats(8,i) = falseNegative.sum();
		stats(9,i) = int(falseNegative.sum() / gt.sum() * 100 + 0.5);

		
		// Save images
		CImg<float> eval(gt.width(),gt.height(),1,3);
		cimg_forXY(eval,x,y)
		{
			eval(x,y,0) = falseNegative(x,y);
			eval(x,y,1) = positive(x,y);
			eval(x,y,2) = falsePositive(x,y);
		}
		this->predictionEvaluated().push_back( eval.normalize(0,255) );

	}
		
	this->statistic() = stats;

	// Write stats to log

	sout << "--------------------------------------------------------------------" << std::endl;
	sout << "EVALUATION RESULT\nTotal\t\t\tTotDE\t\tDEp\t\tDEfp\t\tDEfpD1\t\tDEfpD2\t\tDEfn" << std::endl;

	for( size_t i = 0; i < size; i++)
	{		 
		sout << stats(0,i) << "\t\t" 
			<< stats(1,i) << "\t\t" 
			<< stats(2,i) << " (" << stats(3,i) << "%)\t\t" 
			<< stats(4,i) << "\t\t" 
			<< stats(5,i) << " (" << stats(6,i) << "%)\t\t" 
			<< stats(7,i) << "\t\t" 
			<< stats(8,i) << " (" << stats(9,i) << "%)\t\t" 
			<< std::endl;
	}
	
	sout << "--------------------------------------------------------------------" << std::endl;
	sout << "Labels:\n\tTotal = Total number of evaluated pixels\n"
		<< "\tTotDE = Depth edges pixels (number of given ground truth pixels)\n"
		<< "\tDEp = Depth edges predicted positive\n"
		<< "\tDEfp = Depth edges predicted false positive\n"
		<< "\tDEfpD1 = False positive pixels with a distance to ground truth taller than 1px\n"
		<< "\tDEfpD2 = False positive pixels with a distance to ground truth taller than 2px\n"
		<< "\tDEfn = Depth edges predicted false negative\n" << std::endl;
	sout << "--------------------------------------------------------------------" << std::endl;
	sout << "Percentages: DEfn(%) and DEp(%) are calculated regarding TotDE" << std::endl;
	sout << "Percentages: DEfpD1(%) is calculated regarding all predicted pixels" << std::endl;
	sout << "--------------------------------------------------------------------" << std::endl;

}

CImg<float> getDistHistogram( const CImg<float> &fp, const CImg<float> &gt ){

	const int w = fp.width();
	const int h = fp.height();
	CImg<float> dist(w,h,1,1);
	dist.fill(0);

	cimg_forXY(fp,x,y)
	{
		// do for every false positive pixel
		if(fp(x,y) > 0)
			// iterate over increasing distances and crop image patches
			// around the false positive pixel. when a ground truth pixel
			// is found inside that patch, the distance is saved
			for( size_t i = 1; i < std::max(w,h); i++ )
			{
				if(gt.get_crop(x-i,y-i,x+i,y+i).sum() > 0){
					dist(x,y) = i;//std::sqrt(2.*float(i)*float(i));
					break;
				}
			}
	}

	CImg<float> hist = dist.get_histogram(std::max(w,h),1,300);
	/*
	sout << "III" << std::endl;
	for(size_t i = 0; i < hist.width(); i++)
		sout << hist[i] << " ";
	sout << std::endl;
	*/
	return hist;
}

#endif /* _DEPTHEDGEEVALUATION_HXX_ */
