/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeFeatures.h
 *  Declaration of the parameter class DepthEdgeFeatures.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.02.2013
 */

#ifndef _DEPTHEDGEFEATURES_H_
#define _DEPTHEDGEFEATURES_H_

#ifdef _MSC_VER
#ifdef depthedgefeatures_EXPORTS
/// Visual C++ specific code
#define depthedgefeatures_DECLDIR __declspec(dllexport)
#else
#define depthedgefeatures_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else
/// Not needed without MSVC
#define depthedgefeatures_DECLDIR
#endif

#include <charon-core/ParameteredObject.h>
#include <charon-utils/CImg.h>

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include <vigra/windows.h>
#include <vigra/gaborfilter.hxx>
#include <vigra/multi_array.hxx>
#include <vigra/multi_fft.hxx>
/*
#include <vigra/windows.h>
#include <vigra/random_forest.hxx>
#include <vigra/multi_array.hxx>
*/
using namespace cimg_library;
//using namespace vigra;
//using namespace rf;


/// Features to learn depth edges from
/** This module provides methods to calculate features saved in a CImgList.
 *  The features then can be used to learn and predict depth edges. The output
 *  slot is a pointer on the module itself. In this way, the features can be
 *  calculated for single images of an image list.
 */
template <typename T>
class depthedgefeatures_DECLDIR DepthEdgeFeatures :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	DepthEdgeFeatures(const std::string& name = "");

	/// images for feature generation
	InputSlot< cimg_library::CImgList<T> > images;
	/// edge images with edges corresponding to the given images
	InputSlot< cimg_library::CImgList<T> > edges;
	/// list of mask images
	InputSlot< cimg_library::CImgList<T> > masks;
	/// list of flow files
	InputSlot< cimg_library::CImgList<T> > flow;
	/// List of segmentation files required for segmentation features
	InputSlot< cimg_library::CImgList<T> > segmentationLabels;

	/// Path to folder
	Parameter< std::string > filename;
	/// Path to feature name file
	Parameter< std::string > filename2;

	/// Ignore pattern to save file
	Parameter < bool > ignorePattern;

	/// If of first image in sequence
	Parameter < int > startFrame;
	
	/// Use original source image for feature calculation
	Parameter < bool > useOriginal;
	/// Use normalized source image for feature calculation
	Parameter < bool > useNormalized;
	/// Use equalized source image for feature calculation
	Parameter < bool > useEqualized;

	/// Sigma for gaussian blur of color image
	Parameter < float > sigma1;
	/// Sigma for gaussian blur of scale levels
	Parameter < float > sigma2;
	
	/// Use RGB as feature
	Parameter < bool > useRGB;
	/// Use gradient as feature
	Parameter < bool > useRGBGradient;
	/// Use histogram of RGB gradient as feature
	Parameter < bool > useRGBGradientHistogram;
	/// Use histogram of RGB gradient as feature
	Parameter < bool > useRGBGradientHistogramXY;
	/// Use RGB Eigen Values of structure tensor
	Parameter < bool > useEigenValuesRGB;
	/// Use Histogram of Eigen Values of structure tensor
	Parameter < bool > useEigenValuesRGBHistogram;
	/// Use RGB Eigen Vectors of structure tensor
	Parameter < bool > useEigenVectorsRGB;
	/// Use RGB ratio between Eigen Values of structure tensor
	Parameter < bool > useEigenValueRatioRGB;
	/// Use RGB angles between edges and Y-axis
	Parameter < bool > useEdgeAnglesRGB;
	/// Use Difference between edge angles of neighbouring pixels
	Parameter < bool > useEdgeAnglesDiffRGB;
	/// Use RGB strcuture tensor for neighbour pixels
	Parameter < bool > useNeighbourEigenPropertiesRGB;
	/// Use neighbour RGB as feature
	Parameter < bool > useNeighbourRGB;

	/// Use gabor filter
	Parameter < bool > useGaborFilter;
	
	/// Use HSV hue values as feature
	Parameter < bool > useHSV_H;
	/// Use HSV saturation values as feature
	Parameter < bool > useHSV_S;
	/// Use HSV brightness values as feature
	Parameter < bool > useHSV_V;
	/// Use histogram of pixel and neighbour pixel hues as feature
	Parameter < bool > useColorHistogram;
	/// Use gradient of hue as feature
	Parameter < bool > useHSVGradientH;
	/// Use gradient of saturation as feature
	Parameter < bool > useHSVGradientS;
	/// Use gradient of brightness as feature
	Parameter < bool > useHSVGradientV;
	/// Use gradient of hue histogram as feature
	Parameter < bool > useHSVGradientHistogramH;
	/// Use gradient of saturation histogram as feature
	Parameter < bool > useHSVGradientHistogramS;
	/// Use gradient of brightness histogram as feature
	Parameter < bool > useHSVGradientHistogramV;
	/// Use HSV Eigen Values of structure tensor
	Parameter < bool > useEigenValuesHSV_H;
	/// Use HSV Eigen Values of structure tensor
	Parameter < bool > useEigenValuesHSV_S;
	/// Use HSV Eigen Values of structure tensor
	Parameter < bool > useEigenValuesHSV_V;
	/// Use Histogram of Eigen Values of hue structure tensor
	Parameter < bool > useEigenValuesHSVHistogram_H;
	/// Use Histogram of Eigen Values of saturation structure tensor
	Parameter < bool > useEigenValuesHSVHistogram_S;
	/// Use Histogram of Eigen Values of brightness structure tensor
	Parameter < bool > useEigenValuesHSVHistogram_V;
	/// Use HSV Eigen Vectors of structure tensor
	Parameter < bool > useEigenVectorsHSV_H;
	/// Use HSV Eigen Vectors of structure tensor
	Parameter < bool > useEigenVectorsHSV_S;
	/// Use HSV Eigen Vectors of structure tensor
	Parameter < bool > useEigenVectorsHSV_V;
	/// Use HSV ratio between Eigen Values of structure tensor
	Parameter < bool > useEigenValueRatioHSV_H;
	/// Use HSV ratio between Eigen Values of structure tensor
	Parameter < bool > useEigenValueRatioHSV_S;
	/// Use HSV ratio between Eigen Values of structure tensor
	Parameter < bool > useEigenValueRatioHSV_V;
	/// Use HSV angles between edges and Y-axis
	Parameter < bool > useEdgeAnglesHSV_H;
	/// Use HSV angles between edges and Y-axis
	Parameter < bool > useEdgeAnglesHSV_S;
	/// Use HSV angles between edges and Y-axis
	Parameter < bool > useEdgeAnglesHSV_V;
	/// Use Difference between edge angles of neighbouring pixels
	Parameter < bool > useEdgeAnglesDiffHSV_H;
	/// Use Difference between edge angles of neighbouring pixels
	Parameter < bool > useEdgeAnglesDiffHSV_S;
	/// Use Difference between edge angles of neighbouring pixels
	Parameter < bool > useEdgeAnglesDiffHSV_V;
	/// Use HSV strcuture tensor for neighbour pixels
	Parameter < bool > useNeighbourEigenPropertiesHSV_H;
	/// Use HSV strcuture tensor for neighbour pixels
	Parameter < bool > useNeighbourEigenPropertiesHSV_S;
	/// Use HSV strcuture tensor for neighbour pixels
	Parameter < bool > useNeighbourEigenPropertiesHSV_V;
	/// Use neighbour HSV as feature
	Parameter < bool > useNeighbourHSV_H;
	/// Use neighbour HSV as feature
	Parameter < bool > useNeighbourHSV_S;
	/// Use neighbour HSV as feature
	Parameter < bool > useNeighbourHSV_V;
	
	/// Use hessian as feature
	Parameter < bool > useHessian;
	/// Use hessian as feature
	Parameter < bool > useDiffusion;
	/// Use bilateral filter as feature
	Parameter < bool > useBilateral;



	/// Use pixel positions as feature
	Parameter < bool > usePositions;

	/// Use optical flow as feature (Optical flow given)
	Parameter < bool > useOpticalFlow;
	/// Use segmentation as feature
	Parameter < bool > useSegmentation;
	
	/// Use warped mask (or original mask to train) as feature
	Parameter < bool > useWarpedMask;
	/// Use histogram of warped mask (or original mask to train) as feature
	Parameter < bool > useWarpedMaskHistogram;
	/// Use edges as features
	Parameter < bool > useEdges;
	/// Use number of nearest neighbor edges
	Parameter < bool > useEdgeNeighbours;
	/// Use number of pixels of different segments
	Parameter < bool > useEdgeSegments;
	
	/// Number of pixels taken into account for histograms
	Parameter < unsigned int > histogramNearesNeighbours;
	/// Number of eigen value histogram bins hue
	Parameter < unsigned int > eigenvalueHistogramBins;
	/// Number of gradient histogram bins hue
	Parameter < unsigned int > gradientHistogramBinsHSV_H;
	/// Number of gradient histogram bins saturation
	Parameter < unsigned int > gradientHistogramBinsHSV_S;
	/// Number of gradient histogram bins brightness
	Parameter < unsigned int > gradientHistogramBinsHSV_V;
	/// Number of gradient histogram bins RGB
	Parameter < unsigned int > gradientHistogramBinsRGB;
	/// Number of absolute gradient histogram bins RGB
	Parameter < unsigned int > gradientHistogramBinsRGBXY;
	/// Number of color histogram bins
	Parameter < unsigned int > colorHistogramBins;
	/// Number of color histogram bins for hue
	Parameter < unsigned int > colorHistogramBinsHue;
	/// Distance of used neighbour pixels
	Parameter < float > neighbourDistance;
	/// Distance of used neighbour pixels for angle differences
	Parameter < float > neighbourDistanceAngDiff;
	/// Number of nearest neighbours for angle differences
	Parameter < unsigned int > angDiffNearestNeighbours;

	/// variance for gradient calculation
	Parameter < float > sigmaGradient;
	/// variance for structure tensor calculation in edge direction
	Parameter < float > sigmaStructureTensor;
	/// variance for structure tensor calculation perpendicular to edge directon
	Parameter < float > sigmaStructureTensor2;
	/// width of gaussian for structure tensor calculation in edge direction
	Parameter < int > gaussStructureTensor;
	/// print debug info to log
	Parameter < bool > verbose;
	
	/// Number of scale spaces
	Parameter < int > numberOfScaleSpaces;

	/// Save features as PNG for debuging purposes
	Parameter < bool > debugSavePNG;

	/// Calculate features and append them to feature images
	void saveFeatures( const unsigned int index ); //, std::vector<std::string> &names = 0);
	
protected:

	/// Check dimensions of input slots and define parameters
	void init();

	void getEValHistograms( CImgList<float> &hist, const CImg<float> &eVal, const CImg<float> &eRat, 
							const unsigned int bins, const unsigned int size, const unsigned int scaleSpace );

	/// Update object.
	void execute();
	
	/// Save file
	bool writeToFile( CImgList<float> &features, const int index );
	/// Save name file
	bool writeNamesToFile( const std::string &feature_str );

	/// Returns grey scale image
	inline CImg<T> getGreyScale( const CImg<T> &img );
	
	/// Maximum values for eigen values and eigen value ratio of first calculated image
	/** First dimension of vector represents the scale factor */
	std::vector<std::vector<float>> _maxEV_HSV;
	std::vector<std::vector<float>> _maxEVRatio_HSV;
	std::vector<float> _maxEV_RGB;
	std::vector<float> _maxEVRatio_RGB;

	/// Get structure tensor
	int structureTensor( const CImg<float> &dx, const CImg<float> &dy, 
					 const int gwidth, const float sigma1,  const float sigma2, 
					 CImg<float> &st, const CImg<float> &evec, const bool useEVec = false );
	/// Get eigen vectors and values
	int eigenv( const CImg<float> &st, CImg<float>& eval, CImg<float> &evec1, CImg<float> &evec2, CImg<float> &evalRatio );
	
};

#endif // _DEPTHEDGEFEATURES_H_
