/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeFeatures.hxx
 *  Implementation of the parameter class DepthEdgeFeatures.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.02.2013
 */

#ifndef _DEPTHEDGEFEATURES_HXX_
#define _DEPTHEDGEFEATURES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/DepthEdgeFeatures.h>

#include <charon-utils/ImgTool.hxx>
#include <hekate/openCvUtil.hxx>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>

#include <ctime>

template <typename T>
DepthEdgeFeatures<T>::DepthEdgeFeatures(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthEdgeFeatures", name,
			"<h2>Calculate different feature images used for the depth edge learn algorithm</h2><br>"
			"This module provides a function which calculates different feature images of a image sequence. "
			"The function has to be accessed by the given output pointer which refers to the module itself. "
		)
{
	
	ParameteredObject::_addInputSlot(
		images, "images",
		"Path to image sequence for feature generation",
		"CImgList<T>");
	
	ParameteredObject::_addParameter(filename, "filename", "Path and filename pattern",
			"filewrite");
	ParameteredObject::_addParameter(filename2, "featurenames", "Path to feature name file",
			"filewrite");
		
	ParameteredObject::_addParameter< int >(
		startFrame, "startFrame", "Id of the first image in sequence.", 0);
	
	ParameteredObject::_addParameter< bool >(
		ignorePattern, "ignorePattern", "Ignore pattern to save file.", true);
	
	ParameteredObject::_addParameter< bool >(
		useOriginal, "useOriginal", "Use original source image for feature calculation.", true);
	
	ParameteredObject::_addParameter< bool >(
		useNormalized, "useNormalized", "Use normalized source image for feature calculation.", false);
	
	ParameteredObject::_addParameter< bool >(
		useEqualized, "useEqualized", "Use equalized source image for feature calculation.", false);

	ParameteredObject::_addParameter< float >(
		sigma1, "sigma1", "Sigma for gaussian blur of color image.", 2);
	ParameteredObject::_addParameter< float >(
		sigma2, "sigma2", "Sigma for gaussian blur of scale levels	.", 2);

	ParameteredObject::_addParameter< bool >(
		useRGB, "useRGB", "Use RGB values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useRGBGradient, "useRGBGradient", "Use gradient as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useRGBGradientHistogram, "useRGBGradientHistogram", "Use gradient x and y histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useRGBGradientHistogramXY, "useRGBGradientHistogramXY", "Use gradient sqrt(xx+yy) histogram as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesRGB, "useEigenValuesRGB", "Use Eigen Values of structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesRGBHistogram, "useEigenValuesRGBHistogram", "Use Histogram of Eigen Values of structure tensor", false);
	/*
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsRGB, "useEigenVectorsRGB", "Use Eigen Vectors of structure tensor", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioRGB, "useEigenValueRatioRGB", "Use ratio between Eigen Values of structure tensor", true);
	/*
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesRGB, "useEdgeAnglesRGB", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffRGB, "useEdgeAnglesDiffRGB", "Use Difference between edge angles of neighbouring pixels", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesRGB, "useNeighbourEigenPropertiesRGB", 
		"Apply all selected Eigen features to neighbour pixels", true);
	ParameteredObject::_addParameter< bool >(
		useNeighbourRGB, "useNeighbourRGB", "Use neighbour RGB as feature.", true);

	ParameteredObject::_addParameter< bool >(
		useGaborFilter, "useGaborFilter", "Use Gabor Filter as feature.", true);

	ParameteredObject::_addParameter< bool >(
		useHSV_H, "useHSV_H", "Use HSV hue values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSV_S, "useHSV_S", "Use HSV saturation values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSV_V, "useHSV_V", "Use HSV brightness values as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useColorHistogram, "useHSVHistogram", "Use histogram of neighbour hues as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientH, "useHSVGradientH", "Use gradient of hue ( HSV image type ) as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientS, "useHSVGradientS", "Use gradient of saturation ( HSV image type ) as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientV, "useHSVGradientV", "Use gradient of brightness ( HSV image type ) as feature.", true);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientHistogramH, "useHSVGradientHistogramH", "Use gradient of hue histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientHistogramS, "useHSVGradientHistogramS", "Use gradient of saturation histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHSVGradientHistogramV, "useHSVGradientHistogramV", "Use gradient of brightness histogram as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSV_H, "useEigenValuesHSV_H", "Use Eigen Values of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSV_S, "useEigenValuesHSV_S", "Use Eigen Values of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSV_V, "useEigenValuesHSV_V", "Use Eigen Values of brightness structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSVHistogram_H, "useEigenValuesHSVHistogram_H", "Use Histogram of Eigen Values of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSVHistogram_S, "useEigenValuesHSVHistogram_S", "Use Histogram of Eigen Values of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValuesHSVHistogram_V, "useEigenValuesHSVHistogram_V", "Use Histogram of Eigen Values of brightness structure tensor", false);
	/*
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsHSV_H, "useEigenVectorsHSV_H", "Use Eigen Vectors of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsHSV_S, "useEigenVectorsHSV_S", "Use Eigen Vectors of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenVectorsHSV_V, "useEigenVectorsHSV_V", "Use Eigen Vectors of bightness structure tensor", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioHSV_H, "useEigenValueRatioHSV_H", "Use ratio between Eigen Values of hue structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioHSV_S, "useEigenValueRatioHSV_S", "Use ratio between Eigen Values of saturation structure tensor", false);
	ParameteredObject::_addParameter< bool >(
		useEigenValueRatioHSV_V, "useEigenValueRatioHSV_V", "Use ratio between Eigen Values of bightness structure tensor", false);
	/*
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesHSV_H, "useEdgeAnglesHSV_H", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesHSV_S, "useEdgeAnglesHSV_S", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesHSV_V, "useEdgeAnglesHSV_V", "Use angles between edges and Y-axis", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffHSV_H, "useEdgeAnglesDiffHSV_H", "Use Difference between edge angles of neighbouring pixels", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffHSV_S, "useEdgeAnglesDiffHSV_S", "Use Difference between edge angles of neighbouring pixels", false);
	ParameteredObject::_addParameter< bool >(
		useEdgeAnglesDiffHSV_V, "useEdgeAnglesDiffHSV_V", "Use Difference between edge angles of neighbouring pixels", false);
	*/
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesHSV_H, "useNeighbourEigenPropertiesHSV_H", 
		"Apply all selected Eigen features to neighbour pixels", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesHSV_S, "useNeighbourEigenPropertiesHSV_S", 
		"Apply all selected Eigen features to neighbour pixels", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourEigenPropertiesHSV_V, "useNeighbourEigenPropertiesHSV_V", 
		"Apply all selected Eigen features to neighbour pixels", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourHSV_H, "useNeighbourHSV_H", "Use neighbour HSV as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourHSV_S, "useNeighbourHSV_S", "Use neighbour HSV as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useNeighbourHSV_V, "useNeighbourHSV_V", "Use neighbour HSV as feature.", false);
	ParameteredObject::_addParameter< bool >(
		useHessian, "useHessian", "Use hessian as feature", false);
	ParameteredObject::_addParameter< bool >(
		useDiffusion, "useDiffusion", "Use diffusion as feature", false);
	ParameteredObject::_addParameter< bool >(
		useBilateral, "useBilateral", "Use bilateral filter as feature", false);
	
	ParameteredObject::_addParameter< unsigned int >(
		histogramNearesNeighbours, "histogramNearesNeighbours", "Number of pixels in each direction taken into account for histograms. "
		"For example: histogramNearesNeighbours=1 -> histogram patch is 3x3, histogramNearesNeighbours=2 -> patch = 5x5", 2);
	ParameteredObject::_addParameter< unsigned int >(
		eigenvalueHistogramBins, "eigenvalueHistogramBins", "Number of eigen value histogram bins hue", 5);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsHSV_H, "gradientHistogramBinsHSV_H", "Number of gradient of hue histogram bins. "
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsHSV_S, "gradientHistogramBinsHSV_S", "Number of gradient of saturation histogram bins"
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsHSV_V, "gradientHistogramBinsHSV_V", "Number of gradient of brightness histogram bins"
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsRGB, "gradientHistogramBinsRGB", "Number of gradient of RGB histogram bins"
		"Only central bin will be used as feature.", 19);
	ParameteredObject::_addParameter< unsigned int >(
		gradientHistogramBinsRGBXY, "gradientHistogramBinsRGBXY", "Number of absolute gradient of RGB histogram bins", 3);
	ParameteredObject::_addParameter< unsigned int >(
		colorHistogramBinsHue, "colorHistogramBinsHue", "Number of histogram bins for hue", 6);
	ParameteredObject::_addParameter< unsigned int >(
		colorHistogramBins, "colorHistogramBins", "Number of histogram bins", 3);
	ParameteredObject::_addParameter< float >(
		neighbourDistance, "neighbourDistance", "Distance of used neighbour pixels", 3.0);
	ParameteredObject::_addParameter< float >(
		neighbourDistanceAngDiff, "neighbourDistanceAngDiff", "Distance of used neighbour pixels for angle differences", 0.5);
	ParameteredObject::_addParameter< unsigned int >(
		angDiffNearestNeighbours, "angDiffNearestNeighbours", "Number of nearest neighbours for angle differences", 6);
	
	ParameteredObject::_addParameter< int >(
		numberOfScaleSpaces, "numberOfScaleSpaces", "Number of different scale levels", 1);

	ParameteredObject::_addParameter< bool >(
		debugSavePNG, "debugSavePNG", "Save features as PNG for debuging purposes", false);
	
	ParameteredObject::_addParameter< float >(
		sigmaGradient, "sigmaGradient", "variance for gradient calculation", 1);
	ParameteredObject::_addParameter< float >(
		sigmaStructureTensor, "sigmaStructureTensor", "variance for structure tensor calculation "
		"in edge direction", 1.6);
	ParameteredObject::_addParameter< float >(
		sigmaStructureTensor2, "sigmaStructureTensor2", "variance for structure tensor calculation "
		"perpendicular to edge direction", 0.5);
	ParameteredObject::_addParameter< int >(
		gaussStructureTensor, "gaussStructureTensor", "width of gaussian for structure tensor calculation", 5);
	ParameteredObject::_addParameter< bool >(
		verbose, "verbose", "print debug info to log", false);

	ParameteredObject::_setTags("Hekate;Edges") ;

}

template <typename T>
void DepthEdgeFeatures<T>::execute() {

	// Check dimensions of input slots and define parameters
	init();

	for( size_t i = 0; i < this->images().size(); i++ )
		saveFeatures( i );

}


// Calculate features and append them to feature images
template <typename T>
void DepthEdgeFeatures<T>::saveFeatures( const unsigned int index ){

	// List of feature images
	CImgList<float> features;
	
	const CImg<float> &img = this->images()[index];
	CImg<float> img_normalized, img_equalized;
	if( this->useNormalized() )
		img_normalized = this->images()[index].get_normalize(0,255);
	if( this->useEqualized() )
		img_equalized = this->images()[index].get_equalize(256);

	const unsigned int w = img.width();
	const unsigned int h = img.height();
	const unsigned int d = img.depth();
	
	// Number of different scale levels
	unsigned int levels = std::max( int(1), this->numberOfScaleSpaces() );
	sout << "Count of pyramid levels: " << levels << std::endl;

	// debug log
	std::string feature_str = "";

	// There are different types of images to use for calculation:
	// - original
	// - normalized 
	// - equalized
	// One iteration for each of them

	for( int i_type = 0; i_type < 3; i_type++ ){

		if( i_type == 0 && !this->useOriginal() && !this->useNormalized() && !this->useEqualized()  )
		{}
		else if( i_type == 0 && !this->useOriginal() ){
			continue;
		}
		else if( i_type == 1 && !this->useNormalized() ){
			continue;
		}
		else if( i_type == 2 && !this->useEqualized() ){
			continue;
		}
			

	// If pyramid level > 1 --> iterate over scaled images
	for( size_t i = 0; i < levels; i++ )
	{

		// Scale training image
		float blur = this->sigma1();
		if( i != 0)
			blur = pow( this->sigma2(), float(i) );
		sout << "Calculate features on blurred image with stddev of " << blur << std::endl;
		
		const CImg<float> trainScaled = img.get_blur( blur ); 
		const CImg<float> trainScaledGrey = getGreyScale( trainScaled );
		CImg<float> trainScaledHSV;
		if( img.spectrum() != 3 )
			sout << "No RGB (3) channels given" << std::endl;
		else 
			trainScaledHSV = trainScaled.get_RGBtoHSV();

		cimg_forXY(trainScaledHSV,x,y){
			trainScaledHSV(x,y,0,0) -= 180;
		}

		sout << "trainScaledHSV.get_channel(0).max() = " << trainScaledHSV.get_channel(0).max() << std::endl;


		// Use Gabor Filter
		if (this->useGaborFilter()){

			vigra::MultiArray<2, short> gabor( vigra::Shape2(w,h) );	
			
			cimg_forXY(trainScaledGrey, x, y){
				gabor(x, y) = trainScaledGrey(x, y);
			}


			int directionCount = 10;	// 6

			int scaleCount = 4;		// 4

			double orientation = PI / 2; // radian

			double centerFrequency = 0.375; // 0.375 or smaller / 3.0 / 8.

			/* 
			createGaborFilter(MultiArrayView<2, T, S> dest,
			double orientation, double centerFrequency,
			double angularSigma, double radialSigma);
			*/
			//double radialSigma = vigra::radialGaborSigma( centerFrequency);
			//double angularSigma = vigra::angularGaborSigma(directionCount, centerFrequency);
			//vigra::createGaborFilter(gabor, orientation, centerFrequency, angularSigma, radialSigma);

			//vigra::srcImageRange()

			//CImg<float> dbg(w, h, 1, 1, 0), 
			CImg<float> dbg2(w, h, 1, 1, 0);
			//cimg_forXY(trainScaled, x, y){
			//	dbg(x, y) = gabor(x, y);
			//}
			//dbg.get_normalize(0,255).save("gabor.png");

			//int height = 128;
			//int width = 128;

			//vigra::ImageArray<vigra::FImage> filters(16, image.size());

			// filters
			vigra::GaborFilterFamily<vigra::FImage> 
				myfilterfamily(	w, 
								h,
								directionCount,
								scaleCount,
								centerFrequency
								);

			//{
			//	auto it = myfilterfamily.begin();
			//	auto itEnd = myfilterfamily.end();
			//	int dbgcnt = 0;
			//	for (; it != itEnd; it++){
			//		sout << dbgcnt++ << std::endl;
			//		cimg_forXY(trainScaled, x, y){
			//			dbg(x, y) += (*it)(x, y);
			//			if (x < 50 && y < 50)
			//				sout << (*it)(x, y) << std::endl;
			//		}
			//	}
			//}

			// results array for filter results
			vigra::ImageArray<vigra::FImage >
				gaborresultsf(myfilterfamily.size(), vigra::Diff2D(w, h));
			
			// // get gray value image
			// cimg_library::CImg<short> buffer(trainScaled.width(), trainScaled.height(), 1, 1, 0);
			// buffer = trainScaled.get_RGBtoHSI().get_channel(2);

			// get vigra image
			vigra::BasicImageView<short> cur_frame(gabor.data(), w, h);
			
			// Fimage
			vigra::FImage cur_framef(cur_frame.width(), cur_frame.height());
			
			// image -> fimage
			vigra::copyImage(vigra::srcImageRange(cur_frame),
				vigra::destImage(cur_framef));

			// apply filters
			vigra::applyFourierFilterFamily(vigra::srcImageRange(cur_framef),
				myfilterfamily, gaborresultsf);

			cimg_library::CImgList<short> dbglist;

			// Save to CImg
			auto it = gaborresultsf.begin();
			auto itEnd = gaborresultsf.end();
			int dbgcnt = 0;
			for (; it != itEnd; it++){
				//sout << dbgcnt++ << std::endl;
				dbg2.fill(0);
				cimg_forXY(trainScaled, x, y){
					//dbg(x, y) += std::abs((*it)(x, y));
					dbg2(x, y) += (*it)(x, y);
					//if (x < 50 && y < 50)
					//	sout << (*it)(x, y) << std::endl;
				}
				dbglist.push_back(dbg2.get_normalize(0, 255));
				features.push_back(dbg2.get_channel(0));
				feature_str += "useGabor\n";
			}
			//dbglist.save("gabor/gabor2.png");
			//dbg.get_normalize(0, 255).save("gabor.png");
			//dbg.get_normalize(0, 1).mul(trainScaled).get_normalize(0, 255).save("gaborMul.png");


		}


		// Use Bilateral
		if (this->useBilateral())
		{
			if( img.spectrum() == 3 ){
				sout << "\tCalculate bilateral filter features on HSV" ;
				const clock_t begin_time = clock();

				CImg<float> bilH = trainScaledHSV.get_channel(0).blur_bilateral(blur+0.5, 0.2, -33, 32);
				CImg<float> bilS = trainScaledHSV.get_channel(1).blur_bilateral(blur+0.5, 0.2, -33, 32);
				CImg<float> bilV = trainScaledHSV.get_channel(2).blur_bilateral(blur+0.5, 0.2, -33, 32);
				features.push_back( bilH.get_channel(0) );
				features.push_back( bilS.get_channel(0) );
				features.push_back( bilV.get_channel(0) );
				feature_str += "useBilateralH\nuseBilateralS\nuseBilateralV\n";
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
			else{
				sout << "\tCalculate bilateral filter features on Grey values" ;
				const clock_t begin_time = clock();
				CImg<float> bil = trainScaled.get_channel(0).blur_bilateral(blur+0.5, 0.2, -33, 32);
				features.push_back( bil.get_channel(0) );
				feature_str += "useBilateral\n";
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
		}

		// Use diffusion
		if(this->useDiffusion())
		{
			if( img.spectrum() == 3 ){
				sout << "\tCalculate diffusion features on HSV";
				const clock_t begin_time = clock();

				const float  sharpness = 0.7;
				const float  anisotropy = 0.6;
				const float  alpha = 0.6 * ( 1 + blur );
				const float  sigma = 1.1 * ( 1 + blur );
				CImg<float> diffH = trainScaledHSV.get_channel(0).edge_tensors(sharpness,anisotropy,alpha,sigma);
				CImg<float> diffS = trainScaledHSV.get_channel(1).edge_tensors(sharpness,anisotropy,alpha,sigma);
				CImg<float> diffV = trainScaledHSV.get_channel(2).edge_tensors(sharpness,anisotropy,alpha,sigma);
				CImg<float> diffHa = diffH.get_channel(0).abs() + diffH.get_channel(1).abs() + diffH.get_channel(2).abs() ;
				CImg<float> diffSa = diffS.get_channel(0).abs() + diffS.get_channel(1).abs() + diffS.get_channel(2).abs() ;
				CImg<float> diffVa = diffV.get_channel(0).abs() + diffV.get_channel(1).abs() + diffV.get_channel(2).abs() ;
				features.push_back( diffHa.get_channel(0) );
				features.push_back( diffSa.get_channel(0) );
				features.push_back( diffVa.get_channel(0) );
				feature_str += "useDiffusionH\nuseDiffusionS\nuseDiffusionV\n";			
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
			else{
				sout << "\tCalculate diffusion features on Grey";
				const clock_t begin_time = clock();
				const float  sharpness = 0.7;
				const float  anisotropy = 0.6;
				const float  alpha = 0.6 * ( 1 + blur );
				const float  sigma = 1.1 * ( 1 + blur );
				CImg<float> diff = trainScaled.get_channel(0).edge_tensors(sharpness,anisotropy,alpha,sigma);
				features.push_back( diff.get_channel(0) );		
				feature_str += "useDiffusion\n";
				sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
		}

		
		// RGB image
		if(this->useRGB()){
			sout << "\tCalculate RGB features" ;
			const clock_t begin_time = clock();

			features.push_back( trainScaled.get_channel(0) );
			features.push_back( trainScaled.get_channel(1) );
			features.push_back( trainScaled.get_channel(2) );
			feature_str += "useRGBR\nuseRGBG\nuseRGBB\n";
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}
	
		// HSV image
		if( this->useHSV_H() || this->useHSV_S() ||this->useHSV_V() ){
			sout << "\tCalculate HSV features" ;
			const clock_t begin_time = clock();

			/*
			HSV encoding:
			H = [0,360) / S = [0,1] / V = [0,1]
			*/

			if( this->useHSV_H() )
			{
				features.push_back( trainScaledHSV.get_channel(0)  );
				feature_str += "useHSV_H\n";
			}
			if( this->useHSV_S() ) 
			{
				features.push_back( trainScaledHSV.get_channel(1) );
				feature_str += "useHSV_S\n";
			}
			if( this->useHSV_V() ) 
			{
				features.push_back( trainScaledHSV.get_channel(2) );
				feature_str += "useHSV_V\n";
			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;			
		}

		// Histogram of gradients
		if((this->useRGBGradientHistogram() || this->useRGBGradientHistogramXY()) && i == 0)
		{
			sout << "\tCalculate RGB gradient histogram features" ;
			const clock_t begin_time = clock();

			// calculate number of pixels to crop
			const int size = this->histogramNearesNeighbours();
			cimg_library::CImgList<float> gradient = trainScaledGrey.get_gradient();
			
			// maximum gradient can be -255, minimum 255;
			if( gradient[0].max() > 255. )
				throw( std::runtime_error("useRGBGradientHistogram: maximum gradient > 255 ") );
			if( gradient[0].min() < -255. )
				throw( std::runtime_error("useRGBGradientHistogram: minimum gradient < -255 ") );
			if( gradient[1].max() > 255. )
				throw( std::runtime_error("useRGBGradientHistogram: maximum gradient > 255 ") );
			if( gradient[1].min() < -255. )
				throw( std::runtime_error("useRGBGradientHistogram: minimum gradient < -255 ") );
						
			// histogram of gradients in x and y direction
			if(this->useRGBGradientHistogram())
			{
			
				int bins = this->gradientHistogramBinsRGB();
				// Check if value is evan
				if( bins % 2 == 0) bins++;
				size_t centralBin = int(bins / 2);

				int binWidth = 511 / 2;
				CImg<float> gradXCut = gradient[0].get_min(binWidth).max(-binWidth);
				CImg<float> gradYCut = gradient[1].get_min(binWidth).max(-binWidth);

				// Create histogram image (width,height,numberOfBins);
				cimg_library::CImg<float> histogramX(w,h,3,1);
				cimg_library::CImg<float> histogramY(w,h,3,1);
				
				#pragma omp parallel for
				for( int xy = 0; xy < w*h; xy++ )
				{
					
					const int x = xy % w;
					const int y = xy / w;

					CImg<float> histX = gradXCut.get_crop(x-size,y-size,x+size,y+size).histogram(bins,-255.,255.);
					CImg<float> histY = gradYCut.get_crop(x-size,y-size,x+size,y+size).histogram(bins,-255.,255.);
					// Only save central bin

					histogramX(x,y,0,0) = histX(centralBin,0);
					histogramY(x,y,0,0) = histY(centralBin,0);
					histogramX(x,y,1,0) = histX(centralBin-1,0);
					histogramY(x,y,1,0) = histY(centralBin-1,0);
					histogramX(x,y,2,0) = histX(centralBin+1,0);
					histogramY(x,y,2,0) = histY(centralBin+1,0);
				}
				features.push_back( histogramX );
				features.push_back( histogramY );
				feature_str += "useRGBGradientHistogramX\n";
				feature_str += "useRGBGradientHistogramY\n";
			}

			// historgram of sqrt( gradX + gradY )
			if(this->useRGBGradientHistogramXY())
			{

				unsigned int binsXY = this->gradientHistogramBinsRGBXY();

				cimg_library::CImg<float> gradXY = (gradient[0].get_sqr()+gradient[1].get_sqr()).sqrt();
				// Create histogram image (width,height,numberOfBins);
				cimg_library::CImg<float> histogramXY(w,h,binsXY,1);

				#pragma omp parallel for
				for( int xy = 0; xy < w*h; xy++ )
				{					
					const int x = xy % w;
					const int y = xy / w;

					CImg<float> histXY = gradXY.get_crop(x-size,y-size,x+size,y+size).histogram(binsXY,0,std::sqrt(360.7));
					for(size_t j = 0; j < binsXY; j++)
					{
						histogramXY(x,y,j,0) = histXY(j,0);
					}
				}

				for(size_t j = 0; j < binsXY; j++)
				{
					features.push_back( histogramXY.get_slice(j) );
					feature_str += "useRGBGradientHistogramXY\n";
				}

			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}

		// Histogram of Hue (HSV)
		if(this->useColorHistogram() && i==0)
		{
			if( img.spectrum() == 3 ){
			sout << "\tCalculate HSV histogram features" ;
			const clock_t begin_time = clock();
			
			const unsigned int binsHue = this->colorHistogramBinsHue();
			const unsigned int bins = this->colorHistogramBins();
			// calculate number of pixels to crop
			const unsigned int size = this->histogramNearesNeighbours();
			// Get hue channel of HSV image
			cimg_library::CImg<float> hue = trainScaledHSV.get_channel(0);
			cimg_library::CImg<float> sat = trainScaledHSV.get_channel(1);
			cimg_library::CImg<float> val = trainScaledHSV.get_channel(2);
			// Create histogram image (width,height,numberOfBins);
			cimg_library::CImg<float> histogram(w,h,bins,3);

			#pragma omp parallel for
			for( int xy = 0; xy < w*h; xy++ )
			{					
				const unsigned int x = xy % w;
				const unsigned int y = xy / w;

				CImg<float> hist = hue.get_crop(x-size,y-size,x+size,y+size).histogram(binsHue,0,360);
				CImg<float> hist2 = sat.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,1);
				CImg<float> hist3 = val.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,1);
				for(size_t j = 0; j < bins; j++)
				{
					histogram(x,y,j,0) = hist(j,0);
					histogram(x,y,j,1) = hist2(j,0);
					histogram(x,y,j,2) = hist3(j,0);
				}

			}
			for(size_t j = 0; j < bins; j++)
			{
				features.push_back( histogram.get_slice(j).channel(0) );
				features.push_back( histogram.get_slice(j).channel(1) );
				features.push_back( histogram.get_slice(j).channel(2) );
				feature_str += "useColorHistogram\n";
				feature_str += "useColorHistogram\n";
				feature_str += "useColorHistogram\n";
			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
			}
		}

		// Gradient images
		if(this->useRGBGradient()){
			sout << "\tCalculate RGB gradient features" ;
			const clock_t begin_time = clock();

			// Gradient of train image
			cimg_library::CImgList<float> gradient = trainScaledGrey.get_gradient();
			features.push_back( gradient[0] );
			features.push_back( gradient[1] );
			feature_str += "useRGBGradient\n";
			feature_str += "useRGBGradient\n";
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}

		// Gradient of HSV values
		if( this->useHSVGradientH() || this->useHSVGradientS() || this->useHSVGradientV()
				|| this->useHSVGradientHistogramH() || this->useHSVGradientHistogramS() || this->useHSVGradientHistogramV() )
		{
			sout << "\tCalculate HSV gradient features" ;
			const clock_t begin_time = clock();

			// Calculate HSV gradients
			CImgList<float> gradients = getHSVGradients( trainScaledHSV );

			// Save gradients
			if( this->useHSVGradientH() )
			{
				features.push_back( gradients[0].get_channel(0) );
				features.push_back( gradients[1].get_channel(0) );
				feature_str += "useHSVGradientH\n";
				feature_str += "useHSVGradientH\n";
			}
			if( this->useHSVGradientS() )
			{
				features.push_back( gradients[0].get_channel(1) );
				features.push_back( gradients[1].get_channel(1) );
				feature_str += "useHSVGradientS\n";
				feature_str += "useHSVGradientS\n";
			}
			if( this->useHSVGradientV() )
			{
				features.push_back( gradients[0].get_channel(2) );
				features.push_back( gradients[1].get_channel(2) );
				feature_str += "useHSVGradientV\n";
				feature_str += "useHSVGradientV\n";
			}
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;

			// Calculate gradient histograms
			if( this->useHSVGradientHistogramH() || this->useHSVGradientHistogramS() || this->useHSVGradientHistogramV() )
			{
				sout << "\tCalculate HSV gradient histogram features" ;
				const clock_t begin_time2 = clock();

				// calculate number of pixels to crop
				const unsigned int size = this->histogramNearesNeighbours();

				// Iterate over H, S and V
				for( size_t k = 0; k < 3; k++ )
				{
					
					if( !this->useHSVGradientHistogramH() && k == 0)
						continue;
					if( !this->useHSVGradientHistogramS() && k == 1)
						continue;
					if( !this->useHSVGradientHistogramV() && k == 2)
						continue;

					// get min max values of pixels for histogram
					float min_value, max_value;
					if( this->useHSVGradientHistogramH() && k == 0)
					{
						min_value = -180.;
						max_value = 180.;
					}
					if( this->useHSVGradientHistogramS() && k == 1)
					{
						min_value = -1.;
						max_value = 1.;
					}
					if( this->useHSVGradientHistogramV() && k == 2)
					{
						min_value = -1.;
						max_value = 1.;
					}

					unsigned int bins = this->gradientHistogramBinsHSV_H();
					if(k==1) bins = this->gradientHistogramBinsHSV_S();
					if(k==2) bins = this->gradientHistogramBinsHSV_V();
					
					// Check if value is evan
					if( bins % 2 == 0) bins++;
					size_t centralBin = int(bins / 2);

					// Create histogram image (width,height,numberOfBins);
					cimg_library::CImg<float> histogramX(w,h,3,1);
					cimg_library::CImg<float> histogramY(w,h,3,1);

					int binWidth = std::abs(max_value-min_value) / 2;
					CImg<float> gradXCut = gradients[0].get_min(binWidth).max(-binWidth);
					CImg<float> gradYCut = gradients[1].get_min(binWidth).max(-binWidth);
									
					#pragma omp parallel for
					for( int xy = 0; xy < w*h; xy++ )
					{					
						const unsigned int x = xy % w;
						const unsigned int y = xy / w;

						const CImg<float> histX = gradXCut.get_crop(x-size,y-size,x+size,y+size).channel(k).histogram(bins,min_value,max_value);
						const CImg<float> histY = gradYCut.get_crop(x-size,y-size,x+size,y+size).channel(k).histogram(bins,min_value,max_value);
						// save central 3 bins
						histogramX(x,y,0,0) = histX(centralBin,0);
						histogramY(x,y,0,0) = histY(centralBin,0);
						histogramX(x,y,1,0) = histX(centralBin-1,0);
						histogramY(x,y,1,0) = histY(centralBin-1,0);
						histogramX(x,y,2,0) = histX(centralBin+1,0);
						histogramY(x,y,2,0) = histY(centralBin+1,0);
					}
					
					// if scale space 0, save two more bins
					if(i == 0){
						features.push_back( histogramX.get_slice(0) );
						features.push_back( histogramY.get_slice(0) );
						feature_str += "useHSVGradient\n";
						feature_str += "useHSVGradient\n";
					}
					
				}
				sout << " (" << float( clock () - begin_time2 ) / CLOCKS_PER_SEC << ")" << std::endl;
	
			}
			
		}

		// Hessian images
		if(this->useHessian()){
			sout << "\tCalculate hessian features" ;
			const clock_t begin_time = clock();

			// Gradient of train image
			cimg_library::CImgList<float> hessian = trainScaledGrey.get_hessian();
			features.push_back( hessian[0].abs() + hessian[1].abs() + hessian[2].abs() );
			feature_str += "useHessian\n";
			
			sout << " (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << ")" << std::endl;
		}

		sout << "\tChecking Eigen Features" << std::endl;
		// two iterations ( HSV and RGB )
		for( size_t l=0; l<2; l++)
		{

			// Flags which activate structure tensor calculation
			bool uEVal=false, uEVec=false, uERat=false, uEAng=false, uEAngDiff=false, uENei=false, uENeiProp=false, uEValHist=false;
			
			bool uEVal_H = this->useEigenValuesHSV_H();
			bool uEVal_S = this->useEigenValuesHSV_S();
			bool uEVal_V = this->useEigenValuesHSV_V();
			bool uEVec_H = this->useEigenVectorsHSV_H();
			bool uEVec_S = this->useEigenVectorsHSV_S();
			bool uEVec_V = this->useEigenVectorsHSV_V();
			bool uERat_H = this->useEigenValueRatioHSV_H();
			bool uERat_S = this->useEigenValueRatioHSV_S();
			bool uERat_V = this->useEigenValueRatioHSV_V();
			bool uEAng_H = this->useEdgeAnglesHSV_H();
			bool uEAng_S = this->useEdgeAnglesHSV_S();
			bool uEAng_V = this->useEdgeAnglesHSV_V();
			bool uEAngDiff_H = this->useEdgeAnglesDiffHSV_H();
			bool uEAngDiff_S = this->useEdgeAnglesDiffHSV_S();
			bool uEAngDiff_V = this->useEdgeAnglesDiffHSV_V();
			bool uENei_H = this->useNeighbourHSV_H();
			bool uENei_S = this->useNeighbourHSV_S();
			bool uENei_V = this->useNeighbourHSV_V();
			bool uENeiProp_H = this->useNeighbourEigenPropertiesHSV_H();
			bool uENeiProp_S = this->useNeighbourEigenPropertiesHSV_S();
			bool uENeiProp_V = this->useNeighbourEigenPropertiesHSV_V();
			bool uEValHist_H = this->useEigenValuesHSVHistogram_H();
			bool uEValHist_S = this->useEigenValuesHSVHistogram_S();
			bool uEValHist_V = this->useEigenValuesHSVHistogram_V();

			// Set RGB flags
			if( l == 0 )
			{
				// RGB Structure Tensor features
				uEVal = this->useEigenValuesRGB();
				uEVec = this->useEigenVectorsRGB();
				uERat = this->useEigenValueRatioRGB();
				uEAng = this->useEdgeAnglesRGB();
				uEAngDiff = this->useEdgeAnglesDiffRGB();
				uENei = this->useNeighbourRGB();
				uENeiProp = this->useNeighbourEigenPropertiesRGB();
				uEValHist = this->useEigenValuesRGBHistogram();
			}
			// Set HSV flags
			else
			{
				// use eigen values
				if(uEVal_H || uEVal_S || uEVal_V)
					uEVal = true;
				else 
					uEVal = false;

				// use eigen vectors
				if(uEVec_H || uEVec_S || uEVec_V)
					uEVec = true;
				else 
					uEVec = false;

				// use eigen ratio
				if(uERat_H || uERat_S || uERat_V)
					uERat = true;
				else 
					uERat = false;

				// use eigen angles
				if(uEAng_H || uEAng_S || uEAng_V)
					uEAng = true;
				else 

				// use eigen angles differences
				if(uEAngDiff_H || uEAngDiff_S || uEAngDiff_V)
					uEAngDiff = true;
				else 
					uEAngDiff = false;

				// use HSV neighbors
				if(uENei_H || uENei_S || uENei_V)
					uENei = true;
				else 
					uENei = false;

				// use HSV eigen neighbours
				if(uENeiProp_H || uENeiProp_S || uENeiProp_V)
					uENeiProp = true;
				else 
					uENeiProp = false;

				// use HSV eigenvalue histograms
				if(uEValHist_H || uEValHist_S || uEValHist_V)
					uEValHist = true;
				else
					uEValHist = false;


			}

			if( uEVal || uEVec || uERat || uEAng || uEAngDiff || uENei || uENeiProp || uEValHist ){
				sout << "\tCalculate Eigen features" ;
				const clock_t begin_time = clock();

				if( l == 0 ) sout << " for RGB values" << std::endl;
				else sout << " for HSV values" << std::endl;
			
				const float dist = this->neighbourDistance() ;
				CImg<float> colorImage;	// color image (RBG/HSV)
				CImg<float> st;			// structure tensor (RBG/HSV)
			
				// get eigen entities from structure tensor
				CImg<float> eVal, angle, eVec1, eVec2, eAng, evSum, dx, dy, dxy1, dxy2;
				//CImg<float> eRat;
				CImg<float> eVal2, eVal3, eVal4, eVal5, eVal6;
				CImgList<float> eValHist;
				sout << "\t\tCalculate eigen values, vectors, ratios and angles" << std::endl;

				CImg<float> stOld, eValOld;
				// calculate RGB / HSV structure tensor
				if( l == 0 )
				{
					stOld = trainScaledGrey.get_structure_tensors(0);
					colorImage = trainScaled;
				}
				else
				{
					// the structure tensor of an HSV image always has three channels
					// flags for H, S, V are checked afterwards for the push_back function
					colorImage = trainScaled.get_RGBtoHSV();
					stOld = getHSVStructureTensor( colorImage );
				}
				getEigenEntities(stOld, eValOld);
							
				// calculate RGB / HSV structure tensor
				if( l == 0 )
				{
					colorImage = trainScaled;
				}
				else
				{
					// the structure tensor of an HSV image always has three channels
					// flags for H, S, V are checked afterwards for the push_back function
					colorImage = trainScaled.get_RGBtoHSV();
				}

				// Get SCHAEFER Eigen Entities
				int gwidth = this->gaussStructureTensor();
				if( gwidth % 2 ) gwidth ++; // check if gwidth value is odd
				getEigenEntitiesSchaefer(colorImage, st, eVal, angle, eVec1, eVec2, dx, dy, dxy1, dxy2, 
					this->gaussStructureTensor(), this->sigmaGradient(), this->sigmaStructureTensor(), 
					this->sigmaStructureTensor2(), l==1, eVal2, eVal3, eVal4, eVal5, eVal6);

				//ImgTool::printInfo(sout,st,"st");
				//ImgTool::printInfo(sout,eVal,"eVal");
				//ImgTool::printInfo(sout,eVec1,"eVec1");
				//ImgTool::printInfo(sout,dx,"dx");
				//ImgTool::printInfo(sout,dy,"dy");

				// Get histograms
				//if( uEValHist ){
				//	sout << "\t\tCalculate histograms for eigen values etc." << std::endl;
				//	getEValHistograms( eValHist, eVal, eRat, this->eigenvalueHistogramBins(), this->histogramNearesNeighbours(), i );
				//}

				// Get difference between edge angles of neighbour pixels
				if( uEVal ){
					sout << "\t\tGet sum of EV neighbourhood" << std::endl;
					evSum = getNeighbourEV( st, eVal, eVec2, angle, this->angDiffNearestNeighbours(), this->neighbourDistanceAngDiff() );
						//sout<< evSum.spectrum() << std::endl;
				}
				
				// get eigen entities of neighbour pixels from structure tensor
				CImgList<float> eValN, eVec1N, eVec2N, eRatN, eAngN;
				CImg<float> colorN;
				if(uENeiProp){
					sout << "\t\tCalculate neighbour properties" << std::endl;
					getEigenEntitiesN(colorImage, st, dist, eVec1, eVec2, eValN, eVec1N, eVec2N, eRatN, eAngN);
				}
				
				// Save calculated features for every channel separately
				// ! Different channel results were saved in z-axis !

				sout << "\t\tSave feature images" << std::endl;
				for( unsigned int k = 0; k < st.spectrum(); k++ )
				{
					std::string type;
					if( l== 0 ){ 
						if( k==0 ) type = "r-";
						else if( k==1 ) type = "g-";
						else if( k==2 ) type = "b-";
					}
					else { 
						if( k==0 ) type = "h-";
						else if( k==1 ) type = "s-";
						else if( k==2 ) type = "v-";
					}

					if( i == 0 )
						type += "blur0";
					else if( i == 1 )
						type += "blur1";
					else if( i == 2 )
						type += "blur2";
					else if( i == 3 )
						type += "blur3";
			
					//if(i==0)
					//{
					//	if( (l==0 && uEValHist) || (l==1 && k==0 && uEValHist_H) || (l==1 && k==1 && uEValHist_S) || (l==1 && k==2 && uEValHist_V) ){
					//		for( size_t bins = 0; bins < eValHist.size(); bins++ )
					//		{
					//			features.push_back( eValHist[bins].get_channel(k) );
					//		}
					//	}							
					//}
					if( (l==0 && uEVal) || (l==1 && k==0 && uEVal_H) || (l==1 && k==1 && uEVal_S) || (l==1 && k==2 && uEVal_V) ){
						sout << "Save eigen values features" << std::endl;
						
						if(!eValOld.is_empty()){
							features.push_back( eValOld.get_channel(k).slice(0) );
							feature_str += type + "eValOld\n";
						}
						else
							sout << "!! cimg eval is empty !! -- !! -- !! -- !! " << std::endl;
						features.push_back( eVal.get_channel(k).slice(0).sqr().sqr() );
						features.push_back( eVal.get_channel(k).slice(0) );
						features.push_back( eVal.get_channel(k).slice(1) );
						features.push_back( eVal2.get_channel(k).slice(0) );
						features.push_back( eVal2.get_channel(k).slice(1) );
						features.push_back( eVal3.get_channel(k).slice(0) );
						features.push_back( eVal3.get_channel(k).slice(1) );
						features.push_back( eVal4.get_channel(k).slice(0) );
						features.push_back( eVal4.get_channel(k).slice(1) );
						features.push_back( eVal5.get_channel(k).slice(0) );
						features.push_back( eVal5.get_channel(k).slice(1) );
						features.push_back( eVal6.get_channel(k).slice(0) );
						features.push_back( eVal6.get_channel(k).slice(1) );
						features.push_back( evSum.get_channel(k).slice(0) );
						features.push_back( evSum.get_channel(k).slice(1) );
						features.push_back( angle.get_channel(k).slice(0) );
						
						feature_str += type + "eValMax-sqrt4-1\n";
						feature_str += type + "eValMax-1\n";
						feature_str += type + "eValMax-2\n";
						feature_str += type + "eVal-normal-1\n";
						feature_str += type + "eVal-normal-2\n";
						feature_str += type + "eVal-xdirection-1\n";
						feature_str += type + "eVal-xdirection-2\n";
						feature_str += type + "eVal-ydirection-1\n";
						feature_str += type + "eVal-ydirection-2\n";
						feature_str += type + "eVal-xy1-direction-1\n";
						feature_str += type + "eVal-xy1-direction-2\n";
						feature_str += type + "eVal-xy2-direction-1\n";
						feature_str += type + "eVal-xy2-direction-2\n";
						feature_str += type + "evSum\n";
						feature_str += type + "evAngleDiff\n";
						feature_str += type + "evAngle\n";



						//features.push_back( eVal.get_slice(1).channel(k) );
						//features.push_back( dx.get_channel(k) );
						//features.push_back( dy.get_channel(k) );
						//features.push_back( dxy1.get_channel(k) );
						//features.push_back( dxy2.get_channel(k) );
						//features.push_back( (dx.get_abs()+dy.get_abs()+dxy1.get_abs()+dxy2.get_abs()).get_channel(k).sqrt() );						
					}
					//if( (l==0 && uERat) || (l==1 && k==0 && uERat_H) || (l==1 && k==1 && uERat_S) || (l==1 && k==2 && uERat_V) ){
					//	sout << "Save eigen ratio features" << std::endl;
					//	features.push_back( eRat.get_channel(0).slice(k) );
					//	feature_str += type + "eRat\n";
					//}
					// RGB eigen entities of neigbour pixels
					if( (l==0 && uENeiProp) || (l==1 && k==0 && uENeiProp_H) || (l==1 && k==1 && uENeiProp_S) || (l==1 && k==2 && uENeiProp_V) )
					{
						for( size_t j=0; j < 4; j++)
						{
							sout << "Save eigen values neighbours and eRatN features" << std::endl;
							features.push_back( eValN[i].get_channel(0).slice(k) );
							features.push_back( eRatN[j].get_channel(0).slice(k) );
							feature_str += type + "eValN\n";
							feature_str += type + "eRatN\n";
						}
					}
					// RGB/HSV histogramms of neigbour pixels to putative edge
					if( (l==0 && uENei) || (l==1 && k==0 && uENei_H) || (l==1 && k==1 && uENei_S) || (l==1 && k==2 && uENei_V) )
					{
						sout << "Save color difference features" << std::endl;
						getColourDifference(colorImage, dist, eVec1, eVec2, colorN, k);
						sout << "colorN.depth = " <<	colorN.depth() << std::endl;
						features.push_back( colorN.get_slice(0) );
						features.push_back( colorN.get_slice(1) );
						features.push_back( colorN.get_slice(2) );
						feature_str += type + "color-patch-left\n";
						feature_str += type + "color-patch-right\n";
						feature_str += type + "color-variance\n";
					}


				}
					
				sout << "\tTook " << float( clock () - begin_time ) / CLOCKS_PER_SEC << "seconds" << std::endl;
			
			}

		}

	}

	}

	if(verbose())
		sout << feature_str;

	// SAVE FILE
	writeToFile( features, index );
	writeNamesToFile( feature_str );

}

/// Save name file
template <typename T>
bool DepthEdgeFeatures<T>::writeNamesToFile( const std::string &feature_str ){

	// Get working directory (should be the one containing the *.wrp file)
	std::string workingPath = FileTool::getCurrentDir();

	setlocale(LC_NUMERIC, "C");
	
	// Create binary file containing ids, names of randomforest hdf5 files and its coordinates
	std::string fname( this->filename2() ) ;
	if(fname == "")
	{	return 0;	}	
	
	std::ofstream indexFile;
	indexFile.open (fname, std::ios::trunc | std::ios::out );
	
	if( !indexFile.is_open()) {
		throw std::runtime_error(std::string("Could not open file ") + fname);
	}
	
	// Write meta data
	indexFile << feature_str;
	
	indexFile.close();
	
	// Reset working directory to that one containing the *.wrp file
	FileTool::changeDir( workingPath );

}

/// Save file
template <typename T>
bool DepthEdgeFeatures<T>::writeToFile( CImgList<float> &features, const int index){
	
	std::string workingPath = FileTool::getCurrentDir();
	
	std::string base = this->filename();
	//if(base == "") throw(std::runtime_error("FileWriterSequence<T>::execute(): No filename given"));
	
	std::string ext;
	try{ext =  this->filename().substr( this->filename().find_last_of('.'),  this->filename().length() );}
	catch(...){sout << "Given file path has to have an file extension" << std::endl;}

	std::string name =  this->filename().substr( 0, this->filename().find_last_of('.') );
	
	if( this->images().size() > 1 ) 
	{
		int id  = index ;
		if( !this->ignorePattern() )
			id += this->startFrame();

		std::stringstream s;
		s << std::setw(6) << std::setfill('0') << id;
		name = name + "_" + s.str();
	}
	
	sout << "	" << name << ext << std::endl;
	
	if( !this->debugSavePNG() )
		features.save( ( name + ".cimg" ).c_str() );
	else{
		for( int i = 0; i < features.size(); i++ )
		{
			std::stringstream s2;
			s2 << std::setw(6) << std::setfill('0') << i;
			std::string name2 = name + "_" + s2.str();
			features[i].get_normalize(0,255).save( ( name2 + ".png" ).c_str() );
		}
	}
	
	FileTool::changeDir( workingPath );

	return 1;
}


/// Get sum of neighbour EV in edge direction
CImg<float> getNeighbourEV( const CImg<float> &st, const CImg<float> &eVal, const CImg<float> &eVec1, 
						const CImg<float> &angle, const int non, const float nndist )
{

	// 101 158
	
	const int w = st.width();
	const int h = st.height();
	const int d = st.spectrum();

	CImg<float> sumEV(w, h, 3, d, 0);

	// Temporary neighbour buffers
	CImg<float> mat2, val2;
	CImg<float> vec3;
	mat2.assign(2,2);

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{			
		CImg<float> vec2(2,2);
		float newX;
		float newY;

		const int x = xy % w;
		const int y = xy / w;

		for( int c = 0; c < d; c++ )
		{
			
			float ang = angle(x,y,0,c);
			float val = eVal(x,y,0,c);
			
			sumEV(x,y,0,c) = val;
			sumEV(x,y,1,c) = ang ;


			// DEBUG
			if( x == 34 && y == 24 ){
				sumEV(int(x),int(y),2,c) = 1;
			}
			if( x == 250 && y == 71 ){
				sumEV(int(x),int(y),2,c) = 1;
			}
			if( x == 114 && y == 46 ){
				sumEV(int(x),int(y),2,c) = 1;
			}

			// build differences in both directions of eigen vector
			for( int s = -1; s <= 1; s = s+2 )
			{

				newX = x;
				newY = y;
				vec2(0,0) = eVec1( x,y,0,c );
				vec2(1,0) = eVec1( x,y,1,c );
							
				for( int i = 0; i < non; i++ )
				{
					// Fill structure tensor at neighbour position (x,y)
					newX += s*nndist*vec2(0,0) ;
					newY += s*nndist*vec2(1,0) ;

					vec2(0,0) =  eVec1.linear_atXY(newX,newY,0,c );
					vec2(1,0) =  eVec1.linear_atXY(newX,newY,1,c );

					float length = std::sqrt(vec2(0,0)*vec2(0,0)+vec2(1,0)*vec2(1,0));
					vec2(0,0) /=  length;
					vec2(1,0) /=  length;
			
					// Check if new point in inside of image border
					if( newX > w-1 || newX < 0 || newY > h-1 || newY < 0 )
						break;

					//// Get next point on edge
					//mat2.fill(
					//		st.linear_atXY(newX,newY,0,c),st.linear_atXY(newX,newY,2,c),
					//		st.linear_atXY(newX,newY,2,c),st.linear_atXY(newX,newY,1,c)
					//);
					//
					//// calculate eigenvectors and eigenvalues
					//mat2.symmetric_eigen(val2,vec3);
					
					//vec2(0,0) =  vec3(0,0);
					//vec2(1,0) =  vec3(1,0);
					//vec2(0,1) =  vec3(0,1);
					//vec2(1,1) =  vec3(1,1);
			
					// Add difference
					sumEV(x,y,0,c) += eVal.linear_atXY(newX,newY,0,c);
					sumEV(x,y,1,c) += angle.linear_atXY(newX,newY,0,c ) - ang;
					//sumEV(x,y,0,c) += std::abs( eVal.linear_atXY( newX, newY, 0, c ) );
					//sumEV(x,y,1,c) += acos( eVal.linear_atXY( newX, newY, 0, c ) );			
					
					// DEBUG
					if( x == 34 && y == 24 ){
						sumEV(int(newX),int(newY),2,c) = 1;
						//sout << "new = " << newX << " / " << newY << std::endl;
					//sout << std::sqrt(vec2(0,0)*vec2(0,0)+vec2(1,0)*vec2(1,0)) << std::endl;
					}
					if( x == 250 && y == 71 ){
						sumEV(int(newX),int(newY),2,c) = 0.5;
					//sout << std::sqrt(vec2(0,0)*vec2(0,0)+vec2(1,0)*vec2(1,0)) << std::endl;
					}
					if( x == 114 && y == 46 ){
						sumEV(int(newX),int(newY),2,c) = 0.5;
					//sout << std::sqrt(vec2(0,0)*vec2(0,0)+vec2(1,0)*vec2(1,0)) << std::endl;
					}
					
				}
			
			}

			//sumEV(x,y,0,c) /= (2 * non + 1 );

		}

	}

	sout << "done" << std::endl;

	return sumEV;


}


// Returns grey scale image
template <typename T>
CImg<T> DepthEdgeFeatures<T>::getGreyScale( const CImg<T> &img )
{



	// Create grey scale image
	CImg<T> out( img.width(), img.height(), 1, 1 );
	out.fill( T(0) );

	if( img.spectrum() == 0 )
		throw( std::range_error("img.spectrum() == 0") );

	// sum up image channels
	for( unsigned int i = 0; i < img.spectrum(); i++)
	{
		out += img.get_channel(i) ;
	}

	// divide by number of channels
	out /= img.spectrum() ;
	
	return out;

}


/// Returns gradient for hue of HSV values
/** 
Returns a CImgList with two images. First image contains the x deviations, second the
the y deviations. The channels of the images contain the gradients of H, S, and V.
*/
template <typename T>
CImgList<float> getHSVGradients( const CImg<T> &img )
{

	if( img.depth() != 1)
		throw( std::range_error("getHSVGradient: image depth != 1") );

	cimg_library::CImgList<float> grad;
	CImg<float> gradX(img.width(), img.height(), 1, 3 );
	CImg<float> gradY(img.width(), img.height(), 1, 3 );

	// Create convolve mask
	CImg<float> cX(3,1), cY(1,3);
	cX[0] = -0.5;	cX[1] = 0;		cX[2] = 0.5;
	cY[0] = -0.5;	cY[1] = 0;		cY[2] = 0.5;
	
	gradX = img.get_convolve( cX );
	gradY = img.get_convolve( cY );

	// Correct hue values due to overlap
	const int w = img.width();
	const int h = img.height();

	// Iterate over first channel, to correct Hue gradients
	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{					
		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		// Correct gradX (overlap between [0;369) )
		const float xl = gradX(x,y,0,0);
		const float xlSign = ( xl < 0 ? -1 : 1 );
		gradX(x,y,0,0) = xlSign * std::min( std::abs(xl), float(360.) - std::abs(xl) );
		// Correct gradY (overlap between [0;369) )
		const float yl = gradY(x,y,0,0);
		const float ylSign = ( yl < 0 ? -1 : 1 );
		gradY(x,y,0,0) = ylSign * std::min( std::abs(yl), float(360.) - std::abs(yl) );
	}
	
	grad.push_back( gradX );
	grad.push_back( gradY );

	return grad;

}


/// Returns structure tensor for HSV image type
/** 
Returns the structure tensors for an HSV image type. Attention: The three matrix entries 
are save in channel-axis while if the structure tensor is calculated on an
image with several channels, the different channel values are saved in z-axis.
z = ( H, S, V )
c = ( Ix*Ix, Ix*Iy, Iy*Iy )
*/
template <typename T>
CImg<float> getHSVStructureTensor( const CImg<T> &img )
{

	const CImgList<float> gradients = getHSVGradients( img );
	CImg<float> st(img.width(), img.height(), 3, 3);

	const unsigned int w = img.width();
	const unsigned int h = img.height();

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{				
		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		for( int c = 0; c < 3; c++ )
		{
			st(x,y,c,0) = gradients[0](x,y,0,c) * gradients[0](x,y,0,c);
			st(x,y,c,1) = gradients[0](x,y,0,c) * gradients[1](x,y,0,c);
			st(x,y,c,2) = gradients[1](x,y,0,c) * gradients[1](x,y,0,c);
		}

	}
	
	return st;

}


/// Calculate eigen value histograms
template <typename T>
void DepthEdgeFeatures<T>::getEValHistograms( CImgList<float> &hist, const CImg<float> &eVal, const CImg<float> &eRat, 
											 const unsigned int bins, const unsigned int size, const unsigned int scaleSpace )
{
		
	const unsigned int w = eVal.width();
	const unsigned int h = eVal.height();
	const unsigned int d = eVal.spectrum();

	CImgList<float> list( bins, w, h, 2, d, 0 );

	// Iterate over number of given eigen values
	for( size_t k = 0; k < d; k++ )
	{

		CImg<float> ev = eVal.get_channel(k);
		CImg<float> rat = eRat.get_channel(k);

		// maximum value of eigen values in the current image.
		// Actually, the maximum of all images had to be taken
		float maxEV, maxRatio;

		if( d > 1 )
		{
			if(_maxEV_HSV[scaleSpace][k] == 0) _maxEV_HSV[scaleSpace][k] = ev.max();
			if(_maxEVRatio_HSV[scaleSpace][k] == 0) _maxEVRatio_HSV[scaleSpace][k] = rat.max();
			maxEV = _maxEV_HSV[scaleSpace][k];
			maxRatio = _maxEVRatio_HSV[scaleSpace][k];
		}
		else
		{
			if(_maxEV_RGB[scaleSpace] == 0) _maxEV_RGB[scaleSpace] = ev.max();
			if(_maxEVRatio_RGB[scaleSpace] == 0) _maxEVRatio_RGB[scaleSpace] = rat.max();
			maxEV = _maxEV_RGB[scaleSpace];
			maxRatio = _maxEVRatio_RGB[scaleSpace];
		}
		
		ev.cut(0,maxEV);
		rat.cut(0,maxRatio);

		#pragma omp parallel for
		for( int xy = 0; xy < ev.width() * ev.height(); xy++ )
		{					
			const unsigned int x = xy % w;
			const unsigned int y = xy / w;

			const CImg<float> hist1 = ev.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,maxEV);
			const CImg<float> hist2 = rat.get_crop(x-size,y-size,x+size,y+size).histogram(bins,0,maxRatio);
			for(size_t i = 0; i < bins; i++)
			{
				list[i](x,y,0,k) = hist1(i,0);
				list[i](x,y,1,k) = hist2(i,0);
			}

		}

	}

	// Only take the first two bins
	//hist = list;
	
	//hist.push_back(list[0]);
	//hist.push_back(list[1]);
	for(size_t i = 0; i < bins; i++)
	{
		hist.push_back(list[i]);
	}

}

/// Calculates histograms for neighbour pixels
/** 
	Uses first eigenvectors which point perpendicular to edge direction
*/
void getColourDifference(const CImg<float> &img, const float dist, const CImg<float> &eVec1, 
						 const CImg<float> &eVec2, CImg<float> &colorN, const int channel){

	sout << "\t\t\t\tCalculate neighbour colour differences for channel " << channel << std::endl;
				
	// dimensions of structure tensor
	const int w = eVec1.width();
	const int h = eVec1.height();
	const int d = eVec1.spectrum();

	if( channel > d )
		throw( std::range_error("channel > st.depth()") );
		
	//	sout << "min = " << img.get_channel(channel).min() << std::endl;
	//	sout << "max = " << img.get_channel(channel).max() << std::endl;

	// One colour feature which represents the difference between colors to the left and the right of a putative edge
	colorN.assign(w,h,3,1,0);

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{					
		const int x = xy % w;
		const int y = xy / w;

		colorN(x,y,2,0) = img.get_crop(x-3,y-3,0,channel,x+3,y+3,0,channel).variance();

		int z = channel;

		// Get mean neighbourhood colors by eigen vectors	
		try{

			//CImg<float> vec(1,2,1,2);
			//vec(0,0,0,0) = eVec1(x,y,0,channel);
			//vec(0,1,0,0) = eVec1(x,y,1,channel);
			//vec(0,0,0,1) = eVec2(x,y,0,channel);
			//vec(0,1,0,1) = eVec2(x,y,1,channel);

			float sum_patch [2] = {0,0};
			int count [2] = {0,0};

			// 1 or -1
			for( int sign = -1; sign <= 1; sign = sign+2 ){

				// -1 -> 0
				int idd = ( sign == -1 ? 0 : 1 );

				// Get new position
				float newX = float(x) + sign*dist*eVec1(x,y,0,channel) ;
				float newY = float(y) + sign*dist*eVec1(x,y,1,channel) ;

				// Border conditions: if outside, continue
				if( newX < 0 || newY < 0 || newX > w-1 || newY > h-1 )
					continue;

				// Iterate over neighbourhood of newX and newY
				// -1 or 1
				for( int k1 = -1; k1 <= 1; k1++ ){
					// -1 or 1
					for( int k2 = -1; k2 <= 1; k2++ ){

						// translate in eVec1 direction
						float cx1 = newX + k1 * eVec1(x,y,0,channel);
						float cy1 = newY + k1 * eVec1(x,y,1,channel);
						// superposition with eVec2 direction
						cx1 += k2 * eVec2(x,y,0,channel);
						cy1 += k2 * eVec2(x,y,1,channel);	

						// border conditions
						if( cx1 >= 0 && cy1 >= 0 && cx1 <= w-1 && cy1 <= h-1 ){

							float val = img.linear_atXY(cx1,cy1,0,channel);//,std::numeric_limits<float>::quiet_NaN());
							//if( val != std::numeric_limits<float>::quiet_NaN() ){
								sum_patch[idd] += val ;
								count[idd]++;
							//}

						}
					}
				}

				if( count[idd] != 0 )
					sum_patch[idd] /= count[idd];
			}

			colorN(x,y,0,0) = sum_patch[0];
			colorN(x,y,1,0) = sum_patch[1];

		}
		catch(...){
			continue;
		}	

	}

}

/// Calculates eigen entities such as eigen values and vectors from structure tensor of neighbour pixels
/** The neighbour pixels are determined with the given eigen vectors of the source pixel.*/
void getEigenEntitiesN(const CImg<float> &img, const CImg<float> &st, const float dist, const CImg<float> &eVec1, 
			const CImg<float> &eVec2, CImgList<float> &eValN, CImgList<float> &eVec1N, CImgList<float> &eVec2N, 
			CImgList<float> &eRatN, CImgList<float> &eAngN){

	sout << "\t\t\t\tCalculate neighbour values" << std::endl;
				
	// dimensions of structure tensor
	const int w = st.width();
	const int h = st.height();
	const int d = st.spectrum();

	// Resize for 4 nearest neighbours
	for( size_t i=0; i < 4; i++)
	{
		eValN.push_back( CImg<float>(w, h, 2, d).fill(0) );
		eVec1N.push_back( CImg<float>(w, h, 2, d).fill(0) );
		eVec2N.push_back( CImg<float>(w, h, 2, d).fill(0) );
		eRatN.push_back( CImg<float>(w, h, 1, d).fill(0) );
		eAngN.push_back( CImg<float>(w, h, 1, d).fill(0) );
	}

	sout << "\t\t\t\tCalculate neighbour values 1" << std::endl;

	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{					
		const int x = xy % w;
		const int y = xy / w;

		for( int z=0; z < d; z++)
		{

			// Temporary neighbour buffers
			CImg<float> vec, mat2, val2, vec2;
			mat2.assign(2,2);
			val2.assign(1,2);
			vec2.assign(2,2);
			vec.assign(2,2);

			vec(0,0) = eVec1(x,y,0,z);
			vec(0,1) = eVec1(x,y,1,z);
			vec(1,0) = eVec2(x,y,0,z);
			vec(1,0) = eVec2(x,y,1,z);
		
			int k = 0;
			// iterate over four neighbour pixels
			for( int i = -1; i<2; i+=2 )
			{
				for( int j = 0; j<2; j++, k++ )
				{
					try{
						// Fill structure tensor at neighbour position (x,y)
						float newX = float(x) + float(i*dist)*vec(j,0) ;
						float newY = float(y) + float(i*dist)*vec(j,1) ;

						// Border conditions: if outside, continue
						if( newX < 0 || newY < 0 || newX > w-1 || newY > h-1 )
							continue;

						mat2.fill(
								st.linear_atXY(newX,newY,z,0,0),st.linear_atXY(newX,newY,z,1,0),
								st.linear_atXY(newX,newY,z,1,0), st.linear_atXY(newX,newY,z,2,0)
						);
					
						// calculate eigenvectors and eigenvalues
						mat2.symmetric_eigen(val2,vec2);
					
						// fill into output slots
						assert(val2.width()==1);
						assert((val2[1] <= val2[0]));
						assert(vec2.width()==vec2.height());
						assert(vec2==vec2);
						assert(val2[0]==val2[0]);
						if(val2[1]!=val2[1] ||val2[0]!=val2[0])
							continue;
				
						eValN[k](x,y,z,0) = val2[0];
						eValN[k](x,y,z,1) = val2[1];
						eVec1N[k](x,y,z,0) = vec2(0,0);
						eVec1N[k](x,y,z,1) = vec2(0,1);
						eVec2N[k](x,y,z,0) = vec2(1,0);
						eVec2N[k](x,y,z,1) = vec2(1,1);	

						eRatN[k](x,y,z,0) = val2[1]*val2[1]+val2[0]*val2[0] ;

						eAngN[k](x,y,z,0) = acos( vec2(0,1) );
					
						// calculate colors of both sides of the gradient
						// cimg_library::cubic_atXY is not stable against OpenMP
						//int cropx = int(newX), cropy = int(newY);
						//CImg<float> crop = img.get_crop(newX-1,newY-1,newY+1,newY+1);

					}
					catch(...){
						continue;
					}	

				}

			}

		}
				
	}
	
}

/// Calculates eigen entities such as eigen values and vectors from structure tensor
void getEigenEntities(const CImg<float> &st, CImg<float> &eVal){
				
	const int w = st.width();
	const int h = st.height();
	const int d = st.depth();

	// Eigen values separated in channel 0 and 1
	eVal.resize( w, h, 2, d, 0 );	

	sout << "pre calc " << (eVal.is_empty() ? " empty " : " noempty " ) << std::endl;

	// Calculate values for every pixel
	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{

		// Temporary buffers
		CImg<float> mat, val, vec;
		mat.assign(2,2);
		val.assign(1,2);
		vec.assign(2,2);	

		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		for( int z = 0; z < d; z++)
		{

			// Fill structure tensor at pixel position (x,y)
			mat.fill(
					st(x,y,z,0),st(x,y,z,1),
					st(x,y,z,1),st(x,y,z,2)
			);

			// calculate eigenvectors and eigenvalues
			mat.symmetric_eigen(val,vec);

			// fill into output slots
			assert(val.width()==1);
			assert((val[1] <= val[0]));
			assert(vec.width()==vec.height());
			assert(vec==vec);
			assert(val[0]==val[0]);
			assert(val[1]==val[1]);
				
			eVal(x,y,0,z) = val[0];
			eVal(x,y,1,z) = val[1];

		}
				
	}	

	
	sout << "pre calc " << (eVal.is_empty() ? " empty " : " noempty " ) << std::endl;

}


/// Calculates eigen entities such as eigen values and vectors from structure tensor
void getEigenEntities(const CImg<float> &st, CImg<float> &eVal, CImg<float> &eValDiff, CImg<float> &eVec1, 
			CImg<float> &eVec2, CImg<float> &eRat, CImg<float> &eAng){
				
	const int w = st.width();
	const int h = st.height();
	const int d = st.spectrum();

	// Eigen values separated in channel 0 and 1
	eVal.resize( w, h, d, 2 );
	eValDiff.resize( w, h, d, 1 );

	// Eigen vectors. Coordinates separated in channels 0 and 1. (channel0 =x, channel1=y)
	eVec1.resize(w, h, d, 2);
	eVec2.resize(w, h, d, 2);

	// Ratio between eigen values and angles between eigen vectors
	eRat.resize(w, h, d, 1);
	eAng.resize(w, h, d, 1);		

	// Calculate values for every pixel
	#pragma omp parallel for
	for( int xy = 0; xy < w*h; xy++ )
	{

		// Temporary buffers
		CImg<float> mat, val, vec;
		mat.assign(2,2);
		val.assign(1,2);
		vec.assign(2,2);	

		const unsigned int x = xy % w;
		const unsigned int y = xy / w;

		for( int z = 0; z < st.depth(); z++)
		{

			// Fill structure tensor at pixel position (x,y)
			mat.fill(
					st(x,y,z,0),st(x,y,z,1),
					st(x,y,z,1),st(x,y,z,2)
			);

			// calculate eigenvectors and eigenvalues
			mat.symmetric_eigen(val,vec);

			// fill into output slots
			assert(val.width()==1);
			assert((val[1] <= val[0]));
			assert(vec.width()==vec.height());
			assert(vec==vec);
			assert(val[0]==val[0]);
			assert(val[1]==val[1]);
				
			eVal(x,y,z,0) = val[0];
			eVal(x,y,z,1) = val[1];

			if( val[1] != 0 )
				eValDiff(x,y,z,0) = std::abs( val[0] / val[1] );
			else
				eValDiff(x,y,z,0) = FLT_MAX;

			CImg<float> eVec1_temp(1,2,1,1), eVec2_temp(1,2,1,1);

			//eVec1(x,y,z,0) = vec(0,0);
			//eVec1(x,y,z,1) = vec(0,1);
			//eVec2(x,y,z,0) = vec(1,0);
			//eVec2(x,y,z,1) = vec(1,1);


			eVec1_temp(0,0) = vec(0,0);
			eVec1_temp(0,1) = vec(0,1);
			eVec2_temp(0,0) = vec(1,0);
			eVec2_temp(0,1) = vec(1,1);	

			//CImg<float> res = mat * eVec1_temp;
			//sout << res(0,0)*res(0,0) + res(0,1)*res(0,1) << std::endl;
			
			float m1 = eVec1_temp.magnitude();
			if( m1 != 0 )
				eVec1_temp /= m1;
			float m2 = eVec2_temp.magnitude();
			if( m2 != 0 )
				eVec2_temp /= m2;
			
			eVec1(x,y,z,0) = eVec1_temp(0,0);
			eVec1(x,y,z,1) = eVec1_temp(0,1);
			eVec2(x,y,z,0) = eVec2_temp(0,0);
			eVec2(x,y,z,1) = eVec2_temp(0,1);	




			//
			//eVec1 /= eVec1.magnitude();
			//eVec2 /= eVec2.magnitude();

			//CImg<float>  vec22(1,2,1,1);
			//vec22[0] = vec(0,0);
			//vec22[1] = vec(0,1);
			//
			//vec22 /= vec22.magnitude();
			//
			//CImg<float> res = ( mat * vec22 );
			//ImgTool::printInfo(sout,res,"res");
			//sout << ( val[0] * eVec1 )[0] << " / " << ( val[0] * eVec1 )[1] << std::endl;
			
			//sout << eVec1.magnitude() << std::endl;
			//sout << vec22.magnitude() << std::endl;
			//sout << std::sqrt( vec22.dot( vec22 ) ) << std::endl;
			//sout << std::sqrt( eVec1.dot( eVec2 ) ) << std::endl;
			//sout << "--" << std::endl;
			
			// ratio between different eigen values
			eRat(x,y,z,0) = val[1]*val[1]+val[0]*val[0] ;

			// calculate angles
			eAng(x,y,z,0) = acos( vec(0,1) );

		}
				
	}				

}


// Check dimensions of input slots and define parameters
template <typename T>
void DepthEdgeFeatures<T>::init(){
	
	if( this->filename() == "" )
		throw(std::runtime_error("DepthEdgeFeaturesWrite: empty path"));
	
	_maxEV_HSV.resize(this->numberOfScaleSpaces());
	_maxEVRatio_HSV.resize(this->numberOfScaleSpaces());
	_maxEV_RGB.resize(this->numberOfScaleSpaces());
	_maxEVRatio_RGB.resize(this->numberOfScaleSpaces());
	for( size_t i = 0; i < this->numberOfScaleSpaces(); i++ )
	{
		_maxEV_HSV[i].push_back(0);
		_maxEV_HSV[i].push_back(0);
		_maxEV_HSV[i].push_back(0);
		_maxEVRatio_HSV[i].push_back(0);
		_maxEVRatio_HSV[i].push_back(0);
		_maxEVRatio_HSV[i].push_back(0);
		_maxEV_RGB[i] = 0;
		_maxEVRatio_RGB[i] = 0;
	}
	
}


/// Returns an image containing the mean HSV colors of every pixel
CImg<float> getSegmentMeanColor(const CImg<float> &labels, const CImg<float> &img){
	
	CImg<float> labelSum(labels.max()+1,1,1,img.spectrum());
	std::vector<float> labelCount(labels.max()+1,0);
	cimg_forXYC(img,x,y,c)
	{
		labelSum(labels(x,y),0,0,c) += img(x,y,0,c);
		labelCount[labels(x,y)] ++;
	}
	cimg_forXYC(labelSum,x,y,c)
	{
		if(labelCount[x] != 0)
			labelSum(x,0,0,c) /= labelCount[x];
	}

	CImg<float> meanColor(img.width(), img.height(), 1, img.spectrum() );
	cimg_forXYC(meanColor,x,y,c)
	{
		meanColor(x,y,0,c) = labelSum( labels(x,y), 0, 0, c );
	}

	if(img.spectrum() == 3)
		meanColor.RGBtoHSV();
	return meanColor;
}


/// Returns the mean color of pixel neighbours in and against the given vector direction. Saved in z-channel
CImg<float> getNeighbourMeanColor( const CImg<float> &meanColor, const CImg<float> &eVec2 ){

	CImg<float> neighbourColor(meanColor.width(),meanColor.height(),2,meanColor.spectrum());
	cimg_forXYC(meanColor,x,y,c)
	{
		float color1 = meanColor.linear_atXYZC(float(x)+1.41*eVec2(x,y,0,0),float(y)+1.41*eVec2(x,y,1,0),0,c);
		float color2 = meanColor.linear_atXYZC(float(x)-1.41*eVec2(x,y,0,0),float(y)-1.41*eVec2(x,y,1,0),0,c);
		neighbourColor(x,y,0,c) = color1;
		neighbourColor(x,y,1,c) = color2;
	}
	return neighbourColor;

}


/// calculate derivatives, if hue is set, the gradient of the hue channel is calculated
/** 
	dx and dy are the derivatives in x and y direction while dxy1 is the derivative in lower right direction
	and dy2 is the derivative in upper right direction
*/
void get_gradients( const CImg<float> &image, 
			   const unsigned gwidth, const float sigma, 
			   CImg<float> &dx, CImg<float> &dy, CImg<float> &dxy1, CImg<float> &dxy2, const bool hue = false )
{ 
		
	const CImg<float> color(1,1,1,1,1);
	CImg<float> gaussX,gaussXY1,gaussXY2;		// gaus for smooting in X and XY direction
	CImg<float> dgaussX,dgaussXY1,dgaussXY2;	// gauss for derivatives in X and XY direction

	/*
		gaussXY1  --> smooths in upper right direction 
		gaussXY2  --> smooths in lower right direction
		dgaussXY1 --> derivative in lower right direction
		dgaussXY2 --> derivative in upper right direction
	*/

	float temp = 0;
		
	const int offset = gwidth / 2. - 0.5;

	// create smoothing and derivative kernels in x direction
	gaussX.assign(gwidth,1,1,1,0);
	gaussX.draw_gaussian(offset,0,sigma,&color(0,0,0));
	gaussX /= gaussX.sum();

	// Get diagonal gaussian by covariance matrix
	float s1 = sigma;
	float s2 = 0.1; // variance in gradient direction
	float c = std::cos( PI / 4 ) * std::cos( PI / 4 );
	CImg<float> tensor1(2,2,1,1,0);
	tensor1(0,0) = s1*c + s2*c;
	tensor1(1,0) = s1*c-s2*c;
	tensor1(0,1) = s1*c-s2*c;
	tensor1(1,1) = s1*c + s2*c;

	gaussXY1.assign(gwidth,gwidth,1,1,0);
	gaussXY2.assign(gwidth,gwidth,1,1,0);

	gaussXY1.draw_gaussian(offset,offset,tensor1,&color(0,0,0));
	gaussXY2.draw_gaussian(offset,offset,tensor1.get_mirror('y'),&color(0,0,0));
	gaussXY1 /= gaussXY1.sum();
	gaussXY2 /= gaussXY2.sum();
				
	// Create derivative gaussian in x direction
	dgaussX = gaussX;
	cimg_forX(dgaussX, x){
		if( x - gwidth / 2. + 0.5 ) // do for all but the center pixel
			dgaussX(x) /= -( x - gwidth / 2. + 0.5 );
		else
			dgaussX(x) = 0; // set center pixel to 0
		temp += dgaussX(x)*(-x);
	}
	dgaussX /= temp;

	// Copy values of dgaussX to dgaussXY
	dgaussXY1.assign(gwidth,gwidth,1,1,0);
	dgaussXY2.assign(gwidth,gwidth,1,1,0);
	cimg_forX(dgaussXY1,x){
		const int y_xy2 = gwidth - x - 1 ; // y coordinate for dgaussXY2
		dgaussXY1(x,x) = dgaussX(x,0);
		dgaussXY2(x,y_xy2) = dgaussX(x,0);
	}
		
	dx.assign(image); 
	dy.assign(image); 
	dxy1.assign(image); 
	dxy2.assign(image); 

	dx.convolve(gaussX.get_permute_axes("yxzc"));
	dx.convolve(dgaussX);

	dy.convolve(gaussX);
	dy.convolve(-dgaussX.mirror('y').get_permute_axes("yxzc"));

	dxy1.convolve(gaussXY1);
	dxy1.convolve(dgaussXY1);

	dxy2.convolve(gaussXY2);
	dxy2.convolve(dgaussXY2);

	// gradients for [0;x) channels
	if(hue){

		sout << "Checking hue gradient for correct differences" << std::endl;

		sout << "dx.max() = " << dx.max() 
		 << "\ndy.max() = " << dy.max() 
		 << "\ndxy1.max() = " << dxy1.max() 
		 << "\ndxy2.max() = " << dxy2.max() << std::endl;

		int count = 0;
			
		#pragma omp parallel for
		for( int xy = 0; xy < dx.width()*dx.height(); xy++ ){
	
			const int x = xy % dx.width();
			const int y = xy / dx.width();

			if( std::abs(dx(x,y)) > 180  ){
				int val = std::abs(dx(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				dx(x,y) = sgn * diff;
				count++;
			}
		
			if( dy(x,y) > 180 || dy(x,y) < -180 ){
				int val = std::abs(dy(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				count++;
				dy(x,y) = sgn * diff;
			}
		
			if( dxy1(x,y) > 180 || dxy1(x,y) < -180 ){
				int val = std::abs(dxy1(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				count++;
				dxy1(x,y) = sgn * diff;
			}
		
			if( dxy2(x,y) > 180 || dxy2(x,y) < -180 ){
				int val = std::abs(dxy2(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				count++;
				dxy2(x,y) = sgn * diff;
			}

		}

		sout << "count = " << count << std::endl;

	}

}


/// Calculate structure tensor using an optional vector field for gauss orientation
/** 
	If weakenSecondAxis is activated, the sigma of gaussian for second axis 'dy' is reduced.
	In this case, edges in dx direction are privileged.
	If assumeXYAxis is activated, dx and dy are assumed to be dxy1 and dxy2 which means 
	a gradient from upper left to lower right and one from lower left to upper right
*/
int structureTensor( const CImg<float> &dx, const CImg<float> &dy, 
					 const int gwidth, const float sigma1,  const float sigma2, 
					 CImg<float> &st, const float angle = 0, const std::string debug = "0" )
{
	CImg<float> color(1,1,1,1,1);

	const float c = std::cos( angle ), s = std::sin( angle );
	const float cc = c*c, ss = s*s, cs=c*s;
	const float s1 = sigma1, s2 = sigma2;
	CImg<float> tensor1(2,2,1,1,0);
	tensor1(0,0) = s1*cc + s2*ss;
	tensor1(1,0) = s1*cs-s2*cs;
	tensor1(0,1) = s1*cs-s2*cs;
	tensor1(1,1) = s1*ss + s2*cc;
		
	const int offset = gwidth / 2. - 0.5;
	CImg<float> gauss(gwidth,gwidth,1,1,0);
	gauss.draw_gaussian(offset,offset,tensor1,&color(0,0,0));
	gauss /= gauss.sum();

	/*
	if( debug != "0" ){
		sout << "ST_" << debug << std::endl;
		gauss.get_normalize(0,255).save(("gauss_"+debug+".png").c_str());
		ImgTool::printInfo( sout, tensor1, "tensor");
	}
	*/

	// define structure tensor
	st.clear();
	st.assign(dx.get_pow(2).convolve(gauss));
	st.append(
		dy.get_pow(2).convolve(gauss),'z');
	st.append(
		dx.get_mul(dy).convolve(gauss),'z');

	return 0;
}

/// Get eigen vectors for structure tensor.
/** Also calculates a ratio between eigen values and considers an optional angle to rotate eigen vectors.
	This is useful if the structure tensor bases on
	diagonal gradients. After this rotation, eigen vectors can be compared to those
	which are based on horizontal and vertical gradients
*/
int eigenv( const CImg<float> &st, CImg<float>& eval, CImg<float> &evec1, CImg<float> &evec2, CImg<float> &evalRatio, const float angle = 0, const std::string debug = "" )
{

	// calculate eigen values and eigenvectors
	eval.assign(st.width(),st.height(),2,1,0);
	evec1.assign(st.width(),st.height(),2,1,0);
	evec2.assign(st.width(),st.height(),2,1,0);
	evalRatio.assign(st.width(),st.height(),2,1,0);
	
	#pragma omp parallel for
	for( int xy = 0; xy < st.width()*st.height(); xy++ ){
	
		const int x = xy % st.width();
		const int y = xy / st.width();
	
		CImg<float> stemp(2,2);
		CImg<float> eigval_t;
		CImg<float> eigvec_t;

		stemp(0,0)=st(x,y,0);
		stemp(1,0)=st(x,y,2);
		stemp(0,1)=st(x,y,2);
		stemp(1,1)=st(x,y,1);
		stemp.symmetric_eigen(eigval_t,eigvec_t);

		eval(x,y,0)=eigval_t(0,0);
		eval(x,y,1)=eigval_t(0,1);

		assert( eigval_t(0,0) >= eigval_t(0,1) );

		//evec1(x,y,0)=   eigvec_t(0,0);
		//evec1(x,y,1)=  -eigvec_t(1,0);
		//evec2(x,y,0)=   eigvec_t(0,1);
		//evec2(x,y,1)=  -eigvec_t(1,1);

		CImg<float> ev1(1,2,1,1), ev2(1,2,1,1);
		ev1(0,0) = eigvec_t(0,0);
		ev1(0,1) = eigvec_t(1,0);
		ev2(0,0) = eigvec_t(0,1);
		ev2(0,1) = eigvec_t(1,1);

		// Rotate eigen vector by angle, this is useful if the structure tensor bases on
		// diagonal gradients. After this rotation, eigen vectors can be compared to those
		// which are based on horizontal and vertical gradients
		// Check if eigen vector != 1 since this implies there is no structure at this pixel.
		if( ev1(0,0) != 1 ){
			CImg<float> rotate(2,2,1,1);
			rotate(0,0) = (cos(angle));
			rotate(1,0) = (-sin(angle));
			rotate(0,1) = (sin(angle));
			rotate(1,1) = (cos(angle));
			ev1 = rotate * ev1;
			ev2 = rotate * ev2;
		}

		// Save eigen vectors
		evec1(x,y,0) = ev1(0,0);
		evec1(x,y,1) = ev1(0,1);
		evec2(x,y,0) = ev2(0,0);
		evec2(x,y,1) = ev2(0,1);

		float length1 = std::sqrt( evec1(x,y,0) * evec1(x,y,0) + evec1(x,y,1) * evec1(x,y,1) );
		float length2 = std::sqrt( evec2(x,y,0) * evec2(x,y,0) + evec2(x,y,1) * evec2(x,y,1) );
		if( length1 < 0.99999 && length1 > 1.00001 )
			sout << std::sqrt( evec1(x,y,0) * evec1(x,y,0) + evec1(x,y,1) * evec1(x,y,1) ) << std::endl;
		if( length2 < 0.99999 && length2 > 1.00001 )
			sout << std::sqrt( evec2(x,y,0) * evec2(x,y,0) + evec2(x,y,1) * evec2(x,y,1) ) << std::endl;

		//if( ( eval(x,y,0) + eval(x,y,1) ) != 0 )
		//	evalRatio(x,y) = std::pow( float( eval(x,y,0) - eval(x,y,1))  / float( eval(x,y,0) + eval(x,y,1) ) , float(2.0) );
		//evalRatio(x,y) = acos( evec1(x,y,0) );

		// Get angle and normalize it to positive angle
		// (only the absolute angle is of importance)		
		float vx = evec1(x,y,0), vy = evec1(x,y,1);

		float ang = acos( vx ) / PI * 180;
		if( vy <= 0 ) ang = abs(ang);
		else ang = - abs(ang);
		if( ang < 0 ) ang = 180 + ang;

		float ang2 = asin( vy ) / PI * 180;
		if( vx < 0 ) ang2 = -ang2;
		
		evalRatio(x,y,0) = ang;
		evalRatio(x,y,1) = ang2;

		//// DEBUG: Draw line in eigen vector 1 direction (perpendicular to edge)
		//if(x==23 & y==22){
		//	int ex = x+100*vx, ey = y+100*vy;
		//	sout << ex << "/" << ey << std::endl;
		//
		//	sout << "Debug angle: "<< vx << "," << vy << ": " << ang << ": " << ang2 << std::endl;
		//	
		//	CImg<float> color(1,1,1,1,1);
		//	CImg<float> deb(st.width(),st.height(),1,1,0);
		//	deb.draw_line(x,y,ex,ey,&color(0,0,0));
		//	deb.normalize(0,255).save(("debug"+debug+".png").c_str());
		//}

	}

	return 0;
}


/// Get eigen values and vectors for each channel using Schaefer Implementation
/** gaussWidth for gradients and structure tensor. Sigma 1 for gradients, sigma 2 for structure tensor */
void getEigenEntitiesSchaefer(const CImg<float> &img, CImg<float> &st, CImg<float> &eVal, CImg<float> &eValRatio, CImg<float> &eVec1, 
			CImg<float> &eVec2, CImg<float> &dx, CImg<float> &dy, CImg<float> &dxy1, CImg<float> &dxy2, const int gaussWidth, 
			const float sigma1, const float sigma_st1, const float sigma_st2, const bool hsv, 
			CImg<float> &eVal2, CImg<float> &eVal3, CImg<float> &eVal4, CImg<float> &eVal5, CImg<float> &eVal6){
				
	const int w = img.width();
	const int h = img.height();
	const int d = img.spectrum();

	sout << "Calculate eigen values and vectors for " << d << (hsv ? " hsv" : " rgb" ) << " channels" << std::endl;
	
	st.assign(w,h,3,d,1);
	eVal.assign(w,h,2,d,1);		// save first and second component in z-axis
	eVec1.assign(w,h,2,d,1);	// save first and second component in z-axis
	eVec2.assign(w,h,2,d,1);	// save first and second component in z-axis
	dx.assign(w,h,1,d,1);
	dy.assign(w,h,1,d,1);
	dxy1.assign(w,h,1,d,1);
	dxy2.assign(w,h,1,d,1);
	eValRatio.assign(w,h,2,d,1);

	eVal2.assign(w,h,2,d,1);		// save first and second component in z-axis
	eVal3.assign(w,h,2,d,1);		// save first and second component in z-axis
	eVal4.assign(w,h,2,d,1);		// save first and second component in z-axis
	eVal5.assign(w,h,2,d,1);		// save first and second component in z-axis
	eVal6.assign(w,h,2,d,1);		// save first and second component in z-axis

	// iterate over colour channels
	for( int i = 0; i < d; i++ ){

		sout << "Channel " << (hsv?"hsv":"rgb") << i << " ----------------------------" << std::endl;
		
		std::stringstream string;
		if(hsv) string << "hsv_";
		else string << "rgb_";
		string << "channel_" << i << "_";
		// img.get_channel(i).get_normalize(0,255).save((string.str()+"img.png").c_str());

		// Eigenimplementation
		CImg<float> _dx(w,h,1,1), _dy(w,h,1,1), _dxy1(w,h,1,1), _dxy2(w,h,1,1), _st(w,h), _stX(w,h), _stY(w,h), _stXY1(w,h), _stXY2(w,h);
		CImg<float> _eval(w,h,2,d,0), _evec1(w,h,2,d,0), _evec2(w,h,2,d,0), _evalRatio(w,h,2,d,0);
		
		// save first and second component in z-axis
		CImg<float> _evalX(w,h,2,d,0), _evec1X(w,h,2,d,0), _evec2X(w,h,2,d,0), _evalRatioX(w,h,2,d,0); 
		CImg<float> _evalY(w,h,2,d,0), _evec1Y(w,h,2,d,0), _evec2Y(w,h,2,d,0), _evalRatioY(w,h,2,d,0);
		CImg<float> _evalXY1(w,h,2,d,0), _evec1XY1(w,h,2,d,0), _evec2XY1(w,h,2,d,0), _evalRatioXY1(w,h,2,d,0);
		CImg<float> _evalXY2(w,h,2,d,0), _evec1XY2(w,h,2,d,0), _evec2XY2(w,h,2,d,0), _evalRatioXY2(w,h,2,d,0);

		// get gradients of channel i
		bool _h = ( i == 0 && hsv ? true : false );
		get_gradients(img.get_channel(i), gaussWidth, sigma1, _dx, _dy, _dxy1, _dxy2, _h);
		
		// get structure tensor
		structureTensor( _dx, _dy, gaussWidth, sigma_st1, sigma_st1, _st, 0, "1");	// symmetric gaussian
		// get eigen values / vectors
		eigenv( _st, _eval, _evec1, _evec2, _evalRatio, 0, (string.str()+"0").c_str());
		// DEBUG: save image
		// _eval.get_normalize(0,255).save((string.str()+"eval.png").c_str());
			
		structureTensor( _dx, _dy, gaussWidth, sigma_st1, sigma_st2, _stX, 0);	// gaussian with smaller y smoothing
		structureTensor( _dx, _dy, gaussWidth, sigma_st2, sigma_st1, _stY, 0);	// gaussian with smaller x smoothing
		structureTensor( _dxy1, _dxy2, gaussWidth, sigma_st1, sigma_st2, _stXY1, PI / 4.);	// gaussian  45 rotated
		structureTensor( _dxy1, _dxy2, gaussWidth, sigma_st2, sigma_st1, _stXY2, PI / 4.);	// gaussian -45 rotated

		eigenv( _stX, _evalX, _evec1X, _evec2X, _evalRatioX, 0, (string.str()+"1").c_str());
		eigenv( _stY, _evalY, _evec1Y, _evec2Y, _evalRatioY, 0, (string.str()+"2").c_str());
		eigenv( _stXY1, _evalXY1, _evec1XY1, _evec2XY1, _evalRatioXY1, PI / 4., (string.str()+"3").c_str());
		eigenv( _stXY2, _evalXY2, _evec1XY2, _evec2XY2, _evalRatioXY2, PI / 4., (string.str()+"4").c_str());
		
		//  _evalX.get_normalize(0,255).save((string.str()+"evalX.png").c_str());
		//  _evalY.get_normalize(0,255).save((string.str()+"evalY.png").c_str());
		//  _evalXY1.get_normalize(0,255).save((string.str()+"evalXY1.png").c_str());
		//  _evalXY2.get_normalize(0,255).save((string.str()+"evalXY2.png").c_str());
		//  
		//  
		//  // Build sums of evalX and evalY / evalXY1 and evalXY2
		//  // This is build sums of eigen value images which base on the same base vectors (horizontal/vertical or diagonal)
		//  (_evalX+_evalY).normalize(0,255).save((string.str()+"evalXandY.png").c_str());
		//  (_evalXY1+_evalXY2).normalize(0,255).save((string.str()+"evalXY1and2.png").c_str());
		//  
		//  // Build sum of all eigen value images		
		//  (_evalXY1+_evalXY2+_evalX+_evalY).normalize(0,255).save((string.str()+"evalALL.png").c_str());
		//  
		//  // Take maximum from all available eigen value images
		CImg<float> _evalMAX = _eval.get_max(_evalX).max(_evalY).max(_evalXY1).max(_evalXY2);
		//  _evalMAX.get_normalize(0,255).save((string.str()+"evalMAX.png").c_str());

		// Get mean angle
		CImg<float> _angle = ( _evalRatioX + _evalRatioY + _evalRatioXY1 + _evalRatioXY2 + _evalRatio ) / 5.;

		CImg<float> angleHSV = _angle.get_normalize(0,359);
		angleHSV.resize(w,h,1,3).HSVtoRGB();
		angleHSV.save((string.str()+"angle.png").c_str());

		// Get mean vectors
		CImg<float> _eVecMean1, _eVecMean2;
		_eVecMean1 = ( _evec1X + _evec1Y + _evec1XY1 + _evec1XY2 + _evec1 ) / 5.;
		_eVecMean2 = ( _evec2X + _evec2Y + _evec2XY1 + _evec2XY2 + _evec2 ) / 5.;

		// save images
		#pragma omp parallel for
		for( int xy = 0; xy < st.width()*st.height(); xy++ ){
	
			const int x = xy % st.width();
			const int y = xy / st.width();

			st(x,y,0,i) = _st(x,y,0);
			st(x,y,1,i) = _st(x,y,1);
			st(x,y,2,i) = _st(x,y,2);

			dx(x,y,0,i) = _dx(x,y);
			dy(x,y,0,i) = _dy(x,y);

			dxy1(x,y,0,i) = _dxy1(x,y);
			dxy2(x,y,0,i) = _dxy2(x,y);
			
			eValRatio(x,y,0,i) = _angle(x,y,0);
			eValRatio(x,y,1,i) = _angle(x,y,1);

			eVal(x,y,0,i) = _evalMAX(x,y,0);
			eVal(x,y,1,i) = _evalMAX(x,y,1);

			eVal2(x,y,0,i) = _eval(x,y,0);
			eVal2(x,y,1,i) = _eval(x,y,1);

			eVal3(x,y,0,i) = _evalX(x,y,0);
			eVal3(x,y,1,i) = _evalX(x,y,1);

			eVal4(x,y,0,i) = _evalY(x,y,0);
			eVal4(x,y,1,i) = _evalY(x,y,1);

			eVal5(x,y,0,i) = _evalXY1(x,y,0);
			eVal5(x,y,1,i) = _evalXY1(x,y,1);

			eVal6(x,y,0,i) = _evalXY2(x,y,0);
			eVal6(x,y,1,i) = _evalXY2(x,y,1);

			eVec1(x,y,0,i) = _eVecMean1(x,y,0);
			eVec1(x,y,1,i) = _eVecMean1(x,y,1);
			eVec2(x,y,0,i) = _eVecMean2(x,y,0);
			eVec2(x,y,1,i) = _eVecMean2(x,y,1);
		}

	}

}


#endif /* _DEPTHEDGEFEATURES_HXX_ */
