/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeLearn.h
 *  Declaration of the parameter class DepthEdgeLearn.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.08.2013
 */

#ifndef _DEPTHEDGELEARN_H_
#define _DEPTHEDGELEARN_H_

#ifdef _MSC_VER
#ifdef depthedgelearn_EXPORTS
/// Visual C++ specific code
#define depthedgelearn_DECLDIR __declspec(dllexport)
#else
#define depthedgelearn_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else
/// Not needed without MSVC
#define depthedgelearn_DECLDIR
#endif

#include <charon-core/ParameteredObject.h>
#include <charon-utils/CImg.h>

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include <vigra/windows.h>
#include <vigra/random_forest.hxx>

#include <hekate/DepthEdgeFeatures.h>

#include <charon-utils/ImgTool.hxx>
#include <cmath>

#include <hekate/DepthEdgeModel.hxx>

using namespace cv ;
using namespace vigra;
using namespace rf;
using namespace cimg_library;

typedef float feature_t;
typedef int label_t;

/// Learn masked pixels
/** This module learns one ore more random forests using masked pixels. A
 *  frame can be divided in several patches where every patch is assigned to
 *  one random forest. So the models can be specified on specific image areas.
 *  In this case, the according prediction should use optical flow to
 *  determine the new patch positions in subsequent frames. 
 *  The module also can rely on nearest neighbour pixels of the masks only. So the
 *  model is more specific for the edges of interest. The model also learns pixels
 *  of neighbour frames and uses already learned models of neighbour frames if available.
 */
template <typename T>
class depthedgelearn_DECLDIR DepthEdgeLearn :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	DepthEdgeLearn(const std::string& name = "");	

	/// The depth edge model which was already learned.
	/** This model can be empty, if learning module runs for the first time*/
	InputSlot< hekate::DepthEdgeModel > depthEdgeModelIn;
	/// Pointer to the feature module.
	InputSlot< std::vector<std::string> > features;
	/// Images with highlighted edges (for example CannyEdges)
	InputSlot< cimg_library::CImgList<T> > imgEdge;
	/// Images with annotated masks
	InputSlot< cimg_library::CImgList<T> > imgMask;
	/// A vector with ids according to which images of the imgMask list
	/// should be used for the learn process
	InputSlot< std::vector<int> > idMask;
	/// Optical flow images
	InputSlot< cimg_library::CImgList<T> > imgMaskWarped;

	/// The depth edge model which keeps random forests, its patch coordinates
	///	and its ids and the feature setup which was chosen at learning runtime
	OutputSlot< hekate::DepthEdgeModel > depthEdgeModelOut;

	/// A vector of random forests
	OutputSlot< CImgList<T> > sample;
	
	/// Relearn random forests which already exists
	Parameter< bool > relearn;
	/// Use open mp to parallize patch calculation
	Parameter< bool > useOpenMP;

	/// Number of random forest trees
	Parameter< int > treeCount;
	/// Number of samples per tree
	Parameter< int > samplesPerTree;
	/// Split node every minSplitNodeSize nodes
	Parameter< int > minSplitNodeSize;
	/// Take equal numbers of samples for different labels
	Parameter< bool > equalizeSampleCount;
	/// Learn edge labels beside depth edges and non edges
	Parameter< bool > learnSeparateEdgeLabel;
	
	/// Learn all mask pixels without edge intersection
	Parameter< bool > learnAllPixels;
	/// Learn all mask pixels without edge intersection
	Parameter< bool > learnAllMaskPixels;
	/// Learn multiple depth labels which are separated by connectivity
	Parameter< bool > learnMultipleDepthLabels;
	/// Only learn pixels near to the ground truth edges
	Parameter< bool > learnEdgeNeighboursOnly;

	/// Size of neighbourhood around ground truth edges
	/** Neighbourhood is determined by a gaussian blur with a sigma of this value */
	Parameter< int > neighbourhoodSize;
	/// Offset around warped ground truth which defines the area 
	/// considered not to contain non depth edge pixels
	Parameter< int > neighbourhoodOffset;
	/// Offset around ground truth which defines the area 
	/// considered not to contain non depth edge pixels
	Parameter< int > maskOffset;
	/// Offset around edges which defines the area 
	/// considered not to contain edge pixels
	Parameter< int > edgeOffset;
	
	/// Number of neighbour frames to include into RF model
	Parameter< int > numberOfNeighbours;

	/// Print info to log
	Parameter< bool > verbose;


protected:
	/// Update object.
	virtual void execute();

private:
	/// Check whether the sizes of all input images are consistent
	void CheckInputConsistency();
		
	/// Get label and feature vectors
	bool getFeaturesAndLabels(
		const int index,											// current id
		const std::vector<int> &nnFeatures,							// vector of ids
		std::vector<int> &_vecImgIds,								// coordinate of samples
		std::vector<int> &_vecIds,									// coordinate of samples
		std::vector<int> &_vecSamples,								// label of samples
		int &_nof,													// number of features
		int &_nos,													// number of samples
		std::vector<int> &labels									// vector of labels
		);

	// Copy values from vector to vigra
	void getFeaturesAndLabels( 
		const std::vector<int> &_vecImgIds,	// vector of img id
		const std::vector<int> &_vecIds,	// vector of pixel id
		const std::vector<int> &vecl,		// vector of label
		const std::vector<int> nnFeatures,	// permutation between iterators and ids
		MultiArray<2, float> &f,			// multiarray to write to
		MultiArray<2, int> &l				// multiarray to write to
		);

};

#endif // _DEPTHEDGELEARN_H_
