/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeLearn.hxx
 *  Implementation of the parameter class DepthEdgeLearn.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.08.2013
 */

#ifndef _DEPTHEDGELEARN_HXX_
#define _DEPTHEDGELEARN_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "DepthEdgeLearn.h"
#include <ctime>

template <typename T>
DepthEdgeLearn<T>::DepthEdgeLearn(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthEdgeLearn", name,
			"<h2>Learn masked pixels</h2><br>"
			"This module learns one ore more random forests using masked "
			"pixels. A frame can be divided in several patches where every "
			"patch is assigned to one random forest. So the models can be "
			"specified on specific image areas. In this case, the according "
			"prediction should use optical flow to determine the new patch "
			"positions in subsequent frames."
		), depthEdgeModelIn(true,false)
{

	ParameteredObject::_addInputSlot(
		depthEdgeModelIn, "depthEdgeModelIn",
		"The depth edge model which was already learned."
		" This model can be empty, if learning module runs for the first time.");
	ParameteredObject::_addInputSlot(
		features, "features",
		"Pointer to the feature module.",
		"vector<string>");
	ParameteredObject::_addInputSlot(
		imgEdge, "imgEdge",
		"Images with highlighted edges (for example CannyEdges)",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		imgMask, "imgMask",
		"Images with annotated masks",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		idMask, "idMask",
		"A vector with ids according to which images of the imgMask list"
		"should be used for the learn process",
		"vector<int>");
	ParameteredObject::_addInputSlot(
		imgMaskWarped, "imgMaskWarped",
		"Optical flow images",
		"CImgList<T>");	

	ParameteredObject::_addOutputSlot(
		depthEdgeModelOut, "depthEdgeModelOut",
		"The depth edge model which keeps random forests, its patch coordinates"
		" and its ids. It also contains the feature setup which was chosen"
		" at learning runtime.");

	ParameteredObject::_addOutputSlot(
		sample, "sample",
		"sample",
		"CImgList<T>");

	ParameteredObject::_addParameter< bool >(
		relearn, "relearn",
		"Relearn available models",
		true);

	ParameteredObject::_addParameter(
		useOpenMP, "useOpenMP",
		"Use open mp to parallize patch calculation",
		false, "bool");
	ParameteredObject::_addParameter< int >(
		treeCount, "treeCount",
		"Number of random forest trees",
		150, "int");
	ParameteredObject::_addParameter< int >(
		samplesPerTree, "samplesPerTree",
		"Number of samples per tree",
		1000, "int");
	ParameteredObject::_addParameter< int >(
		minSplitNodeSize, "minSplitNodeSize",
		"Split node every minSplitNodeSize nodes",
		1, "int");
	ParameteredObject::_addParameter< bool >(
		equalizeSampleCount, "equalizeSampleCount",
		"Take equal numbers of samples for texture and depth edges.",
		true);
	ParameteredObject::_addParameter< bool >(
		learnSeparateEdgeLabel, "learnSeparateEdgeLabel",
		"Learn edge labels beside depth edges and non edges.",
		true);
	ParameteredObject::_addParameter< bool >(
		learnAllPixels, "learnAllPixels",
		"Learn all pixels.",
		false);
	ParameteredObject::_addParameter< bool >(
		learnAllMaskPixels, "learnAllMaskPixels",
		"Learn all mask pixels.",
		false);
	ParameteredObject::_addParameter< bool >(
		learnMultipleDepthLabels, "learnMultipleDepthLabels",
		"Learn multiple depth labels which are separated by connectivity.",
		false);
	ParameteredObject::_addParameter< bool >(
		learnEdgeNeighboursOnly, "learnEdgeNeighboursOnly",
		"Only learn pixels near to the ground truth edges.",
		true);
	ParameteredObject::_addParameter< int >(
		neighbourhoodSize, "neighbourhoodSize",
		"Size of neighbourhood around ground truth edges. "
		"Neighbourhood is determined by a gaussian blur with a sigma of this value",
		25, "int");
	ParameteredObject::_addParameter< int >(
		neighbourhoodOffset, "neighbourhoodOffset",
		"Offset around WARPED ground truth which defines the area which is not"
		" considered not to contain non depth edge pixels",
		7, "int");
	ParameteredObject::_addParameter< int >(
		maskOffset, "maskOffset",
		"Offset around ground truth which defines the area which is not"
		" considered not to contain non depth edge pixels",
		1, "int");
	ParameteredObject::_addParameter< int >(
		edgeOffset, "edgeOffset",
		"Offset around edges which defines the area which is not"
		" considered not to contain non depth edge pixels",
		1, "int");
	ParameteredObject::_addParameter< int >(
		numberOfNeighbours, "numberOfNeighbours",
		"Number of neighbour frames to include into RF model",
		1, "int");

	ParameteredObject::_addParameter< bool >(
		verbose, "verbose",
		"Print info to log",
		false, "bool");
	
}

template <typename T>
void DepthEdgeLearn<T>::execute() {

	CheckInputConsistency(); // Check whether input dimensions are consistent

	bool _existModel = this->depthEdgeModelIn.connected();	// is there already a model?
	if( _existModel && this->relearn() ){
		sout << "Found an existing model. RELEARN is activated. The existing model won't be considered" << std::endl;
	}
	if( _existModel && !this->relearn() ){
		sout << "Found an existing model. Taking appropriate models from there" << std::endl;
	}

	hekate::DepthEdgeModel &_modelOut = this->depthEdgeModelOut();		// output model
	_modelOut.clear();

	// Iterate over all key frames which contain annotated edges
	for( size_t keyIndex = 0; keyIndex < this->idMask().size(); keyIndex++ )
	{
		const clock_t begin_time = clock();

		
		const int index = this->idMask()[keyIndex] ;	// get index of current key frame
		
		// If a known model for index frame is connected and compatible to current models
		// take this model and skip relearn
		int pos2;
		if( !this->relearn() && _existModel && this->depthEdgeModelIn().exists(index,pos2) ){
			CImgList<float> features( this->features()[index].c_str() );
			if( this->depthEdgeModelIn().randomForests[pos2].feature_count() == features.size() )
			{
				sout << "Take available model for random forest " << index << std::endl;
				_modelOut.saveModel(index,this->depthEdgeModelIn().randomForests[pos2]);
				continue;
			}
			else{
				sout << "A random forest " << index << " is available but does NOT match to current "
					<< "settings. Relearning model" << std::endl;
			}
		}

		std::vector<int>	_rf_labels;					// list of used labels
		std::vector<int>	_vecImgIds;					// vector with image ids of features
		std::vector<int>	_vecIds;					// vector with pixel ids of features
		std::vector<int>	_vecSamples;				// vector with labels of features
		int					_nof, _nos;					// total number of features and samples

		// The new random forest is combined by new learned trees based on the
		// feature list and already learned trees of nearest neighbour models.
		// Therefore, a list of nearby models and feature lists has to be determined
		// Actually, if a model for the current frame was known, a relearn could
		// also be based on keeping some trees of this id. For now, all samples of the
		// current id are relearnd and non of the trees of this id are kept.		
		std::vector<const RandomForest<int>*>	nnRF;		// vector neighbour models
		std::vector<int>						nnFeatures;	// vector frames ids to learn

		nnFeatures.push_back( index );						// save id of current frame

		// If a model of the neighbour frames is available, take existing trees, otherwise, save
		// neighbour id to learn features of that frame.
		for( int i = (-1)*this->numberOfNeighbours(); i <= this->numberOfNeighbours(); i++ ){

			if( i == 0 ) continue;	// skip 'index' frame

			if( index+i >= 0 && index+i < this->features().size() ){	
				int pos;
				if( _existModel && this->depthEdgeModelIn().exists(index+i, pos) ){
					nnRF.push_back( &this->depthEdgeModelIn().randomForests[pos] );
					sout << "Use RF of frame " << index+i << " for additional trees" << std::endl;
				}
				else{
					nnFeatures.push_back(index+i);
					sout << "Use pixles of frame " << index+i << " as additional samples" << std::endl;
				}
			}

		}


		// Get labels and pixel ids of all features to learn. (this is an intermediate step since the
		// vigra multi arrays do not support push backs. Push backs are used, since the number of 
		// features is not known yet.
		getFeaturesAndLabels( index, nnFeatures, _vecImgIds, _vecIds, _vecSamples, _nof, _nos, _rf_labels);

		sout << "Found " << _nof << " differen features for " << _nos << " sample points" << std::endl;

		// Continue, if not enough samples were found
		if( !_nos ){
			sout << "DepthEdgeLearn<T>::execute(): Image " << index << ": Not enough samples" << std::endl;
			continue;
		}

		MultiArray<2, feature_t> f( vigra::MultiArrayShape<2>::type(_nos,_nof) );	// feature multi array
		MultiArray<2, label_t> l( vigra::MultiArrayShape<2>::type(_nos,1) );		// label multi array

		sout << "Number of features: " << _nof << " Number of samples" << _nos << std::endl;

		//sout << "Get features and labels" << std::endl;
		getFeaturesAndLabels( _vecImgIds, _vecIds, _vecSamples, nnFeatures, f, l );
			
		// Create RF object
		/* ProblemSpec<... will keep label ids, even if some labels do not have any samples */
		RandomForest<int> rf(	RandomForestOptions().tree_count( this->treeCount() ) , 
								ProblemSpec<int>().classes_(_rf_labels.begin(),_rf_labels.end()) 
								); 

		// activate equal sample numbers for the labels only if nEdges ~= 2* nMaskEdges
		// this condition prevents the case, that all edge pixels also are mask pixels
		// so there wouldnt be any textureEdge pixel to learn
		if( this->equalizeSampleCount() )
		{
			//sout << "\tnOfMaskPixels / nOfEdgePixels < 0.7 --> rf.use_stratification(RF_EQUAL)" << std::endl;
			rf.set_options().use_stratification(vigra::RF_EQUAL);
		}
				
		// Create random forest objects and learn the features
		const int spt = std::min( unsigned int(_nos), unsigned int(this->samplesPerTree()) );
		//sout << "\tUsing " << spt << " samples per tree" << std::endl;
		//sout << "\tUsing " << this->minSplitNodeSize() << " minSplitNodeSize" << std::endl;
			
		rf.set_options().samples_per_tree( spt );
		rf.set_options().min_split_node_size( this->minSplitNodeSize() );
				
		// Create progress visitors
		visitors::RandomForestProgressVisitor progress;
		auto visitor1 = visitors::create_visitor(progress);
		visitors::OOB_Error oob_v;
		visitors::VariableImportanceVisitor importance(1);
		auto visitor2 = visitors::create_visitor(progress, oob_v, importance);

		sout << "Learn RandomForest" << std::endl;
				
		// Learn RF
		if( !verbose() )
			rf.learn(f, l, visitor1);
		else {
			sout << "Learning... (with importance vector generation)" << std::endl;
			rf.learn(f, l, visitor2);
			sout << "Importance Vector:" << std::endl;
			MultiArray< 2, double > imp = importance.variable_importance_;
			for(size_t i=0; i < _nof; i++)
			{
				sout << "\t" << imp[i] << "\t" << imp[_nof+i] << "\t" << imp[2*_nof+i] 
						<< "\t" << imp[3*_nof+i] << "\t" << imp[4*_nof+i] << std::endl;
			}
		}




		// Determine tree count
		const int tree_count_usr = this->treeCount();

		// Add trees from former random forests
		if( nnRF.size() > 0 ){
			sout << "Adding additional trees from former models" << std::endl;
			for( int j = 0; j < nnRF.size(); j++ ){
				const int tree_count_add = std::min(nnRF[j]->tree_count(), tree_count_usr) / 3;
				const int tree_count_current = rf.tree_count();
				for( int k = 0; k < tree_count_add; k++ ){
					rf.trees_.push_back(nnRF[j]->tree(k));
				}
				rf.set_options().tree_count_ = tree_count_current + tree_count_add ;
			}
		}
				
		sout << "Leared frame " << index << " in " << float( clock () - begin_time ) / CLOCKS_PER_SEC << " seconds" << std::endl;
		
		// Save random forest ---------------------------------------------------------
		_modelOut.saveModel(index,rf);

	}

}


/// Find sample point which have to be learned and save their lables, coordinates (pixel ids) and image ids
template <typename T>
bool DepthEdgeLearn<T>::getFeaturesAndLabels(
		const int index,											// current frame id
		const std::vector<int> &nnFeatures,							// vector of ids to learn
		std::vector<int> &_vecImgIds,								// vector with image ids of feature
		std::vector<int> &_vecIds,									// vector with pixel ids of feature
		std::vector<int> &_vecSamples,								// vector with labels of features
		int &_nof,													// number of features
		int &_nos,													// number of samples
		std::vector<int> &labels									// list of used labels
		){

	const clock_t begin_time = clock();		// start timer

	unsigned int width  = this->imgMask()[index].width();
	unsigned int height = this->imgMask()[index].height();
	
	const bool _learnAll			= this->learnAllPixels();			// assign all pixels
	const bool _learnAllMasks		= this->learnAllMaskPixels();		// assign all masked pixels
	const bool _learnNeighbours		= this->learnEdgeNeighboursOnly();	// assign only samples near to ground truth
	const bool _learnMultiLabels	= this->learnMultipleDepthLabels();	// assign multiple labels for depth edges

	// iterate over all frames used to learn
	for( int i = 0; i < nnFeatures.size(); i++ ){

		int nDE = 0, nE = 0, nNE = 0, id = 0;	// label counters
	
		// debugging images
		CImg<float> smpl(width,height,1,1,0);
		CImg<float> smpl2(width,height,1,1,0);
		CImg<float> smpl3(width,height,1,1,0);
		CImg<float> smpl4(width,height,1,1,0);

		// index of current frame iteration
		const int index2 = nnFeatures[i] ;
		
		sout << "Assign features of frame " << index2 << std::endl;

		// Get edge labels
		CImg<int> edge_labels;
		if( _learnMultiLabels && index2 == index ) 
			edge_labels = this->imgMask()[index].get_label();
		
		// Read mask image (black if index != index2) 
		CImg<int> _mask(width,height,1,1,0);
		// Read edge image
		const CImg<int>		_edge			= this->imgEdge()[index2].get_channel(0).quantize(2).normalize(0,1);
		// Read edge image offset
		const CImg<int>		_edge_offset	= this->imgEdge()[index2].get_blur( this->edgeOffset() ).channel(0).normalize(0,1).cut(0,0.1).quantize(2).normalize(0,1);
		// Read neighbourhood mask
		const CImg<float>	_mask_blur		= this->imgMaskWarped()[index2].get_blur( this->neighbourhoodSize() ).channel(0).normalize(0,1).cut(0,0.1).quantize(2).normalize(0,1);
		// Read depth edge offset
		CImg<float>			_mask_offset	= this->imgMaskWarped()[index2].get_blur( this->neighbourhoodOffset() ).channel(0).normalize(0,1).cut(0,0.1).quantize(2).normalize(0,1);


		// If current frame is equal to current iteration frame
		if( index == index2 ){ 
			// Get ground truth mask
			_mask = this->imgMask()[index].get_channel(0).quantize(2).normalize(0,1);
			// Replace rough offset with exactly known offset from ground truth
			_mask_offset = this->imgMask()[index].get_blur( this->maskOffset() ).channel(0).normalize(0,1).cut(0,0.1).quantize(2).normalize(0,1);
		}
		
		//this->sample().push_back( _edge.get_normalize(0,255) ) ;
		//this->sample().push_back( _edge_offset.get_normalize(0,255) ) ;
		//this->sample().push_back( _mask_blur.get_normalize(0,255) ) ;
		//this->sample().push_back( _mask_offset.get_normalize(0,255) ) ;
		//this->sample().push_back( _mask.get_normalize(0,255) ) ;

		int OKO = 0;

		// Iterate over frame
		cimg_forXY( _edge, x, y){

			const unsigned int pixel = y * width + x;	// coordinate id

			/*
				Different labels:
				No edge			--> label 0
				Texture edge	--> label 1
				Depth edge		--> label 2 - x
				
				Depth edges can have different labels separated by connectivity
			*/

			//if	( _edge[pixel] > 0 && _mask_offset[pixel] == 0 && !_learnNeighbours ) sout << "111" << std::endl;							// any edge != DE
			//if	(_edge[pixel] > 0 && _mask_offset[pixel] == 0 && _learnNeighbours && _mask_blur[pixel] != 0 ) sout << "222" << std::endl;					// edge != DE && in DE neighbourhood	
			//if	( index != index2 && _edge[pixel] > 0 && _mask_offset[pixel] == 0 && _learnAll ) OKO++;				// edge != DE && in DE neighbourhood		

			// Detect depth edges for frame id 'index'
			if( index == index2 ){
				if ( 
					( _mask[pixel] > 0 && _edge[pixel] > 0 )||	// intersection Depth Edges with Edges
					( _mask[pixel] > 0 && _learnAllMasks  )	||	// all depth edges
					( _mask[pixel] > 0 && _learnAll  )			// all depth edges
				)
				{
					smpl(x,y)	= 255; // for debug sample image
					smpl2(x,y)	= 255; // for debug sample image

					// save label
					if( _learnMultiLabels )
						_vecSamples.push_back(edge_labels( x , y ) + 2); //depth edge label = id of edge segment
					else
						_vecSamples.push_back(2); //depth edge label = 2

					_vecIds.push_back( pixel );	// save pixel id
					_vecImgIds.push_back(i);	// save image id

					id++;
					nDE++;
					continue;
				}
			}
			// texture edge detected
			if ( 
				( _edge[pixel] > 0 && _mask_offset[pixel] == 0 && !_learnNeighbours ) ||							// any edge != DE
				( _edge[pixel] > 0 && _mask_offset[pixel] == 0 && _learnNeighbours && _mask_blur[pixel] != 0 ) ||	// edge != DE && in DE neighbourhood	
				( _edge[pixel] > 0 && _mask_offset[pixel] == 0 && _learnAll ) 	// edge != DE && in DE neighbourhood				
			)
			{
				smpl(x,y) = 255; // for sample image
				smpl3(x,y) = 255; // for sample image
				if( this->learnSeparateEdgeLabel() )
					_vecSamples.push_back(1); //texture edge label = 1	
				else
					_vecSamples.push_back(0); //texture edge label = 1	
				_vecIds.push_back( pixel );	
				_vecImgIds.push_back(i);
				id++;
				nE++;
				continue;
			}

			if( 
					( !_learnNeighbours && _mask_offset[pixel] == 0 && _edge_offset[pixel] == 0 ) ||  
					( _learnNeighbours && _mask_blur[pixel] != 0 && _mask_offset[pixel] == 0 && _edge_offset[pixel] == 0 )
				){
				smpl(x,y) = 255; // for sample image
				smpl4(x,y) = 255; // for sample image
				_vecSamples.push_back(0); // non-edge label = 1				
				_vecIds.push_back( pixel );
				_vecImgIds.push_back(i);
				id++;
				nNE++;
				continue;
			}
		}

		//sout << "OKO = " << OKO << std::endl;
	
		//this->sample().push_back( smpl );
		this->sample().push_back( smpl2 );
		this->sample().push_back( smpl3 );
		this->sample().push_back( smpl4 );

		sout << "\tAssigned " << nDE << " depth edge samples, " << nE << " edge samples and " << nNE << " non-edge samples" << std::endl;
		sout << "\tAssigned " << nDE+nE+nNE << " samples overall" << std::endl;

		if( index == index2 && !nDE )
			return false;

	}

	sout << "\tDetermined pixels to learn (" << float( clock () - begin_time ) / CLOCKS_PER_SEC << "s)" << std::endl;

	// Get number of samples
	_nos = _vecSamples.size();

	// Get number of features
	_nof = CImgList<float>( this->features()[0].c_str() ).size();

	// Get unique vector of label ids
	std::set<int> s;
	int size = _vecSamples.size();
	for( unsigned i = 0; i < size; ++i ) s.insert( _vecSamples[i] );
	labels.assign( s.begin(), s.end() );
	return 1;

}


// getFeaturesAndLabels( index, nnFeatures, nnRF, _vecFeatures, _vecSamples _nof, _nos, _rf_labels);
// Create list of sample pixels containing edge and random pixels
template <typename T>
void DepthEdgeLearn<T>::getFeaturesAndLabels(
		const std::vector<int> &_vecImgIds,	// vector of img id
		const std::vector<int> &_vecIds,	// vector of pixel id
		const std::vector<int> &vecl,		// vector of label
		const std::vector<int> nnFeatures,	// permutation between iterators and ids
		MultiArray<2, float> &f,			// multiarray to write to
		MultiArray<2, int> &l				// multiarray to write to
		){
	
	sout << "Read assigned pixel values" << std::endl;

	unsigned int width  = this->imgMask()[0].width();
	unsigned int height = this->imgMask()[0].height();

	const clock_t begin_time = clock();

	if(_vecImgIds.size() != _vecIds.size() || _vecIds.size() != vecl.size() )
		sout << "!!!!!!!!!!!!!!" << std::endl;

	sout << " nnFeatures.size() = " << nnFeatures.size() << std::endl;
	
	// Get feature images
	std::vector< CImgList<float> > featureList;
	for( int i = 0; i < nnFeatures.size(); i++ ){
		sout << "nnFeatures[i] = " << nnFeatures[i] << std::endl;
		sout <<  this->features()[nnFeatures[i]].c_str() << std::endl;
		CImgList<float> list( this->features()[nnFeatures[i]].c_str() );
		featureList.push_back( list );
		sout << list.size() << std::endl;
	}
	const int _nof = featureList[0].size();
	sout << _nof << std::endl;

	for( int i = 0; i < _vecImgIds.size(); i++ ){

		l(i,0) = vecl[i];	
		for( size_t j = 0; j < _nof; j++){
			f(i,j) = featureList[_vecImgIds[i]][j][_vecIds[i]];		
		}


	}

	sout << "done" << std::endl;

}


// Check whether the sizes of all input images are consistent
template <typename T>
void DepthEdgeLearn<T>::CheckInputConsistency() {

	sout << "Check input values and images" << std::endl;

	if( this->imgEdge().size() == 0 || this->imgMask().size() == 0 || this->features().size() == 0
		|| this->imgMaskWarped().size() == 0 )
		throw(std::range_error("DepthEdgeLearn<T>::CheckInputConsistency: No images given"));

	// Compare size of mask and edge image
	if( this->imgEdge()[0].width() != this->imgMask()[0].width() || this->imgEdge()[0].width() != this->imgMaskWarped()[0].width() )
		throw( std::runtime_error("DepthEdgeLearn<T>::CheckInputConsistency(): mask image dimension != edge image dimension") );

	// Compare size of mask and edge image
	sout << "Check feature image dimension" << std::endl;
	if(this->imgMask()[0].width() != cimg_library::CImgList<float>(this->features()[0].c_str())[0].width() )
		throw( std::runtime_error("DepthEdgeLearn<T>::CheckInputConsistency(): edge image dimension != dimension of images assigned to feature module") );

	sout << "end" << std::endl;
}

#endif /* _DEPTHEDGELEARN_HXX_ */
