/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgePredict.h
 *  Declaration of the parameter class DepthEdgePredict.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.02.2013
 */

#ifndef _DEPTHEDGEPREDICT_H_
#define _DEPTHEDGEPREDICT_H_

#ifdef _MSC_VER
#ifdef depthedgepredict_EXPORTS
/// Visual C++ specific code
#define depthedgepredict_DECLDIR __declspec(dllexport)
#else
#define depthedgepredict_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else
/// Not needed without MSVC
#define depthedgepredict_DECLDIR
#endif

#include <charon-utils/CImg.h>
#include <charon-core/ParameteredObject.h>

#include <opencv/cv.h>
#include <opencv/highgui.h>

#include <vigra/windows.h>
#include <vigra/random_forest.hxx>

#include <hekate/DepthEdgeFeatures.h>
#include <hekate/DepthEdgeModel.hxx>

#include <ctime>

using namespace cimg_library;
using namespace vigra;
using namespace rf;

typedef float feature_t; 
typedef int label_t;

/// Predicts depth edge by random forest model given by LearnDepthEdgeByMask
/** Predicts depth edge by random forest model given by LearnDepthEdgeByMask
 */
template <typename T>
class depthedgepredict_DECLDIR DepthEdgePredict :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	DepthEdgePredict(const std::string& name = "");

	/// The depth edge model which keeps random forests, its patch coordinates
	///	and its ids and the feature setup which was chosen at learning runtime
	InputSlot< hekate::DepthEdgeModel > depthEdgeModel;
	/// Pointer to feature module
	InputSlot< std::vector<std::string> > features;
	/// Edges of images (for example Canny edges)
	InputSlot< cimg_library::CImgList<T> > edges;
	/// List of flow files required to track image patches
	InputSlot< CImgList<T> > flow;
	/// Ground truth images and warped images if GT was not known
	InputSlot< CImgList<T> > groundTruth;
	
	/// Probability of depth edges
	OutputSlot< cimg_library::CImgList<T> > probability;
	/// Labels of predicted of depth edges
	OutputSlot< cimg_library::CImgList<T> > labels;

	/// Use open mp to parallize patch calculation
	Parameter< bool > useOpenMP;
	/// Number of OpenMP thread
	Parameter< int > threads;
	/// Start index of images to predict
	Parameter< int > inFrame;
	/// End index of images to predict
	Parameter< int > outFrame;

	/// Only predict edge pixels given by edge images
	Parameter< bool > predictEdgePixelsOnly;
	/// Only learn pixels near to the warped ground truth edges
	Parameter< bool > predictEdgeNeighboursOnly;
	/// Amount of blur value to determine a neighbourhood of warped ground truth depth edges
	Parameter< int > edgeNeighbourBlur;

	/// If only one model is available, though split frame to patches to parallely compute labels
	Parameter< bool > multiThreadsModel;

	/// Predict also frames for which the ground truth is known (for testing purposes only)
	Parameter< bool > predictGroundTruth;

protected:
	/// Update object.
	virtual void execute();
	size_t getFeatureAndSamplesPredict( CImgList< float > &featureList, 
				std::vector< unsigned int > &sampleList, const unsigned int index);

	/// Check input consistency
	void init(int &i_start, int &i_end);


	/// Assign frames to masks for prediction
	void assignFrames( unsigned int &i_start, unsigned int &i_end, unsigned int i, unsigned int &predstart);

	/// Get coordinates for number_of_patches patches
	std::vector<std::pair< int, int> > trackPatch( 
						const int index,
						const int number_of_patches );

	/// Write probability values to output slot
	void saveProbabilities( const unsigned index, 
							const MultiArray<2, float> &probability, 
							const std::vector< int > &permutation,
							const int _nol );

	/// Write prediction values to output slot removing label 0 and 1
	void savePrediction( const unsigned index, 
							const MultiArray<2, float> &prediction, 
							const std::vector< int > &permutation );

	/// Determine number of samples
	/** Takes coordinates of according image patch */
	unsigned int getNumberOfSamples(const unsigned int index, const std::vector< std::pair <int, int> > &coord);
	
	/// Determine coordinates of samples to predict
	void getFeatureCoordinates(
							const int index,			// current id
							std::vector<int> &_vecIds,	// coordinate of samples
							const std::vector< std::pair< int, int> > &fc,
							const CImg<int>	&_edge,
							const CImg<float> &_mask_blur
	);

	/// Fill feature multi array with pixels of already determined cooridnates
	void getFeatures(	
							MultiArray<2, feature_t> &f, 
							const CImgList<T> &featureList, 
							const unsigned int _nof,
							const std::vector< int > &coord 
	);

};
#endif // _DEPTHEDGEPREDICT_H_
