/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgePredict.hxx
 *  Implementation of the parameter class DepthEdgePredict.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.02.2013
 */

#ifndef _DEPTHEDGEPREDICT_HXX_
#define _DEPTHEDGEPREDICT_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "DepthEdgePredict.h"
#include <charon-utils/ImgTool.hxx>

template <typename T>
DepthEdgePredict<T>::DepthEdgePredict(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthEdgePredict", name,
			"<h2>Predicts depth edges by a random forest model given by the module"
			"LearnDepthEdgeByMask</h2><br>"
			"Predicts depth edges by a random forest model given by the module"
			"LearnDepthEdgeByMask. Probabilities of depth edges are calculated for every pixel of the given image sequence."
		),
		flow(true,false)
{

	ParameteredObject::_addInputSlot(
		depthEdgeModel, "depthEdgeModel",
		"The depth edge model which keeps random forests, its patch coordinates"
		" and its ids. It also contains the feature setup which was chosen"
		" at learning runtime.");
	ParameteredObject::_addInputSlot(
		features, "features",
		"Pointer to feature module",
		"vector<string>");
	ParameteredObject::_addInputSlot(
		edges, "edges",
		"Edges of images (for example Canny edges)",
		"CimgList<T>");
	ParameteredObject::_addInputSlot(
		flow, "flow",
		"list of flow files required to track image patches",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		groundTruth, "groundTruth",
		"Ground truth images and warped images if GT was not known",
		"CImgList<T>");
	
	ParameteredObject::_addOutputSlot(
		probability, "probability",
		"Probability of predicted depth edges",
		"CimgList<T>");
	ParameteredObject::_addOutputSlot(
		labels, "labels",
		"Labels of predicted of depth edges",
		"CimgList<T>");

	ParameteredObject::_addParameter(
		useOpenMP, "useOpenMP",
		"Use open mp to parallize patch calculation",
		false, "bool");
	ParameteredObject::_addParameter< int >(
		threads, "threads",
		"Number of OpenMP threads. Be careful when using large feature sets.",
		4, "int");

	ParameteredObject::_addParameter< int >(
		inFrame, "inFrame",
		"Start index of images to predict",
		0, "int");
	ParameteredObject::_addParameter< int >(
		outFrame, "outFrame",
		"End index of images to predict",
		-1, "int");
	ParameteredObject::_addParameter(
		predictEdgePixelsOnly, "predictEdgePixelsOnly",
		"Only predict edge pixels given by edge images",
		false, "bool");
	
	ParameteredObject::_addParameter< bool >(
		predictEdgeNeighboursOnly, "predictEdgeNeighboursOnly",
		"Only learn pixels near to the ground truth edges.",
		false);
	ParameteredObject::_addParameter< int >(
		edgeNeighbourBlur, "edgeNeighbourBlur",
		"Amount of blur value to determine a neighbourhood of depth edges. (Only active if 'learnEdgeNeighboursOnly' is set.",
		25, "int");	
	ParameteredObject::_addParameter< bool >(
		multiThreadsModel, "multiThreadsForSingleModel",
		"If only one model is available, though split frame to patches to parallely compute labels.",
		false);
	ParameteredObject::_addParameter< bool >(
		predictGroundTruth, "predictGroundTruth",
		"Predict also frames for which the ground truth is known (for testing purposes only).",
		false);

	ParameteredObject::_setTags("Hekate;Edges") ;

}

template <typename T>
void DepthEdgePredict<T>::execute() {

	// Initialize frames which have to be predicted
	int i_start, i_end;
	init(i_start, i_end);

	// Reference to random forests
	const RF_list &model_rf = this->depthEdgeModel().randomForests;
	const RF_ids &model_ids = this->depthEdgeModel().imageIds;

	// Number of frames a model was learned for
	const unsigned int n_models = model_ids.size();

	// Iterate through frames, determine according key_sequence frame and predict
	for (unsigned int i = i_start; i <= i_end; i++)
	{

		const clock_t begin_time = clock();

		// index of nearest RF frame according to the depthEdgeModel
		const unsigned int key_model = getKeyFrame( i, model_ids );
		// index of nearest RF frame according to sequence size
		const unsigned int key_sequence = model_ids[ key_model ];

		// If a ground truth is known for this frame, continue with next frame prediction
		//if( key_sequence == i && !this->predictGroundTruth() ){
		//	if(this->labels.connected())
		//		this->labels()[i] = this->groundTruth()[i];
		//	if(this->probability.connected())
		//		this->probability()[i] = this->groundTruth()[i];
		//	continue;
		//}
		
		sout << "Predicting frame " << i << " of sub sequence" << std::endl;
		sout << "index of nearest RF frame according to the depthEdgeModel = " << key_model << std::endl;
		sout << "index of nearest RF frame according to sequence size      = " << key_sequence << std::endl;
		
		if(model_rf.size() == 0)
			throw(std::runtime_error("DepthEdgePredict<T>::execute(): !n_model_patches"));

		int n_model_patches = 1;
		if( this->multiThreadsModel() ){
			n_model_patches = std::max(int(1),this->threads());
			sout << "OpenMP is used to process " << n_model_patches << " sub frames." << std::endl;
		}

		// Get feature images of current frame
		const CImgList< float > featureList( this->features()[ i ].c_str() );
		const int _nof = featureList.size();				// number of features
		const int _nofRF = model_rf[key_model].feature_count();	// number of features
		const int _nol	 = model_rf[key_model].class_count();	// number of labels

		if( _nof != _nofRF ){
			sout << "Available feature images: " << _nof << std::endl;
			sout << "Available features in given model: " << _nofRF << std::endl;
			throw(std::runtime_error("DepthEdgePredict::execute: number of calculated features does not fit to rf model"));
		}
		if( !_nol ){
			throw(std::runtime_error("DepthEdgePredict::execute: RF does not contain any labels"));
		}

		sout << "Predict patches..." << std::endl;
		

		// Read edge image
		const CImg<int>	edge		= this->edges()[i].get_channel(0).quantize(2).normalize(0,1);
		// Read neighbourhood mask
		const CImg<float> mask_blur	= this->groundTruth()[i].get_blur( this->edgeNeighbourBlur() ).channel(0).normalize(0,1).cut(0,0.1).quantize(2).normalize(0,1);

		#pragma omp parallel for
		for( int j = 0; j < n_model_patches; j++ )
		{

			const clock_t begin_time2 = clock();

			// Frame coordinates of the image patch
			std::vector< std::pair< int, int> > fc;
			//#pragma omp critical(GET_TRACK)
			//{
				fc = trackPatch( j, n_model_patches );
			//}
			
			// get number of samples to determine
			std::vector<int> _vecIds;
			getFeatureCoordinates(i,_vecIds,fc,edge,mask_blur);
			const int _nos = _vecIds.size();

			if( !_nos ){
				sout << "Not enough samples assigned to prediction list" << std::endl;
				sout << "SKIP prediction for frame " << i << std::endl;
				continue;
			}

			sout << "Chose " << _nos << " samples to predict" << std::endl;
			
			// Multi arrays for probability prediction and feature values
			MultiArray<2, feature_t> f( vigra::MultiArrayShape<2>::type(_nos,_nof) );
			MultiArray<2, float> probability( vigra::MultiArrayShape<2>::type(_nos,_nol) );
			MultiArray<2, label_t> prediction( vigra::MultiArrayShape<2>::type(_nos,1) );
			
			// Get feature values for this patch
			getFeatures( f, featureList, _nof, _vecIds );

			sout << "\tpredict labels and probabilities..." << std::endl;
			
			//#pragma omp critical(PRD)
			//{
				// Calculated probability if slot is connected
				if( this->probability.connected() )
				{
					model_rf[key_model].predictProbabilities( f, probability );	
					// write probability values to output slot
						saveProbabilities( i, probability, _vecIds, _nol );
				}

				// PREDICTION OF LABELS HERE!
				if( this->labels.connected() )
				{
					model_rf[key_model].predictLabels( f, prediction );	
					#pragma omp critical(SAVE_PREDICT)
					{
						savePrediction( i, prediction, _vecIds );
					}
				}
			//}

			sout << "\tPredicted patch " << j << " (" << float( clock () - begin_time2 ) / CLOCKS_PER_SEC << ")" << std::endl;

		}	
		
		sout << "Took " << float( clock () - begin_time ) / CLOCKS_PER_SEC << "sseconds to predict frame " << i << std::endl;

	}

}


// Determine coordinates of samples to predict
template <typename T>
void DepthEdgePredict<T>::getFeatureCoordinates(
		const int index,			// current frame id
		std::vector<int> &_vecIds,	// vector with pixel ids of feature
		const std::vector< std::pair< int, int> > &fc,
		const CImg<int>	&_edge,
		const CImg<float> &_mask_blur
		){

	const clock_t begin_time = clock();		// start timer
		
	sout << "Assign features of frame " << index << std::endl;

	const int width  = _edge.width();
	const int height = _edge.height();

	const bool _predictEdges			= this->predictEdgePixelsOnly();		
	const bool _predictNeighbours		= this->predictEdgeNeighboursOnly();	

	int nos = 0;	// label counters

	sout << fc[0] << " / " << fc[1] << std::endl;

	// Iterate over frame
	for( int x = fc[0].first; x <= fc[1].first; x++ ){
		for( int y = fc[0].second; y <= fc[1].second; y++ ){

			const int pixel = y * width + x;	// coordinate id
			if ( 
				( !_predictEdges && !_predictNeighbours ) ||
				( _predictEdges && !_predictNeighbours && _edge[pixel] > 0 )||
				( _predictEdges && _predictNeighbours  && _edge[pixel] > 0 && _mask_blur[pixel] > 0 )||
				( !_predictEdges && _predictNeighbours && _mask_blur[pixel] > 0 )
			)
			{
					_vecIds.push_back( pixel );	// save pixel id
					nos++;
			}
		}
	}

}


// Fill feature multi array with pixels of already determined cooridnates
template <typename T>
void DepthEdgePredict<T>::getFeatures(MultiArray<2, feature_t> &f,
		const CImgList<T> &featureList, 
		const unsigned int _nof,
		const std::vector< int > &_vecIds ){
	
	// Fill feature array with edge pixel values
	for( int i = 0; i < _vecIds.size(); i++ ){
				
		for( size_t j = 0; j < _nof; j++)
			f(i,j) = featureList[j][_vecIds[i]];

	}
	
}


// Check input consistency
template <typename T>
void DepthEdgePredict<T>::init(int &i_start, int &i_end){
	
	// Clear output slot (for the case the workflow is run more than once)
	this->probability().clear();
	this->labels().clear();
	// Assign black images
	this->probability().assign( this->edges().size(), this->edges()[0].width(), this->edges()[0].height(), 1, 1, T(0) );
	this->labels().assign( this->edges().size(), this->edges()[0].width(), this->edges()[0].height(), 1, 1, T(0) );

	// Determine start and end indices
	if( this->outFrame() < 0 )
	{
		i_end = this->edges().size() -1;
		i_start = std::max( int( this->inFrame() ), int(0) );
	}
	else
	{
		std::pair< int, int > rng = 
			std::minmax( int(this->inFrame()), int(this->outFrame()) );
		const unsigned int rng_max = this->edges().size() - 1;
		i_start = ( rng.first < 0 || rng.first > rng_max ? 0 : rng.first );
		i_end = ( rng.second < 0 || rng.second > rng_max ? rng_max : rng.second );
	}

	sout << "Predicting frame " << i_start << " to " << i_end << std::endl;

	// Check flow dimensions
	if( this->flow.connected() )
	{
		if( this->flow().size() != 2 ){
			sout << "\tflow.size(): " << this->flow().size() << std::endl;
			sout << "\tflow[0].spectrum(): " << this->flow()[0].spectrum() << std::endl;
			sout << "\tflow[0].channels(): " << this->flow()[0].depth() << std::endl;
			throw( std::range_error("DepthEdgePredictLocal<T>::init: flow.size != 2") );
		}
		else if( this->flow()[0].spectrum() < this->edges().size() -1 ){
			sout << "flow()[0].spectrum() = " << this->flow()[0].spectrum() << std::endl;
			throw( std::range_error("DepthEdgePredictLocal<T>::init: >flow.spectrum < edges.size -1") );
		}
	}
	else
		sout << "ATTENTION: No flow given, for reasonable results, a flow field "
		<< "should be available" << std::endl;

	if( this->predictEdgeNeighboursOnly() && !this->groundTruth.connected() ){
		throw( std::runtime_error("No groundTruth files given. Attach groundTruth files or deactivate 'predictEdgeNeighboursOnly'") );
	}

}


// Determine nearest key_sequence frame id
unsigned int getKeyFrame(const unsigned int id, const RF_ids &ids){

	// Distance between current frame and first entry of ids
	unsigned int distance = std::abs( int(ids[0] - id) );
	unsigned int key_sequence = 0;

	// If more entries are available, search for smallest distance
	if( ids.size() > 1 ){

		for( size_t i = 0; i < ids.size(); i++ )
		{
			unsigned int distance2 = std::abs( int(ids[i] - id) );
			if( distance2 < distance )
			{
				distance = distance2;
				key_sequence = i;
			}
		}

	}

	return key_sequence;

}


/// Get coordinates for number_of_patches patches
template <typename T>
std::vector<std::pair< int, int> > DepthEdgePredict<T>::trackPatch(const int index, const int number ){

	const int width		= this->edges()[0].width() ;
	const int height	= this->edges()[0].height() ;
	const int patch_width	= width / number ;
	sout << "patchwidth = " << patch_width << std::endl;
	std::pair< int, int> coord1, coord2;
	coord1.first	= index * patch_width;
	coord1.second	= 0;
	sout << coord1 << std::endl;
	coord2.first	= ( (index+1) == number ? width-1 : (index+1) * patch_width );
	coord2.second	= height - 1;
	sout << coord2 << std::endl;
	std::vector< std::pair< int, int> > vec;
	vec.push_back( coord1 );
	vec.push_back( coord2 );

	return vec;
}


// Write probability values to output slot
template <typename T>
void DepthEdgePredict<T>::saveProbabilities( 
						const unsigned index, 
						const MultiArray<2, float> &probability,
						const std::vector< int > &permutation,
						const int _nol ){
	
	// reference to output image
	CImg<T> &prob = this->probability()[index];

	// Number of depth labels
	int nodl;
	if( _nol < 3 )
		nodl = 1;
	else
		nodl = _nol - 2;

	int nodl2;
	if( _nol < 3 )
		nodl2 = 1;
	else
		nodl2 = 2;



	sout << "nodl = " << nodl << std::endl;
	
	CImg<T> debug(this->edges()[0].width(),this->edges()[0].height(),1,1,0);
	CImg<T> debug2(this->edges()[0].width(),this->edges()[0].height(),1,1,0);

	for( size_t i = 0; i < permutation.size(); i++)
	{
		
		for( size_t j = 0; j < nodl; j++ )
			prob[ permutation[i] ] = T( 255 * probability(i,j+nodl2) );
		
		//debug[ permutation[i] ] = T( 255 * probability(i,0) );
		//debug2[ permutation[i] ] = T( 255 * probability(i,1) );

	}
	
	//this->probability().push_back( debug );
	//this->probability().push_back( debug2 );

}


// Write prediction values to output slot removing label 0 and 1
template <typename T>
void DepthEdgePredict<T>::savePrediction( 
						const unsigned index, 
						const MultiArray<2, float> &prediction, 
						const std::vector< int > &permutation ){
	
	// reference to output image
	CImg<T> &pred = this->labels()[index];

	for( size_t i = 0; i < permutation.size(); i++)
	{
		// if labels belongs to depth edge ( 2 - number of depth edge labels)
		if( prediction(i,0) > 1 )		
			pred[ permutation[i] ] = T( prediction(i,0) );

	}

}


#endif /* _DEPTHEDGEPREDICT_HXX_ */
