/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeWarp.hxx
 *  Implementation of the parameter class DepthEdgeWarp.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 20.11.2013
 */

#ifndef _DEPTHEDGEWARP_HXX_
#define _DEPTHEDGEWARP_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "DepthEdgeWarp.h"

using namespace cimg_library;

template <typename T>
DepthEdgeWarp<T>::DepthEdgeWarp(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthEdgeWarp", name,
			"<h2>Warp ground truthmasks foward and backward to get a warped "
			"mask image for every frame</h2><br>"
			"Warp ground truthmasks foward and backward to get a warped mask "
			"image for every frame"
		)
{
	ParameteredObject::_addInputSlot(
		masks, "masks",
		"Depth Edge Masks which should be warped to the other frames",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		ids, "ids",
		"Ids of the given masks",
		"vector<int>");
	ParameteredObject::_addInputSlot(
		flow, "flow",
		"Flow to warp masks. CImgList dimension != 2 for x and y flow and the"
		"channel dimension contains the different frame values.",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		warped, "warped",
		"Warped masks",
		"CImgList<T>");

}

template <typename T>
void DepthEdgeWarp<T>::execute() {

	this->warped().clear();

	// Get image dimensions
	const int _w = this->masks()[0].width();
	const int _h = this->masks()[0].height();

	// Get flow dimensions
	const int _wf = this->flow()[0].width();
	const int _hf = this->flow()[0].height();

	if( _w != _wf || _h != _hf )
		throw( std::range_error("DepthEdgeWarp<T>::execute: Flow img size does not match to mask"));
	
	const int nwarps = this->flow()[0].spectrum() +1;	// +1 for the new warped image at the end of the sequence
	const int nids = this->ids().size();
		
	sout << "warping " << nwarps-nids << " frames from "
	 << nids << " ground truth images." << std::endl;

	CImgList<T> &warped = this->warped();
	warped.assign(nwarps,_w,_h,1,1,0);

	// Start at first ground truth frame
	for( int i = 0; i < nids; i++ ){

		// id of current ground truth reference
		const int id = this->ids()[i];

		if(id < 0 || id > nwarps -1 )
			throw( std::range_error("DepthEdgeWarp<T>::execute() ID out of range"));

		// Write ground truth to output image
		warped[id] = this->masks()[id];

		// Forward search
		for( int j = id +1; j < nwarps; j++ ){
			
			if( !checkReference(j,id) ){
				break;
			}

			// Warp image of id 'j-1' to frame j
			getWarp(j,true);

		}

		// Backward search
		for( int j = id -1; j >= 0; j-- ){

			if( !checkReference(j,id) ){
				break;
			}

			// Warp image of id 'j-1' to frame j
			getWarp(j,false);

		}

	}

}

// Warp frame of id-1 to frame id using forward flow if forward == true and otherwise fram id+1
template <typename T>
void DepthEdgeWarp<T>::getWarp(const int id, const bool forward){

	const int id2 = ( forward ? id-1 : id+1 );

	if( id2 < 0 || id2 > this->flow()[0].spectrum() )
		throw(std::range_error("DepthEdgeWarp<T>::getWarp: id2 < 0 || id2 >= this->flow().size()"));
	
	CImg<T> &warp	= this->warped()[id];
	CImg<T> &source = this->warped()[id2];

	sout << "Warp frame " << id2 << " to " << id << std::endl;

	if( forward ){
		const CImg<T> &u		= this->flow()[0].get_channel(id2);
		const CImg<T> &v		= this->flow()[1].get_channel(id2);
		cimg_forXY( warp, x, y ){
			warp(x,y) = source.linear_atXY(x-u(x,y),y-v(x,y));
		}
	}
	else{
		const CImg<T> &u		= this->flow()[0].get_channel(id);
		const CImg<T> &v		= this->flow()[1].get_channel(id);
		cimg_forXY( warp, x, y ){
			warp(x,y) = source.linear_atXY(x+u(x,y),y+v(x,y));
		}
	}

}

// Check if current reference frame did not change
template <typename T>
bool DepthEdgeWarp<T>::checkReference( const int id, const int id_ref ){	

	// Get new ground truth reference
	int id_seq, id_vec;
	getNextMaskId(id, id_seq, id_vec);

	// If another ground truth frame is nearer to the current
	// frame iteration, break
	if( id_seq != id_ref ){
		sout << "Current reference id outdated!" << std::endl;
		sout << "Current reference: " << id << std::endl;
		sout << "New reference: " << id_seq << std::endl;
		return false;
	}

	return true;
}

// Get ground truth mask next to this frame. id_sequence is the id according to the sequence size.
// id_vector is the according entry of this->ids() and this->masks().
template <typename T>
void DepthEdgeWarp<T>::getNextMaskId(const int index, int &id_sequence, int &id_vector) {

	int dist = std::abs( index - this->ids()[0] );	// distance to first mask
	id_sequence = this->ids()[0];					// id of first mask
	id_vector = 0;									// id according to vector size

	for( int i = 0; i < this->ids().size(); i++ ){

		int dist2 = std::abs( index - this->ids()[i] );	// distance to mask i
		if( dist2 < dist ){
			dist = dist2;					// replace by new dist
			id_sequence = this->ids()[i];	// replace by new seq id
			id_vector = i;					// replace by new vector id
		}
	}

}

#endif /* _DEPTHEDGEWARP_HXX_ */
