/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthMapFrom3DPointCloud.hxx
 *  Implementation of the parameter class DepthMapFrom3DPointCloud.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 06.03.2013
 */

#ifndef _DEPTHMAPFROM3DPOINTCLOUD_HXX_
#define _DEPTHMAPFROM3DPOINTCLOUD_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/DepthMapFrom3DPointCloud.h>
#include <ImgTool.hxx>

template <typename T>
DepthMapFrom3DPointCloud<T>::DepthMapFrom3DPointCloud(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthMapFrom3DPointCloud", name,
			"<h2>Creates a depth map from a 3d point cloud and given "
			"cameras</h2><br>"
			"Creates a depth map from a 3d point cloud and given cameras. Backprojected points are painted to an image. "
			"All pixels without backprojections are painted in a default color (default = black)"
		)
{
	ParameteredObject::_addInputSlot(
		pointCloud, "pointCloud",
		"3D point cloud to calculate depths of points",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"2D points corresponding with the 3D point cloud. The depth map is"
		"calculated by the 2d position and the 3d depth.",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		cameras, "cameras",
		"Camera matrices",
		"IndexedMatrixList");
	
	ParameteredObject::_addOutputSlot(
		depthMaps, "depthMaps",
		"Calculated depth maps vizualised by pixel values of [0,255]",
		"CImgList<T>");
	
	ParameteredObject::_addParameter(
		width, "width",
		"Width of back projected depth map", 1920,
		"int");
	ParameteredObject::_addParameter(
		height, "height",
		"Height of back projected depth map", 1080,
		"int");
	ParameteredObject::_addParameter(
		range, "range",
		"Number of maximal images to backproject. (0 = all)", 0,
		"int");
	ParameteredObject::_addParameter(
		averageDepthColor, "averageDepthColor",
		"Maps average depth to this specific value", 100,
		"int");
	ParameteredObject::_addParameter(
		backgroundColor, "backgroundColor",
		"Background color if not every pixel has a depth", 0,
		"int");
	ParameteredObject::_addParameter<float>(
		confidence, "confidence",
		"Amount of putative inliers. This values specifies, how many points are not sorted out by the flags "
		"\"sortOutFarestPoints\" and \"sortOutNearestPoints\"."
		"", 0.99 );
	ParameteredObject::_addParameter<bool>(
		sortOutFarestPoints, "sortOutFarestPoints",
		"Sort out farest points by confidence level to prevent outliers", true );
	ParameteredObject::_addParameter<bool>(
		sortOutNearestPoints, "sortOutNearestPoints",
		"Sort out nearest points by confidence level to prevent outliers", false );
	ParameteredObject::_addParameter<bool>(
		useAllPoints, "useAllPoints",
		"Back projects all 3D points without checking for occlusions. Otherwise, only points which initially were seen in each frame are backprojected.", false );
	ParameteredObject::_addParameter<bool>(
		useNeighbourFramePoints, "useNeighbourFramePoints",
		"Back projects 3D points from neighbouring frames without checking for occlusions", false );

	

	ParameteredObject::_setTags("Hekate") ;
}

template <typename T>
void DepthMapFrom3DPointCloud<T>::execute() {

	const IndexedMatrixList &cameras = this->cameras();
	const CorrespondenceList<T> &points = this->pointCloud();
	const CorrespondenceList<T> &corrs = this->correspondences();	
	
	size_t limit = std::min(cameras.size(), corrs.sizeImages()); 
	if( this->range() > 0 && this->range() < limit )
		limit = this->range();
	sout << "Backproject " << limit << " images" << std::endl;

	std::vector< std::vector< std::pair<int,int> > > dmCoordinatesList;
	std::vector< std::vector< float > > dmDepthsList;
	std::vector< float > dmDepthsListOverall;
	
	for(size_t i = 0; i < limit; i++)
	{

		unsigned int index = cameras[i].getIndex1();

		// Write camera information to log
		sout << "Camera " << index << std::endl;
		sout << "Points seen by camera: " << points.sizePoints(index) << std::endl;
		sout << "Points seen by camera: " << points.sizePoints(i) << std::endl;

		if( index != i ){
			sout << "Current camera index != current loop iteration:\n"
				<< "\ti = " << i << " / cameras[i].getIndex1() = " << index << std::endl;
		}
		
		// Create backprojection image
		CImg<float> dm(this->width(), this->height(), 1, 1);
		dm.fill( this->backgroundColor() );

		// Create containers for point coordinates and depth values
		std::vector< std::pair<int,int> > dmCoordinates;
		std::vector< float > dmDepths;

		float dmDepthSum = 0;
		int numberOfIt = 0;
		int notOnPlane = 0;
		int notInImg = 0;
		int negative = 0;	// number of points behind camera

		// Iterate over points initially seen by current camera
		auto it = corrs.get<imgIndex>().lower_bound(index) ;
		for(; it != corrs.get<imgIndex>().upper_bound(index) ; it++)
		{
			Correspondence<T> pt;
			try{  pt = points.getPointByIndex(it->pointIndex); }
			catch(...){ continue; }

			// Get backprojected point
			CImg<double> X = (CImg<double>) pt.pt3d();
			CImg<double> x = cameras[i] * X;
			x /= x[2];
			
			float depth = getDepth( cameras[i], X );

			// Write back projected coordinate to depth map
			if( depth > 0 )
			{
				try{
					if( x[0] >= 0 && x[0] < this->width()-1 && x[1] >= 0 && x[1] < this->height()-1 )
					{
						std::pair< int, int > coord;
						coord.first = int( x[0] + 0.5 );
						coord.second = int( x[1] + 0.5 );
						dmCoordinates.push_back( coord );
						dmDepths.push_back( depth );
						dmDepthsListOverall.push_back( depth );
						dmDepthSum += depth;
						numberOfIt++;
					}
					else
					{
						notOnPlane++;
					}
				}
				catch(...)
				{
				}
			}
			else
				negative++;
		}
		
		
		sout << numberOfIt << " iterations " << std::endl;
		sout << negative << " points lying behind camera " << i << std::endl;
		sout << notOnPlane << " points not on plane " << i << std::endl;
		
		sout << "dmCoordinates: " << dmCoordinates.size() << std::endl;
		sout << "dmDepths: " << dmDepths.size() << std::endl;

		dmCoordinatesList.push_back( dmCoordinates );
		dmDepthsList.push_back( dmDepths );


		/*
		sout << "\t" << notOnPlane << " points not on plane" << std::endl;
		sout << "\t" << notInImg << " 3d points without 2d correspondence in this image" << std::endl;
		sout << "Number of points painted: " << numberOfIt << std::endl;
		sout << "max depth: " << dmDepthsMax << std::endl;
		*/
		//dm.normalize(0,255);
		this->depthMaps().push_back( dm );

	}
	
	
	// Sort depths to remove deepest points ----------------------------------	
	sout << "Sort depths" << std::endl;	
	std::sort( dmDepthsListOverall.begin(), dmDepthsListOverall.end() );	
	float dmDepthsMax = dmDepthsListOverall[ dmDepthsListOverall.size()-1 ];
	float dmDepthsMin = 0;
	if( this->sortOutFarestPoints() || this->sortOutNearestPoints() )
	{
		sout << "Determine max and min allowed depths" << std::endl;

		// Get index of deepest point to paint to depth map
		const unsigned int maxDepthPoint = int( this->confidence() * dmDepthsListOverall.size() );

		// Get index of nearest point to paint to depth map
		const unsigned int minDepthPoint = int( dmDepthsListOverall.size() - maxDepthPoint );
		
		sout << "number of points : " << dmDepthsListOverall.size() << std::endl;
		sout << "max depth index = " << maxDepthPoint << std::endl;
		sout << "min depth index = " << minDepthPoint << std::endl;

		if( this->sortOutFarestPoints() )
		{
			dmDepthsMax = dmDepthsListOverall[ maxDepthPoint ];
		}
		if( this->sortOutNearestPoints() )
		{
			dmDepthsMin = dmDepthsListOverall[ minDepthPoint ];
		}
		std::vector< float > buffer;
		buffer.assign( dmDepthsListOverall.begin() + minDepthPoint, dmDepthsListOverall.end() - minDepthPoint ); 
		dmDepthsListOverall.clear();
		dmDepthsListOverall = buffer;

	}


	// Determine mapping factor for depth values ----------------------------------
	// The average depth is hereby mapped to "this->averageDepthColor()"

	float dmDepthAvg = 0;

	for( size_t i=0; i < dmDepthsListOverall.size(); i++)
		dmDepthAvg += dmDepthsListOverall[i];

	dmDepthAvg /= dmDepthsListOverall.size();
	float depthmapping = this->averageDepthColor() / dmDepthAvg;
	sout << "Averag Depth: " << dmDepthAvg << std::endl;
	sout << "Depth conversion factor: " << depthmapping << " [color value / epipolar depth]" << std::endl;


	// paint depth maps ------------------------------------------------------------
	
	int badDepth = 0;
	for(size_t i = 0; i < limit; i++)
	{

		sout << "Paint depth map" << i << std::endl;

		for( size_t j = 0; j < dmDepthsList[i].size(); j++ )
		{

			if( this->sortOutNearestPoints() && dmDepthsList[i][j] < dmDepthsMin  )
			{
				badDepth++;
				continue;
			}

			if( this->sortOutFarestPoints() && dmDepthsList[i][j] > dmDepthsMax )
			{
				badDepth++;
				continue;
			}

			this->depthMaps()[i]( dmCoordinatesList[i][j].first, dmCoordinatesList[i][j].second, 0, 0) = dmDepthsList[i][j] * depthmapping;
			
		}

	}

	sout << badDepth << " points weren't painted to depth maps due to bad depths" << std::endl;

}



/// Calculate depth between x and principle plane of P.
template <typename T>
double DepthMapFrom3DPointCloud<T>::getDepth(const CImg<T> &P, const CImg<T> &X) {

	int signP;
	(P.get_crop(0,0,2,2).det() < 0 ? signP = -1 : signP = 1);

	double P_m3_abs = P(0,2)*P(0,2)+P(1,2)*P(1,2)+P(2,2)*P(2,2);
	P_m3_abs = pow(P_m3_abs,.5);

	double depth;
	depth = signP * (P*X)(0,2);
	if(P_m3_abs!=0)
	{
		depth /= P_m3_abs;
		return depth;
	}
	else
		return 0;

}

#endif /* _DEPTHMAPFROM3DPOINTCLOUD_HXX_ */
