/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthMapProjection.hxx
 *  Implementation of the parameter class DepthMapProjection.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 08.04.2013
 */

#ifndef _DEPTHMAPPROJECTION_HXX_
#define _DEPTHMAPPROJECTION_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "DepthMapProjection.h"

template <typename T>
DepthMapProjection<T>::DepthMapProjection(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DepthMapProjection", name,
			"<h2>Projects a depth map with given cameras</h2><br>"
			"Projects a depth map with given cameras generating a point cloud. The point cloud can be saved as a *.wrl file."
		)
{
	ParameteredObject::_addInputSlot(
		depthMaps, "depthMaps",
		"DepthMaps given as CImgList",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		cameras, "cameras",
		"Cameras according to depth maps",
		"IndexedMatrixList");

	ParameteredObject::_addOutputSlot(
		pointCloud, "pointCloud",
		"Projected point cloud",
		"CorrespondenceList");

	ParameteredObject::_addParameter<float>(
		mappingFactor, "mappingFactor",
		"Mapping factor which defines transformation between depth map values "
		"and depth values", 1 );

	ParameteredObject::_addParameter<int>(
		stepSize, "stepSize",
		"Step size for pixel selection. Choose a higher value for "
		"a sparse point cloud and faster processing time", 1 );
	
	ParameteredObject::_setTags("Hekate") ;
}

template <typename T>
void DepthMapProjection<T>::execute() {
	
	size_t ndm = this->depthMaps().size();

	// Iterate over every camera
	for( size_t i = 0; i < this->cameras().size(); i++ )
	{
		
		const IndexedMatrix &cam = this->cameras()[i];
		if( ndm -1 < cam.getIndex1() )
			continue;

		sout << "Project with camera " << cam.getIndex1() << std::endl;

		// camera matrix
		cimg_library::CImg<float> P = cam;
		float p0 = P[0],  p1 = P[1],  p2 = P[2],  p3 = P[3];
		float p4 = P[4],  p5 = P[5],  p6 = P[6],  p7 = P[7];
		float p8 = P[8],  p9 = P[9],  p10 = P[10],  p11 = P[11];
		
		int signP;
		(P.get_crop(0,0,2,2).det() < 0 ? signP = -1 : signP = 1);

		float cc = P(0,2)*P(0,2)+P(1,2)*P(1,2)+P(2,2)*P(2,2);
		cc = pow(float(cc),float(.5));

		cc *= signP;

		// depth map
		const cimg_library::CImg<float> &dm = this->depthMaps()[ cam.getIndex1() ] ;
		
		float bpe = 0;
		float bpeIt = 0;

		// Project every pixel
		cimg_forXY( dm, x, y )
		{
			if( dm(x,y) == 0 ) continue;

			if( x%this->stepSize() != 0 || y%this->stepSize() != 0 ) continue;

			float X, Y, Z, dd;
			dd = dm(x,y) / this->mappingFactor();
			//sout << dd << ", " << dm(x,y) << std::endl;
			float a = x, b = y;

			X =		(-b * dd * p1 * p10 + a * dd * p10 * p5 - dd * p2 * p5 + cc * p11 * p2 * p5 - 
					cc * p10 * p3 * p5 +
					dd * p1 * p6 - cc * p1 * p11 * p6 + cc * p1 * p10 * p7 + b * dd * p2 * p9 - 
					a * dd * p6 * p9 +
					cc * p3 * p6 * p9 - cc * p2 * p7 * p9) / 
					(cc * (-p1 * p10 * p4 + p0 * p10 * p5 - p2 * p5 * p8 + p1 * p6 * p8 + 
					p2 * p4 * p9 - p0 * p6 * p9)) ;

			Y =		-(b * dd * p0 * p10 - a * dd * p10 * p4 + dd * p2 * p4 - cc * p11 * p2 * p4 + 
					cc * p10 * p3 * p4 -
					dd * p0 * p6 + cc * p0 * p11 * p6 - cc * p0 * p10 * p7 - b * dd * p2 * p8 + 
					a * dd * p6 * p8 - cc * p3 * p6 * p8 + 
					cc * p2 * p7 * p8)/(cc * (p1 * p10 * p4 - p0 * p10 * p5 + p2 * p5 * p8 - p1 * p6 * p8 - 
					p2 * p4 * p9 + p0 * p6 * p9)) ;


			Z =		-(-dd * p1 * p4 + cc * p1 * p11 * p4 + dd * p0 * p5 - cc * p0 * p11 * p5 + b * dd * p1 * p8 - 
					a * dd * p5 * p8 + cc * p3 * p5 * p8 - cc * p1 * p7 * p8 - b * dd * p0 * p9 + a * dd * p4 * p9 -
					cc * p3 * p4 * p9 + 
					cc * p0 * p7 * p9)/(cc * (p1 * p10 * p4 - p0 * p10 * p5 + p2 * p5 * p8 - p1 * p6 * p8 - 
					p2 * p4 * p9 + p0 * p6 * p9)) ;
			/*
			X = -(b * p1 * p11 - a * p11 * p5 + p3 * p5 - p1 * p7 - b * p3 * p9 + a * p7 * p9 + 
					b * p1 * p10 * z - a * p10 * p5 * z + p2 * p5 * z - p1 * p6 * z - b * p2 * p9 * z + 
					a * p6 * p9 * z)/(-p1 * p4 + p0 * p5 + b * p1 * p8 - a * p5 * p8 - b *  p0* p9 + 
					a * p4 * p9);

			Y = -(b * p0 * p11 - a * p11 * p4 + p3 * p4 - p0 * p7 - b * p3 * p8 + a * p7 * p8 + 
					 b * p0 * p10 * z - a * p10 * p4 * z + p2 * p4 * z - p0 * p6 * z - b * p2 * p8 * z + 
					 a * p6 * p8 * z)/(p1 * p4 - p0 * p5 - b * p1 * p8 + a * p5 * p8 + b * p0 * p9 - 
					 a * p4 * p9);
			*/

			// check back projection
			cimg_library::CImg<float> vec(1,4);
			vec[0] = X, vec[1] = Y, vec[2] = Z, vec[3] = 1;
			cimg_library::CImg<float> vec2 = P * vec;
			vec2 /= vec2[2];
			bpe += std::sqrt( float(vec2[0]-x)*float(vec2[0]-x) + float(vec2[1]-y)*float(vec2[1]-y) );
			bpeIt++;

			float depth = getDepth(P, vec);
			//sout << "depth = " << depth << std::endl;
		//	if( depth <= 0 )
		//		continue;

			this->pointCloud().push_back( hekate::Correspondence<T>( X, Y, Z, 1, x+x*y, cam.getIndex1(), false, dm(x,y), dm(x,y), dm(x,y) ) );

		}

		sout << "bpe overall = " << bpe/bpeIt << std::endl;

	}

}

/// Calculate depth between x and principle plane of P.
template <typename T>
double getDepth(const cimg_library::CImg<T> &P, const cimg_library::CImg<T> &X) {

	int signP;
	(P.get_crop(0,0,2,2).det() < 0 ? signP = -1 : signP = 1);

	double P_m3_abs = P(0,2)*P(0,2)+P(1,2)*P(1,2)+P(2,2)*P(2,2);
	P_m3_abs = pow(P_m3_abs,.5);

	double depth;
	depth = signP * (P*X)(0,2);
	if(P_m3_abs!=0)
	{
		depth /= P_m3_abs;
		return depth;
	}
	else
		return 0;

}

#endif /* _DEPTHMAPPROJECTION_HXX_ */
