/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DynamicSuperpixels.hxx
 *  Implementation of the parameter class DynamicSuperpixels.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 04.12.2012
 */

// This file contains parts of Dynamic Superpixels.
//
// Copyright 2012 Visual Sensorics and Information Processing Lab, Goethe University, Frankfurt
//
// Dynamic Superpixels is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Dynamic Superpixels is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Dynamic Superpixels.  If not, see <http://www.gnu.org/licenses/>.


#ifndef _DYNAMICSUPERPIXELS_HXX_
#define _DYNAMICSUPERPIXELS_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "DynamicSuperpixels.h"

#include <hekate/openCvUtil.hxx>
#include <charon-utils/ImgTool.hxx>

#include <iostream>
#include <cstdlib>
#include <boost/version.hpp>
#include <boost/cstdint.hpp>
#include <boost/filesystem.hpp>
#include <cv.h>
#include <highgui.h>
#include <math.h>

#include "contourRelaxation/FeatureType.h"
#include "contourRelaxation/ContourRelaxation.h"
#include "contourRelaxation/InitializationFunctions.h"
#include "VisualizationHelper.h"

using namespace openCvUtil ;

template <typename T>
DynamicSuperpixels<T>::DynamicSuperpixels(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DynamicSuperpixels", name,
			"<h2>Calculates segmentation of an image sequence</h2><br>"
			"This module is kind of a wrapper of \"DynamicSuperpixels\" from "
			"Goethe University, Frankfurt "
			"(http://www.vsi.cs.uni-frankfurt.de/). It calculates "
			"segmentations of an image sequence and optionally propagates "
			"subsequent segmentations, so the segmentation of frame 1 is "
			"used as initial guess for frame 2."
		)
{
	ParameteredObject::_addInputSlot(
		imagePath, "imagePath",
		"Input image paths",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		labels, "labels",
		"segmentation label ids",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		segmentation, "segmentation",
		"Segmented images",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		countour, "countour",
		"countour",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		countourOverlay, "countourOverlay",
		"countourOverlay",
		"CImgList<T>");

	ParameteredObject::_addParameter< bool >(
		propagateSegmentation, "propagateSegmentation",
		"Use subsequent frames to propagate segmentations",
		1, "bool");
	ParameteredObject::_addParameter< int >(
		blockWidth, "blockWidth",
		"Widths of blocks used for initial segments",
		20, "int");
	ParameteredObject::_addParameter< int >(
		blockHeight, "blockHeight",
		"Heights of blocks used for initial segments",
		20, "int");
	ParameteredObject::_addParameter< double >(
		directCliqueCost, "directCliqueCost",
		"Markov clique cost for horizontal and vertical cliques",
		0.3, "double");
	ParameteredObject::_addParameter< double >(
		compactnessWeight, "compactnessWeight",
		"Weight for the compactness feature, optional, default value = 0.045"
		"for color images, 0.015 for grayscale images", 0, "double");
	
	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void DynamicSuperpixels<T>::execute() {

	// Label image
	cv::Mat labelImage;

	// Set some parameters for Contour Relaxation.
	int const blockWidth = this->blockWidth();
	int const blockHeight = this->blockHeight();
	if (blockWidth < 1 || blockHeight < 1){
		std::cout << "blockWidth and blockHeight must be positive. Aborting." << std::endl;
		throw( std::runtime_error("blockWidth and blockHeight must be positive. Aborting.") );
	}
	sout << "Block width \t\t= " << blockWidth << std::endl;
	sout << "Block height \t\t= " << blockHeight << std::endl;

	// Get clique costs
	double const directCliqueCost = this->directCliqueCost();
	if (directCliqueCost < 0){
		std::cout << "directCliqueCost must be non-negative. Aborting." << std::endl;
		throw( std::runtime_error("directCliqueCost must be non-negative. Aborting.") );
	}
	sout << "Direct clique cost \t= " << directCliqueCost << std::endl;

	// Diagonal clique cost is set relative to the direct clique cost, weighted with the distance between pixel centers.
	double const diagonalCliqueCost = directCliqueCost / std::sqrt(2.);

	// Number of iterations to perform. 3 is usually more than enough but still reasonably fast.
	unsigned int const numIterations = 3;

	// Set the desired type of the label image here.
	// Must be unsigned integer type since it will be used to index vectors.
	// A minimum size should be guaranteed so it can accomodate enough different labels.
	typedef boost::uint16_t TLabelImage;

	// Iterate over all images of the image sequence
	for( size_t i = 0; i < this->imagePath().size(); i++ ){

		sout << "Running dynamic superpixels on " << this->imagePath()[i].c_str() << std::endl;

		boost::filesystem::path inputImage( this->imagePath()[i].c_str() );

		if (!boost::filesystem::exists(inputImage)){
			sout << "Input image " << this->imagePath()[i] << " does not exist. Aborting." << std::endl;
			throw(std::runtime_error("Input image does not exist. Aborting."));
		}

		// Load image and detect type.
		cv::Mat image = cv::imread(inputImage.string(), -1);
		bool isColorImage = false;

		if (image.channels() == 3){
			isColorImage = true;
		}
		else if (image.channels() != 1){
			std::cout << "Input image must have 1 (gray) or 3 (color) channels, format not recognized. Aborting." << std::endl;
			throw( std::runtime_error("Input image must have 1 (gray) or 3 (color) channels, format not recognized. Aborting.") );
		}

		// Set desired features.
		// The set of features enabled here must also be provided with data below!
		std::vector<FeatureType> enabledFeatures;

		if (isColorImage){
			enabledFeatures.push_back(Color);
		}
		else{
			enabledFeatures.push_back(Grayvalue);
		}

		enabledFeatures.push_back(Compactness);

		// Default value for compactness feature weight depends on the type of image.
		// Since color images generate a total of 3 cost terms for the separate channels, we choose
		// the compactness weight at 3 times the value for grayvalue images to preserve the relative weighting.
		double const defaultCompactnessWeight = isColorImage ? 0.045 : 0.015;
		double const compactnessWeight = ( this->compactnessWeight() != 0 ? this->compactnessWeight() : defaultCompactnessWeight );
		if (compactnessWeight < 0)
		{
			std::cout << "compactnessWeight must be non-negative. Aborting." << std::endl;
			throw( std::runtime_error("compactnessWeight must be non-negative. Aborting.") );
		}
		//sout << "Compactness weight \t= " << compactnessWeight << std::endl;

		// Create an instance of the Contour Relaxation class with the selected features.
		ContourRelaxation<TLabelImage> contourRelaxationInst(enabledFeatures);

		// Initialize the label image with a "blind segmentation" of rectangular blocks.
		if( i == 0 || !this->propagateSegmentation() ){
			labelImage = createBlockInitialization<TLabelImage>(image.size(), blockWidth, blockHeight);
		}

		// Set data for all features.
		contourRelaxationInst.setCompactnessData(compactnessWeight);

		if (isColorImage)
		{
			// Convert image to YUV-like YCrCb for uncorrelated color channels.
			cv::Mat imageYCrCb;
			cv::cvtColor(image, imageYCrCb, CV_BGR2YCrCb);
			std::vector<cv::Mat> imageYCrCbChannels;
			cv::split(imageYCrCb, imageYCrCbChannels);

			contourRelaxationInst.setColorData(imageYCrCbChannels[0], imageYCrCbChannels[1], imageYCrCbChannels[2]);
		}
		else
		{
			// Generate a 3-channel version of the grayscale image, which we will need later on
			// to generate the boundary overlay. Save it in the "image" variable so we won't
			// have to care about the original type of the image anymore.
			cv::Mat imageGray = image.clone();
			cv::cvtColor(imageGray, image, CV_GRAY2BGR);

			contourRelaxationInst.setGrayvalueData(imageGray);
		}
		
		// Perform contour relaxation.
		cv::Mat relaxedLabelImage;
		cv::Mat regionMeanImage;
		contourRelaxationInst.relax(labelImage, directCliqueCost, diagonalCliqueCost,
									numIterations, relaxedLabelImage, regionMeanImage);
		
		// Save result in labelImage for initial labels of next iteration
		if( this->propagateSegmentation() ){
			labelImage = relaxedLabelImage;
		}

		if (isColorImage)
		{
			// Convert region-mean image back to BGR.
			// If we are dealing with a grayscale image, the region-mean image is grayscale, too.
			cv::cvtColor(regionMeanImage, regionMeanImage, CV_YCrCb2BGR);
		}

		// Get the name of the input image without the extension for creating the output filenames.
		// Since the boost library has changed starting with version 1.46 and stem() now returns a
		// path object instead of a string, we need to handle both cases.
		#if BOOST_VERSION < 104600
		std::string shortInputName = inputImage.stem();
		#else
		std::string shortInputName = inputImage.stem().string();
		#endif
		
		cv::Mat output = regionMeanImage;
		cv::Mat output2;
		//output.convertTo(output2, CV_8U);
		cv::cvtColor( regionMeanImage, output2, CV_RGB2GRAY );
		sout << "output.channels() = " << output2.channels() << std::endl; ;
		this->segmentation().push_back( Mat2CImg<double>(output2) );
		cimg_library::CImg<TLabelImage> label =  Mat2CImg<TLabelImage>(labelImage);
		sout << "Different labels = " << float(label.max()) << std::endl;
		this->labels().push_back( label );
		
		// Save label image.
		// For TLabelImage = boost::uint16_t, this saves a 16bit PNG file which conserves all the labels.
		//cv::imwrite("D:/Dokumente/Universitt/Doktorarbeit/Daten/workflows/superpixel/results/_labelImage.png", relaxedLabelImage);

		// Save region mean image.
		//cv::imwrite("D:/Dokumente/Universitt/Doktorarbeit/Daten/workflows/superpixel/results/_regionMean.png", regionMeanImage);

		// Create and save boundary image and boundary overlay visualization.
		cv::Mat boundaryImage;
		computeBoundaryImage<TLabelImage>(relaxedLabelImage, boundaryImage);
		// Changing the boundaries from 1 to 255 so the resulting image can be interpreted directly in image viewers.
		boundaryImage *= 255;
		this->countour().push_back( Mat2CImg<TLabelImage>(boundaryImage)  );
		//cv::imwrite("D:/Dokumente/Universitt/Doktorarbeit/Daten/workflows/superpixel/results/resultsboundaryImage.png", boundaryImage);

		cv::Mat boundaryOverlay;
		computeBoundaryOverlay<TLabelImage>(image, relaxedLabelImage, boundaryOverlay);
		//cv::imwrite(shortInputName + "_boundaryOverlay.png", boundaryOverlay);
		this->countourOverlay().push_back( Mat2CImg<TLabelImage>(boundaryOverlay)  );

		
		
	}

}

#endif /* _DYNAMICSUPERPIXELS_HXX_ */
