/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file EigenValuesSchaefer.hxx
 *  Implementation of the parameter class EigenValuesSchaefer.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 11.03.2014
 */

#ifndef _EIGENVALUESSCHAEFER_HXX_
#define _EIGENVALUESSCHAEFER_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "EigenValuesSchaefer.h"

template <typename T>
EigenValuesSchaefer<T>::EigenValuesSchaefer(const std::string& name) :
		TemplatedParameteredObject<T>(
			"EigenValuesSchaefer", name,
			"<h2>An eigenvalue of structure tensor implementation</h2><br>"
			"An eigenvalue of structure tensor implementation originally "
			"written by Henrik Schfer from HCI, Heidelberg, Germany."
		)
{
	ParameteredObject::_addInputSlot(
		images, "images",
		"Images to process",
		"CImgList<T>");
	
	ParameteredObject::_addOutputSlot(
		eigenvalues, "eigenvalues",
		"Eigenvalues saved in z-component",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		eigenvaluesScaleSpaces, "eigenvaluesScaleSpaces",
		"Draws the scale space of maximal eigen values for each pixel",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		eigenvaluesMul, "eigenvaluesMul",
		"Multiplication of eigenvalue 1 and 2",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		eigenvector1, "eigenvector1",
		"Eigenvector perpendicular to an edge. Components saved in Z-Axis.",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		eigenvector2, "eigenvector2",
		"Eigenvector parallel to an edge. Components saved in Z-Axis.",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		angle, "angle",
		"Angle perpendicular to an edge",
		"CImgList<T>");
	
	ParameteredObject::_addParameter< float >(
		gradientSigma, "gradientSigma",
		"Gaussian variance for gradient estimation.",
		1, "float");
	ParameteredObject::_addParameter< float >(
		gradientSigma2, "gradientSigma2",
		"Gaussian variance for gradient estimation in edge direction.",
		0.5, "float");
	ParameteredObject::_addParameter< int >(
		gradientFilterWidth, "gradientFilterWidth",
		"Width of the gradient filter.",
		3, "int");
	ParameteredObject::_addParameter< float >(
		structureTensorSigma1, "structureTensorSigma1",
		"Gaussian variance for structure tensor estimation.",
		1, "float");
	ParameteredObject::_addParameter< float >(
		structureTensorSigma2, "structureTensorSigma2",
		"Second gaussian variance for structure tensor estimation. This one"
		"is used, if the filter is weighted for a specific direction. In this"
		"case, sigma2 is used to smooth in gradient direction while sigma1"
		"smooths perpendicular to gradient direction",
		0.5, "float");
	ParameteredObject::_addParameter< int >(
		structureTensorWidth, "structureTensorWidth",
		"Width of the gaussian filter for structure tensor estimation",
		3, "int");
	ParameteredObject::_addParameter< int >(
		scaleSpaces, "scaleSpaces",
		"Numer of scale spaces for which, eigen values are estimated. While"
		"the results of the output slot rely on the given parameters above,"
		"this value can be used to evaluate correlations between eigen values"
		"of different scale spaces.",
		1, "int");
	ParameteredObject::_addParameter(
		evaluateScaleSpaces, "evaluateScaleSpaces",
		"Activate evaluation using the parameter scaleSpaces for different"
		"scale spaces.",
		"bool");
	ParameteredObject::_addParameter(
		useOrientedFilters, "useOrientedFilters",
		"Use oriented filters determined by initial structure tensor values.",
		"bool");

}

template <typename T>
void EigenValuesSchaefer<T>::execute() {

	const int n = this->images().size();
	this->eigenvalues().assign(n);
	this->eigenvaluesMul().assign(n);
	this->eigenvaluesScaleSpaces().assign(n);
	this->eigenvector1().assign(n);
	this->eigenvector2().assign(n);
	this->angle().assign(n);
	
	// iterate over cimglist
	//#pragma omp parallel for
	for( int i = 0; i < n; i++ ){

		int scales = 1;
		if( this->evaluateScaleSpaces() )
			scales = std::max(1, this->scaleSpaces() );

		const CImg<float> &img = this->images()[i];
		CImg<float> st, eVal, eValRatio, eVec1, eVec2, dx, dy, dxy1, dxy2;
		
		const int w = img.width();
		const int h = img.height();
		const int spectrum = img.spectrum();

		// iterate over scale spaces
		for( int j = 0; j < scales; j++ ){

			sout << j << " --------------------------------" << std::endl;

			
			int width_g = this->gradientFilterWidth();// * (j+1);
			float sigma_g = this->gradientSigma();// * (j+1);
			float sigma_g2 = this->gradientSigma2();// * (j+1);
			float sigma_st1 = this->structureTensorSigma1() + (j*1.);
			float sigma_st2 = this->structureTensorSigma2() + (j*1.);
			int ww = (this->structureTensorSigma1() + (j*1.)) *3 + 2;
			if(ww % 2 == 0) ww++;
			int width_st = ww;

			sout << "sigma1 = " << sigma_st1 << " / sigma2 = " << sigma_st2 << " width: " << ww << std::endl;
		
			getEigenEntitiesSchaefer(img, st, eVal, eValRatio, eVec1, 
				eVec2, dx, dy, dxy1, dxy2, width_g, sigma_g, sigma_g2, width_st, sigma_st1, sigma_st2, false, this->useOrientedFilters());

			sout << "scale " << j << ": " << eVal.max() << std::endl;
		
			// save
			this->eigenvalues()[i].append(eVal,'z');
			this->eigenvaluesMul()[i].append(eVal.get_slice(0).mul(eVal.get_slice(1)),'z');
			if( j == 0 ){
				this->eigenvector1()[i].append(eVec1,'z');
				this->eigenvector2()[i].append(eVec2,'z');
				this->angle()[i].append(eValRatio,'z');
			}

		}


		// evaluation output via cimg display
		if( this->evaluateScaleSpaces() && scales > 1 ){

			CImg<float> maxEV(w,h,1,spectrum,0);


			// Draws the scale space of maximal eigen values for each pixel
			this->eigenvaluesScaleSpaces()[i].assign(w,h,2,spectrum,0);
			#pragma omp parallel for
			for(int j = 0; j < w * h; j++ ){
			
				const int x = j % w;
				const int y = j / w;
			
				// channels
				for( int c = 0; c < spectrum; c++ ){

					float max = this->eigenvalues()[i](x,y,0,c);

					// scales
					for( int z = 1; z < scales; z++ ){
						if( this->eigenvalues()[i](x,y,z,c) > max ){
							max = this->eigenvalues()[i](x,y,z,c);
							this->eigenvaluesScaleSpaces()[i](x,y,0,c) = float (z) / scales * 255;
						}
					}

					maxEV(x,y,1,c) = max;							

				}
			
			}

			/*
			cimg_forC(maxEV,c){	
				float min = maxEV.get_channel(c).min();
				cimg_forXY(maxEV, x,y){	
					float max = maxEV(x,y,1,c);
					maxEV(x,y,1,c) = std::log(std::max(min,max));
				}
			}

			maxEV.normalize(0,255);
			this->eigenvaluesScaleSpaces()[i].append(maxEV,'z');

			*/

			//this->eigenvaluesScaleSpaces()[i].normalize(0,255);

			//imglist[0].save("imlist[0].png");
			//imglist[0].get_channel(0).save("imlist[0].get_channel(0).png");
			//imglist[0].get_channel(1).save("imlist[0].get_channel(1).png");
			//imglist[0].get_channel(2).save("imlist[0].get_channel(2).png");

			const int w = img.width();
			const int h = img.height();
						
		    CImgDisplay disp3(w,h);
		    CImgDisplay disp4(400,400);
		    CImgDisplay disp(w,h);
			CImgDisplay disp2(400,400);
			disp.move( disp3.window_x()+disp3.window_width() +50, disp3.window_y() );
			disp2.move( disp.window_x()+disp.window_width() +50, disp.window_y() );
			disp4.move( disp2.window_x()+disp2.window_width() +50, disp2.window_y() );
			img.display(disp3);
			
			CImg<float> stats(400,400,1,1,0),stats2(400,400,1,3,0),stats3(400,400,1,3,0);

			int img_slice = 0;
			int img_index = 0;
			int img_channel = 0;
			//disp.display( imglist[img_index].get_slice(img_slice) );
			
			int _x = 0;
			int _y = 0;



			//CImg<unsigned char> visu(500,400,1,3,0);
			//const unsigned char red[] = { 255,0,0 }, green[] = { 0,255,0 }, blue[] = { 0,0,255 };
			//
			//CImgDisplay main_disp(img,"Click a point"), draw_disp(visu,"Intensity profile");
			//while (!main_disp.is_closed() && !draw_disp.is_closed()) {
			//	main_disp.wait();
			//	if (main_disp.button() && main_disp.mouse_y()>=0) {
			//	const int y = main_disp.mouse_y();
			//	visu.fill(0).draw_graph(img.get_crop(0,y,0,0,img.width()-1,y,0,0),red,1,1,0,255,0);
			//	visu.draw_graph(img.get_crop(0,y,0,1,img.width()-1,y,0,1),green,1,1,0,255,0);
			//	visu.draw_graph(img.get_crop(0,y,0,2,img.width()-1,y,0,2),blue,1,1,0,255,0).display(draw_disp);
			//	}
			//}
			
			const unsigned char red[] = { 255,0,0 }, green[] = { 0,255,0 }, blue[] = { 0,0,255 }, white[] = { 255,255,255 };


			while (!disp.is_closed() && !disp2.is_closed() && !disp3.is_closed() && !disp4.is_closed()) {
				disp.wait();

				bool b1=0, b2=0, b3=0, b4=0;
				
				// switch channel
				if(disp.is_key(cimg::keyARROWUP) ||disp2.is_key(cimg::keyARROWUP) ||disp3.is_key(cimg::keyARROWUP) ){
					if( img_channel < this->eigenvalues()[i].spectrum()-1 ) img_channel++;
					else img_channel = 0;
				}
				if(disp.is_key(cimg::keyARROWDOWN) || disp2.is_key(cimg::keyARROWDOWN)|| disp3.is_key(cimg::keyARROWDOWN) ){
					if( img_channel > 0 ) img_channel--;
					else img_channel = this->eigenvalues()[i].spectrum()-1;
				}

				// switch depth
				if(disp.is_key(cimg::keyTAB) || disp2.is_key(cimg::keyTAB) || disp3.is_key(cimg::keyTAB) ){
					if( img_slice < this->eigenvalues()[i].depth()-1 ) img_slice++;
					else img_slice = 0;
					b3=true;
				}
				// switch depth
				//const unsigned int key_seq[] = { cimg::keyCTRLLEFT, cimg::keyTAB };
				//if(disp.is_key_sequence(key_seq,2) || disp2.is_key_sequence(key_seq,2)){
				//	if( img_slice > 0  ) img_slice--;
				//	else img_slice = imglist[img_index].depth()-1;
				//	b3=true;
				//}
				

				// check mouse key event
				//if (disp.button()&1) { // Left button clicked.
				//	_x = disp.mouse_x();
				//	_y = disp.mouse_y();
				//	sout << "Mouse hit: " << _x << " / " << _y << std::endl;
				//	ImgTool::printInfo(sout, stats, "stats");
				//	//(/disp2.flush();
				//}

				// replace image according to new parameters
				std::stringstream s;
				s << "Image : ";
				s << " Slice: " << img_slice/2 ;
				s << " Channel: " << img_channel ;
				//s << " : channel " << img_channel;
				
				//sout << "title: " << s.str() << std::endl;
				//sout << "img_index: " << img_index << " / img_slice: " << img_slice << std::endl;
				//if(b1) sout << "bt1" << std::endl;
				//if(b2) sout << "bt2" << std::endl;
				//if(b3) sout << "bt3" << std::endl;
				//sout << "---" << std::endl;
				
				_x = disp.mouse_x();
				_y = disp.mouse_y();
				
				stats2.fill(0);
				for( int c = 0; c < spectrum; c++ ){
					CImg<float> hist = img.get_crop(_x-5,_y-5,0,0,_x+5,_y+5,0,c).histogram(5,0,255);
					if( c == 1 )
						stats2.draw_graph(hist,green,1,1,0,255,0);
					else if(c==2)
						stats2.draw_graph(hist,blue,1,1,0,255,0);
					else
						stats2.draw_graph(hist,red,1,1,0,255,0);
				}
				stats2.display(disp4);
				
				const float maxVal = this->eigenvalues()[i].max();

				int subdivisionX = stats3.width() / this->scaleSpaces();

				getStatistics(stats,this->eigenvalues()[i],_x,_y,img_channel);
				stats3.fill(0).draw_axes(0,this->scaleSpaces(),maxVal,0,white,0.5,10);//subdivisionX);
				if( img_channel == 0 ) stats3.draw_graph(stats,red,1,3,0,maxVal);//,0,255,0);
				else if( img_channel == 1 ) stats3.draw_graph(stats,green,1,3,0,maxVal);//,0,255,0);
				else stats3.draw_graph(stats,blue,1,3,0,maxVal);//,0,255,0);
				stats3.display(disp2);//,0,255,0);

				disp.display( this->eigenvalues()[i].get_slice(img_slice).channel(img_channel).normalize(0,255) );
				disp.resize(this->eigenvalues()[i]);
				disp.set_title("%s",s.str().c_str());
				disp.move( disp3.window_x()+disp3.window_width() +50, disp3.window_y() );
				
				//disp2.display( stats.normalize(0,255) );
				//stats.display_graph(disp2,3);
				disp2.move( disp.window_x()+disp.window_width()+50, disp.window_y() );
				disp4.move( disp2.window_x()+disp2.window_width() +50, disp2.window_y() );
	
				//disp2.wait();
			}
			
			if(!disp.is_closed())
				disp.close();
			if(!disp2.is_closed())
				disp2.close();

		}

	}

}

/// Get histogram for current pixel along the z axis
void getStatistics(CImg<float> &out, const CImg<float> &img, const int _x, const int _y, const int channel = -1){
	
	int x = std::min( std::max( 0, _x ), img.width()-1 );
	int y = std::min( std::max( 0, _y ), img.height()-1 );

	out.assign();
	//for( int i = 0; i < img.spectrum(); i++ ){
		CImg<float> temp = img.get_channel(channel);
		out.append( temp.crop(x,y,0,x,y,img.depth()-1).permute_axes("zyxc"),'c');
		//sout << "Channel[" << channel << ":" ;
		//for(int j = 0 ; j < img.depth(); j++ ){
		//	sout << out(j,0,0,channel) << " / ";
		//}
		//sout << std::endl;
	//}
	
}


/// calculate derivatives, if hue is set, the gradient of the hue channel is calculated
/** 
	dx and dy are the derivatives in x and y direction 
*/
void get_gradients( const CImg<float> &image, 
			   const unsigned gwidth, const float sigma,
			   CImg<float> &dx, CImg<float> &dy, const bool hue = false )
{ 
		
	const CImg<float> color(1,1,1,1,1);
	CImg<float> gaussX;		// gaus for smooting in X 
	CImg<float> dgaussX;	// gauss for derivatives in X 

	float temp = 0;
		
	const int offset = gwidth / 2. - 0.5;

	// create smoothing and derivative kernels in x direction
	gaussX.assign(gwidth,1,1,1,0);
	gaussX.draw_gaussian(offset,0,sigma,&color(0,0,0));
	gaussX /= gaussX.sum();

	// Create derivative gaussian in x direction
	dgaussX = gaussX;
	cimg_forX(dgaussX, x){
		if( x - gwidth / 2. + 0.5 ) // do for all but the center pixel
			dgaussX(x) /= -( x - gwidth / 2. + 0.5 );
		else
			dgaussX(x) = 0; // set center pixel to 0
		temp += dgaussX(x)*(-x);
	}
	dgaussX /= temp;

		
	dx.assign(image); 
	dy.assign(image); 

	dx.convolve(gaussX.get_permute_axes("yxzc"));
	dx.convolve(dgaussX);

	dy.convolve(gaussX);
	dy.convolve(-dgaussX.mirror('y').get_permute_axes("yxzc"));

	// gradients for [0;x) channels
	if(hue){

		sout << "Checking hue gradient for correct differences" << std::endl;

		sout << "dx.max() = " << dx.max() 
		 << "\ndy.max() = " << dy.max()  << std::endl;

		int count = 0;
			
		#pragma omp parallel for
		for( int xy = 0; xy < dx.width()*dx.height(); xy++ ){
	
			const int x = xy % dx.width();
			const int y = xy / dx.width();

			if( std::abs(dx(x,y)) > 180  ){
				int val = std::abs(dx(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				dx(x,y) = sgn * diff;
				count++;
			}
		
			if( dy(x,y) > 180 || dy(x,y) < -180 ){
				int val = std::abs(dy(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				count++;
				dy(x,y) = sgn * diff;
			}
		
		}

		sout << "count = " << count << std::endl;

	}

}


/// calculate derivatives, if hue is set, the gradient of the hue channel is calculated
/** 
	dx and dy are the derivatives in x and y direction while dxy1 is the derivative in lower right direction
	and dy2 is the derivative in upper right direction
*/
void get_gradients( const CImg<float> &image, 
			   const unsigned gwidth, const float sigma, const float sigma2, 
			   CImg<float> &dx, CImg<float> &dy, CImg<float> &dxy1, CImg<float> &dxy2, const bool hue = false )
{ 

	// get normal gradients
	get_gradients( image, gwidth, sigma, dx, dy, hue );
	
	const CImg<float> color(1,1,1,1,1);
	CImg<float> gaussXY1,gaussXY2;		// gaus for smooting in X and XY direction
	CImg<float> dgaussXY1,dgaussXY2;	// gauss for derivatives in X and XY direction
	CImg<float> gaussX;		// gaus for smooting in X 
	CImg<float> dgaussX;	// gauss for derivatives in X 

	float temp = 0;
		
	const int offset = gwidth / 2. - 0.5;

	// create smoothing and derivative kernels in x direction
	gaussX.assign(gwidth,1,1,1,0);
	gaussX.draw_gaussian(offset,0,sigma,&color(0,0,0));
	gaussX /= gaussX.sum();

	// Create derivative gaussian in x direction
	dgaussX = gaussX;
	cimg_forX(dgaussX, x){
		if( x - gwidth / 2. + 0.5 ) // do for all but the center pixel
			dgaussX(x) /= -( x - gwidth / 2. + 0.5 );
		else
			dgaussX(x) = 0; // set center pixel to 0
		temp += dgaussX(x)*(-x);
	}
	dgaussX /= temp;
		

	/*
		gaussXY1  --> smooths in upper right direction 
		gaussXY2  --> smooths in lower right direction
		dgaussXY1 --> derivative in lower right direction
		dgaussXY2 --> derivative in upper right direction
	*/
	
	// Get diagonal gaussian by covariance matrix
	float s1 = sigma;
	float s2 = sigma2; // variance in gradient direction
	float c = std::cos( PI / 4 ) * std::cos( PI / 4 );
	CImg<float> tensor1(2,2,1,1,0);
	tensor1(0,0) = s1*c + s2*c;
	tensor1(1,0) = s1*c-s2*c;
	tensor1(0,1) = s1*c-s2*c;
	tensor1(1,1) = s1*c + s2*c;

	gaussXY1.assign(gwidth,gwidth,1,1,0);
	gaussXY2.assign(gwidth,gwidth,1,1,0);

	gaussXY1.draw_gaussian(offset,offset,tensor1,&color(0,0,0));
	gaussXY2.draw_gaussian(offset,offset,tensor1.get_mirror('y'),&color(0,0,0));
	gaussXY1 /= gaussXY1.sum();
	gaussXY2 /= gaussXY2.sum();
				
	// Copy values of dgaussX to dgaussXY
	dgaussXY1.assign(gwidth,gwidth,1,1,0);
	dgaussXY2.assign(gwidth,gwidth,1,1,0);
	cimg_forX(dgaussXY1,x){
		const int y_xy2 = gwidth - x - 1 ; // y coordinate for dgaussXY2
		dgaussXY1(x,x) = dgaussX(x,0);
		dgaussXY2(x,y_xy2) = dgaussX(x,0);
	}
		
	dxy1.assign(image); 
	dxy2.assign(image); 

	dxy1.convolve(gaussXY1);
	dxy1.convolve(dgaussXY1);

	dxy2.convolve(gaussXY2);
	dxy2.convolve(dgaussXY2);

	// gradients for [0;x) channels
	if(hue){

		sout << "Checking hue gradient for correct differences" << std::endl;

		sout << "\ndxy1.max() = " << dxy1.max() 
		 << "\ndxy2.max() = " << dxy2.max() << std::endl;

		int count = 0;
			
		#pragma omp parallel for
		for( int xy = 0; xy < dx.width()*dx.height(); xy++ ){
	
			const int x = xy % dx.width();
			const int y = xy / dx.width();

			if( dxy1(x,y) > 180 || dxy1(x,y) < -180 ){
				int val = std::abs(dxy1(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				count++;
				dxy1(x,y) = sgn * diff;
			}
		
			if( dxy2(x,y) > 180 || dxy2(x,y) < -180 ){
				int val = std::abs(dxy2(x,y));
				int sgn = ( val < 0 ? -1 : 1 );
				int res = int(val) % 360;
				int diff = std::abs( 180 - res );
				count++;
				dxy2(x,y) = sgn * diff;
			}

		}

		sout << "count = " << count << std::endl;

	}

}


/// Calculate structure tensor using an optional vector field for gauss orientation
/** 
	If weakenSecondAxis is activated, the sigma of gaussian for second axis 'dy' is reduced.
	In this case, edges in dx direction are privileged.
	If assumeXYAxis is activated, dx and dy are assumed to be dxy1 and dxy2 which means 
	a gradient from upper left to lower right and one from lower left to upper right
*/
int structureTensor( const CImg<float> &dx, const CImg<float> &dy, 
					 const int gwidth, const float sigma1,  const float sigma2, 
					 CImg<float> &st, const float angle = 0, const std::string debug = "0" )
{
	CImg<float> color(1,1,1,1,1);

	const float c = std::cos( angle ), s = std::sin( angle );
	const float cc = c*c, ss = s*s, cs=c*s;
	const float s1 = sigma1, s2 = sigma2;
	CImg<float> tensor1(2,2,1,1,0);
	tensor1(0,0) = s1*cc + s2*ss;
	tensor1(1,0) = s1*cs-s2*cs;
	tensor1(0,1) = s1*cs-s2*cs;
	tensor1(1,1) = s1*ss + s2*cc;
		
	const int offset = gwidth / 2. - 0.5;
	CImg<float> gauss(gwidth,gwidth,1,1,0);
	gauss.draw_gaussian(offset,offset,tensor1,&color(0,0,0));
	gauss /= gauss.sum();

	/*
	if( debug != "0" ){
		sout << "ST_" << debug << std::endl;
		gauss.get_normalize(0,255).save(("gauss_"+debug+".png").c_str());
		ImgTool::printInfo( sout, tensor1, "tensor");
	}
	*/

	// define structure tensor
	st.clear();
	st.assign(dx.get_pow(2).convolve(gauss));
	st.append(
		dy.get_pow(2).convolve(gauss),'z');
	st.append(
		dx.get_mul(dy).convolve(gauss),'z');

	return 0;
}

/// Get eigen vectors for structure tensor.
/** Also calculates a ratio between eigen values and considers an optional angle to rotate eigen vectors.
	This is useful if the structure tensor bases on
	diagonal gradients. After this rotation, eigen vectors can be compared to those
	which are based on horizontal and vertical gradients
*/
int eigenv( const CImg<float> &st, CImg<float>& eval, CImg<float> &evec1, CImg<float> &evec2, CImg<float> &evalRatio, const float angle = 0, const std::string debug = "" )
{

	// calculate eigen values and eigenvectors
	eval.assign(st.width(),st.height(),2,1,0);
	evec1.assign(st.width(),st.height(),2,1,0);
	evec2.assign(st.width(),st.height(),2,1,0);
	evalRatio.assign(st.width(),st.height(),2,1,0);
	
	#pragma omp parallel for
	for( int xy = 0; xy < st.width()*st.height(); xy++ ){
	
		const int x = xy % st.width();
		const int y = xy / st.width();
	
		CImg<float> stemp(2,2);
		CImg<float> eigval_t;
		CImg<float> eigvec_t;

		stemp(0,0)=st(x,y,0);
		stemp(1,0)=st(x,y,2);
		stemp(0,1)=st(x,y,2);
		stemp(1,1)=st(x,y,1);
		stemp.symmetric_eigen(eigval_t,eigvec_t);

		eval(x,y,0)=eigval_t(0,0);
		eval(x,y,1)=eigval_t(0,1);

		assert( eigval_t(0,0) >= eigval_t(0,1) );

		//evec1(x,y,0)=   eigvec_t(0,0);
		//evec1(x,y,1)=  -eigvec_t(1,0);
		//evec2(x,y,0)=   eigvec_t(0,1);
		//evec2(x,y,1)=  -eigvec_t(1,1);

		CImg<float> ev1(1,2,1,1), ev2(1,2,1,1);
		ev1(0,0) = eigvec_t(0,0);
		ev1(0,1) = eigvec_t(1,0);
		ev2(0,0) = eigvec_t(0,1);
		ev2(0,1) = eigvec_t(1,1);

		// Rotate eigen vector by angle, this is useful if the structure tensor bases on
		// diagonal gradients. After this rotation, eigen vectors can be compared to those
		// which are based on horizontal and vertical gradients
		// Check if eigen vector != 1 since this implies there is no structure at this pixel.
		if( ev1(0,0) != 1 ){
			CImg<float> rotate(2,2,1,1);
			rotate(0,0) = (cos(angle));
			rotate(1,0) = (-sin(angle));
			rotate(0,1) = (sin(angle));
			rotate(1,1) = (cos(angle));
			ev1 = rotate * ev1;
			ev2 = rotate * ev2;
		}

		// Save eigen vectors
		evec1(x,y,0) = ev1(0,0);
		evec1(x,y,1) = ev1(0,1);
		evec2(x,y,0) = ev2(0,0);
		evec2(x,y,1) = ev2(0,1);

		float length1 = std::sqrt( evec1(x,y,0) * evec1(x,y,0) + evec1(x,y,1) * evec1(x,y,1) );
		float length2 = std::sqrt( evec2(x,y,0) * evec2(x,y,0) + evec2(x,y,1) * evec2(x,y,1) );
		if( length1 < 0.99999 && length1 > 1.00001 )
			sout << std::sqrt( evec1(x,y,0) * evec1(x,y,0) + evec1(x,y,1) * evec1(x,y,1) ) << std::endl;
		if( length2 < 0.99999 && length2 > 1.00001 )
			sout << std::sqrt( evec2(x,y,0) * evec2(x,y,0) + evec2(x,y,1) * evec2(x,y,1) ) << std::endl;

		//if( ( eval(x,y,0) + eval(x,y,1) ) != 0 )
		//	evalRatio(x,y) = std::pow( float( eval(x,y,0) - eval(x,y,1))  / float( eval(x,y,0) + eval(x,y,1) ) , float(2.0) );
		//evalRatio(x,y) = acos( evec1(x,y,0) );

		// Get angle and normalize it to positive angle
		// (only the absolute angle is of importance)		
		float vx = evec1(x,y,0), vy = evec1(x,y,1);

		float ang = acos( vx ) / PI * 180;
		if( vy <= 0 ) ang = abs(ang);
		else ang = - abs(ang);
		if( ang < 0 ) ang = 180 + ang;

		float ang2 = asin( vy ) / PI * 180;
		if( vx < 0 ) ang2 = -ang2;
		
		evalRatio(x,y,0) = ang;
		evalRatio(x,y,1) = ang2;

		//// DEBUG: Draw line in eigen vector 1 direction (perpendicular to edge)
		//if(x==23 & y==22){
		//	int ex = x+100*vx, ey = y+100*vy;
		//	sout << ex << "/" << ey << std::endl;
		//
		//	sout << "Debug angle: "<< vx << "," << vy << ": " << ang << ": " << ang2 << std::endl;
		//	
		//	CImg<float> color(1,1,1,1,1);
		//	CImg<float> deb(st.width(),st.height(),1,1,0);
		//	deb.draw_line(x,y,ex,ey,&color(0,0,0));
		//	deb.normalize(0,255).save(("debug"+debug+".png").c_str());
		//}

	}

	return 0;
}


/// Get eigen values and vectors for each channel using Schaefer Implementation
/** gaussWidth for gradients and structure tensor. Sigma 1 for gradients, sigma 2 for structure tensor */
void getEigenEntitiesSchaefer(const CImg<float> &img, CImg<float> &st, CImg<float> &eVal, CImg<float> &eValRatio, CImg<float> &eVec1, 
			CImg<float> &eVec2, CImg<float> &dx, CImg<float> &dy, CImg<float> &dxy1, CImg<float> &dxy2, const int gaussWidth, 
			const float sigma1, const float sigma2, const int gaussWidthST, const float sigma_st1, const float sigma_st2, const bool hsv, const bool orientated = false){
				
	const int w = img.width();
	const int h = img.height();
	const int d = img.spectrum();

	sout << "Calculate eigen values and vectors for " << d << (hsv ? " hsv" : " rgb" ) << " channels" << std::endl;
	
	st.assign(w,h,3,d,1);
	eVal.assign(w,h,2,d,1);		// save first and second component in z-axis
	eVec1.assign(w,h,2,d,1);	// save first and second component in z-axis
	eVec2.assign(w,h,2,d,1);	// save first and second component in z-axis
	dx.assign(w,h,1,d,1);
	dy.assign(w,h,1,d,1);
	dxy1.assign(w,h,1,d,1);
	dxy2.assign(w,h,1,d,1);
	eValRatio.assign(w,h,2,d,1);

	// iterate over colour channels
	for( int i = 0; i < d; i++ ){

		sout << "Channel " << (hsv?"hsv":"rgb") << i << " ----------------------------" << std::endl;
		
		std::stringstream string;
		if(hsv) string << "hsv_";
		else string << "rgb_";
		string << "channel_" << i << "_";
		// img.get_channel(i).get_normalize(0,255).save((string.str()+"img.png").c_str());

		// Eigenimplementation
		CImg<float> _dx(w,h,1,1), _dy(w,h,1,1), _dxy1(w,h,1,1), _dxy2(w,h,1,1), _st(w,h), _stX(w,h), _stY(w,h), _stXY1(w,h), _stXY2(w,h);
		CImg<float> _eval(w,h,2,d,0), _evec1(w,h,2,d,0), _evec2(w,h,2,d,0), _evalRatio(w,h,2,d,0);
		CImg<float> _evalMAX , _angle , angleHSV , _eVecMean1, _eVecMean2;

		// get gradients of channel i
		bool _h = ( i == 0 && hsv ? true : false );
				
		if( !orientated ){
			get_gradients(img.get_channel(i), gaussWidth, sigma1, _dx, _dy, _h);
		
			// get structure tensor
			structureTensor( _dx, _dy, gaussWidthST, sigma_st1, sigma_st1, _st, 0, "1");	// symmetric gaussian

			// get eigen values / vectors
			eigenv( _st, _eval, _evec1, _evec2, _evalRatio, 0, (string.str()+"0").c_str());

			_evalMAX = _eval;
			_angle = _evalRatio;
			_eVecMean1 = _evec1;
			_eVecMean2 = _evec2;

		}
		else{

			get_gradients(img.get_channel(i), gaussWidth, sigma1, sigma2, _dx, _dy, _dxy1, _dxy2, _h);
		
			// get structure tensor
			structureTensor( _dx, _dy, gaussWidthST, sigma_st1, sigma_st1, _st, 0, "1");	// symmetric gaussian

			// get eigen values / vectors
			eigenv( _st, _eval, _evec1, _evec2, _evalRatio, 0, (string.str()+"0").c_str());
		
			// Eigenimplementation
			CImg<float> _stX(w,h), _stY(w,h), _stXY1(w,h), _stXY2(w,h);
			// save first and second component in z-axis
			CImg<float> _evalX(w,h,2,d,0), _evec1X(w,h,2,d,0), _evec2X(w,h,2,d,0), _evalRatioX(w,h,2,d,0); 
			CImg<float> _evalY(w,h,2,d,0), _evec1Y(w,h,2,d,0), _evec2Y(w,h,2,d,0), _evalRatioY(w,h,2,d,0);
			CImg<float> _evalXY1(w,h,2,d,0), _evec1XY1(w,h,2,d,0), _evec2XY1(w,h,2,d,0), _evalRatioXY1(w,h,2,d,0);
			CImg<float> _evalXY2(w,h,2,d,0), _evec1XY2(w,h,2,d,0), _evec2XY2(w,h,2,d,0), _evalRatioXY2(w,h,2,d,0);

			structureTensor( _dx, _dy, gaussWidthST, sigma_st1, sigma_st2, _stX, 0);	// gaussian with smaller y smoothing
			structureTensor( _dx, _dy, gaussWidthST, sigma_st2, sigma_st1, _stY, 0);	// gaussian with smaller x smoothing
			structureTensor( _dxy1, _dxy2, gaussWidthST, sigma_st1, sigma_st2, _stXY1, PI / 4.);	// gaussian  45 rotated
			structureTensor( _dxy1, _dxy2, gaussWidthST, sigma_st2, sigma_st1, _stXY2, PI / 4.);	// gaussian -45 rotated

			eigenv( _stX, _evalX, _evec1X, _evec2X, _evalRatioX, 0, (string.str()+"1").c_str());
			eigenv( _stY, _evalY, _evec1Y, _evec2Y, _evalRatioY, 0, (string.str()+"2").c_str());
			eigenv( _stXY1, _evalXY1, _evec1XY1, _evec2XY1, _evalRatioXY1, PI / 4., (string.str()+"3").c_str());
			eigenv( _stXY2, _evalXY2, _evec1XY2, _evec2XY2, _evalRatioXY2, PI / 4., (string.str()+"4").c_str());
		
			//  _evalX.get_normalize(0,255).save((string.str()+"evalX.png").c_str());
			//  _evalY.get_normalize(0,255).save((string.str()+"evalY.png").c_str());
			//  _evalXY1.get_normalize(0,255).save((string.str()+"evalXY1.png").c_str());
			//  _evalXY2.get_normalize(0,255).save((string.str()+"evalXY2.png").c_str());
			//  
			//  
			//  // Build sums of evalX and evalY / evalXY1 and evalXY2
			//  // This is build sums of eigen value images which base on the same base vectors (horizontal/vertical or diagonal)
			//  (_evalX+_evalY).normalize(0,255).save((string.str()+"evalXandY.png").c_str());
			//  (_evalXY1+_evalXY2).normalize(0,255).save((string.str()+"evalXY1and2.png").c_str());
			//  
			//  // Build sum of all eigen value images		
			//  (_evalXY1+_evalXY2+_evalX+_evalY).normalize(0,255).save((string.str()+"evalALL.png").c_str());
			//  
			//  // Take maximum from all available eigen value images
			_evalMAX = _eval.get_max(_evalX).max(_evalY).max(_evalXY1).max(_evalXY2);
			//  _evalMAX.get_normalize(0,255).save((string.str()+"evalMAX.png").c_str());

			// Get mean angle
			_angle = ( _evalRatioX + _evalRatioY + _evalRatioXY1 + _evalRatioXY2 + _evalRatio ) / 5.;

			angleHSV = _angle.get_normalize(0,359);
			angleHSV.resize(w,h,1,3).HSVtoRGB();
			angleHSV.save((string.str()+"angle.png").c_str());

			// Get mean vectors
			_eVecMean1, _eVecMean2;
			_eVecMean1 = ( _evec1X + _evec1Y + _evec1XY1 + _evec1XY2 + _evec1 ) / 5.;
			_eVecMean2 = ( _evec2X + _evec2Y + _evec2XY1 + _evec2XY2 + _evec2 ) / 5.;

		}

		// save images
		#pragma omp parallel for
		for( int xy = 0; xy < st.width()*st.height(); xy++ ){
	
			const int x = xy % st.width();
			const int y = xy / st.width();

			st(x,y,0,i) = _st(x,y,0);
			st(x,y,1,i) = _st(x,y,1);
			st(x,y,2,i) = _st(x,y,2);

			dx(x,y,0,i) = _dx(x,y);
			dy(x,y,0,i) = _dy(x,y);

			dxy1(x,y,0,i) = _dxy1(x,y);
			dxy2(x,y,0,i) = _dxy2(x,y);
			
			eValRatio(x,y,0,i) = _angle(x,y,0);
			eValRatio(x,y,1,i) = _angle(x,y,1);
			
			//eVal(x,y,0,i) = _evalMAX(x,y,0);
			//eVal(x,y,1,i) = _evalMAX(x,y,1);
			eVal(x,y,0,i) = _eval(x,y,0);
			eVal(x,y,1,i) = _eval(x,y,1);

			eVec1(x,y,0,i) = _eVecMean1(x,y,0);
			eVec1(x,y,1,i) = _eVecMean1(x,y,1);
			eVec2(x,y,0,i) = _eVecMean2(x,y,0);
			eVec2(x,y,1,i) = _eVecMean2(x,y,1);
		}

	}

}



#endif /* _EIGENVALUESSCHAEFER_HXX_ */
