/*  Copyright (C) 2009 SasiKanth

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file epipolarrectification.hxx
 *  Implementation of the parameter class epipolarrectification.
 *  \author SasiKanth
 *  \date 18.06.2010
 */

#ifndef _EPIPOLARRECTIFICATION_HXX_
#define _EPIPOLARRECTIFICATION_HXX_

#include <hekate/EpipolarRectification.h>
#include <charon-utils/ImgTool.hxx>

#include <levmar/levmar.h>

template <typename T>
epipolarrectification<T>::epipolarrectification(const std::string& name) :
		TemplatedParameteredObject<T>(
				"epipolarrectification", name,
				"This plugin warps the images taking the fundamental matix as the input so that, the epipolar lines in both the cases become horizontal and Images are effectively rectified (Hartley pg. 304-307)"
		)
			,Input_Images(false, false) 
			,fmatrix(false, false) 
			
			
{
	ParameteredObject::_addParameter (inlier_confidence, "inlier_confidence", "The threshold for confident inlier determination", 0.005);
	

	

	ParameteredObject::_addInputSlot(Input_Images, "Input_Images", 
	"The set of input images", 
	"CImgList<T>"); 
	ParameteredObject::_addInputSlot(fmatrix, "fmatrix", 
	"The fundamental matrix", 
	"CImgList<T>"); 
	ParameteredObject::_addInputSlot(correspondences_input, "correspondences_input", 
	"The input correspondences", 
	"std::vector<std::vector<std::pair<T, T> > >"); 
	ParameteredObject::_addOutputSlot(Warping_Matrices, "Warping_Matrices", 
	"The warping matrices for the two images", 
	"CImgList<T>"); 
	
}

template <typename T>
void epipolarrectification<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	// Beginning of plugin code

	// Extracting the epipoles from the fundamental matrix

	cimg_library::CImg<T> epipole_left(1,3), epipole_right(1,3), epipole_right_translated(1,3), U, S, V, f_matt(3,3);

	f_matt = fmatrix()[0].get_transpose();

	// Left Epipole

	fmatrix()[0].SVD(U,S,V);

	epipole_left(0,0) = V(2,0)/V(2,2);
	epipole_left(0,1) = V(2,1)/V(2,2);
	epipole_left(0,2) = 1;

	//Right Epipole

	f_matt.SVD(U,S,V);

	epipole_right(0,0) = V(2,0)/V(2,2);
	epipole_right(0,1) = V(2,1)/V(2,2);
	epipole_right(0,2) = 1;

	// Finding the homography of the right image, H'

	cimg_library::CImg<T> G(3,3), R(3,3), Tr(3,3), H_right(3,3), H_left(3,3);

	// Generating a translation matrix which maps the point of interest to the origin (Hartley pg. 305)

	Tr(0,0) = 1;		Tr(1,0) = 0;		Tr(2,0) = -Input_Images()[1].width()/2;
	Tr(0,1) = 0;		Tr(1,1) = 1;		Tr(2,1) = -Input_Images()[1].height()/2;
	Tr(0,2) = 0;		Tr(1,2) = 0;		Tr(2,2) = 1;

	epipole_right_translated = Tr*epipole_right;

	// Generating the rotation matrix (Hartley pg. 305)

	double theta, f;

	theta = atan(-(double)epipole_right_translated(1)/(double)epipole_right_translated(0));

	R(0,0) = cos(theta);		R(1,0) = -sin(theta);		R(2,0) = 0;
	R(0,1) = sin(theta);		R(1,1) = cos(theta);		R(2,1) = 0;
	R(0,2) = 0;					R(1,2) = 0;					R(2,2) = 1;

	// Generating the G matrix

	f = epipole_right_translated(0)*cos(theta) - epipole_right_translated(1)*sin(theta);

	G(0,0) = 1;			G(1,0) = 0;		G(2,0) = 0;
	G(0,1) = 0;			G(1,1) = 1;		G(2,1) = 0;
	G(0,2) = -1/f;		G(1,2) = 0;		G(2,2) = 1;

	// Generating the composite right homography

	H_right = G*R;
	H_right = H_right*Tr;

	// Computation of the left homography begins here

	// Computation of the M matrix (refer to http://www.robots.ox.ac.uk/~vgg/hzbook/hzbook2/clarification_rectification.pdf)

	cimg_library::CImg<T> epipole_mat(3,3), M(3,3), Ho(3,3), S_mat(3,3), Z(3,3);

	fmatrix()[0].SVD(U,S,V);
	
	S(2) = (S(1)+S(0))/2;

	S_mat(0,0) = S(0);		S_mat(1,0) = 0;		S_mat(2,0) = 0;
	S_mat(0,1) = 0;			S_mat(1,1) = S(1);	S_mat(2,1) = 0;
	S_mat(0,2) = 0;			S_mat(1,2) = 0;		S_mat(2,2) = S(2);

	Z(0,0) = 0;				Z(1,0) = -1;		Z(2,0) = 0;
	Z(0,1) = 1;				Z(1,1) = 0;			Z(2,1) = 0;
	Z(0,2) = 0;				Z(1,2) = 0;			Z(2,2) = 1;

	V.transpose();

	M = U*Z;
	M = M*S_mat;
	M = M*V;

	Ho = H_right*M;

	// Determining the values of a,b and c

	const int m=correspondences_input()[0].size(), n=3;
	int i, *detail_array;
	double *data;
	data = new double[4*m];
	detail_array = new int[m];

	//Transforming and loading values of correspondences into array

	for(i=0; i<m; i++)
		{
			cimg_library::CImg<T> input_cor(1,3), output_cor(1,3);

			input_cor(0,0) = correspondences_input()[0][i].first;
			input_cor(0,1) = correspondences_input()[0][i].second;
			input_cor(0,2) = 1;

			output_cor = Ho*input_cor;

			data[4*i] = output_cor(0,0)/output_cor(0,2);
			data[4*i+1] = output_cor(0,1)/output_cor(0,2);

			input_cor(0,0) = correspondences_input()[1][i].first;
			input_cor(0,1) = correspondences_input()[1][i].second;
			input_cor(0,2) = 1;

			output_cor = H_right*input_cor;

			data[4*i+2] = output_cor(0,0)/output_cor(0,2);
			data[4*i+3] = output_cor(0,1)/output_cor(0,2);
		}

	// Finding the least squares estimate of a, b and c (Hartley. pg 589 Algorithm A5.1)

	// Finding the high-confidence correspondences

	for(i=0; i<m; i++)
	{
		cimg_library::CImg<double> num1(1), den1(1), den2(1), den3(1), den4(1), temp1(3,1), temp2(1,3), xs_temp(1,3), xt_temp(1,3), xtt_temp(3,1), f_tempt(3,3);
		double error_value;
		
		xs_temp(0,0) = correspondences_input()[0][i].first;
		xs_temp(0,1) = correspondences_input()[0][i].second;
		xs_temp(0,2) = 1;

		xt_temp(0,0) = correspondences_input()[1][i].first;
		xt_temp(0,1) = correspondences_input()[1][i].second;
		xt_temp(0,2) = 1;

		//Energy calculation and storing values of x

			f_tempt = fmatrix()[0].get_transpose();
			xtt_temp = xt_temp.get_transpose();

			temp1 = xtt_temp*fmatrix()[0];
			num1 = (temp1*xs_temp);
			num1 = num1(0)*num1(0);

			temp1 = (fmatrix()[0]*xs_temp);
			den1 = temp1(0)*temp1(0);
			den2 = temp1(1)*temp1(1);

			temp2 = (f_tempt*xt_temp);
			den3 = temp2(0)*temp2(0);
			den4 = temp2(1)*temp2(1);

			error_value = ((double)num1(0)/(((double)den1(0)+(double)den2(0)+(double)den3(0)+(double)den4(0))));

			if(error_value < inlier_confidence)
			{
				detail_array[i] = 1;
			}
			else
			{
				detail_array[i] = 0;
			}
	}

	// Calculating the number of high-confidence correspondences
	
	int sum=0;
	for(i=0; i<m; i++)
		sum = sum + detail_array[i];

	// Loading only the high-confidence correspondences into the matrices

	cimg_library::CImg<T> A(n,sum), B(1,sum), Ut(sum,sum), B_dash(1,sum), Y(1,n), P(1,n);
	int j=0;

	for(i=0; i<m; i++)
	{
		if(detail_array[i] == 1)
		{
			A(0,j) = data[4*i];
			A(1,j) = data[4*i+1];
			A(2,j) = 1;

			B(0,j) = data[4*i+2];

			j++;
		}
	}

	A.SVD(U,S,V);
	
	Ut = U.get_transpose();

	B_dash = Ut*B;

	for(i=0; i<n; i++)
	{
		Y(0,i) = B_dash(0,i)/S(0,i);
	}

	P = V*Y;

	// Generating the composite left homography

	cimg_library::CImg<T> Ha(3,3);

	Ha(0,0) = P(0,0);		Ha(1,0) = P(0,1);		Ha(2,0) = P(0,2);
	Ha(0,1) = 0;			Ha(1,1) = 1;			Ha(2,1) = 0;
	Ha(0,2) = 0;			Ha(1,2) = 0;			Ha(2,2) = 1;

	H_left = Ha*H_right;
	H_left = H_left*M;

	Warping_Matrices().push_back(H_left);
	Warping_Matrices().push_back(H_right);

	ImgTool::printInfo(sout,H_left,"H_l");

	ImgTool::printInfo(sout,H_right,"H_r");
}

#endif /* _EPIPOLARRECTIFICATION_HXX_ */



