/*  Copyright (C) 2011 Stephan Meister

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file EstimateCamMatrixOCV.hxx
 *  Implementation of the parameter class EstimateCamMatrixOCV.
 *  \author <a href="mailto:stephan.meister@iwr.uni-heidelberg.de">
 *      Stephan Meister</a>
 *  \date 31.10.2011
 */

#ifndef _ESTIMATECAMMATRIXOCV_HXX_
#define _ESTIMATECAMMATRIXOCV_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/EstimateCamMatrixOCV.h>

#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

#include <hekate/openCvUtil.hxx>

#include <boost/smart_ptr.hpp>

template <typename T>
EstimateCamMatrixOCV<T>::EstimateCamMatrixOCV(const std::string& name) :
		TemplatedParameteredObject<T>(
			"EstimateCamMatrixOCV", name,
			"<h2>Estimate internal camera matrix from several images of a"
			"grid pattern using OpenCV</h2><br>"
			"Remark: A Checkerboard with 4x4 tiles will have 3x3 inner corners"
			"set gridRangeX and gridRangeY accordingly<br>"
			"For a circle grid pattern, use the number of circles for each direction"
		),
		initialCamMatrix(true,false)
{
	ParameteredObject::_addOutputSlot(
		camMatrix, "camMatrix",
		"estimated camera matrix, units in pixels",
		"CImgList<T>");
	ParameteredObject::_addOutputSlot(
		distortionCoeffs, "distortionCoeffs",
		"distortion coefficients, values are aligned on CImg x dimension ",
		"CImgList<T>");

	ParameteredObject::_addInputSlot(
		initialCamMatrix, "initialCamMatrix",
		"camera matrix to use as initial guess (optional)",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		gridRangeX, "gridRangeX",
		"number of inner corners in x direction",
		10, "int");
	ParameteredObject::_addParameter< int >(
		gridRangeY, "gridRangeY",
		"number of inner corners in y direction",
		10, "int");
	ParameteredObject::_addParameter< double >(
		gridDist, "gridDist",
		"distance between grid tiles",
		1, "double");
	ParameteredObject::_addParameter(
		gridType, "gridType",
		"type of calibration target",
		"{Checkerboard;Circle Grid}");

	ParameteredObject::_addParameter(
		verbose, "verbose",
		"verbose output",
		false) ;

	ParameteredObject::_addParameter(
		filenames,"filenames",
		"paths to images which show the calibration pattern",
		"string");

	ParameteredObject::_addParameter(
		tryRotatedGrid, "tryRotatedGrid",
		"try to find the grid with switched gridRangeX/Y Parameters<br>"
		"Usually the search does fail when the grid is rotated more than 45 degrees",
		false);

	ParameteredObject::_addParameter(
		aggresiveFeatureSearch, "aggresiveFeatureSearch",
		"use more aggresive settings for feature detection, usefull if you have for example very small circles etc.",
		false);


}

template <typename T>
void EstimateCamMatrixOCV<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	using namespace cv ;

	//synthetic coordinates of grid
	vector<Point3f> gridPoints ;
	vector<Point3f> gridPointsRotated ;
	cv::vector<cv::vector<Point3f> > foundObjPoints ;

	double gridSize = gridDist() ;

	Size boardSize, boardSizeRotated ; 
	boardSize.height = boardSizeRotated.width = gridRangeY() ;
	boardSize.width = boardSizeRotated.height = gridRangeX() ;

	//validate inputs
	if(gridRangeY() < 2 || gridRangeX() < 2 || gridSize <= 1e-6)
	{
		throw std::runtime_error("Checker Board parameters are to small or negative!") ;
	}
	if(initialCamMatrix.connected() && (initialCamMatrix.size() <= 0 ||
			initialCamMatrix()[0].width() != 3 ||
			initialCamMatrix()[0].height() != 3))
	{
		throw std::runtime_error("Initial Camera Matrix must be of size 3x3") ;
	}

	for( int i = 0; i < gridRangeY(); i++ )
		for( int j = 0; j < gridRangeX(); j++ )
			gridPoints.push_back(Point3f(float(j*gridSize), float(i*gridSize), 0.0));

	for( int i = 0; i < gridRangeX(); i++ )
		for( int j = 0; j < gridRangeY(); j++ )
			gridPointsRotated.push_back(Point3f(float(j*gridSize), float(i*gridSize), 0.0));


	vector<vector<Point2f> > foundCorners ;

	Size imageSize ;
	for(size_t ii = 0 ; ii < filenames.size() ; ii++)
	{
		Mat view ;
		sout << "Processing image \"" << filenames[ii] << "\" ...." ;
		view = imread(filenames[ii].c_str(), 0);
		if(view.data == 0)
		{
			raise(std::string("Could not read image!")) ;
		}
		imageSize = view.size();
		foundCorners.push_back(vector<Point2f>()) ;
		bool found = false ;
		if(gridType() == "Checkerboard")
		{	
			try
			{
				found = findChessboardCorners( view, boardSize, foundCorners.back(), CV_CALIB_CB_ADAPTIVE_THRESH );
				if(!found && tryRotatedGrid())
				{	found = findChessboardCorners( view, boardSizeRotated, foundCorners.back(), CV_CALIB_CB_ADAPTIVE_THRESH );
					if(found)
						foundObjPoints.push_back(gridPointsRotated) ;
				}
				else if(found)
				{	foundObjPoints.push_back(gridPoints) ;}
				if(found)
				{
					Mat gray;//(view.size().height,view.size().width,CV_8U,Scalar(0));
					cvtColor(view, gray, CV_RGB2GRAY);	
					cornerSubPix(gray, foundCorners.back(), Size(11, 11), Size(-1, -1),
						TermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));
				}
			}
			catch(cv::Exception& err)
			{
				err.formatMessage() ;
				raise(err.msg) ;
			}
		}
		else if(gridType() == "Circle Grid")
		{
			try
			{	
				cv::SimpleBlobDetector::Params params ;
				if(aggresiveFeatureSearch())
				{	params.minDistBetweenBlobs = 6.0 ;
					params.minArea = 9 ;
					params.minConvexity = 0.8f;
				}
				cv::SimpleBlobDetector* detector =new SimpleBlobDetector(params) ;
				// dont delete detector, findCirclesGrid will do that for you
				found = findCirclesGrid(view, boardSize, foundCorners.back(), CALIB_CB_SYMMETRIC_GRID, detector) ;	
				if(!found && tryRotatedGrid())
				{
					cv::SimpleBlobDetector* detector =new SimpleBlobDetector(params) ;
					found = findCirclesGrid(view, boardSizeRotated, foundCorners.back(), CALIB_CB_SYMMETRIC_GRID, detector) ;
					if(found)
						foundObjPoints.push_back(gridPointsRotated) ;
				}
				else if(found)
				{
					foundObjPoints.push_back(gridPoints) ;
				}

			
			}
			catch(cv::Exception& err)
			{
				err.formatMessage() ;
				raise(err.msg) ;
			}
		}
		else
			raise("Unknown grid type defined!") ;
		if(!found)
		{
			sout << "\n\tCould not find all corners in image " << std::endl ;
			foundCorners.pop_back() ;
		}
		else
			sout << std::endl ;
	}
	if(foundCorners.empty() || foundCorners[0].empty())
	{
		raise("Could not find any usable feature points") ;
	}

	Mat cMatrix = (initialCamMatrix.connected() ? 
						openCvUtil::CImg2Mat<T>(initialCamMatrix()[0],CV_64F) :
						Mat::eye(3, 3, CV_64F)) ;
	Mat distCoeffs = Mat::zeros(5, 1, CV_64F);

	vector<Mat> rvecs ;
	vector<Mat> tvecs ;
	double err = calibrateCamera(foundObjPoints, foundCorners, imageSize, cMatrix,
				distCoeffs, rvecs, tvecs);
				//CV_CALIB_USE_INTRINSIC_GUESS);

	sout << "Camera Matrix: " << cMatrix << std::endl ;
	sout << "Distortion Coefficients: " << distCoeffs << std::endl ;
	sout << "Reprojection Error: " << err << std::endl ;

	if(verbose)
	{	for(size_t ii = 0 ; ii < rvecs.size() && ii < tvecs.size() ; ii++)
		{
			sout << "Rotation " << ii <<": " << rvecs[ii] << std::endl ;
			sout << "Translation " << ii << ": " << tvecs[ii] << std::endl ;
		}
	}

	distortionCoeffs().assign(1) ;
	distortionCoeffs()[0] = openCvUtil::Mat2CImg<T>(distCoeffs) ;

	camMatrix().assign(1) ;
	camMatrix()[0] = openCvUtil::Mat2CImg<T>(cMatrix) ;
}

#endif /* _ESTIMATECAMMATRIXOCV_HXX_ */
