/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file EvaluationCorrespondences3D.hxx
 *  Implementation of the parameter class EvaluationCorrespondences3D.
 *  \author Moritz Becker
 *  \date 16.02.2011
 */

#ifndef _EVALUATIONCORRESPONDENCES3D_HXX_
#define _EVALUATIONCORRESPONDENCES3D_HXX_

#include <hekate/EvaluationCorrespondences3D.h>

template <typename T>
EvaluationCorrespondences3D<T>::EvaluationCorrespondences3D(const std::string& name) :
		TemplatedParameteredObject<T>(
				"testEvaluationCorrespondences3D", name,
				"Evaluates distances between corresponding point clouds."
		)
			,points3D1(true, false) 
			,points3D2(true, false) 
			,points2D1(true, false) 
			,points2D2(true, false) 
			
			
{
	
	
	ParameteredObject::_addParameter(filename, "filename", "filename to write evaluation results to",
			"filewrite");	
	ParameteredObject::_addParameter(filename2, "filename2", "filename to write point evaluation results to",
			"filewrite");	
	ParameteredObject::_addParameter(extendedLog, "extendedLog", "Gives out an extended log with point information.",
			false);

	ParameteredObject::_addInputSlot(points3D1, "points3D1", 
	"Evaluation 3D point cloud.", 
	"vector<vector<T> >"); 
	ParameteredObject::_addInputSlot(points3D2, "points3D2", 
	"Reference 3D point cloud. These points are assumed to be without errors.", 
	"vector<vector<T> >"); 

	ParameteredObject::_addInputSlot(points2D1, "points2D1", 
	"Evaluation 2D point cloud.", 
	"vector<vector<vector<T> > >"); 
	ParameteredObject::_addInputSlot(points2D2, "points2D2", 
	"Reference 2D point cloud. These points are assumed to be without errors.", 
	"vector<vector<vector<T> > >"); 
	
}

template <typename T>
void EvaluationCorrespondences3D<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	sout << "Evaluating points..." << std::endl;
	
    std::string inFile = this->filename() ;
    std::string inFile2 = this->filename2() ;

	//check if input string already contains file type .dat
	if(inFile=="") inFile.append("eval.dat");
	else if(inFile.substr(inFile.size()-4,4)!=".dat") inFile.append(".dat");

	//check if input string already contains file type .dat
	if(inFile2=="") inFile2.append("eval_points.dat");
	else if(inFile2.substr(inFile2.size()-4,4)!=".dat") inFile2.append(".dat");
	
	FILE* fp; 
	fp = fopen(inFile.c_str(),"a");
	if(fp == NULL)
	{	throw std::runtime_error(std::string("Could not open file ") + std::string(inFile)) ;	}
	fseek(fp,0,SEEK_END); 
	
	FILE* fp_point; 
	fp_point = fopen(inFile2.c_str(),"w");
	if(fp_point == NULL)
	{	throw std::runtime_error(std::string("Could not open file ") + std::string(inFile2)) ;	}
	//write column titles
	fprintf(fp_point, "X(real)   Y(real)   Z(real)   X   Y   Z   |Distance3D|   |Z|   DistanceX   DistanceY   DistanceZ   \n");
	

	if(points3D1.connected() && points3D2.connected()){


		// points 1
		typename std::vector< std::vector<T> >::const_iterator pointsIt1 =
			this->points3D1().begin();
		const typename std::vector< std::vector<T> >::const_iterator pointsEnd1 =
			this->points3D1().end();

		// points 2
		typename std::vector< std::vector<T> >::const_iterator pointsIt2 =
			this->points3D2().begin();
		const typename std::vector< std::vector<T> >::const_iterator pointsEnd2 =
			this->points3D2().end();

		// check if both vectors have the same length
		if(points3D1.size() != points3D2.size())
			sout << "ATTENTION: Different 3D vector length!" << std::endl;
		else{
			//number of inliers
			int inliers = 0;
			double mean = 0.;
			double meanDepth = 0.;
			double meanX = 0.;
			double meanY = 0.;
			double meanZ = 0.;
			double stdev = 0.;
			double stdevDepth = 0.;
			double stdevX = 0.;
			double stdevY = 0.;
			double stdevZ = 0.;

			cimg_library::CImg<T> centreOfMass(1,3);
			centreOfMass.fill(T(0));

			std::vector<double> distance;
			std::vector<double> distanceDepth;
			std::vector<double> distanceX;
			std::vector<double> distanceY;
			std::vector<double> distanceZ;

			// calculate distances and mean distance
			for(;pointsIt1 != pointsEnd1 && pointsIt2 != pointsEnd2; pointsIt1++, pointsIt2++){
			
				//check if points are outliers
				if((*pointsIt1)[0]==0 && (*pointsIt1)[1]==0 && (*pointsIt1)[2]==0){
					sout << "Outlier, skipping calculation" << std::endl;
				}
				else if((*pointsIt2)[0]==0 && (*pointsIt2)[1]==0 && (*pointsIt2)[2]==0){
					sout << "Outlier, skipping calculation" << std::endl;
				}
				else{
					inliers++;

					//add to centre of mass
					centreOfMass[0] += (*pointsIt1)[0];
					centreOfMass[1] += (*pointsIt1)[1];
					centreOfMass[2] += (*pointsIt1)[2];

					// calculate distance between both points
					double d = getDistance(*pointsIt1, *pointsIt2);

					// calculate absolute z distance
					double d2 = getDepthDistance(*pointsIt1, *pointsIt2);

					if(this->extendedLog()){
						sout << (*pointsIt1)[0] << ", " << (*pointsIt1)[1] << ", " << (*pointsIt1)[2] << std::endl;
						sout << (*pointsIt2)[0] << ", " << (*pointsIt2)[1] << ", " << (*pointsIt2)[2] << std::endl;
						sout << "Distance between point: " << d << std::endl;
						sout << "-------" << std::endl;
					}

					double distx = (*pointsIt1)[0]-(*pointsIt2)[0];
					double disty = (*pointsIt1)[1]-(*pointsIt2)[1];
					double distz = (*pointsIt1)[2]-(*pointsIt2)[2];

					distance.push_back(d);
					distanceDepth.push_back(d2);

					distanceX.push_back(distx);
					distanceY.push_back(disty);
					distanceZ.push_back(distz);

					mean += d;
					meanDepth += d2;

					meanX += distx;
					meanY += disty;
					meanZ += distz;
					
					//save distances in point evaluation file
					fprintf(fp_point, "%f   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   \n",
						(double)(*pointsIt2)[0], (double)(*pointsIt2)[1],
						(double)(*pointsIt2)[2], (double)(*pointsIt1)[0],
						(double)(*pointsIt1)[1], (double)(*pointsIt1)[2],
						d, d2, distx, disty, distz);
				}
			}

			centreOfMass /= inliers;

			mean /= inliers;
			meanDepth /= inliers;
			meanX /= inliers;
			meanY /= inliers;
			meanZ /= inliers;
			
			//calculate standard deviation
			std::vector<double>::const_iterator meanIt = distance.begin();
			std::vector<double>::const_iterator meanItDepth = distanceDepth.begin();
			std::vector<double>::const_iterator meanItX = distanceX.begin();
			std::vector<double>::const_iterator meanItY = distanceY.begin();
			std::vector<double>::const_iterator meanItZ = distanceZ.begin();
			const std::vector<double>::const_iterator meanEnd = distance.end();

			for(; meanIt != meanEnd; meanIt++, meanItDepth++, meanItX++, meanItY++, meanItZ++){
				stdev += (mean - *meanIt)*(mean - *meanIt) / inliers;
				stdevDepth += (meanDepth - *meanItDepth)*(meanDepth - *meanItDepth) / inliers;
				stdevX += (meanX - *meanItX)*(meanX - *meanItX) / inliers;
				stdevY += (meanY - *meanItY)*(meanY - *meanItY) / inliers;
				stdevZ += (meanZ - *meanItZ)*(meanZ - *meanItZ) / inliers;
			}

			stdev = sqrt(stdev);
			stdevDepth = sqrt(stdevDepth);
			stdevX = sqrt(stdevX);
			stdevY = sqrt(stdevY);
			stdevZ = sqrt(stdevZ);

			//if 2d AND 3d points are connected, merge sout
			if(points2D1.connected() && points2D2.connected()){
				sout << "------------------------------------------------" << std::endl;
				sout << "Evaluation of 2D AND 3D Correspondences:" << std::endl;
				sout << "inliers3D   CentreOfMassX   CentreOfMassY   CentreOfMassZ   mean   meanDepth   meanX   meanY   meanZ   stdev   stdevDepth   "
					<< "stdevX   stdevY   stdevZ   "
					<< "inliers2D   mean   meanX   meanY   stdev   stdevX   stdevY   " << std::endl;
				sout << inliers << "   "  << centreOfMass[0] << "   "  << centreOfMass[1] << "   "  << centreOfMass[2] << "   " 
				 << mean << "   " << meanDepth << "   "
				 << meanX << "   " << meanY << "   "
				 << meanZ << "   " << stdev << "   "
				 << stdevDepth << "   "	<< stdevX << "   "
				 << stdevY << "   " << stdevZ;

				//save values in file
				fprintf(fp, "%d   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   ", 
					inliers,
					(double)centreOfMass[0], (double)centreOfMass[1],
					(double)centreOfMass[2],
					mean, meanDepth, meanX, meanY, meanZ,
					stdev, stdevDepth, stdevX, stdevY, stdevZ);

			}
			//else give out 3d evalutation results serparately
			else{
				sout << "------------------------------------------------" << std::endl;
				sout << "Evaluation of 3D Correspondences:" << std::endl;
				sout << "inliers3D   CentreOfMassX   CentreOfMassY   CentreOfMassZ   mean   meanDepth   meanX   meanY   meanZ   stdev   stdevDepth   "
					<< "stdevX   stdevY   stdevZ " << std::endl;
				sout << inliers << "   " << mean << "   " << meanDepth << "   "
				 << meanX << "   " << meanY << "   "
				 << meanZ << "   " << stdev << "   "
				 << stdevDepth << "   "	<< stdevX << "   "
				 << stdevY << "   " << stdevZ << std::endl;
				sout << "------------------------------------------------" << std::endl;

				//save values in file
				fprintf(fp, "%d   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f   %f  \n", 
					inliers,
					(double)centreOfMass[0], (double)centreOfMass[1],
					(double)centreOfMass[2],
					mean, meanDepth, meanX, meanY, meanZ,
					stdev, stdevDepth, stdevX, stdevY, stdevZ);
			}

		}
	}

	if(points2D1.connected() && points2D2.connected()){

		
		// points 1
		typename std::vector< std::vector<T> >::const_iterator pointsIt1 =
			this->points2D1()[0].begin();
		typename std::vector< std::vector<T> >::const_iterator pointsIt12 =
			this->points2D1()[1].begin();
		const typename std::vector< std::vector<T> >::const_iterator pointsEnd1 =
			this->points2D1()[0].end();

		// points 2
		typename std::vector< std::vector<T> >::const_iterator pointsIt2 =
			this->points2D2()[0].begin();
		typename std::vector< std::vector<T> >::const_iterator pointsIt22 =
			this->points2D2()[1].begin();
		const typename std::vector< std::vector<T> >::const_iterator pointsEnd2 =
			this->points2D2()[0].end();

		// check if both vectors have the same length
		if(points2D1.size() != points2D2.size())
			sout << "ATTENTION: Different 2D vector length!" << std::endl;
		else{
			//number of inliers
			int inliers = 0;
			double mean = 0.;
			double meanX = 0.;
			double meanY = 0.;
			double stdev = 0.;
			double stdevX = 0.;
			double stdevY = 0.;

			std::vector<double> distance;
			std::vector<double> distanceX;
			std::vector<double> distanceY;

			// calculate distances and mean distance
			for(;pointsIt1 != pointsEnd1 && pointsIt2 != pointsEnd2; pointsIt1++, pointsIt12++, pointsIt2++, pointsIt22++){
			
				//check if points are outliers
				if((*pointsIt1)[0]==0 && (*pointsIt1)[1]==0 && (*pointsIt1)[2]==0){
					sout << "Outlier, skipping calculation" << std::endl;
				}
				else if((*pointsIt12)[0]==0 && (*pointsIt12)[1]==0 && (*pointsIt12)[2]==0){
					sout << "Outlier, skipping calculation" << std::endl;
				}
				else if((*pointsIt22)[0]==0 && (*pointsIt22)[1]==0 && (*pointsIt22)[2]==0){
					sout << "Outlier, skipping calculation" << std::endl;
				}
				else if((*pointsIt2)[0]==0 && (*pointsIt2)[1]==0 && (*pointsIt2)[2]==0){
					sout << "Outlier, skipping calculation" << std::endl;
				}
				else{
					inliers++;
					//sout << (*pointsIt1)[0] << ", " << (*pointsIt1)[1] << ", " << (*pointsIt1)[2] << std::endl;
					//sout << (*pointsIt2)[0] << ", " << (*pointsIt2)[1] << ", " << (*pointsIt2)[2] << std::endl;
					double d = getDistance2D(*pointsIt1, *pointsIt2);
					distance.push_back(d);
					mean += d;
					d = getDistance2D(*pointsIt12, *pointsIt22);
					distance.push_back(d);
					mean += d;

					distanceX.push_back((*pointsIt1)[0]-(*pointsIt2)[0]);
					distanceX.push_back((*pointsIt12)[0]-(*pointsIt22)[0]);
					distanceY.push_back((*pointsIt1)[1]-(*pointsIt2)[1]);
					distanceY.push_back((*pointsIt12)[1]-(*pointsIt22)[1]);
					
					meanX += (*pointsIt1)[0]-(*pointsIt2)[0];
					meanX += (*pointsIt12)[0]-(*pointsIt22)[0];
					meanY += (*pointsIt1)[1]-(*pointsIt2)[1];
					meanY += (*pointsIt12)[1]-(*pointsIt22)[1];
					//sout << inliers << "; " << d << ";" << std::endl;

				}
			}
			mean /= (inliers);
			meanX /= (inliers);
			meanY /= (inliers);
			
			//calculate standard deviation
			std::vector<double>::const_iterator meanIt = distance.begin();
			std::vector<double>::const_iterator meanItX = distanceX.begin();
			std::vector<double>::const_iterator meanItY = distanceY.begin();
			const std::vector<double>::const_iterator meanEnd = distance.end();
			for(; meanIt != meanEnd; meanIt++, meanItX++, meanItY++){
				stdev += (mean - *meanIt)*(mean - *meanIt) / (inliers);
				stdevX += (meanX - *meanItX)*(meanX - *meanItX) / (inliers);
				stdevY += (meanY - *meanItY)*(meanY - *meanItY) / (inliers);
			}

			stdev = sqrt(stdev);
			stdevX = sqrt(stdevX);
			stdevY = sqrt(stdevY);

			//if 2d AND 3d points are connected, merge sout
			if(points3D1.connected() && points3D2.connected()){
				sout << "   " << inliers << "   " << mean << "   " <<  meanX << "   " << meanY << "   " << stdev 
				 << "   " << stdevX << "   " << stdevY << std::endl;
				sout << "------------------------------------------------" << std::endl;

				//save values in file
				fprintf(fp, "%d   %f   %f   %f   %f   %f   %f \n", 
					inliers, mean, meanX, meanY, stdev, stdevX, stdevY);
			}
			//else give out 2d evalutation results serparately
			else{
				sout << "------------------------------------------------" << std::endl;
				sout << "Evaluation of 2D Correspondences:" << std::endl;
				sout << "inliers2D   mean   meanX   meanY   stdev   stdevX   stdevY" << std::endl;
				sout << inliers << "   " << mean << "   " <<  meanX << "   " << meanY << "   " << stdev 
				 << "   " << stdevX << "   " << stdevY << std::endl;
				sout << "------------------------------------------------" << std::endl;

				//save values in file
				fprintf(fp, "%d   %f   %f   %f   %f   %f   %f \n", 
					inliers, mean, meanX, meanY, stdev, stdevX, stdevY);
			}
			//give out 2d distances
			if(this->extendedLog()){
				meanIt = distance.begin();
				for(int j=0; meanIt != meanEnd; meanIt++, j++){
					sout << "2D point [" << j << "]: " << (*meanIt) << std::endl;
				}
			}
		}
	}
	
	
	fclose(fp); 
	fclose(fp_point); 
}

//calculate distance
template <typename T>
double getDistance(const std::vector<T> &x1, const std::vector<T> &x2){

	double x, y, z;
	x = x1[0]-x2[0];
	y = x1[1]-x2[1];
	z = x1[2]-x2[2];

	double d = sqrt((double)(x*x+y*y+z*z));

	return d;

}

//calculate distance
template <typename T>
double getDistance2D(const std::vector<T> &x1, const std::vector<T> &x2){

	double x, y;
	x = x1[0]-x2[0];
	y = x1[1]-x2[1];

	double d = sqrt((double)(x*x+y*y));

	return d;

}

//calculate depth distance
template <typename T>
double getDepthDistance(const std::vector<T> &x1, const std::vector<T> &x2){

	double z;

	z = sqrt((double)(x1[2]-x2[2])*(x1[2]-x2[2]));

	return z;

}

#endif /* _EVALUATIONCORRESPONDENCES3D_HXX_ */



