/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file ExtendSparseEdges.hxx
 *  Implementation of the parameter class ExtendSparseEdges.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.10.2012
 */

#ifndef _EXTENDSPARSEEDGES_HXX_
#define _EXTENDSPARSEEDGES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/ExtendSparseEdges.h>
#include <charon-utils/ImgTool.hxx>

template <typename T>
ExtendSparseEdges<T>::ExtendSparseEdges(const std::string& name) :
		TemplatedParameteredObject<T>(
			"ExtendSparseEdges", name,
			"<h2>Extends sparse edges by a given dense edge image</h2><br>"
			"Extends edges of source image by additional edges given by "
			"second image. This is primarily used to densify sparse depth "
			"edges obtained by LearnDepthEdgeByMask. If a given percentage "
			"(threshold1) of sparse edges ly on an edge provided by the "
			"second edge image, the sparse edge is extended by the provided "
			"edge.<br>"
			"<br>The plugin can be used to only refine specific edges."
		)
{
	ParameteredObject::_addInputSlot(
		sourceEdges, "sourceEdges",
		"Image of edges which should be extended",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		extensionEdges, "extensionEdges",
		"Image of edges used to extend source edges",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		extendedEdges, "extendedEdges",
		"Extended edges",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		iterations, "iterations",
		"Number of extension iterations",
		10, "int");

	ParameteredObject::_addParameter< float >(
		threshold, "threshold",
		"Lower threshold for source pixel values to belong to an putative edge",
		0.3, "float");

	ParameteredObject::_addParameter< float >(
		threshold2, "threshold2",
		"Threshold which defines source pixels to be edges.",
		0.5, "float");

	ParameteredObject::_addParameter< float >(
		intersection, "intersection",
		"How much percent of source pixels have to ly on a neighbouring edge, if pixel should be extended.",
		0.5, "float");

	ParameteredObject::_addParameter< float >(
		tolerance, "tolerance",
		"Tolerance for edge segments",
		1.0, "float");

	ParameteredObject::_addParameter< int >(
		minSubEdgeLength, "minSubEdgeLength",
		"Minimum length of sub edges. Defines patch around pixels to look for further edge pixels. "
		"Height and width of patch: (2*minSubEdgeLength+1)",
		3, "int");

	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void ExtendSparseEdges<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	
	const CImgList<T> & imgSrcList = this->sourceEdges();
	const CImgList<T> & imgExtList = this->extensionEdges();

	this->extendedEdges() = this->sourceEdges();

	// Throw error if number of images in input image lists are different
	if( imgExtList.size() != imgSrcList.size() )
		throw(std::range_error("Different numbers of source and extension images."));

	for( size_t i = 0; i < imgSrcList.size(); i++){

		sout << "Processing image " << i << std::endl;

		unsigned int size = this->minSubEdgeLength();	// size of evaluated patch

		for( size_t j = 0; j < this->iterations(); j++ )
		{

			sout << "\titeration " << j << " / " << this->iterations()-1 << std::endl;
			
			CImg<float> src, ext;
			src = this->extendedEdges()[i].get_channel(0).normalize(0,1);
			ext = imgExtList[i].get_channel(0).normalize(0,1).quantize(2);
			
			const int w = src.width();
			const int h = src.height();

			// Number of pixels
			size_t nPixExt = ext.width() * ext.height() ;
			size_t nPixSrc = src.width() * src.height() ;

			// Throw error if images have different dimensions
			if( nPixExt != nPixSrc ){
				sout << "Different sizes of source and extension images!" << std::endl;
				sout << "Scaling images to smaller dimension..." << std::endl;
				if( ext.width() > src.width() )
					ext.resize( src.width(), src.height() );
				else
					src.resize( ext.width(), ext.height() );
			}
	
			int x, y;
			cimg_forXY( ext, x3, y3 ){

				// Change direction of loop
				x = x3, y = y3;
				if( j % 4 == 0 )
				{
					x = w - 1 - x3;
					y = h - 1 - y3;
				} 
				else if( j % 3 == 0 )
				{
					x = w - 1 - x3;
				} 
				else if( j % 2 == 0 )
				{
					y = h - 1 - y3;
				} 

				if( src(x,y) >= this->threshold() && ext(x,y) >= this->threshold2() )
				{

					CImg<float> labels = ext.get_crop(x-size,y-size,x+size,y+size).label(true, tolerance);

					const int label = labels(size,size);
					// continue, if only one label inside the patch
					if( labels.max() == 0)
						continue;

					sout << "1" << std::endl;
				
					int numberOfExt = 0;	// number of extension edge pixels in this patch
					int numberOfSrc = 0;	// number of intersection source pixels in this patch
					float srcValues = 0;	// Sum of source values of pixels on the edge
					cimg_forXY( labels, x2, y2 )
					{				
						if( labels(x2,y2) == label )
						{
							
							int xx = std::max( int(std::min(int(x-size+x2), w)), 0 );
							int yy = std::max( int(std::min(int(y-size+y2), h)), 0 );

							numberOfExt++;
							if( src(xx,yy) > this->threshold2() )
							{
								srcValues += this->extendedEdges()[i](xx,yy);
								numberOfSrc++;
							}
						}					
					}
					sout << "2" << std::endl;

					if( float(numberOfExt) > 0)
						if( float(numberOfSrc) / float(numberOfExt) >= this->intersection() )
						{
							this->extendedEdges()[i](x,y) = std::max( float(this->extendedEdges()[i](x,y)), float(255) );//srcValues / float(numberOfSrc) );
						}

				}

			}

		}

	}

}

#endif /* _EXTENDSPARSEEDGES_HXX_ */
