/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file filewriterWRL.hxx
 *  Implementation of the parameter class filewriterWRL.
 *  \author Moritz Becker, Stephan Meister
 *  \date 09.06.2010
 */

#ifndef _FILEWRITERWRL_HXX_
#define _FILEWRITERWRL_HXX_

#include <hekate/FilewriterWRL.h>
#include <charon-utils/ImgTool.hxx>
#include <sstream>

using namespace hekate;

//set static member vars
template <typename T>
const float filewriterWRL<T>::_defaultColorR = 1.0 ;
template <typename T>
const float filewriterWRL<T>::_defaultColorG = 1.0 ;
template <typename T>
const float filewriterWRL<T>::_defaultColorB = 1.0 ;

template <typename T>
filewriterWRL<T>::filewriterWRL(const std::string& name) :
		TemplatedParameteredObject<T>("filewriterWRL", name,
			"Saves 3D point clouds as VRML 2.0 files"),
		cameraMatrices(true, false),
		calibrationMatrix(true, false),
		pointIndex(true, false),
		imagePath(true, false),
		pointCloud2D(true, false)
		//,colors(true,false)
{

	ParameteredObject::_addParameter(filename, "filename", "filename to write point cloud to",
			"filewrite");
	ParameteredObject::_addParameter(withFaceSet, "withFaceSet", "withFaceSet",
			false);
	ParameteredObject::_addParameter(camX, "cam_x", "x coordinate of camera", 0.0f);
	ParameteredObject::_addParameter(camY, "cam_y", "y coordinate of camera", 0.0f);
	ParameteredObject::_addParameter(camZ, "cam_z", "z coordinate of camera", 0.0f);
	ParameteredObject::_addParameter(viewRotAxisX, "viewRotAxisX", "X coordinate of camera rotation axis", 0.0f) ;
	ParameteredObject::_addParameter(viewRotAxisY, "viewRotAxisY", "Y coordinate of camera rotation axis", 0.0f) ;
	ParameteredObject::_addParameter(viewRotAxisZ, "viewRotAxisZ", "Z coordinate of camera rotation axis", 1.0f) ;
	ParameteredObject::_addParameter(viewRotAngle, "viewRotAngle", "angle of camera rotation [rad]", 0.0f) ;
	ParameteredObject::_addParameter(viewFov, "viewFov", "camera Field-Of-View [rad]", 0.785398f) ;

	ParameteredObject::_addInputSlot(cameraMatrices, "cameraMatrices", "cameraMatrices", "IndexedMatrixList");
	ParameteredObject::_addInputSlot(calibrationMatrix, "calibrationMatrix", "calibrationMatrix", "CImgList<T>");
	ParameteredObject::_addInputSlot(pointCloud, "point_cloud", "point_cloud", "CorrespondenceList"); 
	ParameteredObject::_addInputSlot(pointIndex, "point_index", "point_index", "vector<vector<int> >"); 
	//ParameteredObject::_addInputSlot(colors, "colors", "colors", "vector<vector<T> >");  

	ParameteredObject::_addInputSlot(imagePath, "imagePath", "imagePath", "vector<string>"); 
	ParameteredObject::_addInputSlot(pointCloud2D, "pointCloud2D", 
	"2d correspond. with depth", "vector<vector<T> >"); 
	
}

template <typename T>
void filewriterWRL<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	//Use . as decimal point independ from the system language!
	setlocale(LC_NUMERIC, "C");

	// Get file path
    std::string inFile = this->filename() ;

	// Check if input string already contains file type .wrl
	if(inFile.empty())
	{
		std::string msg = this->ParameteredObject::getClassName() + " : " + this->ParameteredObject::getName() + " : filename is empty!";
		throw std::invalid_argument(msg) ;
	}
	if(inFile.length() > 4 && inFile.substr(inFile.size()-4,4)!=".wrl") 
		inFile.append(".wrl");

	FILE* fp; 
	fp = fopen(inFile.c_str(),"w");
	if(fp == NULL)
	{	throw std::runtime_error(std::string("Could not open file ") + std::string(inFile)) ;	}


	// Write HEADER
	fprintf(fp, "#VRML V2.0 utf8 Hekate filewriterWRL plugin\n") ;
			

	//write Viewpoint/Camera
	if(this->cameraMatrices.connected()){

		sout << "Found " << this->cameraMatrices().size() << " camera matrices." << std::endl;

		for(size_t i = 0; i<this->cameraMatrices().size(); i++){

			//calculate camera centre (hartley p. 163)
			const cimg_library::CImg<double> &cam = cameraMatrices()[i];	
			cimg_library::CImg<double> C1(1,4);
			C1[0] = cam.get_crop(1,0,3,2).det();
			C1[1] = -((cam.get_crop(0,0,0,2)).append(cam.get_crop(2,0,3,2))).det();
			C1[2] = (cam.get_crop(0,0,1,2).append(cam.get_crop(3,0,3,2))).det();
			C1[3] = -cam.get_crop(0,0,2,2).det();
			if(C1[3] != 0) C1 /= C1[3];
			else sout << "Camera centre at infinity" << std::endl;

			// Calculate direction of principal axis
			cimg_library::CImg<double> principalAxis(1,3), crossProduct(1,3), defaultVector(1,3);
			principalAxis = cam.get_crop(0,0,2,2).det() * cam.get_crop(0,2,2,2).get_transpose();
			principalAxis /= principalAxis.magnitude();
			//ImgTool::printInfo(sout,principalAxis,"principalAxis");

			defaultVector.fill(0);
			defaultVector[2] = T(-1);
			crossProduct = defaultVector.get_cross(principalAxis);
			//ImgTool::printInfo(sout,crossProduct,"crossProduct");

			double rotationAngle = std::acos(defaultVector.dot(principalAxis)/
				(principalAxis.magnitude()*defaultVector.magnitude()));
			//sout << "Angle = " << rotationAngle << std::endl;

			sout << "Camera " << i << " at: (" << C1[0] << ", "
				<< C1[1] << ", " << C1[2] << ")" << std::endl;

			// write camera centre to wrl file
			fprintf(fp, "Viewpoint {\n");
			fprintf(fp, "\tposition %f %f %f \n",
				(double)C1[0], (double)C1[1], (double)C1[2]);
			if(crossProduct.magnitude() > 0){
				fprintf(fp, "\torientation %f %f %f %f\n",
				(double)crossProduct[0], (double)crossProduct[1],
				(double)crossProduct[2], (double)rotationAngle);
			}
			else if(principalAxis[2]<0){
				fprintf(fp, "\torientation 0, 1, 0, 0 \n");
			}
			else{
				fprintf(fp, "\torientation 0, 1, 0, 3.141 \n");
			}
			fprintf(fp, "\tfieldOfView %f\n", this->viewFov());
			fprintf(fp, "}\n") ;
			
		
			//ImgTool::printInfo(sout,cam,"P");
		}
	}
	else{
		fprintf(fp, "Viewpoint {\n");
		fprintf(fp, "\tposition %f %f %f \n", this->camX(), this->camY(), this->camZ());
		fprintf(fp, "\torientation %f %f %f %f\n", this->viewRotAxisX(), this->viewRotAxisY(), this->viewRotAxisZ(), this->viewRotAngle());
		fprintf(fp, "\tfieldOfView %f\n", this->viewFov());
		fprintf(fp, "}\n") ;
	}
	

	// write point data -------------------------------------------------------------
	
	//create Shape node to hold the geometry data
	fprintf(fp, "\nShape {\n");

	if(this->imagePath.connected()){
		std::string path = this->imagePath()[0];
		FileTool::slashConvert(path);
		fprintf(fp, "\tappearance Appearance {\n");
		fprintf(fp, "\tmaterial Material { }\n");
		if(this->withFaceSet())
		{
			fprintf(fp, "\ttexture ImageTexture {\n");
			fprintf(fp, "\turl \"%s\"\n",path.c_str());
			fprintf(fp, "\t}\n");
		}
		fprintf(fp, "}\n");			
    }


	//use geometry node of type PointSet for unconnected point cloud -----------------
	if(this->pointIndex.connected() && this->pointIndex().size() != 0 && this->withFaceSet())
	{
		fprintf(fp, "\tgeometry IndexedFaceSet {\n");
		fprintf(fp, "\tsolid FALSE\n");
		fprintf(fp, "\tccw FALSE\n");	
		fprintf(fp, "\tcolorPerVertex FALSE\n");		
	}
	else //use IndexedFaceSet for connected polygons
	{
		fprintf(fp, "\tgeometry PointSet {\n");
	}
	fprintf(fp, "\t\tcoord Coordinate {\n");
	fprintf(fp, "\t\t\tpoint [\n");
	
	
	// Get point information
	const CorrespondenceList<T> &corrs = this->pointCloud();
	size_t nPoints = corrs.size();
	size_t nImages = corrs.sizeImages();


	// string for colors per point
	std::string colors;
	colors = "color Color {\ncolor [\n" ;


	// write point coords to wrl file -------------------------------------------------
	auto itPtCloud = corrs.begin() ;
	auto itPtCloudEnd = corrs.end() ;
	for(; itPtCloud != itPtCloudEnd; itPtCloud++)
	{			

		// Write point coordinates to file
		//sout << *itPtCloud << std::endl;
		fprintf(fp, "\t\t\t%f %f %f,\n",
			(float) itPtCloud->x(),
			(float) itPtCloud->y(),
			(float) itPtCloud->z() );
		
		
		// Save colors of point into the colors string
		colors.append("\t\t\t");

		std::ostringstream rgb_r, rgb_g, rgb_b;
		rgb_r << itPtCloud->R / 255.;
		rgb_g << itPtCloud->G / 255.;
		rgb_b << itPtCloud->B / 255.;
		
		colors.append(rgb_r.str());
		colors.append(" ");
		colors.append(rgb_g.str());
		colors.append(" ");
		colors.append(rgb_b.str());

		colors.append("\n");

	}

	// close indentations
	fprintf(fp, "\t\t\t]\n\t\t}\n");
	

	// close color string
	colors.append("\t\t\t]\n\t\t}\n");

	// print color string to file
	fprintf(fp, colors.c_str());


	
	
	
	// if point index is given, write them to wrl file
	if(this->pointIndex.connected() && this->pointIndex().size()!=0 && this->withFaceSet()){
		fprintf(fp, "coordIndex [\n");

		std::vector< std::vector<int> >::const_iterator pointsIt = this->pointIndex().begin();
		const std::vector< std::vector<int> >::const_iterator pointsEnd = this->pointIndex().end() ;

		for(;pointsIt != pointsEnd ; pointsIt++)
		{
			const std::vector<int> & points = *pointsIt ;
			fprintf(fp, "\t\t\t") ;
			for(std::vector<int>::const_iterator j = points.begin() ; j != points.end() ; j++)
			{	fprintf(fp, "%u, ", *j) ;	}
			fprintf(fp, "-1,\n") ;
		}

		fprintf(fp, "\t\t\t]\n");

		// Give color
		fprintf(fp, "\t\tcolorPerVertex TRUE\n") ;
		fprintf(fp, colors.c_str());

	}
	

	/*
	if(this->imagePath.connected()){
		fprintf(fp, "texCoord TextureCoordinate {\n");
		fprintf(fp, "\tpoint [\n");
		
		const cimg_library::CImg<T> img(imagePath()[0].c_str());
		
		typename std::vector<std::vector<T> >::const_iterator it = pointCloud2D().begin() ;
		typename std::vector<std::vector<T> >::const_iterator end = pointCloud2D().end() ;

		for(;it != end ; it++){
			fprintf(fp, "\t%f %f,\n",
				(double)(*it)[0]/img.width(),
				(double)(img.height()-(*it)[1])/img.height());
		}
		fprintf(fp, "\t]\n");
		fprintf(fp, "}\n");
		fprintf(fp, "texCoordIndex [");
		
		//for(int i=0; i<this->pointIndex().size(); i++){
		//	fprintf(fp, "%u ", i);
		//}
		

		std::vector< std::vector<int> >::const_iterator pointsIt = this->pointIndex().begin();
		const std::vector< std::vector<int> >::const_iterator pointsEnd = this->pointIndex().end() ;
		
		for(;pointsIt != pointsEnd ; pointsIt++)
		{
			const std::vector<int> & points = *pointsIt ;
			fprintf(fp, "\t\t\t") ;
			for(std::vector<int>::const_iterator  j = points.begin() ; j != points.end() ; j++)
			{	fprintf(fp, "%u, ", *j) ;	}
			fprintf(fp, "-1,\n") ;
		}
		fprintf(fp, "]");
	}
	*/


	

	
	
	fprintf(fp, "\t}\n}");
	fclose(fp);

}

#endif /* _FILEWRITERWRL_HXX_ */



