/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Flow2Correspondences.hxx
 *  Implementation of the parameter class Flow2Correspondences.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 18.07.2012
 */

#ifndef _FLOW2CORRESPONDENCES_HXX_
#define _FLOW2CORRESPONDENCES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "Flow2Correspondences.h"
#include <charon-utils/ImgTool.hxx>

template <typename T>
Flow2Correspondences<T>::Flow2Correspondences(const std::string& name) :
		TemplatedParameteredObject<T>(
			"Flow2Correspondences", name,
			"<h2>Converts a flow image to Correspondences</h2><br>"
			"Converts a flow image to the Correspondences<T> class. A step width can be defined to only select a smaller grid from the flow image."
		),
		colorImages(true,0),
		flowMask(true,0)
{
	ParameteredObject::_addInputSlot(
		flowImage, "flowImage",
		"flow image",
		"CImgList<T>");

	ParameteredObject::_addInputSlot(
		colorImages, "colorImages",
		"Optional color images to color the correspondences",
		"CImgList<T>");

	ParameteredObject::_addInputSlot(
		flowMask, "flowMask",
		"A mask to select only areas of the given flow",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");

	ParameteredObject::_addParameter< int >(
		stepSize, "stepSize",
		"Step size for point selection (1 to take every points -> slow). ",
		1, "int");

	ParameteredObject::_addParameter< int >(
		stepSizeNewCorrs, "stepSizeNewCorrs",
		"Start a new correspondence every \"stepSizeNewCorrs\"th frame",
		0, "int");

	ParameteredObject::_addParameter< int >(
		length, "length",
		"Length of a correspondence (number of frames). "
		"Default(0) == whole sequence length.",
		0, "int");

	ParameteredObject::_addParameter(
		indicesFlow, "indicesFlow",
		"List of image indices represented by the flow images",
		"int");
	
	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void Flow2Correspondences<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	hekate::CorrespondenceList<T> &correspondences = this->correspondences();

	// Check consistency of given indices, if false, assume increasing index order
	// as following: flow0: img0 -> img1, flow1: img1 -> img2 ...
	// Only if this function returns true, color images will be used as corr colors
	/*
	bool useIndices = checkInputConsistency();
	if(!useIndices)
		sout << "No given index order. " << std::endl;
		*/
	if( this->flowMask.connected() )
		if( this->flowImage().size() != this->flowMask().size() )
			throw(std::range_error("Inconsistent size of flow and according masks."));

	if( this->colorImages.connected() )
		if( this->colorImages().size() <= this->flowImage().size() )
			throw( std::range_error("Not enough color images.") );


	// Initialize correspondences ----------------------------------

	// Get flow data
	const int w = this->flowImage()[0].width();
	const int h = this->flowImage()[0].height();
	cimg_library::CImg<T> colImg0(w,h,1,3);
	colImg0.fill(0);
	if(this->colorImages.connected() /*&& useIndices*/){
		colImg0 = this->colorImages()[ this->indicesFlow()[0] ];
		colImg0.resize(colImg0.width(),colImg0.height(),1,3);
	}

	// Get step size to read in
	int step = (this->stepSize() > 1 ? this->stepSize() : 1);

	int k=0;
	for(size_t i=0; i < w; i=i+step)
	{
		for(size_t j=0; j < h; j=j+step)
		{

			// get colors
			float r, g, b;
			if(this->colorImages.connected()){
				r = float( colImg0(i,j,0,0) );
				g = float( colImg0(i,j,0,1) );
				b =float(  colImg0(i,j,0,2) );
			}
			else{
				r = 0; g = 0; b = 0;
			}
			hekate::Correspondence<T> corr1( T(i), T(j), 0, 1, k, 0, false, r, g, b);
			correspondences.push_back(corr1);
			k++;
		}
	}

	sout << "Initial number of correspondences = " << k << std::endl;
	sout << "correspondences.sizeImages() = " << correspondences.sizeImages() << std::endl;
	sout << "correspondences.sizePoints(0) = " << correspondences.sizePoints(0) << std::endl;
	sout << "correspondences.sizePoints() = " << correspondences.sizePoints() << std::endl;

	// add correspondences by flow ---------------------------------

	for( size_t ii = 0; ii < this->flowImage().size(); ii++ )
	{

		sout << "Convert flow image " << ii << std::endl;

		// Get flow data
		const cimg_library::CImg<T> &img = this->flowImage()[ii];
		cimg_library::CImg<T> colImg(img.width(),img.height(),1,3);
		colImg.fill(0);
		if(this->colorImages.connected() /*&& useIndices*/){
			colImg = this->colorImages()[ this->indicesFlow()[ii] ];
			colImg.resize(colImg.width(),colImg.height(),1,3);
		}
	
		// Convert flow to correspondences --------------------------------------

		// Schleife durch bestehende Correspondences mit index ii
		//   -> suche an der interpolierten stelle der korrespondenz ii den interpolierten wert fuer flow_x und flow_y
		//   -> addiere diesen flow zu (x,y)_ii und erstelle damit die neue korrespondenz mit index ii+1
		//   -> consistency check: wenn flow werte NAN sind, dann breche korrespondenz ab
		// Damit durch das Abbrechen der Korrespondenzen die Anzahl der Korrespondenzen nicht immer weniger wird
		// kann in jedem i-ten Frame ein neuer Satz an Korrespondenzen begonnen werden.

		sout << "\titerate over flow image" << std::endl;
		sout << "\tWidth: " << img.width() << " / height: " << img.height() << std::endl;

		std::vector< Correspondence<T> > corrs_temp;

		int k2=0, k3=0;
		auto it = correspondences.get<imgIndex>().lower_bound(ii) ;
		auto itEnd = correspondences.get<imgIndex>().upper_bound(ii) ;
		for(; it != itEnd ; it++)
		{

			// Break condition for current correspondence according to given length
			if( this->length() > 0 )
			{
				const std::vector< const Correspondence<T>* > ccorr = correspondences.getListByPointId( it->pointIndex );
				if( ccorr.size() >= this->length() )
					continue;
			}

			double oldX = (*it).x() ;
			double oldY = (*it).y() ;
			
			// Skip pixel because flow mask = 0
			if(this->flowMask.connected())			
				if(this->flowMask()[ii].linear_atXY(oldX, oldY) == 0 )
					continue;

			// Get new position
			double x = oldX + img.linear_atXYZ(oldX, oldY, 0);
			double y = oldY + img.linear_atXYZ(oldX, oldY, 1);

			// Exit tracking of current correspondence due to border conditions
			if( x < 0 || y < 0 || x >= img.width() || y >= img.height() )
			{
				//sout << "Break correspondence " << it->pointIndex << ": x=" << x << ",y=" << y << std::endl;
				k3++;
			}
			else
			{
				
				k2++;
				// get colors
				float r, g, b;
				if(this->colorImages.connected()){
					r = colImg(x,y,0,0);
					g = colImg(x,y,0,1);
					b = colImg(x,y,0,2);
				}
				else{
					r = 0; g = 0; b = 0;
				}
				
				hekate::Correspondence<T> corr1( T(x), T(y), 0, 1, it->pointIndex, ii+1, false, 0, 0, 0);
				corrs_temp.push_back(corr1);

			}

		}

		// Create new correspondences on evan pixel grid
		if( this->stepSizeNewCorrs() > 0 )
		{
			unsigned int stepNewCorrs = this->stepSizeNewCorrs();
			if( this->length() > 0 )
				stepNewCorrs = std::min( this->stepSizeNewCorrs(), this->length() );
			if(ii%stepNewCorrs == 0)
			{
				for(size_t i=0; i < w; i=i+step)
				{
					for(size_t j=0; j < h; j=j+step)
					{

						// get colors
						float r, g, b;
						if(this->colorImages.connected()){
							r = colImg(i,j,0,0);
							g = colImg(i,j,0,1);
							b = colImg(i,j,0,2);
						}
						else{
							r = 0; g = 0; b = 0;
						}
						hekate::Correspondence<T> corr1( T(i), T(j), 0, 1, k, ii+1, false, r, g, b);
						corrs_temp.push_back(corr1);
						k++;
					}
				}
			}
		}

		// Add new correspondences to correspondence list
		for( size_t iii = 0; iii < corrs_temp.size(); iii++)
		{
			correspondences.push_back( corrs_temp[iii] );
		}

		sout << "Broke " << k3 << " correspondences due to border conditions" << std::endl;
		sout << "Correspondences of frame " << ii << ": " << k2 << std::endl;

	}

}

template <typename T>
bool Flow2Correspondences<T>::checkInputConsistency() 
{
	bool consistentIndices = 0;
	if( this->indicesFlow().size() == this->flowImage().size() * 2 )
	{

		// Get maximal given index
		int maxIndex = 0;
		for( size_t i = 0; i < this->indicesFlow().size(); i++)
		{
			maxIndex = ( maxIndex < this->indicesFlow()[i] ? this->indicesFlow()[i] : maxIndex );
		}

		// If maximal given index out of range -> return 0
		if( maxIndex > this->flowImage().size() )
		{
			return 0;
		}

		if( maxIndex == 0 )
		{
			return 0;
		}

		if( this->colorImages.connected() )
		{
			// If maximal given index out of range of color images -> return 0
			if( maxIndex >= this->colorImages().size() )
			{
				return 0;
			}
		}

		return 1;

	}
	else if( this->indicesFlow().size() == 0 )
	{				
		if( this->colorImages.connected() )
		{
			// If maximal given index out of range of color images -> return 0
			if( this->flowImage().size() >= this->colorImages().size() )
			{
				sout << "flow images: " << this->flowImage().size() << std::endl;
				sout << "color images: " << this->colorImages().size() << std::endl;
				return 0;
			}
			return 1;
		}
		return 0;
	}
	else
	{
		return 0;
	}

}

#endif /* _FLOW2CORRESPONDENCES_HXX_ */
