/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Flow2Correspondences.hxx
 *  Implementation of the parameter class Flow2Correspondences.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 18.07.2012
 */

#ifndef _FLOW2CORRESPONDENCESKITTI_HXX_
#define _FLOW2CORRESPONDENCESKITTI_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "Flow2CorrespondencesKitti.h"
#include <charon-utils/ImgTool.hxx>

template <typename T>
Flow2CorrespondencesKitti<T>::Flow2CorrespondencesKitti(const std::string& name) :
		TemplatedParameteredObject<T>(
			"Flow2CorrespondencesKitti", name,
			"<h2>Converts a flow image to Correspondences</h2><br>"
			"Converts a flow image to Correspondences<T>"
		),
		colorImages(true,0)
{
	ParameteredObject::_addInputSlot(
		flowImage, "flowImage",
		"flow image",
		"CImgList<T>");

	ParameteredObject::_addInputSlot(
		colorImages, "colorImages",
		"Optional color images to color the correspondences",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");
	ParameteredObject::_addOutputSlot(
		flow, "flow",
		"flow",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		stepSize, "stepSize",
		"Step size for choosing points (1 to take every points -> slow).",
		1, "int");
	
	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void Flow2CorrespondencesKitti<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	// Get flow data
	const cimg_library::CImg<T> &img = this->flowImage()[0];
	cimg_library::CImg<T> colImg(img.width(),img.height(),1,3);
	colImg.fill(0);
	if(this->colorImages.connected()){
		colImg = this->colorImages()[0];
		colImg.resize(colImg.width(),colImg.height(),1,3);
	}

	// Initialize flow output images
	cimg_library::CImg<float> flow1(img.width(),img.height(),1,3);
	float maxUV = 0, minUV = 0;
	//flow1.fill(0);

	// Initialize correspondence list
	hekate::CorrespondenceList<T> &corrs = this->correspondences();

	// Get step size to read in
	int step = (this->stepSize() > 1 ? this->stepSize() : 1);
	
	// Read in flow
	int k=0;
	for(size_t i=0; i < img.width(); i=i+step){
		for(size_t j=0; j < img.height(); j=j+step){
			if(img.atXYZC(i,j,0,2) == 1){
				float u = ((float)img.atXYZC(i,j,0,0)-std::pow(2.,15.))/64.0;
				float v = ((float)img.atXYZC(i,j,0,1)-std::pow(2.,15.))/64.0;
				//sout << u << " " << v << std::endl;

				// get colors
				float r, g, b;
				if(this->colorImages.connected()){
					r = colImg(i,j,0,0);
					g = colImg(i,j,0,1);
					b = colImg(i,j,0,2);
				}
				else{
					r = 0; g = 0; b = 0;
				}

				// Write flow image
				flow1(i,j,0,0) = u;
				flow1(i,j,0,1) = v;
				if( maxUV < u ) maxUV = u;
				if( maxUV < v ) maxUV = v;
				if( minUV > u ) minUV = u;
				if( minUV > v ) minUV = v;

				// Save correspondence
				hekate::Correspondence<T> corr1( T(i), T(j), 0, 1, k, 0, false, r, g, b);
				hekate::Correspondence<T> corr2( T(i+u), T(j+v), 0, 1, k, 1, false, r, g, b);
				corrs.push_back(corr1);
				corrs.push_back(corr2);
				k++;
			}
		}	
	}
	
	flow1 -= minUV;
	flow1 /= maxUV/255;

	sout << "maxUV = " << maxUV << std::endl;
	sout << "minUV = " << minUV << std::endl;

	sout << k << " correspondences of " 
		<< img.width()*img.height() << " pixels found." << std::endl;
	
	this->flow().push_back( flow1 );

	//ImgTool::printInfo(sout, this->flowImage()[0] , "flowImage");

}

#endif /* _FLOW2CORRESPONDENCESKITTI_HXX_ */
