/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FlowConfidence.hxx
 *  Implementation of the parameter class FlowConfidence.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 15.11.2013
 */

#ifndef _FLOWCONFIDENCE_HXX_
#define _FLOWCONFIDENCE_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "FlowConfidence.h"
#include <charon-utils/ImgTool.hxx>

template <typename T>
FlowConfidence<T>::FlowConfidence(const std::string& name) :
		TemplatedParameteredObject<T>(
			"FlowConfidence", name,
			"<h2>Confidence measure for a given flow field</h2><br>"
			"Confidence measure for a given flow field"
		)
{
	ParameteredObject::_addInputSlot(
		flow, "flow",
		"Flow to measure confidence for",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		confidence, "confidence",
		"Weights according to the measured confidence",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		blockSize, "blockSize",
		"Size of blocks to evaluate",
		5, "int");

	ParameteredObject::_addParameter< int >(
		projectDimensions, "projectDimensions",
		"Number of dimension for PCA projection",
		4, "int");

}

template <typename T>
void FlowConfidence<T>::execute() {

	// number of pixels to add to the block in each direction
	const int _bs = std::max(int(0), int(this->blockSize()));

	// length of the sequence
	const int slen = this->flow().size();
	sout << "Calculating confidence for " << slen << " flow images" << std::endl;

	// percentage of how many eigen vectors to keep for the pca
	const int _entropy = std::max( 1, std::min( _bs*2+1, this->projectDimensions() ) );
		
	// Output cimgs
	CImgList<T> &confidence = this->confidence();
	confidence.assign( slen );

	// Default vector length (without border conditions)
	const int vlen = (2 * _bs + 1)*(2 * _bs + 1)*2;

	// number of dimensions to keep for pca
	const int dim_val	= std::max( 1, std::min( int(vlen-1) , int(vlen*_entropy) ) );
	sout << "Taking " << dim_val << " eigen vectors for pca projection" << std::endl;
	
	#pragma omp parallel for
	for( int i = 0; i < slen; i++ ){

		// Get flow values from filename
		CImg<float> u, v;
		getFlow( u, v, this->flow()[i] );
	
		// Image dimensions
		const int _w = u.width();
		const int _h = u.height();

		sout << "Calculate confidence for flow image " << i << std::endl;

		const T mean = 0.5 * ( u.mean() +  v.mean() );

		CImg<T> &conf = confidence[i];
		conf.resize(_w,_h,1,1,0);

		// vector with - mean value as default vector
		CImg<T> vec(1,vlen,1,1);

		// distance vector
		CImg<T> proj1(1,dim_val,1,1,0);

		// covariance matrix
		CImg<T> mat(vlen, vlen, 1, 1, 0);

		// eigen values and eigen vectors
		CImg<T> eval, evec;
		
		for(int x = _bs; x < _w-_bs; x++ ){

			if( x % 100 == 0 && x != 0 ){
				sout << int(float(x)/float(_w)*100) << "%" << std::endl; ;
			}

			for( int y = _bs; y < _h-_bs; y++ ){
						
				// Get 1d vector containing u and v values in _bs sized neighbourhood
				getVec(vec, u, v, i, x, y, _bs, _w, _h, mean);

				// Calculate covariance matrix of vec				
				for( int k1 = 0; k1 < vlen; k1++ ){
					for( int k2 = k1; k2 < vlen; k2++ ){
						mat(k1,k2) += vec[k1] * vec[k2];
						mat(k2,k1) += vec[k1] * vec[k2];
					}
				}

			}

		}

		mat.symmetric_eigen(eval, evec);

		for(int x = _bs; x < _w-_bs; x++ ){

			if( x % 100 == 0 && x != 0 ){
				sout << int(float(x)/float(_w)*100) << "%" << std::endl; ;
			}

			for( int y = _bs; y < _h-_bs; y++ ){

				proj1.fill(0);
						
				// Get 1d vector containing u and v values in _bs sized neighbourhood
				getVec(vec, u, v, i, x, y, _bs, _w, _h, mean);

				for( int xx = 0; xx < vlen; xx++ ){
					for( int yy = 0; yy < dim_val; yy++ ){
						proj1(0,yy) += evec(yy,xx)*vec(0,xx);	// E_reduced_transposed * vec
					}
				}
				
				// Get distance vector by projecting vec to the reduced eigen vectors
				// Get projection
				float p1=0,p2=0;
				for( int l = 0; l < dim_val; l++ ){
					p1 += evec(l,0) * proj1(0,0);
					p2 += evec(l,1) * proj1(0,1);
				}
				
				// Calculate distance between projection and former vector
				float d1 = p1-vec[0];
				float d2 = p2-vec[1];
				//float c = std::sqrt( d1*d1 + d2*d2 );
				float c = d1*d1 + d2*d2;

				// save confidence
				conf(x,y) = c;

			}

		}

	}

}


/// Read flow values from file
void getFlow( CImg<float> &u, CImg<float> &v, const std::string &path ){

	try {
		const char *fn = path.c_str();
		FILE *stream = fopen(fn, "rb");

		if (!stream) {
			throw(std::runtime_error("Failed to open file!"));
		}

		int width, height;
		float tag;
		size_t read;

		read = fread(&tag,    sizeof(float), 1, stream);
		if (read != 1) {
			throw std::runtime_error("error reading flo tag");
		}
		read = fread(&width,  sizeof(int),   1, stream);
		if (read != 1) {
			throw std::runtime_error("error reading flo width");
		}
		read = fread(&height, sizeof(int),   1, stream);
		if (read != 1) {
			throw std::runtime_error("error reading flo height");
		}

		int nBands = 2;
		u.resize(width, height, 1, 1);
		v.resize(width, height, 1, 1);
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < nBands*width; x++) {
				float fv;
				read = fread(&fv, sizeof(float), 1, stream);
				if (read != 1) {
					throw std::runtime_error("error reading flo values");
				}
				if( x%nBands == 0 )
					u.atXY(x/nBands, y) = fv;
				else if( x%nBands == 1 )
					v.atXY(x/nBands, y) = fv;

			}
		}
		fclose(stream);

	}
	catch (const cimg_library::CImgException& err) {
		throw(std::runtime_error("Could not read flow file"));
	}

}


// Get vector of u and v flow at area (_x,_y) and its neighbourhood
template<typename T>
bool FlowConfidence<T>::getVec(CImg<T> &vec, const CImg<float> &u, const CImg<float> &v, const int index, const int _x, 
	const int _y, const int _bs, const int _w, const int _h, const float _mean){

	int l = _x - _bs;
	int r = _x + _bs;	
	int t = _y - _bs;
	int b = _y + _bs;

	vec(0,0) = u(_x,_y) - _mean;
	vec(0,1) = v(_x,_y) - _mean;

	// Fill vector with neighbour values of u and v
	int it = 1;
	for( int x = l; x <= r; x++ ){
		for( int y = t; y <= b; y++ ){
			if( x == _x && y == _y ) continue;
			vec(0, it*2)	= u(x,y) - _mean;
			vec(0, it*2+1)	= v(x,y) - _mean;
			it++;
		}
	}

	return true;

}

#endif /* _FLOWCONFIDENCE_HXX_ */
