/*  Copyright (C) 2011 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FundamentalMatrix.hxx
 *  Implementation of the parameter class FundamentalMatrix.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 14.10.2011
 */

#ifndef _FUNDAMENTALMATRIX_HXX_
#define _FUNDAMENTALMATRIX_HXX_

#include <opencv2/core/core.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <hekate/openCvUtil.hxx>
#include <charon-core/ParameteredObject.hxx>
#include <hekate/FundamentalMatrix.h>
#include <charon-utils/ImgTool.hxx>
#include <hekate/Correspondence.hxx>
#include <time.h>

using namespace hekate;

template <typename T>
FundamentalMatrix<T>::FundamentalMatrix(const std::string& name) :
		TemplatedParameteredObject<T>(
			"FundamentalMatrix", name,
			"<h2>Calculates a fundamental matrix using the OpenCV "
			"implementation</h2><br>"
			"This module calculates a fundamental matrix based on OpenCV."
		)
{
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"A vector list of corresponding feature point.",
		"CorrespondenceList");

	ParameteredObject::_addOutputSlot(
		fundamentalMatrix, "fundamentalMatrix",
		"The calculated fundamental matrix",
		"IndexedMatrixList");
	ParameteredObject::_addOutputSlot(
		correspondenceInlier, "correspondenceInlier",
		"A correspondence list, where outliers have been sorted out by"
		"epipolar condition.",
		"CorrespondenceList");

	ParameteredObject::_addParameter< int >(
		maxFrames, "maxFrames",
		"Numer of maximal frames which will be considered. (0 = all frames)",
		0, "int");
	ParameteredObject::_addParameter< double >(
		paramRANSAC, "paramRANSAC",
		"Parameter used for RANSAC. It is the maximum distance from a point"
		"to an epipolar line in pixels, beyond which the point is considered"
		"an outlier and is not used for computing the final fundamental"
		"matrix. It can be set to something like 1-3, depending on the"
		"accuracy of the point localization, image resolution, and the image"
		"noise.",
		3, "double");
	ParameteredObject::_addParameter< double >(
		paramConfidence, "paramConfidence",
		"Parameter used for the RANSAC or LMedS methods only. It specifies a"
		"desirable level of confidence (probability) that the estimated"
		"matrix is correct.",
		0.99, "double");
	ParameteredObject::_addParameter< bool >(
		logDetails, "logDetails",
		"Prints details of calculation to log file.",
		0, "bool");

	calcMax = 0;
	calcCurrent = 0;
	
	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void FundamentalMatrix<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	time_t cstart, cend;
	cstart = time(0);

	// Get correspondence dimensions
	CorrespondenceList<T> list = this->correspondences();	
	
	// Number of different images
	const size_t nImages = list.sizeImages();

	// Determine number of fundamental matrices
	// to calculate
	size_t nFM;
	if(maxFrames() < nImages && maxFrames() > 1)
		nFM = maxFrames() - 1 ;
	else
		nFM = nImages - 1 ;
	
	sout << "Calculating  fundamental matrices for " 
		<< nFM +1 << " frames." << std::endl;

	/*
	for(int i=1; i<=nFM; i++)
		calcMax += i;
	calcMax += nFM;
	*/

	calcMax += nFM * (nFM + 3)/2;
	
	int outliers = 0;
	int reversedInliers = 0;
	int nAllPoints = 0;
	// Iterate over all correspondence pairs i
	for(int i=0; i<nFM; i++)
	{

		for(size_t ii = i+1; ii < nFM+1; ii++)
		{

			//time_t time_start, time_end;
			//time_start = time(0);

			int outliersTemp = 0;
			int reversedInliersTemp = 0;
			int nPointsTemp = 0;

			sout << "Calculating fundamental matrix (" << i 
				<< "/" << ii << ") --------------------------------" << std::endl;

			// Read in correspondences
			std::vector<cv::Point2d> feat1;
			std::vector<cv::Point2d> feat2;
			std::vector<size_t> indexList;

			auto it3 = list.get<imgIndex>().lower_bound(i) ;
			auto it3End = list.get<imgIndex>().upper_bound(i) ;

			for(int j=0; it3 != it3End ; it3++, j++){

				if( it3->outlier ) continue;

				try{

					const Correspondence<T> &pt2 = list(ii, it3->pointIndex);

					if( pt2.outlier ) continue;

					feat1.push_back(convertVector(*it3));
					feat2.push_back(convertVector(pt2));
					indexList.push_back(it3->pointIndex);

					nPointsTemp++;

				}
				catch(...){}
			}
			nAllPoints += nPointsTemp;

			if(this->logDetails())
			{
				sout << nPointsTemp << " correspondences found between image "
					<< i << " and " << ii << std::endl;
			}

			if(nPointsTemp < 8)
			{
				sout << "Not enough points available to calculate a fundamental matrix!" << std::endl;
				sout << "At least 8 points are needed for the used 8-point algorithm." << std::endl;
				continue;
				//throw std::range_error("Not enough point for 8-point algorithm!");
			}
		
			// Create fundamental matrix of correspondence i
			cv::Mat mask;
			cv::Mat fm = cv::findFundamentalMat(feat1,feat2,cv::FM_RANSAC,this->paramRANSAC,this->paramConfidence,mask);
				
			// Convert fundamental matrix to CImg and push back to output slot
			cimg_library::CImg<double> fmCImg = openCvUtil::Mat2CImg<double>(fm);

			// Save fundamental matrix
			IndexedMatrix fmIndexed = fmCImg;
			fmIndexed.setIndex(i,ii);
			this->fundamentalMatrix().push_back(fmIndexed);

			// Save inverted fundamental matrix
			IndexedMatrix fmIndexedTransposed = fmCImg.get_transpose();
			fmIndexed.setIndex(ii,i);
			this->fundamentalMatrix().push_back(fmIndexedTransposed);

			if(this->logDetails())
				ImgTool::printInfo(sout,fmCImg,"	Fundamental matrix: ");

			// Check rank of F which should be 2
			if(this->logDetails())
			{
				cimg_library::CImg<T> U,S,V;
				fmCImg.SVD(U,S,V);
				sout << "\tSingular values {s1, s2, s3} = {" << S[0]
						<< ", " << S[1] << ", " << S[2] << "}" << std::endl;
				sout << "\ts1*s3/(s2*s2) = " << (S[0]*S[2]) / (S[1]*S[1]) << std::endl;
			}

			const cimg_library::CImg<float> maskCImg = openCvUtil::Mat2CImg<float>(mask);
			// Mark outliers in CorrepondenceList using the mask
			auto it4 = list.get<imgIndex>().lower_bound(i) ;
			auto it4End = list.get<imgIndex>().upper_bound(i) ;
			for(int j=0; it4 != it4End ; it4++, j++){
				for(int k=0; k<maskCImg.height(); k++){
					if(indexList[k] == it4->pointIndex && maskCImg(0,k) == 0 && i == 0)
					{
						const_cast<Correspondence<T>& >(*it4).outlier = true;
					}
					else if(indexList[k] == it4->pointIndex && maskCImg(0,k) == 1
						&& it4->outlier == true)
					{	
						const_cast<Correspondence<T>& >(*it4).outlier = false;
						outliers--;
						outliersTemp--;
						reversedInliers++;
						reversedInliersTemp++;
					}
				}
			}
			auto it5 = list.get<imgIndex>().lower_bound(ii) ;
			auto it5End = list.get<imgIndex>().upper_bound(ii) ;
			for(int j=0; it5 != it5End ; it5++, j++){
				for(int k=0; k<maskCImg.height(); k++){
					if(indexList[k] == it5->pointIndex && maskCImg(0,k) == 0)
					{
						const_cast<Correspondence<T>& >(*it5).outlier = true;
						outliers++;
						outliersTemp++;
						if(this->logDetails())
						{
							sout << "feat1: " << feat1[k].x << "\t" << feat1[k].y << "\t" << indexList[k] << "\toutlier" << std::endl;
							sout << "feat2: " << feat2[k].x << "\t" << feat2[k].y << "\t" << indexList[k] << "\toutlier" << std::endl;
						}
					}
					else if(indexList[k] == it5->pointIndex && maskCImg(0,k) == 1)
					{
						const_cast<Correspondence<T>& >(*it5).outlier = false;
						if(this->logDetails())
						{
							sout << "feat1: " << feat1[k].x << "\t" << feat1[k].y << "\t" << indexList[k] << "\tinlier" << std::endl;
							sout << "feat2: " << feat2[k].x << "\t" << feat2[k].y << "\t" << indexList[k] << "\tinlier" << std::endl;
						}
					}
				}
			}
			/*
			sout << "\t" << outliersTemp << " outliers of " << nPointsTemp << " correspondences found "
				<< "between image " << i << " and " << ii << std::endl;
			sout << "\t" << reversedInliersTemp << " points could be reversed from outlier to inlier "
				<< "between image " << i << " and " << ii
				<< " since good matches could be found in other image pairs." << std::endl;
			sout << "\t" << nPointsTemp-outliersTemp << " inliers remaining "
				<< "between image " << i << " and " << ii << std::endl;

			time_end = time(0);
			sout << "\t" << "Took " << difftime(time_start, time_end) << " seconds to calculate"
				<< " fundamental matrix." << std::endl;
			sout << std::endl;
			*/
		}
	}

	sout << std::endl;
	sout << "------------------------------------------" << std::endl;
	sout << "OVERALL STATS:" << std::endl;
	sout << outliers << " correspondence outliers of " << nAllPoints << " feature points found." << std::endl;
	sout << reversedInliers << " points could be reversed from outlier to inlier since "
		<< "good matches could be found in other image pairs." << std::endl;
	sout << nAllPoints-outliers << " inliers remaining." << std::endl;
	sout << "------------------------------------------" << std::endl;
	sout << std::endl;

	for(size_t i=0; i<list.sizeImages(); i++){
		try{
			sout << "Frame " << i << ": " << list.sizePoints(i) << std::endl;
		}
		catch(...){
			sout << "Frame " << i << " could not be found. Be aware of wrong indices" << std::endl;
		}
	}

	sout << "-----------------------------------------" << std::endl;
			
	sout << "list.sizePoints() = " << list.size() << std::endl;
		
	// Push back inlier list to output slot
	this->correspondenceInlier() = list.removeOutliers();
	sout << "inlierList.size() = " << this->correspondenceInlier().size() << std::endl;

	cend = time(0);
	sout << "Calculating all fundamental matrices took " << difftime(cend, cstart) << " seconds." << std::endl;

}

template <typename T>
cv::Point2d convertVector(const Correspondence<T> &vec){
	cv::Point2d vecOut;
	vecOut.x = (double)vec.x();
	vecOut.y = (double)vec.y();
	return vecOut;
}

#endif /* _FUNDAMENTALMATRIX_HXX_ */
