/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file ImagePathReader.hxx
 *  Implementation of the parameter class ImagePathReader.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.02.2012
 */

#ifndef _IMAGEPATHREADER_HXX_
#define _IMAGEPATHREADER_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/ImagePathReader.h>
#include <regex>
#include <iostream>

template <typename T>
ImagePathReader<T>::ImagePathReader(const std::string& name) :
		TemplatedParameteredObject<T>(
			"ImagePathReader", name,
			"<h2>Reads in all image paths of a specific image "
			"format</h2><br>"
			"Reads in all image paths of a specific image format. Suppose "
			"one selects an JPG image inside a folder, the plugin will save "
			"every path of a JPG image which is inside that folder to a "
			"vector<string>"
		)
{

	ParameteredObject::_addOutputSlot(
		imagePaths, "imagePaths",
		"imagePaths",
		"vector<string>");

	ParameteredObject::_addParameter(
		imageDir, "referenceFile",
		"Reference file to load a sequence of files.",
		"fileopen");

	ParameteredObject::_addParameter< int >(
		startFrame, "startFrame",
		"ID of first frame to read in.", 0);

	ParameteredObject::_addParameter< int >(
		endFrame, "endFrame",
		"ID of last frame to read in. Default is '-1' which means"
		" read until end of sequence.", -1);

	ParameteredObject::_addParameter< bool >(
		ignorePattern, "ignorePattern",
		"Ignore number pattern for read in", true);

	ParameteredObject::_addParameter< bool >(
		throw_error, "throw_error",
		"Throw exception if error while read", true);

	ParameteredObject::_addParameter< bool >(
		verbose, "verbose",
		"Print log messages", false);

	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void ImagePathReader<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	// Get working directory (should be the one containing the *.wrp file)
	std::string workingPath = FileTool::getCurrentDir();
	//sout << "workingPath = " << workingPath << std::endl;
	
	// Read in images -------------------------------------------
	//sout << "Reading in folder:" << std::endl;

	// Read in specified dir
	std::string dir = this->imageDir() ;
	sout << dir << std::endl;
	sout << this->imageDir() << std::endl;
	FileTool::slashConvert(dir);
	if(dir==""){
		throw(std::runtime_error("No file of directory given."));
	}
	else
		sout << "	Choose " << dir 
			<< " as sequence directory:" << std::endl;


	// Get file path
	this->filepath = dir.substr(0,
				dir.find_last_of(FileTool::slash)+1);
	sout << "	File path = " << filepath << std::endl;


	// Get file name
	std::string filename = dir.substr(dir.find_last_of(FileTool::slash)+1,
				dir.length());
	sout << "	Reference file name = " << filename << std::endl;
	

	// Get file extension
	std::string ext = dir.substr(dir.find_last_of('.'), dir.length());
	sout << "	Filter for *" << ext << " files" << std::endl;


	// Change working directory
	FileTool::changeDir(filepath);
	//sout << "Working Dir: " << FileTool::getCurrentDir() << std::endl;


	// sprintf( str, ""


	// Read in files with specified extension
	this->filenames = FileTool::getFilesWithSuffix(ext);
	this->numberOfFiles = filenames.size();
	sout << "\t" << numberOfFiles << " files found:" << std::endl;

	// Get first frame
	int start = std::max( 0, this->startFrame() );

	// Get last frame
	int end = std::max( start, int(this->endFrame()) );

	if( this->endFrame() < 0 )
		end = numberOfFiles;
	

	if( !this->ignorePattern() ){

		// Clear output slot (for the case the workflow is executed more than once)
		this->imagePaths().clear();
		this->imagePaths().assign(end-start+1,"");

		sout << "\tRead in frame(s) " << start << " to " << end << std::endl;
		sout << "\tWrite into array " << 0 << " to " << end-start+1 << std::endl;
	
		int n_files = 0;

		// Write found files to sequence
		for( size_t i=0; i<numberOfFiles; i++ ){
			int current_id = -1;
			decodeFilename( filenames[i], current_id );
			if( current_id >= start && current_id <= end ){
				this->imagePaths()[current_id-start] = (filepath+filenames[i]).c_str();
				//this->imagePaths()[i] = (filepath+filenames[i]).c_str();
				if( this->verbose() ) sout << "	img" << i << ": " << filenames[i] << " at position " << current_id-start << std::endl;
				n_files++;
			}
		}

		if(!n_files && this->throw_error()){
			throw(std::range_error("ImagePathReader<T>::execute(): Could not read in any Images."));
		}

	}
	else{

		sout << "Ignore file pattern" << std::endl;

		this->imagePaths().clear();

		if( end > numberOfFiles )
			end = numberOfFiles;

		// Write found files to sequence
		for( size_t i=0; i<end; i++ ){
			this->imagePaths().push_back( (filepath+filenames[i]).c_str() );
			if( this->verbose() ) sout << filenames[i] << std::endl;
		}

	}
	
	// Reset working directory to that one containing the *.wrp file
	FileTool::changeDir( workingPath );

}

#endif /* _IMAGEPATHREADER_HXX_ */



/// Decodes the given name of an image file from a sequence into its
/// prefix, index and suffix.
/**
\param filename The filename to decode.
\param index [out] The extracted index part of the filename converted to an integer number.
\param prefix [out] The extracted prefix part of the filename.
\param indexString [out] The extracted index part of the filename.
\param suffix [out] The extracted prefix part of the filename.
*/
void decodeFilename ( const std::string &filename, int &index)
{
	
    // decode filename
	std::regex e ("(\\d+)(\\D*)\\.");
	std::smatch sm;

	index = -1;

	if( std::regex_search(filename,sm,e)) {
		std::stringstream s;
		s << sm[1];
		//sout << sm[1] << std::endl;
		index = stoi(s.str());
	}
	/*
    // decode filename
	//std::basic_regex<char> indexExpression("(\\d+)(\\D*)(.bmp|.png|.jpg|.exr|.cimg)");
	std::regex e (".*[_a-z0]*([0-9]+).*\.[a-z]*");
	//std::regex e ("(\\d+)\\D*\\.\\D*");
	
	index = -1;

	std::cmatch cm;    // same as std::match_results<const char*> cm;
	if( std::regex_match(filename.c_str(),cm,e) ){
		//sout << "found match: " << cm[1] << std::endl;
		std::stringstream s;
		s << cm[1];
		index = stoi(s.str());
	}
	*/
}