/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file InitialFrameSelection.hxx
 *  Implementation of the parameter class InitialFrameSelection.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 27.07.2012
 */

#ifndef _INITIALFRAMESELECTION_HXX_
#define _INITIALFRAMESELECTION_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "InitialFrameSelection.h"
#include <charon-utils/ImgTool.hxx>


template <typename T>
InitialFrameSelection<T>::InitialFrameSelection(const std::string& name) :
		TemplatedParameteredObject<T>(
			"InitialFrameSelection", name,
			"<h2>Selects initial frames of an image sequence for SfM "
			"Reconstruction</h2><br>"
			"Selects three initial frames from an image sequence. Selection "
			"criterion is the number of correspondences between those "
			"frames, but also the mean distance between correspondences."
		)
{
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");

	ParameteredObject::_addOutputSlot(
		initialFrames, "initialFrames",
		"Vector of three initial frames used for SfM",
		"vector<int>");

	ParameteredObject::_addOutputSlot(
		distributionCorrs, "distributionCorrs",
		"Distribution of correspondences between images",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		distributionDisparity, "distributionDisparity",
		"Distribution of disparity between correspondences between images",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		nCorrs1, "nCorrs1",
		"Number of correspondences between frame 1 and 2",
		400, "int");
	ParameteredObject::_addParameter< int >(
		nCorrs2, "nCorrs2",
		"Number of correspondences between frame 1 and 3",
		300, "int");
	ParameteredObject::_addParameter< int >(
		nCorrs3, "nCorrs3",
		"Number of correspondences between frame 2 and 3",
		200, "int");

}

template <typename T>
void InitialFrameSelection<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	hekate::CorrespondenceList<T> corrs = this->correspondences();

	size_t nImg = corrs.sizeImages();

	cimg_library::CImg<int> nCorrs(nImg,nImg);
	cimg_library::CImg<float> nDisparity(nImg,nImg);
	nCorrs.fill(0);
	nDisparity.fill(0.);

	// Iterate over images
	for( size_t i=0; i<nImg; i++ )
	{
		
		// Iterate over points of current image
		auto it = corrs.get<imgIndex>().lower_bound(i) ;
		auto itEnd = corrs.get<imgIndex>().upper_bound(i) ;
		for(; it != itEnd; it++ )
		{
			// Iterate over points with current point id
			const size_t ptId = it->pointIndex;
			auto it2 = corrs.get<pointIndex>().lower_bound(ptId) ;
			auto it2End = corrs.get<pointIndex>().upper_bound(ptId) ;
			for(; it2 != it2End ; it2++)
			{
				if(it2->imgIndex != it->imgIndex){
					nCorrs(it2->imgIndex,i) += 1;
					// calculate euclidean distance for disparity
					float dx = it->x()-it2->x();
					float dy = it->y()-it2->y();
					float d = std::sqrt(dx*dx+dy*dy);
					nDisparity(it2->imgIndex,i) += d;
				}
			}
		}
	}

	// Normalize disparity by number of corrs
	for( size_t i=0; i<nImg; i++ )
	{
		for( size_t j=0; j<nImg; j++ )
		{
			if(nCorrs(i,j) != 0 )
				nDisparity(i,j) /= nCorrs(i,j);
		}
	}
	
	ImgTool::printInfo(sout,nDisparity,"nDisparity");
	ImgTool::printInfo(sout,nCorrs,"nCorrs");

	// Try to find initial frames ----------------------------------------------------------
	sout << "Start evaluation to find initial frames..." << std::endl;

	bool initialFrameFound = false;
	
	// If try to find initial frames is not successful, reduce number of correspondences
	// considered between frames for next iteration
	int nC1 = this->nCorrs1();
	int nC2 = this->nCorrs2();
	int nC3 = this->nCorrs3();
	for( size_t m=1; m<4; m++){
		
		if(m>1){
			sout << "Reducing number of demanded correspondences between frames" << std::endl;
			nC1 /= m;
			nC2 /= m;
			nC3 /= m;
			sout << "nCorrs1 = " << nC1 << std::endl;
			sout << "nCorrs2 = " << nC2 << std::endl;
			sout << "nCorrs3 = " << nC3 << std::endl;
		}

		// Iterate over every image until initial frames are found
		for( size_t i=0; i<nImg; i++){

			sout << "Try to choose image " << i << " as first initial frame"
				<< " ----------------------------" << std::endl;

			cimg_library::CImg<float> dispSorted;
			cimg_library::CImg<int> permutation;
			dispSorted = nDisparity.get_crop(0,0,nImg-1,0).sort(permutation,0);

			// Iterate over remaining images and look for second image with maximum(disp)
			for(size_t j=0; j<nImg; j++){

				int secondId = permutation(j,0);

				// Do not compare frame with itself
				if(secondId == i)
					continue;

				// Continue if current image pair has too less correspondences
				if( nCorrs(secondId,i) < nC1 )
					continue;

				// Continue if disparity between selected first and second frame is 0
				if(nDisparity(secondId,i) == 0)
					continue;

				sout << "Try to choose image " << secondId << " as second initial frame" << std::endl;
				sout << "nCorrs(" << i << "," << secondId << ") = " << nCorrs(secondId,i)
					<< " / nDisparity = " << nDisparity(secondId,i) << std::endl;

				// Image with energy values regarding number of correspondences between
				// image 3 and 2 | image 3 and 1
				cimg_library::CImg<float> energyThirdFrame(nImg,1);
				energyThirdFrame.fill(0);
			
				// Iterate over remaining images and look for a third image where
				// the disparity between the third image and image 1 AND 2 is as big as possible
				for( size_t k = 0; k<nImg; k++){
				
					// Do not compare frames with eachother
					if(k==i || k == secondId)
						continue;
				
					// Continue if third selected frame does not have enough correspondences
					// with frame 1 and 2
					if( nCorrs(k,i) < nC2 || nCorrs(k,secondId) < nC3 )
						continue;

					energyThirdFrame(k,0) = nDisparity(k,i) * nDisparity(k,secondId);

				}

				// Sort energyThirdFrame and evaluate entries with high value first
				cimg_library::CImg<int> permutation2;
				energyThirdFrame.sort(permutation2,0);
				int thirdId = permutation2(0,0);

				// Select initial frames if disparities > 0 and nCorrs > min
				if(energyThirdFrame(0,0) >= 0.001){
					sout << "Choose image " << thirdId << " as third initial frame" << std::endl;
					int firstId = i;
					if(firstId > secondId){
						int firstId_temp;
						firstId = secondId;
						secondId = firstId_temp;
					}
					sout << "----------------------------------" << std::endl;
					sout << "Selected initial frames: (" << firstId << ", " << secondId 
							<< ", " << thirdId << ")" << std::endl;
					sout << "nCorrs(" << firstId << "," << secondId << ") = " << nCorrs(secondId,firstId)
							<< " / nDisparity = " << nDisparity(secondId,i) << std::endl;
					sout << "nCorrs(" << firstId << "," << thirdId << ") = " << nCorrs(thirdId,firstId)
							<< " / nDisparity = " << nDisparity(thirdId,firstId) << std::endl;
					sout << "nCorrs(" << secondId << "," << thirdId << ") = " << nCorrs(thirdId,secondId)
							<< " / nDisparity = " << nDisparity(thirdId,secondId) << std::endl;

					this->initialFrames().push_back(firstId);
					this->initialFrames().push_back(secondId);
					this->initialFrames().push_back(thirdId);
					initialFrameFound = true;
					break;
				}
				else
					sout << "Disparities between 1-3 and 1-2 are too small...";

			}

			if(initialFrameFound)
				break;

		}

		if(initialFrameFound)
			break;

	}

	if(!initialFrameFound){
		sout << "ATTENTION: No good initial frames could be found!" << std::endl;
		sout << "Choosing default values 0,1 and 2" << std::endl;
		this->initialFrames().push_back(0);
		this->initialFrames().push_back(1);
		this->initialFrames().push_back(2);
	}
	
	this->distributionCorrs().push_back(nCorrs.get_normalize(0,255).resize(500,500));
	this->distributionDisparity().push_back(nDisparity.get_normalize(0,255).resize(500,500));

}

#endif /* _INITIALFRAMESELECTION_HXX_ */
