/*  Copyright (C) 2011 Daniel Kondermann

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file ManualCorrespondences.hxx
 *  Implementation of the parameter class ManualCorrespondences.
 *  \author <a href="mailto:daniel@kondermann.de">
 *      Daniel Kondermann</a>
 *  \date 11.10.2011
 */

#ifndef _MANUALCORRESPONDENCES_HXX_
#define _MANUALCORRESPONDENCES_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/ManualCorrespondences.h>
#include <exception>
#include <hekate/Correspondence.hxx>

using namespace std;
using namespace cimg_library;

template <typename T>
ManualCorrespondences<T>::ManualCorrespondences(const std::string& name) :
		TemplatedParameteredObject<T>(
			"ManualCorrespondences", name,
			"<h2>Tool to annotate manual correspondences in "
			"sequences</h2><br>"
			"Works both from left to right" 
			"as well as through time or on single images.<br>"
			" Run workflow threaded in tuchulcha to see the console output."
		),
		right(true, false)
{
	ParameteredObject::_addInputSlot(
		left, "left",
		"Left sequence",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		right, "right",
		"Right sequence.",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"2D correspondences"
		"correspondences for the left input sequence will have the imgIndex #framenumber,<br>"
		"correspondences for the right input sequence will have the imgIndex #framenumber + 1",
		"CorrespondenceList");

}

template <typename T>
void ManualCorrespondences<T>::drawFrame(int frame, CImg<T> &img)
{
	using namespace hekate ;
	img.resize(-100, -100, -100, 3); //color channel
	img.normalize(0,255) ;
	/*
	T m, M;
	m =img.min_max(M);
	*/
	int ll =3;
	T color[] ={255, 0, 0};

	//iterate over all Correspondences for the current frame
	CorrespondenceList<T>::ListByImg::iterator it = correspondences().get<hekate::imgIndex>().lower_bound(frame) ;
	for(; it != correspondences().get<hekate::imgIndex>().upper_bound(frame) ; it++)
	{
		int x = it->x() ;
		int y = it->y() ;
		img.draw_line(x-ll, y, x+ll, y, color);
		img.draw_line(x, y-ll, x, y+ll, color);
	}
}

template <typename T>
void ManualCorrespondences<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	CImgDisplay dsp, dspref;
	bool singleMode = !right.connected() ; //modus for selecting pixel coordinates in a single image
	
	if(left().size() != 1 || (!singleMode && right().size() != 1))
		raise("Only gray scale images allowed (CImgList<>.size()==1)!");
	if(!singleMode && (left()[0].spectrum() != right()[0].spectrum()))
		raise("Sequences vary in temporal length (left.spectrum()!=right.spectrum())!");
	if(!singleMode && (left()[0].depth() != 1 || right()[0].depth() != 1))
		raise("Sequences have depth (left.depth()!=1 || right.depth()!=1)!");

	sout << "Press F to switch between correspondence mode (forward or leftright). Current mode is leftright." << endl;
	sout << "Press L to switch between left and right input slot when in forward mode (no effect when in leftright mode). Standard value is left." << endl;
	bool lr =true; //left-right-matching (as opposed to forward/backward)
	bool l =true; //working on left sequence (as opposed to right)
	bool stop =false;
	int frame =0;
	int mode =0;
	size_t corrIndex = 0 ;
	int w =100;
	int wx1, wy1;
	float x1 =-1.0f, y1 =-1.0f;
	float x2 =-1.0f, y2 =-1.0f;
	float zf =1.0f; //zoom factor
	const CImg<T> &imgL =left()[0];
	CImg<T> dummy ;
	const CImg<T> &imgR = right.connected() ? right()[0] : dummy ;
	CImg<T> ref, cur;
	cur =imgL.get_channel(frame).resize(-100, -100, -100, 3);
	dsp.assign(cur);
	while(!dsp.is_keyESC())
	{
		dsp.wait_all();
		stop =false;
		if(mode == 0)
		{
			dsp.set_title("Select frame=L/R-Arrow, choose feature=Klick, ESC=end");
			if(dsp.button()) {
				mode =1;
				zf =2;
				x1 =dsp.mouse_x();
				y1 =dsp.mouse_y();
				sout << "Initial position is (" << x1 << "," << y1 << ")" << endl;
				wx1 =x1-w; //top left corner of window centered around feature
				wy1 =y1-w;
				cur =(l ? imgL : imgR).get_channel(frame).get_crop(wx1, wy1, x1+w, y1+w);
				cur.resize(-100*zf, -100*zf, -100, -100, 6); //lanczos resize of crop
				dsp.resize(cur);
				dsp.display(cur);
			} else if(dsp.is_keyARROWLEFT()) {
				frame =(frame != 0 ? --frame : imgL.spectrum() - 1);
				cur =(l ? imgL : imgR).get_channel(frame);
				drawFrame(frame, cur);
				dsp.resize(cur);
				dsp.display(cur);
				sout << "Frame: " << frame << (l ? " (left)" : " (right)") << endl;
			} else if(dsp.is_keyARROWRIGHT()) {
				frame =(frame+1)%(imgL.spectrum());
				cur =(l ? imgL : imgR).get_channel(frame);
				drawFrame(frame, cur);
				dsp.resize(cur);
				dsp.display(cur);
				sout << "Frame: " << frame << (l ? " (left)" : " (right)") << endl;
			} else if(dsp.is_keyL() && !singleMode) { //change left/right
				l =!l;
				cur =(l ? imgL : imgR).get_channel(frame);
				drawFrame(frame, cur);
				dsp.resize(cur);
				dsp.display(cur);
				sout << "Frame: " << frame << (l ? " (LEFT)" : " (RIGHT)") << endl;
			} else if(dsp.is_keyF()) { //change forward versus LR
				lr =!lr;
				sout << "Mode change to: " << (lr ? "Left/Right-Correspondences" : "Forward-Correspondences") << endl;
			} else if(dsp.is_keyESC()) {
				sout << "DONE!" << endl;
				break;
			}
			dsp.flush();
		}
		else if(mode == 1) {
			dsp.set_title("Zoom=klick, done=ENTER");
			if(dsp.button()) {
				int mx1 =dsp.mouse_x();
				int my1 =dsp.mouse_y();
				x1 =wx1+mx1/zf; //original position in full image
				y1 =wy1+my1/zf;
				sout << "New position is (" << x1 << "," << y1 << ")" << endl;

				zf++; //increase zoom
				wx1 =(int)(x1-w/zf); //top left corner of window centered around feature with new scale
				wy1 =(int)(y1-w/zf);
				cur =(l ? imgL : imgR).get_channel(frame).get_crop(wx1, wy1, x1+w/zf, y1+w/zf);
				cur.resize(-100*zf, -100*zf, -100, -100, 6); //lanczos resize of crop
				dsp.resize(cur);
				dsp.display(cur);
			} else if(dsp.is_keyENTER()) {
				mode =2;
				//create reference frame view
				ref =cur;
				ref(ref.width()/2, ref.height()/2, 0, 0) =ref.max();
				if(!singleMode)
				{	dspref.resize(ref);
					dspref.display(ref);
					dspref.set_title("Location in first image:");
				}
				zf =1;
				wx1 =wy1 =0;
				if(!singleMode && lr)
				{
					l = !l ;
					cur =imgR.get_channel(frame);
					sout << "Now choose feature-correspondence in right stereo view..." << endl;
				}
				else if(!singleMode && frame+1 < imgL.spectrum())
				{
					frame++;
					if(l)
					{
						cur =imgL.get_channel(frame);
						sout << "Now choose feature-correspondence in next temporal (left-slot) frame..." << endl;
					}
					else
					{
						cur =imgR.get_channel(frame);
						sout << "Now choose feature-correspondence in next temporal (right-slot) frame..." << endl;
					}
				}
				dsp.resize(cur);
				dsp.display(cur);
			} else if(dsp.is_keyESC()) {
				mode =0;
				zf =1;
				wx1 =wy1 =0;
				cur =(l ? imgL : imgR).get_channel(frame);
				drawFrame(frame, cur);
				dsp.resize(cur);
				dsp.display(cur);
				dsp.set_title("Select frame=L/R-Arrow, choose feature=Klick");
				sout << "cancelled input of current feature" << endl;
			}
			dsp.flush();
		}
		else if(mode == 2) {
			dsp.set_title("Find corr.: Zoom=Klick, done=ENTER");
			if(!singleMode && dsp.button()) {
				int mx1 =dsp.mouse_x();
				int my1 =dsp.mouse_y();
				x2 =wx1+mx1/zf; //original position in full image
				y2 =wy1+my1/zf;
				sout << "New position is (" << x2 << "," << y2 << ")" << " ";
				sout << "Displacement is (p1-p2)=(" << (x1-x2) << "," << (y1-y2) << ")" << endl;

				zf++; //increase zoom
				wx1 =(int)(x2-w/zf); //top left corner of window centered around feature with new scale
				wy1 =(int)(y2-w/zf);
				cur =(l ? imgL : imgR).get_channel(frame).get_crop(wx1, wy1, x2+w/zf, y2+w/zf);
				cur.resize(-100*zf, -100*zf, -100, -100, 6); //lanczos resize of crop
				dsp.resize(cur);
				dsp.display(cur);
			} else if(dsp.is_keyENTER()) {
				//store correspondence
				if(lr)
				{
					hekate::Correspondence<T> corL(x1,y1,0.0,1.0,corrIndex,frame) ;
					correspondences().push_back(corL) ;
					if(!singleMode)
					{	hekate::Correspondence<T> corR(x2,y2,0.0,1.0,corrIndex,frame+1) ;
						correspondences().push_back(corR) ;
					}
					
				}
				else if(l)
				{
					hekate::Correspondence<T> corL(x1,y1,0.0,1.0,corrIndex,frame) ;
					correspondences().push_back(corL) ;

				}
				else
				{
					if(!singleMode)
					{	hekate::Correspondence<T> corL(x1,y1,0.0,1.0,corrIndex,frame) ;
						correspondences().push_back(corL) ;
					}
				}
				corrIndex++ ;
				sout << "Stored correspondence, continue with next feature..." << endl;
				mode =0;
				zf =1;
				wx1 =wy1 =0;
				cur =(l ? imgL : imgR).get_channel(frame);
				drawFrame(frame, cur);
				dsp.resize(cur);
				dsp.display(cur);
				dsp.set_title("Select frame=L/R-Arrow, choose feature=Klick");
			} else if(dsp.is_keyESC()) {
				mode =0;
				zf =1;
				wx1 =wy1 =0;
				cur =(l ? imgL : imgR).get_channel(frame);
				drawFrame(frame, cur);
				dsp.resize(cur);
				dsp.display(cur);
				dsp.set_title("Select frame=L/R-Arrow, choose feature=Klick");
				sout << "cancelled input of current feature" << endl;
			}
			dsp.flush();
		}
	}
}

#endif /* _MANUALCORRESPONDENCES_HXX_ */
