/*  Copyright (C) 2011 Julian Wintermayr

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file ManualUpgrade.hxx
 *  Implementation of the parameter class ManualUpgrade.
 *  \author Julian Wintermayr
 *  \date 27.03.2011
 */

#ifndef _MANUALUPGRADE_HXX_
#define _MANUALUPGRADE_HXX_

#include <QApplication> 
#include <QThread>

#include "ManualUpgrade.h"
#include "qt/convToPixmap.hxx"
#include "MainWindow.h"

#include <sstream>

template <typename T>
ManualUpgrade<T>::ManualUpgrade(const std::string& name) :
		TemplatedParameteredObject<T>(
				"ManualUpgrade", name,
				"Define the plane at infinity, parallel lines and orthographic lines to receive much better results!"
		)

	
{
	ParameteredObject::_addInputSlot(imageSequence, "imageSequence", 
	"Add an imagelist as input. Only the first image in the list will be loaded! If there is no input, the program will quit.", 
	"CImgList<T>");
 
	ParameteredObject::_addOutputSlot(pointsVector, "pointsVector", 
	"Vector which contains points at infinity. Stored as two points representing a line on the plane at infinity.", 
	"std::vector<DrawingArea::drawnLines>");

   ParameteredObject::_addOutputSlot(parallelLinesVector, "parallelLinesVector", 
	"Vector which contains parallel lines. Stored as two points for each line and one integer as line group number.", 
	"std::vector<DrawingArea::drawnLines>");

   ParameteredObject::_addOutputSlot(orthographicLinesVector, "orthographicLinesVector", 
	"Vector which contains orthographic lines. Stored as three points for each right angle.", 
	"std::vector<DrawingArea::drawnLines>");
}

template <typename T>
void ManualUpgrade<T>::execute() {
   PARAMETEREDOBJECT_AVOID_REEXECUTION;
   ParameteredObject::execute();
   
   //this will check if the workflow was run from tuchulcha or from the command line
   //in the later case Qt can't be used at all and execution of the module will be skipped
   if(!qApp)
   {    return ;    }
   
   //this will detect if the workflow is executed in a separate thread
   if(QApplication::instance()->thread() != QThread::currentThread())
   {
       throw std::runtime_error("Threaded Exectuion not supported for this widget!") ;
   }

   //get image sequence as QImage
   QPixmap pix;

   const cimg_library::CImgList<T>& image = this->imageSequence();

   if(image.size() <= 0 && image[0].size() <= 0) {

       std::stringstream errorStr;
       errorStr << "Pluginname: " << this->getName() << ": input image is empty!" << std::endl;
       throw std::runtime_error(errorStr.str());
   }

   convToPixmap(image[0], pix);
   QImage img = pix.toImage();

   MainWindow window(img, &(pointsVector()), &(parallelLinesVector()), &(orthographicLinesVector()));
   window.exec(); 
}

#endif /* _MANUALUPGRADE_HXX_ */
