/*  Copyright (C) 2013 Michael Mueller

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file MergePointClouds.hxx
 *  Implementation of the parameter class MergePointClouds.
 *  \author <a href="mailto:michael@fds-team.de">
 *      Michael Mueller</a>
 *  \date 28.02.2013
 */

#ifndef _MERGEPOINTCLOUDS_HXX_
#define _MERGEPOINTCLOUDS_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "MergePointClouds.h"

#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/accumulators/statistics/median.hpp>
#include <boost/accumulators/statistics/variance.hpp>

#include "charon-utils/ImgTool.hxx"

#include <algorithm>

using namespace boost::accumulators;
using namespace cimg_library;

template <typename T>
MergePointClouds<T>::MergePointClouds(const std::string& name) :
		TemplatedParameteredObject<T>(
			"MergePointClouds", name,
			"<h2>Merges several point clouds in a single one</h2><br>"
			"Merges several point clouds in a single one"
		)
{
	ParameteredObject::_addInputSlot(pointCloudsIn, 	"pointCloudsIn", 	"serveral input point clouds",	"std::vector<CorrespondenceList>");
	ParameteredObject::_addOutputSlot(pointCloudsOut, 	"pointCloudsOut",	"one or more input point clouds","std::vector<CorrespondenceList>");
	ParameteredObject::_addInputSlot(cameraMatricesIn, 	"cameraMatricesIn", "cameraMatrices", "std::vector<IndexedMatrixList>");
	ParameteredObject::_addOutputSlot(cameraMatricesOut,"cameraMatricesOut","cameraMatrices", "std::vector<IndexedMatrixList>");

	ParameteredObject::_addOutputSlot(maxPointCloud, 	"maxPointCloud",	"max. result Point Cloud", "CorrespondenceList");
	ParameteredObject::_addOutputSlot(maxCameras,		"maxCameras",		"max. result Point Cloud cameras", "IndexedMatrixList");


	OutputSlot< typename hekate::CorrespondenceList<T> >  	maxPointCloud;
	OutputSlot< IndexedMatrixList > 						maxCameras;

	ParameteredObject::_addParameter (numCor, 			"numCor", 			"number of correspondences for each iteration", 5);
	ParameteredObject::_addParameter (maxIterations, 	"maxIterations", 	"maximal iterations", 							7500);
	ParameteredObject::_addParameter (inlierThreshold, 	"inlierThreshold", 	"The threshold for inlier determination", 		0.10);
	ParameteredObject::_addParameter (minMergeCount, 	"minMergeCount", 	"Minimum number of shared points two point clouds must have to be merged", 100);
}

template <typename T>
bool pointCloudCompareSize(const hekate::CorrespondenceList<T> &a, const hekate::CorrespondenceList<T> &b){
	return (a.size() > b.size());
}

// Simple sorting algorithm to sort the point clouds and cameras simultaneously
template <typename T>
void simplePointCloudSort(	std::vector<hekate::CorrespondenceList<T> > &unmergedClouds, 
							std::vector<IndexedMatrixList> 				&unmergedCameras){
	
	bool changed;

	do{

		changed = false;

		for(int i = 0; i < unmergedClouds.size()-1; i++){

			if (unmergedClouds[i].size() < unmergedClouds[i+1].size()){

				std::swap(unmergedClouds[i], 	unmergedClouds[i+1]);
				std::swap(unmergedCameras[i], 	unmergedCameras[i+1]);

				changed = true;

			} 

		}

	}while(changed);

}

template <typename T>
void MergePointClouds<T>::execute() {

	std::vector<hekate::CorrespondenceList<T> > unmergedClouds 	= pointCloudsIn();
	std::vector<IndexedMatrixList> 				unmergedCameras	= cameraMatricesIn();

	if (unmergedClouds.size() != unmergedCameras.size()){
		throw std::runtime_error("MergePointClouds: Camera and PointCloud size mismatch");
	}

	if(unmergedClouds.size() <= 0)
		return;

	if (unmergedClouds.size() == 1){
		pointCloudsOut() 	= unmergedClouds;
		cameraMatricesOut() = unmergedCameras;
		maxPointCloud()		= unmergedClouds[0];
		maxCameras()		= unmergedCameras[0];		
		return;
	}

	// Sort the point clouds descending by point cloud size
	simplePointCloudSort(unmergedClouds, unmergedCameras);

	std::vector<hekate::CorrespondenceList<T> > mergedClouds;
	std::vector<IndexedMatrixList>				mergedCameras;

	while (unmergedClouds.size() > 0){
		
		mergedClouds.push_back(unmergedClouds[0]);
		mergedCameras.push_back(unmergedCameras[0]);

		//Remove the currently biggest Point Cloud
		unmergedClouds.erase(unmergedClouds.begin());
		unmergedCameras.erase(unmergedCameras.begin());

		auto itPoint 	= unmergedClouds.begin();
		auto itCamera	= unmergedCameras.begin();

		while (itPoint != unmergedClouds.end()){

			// Try to merge output point cloud with the next smaller on
			if (mergeTwoPointClouds(*itPoint, mergedClouds.back(), *itCamera, mergedCameras.back())){
				itPoint 	= unmergedClouds.erase(itPoint);
				itCamera 	= unmergedCameras.erase(itCamera);
			}else{
				itPoint++;
				itCamera++;
			}

		}
		
	}

	sout << "Merging result: " << mergedClouds.size() << " point cloud(s) left" << std::endl;
	for (int i = 0; i < mergedClouds.size(); i++){
		sout 	<< "Point Cloud " << i << ": " << mergedClouds[i].size() << " points " 
				<< mergedCameras[i].size() << "cameras" << std::endl;
	}

	pointCloudsOut() 	= mergedClouds;
	cameraMatricesOut() = mergedCameras;

	// Find point cloud with the maximum number of points
	if(mergedClouds.size() > 0){

		int max = 0;

		for(int i = 1; i < mergedClouds.size(); i++){
			
			if (mergedClouds[i].size() > mergedClouds[max].size()){
				max = i;
			}

		}
		maxPointCloud()		= mergedClouds[max];
		maxCameras()		= mergedCameras[max];

	}

}

/* 
	mergeTwoPointClouds merges two point clouds and returns false, if it's not possible to merge them.
	The result will be stored in the second point/camera input parameter.
*/
template <typename T>
bool MergePointClouds<T>::mergeTwoPointClouds(	hekate::CorrespondenceList<T> &first, 
												hekate::CorrespondenceList<T> &second,
												IndexedMatrixList &firstCameras,
												IndexedMatrixList &secondCameras){

	sout << "Merging Pointcloud (" << first.size() << ") and Pointcloud(" << second.size() << ") " << std::endl;
	
	/* Find shared points between first and second point cloud */

	// Use a vector instead of CorrespondenceList as we can not easily access the elements 
	// in the order we added them to the CorrespondenceList. 
	std::vector< hekate::Correspondence<T> > commonPoints[2];

	/* MSVC 2010 does not support range based loops 
	for(auto &point : first){
	*/
	for(auto pointIt = first.begin(); pointIt != first.end(); pointIt++){
		
		auto &point = *pointIt;

		if (point.isOutlier())
			continue;

		if (second.existPtIndex(point.pointIndex)){

			hekate::Correspondence<T> secondPoint = second.getPointByIndex(point.pointIndex);

			if (secondPoint.isOutlier())
				continue;

			commonPoints[0].push_back(point);
			commonPoints[1].push_back(secondPoint);
		}

	}

	size_t pointCount = commonPoints[0].size();
	
	if (pointCount == second.size() || pointCount == first.size()){
		sout 	<< "Skip merging Pointcloud (" << first.size() << ") and Pointcloud(" << second.size() << ") " 
				<< "as we wouldn't get any new points" << std::endl;
		if (first.size() > second.size()){
			second = first;
		}
		return true;
	}

	if (pointCount < 5){
		sout 	<< "Stop merging, both point clouds do not have at least 5 shared points" << std::endl;
		return false;
	}

	if (pointCount < this->minMergeCount()){
		sout 	<< "Stop merging, both point clouds do not have at least " << this->minMergeCount() << " shared points" << std::endl;
		return false;		
	}

	size_t nCor;
	nCor = std::max(numCor(), 5);
	nCor = std::min(pointCount, nCor);

	std::vector< hekate::Correspondence<T> > selected_correspondences[2]; 
	std::vector<size_t> randomList(nCor);
	randomList.resize(nCor);

	srand( (unsigned) time(NULL) );

	unsigned int max_no_inliers = 0;

	CImg<T> H_ransac(4,4);
	CImg<T> H_final_rot(4,4);

	for(int iteration = 0; iteration < maxIterations(); iteration++){

		unsigned int no_inliers = 0;

		//Check if we have enough correspondences for a good random selection
		if (pointCount > nCor * 2){

			int i = 0;
			while(i < nCor){

				size_t randomIndex = rand() % pointCount;

				//Check if we already got this index in our list
				if(std::find(randomList.begin(), randomList.end(), randomIndex) != randomList.end()){
					continue;
				}

				randomList[i] = randomIndex;
				i++;
			}

		}else{

			for (int i = 0; i < numCor(); i++){
				randomList[i] = i;
			}

		}

		selected_correspondences[0].clear();
		selected_correspondences[1].clear();

		for(size_t i = 0; i < nCor; i++){
			selected_correspondences[0].push_back(commonPoints[0][randomList[i]]);
			selected_correspondences[1].push_back(commonPoints[1][randomList[i]]);
		}

		CImgList<double> centroids = accumulateCentroids(selected_correspondences[0], selected_correspondences[1], false);

		CImg<double> tr1 	= centroids(0);
		CImg<double> tr2 	= centroids(2); 
		CImg<double> sc1 	= centroids(1); 
		CImg<double> sc2 	= centroids(3); 
		CImg<double> T1 	= centroids(4); 
		CImg<double> T2 	= centroids(5);

		CImg<double> H, H_rot;

		H = getPointSetHomography(selected_correspondences[0], selected_correspondences[1], T1, T2);

		H_rot = H ;
		//De-Normalization of the H matrix

		CImg<double> T2t(4,4);
		T2t = T2.get_invert();

		H = T2t	* H;
		H = H 	* T1;

		//----------------------------------------------------------------------------------------------------------------
		// END OF NORMALIZED 5 POINT H MATRIX COMPUTATION

		// Symmetric transfer error computation
		//
		// Refer to Hartley, pg 124 for further details
		//
		//	pt1 - x
		//	pt2 - x'
		//	pt3 - Hinv*x'
		//	pt4 - H*x

		cimg_library::CImg<T> Hinv(4,4), pt1(1,4), pt2(1,4), pt3(1,4), pt4(1,4);
		double error1 = 0, error2 = 0, l1;
	
		double zErrorW = 1.0 ;//_depthErrorWeight() ;

		Hinv = H.get_invert();

		for(size_t point = 0; point < pointCount ; point++){

			for(int i = 0; i < 3; i++){
				pt1(0,i) = commonPoints[0][point].point[i];	
				pt2(0,i) = commonPoints[1][point].point[i];	
			}

			pt1(0,3) = 1;
			pt2(0,3) = 1;

			pt3 = Hinv 	* pt2;
			pt4 = H 	* pt1;

			pt3(0,0) = pt3(0,0) / pt3(0,3);
			pt3(0,1) = pt3(0,1) / pt3(0,3);
			pt3(0,2) = pt3(0,2) / pt3(0,3);
			pt3(0,3) = 1;

			pt4(0,0) = pt4(0,0) / pt4(0,3);
			pt4(0,1) = pt4(0,1) / pt4(0,3);
			pt4(0,2) = pt4(0,2) / pt4(0,3);
			pt4(0,3) = 1;

			l1 		= 	std::pow((double)commonPoints[0][point].x() - commonPoints[0][point].x(), 2.0) + 
						std::pow((double)commonPoints[1][point].y() - commonPoints[1][point].y(), 2.0);

			error1 	= std::pow((double)pt1(0,0) - pt3(0,0), 2.0) + pow((double)pt1(0,1) - pt3(0,1), 2.0) + std::pow((double)pt1(0,2) - pt3(0,2), 2.0);
			error2 	= std::pow((double)pt2(0,0) - pt4(0,0), 2.0) + pow((double)pt2(0,1) - pt4(0,1), 2.0) + std::pow((double)pt2(0,2) - pt4(0,2), 2.0);

			double error_value = pow(error1 + error2, 0.5);

			if(l1 > 10.0){ 
				error_value *= 0.1 ;
			}

			//Checking with inlier consistency
			//sout << "HOMEST: error value " << error_value << std::endl;

			if(error_value < inlierThreshold()){
				no_inliers++;
			}

			
		}

		if (no_inliers > max_no_inliers){
			H_ransac 		= H;
			H_final_rot 	= H_rot;
			max_no_inliers 	= no_inliers;
		}

		// Check if we already got a homography which reprojects all points correctly
		if (no_inliers == pointCount)
			break;

	}

	sout << "RANSAC completed successfully after " << maxIterations() << " iterations with " << max_no_inliers << " inliers (of " << pointCount << " data sets)." << std::endl;

	// Add all missing points from Point Cloud 1 into Point Cloud 2
	size_t mergedPointCount = 0;

	//for(auto &point : first){
	for(auto pointIt = first.begin(); pointIt != first.end(); pointIt++){
		
		auto &point = *pointIt;

		if (!second.existPtIndex(point.pointIndex)){

			hekate::Correspondence<T> newPoint(point);

			CImg<T> pt(1,4);
			for(int i = 0; i < 3; i++){
				pt(0,i) = newPoint.point[i];		
			}

			pt(0,3) = 1;

			newPoint.point = H_ransac * pt;
			newPoint.point /= newPoint.point[3];

			second.push_back(newPoint);
			mergedPointCount++;
    	}
	}

	sout << "Merged " << mergedPointCount << " points" << std::endl;

	// Add all missing cameras from Point Cloud 1 into Point Cloud 
	size_t mergedCameraCount = 0;

	CImg<T> H_ransac_inverse = H_ransac.get_invert();

	//for (IndexedMatrix &firstCamera : firstCameras){
	for (auto firstCameraIt = firstCameras.begin(); firstCameraIt != firstCameras.end(); firstCameraIt++){

		auto &firstCamera = *firstCameraIt;

		size_t index = firstCamera.getIndex1();
		if(!secondCameras.exists( index ) && !firstCamera.degenerated){
			
			/* 	The multiplication is not well defined for Indexed Matrix
			 	Only the CImg part will be multiplcated and the other
				values will get lost, so we have to set them again 			*/

			IndexedMatrix secondCam = firstCamera * H_ransac_inverse;
			secondCam.setIndex(index, index);
			secondCam.f = firstCamera.f;

			secondCameras.push_back(secondCam);

			mergedCameraCount++;
		}

	}
	sout << "Merged " << mergedCameraCount << " cameras" << std::endl;

	sout << "Successfully merged to Pointcloud(" << second.size() << ") " << std::endl;
	return true;
}

template <typename T>
CImgList<double> MergePointClouds<T>::accumulateCentroids(	std::vector< hekate::Correspondence<T> > &pointCloud1, 
															std::vector< hekate::Correspondence<T> > &pointCloud2, bool evMedian){

	CImgList<double> result(6,4,4,1,1,0) ;

	CImg<double> &tr1	= result(0);
	CImg<double> &sc1	= result(1); 
	CImg<double> &tr2	= result(2); 
	CImg<double> &sc2	= result(3); 
	CImg<double> &T1 	= result(4); 
	CImg<double> &T2 	= result(5);

	typedef accumulator_set<T, stats<boost::accumulators::tag::mean, boost::accumulators::tag::variance, boost::accumulators::tag::median> > accum_set;

	accum_set accCentroid1[3]; 				//stats for xyz-coordinates of point cloud 1
	accum_set accCentroid2[3]; 				//stats for xyz-coordinates of point cloud 2
	accum_set accDifferenceCentroid[3]; 	//stats for differences of xyz-coordinates
	accum_set accScale1; 					//stats for vector lengths of point cloud 1
	accum_set accScale2; 					//stats for vector lengths of point cloud 2

	if(pointCloud1.size() != pointCloud2.size())	
		return result;
	

	for (int pointIndex = 0; pointIndex < pointCloud1.size(); pointIndex++){

		for (int i = 0; i < 3; i++){

			accCentroid1[i](pointCloud1[pointIndex].point[i]);
			accCentroid2[i](pointCloud2[pointIndex].point[i]);

			accDifferenceCentroid[i](pointCloud1[pointIndex].point[i] - pointCloud2[pointIndex].point[i]);

		}

	}

	double centroidMedian_1[3] 		= { median(accCentroid1[0]), median(accCentroid1[1]), median(accCentroid1[2]) };
	double centroidMedian_2[3] 		= { median(accCentroid2[0]), median(accCentroid2[1]), median(accCentroid2[2]) };
	double centroidMedian_diff[3] 	= { median(accDifferenceCentroid[0]), median(accDifferenceCentroid[1]), median(accDifferenceCentroid[2]) };

	double centroidMean_1[3] 		= { mean(accCentroid1[0]), mean(accCentroid1[1]), mean(accCentroid1[2]) };
	double centroidMean_2[3] 		= { mean(accCentroid2[0]), mean(accCentroid2[1]), mean(accCentroid2[2]) };
	double centroidMean_diff[3] 	= { mean(accDifferenceCentroid[0]), mean(accDifferenceCentroid[1]), mean(accDifferenceCentroid[2]) };

	for (int pointIndex = 0; pointIndex < pointCloud1.size(); pointIndex++){

		double length1, length2;

		if(evMedian){

			length1 = sqrt(		pow(pointCloud1[pointIndex].x() - centroidMedian_1[0], 2.0) +
								pow(pointCloud1[pointIndex].y() - centroidMedian_1[1], 2.0) +
								pow(pointCloud1[pointIndex].z() - centroidMedian_1[2], 2.0));

			length2 = sqrt(		pow(pointCloud2[pointIndex].x() - centroidMedian_2[0], 2.0) +
								pow(pointCloud2[pointIndex].y() - centroidMedian_2[1], 2.0) +
								pow(pointCloud2[pointIndex].z() - centroidMedian_2[2], 2.0));
		}else{

			length1 = sqrt(		pow(pointCloud1[pointIndex].x() - centroidMean_1[0], 2.0) +
								pow(pointCloud1[pointIndex].y() - centroidMean_1[1], 2.0) +
								pow(pointCloud1[pointIndex].z() - centroidMean_1[2], 2.0));

			length2 = sqrt(		pow(pointCloud2[pointIndex].x() - centroidMean_2[0], 2.0) +
								pow(pointCloud2[pointIndex].y() - centroidMean_2[1], 2.0) +
								pow(pointCloud2[pointIndex].z() - centroidMean_2[2], 2.0));

		}

		accScale1(length1) ;
		accScale2(length2) ;
	}
	/*
	sout	<< "scales:\n\tmean\tstddev\tmedian\n" 
			<< "1\t" << mean(accScale1) << "\t" << sqrt(variance(accScale1)) << "\t" << median(accScale1) << "\n"
			<< "2\t" << mean(accScale2) << "\t" << sqrt(variance(accScale2)) << "\t" << median(accScale2) << "\n"
			<< std::endl;
	*/

	tr1.identity_matrix();
	tr2.identity_matrix();
	sc1.identity_matrix();
	sc2.identity_matrix();

	for(int i = 0; i < 3; i++){

		if(evMedian){
			tr1(3,i) = -centroidMedian_1[i];
			tr2(3,i) = -centroidMedian_2[i];
		}else{
			tr1(3,i) = -centroidMean_1[i];
			tr1(3,i) = -centroidMean_2[i];
		}

	}

	if(evMedian){
		sc1 *= (sqrt(2.00) / median(accScale1));
		sc2 *= (sqrt(2.00) / median(accScale2));
	}else{
		sc1 *= (sqrt(2.00) / mean(accScale1));
		sc2 *= (sqrt(2.00) / mean(accScale2));
	}

	sc1(3,3) = 1.0 ;	
	sc2(3,3) = 1.0 ;

	T1 = sc1 * tr1;
	T2 = sc2 * tr2;

	return result;
}

template <typename T>
CImg<double> MergePointClouds<T>::getPointSetHomography(std::vector< hekate::Correspondence<T> > &pointCloud1, 
														std::vector< hekate::Correspondence<T> > &pointCloud2, 
														const CImg<double> T1, 
														const CImg<double> T2){

	std::vector<std::vector<double> > norm_correspondences;

	norm_correspondences.assign(pointCloud1.size(), std::vector<double>(6,0.0));

	size_t nCor = pointCloud1.size() ;
	//Finding and storing the Normalized correspondences

	for(size_t point = 0; point < pointCloud1.size(); point++){

		CImg<double> x1s_temp(1,4), x1n_temp(1,4), x2s_temp(1,4), x2n_temp(1,4);
		

		for(int i = 0; i < 3; i++){
			x1s_temp(0,i) = pointCloud1[point].point[i];
			x2s_temp(0,i) = pointCloud2[point].point[i];
		}
			
		x1s_temp(0,3) = 1;
		x2s_temp(0,3) = 1;

		x1n_temp = T1 * x1s_temp;
		x2n_temp = T2 * x2s_temp;

		for(int i = 0; i < 3; i++){
			norm_correspondences[point][i] 		= x1n_temp(0,i);
			norm_correspondences[point][i+3] 	= x2n_temp(0,i);
		}

	}

	// Generating the A matrix to solve Ax=0

	CImg<double> A(16, 3 * nCor);

	for(size_t point = 0; point < nCor; point++){

		A(0,  3 * point) = norm_correspondences[point][0];
		A(1,  3 * point) = norm_correspondences[point][1];
		A(2,  3 * point) = norm_correspondences[point][2];
		A(3,  3 * point) = 1;
		A(4,  3 * point) = 0;
		A(5,  3 * point) = 0;
		A(6,  3 * point) = 0;
		A(7,  3 * point) = 0;
		A(8,  3 * point) = 0;
		A(9,  3 * point) = 0;
		A(10, 3 * point) = 0;
		A(11, 3 * point) = 0;
		A(12, 3 * point) = -norm_correspondences[point][3]*norm_correspondences[point][0];
		A(13, 3 * point) = -norm_correspondences[point][3]*norm_correspondences[point][1];
		A(14, 3 * point) = -norm_correspondences[point][3]*norm_correspondences[point][2];
		A(15, 3 * point) = -norm_correspondences[point][3];
			
		A(0,  3 * point + 1) = 0;
		A(1,  3 * point + 1) = 0;
		A(2,  3 * point + 1) = 0;
		A(3,  3 * point + 1) = 0;
		A(4,  3 * point + 1) = norm_correspondences[point][0];
		A(5,  3 * point + 1) = norm_correspondences[point][1];
		A(6,  3 * point + 1) = norm_correspondences[point][2];
		A(7,  3 * point + 1) = 1;
		A(8,  3 * point + 1) = 0;
		A(9,  3 * point + 1) = 0;
		A(10, 3 * point + 1) = 0;
		A(11, 3 * point + 1) = 0;
		A(12, 3 * point + 1) = -norm_correspondences[point][4]*norm_correspondences[point][0];
		A(13, 3 * point + 1) = -norm_correspondences[point][4]*norm_correspondences[point][1];
		A(14, 3 * point + 1) = -norm_correspondences[point][4]*norm_correspondences[point][2];
		A(15, 3 * point + 1) = -norm_correspondences[point][4];
		
		A(0,  3 * point + 2) = 0;
		A(1,  3 * point + 2) = 0;
		A(2,  3 * point + 2) = 0;
		A(3,  3 * point + 2) = 0;
		A(4,  3 * point + 2) = 0;
		A(5,  3 * point + 2) = 0;
		A(6,  3 * point + 2) = 0;
		A(7,  3 * point + 2) = 0;
		A(8,  3 * point + 2) = norm_correspondences[point][0];
		A(9,  3 * point + 2) = norm_correspondences[point][1];
		A(10, 3 * point + 2) = norm_correspondences[point][2];
		A(11, 3 * point + 2) = 1;
		A(12, 3 * point + 2) = -norm_correspondences[point][5]*norm_correspondences[point][0];
		A(13, 3 * point + 2) = -norm_correspondences[point][5]*norm_correspondences[point][1];
		A(14, 3 * point + 2) = -norm_correspondences[point][5]*norm_correspondences[point][2];
		A(15, 3 * point + 2) = -norm_correspondences[point][5];

	}

	//Computing the SVD of A to solve for Ax=0

	CImg<double> U, S, V;
	A.SVD(U,S,V, true, 200);
	//Assigning the values to the homography

	//sout << "smallest Singular values : " << S(0,13) << " " << S(0,14) << " " << S(0,15) << std::endl;
	cimg_library::CImg<double> H(4,4) ;
	
	H(0,0) = V(15,0);	H(1,0) = V(15,1);	H(2,0) = V(15,2);	H(3,0) = V(15,3);
	H(0,1) = V(15,4);	H(1,1) = V(15,5);	H(2,1) = V(15,6);	H(3,1) = V(15,7);
	H(0,2) = V(15,8);	H(1,2) = V(15,9);	H(2,2) = V(15,10);	H(3,2) = V(15,11);
	H(0,3) = V(15,12);	H(1,3) = V(15,13);	H(2,3) = V(15,14);	H(3,3) = V(15,15);

	return H ;
}

//This function defines the energy generating function for implementation of the Levenberg Marquardt algorithm.
void func1(double *p_lm, double *x_lm, int m, int n, void *adata){

	cimg_library::CImg<double> H(4,4), Hinv(4,4), pt1(1,4), pt2(1,4), pt3(1,4), pt4(1,4);
	double error1 = 0, error2 = 0;
	int i;

	// Assigning the H matrix from the parameter vector

	H(0,0) = p_lm[0];	H(1,0) = p_lm[1];	H(2,0) = p_lm[2];	H(3,0) = p_lm[3];
	H(0,1) = p_lm[4];	H(1,1) = p_lm[5];	H(2,1) = p_lm[6];	H(3,1) = p_lm[7];
	H(0,2) = p_lm[8];	H(1,2) = p_lm[9];	H(2,2) = p_lm[10];	H(3,2) = p_lm[11];
	H(0,3) = p_lm[12];	H(1,3) = p_lm[13];	H(2,3) = p_lm[14];	H(3,3) = p_lm[15];

	Hinv = H.get_invert();

	// Recalculating the parametrized energy vector

	for(i=0; i<6*n; i=i+6)
	{
			pt1(0,0) = ((double *)adata)[i];
			pt1(0,1) = ((double *)adata)[i+1];
			pt1(0,2) = ((double *)adata)[i+2];
			pt1(0,3) = 1;

			pt2(0,0) = ((double *)adata)[i+3];
			pt2(0,1) = ((double *)adata)[i+4];
			pt2(0,2) = ((double *)adata)[i+5];
			pt2(0,3) = 1;

			pt3 = Hinv*pt2;
			pt4 = H*pt1;

			pt3(0,0) = pt3(0,0)/pt3(0,3);
			pt3(0,1) = pt3(0,1)/pt3(0,3);
			pt3(0,2) = pt3(0,2)/pt3(0,3);
			pt3(0,3) = 1;

			pt4(0,0) = pt4(0,0)/pt4(0,3);
			pt4(0,1) = pt4(0,1)/pt4(0,3);
			pt4(0,2) = pt4(0,2)/pt4(0,3);
			pt4(0,3) = 1;

			error1 = (pt1(0,0)-pt3(0,0))*(pt1(0,0)-pt3(0,0)) + (pt1(0,1)-pt3(0,1))*(pt1(0,1)-pt3(0,1)) + (pt1(0,2)-pt3(0,2))*(pt1(0,2)-pt3(0,2));

			error2 = (pt2(0,0)-pt4(0,0))*(pt2(0,0)-pt4(0,0)) + (pt2(0,1)-pt4(0,1))*(pt2(0,1)-pt4(0,1)) + (pt2(0,2)-pt4(0,2))*(pt2(0,2)-pt4(0,2));

			x_lm[i / 6] = error1 + error2;
	}
}

#endif /* _MERGEPOINTCLOUDS_HXX_ */
