/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file MeshTriangulationPCL.hxx
 *  Implementation of the parameter class MeshTriangulationPCL.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 15.03.2012
 */

#ifndef _MESHTRIANGULATIONPCL_HXX_
#define _MESHTRIANGULATIONPCL_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "MeshTriangulationPCL.h"

template <typename T>
MeshTriangulationPCL<T>::MeshTriangulationPCL(const std::string& name) :
		TemplatedParameteredObject<T>(
			"MeshTriangulationPCL", name,
			"<h2>Generates a mesh using the PCL</h2><br>"
			"Generates a mesh by fast triangulation of unordered point "
			"cloud, using the point cloud library."
		)
{
	ParameteredObject::_addInputSlot(
		pointCloud, "pointCloud",
		"A 3D point cloud",
		"CorrespondenceList");

	ParameteredObject::_addInputSlot(
		cameraMatrices, "cameraMatrices",
		"cameraMatrices",
		"IndexedMatrixList");

	ParameteredObject::_addOutputSlot(
		pointIndex, "pointIndex",
		"Indices of point mesh",
		"vector<vector<int> >");

}

template <typename T>
void MeshTriangulationPCL<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	CorrespondenceList<T> points = this->pointCloud();
	normalizeTriangulatedPoints(points);
	const IndexedMatrixList &cameras = this->cameraMatrices();
	
	pcl::PointCloud<pcl::Normal>::Ptr normals (new pcl::PointCloud<pcl::Normal>);
	pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ>);
	auto it = this->pointCloud().begin();
	auto itEnd = this->pointCloud().end();
	for(; it != itEnd; it++)
	{
		if(!it->outlier)
		{
			pcl::PointXYZ pt;
			pt.x = it->x() ;
			pt.y = it->y() ;
			pt.z = it->z() ;
			cloud->push_back(pt);

			CImg<double> cam = cameras(it->imgIndex);			
			CImg<double> normal = cam.get_crop(0,0,2,2).det() * cam.get_crop(0,2,2,2).get_transpose();
			pcl::Normal normal_pcl;
			normal_pcl.normal_x = normal[0];
			normal_pcl.normal_y = normal[1];
			normal_pcl.normal_z = normal[2];
			normals->push_back(normal_pcl);
		}
	}


	sout << "Load " << cloud->size() << " 3D points to pcl::PointCloud" << std::endl;
	sout << cloud->points.size() << std::endl;
	//sout << cloud->points[0].x << cloud->points[0].y << cloud->points[0].z << std::endl;



	// Normal estimation*
	/*
	pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> n;
	pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ>);
	tree->setInputCloud (cloud);
	n.setInputCloud (cloud);
	n.setSearchMethod (tree);
	n.setKSearch (20);
	n.compute (*normals);
	*/
	//* normals should not contain the point normals + surface curvatures

	// Concatenate the XYZ and normal fields*
	pcl::PointCloud<pcl::PointNormal>::Ptr cloud_with_normals (new pcl::PointCloud<pcl::PointNormal>);
	pcl::concatenateFields (*cloud, *normals, *cloud_with_normals);
	//* cloud_with_normals = cloud + normals

	// Create search tree*
	pcl::search::KdTree<pcl::PointNormal>::Ptr tree2 (new pcl::search::KdTree<pcl::PointNormal>);
	tree2->setInputCloud (cloud_with_normals);

	// Initialize objects
	pcl::GreedyProjectionTriangulation<pcl::PointNormal> gp3;
	pcl::PolygonMesh triangles;

	// Set the maximum distance between connected points (maximum edge length)
	gp3.setSearchRadius (1.0);

	// Set typical values for the parameters
	gp3.setMu (2.5);
	gp3.setMaximumNearestNeighbors (100);
	gp3.setMaximumSurfaceAngle(M_PI/4); // 45 degrees
	gp3.setMinimumAngle(M_PI/18); // 10 degrees
	gp3.setMaximumAngle(2*M_PI/3); // 120 degrees
	gp3.setNormalConsistency(false);

	// Get result
	gp3.setInputCloud (cloud_with_normals);
	gp3.setSearchMethod (tree2);
	gp3.reconstruct (triangles);

	//triangles.polygons[0].vertices.

	// Additional vertex information
	std::vector<int> parts = gp3.getPartIDs();
	std::vector<int> states = gp3.getPointStates();

	//sout << triangles << std::endl;

	int nr_points  = triangles.cloud.width * triangles.cloud.height;
	int point_size = triangles.cloud.data.size () / nr_points;

	//sout << triangles.cloud.data.size() << std::endl;
	
	sout << "Found " << triangles.polygons.size() << " polygons." << std::endl;

	// Save triangles to output slot
	size_t triangle_size = triangles.polygons.size ();
	for (size_t i = 0; i < triangle_size; ++i)
	{
		std::vector<int> vertex;
		for (size_t j = 0; j < triangles.polygons[i].vertices.size (); ++j)
		{
			vertex.push_back( triangles.polygons[i].vertices[j] );
		}
		this->pointIndex().push_back( vertex );
	}

}

// Recover camera matrices by x, x', X
template <typename T>
void normalizeTriangulatedPoints(CorrespondenceList<T> &points){
	
	double dist = 0;
	int nPoints = 0;

	auto it = points.begin();
	auto itEnd = points.end();
	for(; it != itEnd; it++)
	{
		if(!it->outlier)
		{
			double x, y, z;
			x = it->x();
			y = it->y();
			z = it->z();

			dist += std::sqrt( x*x + y*y + z*z );
			nPoints++;
		}

	}

	dist /= nPoints;
	
	it = points.begin();
	for(; it != itEnd; it++)
	{
		it->point /= dist;
	}

}

#endif /* _MESHTRIANGULATIONPCL_HXX_ */
