/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file NVMReader.hxx
 *  Implementation of the parameter class NVMReader.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.02.2013
 */

#ifndef _NVMREADER_HXX_
#define _NVMREADER_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/NVMReader.h>
#include <charon-utils/ImgTool.hxx>

template <typename T>
NVMReader<T>::NVMReader(const std::string& name) :
		TemplatedParameteredObject<T>(
			"NVMReader", name,
			"<h2>Reads NVM files from Visual SfM</h2><br>"
			"Read NVM files from Visual SfM and forwards camera matrices and "
			"point lists. Principal points have to be added, since the NVM file format "
			"normally does not contain its values / VisualSFM uses calibration matrix "
			"K = [fx 0 0, 0 fy 0, 0 0 1]"
		)
{
	ParameteredObject::_addInputSlot(
		imagePathSource, "imagePathSource",
		"Paths to images used by the NVM file. Since the sorting of "
		"the used images is not necessarily in increasing order, the"
		"names saved in the NVM file will be compared with the given ones "
		"so all following pluings use consistent images.",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		cameras, "cameras",
		"cameras",
		"IndexedMatrixList");
	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");
	ParameteredObject::_addOutputSlot(
		points3D, "points3D",
		"points3D",
		"CorrespondenceList");
	
	ParameteredObject::_addParameter (filename, "filename",
		"Filename to read NVM from", "fileopen");
	ParameteredObject::_addParameter (width, "width",
		"Width of image to calculate principal point. The principal points are not known in NVM files.", 0, "int");
	ParameteredObject::_addParameter (height, "height",
		"Height of image to calculate principal point. The principal points are not known in NVM files.", 0, "int");
	
	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void NVMReader<T>::execute() {

	vector<CameraT> camera_data;
	vector<Point3D> point_data;
	vector<Point2D> measurements;
	vector<int> ptidx;
	vector<int> camidx;
	vector<string> names;
	vector<int> ptc;

	ifstream in;
	in.open( this->filename().c_str() );
	
	float ppx = float( this->width() / 2. ) ;
	float ppy = float( this->height() / 2. )  ;

	bool ppGiven = false;
	if( ppx != 0 && ppy != 0)
	{
		ppGiven = true;
		sout << "Chosen principal point: (" << ppx << ", " << ppy << ")" << std::endl;
	}
	else
		sout << "ATTENTION: No Principal point set. Cameras may not be compatible with other data." << std::endl;


	// Read NVM file -----------------------------------------------------------------
	sout << "Read in NVM file: " << this->filename() << std::endl;
	bool completed = LoadNVM(in, camera_data, point_data, measurements, ptidx, camidx, names, ptc);

	if( !completed )
	{
		sout << "NVM read in unsuccessful!" << std::endl;
		throw( std::runtime_error("NVM read in unsuccessful!") );
	}
	
	sout << "point_data.size()=" << point_data.size() << std::endl;
	sout << "ptidx.size()=" << ptidx.size() << std::endl;
	sout << "camidx.size()=" << camidx.size() << std::endl;
	sout << "names.size()=" << names.size() << std::endl;

	int maxx = 0;
	for( size_t i = 0; i < ptidx.size(); i++)
	{
		maxx = std::max(ptidx[i],maxx);
	}
	sout << "maxx = " << maxx << std::endl;


	// ---------------- read cameras ----------------------------------------------

	// Read image names from given image paths
	std::vector<std::string> imageNames;
	for( size_t i = 0; i < this->imagePathSource().size(); i++ )
	{
		std::string name = this->imagePathSource()[i].substr(this->imagePathSource()[i].find_last_of(FileTool::slash)+1,
				this->imagePathSource()[i].length());
		imageNames.push_back( name );
	}
	
	sout << "Read " << imageNames.size() << " image names from imagePathSource" << std::endl;
	sout << "Read " << names.size() << " cameras from NVM file" << std::endl;

	if( imageNames.size() < names.size())
		throw( std::range_error("NVMReader<T>::execute(): More cameras than given image files!") );

	// Create permutation between NVM cameras and given image list
	std::vector<unsigned int> permutation;
	permutation.resize(names.size());

	// Correct read in names given by NVM file
	for( size_t i = 0; i < names.size(); i++ )
	{		
		try
		{
			names[i] = names[i].substr(names[i].find_last_of(FileTool::slash)+1, names[i].length());
		}
		catch(...){}
	}
	
	// Compare image names with names of NVM file
	int count = 0;
	for( size_t i = 0; i < imageNames.size(); i++ )
	{
		for( size_t j = 0; j < names.size(); j++ )
		{
			if( names[j] == imageNames[i] )
			{
				permutation[j] = i;
				count ++;
			}
		}
	}
	// Check, if the images required by NVM file are given.
	if( count != names.size() )
	{
		sout << count << " NVM names found in given image list. But number of NVM count is << " << names.size() 
			<< ". Images in path list missing..." << std::endl;
		throw( std::range_error("NVMReader<T>::execute(): Missing images in image path list") );
	}
	

	// Convert vectors to IndexedMatrixList and CorrespondenceList -------------------

	for( size_t i = 0; i < camera_data.size(); i++)
	{
		CImg<float> Rt(4,3), K(3,3), Kc(3,3), cam_cimg;

		// Get rotation and translation (R|t)
		for(int k = 0; k < 3; ++k)
		{
			for(int j = 0; j < 3; ++j)
			{
				Rt(j,k) = camera_data[i].m[k][j];
			}
			Rt(3,k) = camera_data[i].t[k];
		}

		// Get calibration K(f,0,0; 0,f,0; 0,0,1)
		K.fill(0);
		K(0,0) = camera_data[i].f ;
		K(1,1) = camera_data[i].f ;
		K(2,2) = 1;

		// Define additional calibration to add principal point
		if(ppGiven)
		{
			Kc.fill(0);
			Kc(0,0) = 1;
			Kc(2,0) = ppx;
			Kc(1,1) = 1;
			Kc(2,1) = ppy;
			Kc(2,2) = 1;
			K = Kc * K;
		}
		
		// Get K(R|t)
		cam_cimg = K * Rt;

		// Create IndexedMatrix
		IndexedMatrix cam(cam_cimg);
		cam.setIndex( permutation[i] );

		cam.f = camera_data[i].f;
		
		cam.radial = camera_data[i].radial ;

		cam.distortion_type = 0;
		if( camera_data[i].distortion_type != 4294967295 )
			cam.distortion_type = camera_data[i].distortion_type ;

		this->cameras().push_back( cam );

		//ImgTool::printInfo( sout, cam_cimg, "cam");

	}
	// Sort cameras in output slot by increasing index
	const IndexedMatrixList camerasUnsorted = this->cameras();
	for( size_t i = 0; i < camerasUnsorted.size(); i++ )
	{
		// Write matrix from position i in camerasUnsorted to postion .getIndex1() in cameras()
		this->cameras()[ camerasUnsorted[i].getIndex1() ] = camerasUnsorted[i] ;
	}

	// Get 3D points
	for( size_t i = 0; i < point_data.size(); i++)
	{
		Correspondence<T> pt3D(	
					T( point_data[i].xyz[0]),
					T( point_data[i].xyz[1]),
					T( point_data[i].xyz[2]),
					T(1.),
					i,
					0,
					0,
					float(ptc[i*3]),
					float(ptc[i*3 + 1]),
					float(ptc[i*3 + 2])
			);
		this->points3D().push_back( pt3D );
	}

	// Get 2D Points
	for( size_t i = 0; i < measurements.size(); i++)
	{
		float x = measurements[i].x;
		float y = measurements[i].y;
		if(ppGiven)
		{
			x += ppx;
			y += ppy;
		}

		Correspondence<T> pt2D(	
					T(x),
					T(y),
					T(0),
					T(1.),
					ptidx[i],
					permutation[ camidx[i] ],
					0,
					float(ptc[ptidx[i]*3]),
					float(ptc[ptidx[i]*3 + 1]),
					float(ptc[ptidx[i]*3 + 2])
			);
		this->correspondences().push_back( pt2D );
	}

}

#endif /* _NVMREADER_HXX_ */
