/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file NormalizeLocal.hxx
 *  Implementation of the parameter class NormalizeLocal.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 16.10.2013
 */

#ifndef _NORMALIZELOCAL_HXX_
#define _NORMALIZELOCAL_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "NormalizeLocal.h"

template <typename T>
NormalizeLocal<T>::NormalizeLocal(const std::string& name) :
		TemplatedParameteredObject<T>(
			"NormalizeLocal", name,
			"<h2>Normalizes image by local histograms</h2><br>"
			"This module takes local histograms of a chosen size and "
			"accordingly normalizes the image."
		)
{
	ParameteredObject::_addInputSlot(
		images, "images",
		"Images to normalize",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		normalized, "normalized",
		"Normalized images",
		"CImgList<T>");

	ParameteredObject::_addParameter< int >(
		patchSize, "patchSize",
		"Width and height of the patches used for local histograms.",
		5, "int");

	ParameteredObject::_addParameter< float >(
		blur, "blur",
		"Blur befor normalize",
		0, "float");

	ParameteredObject::_addParameter(
		weightByDistance, "weightByDistance",
		"Weight pixels in normalization step by its distance to center pixel",
		"bool");

	ParameteredObject::_addParameter< float >(
		weight, "weight",
		"Weight for pixels with maximal distance. Only affects if weightByDistance is true. Possible value range (0,1)",
		1, "float");

	ParameteredObject::_addParameter(
		useSingleChannels, "useSingleChannels",
		"Normalize channels seperately",
		"bool");

	ParameteredObject::_addParameter(
		useHSV, "useHSV",
		"Convert to HSV, normalize and convert back to RGB",
		"bool");

}

template <typename T>
void NormalizeLocal<T>::execute() {

	const unsigned int _s = std::max( int(1), int(this->patchSize() / 2) );
	sout << "Use patches of width/height: " << _s*2+1 << " pixels" << std::endl;
	const float _max_dist = std::sqrt(2.)*_s;
	sout << "Maximal distance to center pixel: " << _max_dist << std::endl;
	const float _weight = std::max( float(0), std::min( float(1), float(this->weight()) ) );
	if( this->weightByDistance() )
		sout << "Maximal weight for farest pixel: " << 1-_weight << std::endl;

	// RGB normalization
	if( !this->useHSV() ){

		for( size_t i = 0; i < this->images().size(); i++ ){
	
			const cimg_library::CImg<T> img = this->images()[i].get_blur( this->blur() );
			cimg_library::CImg<float> normalized( img.width(), img.height(), 1, img.spectrum() );

			cimg_forXY( img, x, y){

				// crop patch for all channels
				cimg_library::CImg<float> patch = img.get_crop(x-_s,y-_s,x+_s,y+_s);

				// Get global coords for patch center
				int xs = _s, ys = _s;
				if(x-_s < 0 || y-_s < 0 )
				{
					xs = x - _s;
					ys = y - _s;
				}

				// normalize
				if( this->weightByDistance() ){
					cimg_forXYC( patch, xx, yy, cc ){
						float weight = std::sqrt( float( (xx-xs)*(xx-xs) + (yy-ys)*(yy-ys) ) ) / _max_dist * _weight ; // distance to center divided by max distance
						patch(xx,yy,0,cc) = patch(xs,ys,0,cc) - (1-weight)* ( patch(xs,ys,0,cc) - patch(xx,yy,0,cc) );
					}
				}

				// normalize keeping channel relations
				if( !this->useSingleChannels() ){
					//patch.normalize(0,255);
					//for(size_t c=0; c < patch.spectrum(); c++)
					//	normalized(x,y,0,c) = patch(xs,ys,0,c);
					patch.equalize(255,0,255);
					for(size_t c=0; c < patch.spectrum(); c++)
						normalized(x,y,0,c) = patch(xs,ys,0,c);
				}
				// normalize for each channel seperately
				else{
					for(size_t c=0; c < patch.spectrum(); c++){						
						normalized(x,y,0,c) = (patch.get_channel(c).normalize(0,255))(xs,ys,0,0);
					}
				}

			}

			this->normalized().push_back( normalized.normalize(0,255) );
	
		}

	}

	// use HSV values to normalize
	else{
		sout << "Not implemented HSV normalization yet " << std::endl;

		for( size_t i = 0; i < this->images().size(); i++ ){
	
			const cimg_library::CImg<T> img = this->images()[i].get_RGBtoHSV().blur( this->blur() );
			cimg_library::CImg<float> normalized = img;

			cimg_forXY( img, x, y){

				// crop patch for all channels
				cimg_library::CImg<float> patch = img.get_crop(x-_s,y-_s,x+_s,y+_s);

				// Get global coords for patch center
				int xs = _s, ys = _s;
				if(x-_s < 0 || y-_s < 0 )
				{
					xs = x - _s;
					ys = y - _s;
				}				

				// normalize for each channel seperately
				float pnew = (patch.get_channel(2).normalize(0,1))(xs,ys);
				float pold = patch(xs,ys,0,2);
				normalized(x,y,0,2) = pold * ( 0.5 + pnew );

				if( x == 10 && y == 10)
					cimg_forXY( patch, xx, yy ){
						sout << patch(xx,yy,0,2) << std::endl;
					}

			}
			
			this->normalized().push_back( normalized.HSVtoRGB().normalize(0,255)  );
			
			//this->normalized().push_back( normalized.get_HSVtoRGB().get_channel(0).normalize(0,255) );
			//this->normalized().push_back( normalized.get_HSVtoRGB().get_channel(1).normalize(0,255) );
			//this->normalized().push_back( normalized.get_HSVtoRGB().get_channel(2).normalize(0,255) );
	
		}

	}



}

#endif /* _NORMALIZELOCAL_HXX_ */
