/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file OptimizeTrackingMatrix.h
 *  Declaration of the parameter class OptimizeTrackingMatrix.
 *  \author Moritz Becker
 *  \date 04.08.2011
 */

#ifndef _OPTIMIZETRACKINGMATRIX_H_
#define _OPTIMIZETRACKINGMATRIX_H_

#if defined(MSVC) && defined(HANDLE_DLL)
#ifdef optimizetrackingmatrix_EXPORTS
/// Visual C++ specific code
#define optimizetrackingmatrix_DECLDIR __declspec(dllexport)
#else
#define optimizetrackingmatrix_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else /* No DLL handling or GCC */
/// Not needed with GCC
#define optimizetrackingmatrix_DECLDIR
#endif

#include <charon-core/ParameteredObject.hxx>
#include <charon-utils/CImg.h>

/// This plugin is based on Wang et al., Guide to Three Dimensional Structure and Motion Factorization, page 92.
/// 
/// This plugin optimizes the input tracking matrix W(=PX) by weighting the projective depths. Then, P and X can be obtained by rank-4-factorization of W.
/// 
/// Therefore an initial guess of projective matrices P1 and P2 from fundamental matrix F is achieved to triangulate the 2D correspondences from W to also retrieve a first 3D point estimation. Then with W and 3D Points, also the other projective matrices P_i are estimated. Now the projective depths are optimized by minimazing a back-projection cost function. With the new projective depths, the tracking matrix gets weighted and factorized by a rank-4-factorization to receive the projective motion matrix M and shape S. After that another cost function (W-MS) is minimized by alterning the projective depth.
/**	The algorithm performs steps as follows:
	1) normalize correspondences by transforming
	   them with T so the centroid lies at origin and the mean 
	   distance is sqrt(2)
    2) Triangulate 3D points by using the estimated
	   projective matrices P1 and P2 of frame 1 and 2.
	   Once all 3D Points are recovered, they can be used
	   togehter with the 2D correspondences, to retrieve
	   all other projective matrices Pi with i > 2.
    3) Estimate the projective depths lambda_ij as in
	   Wang et al., page 93
    4) Retrieve structure and motion matrices from 
	   tracking matrix and optimize them by rank-4-factorization
	5) unnormalize recovered cameras and points with T^-1
*/

template <typename T>
class optimizetrackingmatrix_DECLDIR OptimizeTrackingMatrix :
		public TemplatedParameteredObject<T> {
public:
	

	/// Number of iterations of normalization of the tracking matrix point and image wise, before the rank-4-factorization is performed.
	Parameter < int > parNorm;	
	/// Number of iterations of optimization of the tracking matrix.
	Parameter < int > parConvIt;	
	/// Convergence value which has to be reached by the difference of subsequent projective depths of the optimization algorithm.
	Parameter < double > parConv;		

	/// The (nxm) tracking matrix contains the detected 2D points where n is the index of the point and m the index of the frame. Thus there are n points in every m frames. 
	InputSlot < cimg_library::CImgList<T> > trackingMatrix;
	/// The 3D points are used as for an initial guess to optimize the tracking matrix. 
	InputSlot < cimg_library::CImgList<T> > points3D;
	/// Optimized tracking matrix. 
	OutputSlot < cimg_library::CImgList<T> > trackingMatrixOut;
	/// Optimized 3D points. 
	OutputSlot < std::vector<std::vector<T> > > points3DOut;
	/// Calculated camera matrices by rank-4-factorization. 
	OutputSlot < cimg_library::CImgList<T> > cameraMatrices;
	

	/// create a new OptimizeTrackingMatrix object
	/// \param name          Instance name
	OptimizeTrackingMatrix(const std::string& name);


	/// Update object.
	virtual void execute();
};

#endif // _OPTIMIZETRACKINGMATRIX_H_



