/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file OptimizeTrackingMatrix.hxx
 *  Implementation of the parameter class OptimizeTrackingMatrix.
 *  \author Moritz Becker
 *  \date 04.08.2011
 */

#ifndef _OPTIMIZETRACKINGMATRIX_HXX_
#define _OPTIMIZETRACKINGMATRIX_HXX_

#include <hekate/OptimizeTrackingMatrix.h>
#include <charon-utils/ImgTool.hxx>
#include <levmar/levmar.h>

template <typename T>
OptimizeTrackingMatrix<T>::OptimizeTrackingMatrix(const std::string& name) :
		TemplatedParameteredObject<T>(
				"OptimizeTrackingMatrix", name,
				"This plugin is based on Wang et al., Guide to Three Dimensional Structure and Motion Factorization, page 92."
				"This plugin optimizes the input tracking matrix W(=PX) by weighting the projective depths. Then, P and X can be obtained by rank-4-factorization of W."
				"Therefore an initial guess of projective matrices P1 and P2 from fundamental matrix F is achieved to triangulate the 2D correspondences from W to also retrieve a first 3D point estimation. Then with W and 3D Points, also the other projective matrices P_i are estimated. Now the projective depths are optimized by minimazing a back-projection cost function. With the new projective depths, the tracking matrix gets weighted and factorized by a rank-4-factorization to receive the projective motion matrix M and shape S. After that another cost function (W-MS) is minimized by alterning the projective depth."
		)
			,trackingMatrix(false, false) 
			,points3D(false, false) 
			
			
{
	

	
	ParameteredObject::_addParameter (parNorm, "NormalizationIterations", 
		"Number of iterations of normalization of the tracking matrix"
		" point and image wise, before the rank-4-factorization is performed.", 2);
	ParameteredObject::_addParameter (parConvIt, "ConvergenceIterations", 
		"Number of iterations of optimization of the tracking matrix.", 1000);
	ParameteredObject::_addParameter (parConv, "ConvergenceValue", 
		"Convergence value which has to be reached by the difference of subsequent "
		"projective depths of the optimization algorithm.", 0.0000000001);

	ParameteredObject::_addInputSlot(trackingMatrix, "trackingMatrix", 
	"The (nxm) tracking matrix contains the detected 2D points where n is the in"
	"dex of the point and m the index of the frame. Thus there are n points in every"
	" m frames.", 
	"CImgList<T>");
	ParameteredObject::_addInputSlot(points3D, "points3D", 
	"The 3D points are used as for an initial guess to optimize the tracking matrix.", 
	"CImgList<T>");

	ParameteredObject::_addOutputSlot(trackingMatrixOut, "trackingMatrixOut", 
	"Optimized tracking matrix.", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(points3DOut, "points3DOut", 
	"Optimized 3D points.", 
	"vector<vector<T> >");
	ParameteredObject::_addOutputSlot(cameraMatrices, "cameraMatrices", 
	"Calculated camera matrices by rank-4-factorization.", 
	"CImgList<T>"); 
	
}

template <typename T>
void OptimizeTrackingMatrix<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	double convValue = this->parConv();
	int convIt = this->parConvIt();
	int normIt = this->parNorm();

	const cimg_library::CImgList<T> & points = this->points3D();
	const cimg_library::CImgList<T> & trackMatrix = this->trackingMatrix();

	// Check whether tracking matrix and 3D points are consistent.
	if(points.size() != (unsigned)trackMatrix[0].width()){
		sout << "ERROR: Number of 3D points is not consistent "
			"with number of different 2D points." << std::endl;
		sout << "Number of 3D points: " << points.size() << std::endl;
		sout << "Number of 2D points: " << trackMatrix[0].width() << std::endl;
		throw(std::runtime_error("Number of 3D points is not consistent "
			"with number of different 2D points."));
	}

	// Check whether the 3D input points are homogeneous
	cimg_library::CImgList<T> pointsHom;
	if(points[0].height() == 4)
		pointsHom = points;
	else{
		for(size_t i=0; i<points.size(); i++){
			cimg_library::CImg<T> X(1,4);
			X[0] = points[i][0];
			X[1] = points[i][1];
			X[2] = points[i][2];
			X[3] = T(1);
			pointsHom.push_back(X);
		}
	}

	/* 
	Read in tracking matrix ------------------------------------------------
	*/

	// Write uunormalized W to log
	sout << "Read in tracking matrix:" << std::endl;

	// Read in dimensions
	size_t dimN = trackMatrix[0].width();
	size_t dimM = trackMatrix[0].height()/3;
	size_t dimMN = dimM * dimN;

	sout << "Number of different points = " << dimN << std::endl;
	sout << "Number of different frames = " << dimM << std::endl;

	// Create first row of the tracking matrix
	cimg_library::CImg<T> W = trackMatrix[0];

	ImgTool::printInfo(sout,W,"W unormalized");

	// Throw error message, if less than 4 frames are available.
	if(W.height() < 6)
		throw(std::runtime_error("Insufficient number of frames. At least 2 frames are needed."));	
	else if(W.height()%3 != 0)
		throw(std::runtime_error("Number of tracking matrix lines needs to be a multiple of 3."));	


	// Check for outliers
	cimg_library::CImg<T> Wsorted;
	cimg_library::CImgList<T> pointsHomSorted;
	for(size_t i=0; i<pointsHom.size(); i++){
		if(pointsHom[i][0] != T(0) &&
			pointsHom[i][1] != T(0) &&
			pointsHom[i][2] != T(0)){

				bool isValid = true;
				for(int j=0; j<W.height(); j=j+3){
					if(W(i,j) == 0 && W(i,j+1) == 0
						&& W(i,j+2) == 0){
						isValid = false;												
					}
				}

				if(isValid){
					Wsorted.append(W.get_crop(i,0,i,W.height()-1));
					pointsHomSorted.push_back(pointsHom[i]);
				}

		}
		//else
			//sout << "Sorted out point " << i << std::endl;
			//ImgTool::printInfo(sout,pointsHom[i],"pointsHom[i]");
	}
	W = Wsorted;
	pointsHom = pointsHomSorted;
	ImgTool::printInfo(sout,W,"W outsorted");
	if(W.width() == 0)
		throw(std::runtime_error("Tracking matrix has no points!"));

	dimN = W.width();

	/*
	Normalization of the Correspondences ------------------------------------
	*/

	// create transformation matrix which normalizes the points
	cimg_library::CImgList<T> transformations;
	normalizeTrackingMatrix(W,transformations);

	// Write normalized W to log
	sout << "Normalized tracking matrix. The point cloud of each image is "
		<< "centered at the origin and the mean distance between the points and the "
		<< "origin is sqrt(2)." << std::endl;
	ImgTool::printInfo(sout,W,"W normalized");



	/* ------------------------------------------------------------
	Create linear equation AP=0 to determine
	projective matrices P by SVD (Wang et al., page 92)

	ATTENTION: At this point, these projective matrices
	could also be obtained via RANSAC to get better results
	if 3D points X and 2D points x have outliers.
	*/

	cimg_library::CImgList<T> PList;
	
	recoverProjectiveMatrices(W, PList, pointsHom);

	/* -------------------------------------------------------------------
	Estimate projective depths by costfunction:
	J1 = 0.5 min(Frobeniusnorm(lamda_ij*x_ij-P_i*X_j)^2,labda_ij)
	*/
	
	// Generate matrix with projective depths as entries
	cimg_library::CImg<T> lambdaMatrix(dimN,dimM);

	// Iterate the following code for every frame
	for(size_t i=0; i<dimM; i++){

		// Iterate the following code for every point per frame
		for(size_t j=0; j<dimN; j++){

			cimg_library::CImg<T> u(1,3), v(1,3);
			u = W.get_crop(j,i*3,j,i*3+2);
			v = PList[i]*pointsHom[j];

			// Calculate projective depth of point i,j and weight the
			// tracking matrix W.
			T lambda = T(u.dot(v)/u.dot(u));
			W(j,i*3) *= lambda;
			W(j,i*3+1) *= lambda;
			W(j,i*3+2) *= lambda;
			lambdaMatrix(j,i) = lambda;
			
		}
	}
	
	// Give out weighted tracking matrix
	sout << "Weighted the tracking matrix by estimating "
		<< "projective depths lambda_ij." << std::endl;
	ImgTool::printInfo(sout,W,"W weighted");
	ImgTool::printInfo(sout,lambdaMatrix,"lambdaMatrix");



	/* -----------------------------------------------------------------
	Perform a rank-4-factorization of W by SVD
	Then extract the motion and structure matrices M and S
	*/

	// Calculate SVD of weighted tracking matrix
	cimg_library::CImg<T> U, D, V;
	// Init motion and structure matrices
	cimg_library::CImg<T> M, S, D2, MS;

	//Iterate optimization of tracking matrix until convergence
	double deltaLambda = 1.0;
	int it = 0;

	while(std::abs(deltaLambda) > convValue && it < convIt){
		
		it++;
		sout << "Iteration " << it << " of projective depths optimization." << std::endl;

		// Calculate the SVD of the tracking matrix and
		// perform a rank-4-factorization which is done by cutting the
		// singular value matrix D after the four highest singular values
		// and setting all other singular values to zero.

		W.SVD(U,D,V);
		ImgTool::printInfo(sout,D,"SingularValues(W)");
		D2 = D.get_crop(0,0,0,3).diagonal().sqrt();
		M = U.get_crop(0,0,3,U.height()) * D2;
		S = D2 * V.get_crop(0,0,3,V.height()).transpose();
		MS = M*S;

		// Reset deltaLambda for every iteration
		deltaLambda = 0.0;

		// Iterate the following code for every frame
		for(size_t i=0; i<dimM; i++){

			// Iterate the following code for every point per frame
			for(size_t j=0; j<dimN; j++){

				cimg_library::CImg<T> u(1,3), v(1,3);
				u = W.get_crop(j,i*3,j,i*3+2);
				v = MS.get_crop(j,i*3,j,i*3+2);

				// Calculate projective depth of point i,j and weight the
				// tracking matrix W.
				T lambda = T(u.dot(v)/u.dot(u));
				W(j,i*3) *= lambda;
				W(j,i*3+1) *= lambda;
				W(j,i*3+2) *= lambda;

				//sout << "lambda: " << lambda << std::endl;
				//sout << "lambdaMatrix(j,i): " << lambdaMatrix(j,i) << std::endl;
				
				deltaLambda += (double)((lambda-lambdaMatrix(j,i))*(lambda-lambdaMatrix(j,i))/dimMN);
				//sout << "deltaLambda: " << deltaLambda << std::endl;

				lambdaMatrix(j,i) = lambda;
			
			}
		}

		sout << "Calculated difference between lambdas of subsequent iterations:" << std::endl;
		sout << "Delta(lambda) = " << deltaLambda << std::endl;

	}

	// Give out number of iterations and convergence value if algorithm converged
	// within number of maximal iterations
	if(std::abs(deltaLambda) <= convValue){
		sout << "Iteration converged after " << it << " iterations: deltaLambda = " << deltaLambda << std::endl;
	}

	// Give out message, if iteration stopped due to maximal number of iteration
	else if(it >= convIt){
		sout << "Iteration stopped: maximum number of iterations reached (" << it << ")" << std::endl;
		sout << "Convergence value: " << deltaLambda << std::endl;
	}

	sout << "Optimized tracking matrix: " << std::endl;
	ImgTool::printInfo(sout,W,"W optimizedProjectiveDepths");

	/* ---------------------------------------------------------------
	Perform normalization on each image and also each point separately.
	Normalizing each row is consistent with image wise scaling of the
	projective matrix while normalizing each column is equal to scale 
	every point X_j.
	*/

	sout << "Performing normalization on each image and also each point separately" 
		<< " which is consistent with normalizing every row and every column"
		<< " of the tracking matrix." << std::endl;

	// Create identitiy matrix
	cimg_library::CImg<T> I(3,3), I2(4,4);
	I.identity_matrix();
	I2.identity_matrix();

	// Create transformation matrix which normalizes the rows
	// of the tracking matrix W
	cimg_library::CImgList<T> transformationsRows(dimM,I);

	// Create transformation matrix which normalizes the columns
	// of the tracking matrix W
	cimg_library::CImgList<T> transformationsColumns(dimN,I2);

	// Iterate normalization
	for(int i = 0; i < normIt; i++){

		// Perform normalization of Columns
		normalizeTrackingMatrixColumns(W,transformationsColumns);
		ImgTool::printInfo(sout,W,"W normalizedColumns");

		// Perform normalization of Rows
		normalizeTrackingMatrixRows(W,transformationsRows);
		ImgTool::printInfo(sout,W,"W normalizedRows");

	}



	/* ----------------------------------------------------------------
	Extract motion and structure matrices from optimized,
	weighted and normalized tracking matrix
	*/

	W.SVD(U,D,V);
	D.crop(0,0,0,3).diagonal().sqrt();
	M = U.get_crop(0,0,3,U.height()-1) * D;
	S = D * V.get_crop(0,0,3,V.height()-1).transpose();
	
	ImgTool::printInfo(sout,W,"W");
	ImgTool::printInfo(sout,M,"M");
	ImgTool::printInfo(sout,S,"S");
	ImgTool::printInfo(sout,M*S,"M*S");

	// Create optimized output tracking matrix
	cimg_library::CImg<T> S2, M2;

	// Extract projective matrices
	for(size_t i=0; i<dimM; i++){

		// Back transform the projective matrix which has been
		// normalized image wise twice
		cimg_library::CImg<T> pm;
		pm = M.get_crop(0,i*3,3,i*3+2);
		pm = transformations[i].get_invert() * pm;

		// Push back transformed projective matrix to output slot
		this->cameraMatrices().push_back(pm);

		// Append transformed matrix to M2
		M2.append(pm,'y');

	}
	
	// Extract 3D points
	std::vector<std::vector<T> > vectorPoints;
	for(size_t i=0; i<dimN; i++){

		// Back transform 3D points
		cimg_library::CImg<T> pt;
		pt = S.get_crop(i,0,i,3);
		pt /= pt[3];

		// Push back transformed 3D points
		std::vector<T> point;
		point.push_back(pt[0]);
		point.push_back(pt[1]);
		point.push_back(pt[2]);
		point.push_back(pt[3]);
		vectorPoints.push_back(point);

		// Append transformed matrix to S2
		S2.append(pt,'x');

	}
	this->points3DOut() = vectorPoints;

	ImgTool::printInfo(sout,M2,"M2");
	ImgTool::printInfo(sout,S2,"S2");
	ImgTool::printInfo(sout,M2*S2,"M2*S2");


	// Normalize M*S so the homogeneous coordinate is equal 1 ------------------
	// to check whether the original correspondences are still correct
	cimg_library::CImg<T> Wnorm;
	Wnorm = M2*S2;
	for(size_t i=0; i<dimM; i++){
		for(size_t j=0; j<dimN; j++){
			Wnorm(j,i*3) /= Wnorm(j,i*3+2);
			Wnorm(j,i*3+1) /= Wnorm(j,i*3+2);
			Wnorm(j,i*3+2) /= Wnorm(j,i*3+2);
		}
	}
	ImgTool::printInfo(sout,Wnorm,"Wnorm");


	// Give out factorized tracking matrix
	cimg_library::CImg<T> Wout;
	Wout = M2*S2;
	normalizeTrackingMatrixColumns(Wout,transformationsColumns);
	normalizeTrackingMatrixRows(Wout,transformationsRows);
	for(size_t i=0; i<dimM; i++){
		this->trackingMatrixOut().push_back(Wout.get_crop(0,i*3,dimN,i*3+2));
	}
		
}


/// Normalizes 2D correspondences so the points are centered at origin and have a mean length of sqrt(2).
template <typename T>
void normalizeTrackingMatrix(cimg_library::CImg<T> &W, cimg_library::CImgList<T> &transformations){

	// Get number of correspondences (there are 3 values per point)
	size_t dimN = W.width();
	size_t dimM = W.height();
	
	//Iterate the following code for every image
	for(size_t i=0; i<dimM; i=i+3){

		//Calculation of the centroid of the Correspondences
		T centroid_x=0, centroid_y=0;
		for(size_t j=0; j<dimN; j++){

			// Check points for outliers
			int dimNInlier = 0;
			if(W(j,i) != 0 || W(j,i+1) != 0 || W(j,i+2) != 0){
				centroid_x += W(j,i);
				centroid_y += W(j,i+1);
				dimNInlier++;
			}
			centroid_x /= dimNInlier;
			centroid_y /= dimNInlier;

		}	

		// Translating correspondences
		for(size_t j=0; j<dimN; j++){	
			if(W(j,i) != 0 || W(j,i+1) != 0 || W(j,i+2) != 0){
				W(j,i) -= centroid_x;
				W(j,i+1) -= centroid_y;
			}
		}	

		//Normalize RMS distance to sqrt(2)
		T rms1=0;
		for(size_t j=0; j<dimN; j++){
			int dimNInlier = 0;
			if(W(j,i) != 0 || W(j,i+1) != 0 || W(j,i+2) != 0){
				rms1 += T(W.get_crop(j,i,j,i+1).magnitude());
				dimNInlier++;
			}
			rms1 /= dimNInlier;
		}	
		T s1 = rms1/T(sqrt(2.00)); // scale factor
		
		// Apply scale factor
		if(s1 != 0){
			for(size_t j=0; j<dimN; j++){
				if(W(j,i) != 0 || W(j,i+1) != 0 || W(j,i+2) != 0){
					W(j,i) *= 1/s1;
					W(j,i+1) *= 1/s1;
				}
			}
		}
		else{
			throw std::runtime_error("Division by zero.") ;
		}

		//Generation of the translation matrices ------------------------
		cimg_library::CImg<T> tr1(3,3), sc1(3,3);

		//Translation matrix
		tr1(0,0) = (T)1; tr1(1,0) = (T)0; tr1(2,0) = -centroid_x;
		tr1(0,1) = (T)0; tr1(1,1) = (T)1; tr1(2,1) = -centroid_y;
		tr1(0,2) = (T)0; tr1(1,2) = (T)0; tr1(2,2) = (T)1;

		//Scaling matrix of Img 1
		sc1(0,0) = 1/s1;	sc1(1,0) = 0;		sc1(2,0) = 0;
		sc1(0,1) = 0;		sc1(1,1) = 1/s1;	sc1(2,1) = 0;
		sc1(0,2) = 0;		sc1(1,2) = 0;		sc1(2,2) = 1;

		//Generation of Transformation Matrices (combination of scaling and translation)
		cimg_library::CImg<T> trafo;
		trafo = sc1*tr1;
		transformations.push_back(trafo);
	}
}


/// Recovers all projective matrices P_i of all frames using 2D points x_ij and 3D points X_j.
template <typename T>
void recoverProjectiveMatrices(const cimg_library::CImg<T> &W, cimg_library::CImgList<T> &PList, const cimg_library::CImgList<T> &points3D){

	// Read in dimensions
	size_t dimN = W.width();

	// Iterate the following code for every projective matrix
	for(int i=0; i<W.height(); i=i+3){

		// Create A
		cimg_library::CImg<T> A;

		for(size_t j=0; j<dimN; j++){

			// Create A and read in 3D point X and 2d point x.
			cimg_library::CImg<T> X(1,4), x(1,3);

			// Read in j-th 3D point
			X = points3D[j];

			// Read in j-th 2D point from frame i
			x[0] = W(j,i);
			x[1] = W(j,i+1);
			x[2] = W(j,i+2);

			//Check if 2D point is an outlier or not available (0-vector)
			if(x[0]==0 && x[1]==0 && x[2]==0){
				sout << "2D point " << j << " of frame " << i
					<< " does not exist or is an outlier. Ignoring point." << std::endl;
			}

			//Check if 3D point is an outlier or not available (0-vector)
			else if(X[0]==0 && X[1]==0 && X[2]==0 && X[3]==0){
				sout << "3D point " << j << " does not exist or is an outlier."
					<< "Ignoring point." << std::endl;
			}

			else{		
				// Create rowA where which contains two
				// constraints for every point and is appended
				// to matrix A afterwards.
				cimg_library::CImg<T> rowA(12,2);

				// Reset row
				rowA.fill(0);

				// Fill A as on Wang et al., page 92
				rowA(0,0) = X[0];
				rowA(1,0) = X[1];
				rowA(2,0) = X[2];
				rowA(3,0) = X[3];
				rowA(8,0) = -x[0]*X[0];
				rowA(9,0) = -x[0]*X[1];
				rowA(10,0) = -x[0]*X[2];
				rowA(11,0) = -x[0]*X[3];
				rowA(4,1) = X[0];
				rowA(5,1) = X[1];
				rowA(6,1) = X[2];
				rowA(7,1) = X[3];
				rowA(8,1) = -x[1]*X[0];
				rowA(9,1) = -x[1]*X[1];
				rowA(10,1) = -x[1]*X[2];
				rowA(11,1) = -x[1]*X[3];

				// Append rowA to A
				const char direction = 'y';
				A.append(rowA,direction);

			}
		}

		// Create SVD of A to determine the projective matrix i;
		cimg_library::CImg<T> U, D, V;
		A.SVD(U,D,V);

		// Generate projective matrix i from V
		cimg_library::CImg<T> P(4,3);
		P[0] = V(11,0);
		P[1] = V(11,1);
		P[2] = V(11,2);
		P[3] = V(11,3);
		P[4] = V(11,4);
		P[5] = V(11,5);
		P[6] = V(11,6);
		P[7] = V(11,7);
		P[8] = V(11,8);
		P[9] = V(11,9);
		P[10] = V(11,10);
		P[11] = V(11,11);

		// Save projective matrix which is the last row of V in CImgList
		PList.push_back(P);

	}
}

/// Normalizes rows of the tracking matrix which is consistent with scaling every image.
template <typename T>
void normalizeTrackingMatrixRows(cimg_library::CImg<T> &W, cimg_library::CImgList<T> &transformations){

	// Get number of correspondences (there are 3 values per point)
	size_t dimN = W.width();
	size_t dimM = W.height();
	
	//Iterate the following code for every image
	for(size_t i=0; i<dimM; i=i+3){

		//Normalize RMS distance to sqrt(2)
		T rms1=0;
		for(size_t j=0; j<dimN; j++){
			rms1 += T(W.get_crop(j,i,j,i+2).magnitude() / dimN);
		}	
		T s1 = rms1/T(sqrt(2.00)); // scale factor

		// Apply scale factor
		if(s1 != 0){
			for(size_t j=0; j<dimN; j++){
				W(j,i) *= 1/s1;
				W(j,i+1) *= 1/s1;
				W(j,i+2) *= 1/s1;
			}
		}
		else{
			throw std::runtime_error("Division by zero.") ;
		}

		//Generation of the translation matrices ------------------------
		cimg_library::CImg<T> sc1(1,3);
		sc1.fill(1/s1);
		sc1.diagonal();

		//Generation of Transformation Matrices (combination of scaling and translation)
		transformations[i/3]*=sc1;

	}
}

/// Normalizes columns of the tracking matrix which is consistent with scaling every 3D point.
template <typename T>
void normalizeTrackingMatrixColumns(cimg_library::CImg<T> &W, cimg_library::CImgList<T> &transformations){

	// Get number of correspondences (there are 3 values per point)
	size_t dimN = W.width();
	size_t dimM = W.height();
	
	// Iterate the following code for every 3d point
	for(size_t j=0; j<dimN; j++){	

		// Normalize RMS distance to sqrt(2)
		T rms1=0;
		for(size_t i=0; i<dimM; i=i+3){
			rms1 += T(W.get_crop(j,i,j,i+2).magnitude() / dimM);
		}	
		T s1 = rms1/T(sqrt(2.00)); // scale factor

		// Apply scale factor
		if(s1 != 0){
			for(size_t i=0; i<dimM; i=i+3){
				W(j,i) *= 1/s1;
				W(j,i+1) *= 1/s1;
				W(j,i+2) *= 1/s1;
			}
		}
		else{
			throw std::runtime_error("Division by zero.") ;
		}

		// Generation of the translation matrices ------------------------
		cimg_library::CImg<T> sc1(1,4);
		sc1.fill(1/s1);
		sc1.diagonal();

		// Generation of Transformation Matrices (combination of scaling and translation)
		// The order of multiplications does not matter since it is a diagonal matrix
		// which is multiplied.
		transformations[j]*=sc1;

	}
}


#endif /* _OPTIMIZETRACKINGMATRIX_HXX_ */



