/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file ParallelBundleAdjustment.hxx
 *  Implementation of the parameter class ParallelBundleAdjustment.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 29.02.2012
 */

#ifndef _PARALLELBUNDLEADJUSTMENT_HXX_
#define _PARALLELBUNDLEADJUSTMENT_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/ParallelBundleAdjustment.h>
#include <charon-utils/ImgTool.hxx>
#include <hekate/Create3DPointCloud_plugins.h>

template <typename T>
ParallelBundleAdjustment<T>::ParallelBundleAdjustment(const std::string& name) :
		TemplatedParameteredObject<T>(
			"ParallelBundleAdjustment", name,
			"<h2>Performs a bundle adjustment for a given set of Cameras, "
			"3DPoints, 2DPoints</h2><br>"
			"Performs a bundle adjustment for a given set Cameras, 3DPoints, "
			"2DPoints using several threads. Indices of output points are NOT "
			"consistent with input points"
		), calibrationMatrix(1,0)
{
	
	ParameteredObject::_addParameter< int > (minPtViews, "minPtViews",
		"Minimum number of views, a point is visible in.", 3, "int");
	ParameteredObject::_addParameter< bool > (fixedFocalLength, "fixedFocalLength",
		"Focal length is fixed for all cameras (calibrated case).", true, "bool");
	ParameteredObject::_addParameter< bool > (useCUDA, "useCUDA",
		"useCUDA", false, "bool");
	ParameteredObject::_addParameter< bool > (calculateDataOverlap, "calculateDataOverlap",
		"Calculate consistent data by taking the overlap between image and point indices of 2d, 3d and cams", false, "bool");
	ParameteredObject::_addParameter< bool > (constantCameras, "constantCameras",
		"Flag cameras as constant. Only 3D points will be modified by PBA", false, "bool");

	ParameteredObject::_addInputSlot(
		cameraMatrices, "cameraMatrices",
		"Projection matrices",
		"IndexedMatrixList");
	ParameteredObject::_addInputSlot(
		calibrationMatrix, "calibrationMatrix",
		"calibrationMatrix",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		pointCloud, "pointCloud",
		"3D point clouds",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"2D correspondences",
		"CorrespondenceList");

	ParameteredObject::_addOutputSlot(
		cameraMatricesOut, "cameraMatricesOut",
		"Adjusted camera matrices",
		"IndexedMatrixList");
	ParameteredObject::_addOutputSlot(
		pointCloudOut, "pointCloudOut",
		"Adjusted 3D point clouds",
		"CorrespondenceList");
	ParameteredObject::_addOutputSlot(
		calibrationMatrixOut, "calibrationMatrixOut",
		"calibrationMatrixOut",
		"CImgList<T>");

}

template <typename T>
void ParallelBundleAdjustment<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	    
	if (this->calibrationMatrix.connected()){
		ParallelBundleAdjustmentPlugin(	this->cameraMatrices(), 	this->calibrationMatrix(), this->pointCloud(),
										this->correspondences(),	this->cameraMatricesOut(), 
										this->pointCloudOut(), 		this->calibrationMatrixOut(), this->minPtViews(), 
										this->fixedFocalLength(), 	this->useCUDA(), this->calculateDataOverlap(), this->constantCameras() );
	}else{
		ParallelBundleAdjustmentPlugin(	this->cameraMatrices(), 	CImgList<T>(), this->pointCloud(),
										this->correspondences(), 	this->cameraMatricesOut(), 
										this->pointCloudOut(), 		this->calibrationMatrixOut(), this->minPtViews(), 
										this->fixedFocalLength(), 	this->useCUDA(), this->calculateDataOverlap(), this->constantCameras() );		
	}
}

#endif /* _PARALLELBUNDLEADJUSTMENT_HXX_ */
