/*  Copyright (C) 2013 Philipp Geier

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file PointCloudCorrection.h
 *  Declaration of the parameter class PointCloudCorrection.
 *  \author <a href="mailto:P.Geier@stud.uni-heidelberg.de">
 *      Philipp Geier</a>
 *  \date 11.03.2013
 */

#ifndef _POINTCLOUDCORRECTION_H_
#define _POINTCLOUDCORRECTION_H_

#ifdef _MSC_VER
#ifdef pointcloudcorrection_EXPORTS
/// Visual C++ specific code
#define pointcloudcorrection_DECLDIR __declspec(dllexport)
#else
#define pointcloudcorrection_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else
/// Not needed without MSVC
#define pointcloudcorrection_DECLDIR
#endif

#include <charon-core/ParameteredObject.h>
#include <charon-utils/CImg.h>
#include <hekate/IndexedMatrix.h>
#include <hekate/Correspondence.hxx>
#include <charon-utils/CImg.h>
#include <map>
#include <vector>
#include <hekate/DepthMapFrom3DPointCloud.h>
#include <string>


/// Takes a set of cameras and decides which can be expulsed to correct the 3D
/// Point Cloud
/** Takes a set of cameras and decides which can be expulsed to correct the 3D
 *  Point Cloud
 */
template <typename T>
class pointcloudcorrection_DECLDIR PointCloudCorrection :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	PointCloudCorrection(const std::string& name = "");

	/// cameras......
	InputSlot< IndexedMatrixList > cameras;
	/// 3d point cloud....
	InputSlot< hekate::CorrespondenceList<T> > points3d;
	/// 2d points.....
	InputSlot< hekate::CorrespondenceList<T> > points2d;

	/// corrected 3d points...
	OutputSlot< hekate::CorrespondenceList<T> > correctedPointCloud;

	Parameter< bool> depthFilterEnabled;
	Parameter< bool> angleFilterEnabled;

	typename struct AngleInf;
	typename struct CorrInf;
	typename struct EvalInf;

	Parameter< bool > relativeBinPicking;

	Parameter< double> histogramBins;

	Parameter< double> minBinsToPick;

	Parameter<unsigned int> minCorrespondencesToPick;

	Parameter<double> depthSubsetLowerThreshold;
	Parameter<double> depthSubsetUpperThreshold;
	Parameter<unsigned int> angleSubsetcount;

	Parameter< double> minAngle;
	Parameter< double> maxAngle;


	Parameter< double> backProjectionErrorThreshold;

	Parameter< unsigned int> errorDiffHistogramBinCount;

	Parameter< bool > suppressOutputForEachPoint;


protected:
	/// Update object.
	virtual void execute();

	/// Calculate depth between x and principle plane of P.
	double getDepth(const CImg<T> &P, const CImg<T> &X);

	cimg_library::CImg<T> getPrincipalAxis(const IndexedMatrix& cam);
	//returns the angle between two vectors
	double getAngle(const CImg<T>  &v1,const CImg<T>  &v2);

	double radToDeg(double rad);

	double degToRad(double deg);


	std::vector< struct PointCloudCorrection<T>::CorrInf *>   depthRoughFilter(std::vector< const hekate::Correspondence<T>*> &p2dList,const  hekate::Correspondence<T> &p3d, const IndexedMatrixList &cameraset,  bool relPicking, double histoBins, double minBinsToPick, unsigned int minCorrespondencesToPick);
	std::vector< struct PointCloudCorrection<T>::AngleInf *> PointCloudCorrection<T>::angleSorter(std::vector< const hekate::Correspondence<T>*> &p2dList,const IndexedMatrixList &cameraset);
	std::vector<  std::vector< const hekate::Correspondence<T>*> >PointCloudCorrection<T>::angleSubsets(std::vector< struct PointCloudCorrection<T>::AngleInf *>& ordAngleList, double minAngle, double maxAngle,  unsigned int minCamCount, unsigned int subsets)  ;



	hekate::Correspondence<T>* triangulation(const IndexedMatrixList& cameraList, std::vector< const hekate::Correspondence<T>*>& points);
	double bpError(std::vector< const hekate::Correspondence<T>*> &p2dList, const hekate::Correspondence<T> p3d, const IndexedMatrixList &cameraset, double percentage);

    EvalInf* bpEvaluation(std::string name, std::vector< const hekate::Correspondence<T>*> &p2dList,  const IndexedMatrixList &cameraset, double percentage, unsigned int minCount);

	
};

#endif // _POINTCLOUDCORRECTION_H_
