/*  Copyright (C) 2013 Philipp Geier

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file PointCloudCorrection.hxx
 *  Implementation of the parameter class PointCloudCorrection.
 *  \author <a href="mailto:P.Geier@stud.uni-heidelberg.de">
 *      Philipp Geier</a>
 *  \date 11.03.2013
 */

#ifndef _POINTCLOUDCORRECTION_HXX_
#define _POINTCLOUDCORRECTION_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "PointCloudCorrection.h"
#include <string>



template <typename T>
PointCloudCorrection<T>::PointCloudCorrection(const std::string& name) :
		TemplatedParameteredObject<T>(
			"PointCloudCorrection", name,
			"<h2>Takes a set of cameras and decides which can be expulsed to "
			"correct the 3D Point Cloud</h2><br>"
			"Takes a set of cameras and decides which can be expulsed to "
			"correct the 3D Point Cloud"
		)
{
	ParameteredObject::_addInputSlot(
		cameras, "cameras",
		"cameras......",
		"IndexedMatrixList");
	ParameteredObject::_addInputSlot(
		points3d, "points3d",
		"3d point cloud....",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		points2d, "points2d",
		"2d points.....",
		"CorrespondenceList");

	ParameteredObject::_addOutputSlot(
		correctedPointCloud, "correctedPointCloud",
		"corrected 3d points...",
		"CorrespondenceList");


		ParameteredObject::_addParameter<unsigned int >(
		minCorrespondencesToPick, "minCorrespondencesToPick",
		"Minium count of  cameras/correspondences",
		3, "unsigned int");


	ParameteredObject::_addParameter< bool >(
		depthFilterEnabled, "depthFilterEnabled",
		"Enable if distant cameras should be removed"
		"",
		1, "bool");



		ParameteredObject::_addParameter< bool >(
		relativeBinPicking, "relativeBinPicking",
		"The Depth of all cameras to each approximated 3dPoint will be calculated, then a histogram will be created to pick the the cameras with the smallest depth up to the depth area where most cameras are. Therefore the histogram will be split in bins. You can split it in a absolute (uncheck this) Count of Segments (e.g. always 4 Segments) or can take the count of Cameras and divide it by a value (relative, check this)."
		"",
		1, "bool");


	ParameteredObject::_addParameter< double >(
		histogramBins, "histogramBins",
		"Absolute: Divides the Area from DepthMin to DepthMax and creates a Histogramm with that amount of Segments"
		"Relative: (0.0 - 1.0) Divides the Area from DepthMin to DepthMax and creates a Histogramm with Camera count multiplied by that amount  of Segments",
		0.5, "double");

	ParameteredObject::_addParameter<double >(
		minBinsToPick, "minBinsToPick",
		"Absolute: Count how many are going to be picked (rounded to integer)"
		"Relative: (Percentage 0.0 to 1.0) to determine how many bins are going to be picked",
		0.5, "double");



	ParameteredObject::_addParameter<double>(
		depthSubsetUpperThreshold, "depthSubsetUpperThreshold",
		"After the rough filtering with a histogram, subsets are created and evaluted and the subset with the lowest backprojection error is picked. Therefore this percentage to 100 percent (with increasing depth) is picked and tested. See depthSubsetLowerThreshold as well",
		0.5, "double");

	ParameteredObject::_addParameter<double>(
		depthSubsetLowerThreshold, "depthSubsetLowerThreshold",
		"After the rough filtering with a histogram, subsets are created and evaluted and the subset with the lowest backprojection error is picked. Therefore  0 percent up to this percentage (with decreasing depth) is picked and tested. See depthSubsetUpperThreshold as well",
		0.5, "double");



	
	ParameteredObject::_addParameter< bool >(
		angleFilterEnabled, "angleFilterEnabled",
		"Enable if distant cameras should be removed"
		"",
		1, "bool");


	ParameteredObject::_addParameter<double >(
		minAngle, "minAngle",
		"Minimal Angle between 2 Cams in degrees. Default 11",
		5.0, "double");

	ParameteredObject::_addParameter<double >(
		maxAngle, "maxAngle",
		"Maximal Angle between 2 Cams in degrees. Default 40",
		50.0, "double");




	ParameteredObject::_addParameter<unsigned int >(
		angleSubsetcount, "angleSubsetcount",
		"Count of subsets that are going to be tested - 0 will pick the one with most",
		0, "unsigned int");

	ParameteredObject::_addParameter<double >(
		backProjectionErrorThreshold, "backProjectionErrorThreshold",
		"Comparing error with backprojection, to pick the best subset..... ",
		0.9, "double");


	ParameteredObject::_addParameter<unsigned int >(
		errorDiffHistogramBinCount, "errorDiffHistogramBinCount",
		"Number of bins for the error histogram..",
		10, "unsigned int");


	ParameteredObject::_addParameter<bool >(
		suppressOutputForEachPoint, "suppressOutputForEachPoint",
		"If enabled, there won't be an output (with sout) for each point",
		1, "bool");



}

//Returns the principal axis of a camera
template <typename T>
cimg_library::CImg<T> PointCloudCorrection<T>::getPrincipalAxis(const IndexedMatrix& cam) {
			cimg_library::CImg<double> principalAxis(1,3);
			principalAxis = cam.get_crop(0,0,2,2).det() * cam.get_crop(0,2,2,2).get_transpose();
			principalAxis /= principalAxis.magnitude();

			return principalAxis;
}

//gets the angle between two cameras
template <typename T>
double PointCloudCorrection<T>::getAngle(const CImg<T> &v1,const CImg<T> &v2) {
	return std::acos(   (v1.dot(v2))/(v1.magnitude() * v2.magnitude() )   );
}

const double PI = 4.0*atan(1.0);
// Calculate depth between x and principle plane of P.
template <typename T>
double PointCloudCorrection<T>::getDepth(const CImg<T> &P, const CImg<T> &X) {

	int signP;
	(P.get_crop(0,0,2,2).det() < 0 ? signP = -1 : signP = 1);

	double P_m3_abs = P(0,2)*P(0,2)+P(1,2)*P(1,2)+P(2,2)*P(2,2);
	P_m3_abs = pow(P_m3_abs,.5);

	double depth;
	depth = signP * (P*X)(0,2);
	if(P_m3_abs!=0) depth /= P_m3_abs;

	return depth;
}


template <typename T>
double PointCloudCorrection<T>::radToDeg(double rad) {
	return  180/PI*rad;
}

template <typename T>
double PointCloudCorrection<T>::degToRad(double deg) {
	return PI/180*deg;
}


	//Struct to associate more Values to each 2dPoint

template <typename T>
	 struct PointCloudCorrection<T>::CorrInf {
			const hekate::Correspondence<T> * cor;
			double depth;
			unsigned int histoField;

			static bool depthComp (const CorrInf* i,const CorrInf* j) { return (i->depth < j->depth);}
		} ;

//Struct to associate an angle to 2 Cams
template <typename T>
	 struct PointCloudCorrection<T>::AngleInf {
			const IndexedMatrix* cam1;
			const IndexedMatrix* cam2;
			const hekate::Correspondence<T> * cor1;
			const hekate::Correspondence<T> * cor2;
			double angle;	

			static bool angleComp (const AngleInf* i,const AngleInf* j) { return (i->angle < j->angle);}
			static bool revAngleComp (const AngleInf* i,const AngleInf* j) { return (j->angle < i->angle);}
		} ;

// Struct to evaluate and compare subsets
template <typename T>
	struct PointCloudCorrection<T>::EvalInf {
		
		std::string name;
		std::vector< const hekate::Correspondence<T>* > points2d;
		//const hekate::Correspondence<T>* point3d;
		double bperror;

		static bool errorComp (const EvalInf* i,const EvalInf* j) { return (i->bperror < j->bperror);}
		static bool revErrorComp (const EvalInf* i,const EvalInf* j) { return (j->bperror < i->bperror);}
		

	};


//Creates a Histogram and picks all bins up to the bin with most of correspondences so correspondence with larger distance are filtered out and in case of moving straight onto an object not all, just the cameras in front are picked
template<typename T> 
std::vector< struct PointCloudCorrection<T>::CorrInf*> PointCloudCorrection<T>::depthRoughFilter(std::vector< const hekate::Correspondence<T>*> &p2dList,const hekate::Correspondence<T> &p3d, const IndexedMatrixList &cameraset, bool relPicking, double histoBins, double minBinsToPick, unsigned int minCorrespondencesToPick) {
		
	 //sout << " Start Depth ";
					 unsigned int minBins;			
					 unsigned int binc;
					//std::vector< const hekate::Correspondence<T>*> p2dDepthFilterList;
					CorrInf *corInf;
				    IndexedMatrix cam;
					std::vector< CorrInf*>  depthList, returnList;

		
					for(unsigned int j =0;j<p2dList.size();j++) {
				
					//Getting the cam for each 2dPoint
		
						cam = cameraset(    (*(p2dList[j])).imgIndex);
						
						//Calc distances				
						corInf = new CorrInf();
							corInf->cor = p2dList[j]; 
							corInf->depth = getDepth(cam, p3d.pt3d() ) ;  //Depth Calculation
							corInf->histoField = 0;
						
						depthList.push_back(corInf);
						
					}
				


					//Determining Depth Range for Histogram
					unsigned int size = p2dList.size();
					double maxD, minD, range, binrange;

					if(!depthList.empty()) {
						maxD = 0; 
						minD = depthList[0]->depth; 
					}
				
					for(unsigned int i= 0;i<depthList.size();i++) {
						if( depthList[i]->depth > maxD) {
							maxD = depthList[i]->depth;
						}
						if( depthList[i]->depth < minD) {
							minD = depthList[i]->depth;
						}
					}
					range = maxD - minD;
		

				//Determine Bins
				if(relPicking) {
					//sout << "relPicking" << std::endl;
					binc = ((double) p2dList.size() )* histoBins ;
					minBins = ((double) binc) * minBinsToPick;
					if(minBins > binc) minBins = binc;
				} else {
					//sout << "Absolutepicking " << std::endl;
					binc = histoBins;
					minBins = minBinsToPick;
					if(minBinsToPick > binc) minBins = binc;
				}


				binrange = range/ ((double) binc);
			
				//Alloc Histogram
				unsigned int *histogram = new unsigned int[binc];
				for(unsigned int i= 0;i<binc;i++)  histogram[i] = 0;

				//Creating Histogram
				for(unsigned int i= 0;i<depthList.size();i++) {
					depthList[i]->histoField = (int) ((depthList[i]->depth - minD) /binrange) +1 ;
					if(depthList[i]->histoField > binc) {
						depthList[i]->histoField = binc;
					}
			
					if( depthList[i]->histoField-1 >= 0 && depthList[i]->histoField-1 < binc)
						(histogram[depthList[i]->histoField-1]  )++;
				}
				/*
				sout <<"histo: " ;
				for(unsigned int i =0;i<segc;i++) {				
					sout << i+1<< ": " <<histogram[i] << " || ";
				}

				sout << std::endl;*/


				//Determines how many segments are going to be picked
				unsigned int biggestBin=0, biggestBinCnt=0;
				for(unsigned int i=0;i<binc;i++) {
					if(histogram[i] >= biggestBinCnt) {
						biggestBinCnt = histogram[i];
						biggestBin = i+1;
					}
				}
				if(biggestBin < minBins) {
					biggestBin = minBins;
				}

			
				//sout << "picking " << biggestSegment << "/" << segc << " || Segrange: " << segrange << std::endl;
				
				//Adding all to ouput			
				unsigned int count = 0;
				for(unsigned int i= 0;i<depthList.size(); ) {
					if(depthList[i]->histoField <= biggestBin) {
						count++;
						returnList.push_back(  depthList[i]  );


						
						depthList.erase(depthList.begin()+i);
					}
					else {
						i++;
					}
				}
			
			    
				//If there are too less correspondences more and more bins are added
				unsigned int lowBin = biggestBin;
				while(count <= minCorrespondencesToPick && biggestBin <= binc) {
					biggestBin++;
					for(unsigned int i= 0;i<depthList.size();i++) {
						if(depthList[i]->histoField <= biggestBin && depthList[i]->histoField > lowBin) {
							count++;						
							returnList.push_back( depthList[i] );
														
							depthList.erase(depthList.begin()+i);
						}
						else {
							i++;
						}
					}
				}
		
			
				//Free Old Memory
				delete [] histogram;			
			  
				for(unsigned int i = 0; i<depthList.size();i++) 
						delete depthList[i];		
				depthList.clear();
				//sout << " end depth ";
				//sout << "depth filtered " << p2dDepthFilterList.size() << " remain"<< std::endl;
			
				return (returnList);
}

//Sorts the cameras by angle (approximation)
template<typename T>
std::vector< struct PointCloudCorrection<T>::AngleInf *> PointCloudCorrection<T>::angleSorter(std::vector< const hekate::Correspondence<T>*> &p2dList,const IndexedMatrixList &cameraset) {
					
				AngleInf *angleInf;
	
				//sout << " Start angle ";
				double tangle;
				const IndexedMatrix* cam1, *cam2;
				const hekate::Correspondence<T> * cor1, *cor2;
				cam1 = &cameraset(    (*(p2dList[0])).imgIndex);
				cor1 = p2dList[0];
			
				std::vector< AngleInf*> tmpList, tmpList2, ordAngleList;	
			
				for(unsigned int i=1;i<p2dList.size();i++) {
					angleInf = new AngleInf();
						angleInf->cam1 = cam1; 
						angleInf->cor1 = cor1;
						angleInf->cam2 = &cameraset(    (*(p2dList[i])).imgIndex);
						angleInf->cor2 = p2dList[i];
						angleInf->angle = getAngle(  getPrincipalAxis(*(angleInf->cam1)) , getPrincipalAxis(*(angleInf->cam2) )   ) ;
				
					tmpList.push_back(angleInf);
				}

				std::sort (tmpList.begin(), tmpList.end(),AngleInf::angleComp);
				cam2 = tmpList.back()->cam2;
				cor2 = tmpList.back()->cor2;
				tangle = tmpList.back()->angle;

				delete tmpList.back();
				tmpList.pop_back();


				//Adding itself to ordAngleList with angle 0.0
				angleInf = new AngleInf();
						angleInf->cam1 = cam2;
						angleInf->cor1 = cor2;
						angleInf->cam2 = cam2;
						angleInf->cor2 = cor2;
						angleInf->angle = 0.0;
						ordAngleList.push_back(angleInf);

				//adding neighbours to ordlist and farer distants to tmpList
				while(tmpList.size()!=0) {
					angleInf = new AngleInf();
						angleInf->cam1 = cam2;
						angleInf->cor1 = cor2;
						angleInf->cam2 = tmpList.back()->cam2;
						angleInf->cor2 = tmpList.back()->cor2;
						angleInf->angle = getAngle(  getPrincipalAxis(*(angleInf->cam1)  ),  getPrincipalAxis(*(angleInf->cam2)  )  ) ;
					
	
						if( std::abs(  angleInf->angle + tmpList.back()->angle   - tangle )< 1e-8) {
							angleInf->angle = tangle + tmpList.back()->angle;
							tmpList2.push_back(angleInf);
						} else {
							ordAngleList.push_back(angleInf);
						}
						delete tmpList.back();
						tmpList.pop_back();
				}
				//ADDING cam1..
				angleInf = new AngleInf();
						angleInf->cam1 = cam2;
						angleInf->cor1 = cor2;
						angleInf->cam2 = cam1;
						angleInf->cor2 = cor1;
						angleInf->angle = getAngle( getPrincipalAxis(*( angleInf->cam1) ) ,  getPrincipalAxis(*(angleInf->cam2)   ) ) ;
							
						ordAngleList.push_back(angleInf);
				
				//running through tmpList
				if(tmpList2.size() > 0) {
					tangle = tmpList2[0]->angle;
				}
	
				
				while(tmpList2.size() != 0) {
					ordAngleList.push_back( tmpList2.back() );
					tmpList2.pop_back();
				}
	
				//delete &tmpList;
				//delete &tmpList2;
	
				//sout << " sorted ";
	
			     return ordAngleList;

				
				
}

template<typename T>
std::vector<  std::vector< const hekate::Correspondence<T>*> >PointCloudCorrection<T>::angleSubsets(std::vector< struct PointCloudCorrection<T>::AngleInf *>& ordAngleList, double minAngle, double maxAngle,  unsigned int minCamCount, unsigned int subsets)  {
	
	std::vector< std::vector< const hekate::Correspondence<T>*>  > p2dAngleFilterList;
	std::vector< std::vector< AngleInf* >> angleList;
			
	//sout <<std::endl << " ANGLE: " ;
	//Checking minimum and maximum angle
	//sout << ordAngleList.size() << std::endl;
	unsigned int innerIndex;
	unsigned int maxSize = 0, msIndex =0;
	//sout << "  PointList "  << p2dList.size() << " OrdList " << ordAngleList.size();
	for(unsigned int i = 0;i<ordAngleList.size();i++) {
		//sout << "OrdListRun "<< i << "  ID1: " << ordAngleList[i]->cam1->getIndex1() << "  Id2: " << ordAngleList[i]->cam2->getIndex1() << "  Angle: " << ordAngleList[i]->angle << std::endl;
		std::vector< AngleInf*> tmpList3;
		tmpList3.push_back(ordAngleList[i]);
		//sout << radToDeg(ordAngleList[i]->angle) << "  ";
				     
		for(unsigned int j=1;j<ordAngleList.size();j++) {
						
			innerIndex = j+i;
			if(innerIndex >= ordAngleList.size()) innerIndex = innerIndex - ordAngleList.size();

			//Since list is sorted we can break when the first Element is out of range
			if(ordAngleList[innerIndex]->angle - tmpList3.front()->angle >= degToRad(maxAngle) ) {
				break;
			}

			//Checking MinimumAngle
			if(ordAngleList[innerIndex]->angle - tmpList3.back()->angle >= degToRad(minAngle)) {
				tmpList3.push_back(ordAngleList[innerIndex]);
			}
		}
		//sout << " ts " << tmpList3.size();
					
		if(tmpList3.size() >= minCamCount) {
			angleList.push_back(tmpList3);
			if(tmpList3.size() > maxSize) {
				maxSize = tmpList3.size();
				msIndex = angleList.size()-1;
			}
		}
	}
	//sout << std::endl;
	if(angleList.size() == 0 || msIndex >= angleList.size() ){						
			// No Subsets
			//sout << " returning 0 subsets " << std::endl;
			return  p2dAngleFilterList;					
	}
	if(angleList[msIndex].size() ==0) {
		return  p2dAngleFilterList;	
	}

	unsigned int subsetcnt = 0;
	unsigned int incr = 0;
	if( subsets <= 0) {		
					
		subsetcnt = angleList.size()/angleList[msIndex].size() ;	
		if(subsetcnt < 1) subsetcnt = 1;
			incr =  angleList.size() / subsetcnt;
	} 				
	else
	if(angleList.size() <= subsets ) {
		subsetcnt = angleList.size();
		incr = 1;
	}
	else {
		subsetcnt = subsets;
		if(subsetcnt == 0) return  p2dAngleFilterList;	
		incr =  angleList.size() / subsetcnt;
	} 
	//sout << " Subsetcnt: " << subsetcnt << std::endl;
	if(incr <= 0) incr = 1;
	//sout << "Subsetcnt: " << subsetcnt << "/" << angleList.size() << std::endl;
	unsigned int cindex = msIndex;
	unsigned int i=0;
	do {
				   
		if(cindex < angleList.size()) {
			std::vector< const hekate::Correspondence<T>* > tmp;
			while(angleList[cindex].size() !=0) {
				tmp.push_back( angleList[cindex].back()->cor2);
				angleList[cindex].pop_back();
			}
			p2dAngleFilterList.push_back(tmp);
				   

			cindex += incr;
			cindex %= angleList.size();
		}
		i++;
	} while(  (p2dAngleFilterList.size() < subsetcnt )&& (i < angleList.size()) );
			   

	return p2dAngleFilterList;
}


template<typename T>
hekate::Correspondence<T>* PointCloudCorrection<T>::triangulation(const IndexedMatrixList& cameraList, std::vector< const hekate::Correspondence<T>*>& points){
	
	if(cameraList.size() < 2) return 0;
	hekate::Correspondence<T>* point;
			if (points.size() >= 2){
			
			typename std::vector< const hekate::Correspondence<T>*>::iterator pointIt;

			try{
				T *A = new T[4 * points.size() * 2];

				int rowsA = 0;
				
				for(pointIt = points.begin(); pointIt != points.end(); pointIt++){
			
					try{
						const hekate::Correspondence<T> &point2d = **pointIt;
						const IndexedMatrix &camera = cameraList(point2d.imgIndex);

						if (camera.degenerated || point2d.outlier){
							//sout << "camera.degenerated: " << camera.degenerated 
							//	<< "point2d.outlier: " << point2d.outlier << std::endl;
							continue;
						}
						
						//first row of A
						A[rowsA*4+0] = point2d.point[0]*camera(0,2) - camera(0,0);
						A[rowsA*4+1] = point2d.point[0]*camera(1,2) - camera(1,0);
						A[rowsA*4+2] = point2d.point[0]*camera(2,2) - camera(2,0);
						A[rowsA*4+3] = point2d.point[0]*camera(3,2) - camera(3,0);
						rowsA++;

						//second row of A
						A[rowsA*4+0] = point2d.point[1]*camera(0,2) - camera(0,1);
						A[rowsA*4+1] = point2d.point[1]*camera(1,2) - camera(1,1);
						A[rowsA*4+2] = point2d.point[1]*camera(2,2) - camera(2,1);
						A[rowsA*4+3] = point2d.point[1]*camera(3,2) - camera(3,1);
						rowsA++;

					}
					catch(...){
					}
				
				}

				if(rowsA < 2){
					return 0;
				}

				CImg<T> X(1,4);

				// If useSVD is set, use CIMG SVD to solve equation
				//if(this->useSVD()){
					cimg_library::CImg<T> AA(4,rowsA);
					for(size_t i = 0; i < rowsA; i++){
						AA(0,i) = A[i*4];
						AA(1,i) = A[i*4+1];
						AA(2,i) = A[i*4+2];
						AA(3,i) = A[i*4+3];
					}
					CImg<T> U, S, V;
					AA.SVD(U,S,V);
					V.crop(3,0,3,3);
					X(0, 0) = V[0] / V[3]; 				
					X(0, 1) = V[1] / V[3]; 
					X(0, 2) = V[2] / V[3]; 
					X(0, 3) = 1; 
				/*}
				else{
					T result[4];
					totalLstsqMinimize(A, rowsA, (T*)&result, false);	

					CImg<T> X(1,4);
					X(0, 0) = result[0] / result[3]; 				
					X(0, 1) = result[1] / result[3]; 
					X(0, 2) = result[2] / result[3]; 
					X(0, 3) = 1; 
				}
				*/


					
					point = new hekate::Correspondence<T>( X, points[0]->pointIndex, 0, false, points[0]->R, points[0]->G, points[0]->B);
				
				delete[] A;
				
			}
			catch(...){

			}		

		}
		return point;
}


template<typename T>
double PointCloudCorrection<T>::bpError(std::vector< const hekate::Correspondence<T>*> &p2dList, const hekate::Correspondence<T> p3d, const IndexedMatrixList &cameraset, double percentage) {
	cimg_library::CImg<T> bp,img;
	IndexedMatrix cam;

	std::vector< double> distanceList;

	double error = 0.0;
	
	for(unsigned int i=0;i<p2dList.size();i++) {
		cam = cameraset(   (*(p2dList[i])).imgIndex);

		bp = cam * p3d.pt3d();
		if(bp[2] != 0.0) {
			bp[0] /= bp[2];
			bp[1] /= bp[2];
		}
		img = (*p2dList[i]).pt2d();
		double dis = std::sqrt( std::pow(   (double)  img[0] - (double) bp[0],2.0 ) + std::pow((double) img[1] - (double) bp[1],2.0 )  );

		distanceList.push_back(dis);		
	}

	std::sort(distanceList.begin(), distanceList.end() );
	unsigned int ccount = ((double) distanceList.size() * percentage);
	/*if(percentage < 1.0) {
			ccount++;
		}*/
	if(ccount <= 0) ccount = distanceList.size();

	if(ccount <=0) return -1.0;

	for( unsigned int i =0;i<ccount;i++) {			
				error += distanceList[i];		
	}
	error /= ccount;

	return error;
}


template<typename T>
typename PointCloudCorrection<T>::EvalInf* PointCloudCorrection<T>::bpEvaluation(std::string name, std::vector< const hekate::Correspondence<T>*> &p2dList,  const IndexedMatrixList &cameraset, double percentage, unsigned int minCount) {
	
	cimg_library::CImg<T> bp,img;
	IndexedMatrix cam;
	EvalInf* ev = new EvalInf();
	ev->name = name;
	std::vector< CorrInf*> distanceList;

	double error = 0.0;
	const hekate::Correspondence<T>* p3d = triangulation(cameraset,p2dList);

	
	for(unsigned int i=0;i<p2dList.size();i++) {
		cam = cameraset(   (*(p2dList[i])).imgIndex);

		bp = cam * p3d->pt3d();
		if(bp[2] != 0.0) {
			bp[0] /= bp[2];
			bp[1] /= bp[2];
		}
		img = (*p2dList[i]).pt2d();
		double dis = std::sqrt( std::pow(   (double)  img[0] - (double) bp[0],2.0 ) + std::pow((double) img[1] - (double) bp[1],2.0 )  );
		CorrInf * corinf = new CorrInf();
		corinf->cor = p2dList[i];
		corinf->depth = dis;
		distanceList.push_back(corinf);		
	}
	
	std::sort(distanceList.begin(), distanceList.end() , CorrInf::depthComp);
	unsigned int ccount = ((double) distanceList.size() * percentage);
	/*if(percentage < 1.0) {
			ccount++;
		}*/
	if (ccount < minCount) ccount = distanceList.size();


	if (ccount < minCount) { 
		error = -1.0;
	} else {
		for( unsigned int i =0;i<ccount;i++) {			
					error += distanceList[i]->depth;		
					ev->points2d.push_back(distanceList[i]->cor);
					delete distanceList[i];
		}
		error /= ccount;
	}

	ev->bperror = error;

	//Plugin crashes on deleting .. ??
	//delete &p3d;

	if(ev->bperror > 0.0) {
		return ev;
	}else {
		delete ev;
		return 0;
	}
}


template <typename T>
void PointCloudCorrection<T>::execute() {
	
	//Input
	const IndexedMatrixList& cams =this->cameras();
	const hekate::CorrespondenceList<T>& ps2d =this->points2d();
	const hekate::CorrespondenceList<T>& ps3d =this->points3d();	

	//Parameters
	bool relPicking			= this->relativeBinPicking();
	double histoBins		= this->histogramBins();
	double minBinsToPick	= this->minBinsToPick();
	unsigned int minCorrespondencesToPick = this->minCorrespondencesToPick();

	double minAngle = this->minAngle();
	double maxAngle = this->maxAngle();

	bool depthFilterP = this->depthFilterEnabled();
	bool angleFilterP = this->angleFilterEnabled();
	
	double totalErrorBefore=0.0, totalErrorAfter=0.0;
	std::vector<double> errorDifflist, absErrorBefore, absErrorAfter;
	
	std::vector< const hekate::Correspondence<T>*> p2dList;
		
	 
	sout << "########POINTCLOUDCORRECTION#########" << std::endl;
	sout << "Starting with " << ps2d.size() << " correspondences" << std::endl;	 

	if(!depthFilterP && !angleFilterP) {
			sout << " No Filter Selected, triangulating old correspondences " << std::endl;
			this->correctedPointCloud() = this->points3d();
			return;
	}

	//Walking through each Point
	try {
	for(hekate::CorrespondenceList<T>::const_iterator p3d = ps3d.begin(); p3d != ps3d.end(); ++p3d)
	{

		//Get 2d Points for each Point
		p2dList =  ps2d.getListByPointId((*p3d).pointIndex);
		if(p2dList.size() < 2) {
			sout << "Warning less than 2 correspondences for pointIndex ... taking next point." << (*p3d).pointIndex << std::endl;
			this->correctedPointCloud().push_back(*p3d);
			continue;
		}
					
		std::vector< CorrInf* > roughFiltered;			
		std::vector< const Correspondence<T>* > upperDepthList, lowerDepthList, wholelist;
		std::vector< EvalInf* > errorList;
	
		EvalInf* ev  = bpEvaluation("oldlist", p2dList,cams,this->backProjectionErrorThreshold(), p2dList.size());
		if(ev!=0) { 
			totalErrorBefore += ev->bperror;
			absErrorBefore.push_back(ev->bperror);
			errorList.push_back(ev); 
		}
				/*EvalInf* ev = new EvalInf();
				ev->name = "oldlist";
				ev->points2d = p2dList;
				ev->point3d = triangulation(cams,p2dList);
				ev->bperror = bpError(p2dList,*(ev->point3d),cams,this->backProjectionErrorThreshold());
				if(ev->bperror >=0.0) {
					totalErrorBefore += ev->bperror;
					absErrorBefore.push_back(ev->bperror);
					errorList.push_back(ev);
				} else {
					//delete ev->point3d;
					//delete ev;
				}*/
			//absErrorBefore.push_back(*(ev->point3d));
			//Depth Filtering
			if(depthFilterP) {
				// sout << "DepthFiltering" << std::endl;
				 roughFiltered = depthRoughFilter(p2dList,*p3d, cams, relPicking,  histoBins,  minBinsToPick,  minCorrespondencesToPick);
				 unsigned int lowerindex =  (double) roughFiltered.size() * this->depthSubsetLowerThreshold()+1,  upperindex = (double) roughFiltered.size() * this->depthSubsetUpperThreshold()  ;
				 if(this->depthSubsetLowerThreshold() <= 0.0) {
					 lowerindex = 0;
				 }
				 if(this->depthSubsetUpperThreshold() >= 1.0) {
					 upperindex = roughFiltered.size();
				 }

				 for(unsigned int i = 0;i< roughFiltered.size() ;i++) {
					 wholelist.push_back(roughFiltered[i]->cor);
					 if(i < lowerindex) {
						 lowerDepthList.push_back(roughFiltered[i]->cor);
					 }
					 if(i >= upperindex) {
						 upperDepthList.push_back(roughFiltered[i]->cor);
					 }
				 }			

				 //Creation EvalInf
				 if(wholelist.size() >= minCorrespondencesToPick) {
					 	ev  = bpEvaluation("DepthWholeList", wholelist,cams,this->backProjectionErrorThreshold(), minCorrespondencesToPick);
						if(ev!=0) { errorList.push_back(ev); }
				 /*EvalInf* ev = new EvalInf();
				 		ev->name = "DepthWholeList";
						ev->points2d = wholelist;
						ev->point3d = triangulation(cams,wholelist);
						ev->bperror = bpError(wholelist,*(ev->point3d),cams,this->backProjectionErrorThreshold());
						if(ev->bperror >=0.0) {
							errorList.push_back(ev);
						} else {
							//delete ev->point3d;
							//delete ev;
						}*/
				 }

				 if(lowerDepthList.size() >=minCorrespondencesToPick) {
					 	ev  = bpEvaluation("DepthLowerList", lowerDepthList,cams,this->backProjectionErrorThreshold(), minCorrespondencesToPick);
						if(ev!=0) { errorList.push_back(ev); }
					/*ev = new EvalInf();
				 		ev->name = "DepthLowerList";
						ev->points2d = lowerDepthList;
						ev->point3d = triangulation(cams,lowerDepthList);
						ev->bperror = bpError(lowerDepthList,*(ev->point3d),cams,this->backProjectionErrorThreshold());
						if(ev->bperror >=0.0) {
							errorList.push_back(ev);
						} else {
							//delete ev->point3d;
							//delete ev;
						}*/
				 }

				 if(upperDepthList.size() >= minCorrespondencesToPick) {
					 	ev  = bpEvaluation("DepthUpperList", upperDepthList,cams,this->backProjectionErrorThreshold(), minCorrespondencesToPick);
						if(ev!=0) { errorList.push_back(ev); }
					/*ev = new EvalInf();
				 		ev->name = "DepthUpperList";
						ev->points2d = upperDepthList;
						ev->point3d = triangulation(cams,upperDepthList);
						ev->bperror = bpError(upperDepthList,*(ev->point3d),cams,this->backProjectionErrorThreshold());
						if(ev->bperror >=0.0) {
							errorList.push_back(ev);
						} else {
							//delete ev->point3d;
							//delete ev;
						}*/
				 }

				 //Freeing memory
						for(unsigned int i=0;i < roughFiltered.size();i++) {
							//delete roughFiltered[i];
						}
						roughFiltered.clear();
						//sout << "DepthFinish" << std::endl;
			}	
			
		//END DEPTH FILTERING
			
			
			//CAM ANGLE FILTERING			
			if(angleFilterP) {
				//sout << "AngleFiltering" << std::endl;
				unsigned int oldsize = errorList.size();
				for( unsigned int i=0;i< oldsize;i++) {
					std::vector< AngleInf *> ordAngleList = angleSorter(errorList[i]->points2d,  cams);

					std::vector < std::vector < const hekate::Correspondence<T>* > > subsets = angleSubsets(ordAngleList,minAngle,  maxAngle,   minCorrespondencesToPick, this->angleSubsetcount());
					
						for(unsigned int j=0;j<subsets.size();j++) {
							if( subsets[j].size() < 2) continue;
							EvalInf* ev  = bpEvaluation(errorList[i]->name + " anglefiltered", subsets[j],cams,this->backProjectionErrorThreshold(), minCorrespondencesToPick);
							if(ev!=0) { errorList.push_back(ev); }
							/*EvalInf* ev = new EvalInf();
							ev->name = errorList[i]->name + " anglefiltered";
							ev->points2d = subsets[j];
							ev->point3d = triangulation(cams,subsets[j]);
							ev->bperror = bpError(subsets[j],*(ev->point3d),cams,this->backProjectionErrorThreshold());			    
							
								if(ev->bperror >=0.0) {
									errorList.push_back(ev);
								} else {
									delete ev->point3d;
									delete ev;
								}
							*/
						}

						for(unsigned int i=0;i<ordAngleList.size();i++) {
							delete ordAngleList[i];
						}
						//ordAngleList.clear();
						
				}

		
				//sout << "AngleFInish" << std::endl;
			}			

			//Olderror
			double errordiff = errorList[0]->bperror;

			if(!this->suppressOutputForEachPoint()) sout <<"PointIndex: " << (*p3d).pointIndex << " Correspondences: " << p2dList.size() << " Subsets: " << errorList.size() << std::endl;
			

			//Sorting to get lowest backprojection error
			std::sort (errorList.begin(), errorList.end(),EvalInf::errorComp);
			
			//new reduced error (for histogram)
			errordiff -= errorList[0]->bperror;
			errorDifflist.push_back(errordiff);
			//Output....
			if(!this->suppressOutputForEachPoint()) {
			for(unsigned int i=0;i< errorList.size();i++) {
					sout << errorList[i]->name << " " << errorList[i]->bperror << std::endl;
			}
			}
			if(!this->suppressOutputForEachPoint()) sout << "-----" << std::endl;

			totalErrorAfter += errorList[0]->bperror;
			absErrorAfter.push_back(errorList[0]->bperror);
			//this->correctedPointCloud().push_back(*(errorList[0]->point3d));

			// new point
			const hekate::Correspondence<T> * correctedPoint = triangulation(cams,errorList[0]->points2d);
			if(correctedPoint!=0) {
				this->correctedPointCloud().push_back(  *correctedPoint );				
			} 
			//sout << "Added Index: " << (*p3d).pointIndex << std::endl;

			//Deleting
			delete errorList[0];
			for(unsigned int i=1;i<errorList.size();i++) {
				//delete errorList[i]->point3d;
				delete errorList[i];
			}


			//sout << "PointAdded" << std::endl;
			//END CAM ANGLE FILTERING
			
		//	this->correctedPointCloud()
		} //End For each 3dpoint
	}
	catch (...) {
		//throw( std::runtime_error("catch(...)") );
	}


	//Evaluating Error Histogram	
	if(this->errorDiffHistogramBinCount() > 0  && errorDifflist.size() > 0) {
		std::sort(errorDifflist.begin(),errorDifflist.end());
		
		unsigned int *errorhisto = new unsigned int[this->errorDiffHistogramBinCount()];
		for(unsigned int i =0;i<this->errorDiffHistogramBinCount(); i++) {
			errorhisto[i]=0;
		}

		double errorbinrange = errorDifflist.back() - errorDifflist.front();
			   errorbinrange /= this->errorDiffHistogramBinCount();

		errorhisto[0]++;
		errorhisto[this->errorDiffHistogramBinCount()-1]++;

		for(unsigned int i=1;i<errorDifflist.size()-1;i++) {
			unsigned int index = (unsigned int) (errorDifflist[i] - errorDifflist.front() ) / errorbinrange + 1; 
			if(index > 0 && index < this->errorDiffHistogramBinCount()) {
				errorhisto[index]++;
			}
		}

		sout << "### Backprojection Errordifference   Histogram ###" << std::endl;
		sout << "Binrange: Correspondences" << std::endl;
		for(unsigned int i =0;i<this->errorDiffHistogramBinCount(); i++) {
			sout << errorDifflist.front() + i*errorbinrange << " - " << errorDifflist.front() + (i+1) * errorbinrange << ":\t\t" << errorhisto[i] << std::endl;
		}

		//TODO HISTOGRAMM
		//
		//ABsolute Histogramm
		
		std::sort(absErrorBefore.begin(), absErrorBefore.end());
		std::sort(absErrorAfter.begin(), absErrorAfter.end());
		unsigned int *absBeforeErrorHisto = new unsigned int[this->errorDiffHistogramBinCount()];
		unsigned int *absAfterErrorHisto = new unsigned int[this->errorDiffHistogramBinCount()];
		
		
		for(unsigned int i =0;i<this->errorDiffHistogramBinCount(); i++) {
			absBeforeErrorHisto[i]=0;
			absAfterErrorHisto[i]=0;
		}
		double abserrorbinrange = absErrorBefore.back() - absErrorBefore.front();
			   abserrorbinrange /= this->errorDiffHistogramBinCount();
			   
		absBeforeErrorHisto[0]++;
		absBeforeErrorHisto[this->errorDiffHistogramBinCount()-1]++;
		absAfterErrorHisto[0]++;
		absAfterErrorHisto[this->errorDiffHistogramBinCount()-1]++;

		for(unsigned int i=1;i<absErrorBefore.size()-1;i++) {
			unsigned int index = (unsigned int) (absErrorBefore[i] - absErrorBefore.front() ) / abserrorbinrange + 1; 
			if(index > 0 && index < this->errorDiffHistogramBinCount()) {
				absBeforeErrorHisto[index]++;
			}
		}
		for(unsigned int i=1;i<absErrorAfter.size()-1;i++) {
			unsigned int index = (unsigned int) (absErrorAfter[i] - absErrorAfter.front() ) / abserrorbinrange + 1; 
			if(index > 0 && index < this->errorDiffHistogramBinCount()) {
				absAfterErrorHisto[index]++;
			}
		}
		
		sout << "### Absolute BackprojectionError Histogram		        	 ###" << std::endl;
		sout << "---Binrange :  old Correspondences ||  new Correspondences  ---" << std::endl;
		for(unsigned int i =0;i<this->errorDiffHistogramBinCount(); i++) {
			sout << absErrorBefore.front() + i*abserrorbinrange << " - " << absErrorBefore.front() + (i+1) * abserrorbinrange << ":\t\t" << absBeforeErrorHisto[i] << "\t||\t"<< absAfterErrorHisto[i]<< std::endl;
		}
	}

	sout << "TotalError before: " << totalErrorBefore << "  afterwards: " << totalErrorAfter << " (" << (totalErrorAfter/totalErrorBefore*100) << ") " << std::endl;

	sout << "Finishing with " << this->correctedPointCloud().size()<< " 3dpoints" << std::endl;

}




#endif /* _POINTCLOUDCORRECTION_HXX_ */
