/*  Copyright (C) 2009 SasiKanth

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file projectivesba.hxx
 *  Implementation of the parameter class projectivesba.
 *  \author SasiKanth
 *  \date 08.07.2010
 */

#ifndef _PROJECTIVESBA_HXX_
#define _PROJECTIVESBA_HXX_

#include <hekate/ProjectiveSba.h>
#include <charon-utils/ImgTool.hxx>

#include <sba/sba.h>

template <typename T>
projectivesba<T>::projectivesba(const std::string& name) :
		TemplatedParameteredObject<T>(
				"projectivesba", name,
				"This plugin performs sparse bundle adjustment on a projective camera using the sba package"
		)
			,input_3d_points(false, false) 
			,input_correspondences(false, false) 
			,input_p_matrices(false, false) 
			
			
{
	

	

	ParameteredObject::_addInputSlot(input_3d_points, "input_3d_points", 
	"The set of input 3D points", 
	"std::vector<std::vector<T> >"); 
	ParameteredObject::_addInputSlot(input_correspondences, "input_correspondences", 
	"The set of input correspondences", 
	"std::vector<std::vector<std::pair<T, T> > >"); 
	ParameteredObject::_addInputSlot(input_p_matrices, "input_p_matrices", 
	"The set of camera matrices", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(output_3d_points, "output_3d_points", 
	"The adjusted 3D points", 
	"std::vector<std::vector<T> >"); 
	ParameteredObject::_addOutputSlot(output_p_matrices, "output_p_matrices", 
	"The adjusted camera matrices", 
	"CImgList<T>"); 
	
}

void func1(int, int, double* aj, double* bi, double* xij, void*) {
	cimg_library::CImg<double> P_matrix(4,3), point1(1,4), point2(1,3);
	
	// Assign the values to the P-matrix
	P_matrix(0,0) = aj[0];
	P_matrix(0,1) = aj[1];
	P_matrix(0,2) = aj[2];
	P_matrix(1,0) = aj[3];
	P_matrix(1,1) = aj[4];
	P_matrix(1,2) = aj[5];
	P_matrix(2,0) = aj[6];
	P_matrix(2,1) = aj[7];
	P_matrix(2,2) = aj[8];
	P_matrix(3,0) = aj[9];
	P_matrix(3,1) = aj[10];
	P_matrix(3,2) = aj[11];

	// Assigning the values of the 3D point
	point1(0,0) = bi[0];
	point1(0,1) = bi[1];
	point1(0,2) = bi[2];
	point1(0,3) = bi[3];

	// Estimating the relative xij
	point2 = P_matrix*point1;

	// Storing the 2d point xij
	xij[0] = point2(0,0)/point2(0,2);
	xij[1] = point2(0,1)/point2(0,2);
}

template <typename T>
void projectivesba<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	// Beginning of Plugin Code

	int	n = input_3d_points().size(), // no. of 3D points
		m = 2,	// no. of cameras
		mcon = 1, // the first camera is constant [I|O]
		cnp = 12, // no. of camera parameters for a projective camera
		pnp = 4, // no. of parameters for a 3D point (homogeneous)
		mnp = 2,
		itmax = 1000, // maximum no. of iterations for the levenberg marquardt implementation
		i, j; 

	double opts[SBA_OPTSSZ], info[SBA_INFOSZ], *p, *x;
	p = new double[n*pnp + m*cnp];
	x = new double[m*n*mnp];

	char *vmask;
	vmask = new char[m*n];

	// Initializing all values of vmask to 1

	for(i=0; i<m*n; i++)
	{
		vmask[i] = 1;
	}

	// Loading the parameter values

	for(i=0; i<m; i++)
	{
		p[cnp*i] = input_p_matrices()[i](0,0);
		p[cnp*i+1] = input_p_matrices()[i](0,1);
		p[cnp*i+2] = input_p_matrices()[i](0,2);
		p[cnp*i+3] = input_p_matrices()[i](1,0);
		p[cnp*i+4] = input_p_matrices()[i](1,1);
		p[cnp*i+5] = input_p_matrices()[i](1,2);
		p[cnp*i+6] = input_p_matrices()[i](2,0);
		p[cnp*i+7] = input_p_matrices()[i](2,1);
		p[cnp*i+8] = input_p_matrices()[i](2,2);
		p[cnp*i+9] = input_p_matrices()[i](3,0);
		p[cnp*i+10] = input_p_matrices()[i](3,1);
		p[cnp*i+11] = input_p_matrices()[i](3,2);
	}

	for(j=0; j<n; j++)
	{
		p[cnp*m + pnp*j] = input_3d_points()[j][0];
		p[cnp*m + pnp*j + 1] = input_3d_points()[j][1];
		p[cnp*m + pnp*j + 2] = input_3d_points()[j][2];
		p[cnp*m + pnp*j + 3] = 1;
	}

	// Loading the measurement values

	for(i=0; i<n; i++)
	{
		for(j=0; j<m; j++)
		{
			x[mnp*(m*i + j)] = input_correspondences()[j][i].first;
			x[mnp*(m*i + j)+1] = input_correspondences()[j][i].second;
		}
	}

	// Calling the SBA function
	sba_motstr_levmar(n, 0, m, mcon, vmask, p, cnp, pnp, x, NULL, mnp, func1, NULL, NULL, itmax, 2, opts, info);

	// Assigning the new values to the output
	cimg_library::CImg<T> P_left(4,3), P_right(4,3); 

	// Assigning the values to the P matrices
	P_left(0,0) = p[0];
	P_left(0,1) = p[1];
	P_left(0,2) = p[2];
	P_left(1,0) = p[3];
	P_left(1,1) = p[4];
	P_left(1,2) = p[5];
	P_left(2,0) = p[6];
	P_left(2,1) = p[7];
	P_left(2,2) = p[8];
	P_left(3,0) = p[9];
	P_left(3,1) = p[10];
	P_left(3,2) = p[11];

	P_right(0,0) = p[12];
	P_right(0,1) = p[13];
	P_right(0,2) = p[14];
	P_right(1,0) = p[15];
	P_right(1,1) = p[16];
	P_right(1,2) = p[17];
	P_right(2,0) = p[18];
	P_right(2,1) = p[19];
	P_right(2,2) = p[20];
	P_right(3,0) = p[21];
	P_right(3,1) = p[22];
	P_right(3,2) = p[23];

	// Assigning the values to the adjusted 3D points
	std::vector<T> vecX;

	for(i=0; i<n; i++) {
		vecX.push_back(p[cnp*m + pnp*i]/p[cnp*m + pnp*i+3]);
		vecX.push_back(p[cnp*m + pnp*i+1]/p[cnp*m + pnp*i+3]);
		vecX.push_back(p[cnp*m + pnp*i+2]/p[cnp*m + pnp*i+3]);

		output_3d_points().push_back(vecX);

		vecX.clear();
	}

	output_p_matrices().push_back(P_left);
	output_p_matrices().push_back(P_right);

	sout << "SBA returned in " << info[5] << " iterations, reason " <<  info[6] << " , error " << info[1]/n << "[" << info[0]/n << "]" << std::endl;

	ImgTool::printInfo(sout,P_left,"");
	ImgTool::printInfo(sout,P_right,"");
		
}

#endif /* _PROJECTIVESBA_HXX_ */



