/*  This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/

/** @file RectifyUncallibratedImages.hxx
 *  @brief Implementation of the class RectifyUncallibratedImages.
 *  Most Parts adapted from a Matlab code by Andrea Fusiello, 2007
 *  Implementation as charon plugin by Julian Coordts, 2010
 *  This Plugin creates the rectification Matrices for two uncallibrated Images
 *  using the two Images and the feature correspondence vector between them.
 *  Those two Matrices can then be used to a warp the images.
 *  @author <a href="mailto:coordts@stud.uni-heidelberg.de">Julian Coordts</a>
 *
 *  @date 10.04.2009
 */

#ifndef _RECITFYUNCALLIBRATEDIMAGES_HXX_
#define _RECITFYUNCALLIBRATEDIMAGES_HXX_

#include <hekate/RectifyUncallibratedImages.h>
#include <vector>
#include <math.h>
#include <algorithm>

template <typename T>
cimg_library::CImg<T> rectifyUncallibratedImages<T>::star(T x1, T x2, T x3)
{
	/*
	 * STAR Return the skew-symmetric matrix X s.t. cross(x,y) = X*y
	 * X is s.t. ker(X) = x
	 * The name is after the Hodge star operator, of which this is an instance.
	*/
	cimg_library::CImg<T> res(3,3);
	res(0,0)=0;		res(1,0)=-x3;	res(2,0)=x2;
	res(0,1)=x3;	res(1,1)=0;		res(2,1)=-x1;
	res(0,2)=-x2;	res(1,2)=x1;	res(2,2)=0;
	return res;

}
template <typename T>
cimg_library::CImg<T> rectifyUncallibratedImages<T>::calcBB(cimg_library::CImg<T>&H,
															cimg_library::CImg<T>&Image)
{
//Calc Bounding Boxes
//Notice: 	For the calculation of H*v the upper left point of v is(1,1) not (0,0)
//			and the right bottom (width,heigth)!!!!

					 cimg_library::CImg<T> oldPoint(1,3);
					 cimg_library::CImg<T> newPoint(1,3);

					 //Calc Bounding Boxes left
					 int xmax=0;
					 int ymax=0;
					 int xmin=0;
					 int ymin=0;

							//Transform Point(0,0)
					        oldPoint (0,0)= 0; oldPoint (0,1)= 0; oldPoint (0,2)= 1;
	 						newPoint= H* oldPoint;
	 						newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
							newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

							if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
							if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
							if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
							if (newPoint(0,1) > ymax) ymax = newPoint(0,1);


							//Transform Point(0,Image.height()-1)
					        oldPoint (0,0)= 0; oldPoint (0,1)= Image.height(); oldPoint (0,2)= 1;
	 						newPoint= H* oldPoint;
	 						newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
							newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

							if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
							if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
							if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
							if (newPoint(0,1) > ymax) ymax = newPoint(0,1);

							//Transform Point(Image.width()-1,0)
					        oldPoint (0,0)= Image.width(); oldPoint (0,1)= 0; oldPoint (0,2)= 1;
	 						newPoint= H* oldPoint;
	 						newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
							newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

							if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
							if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
							if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
							if (newPoint(0,1) > ymax) ymax = newPoint(0,1);

							//Transform Point(Image.width()-1,Image.height()-1)
					        oldPoint (0,0)= Image.width(); oldPoint (0,1)= Image.height(); oldPoint (0,2)= 1;
	 						newPoint= H* oldPoint;
	 						newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
							newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

							if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
							if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
							if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
							if (newPoint(0,1) > ymax) ymax = newPoint(0,1);


sout << "BB: [" << xmin  <<", "   << ymin << "][" << xmax << ", " << ymax << "]" << std::endl;

cimg_library::CImg<T> res(2,2);
res(0,0)=xmin;res(1,0)=xmax;
res(0,1)=ymin;res(1,1)=ymax;
return res;
}


template <typename T>
void rectifyUncallibratedImages<T>::calcHlHr(double p[6],
											 T width,
											 T height,
											 cimg_library::CImg<T> &Hl,
											 cimg_library::CImg<T> &Hr)
{
//	sout << "calcHlHr called ";
//	sout << p[0] << ", "
//						 << p[1] << ", "
//						<< p[2] << ", "
//						<< p[3] << ", "
//						<< p[4] << ", "
//						<< p[5] << "] width:"
//						<< width << " height: " << height << std::endl;


	double yl=p[0];
	double zl=p[1];
	double xr=p[2];
	double yr=p[3];
	double zr=p[4];
	double f =pow(3,p[5])*(width+height);

	// the old intrinsics
	cimg_library::CImg<T> Kol(3,3);
	Kol(0,0)=f; Kol(1,0)= 0; Kol(2,0)= width/2.0f;
	Kol(0,1)=0; Kol(1,1)= f; Kol(2,1)= height/2.0f;
	Kol(0,2)=0; Kol(1,2)= 0; Kol(2,2)=   1;

	cimg_library::CImg<T> Kor(3,3);
	Kor(0,0)=f; Kor(1,0)= 0; Kor(2,0)= width/2.0f;
	Kor(0,1)=0; Kor(1,1)= f; Kor(2,1)= height/2.0f;
	Kor(0,2)=0; Kor(1,2)= 0; Kor(2,2)=   1.0f;
//	soutCImgMatrix(Kor,"Kor");
//	soutCImgMatrix(Kol,"Kol");

	//rotations
	cimg_library::CImg<T> Rl(3,3), Rr(3,3);
	Rl = eulR(0,yl,zl);
	Rr = eulR(xr,yr,zr);
//	soutCImgMatrix(Rr,"Rr");
//	soutCImgMatrix(Rl,"Rl");

	// new intrinsics: arbitrary
	cimg_library::CImg<T> Knl(3,3);
	Knl = (Kol+Kor)/2.0f;
	cimg_library::CImg<T> Knr(Knl);

	// compute collineations
	Hr = Knr*Rr*(Kor.get_invert());
	Hl = Knl*Rl*(Kol.get_invert());


	if(cntr==true)
	{
		 // centering LEFT image
		 	 cimg_library::CImg<T> px(1,3), tmp(1,3);
		 	 tmp(0,0)=width/2.0f ; tmp(0,1)=height/2.0f ; tmp(0,2)=1.0f ;
		     px = Hl * tmp;
//		     soutCImgMatrix(px,"px");
		     cimg_library::CImg<T> dl(1,2);
		     dl(0,0)= width/2 - px(0,0)/px(0,2);
		     dl(0,1)= height/2 - px(0,1)/px(0,2);

//		 	soutCImgMatrix(dl,"dl");
		    // centering RIGHT image
				px = Hr * tmp;
				cimg_library::CImg<T> dr(1,2);
				dr(0,0) = width/2 - px(0,0)/px(0,2);
				dr(0,1)	= height/2 - px(0,1)/px(0,2);

//				soutCImgMatrix(dr,"dr");
		    //vertical diplacement must be the same
				dr(0,1) = dl(0,1);

		    // modify new intrinsic
				Knl(2,0)=Knl(2,0)+dl(0,0);
				Knl(2,1)=Knl(2,1)+dl(0,1);
				Knr(2,0)=Knr(2,0)+dr(0,0);
				Knr(2,1)=Knr(2,1)+dr(0,1);

//				soutCImgMatrix(Knl,"newKnl");
//				soutCImgMatrix(Knr,"newKnr");
		    // re-compute collineations with centering
				Hr = Knr*Rr*(Kor.get_invert());
				Hl = Knl*Rl*(Kol.get_invert());
	}

//	soutCImgMatrix(Hl,"Hl");
//	soutCImgMatrix(Hr,"Hr");

}


template <typename T>
cimg_library::CImg<T> rectifyUncallibratedImages<T>::eulR(T xa, T ya, T za)
{
	/*
	 *eulR compure a rotation matrix given three Euler angles
	 *eulR  Rotations are applied in the order y-z-x (first y, last x)
	 *Parameter a Format: (rotx,roty,rotz)
	 *TODO explicit calculate res without using R1,R2,R3
	 */

	cimg_library::CImg<T> R1(3,3),R2(3,3),R3(3,3);

	R1(0,0)=1; 			R1(1,0)=0;			R1(2,0)=0;
	R1(0,1)=0; 			R1(1,1)=cos(xa);	R1(2,1)=-sin(xa) ;
	R1(0,2)=0; 			R1(1,2)=sin(xa);	R1(2,2)=cos(xa) ;

	R2(0,0)=cos(za);	R2(1,0)=-sin(za);	R2(2,0)=0;
	R2(0,1)=sin(za);	R2(1,1)=cos(za);	R2(2,1)=0 ;
	R2(0,2)=0;			R2(1,2)=0;			R2(2,2)=1 ;

	R3(0,0)=cos(ya);	R3(1,0)=0;			R3(2,0)=sin(ya);
	R3(0,1)=0;			R3(1,1)=1;			R3(2,1)=0;
	R3(0,2)=-sin(ya);	R3(1,2)=0;			R3(2,2)=cos(ya);

	cimg_library::CImg<T> res(3,3);
	res= R1*R2*R3;
	return res;
}

template <>
inline cimg_library::CImg<int> rectifyUncallibratedImages<int>::eulR(int, int, int ) {
	throw std::invalid_argument("rectifyUncallibratedImages not defined for int");
	return cimg_library::CImg<int>();
}

template <typename T>
void rectifyUncallibratedImages<T>::costRectif(double *p, double *x, int, int, void *data)
{

	/* Calculates Rectify Costs
	 *
	 * p = input vector
	 * x = output vector
	 * m = input dimension
	 * n=  output dimension
	 * *data = specific Data passed through levmar, in our case
	 *         a strukt containing width, height and the point correspondences
	 *
	 * p[0]= Y-Left rot.
	 * p[1]= Z-Left rot.
	 * p[2]= X-Right rot.
	 * p[3]= Y-Right rot.
	 * p[4]= Z-Right rot
	 * p[5]= focal lenght
	 *
	 *
	 */

	rectifyUncallibratedImages<T>* cthis = (rectifyUncallibratedImages<T>*) callBackObject;
// extract data
	int w = ((costData*)data)->w; //width
	int h = ((costData*)data)->h; //Height
	cimg_library::CImg<T>* pml = ((costData*)data)->ml;
	cimg_library::CImg<T>* pmr = ((costData*)data)->mr;
	cimg_library::CImg<T> m1(*pml);
	cimg_library::CImg<T> m2(*pmr);

//extract params

	double yl,zl,xr,yr,zr,f,p5;
	yl= p[0]; zl=p[1]; xr=p[2] ;
	yr= p[3]; zr=p[4]; p5= p[5]; f=pow(3.,p5)*(w+h);

//	sout << "costRectif called(";
//	sout << "w: " << w << " h: " << h << ")";
//	sout << " p: [";
//	sout << p[0] << ", "
//					 << p[1] << ", "
//					<< p[2] << ", "
//					<< p[3] << ", "
//					<< p[4] << ", "
//					<< p[5] << "]" << std::endl;
//	cthis->soutCImgMatrix(m1,"m1");
//	cthis->soutCImgMatrix(m2,"m2");



// estimate of the intrinsic parameters of the old cameras
cimg_library::CImg<T> Kol(3,3);
Kol(0,0)=f; Kol(1,0)= 0; Kol(2,0)= w/2;
Kol(0,1)=0; Kol(1,1)= f; Kol(2,1)= h/2;
Kol(0,2)=0; Kol(1,2)= 0; Kol(2,2)=   1;

//cthis->soutCImgMatrix(Kol,"Kol/Kor");


// eulR applies rotations in the order Y-Z-X
cimg_library::CImg<T> Kor(Kol);
cimg_library::CImg<T> Rl(3,3), Rr(3,3);

Rl = cthis->eulR(0,yl,zl);
Rr = cthis->eulR(xr,yr,zr);

//cthis->soutCImgMatrix(Rl,"Rl");
//cthis->soutCImgMatrix(Rr,"Rr");

// fundamental matrix btw original points
cimg_library::CImg<T> F(3,3);

cimg_library::CImg<T> korinvtrans(Kor.get_invert());
korinvtrans.transpose();
cimg_library::CImg<T> Rrtrans(Rr.get_transpose());
cimg_library::CImg<T> Kolinv(Kol.get_invert());


F= korinvtrans * Rrtrans;
F= F * cthis->star(1,0,0);
F= F * Rl;
F= F * Kolinv;


//cthis->soutCImgMatrix(F,"F");

//cthis->soutCImgMatrix(m1,"m1");
//cthis->soutCImgMatrix(m2,"m2");

cimg_library::CImg<T> U3;
U3=cthis->star(0,0,1);
//cthis->soutCImgMatrix(U3,"U3");
double erg=0;
	for (int i=0; i<m1.width(); ++i)
		{
			cimg_library::CImg<T> ufm1(1,3);
			 ufm1 = F*m1.get_column(i);
			 ufm1=  U3*ufm1;
			 //cthis->soutCImgMatrix(ufm1,"ufm1");
			 cimg_library::CImg<T> m2fu(3,1);
			 m2fu=m2.get_column(i).transpose()*F*U3;
			 //cthis->soutCImgMatrix(m2fu,"m2fu");

			double a= (double) (m2.get_column(i).transpose()*F*m1.get_column(i))(0,0);
            double b =(double) (ufm1.get_transpose()*ufm1)(0,0)  + (double) (m2fu*m2fu.get_transpose())(0,0);
			//sout << "a: " << a << " b: " << b << std::endl;
            double s = sqrt(a*a / b);
            //sout << s << std::endl;
			x[i]=s;
			erg+=s*s;

		}
	sout /*<< "\r"*/ <<  sqrt(erg) << std::endl;
	//cthis->soutVectorArray(x,m1.width(),"x");

}



template <typename T>
void rectifyUncallibratedImages<T>::soutCImgMatrix(cimg_library::CImg<T> &m,const char* text = "")
{
	/*
	 * outputs a CImg in human readable matrix format through sout
	 */

	sout << "Output Matrix: " << text << " ("  << m.height() <<  "x" << m.width() << ")" << std::endl;
	for (int y=0;y<m.height();y++)
	{
		sout << "\t";
		for (int x=0;x<m.width();x++)
		{
			sout.width(5);
			sout << m(x,y) << "  ";
		}
		sout << std::endl;
	}
}

template <typename T>
void rectifyUncallibratedImages<T>::soutVectorArray(const double *ar,const int length, const char* text)
{
	/*
	 * outputs an array vector in human readable format through sout
	 */
	 sout << text <<" =[";
					for (int i=0; i<length-1; ++i)
					{
						sout.width(5);
						sout << ar[i] << ", ";
					}
				sout << ar[length-1] << "]" << std::endl;
}



template <typename T>
cimg_library::CImg<T> rectifyUncallibratedImages<T>::sampson(cimg_library::CImg<T> &F,
											  cimg_library::CImg<T> &ml,
											  cimg_library::CImg<T> &mr)

{
	/* SAMPSON compute Sampson error
	 * 	 sampson(F, m1, m2)  evaluate the first order approximation
	 * 	 of the fit of F  with respect to a set of matched points m1, m2.
	 * 	 Returns an approximation of the **squared** distance from the
	 *   4d joint-space point [m1;m2] to the F manifold.
	 * Author: Extracted and adapted from a piece of code by Peter Kovesi
	 * Author: by A. Fusiello
	 * C++ Translation: Julian Coordts
	 */


	cimg_library::CImg<T> m1(ml);
	cimg_library::CImg<T> m2(mr);


    //add rowvector [1] to the the correspondence matrices mX
	sout << "calculating Sampson Error:" << std::endl;
	//sout << "m1:" << std::endl; soutCImgMatrix(m1);
	//sout << "m2:" << std::endl; soutCImgMatrix(m2);


	sout << "\t calculate m2tFm1..." << std::endl;
	int lng = m1.width();
	cimg_library::CImg<T> m2tFm1(lng,1,1,1,0);
	//soutCImgMatrix(m2tFm1,"m2tFm1(empty)");

	cimg_library::CImg<T> p1;
	cimg_library::CImg<T> p2;

	for (int n = 0;n<lng;++n)
	{
		p1 = m2.get_column(n);
		p1.transpose();
		p2 = m1.get_column(n);
		cimg_library::CImg<T> tmp = p1*F*p2;
	    m2tFm1(n,0) = (p1*F*p2).sum();
	}

	//soutCImgMatrix(m2tFm1,"m2tFm1(filled)");

	sout << "\t calculate Fm1 and Ftm2..." << std::endl;
	cimg_library::CImg<T> Fm1(lng,3,1,1,0);
	cimg_library::CImg<T> Ftm2(lng,3,1,1,0);

	Fm1 = F*m1;
	Ftm2 = F.get_transpose()*m2;

	//soutCImgMatrix(Fm1,"Fm1");
	//soutCImgMatrix(Ftm2,"Ftm2");

	// Evaluate squared distances
	// d =  m2tFm1.^2 ./ (Fm1(1,:).^2 + Fm1(2,:).^2 + Ftm2(1,:).^2 + Ftm2(2,:).^2);
	cimg_library::CImg<T> d(lng,1);
	sout << "\t calculate d..." << std::endl;

	for (int i = 0;i<lng;++i)
	{
		d(i,0)= (m2tFm1(i,0)*m2tFm1(i,0)) / ((Fm1(i,0)*Fm1(i,0) )
				+ (Fm1(i,1)*Fm1(i,1))
				+ Ftm2(i,0) *  Ftm2(i,0)
				+ (Ftm2(i,1)*Ftm2(i,1)));
	}
	sout << "\t finished sampson error" << std::endl;
	return d;



}


template <typename T>
rectifyUncallibratedImages<T>::rectifyUncallibratedImages(const std::string& name) :
		TemplatedParameteredObject<T>("rectifyUncallibratedImages", name,
			"Rectifies 2 Images using the point correspondences and the F matrix belonging to this 2 Images")
{
	ParameteredObject::_addParameter (cntr, "cntr", "automatic centering", true);
	ParameteredObject::_addInputSlot(fmatrix, "fmatrix", "the F Matrix", "CImgList<T>"); 
	ParameteredObject::_addInputSlot(feature_correspondences, "feature_correspondences", "the correspondences", "std::vector<std::vector<std::pair<T, T> > >");

	ParameteredObject::_addInputSlot(image_input, "image_input", "the input image sequence (first left, second right)", "CImgList<T>");
	ParameteredObject::_addInputSlot(solver, "solver", "a solver for non-linear, least squared problems", "lsqnonlinsolver<T>*");
	ParameteredObject::_addOutputSlot(warp_matrices, "warp_matrices", "warp_matrices(first left, second right)", "CImgList<T>");
    called=false;

}

template <typename T>
void rectifyUncallibratedImages<T>::execute() {
	ParameteredObject::execute();
	sout << "rectifying uncallibrated images..." << std::endl;


	//Following lines disable multiple executions in order to minimize
	//execution time and avoid unused calculations.
	//Notice: If you need multiple instances of rectifyUncallibratedImages
	//you have to comment them!
	if (called==true)
	{
		sout << "\trectify already called, skip calculating..." << std::endl;
		return;
	}
	else called = true;



// Test Cases:
//	eulR(a)
//	star(a)
//	sampson(F,ml,mr)
//	costRectif(x,600,600,ml,mr)
//	compRectif(ml,mr,600,600,true)
//	compRectif(ml,mr,600,600,false)
//
//	cimg_library::CImg<T> tml(2,2),tmr(2,2),a(3,1),tx(6,1),tF(3,3);
//	tml(0,0)= 1.5; tml(1,0)=2.5; tml(0,1)=3.5; tml(1,1) =4.5;
//	tmr(0,0)= 5.5; tmr(1,0)=6.5; tmr(0,1)=7.5; tmr(1,1) =8.5;
//	tF(0,0)=10.5; tF(1,0)=5.5; tF(2,0)=3.4;
//	tF(0,1)=21; tF(1,1)=15.2; tF(2,1)=25;
//	tF(0,2)=60; tF(1,2)=12; tF(2,2)=0.5;
//
//	cimg_library::CImg<T> resSams(3,3), resStar(3,3),resEul(3,3),rescost(2,1);
//	resEul=eulR(1.5,2.5,3.5);
//	resStar=star(1.5,2.5,3.5);
//	resSams=sampson(tF,tml,tmr);
//
//	soutCImgMatrix(resEul,"eulr");
//	soutCImgMatrix(resStar,"star");
//	soutCImgMatrix(tml,"tml");
//	soutCImgMatrix(tmr,"tmr");
//	soutCImgMatrix(tF,"tF");
//	soutCImgMatrix(resSams,"sampson");
//
//	costData tdata;
//	tdata.w = 600;
//	tdata.h = 600;
//	tdata.ml = &tml;
//	tdata.mr = &tmr;
//	callBackObject = (void*) this;
//
//	double tp[] = { 3, 26.4, 5, 1, 78.3, 5 };
//	double* ttx = new double[2];
//
//	costRectif(tp, ttx, 6, 2, &tdata);
//	soutVectorArray(ttx, 2, "tx:");
//
//	double pt[] = { 1.4, 0.5, 0.6, 0.1, 0.9, 0.002 };
//	cimg_library::CImg<T> tHl(3, 3), tHr(3, 3);
//	calcHlHr(pt, 600, 600, tHl, tHr);
//
//	return;

	/*Extracting f matrix, ImageLeft and ImageRight and correspondences from input slots
	 * for the first pair of images in image_sequence
	 */
		sout << std::endl << "-----RectifiyUncallibratedImages-----" << std::endl;
	    sout << "extract F, IL, IR  and the correspondences from input slots..." << std::endl;
		cimg_library::CImg<T> F = this->fmatrix()(0);
		cimg_library::CImg<T> IL = this->image_input()(0);
		cimg_library::CImg<T> IR = this->image_input()(1);
    /*
     * create separated matrices of correspondences for each image
     *Format: 2xn
     * ml:           cor. to    mr:
     *[x1,x2,...]              [x1',x2',...]
     *[y1,y2,...]              [y1',y2',...]
     *
    */
		sout << feature_correspondences().size() << " feature vectors" << std::endl;
		//Todo check feature_correspondences()->size() is even number


		sout << "create separated matrices of correspondences for each image...." << std::endl;
		int correspondences_count = std::min(this->feature_correspondences()[0].size(),
											 this->feature_correspondences()[1].size());

		sout << correspondences_count << " pairs of correspondences found" << std::endl;

		cimg_library::CImg<T> ml(correspondences_count,3,1,1,0);
		cimg_library::CImg<T> mr(correspondences_count,3,1,1,0);
		sout << "spreading correspondences..." << std::endl;

		for (int i=0; i< correspondences_count;i++)
		{

				ml(i,0)=this->feature_correspondences()[0][i].first;
				ml(i,1)=this->feature_correspondences()[0][i].second;
				mr(i,0)=this->feature_correspondences()[1][i].first;
				mr(i,1)=this->feature_correspondences()[1][i].second;
				ml(i,2)=1;
				mr(i,2)=1;

		}

//		soutCImgMatrix(ml,"Matchpoints left(x,y)");
//		soutCImgMatrix(mr,"Matchpoints right(x,y)");
//return;
		//Output the first and last pairs of correspondences in order to verify them
		//sout << ml(0,0) << std::endl << ml(0,1) << std::endl;
		//sout << ml(ml.width()-1,0) << std::endl << ml(ml.width()-1,1) << std::endl;
		//sout << mr(0,0) << std::endl << mr(0,1) << std::endl;
		//sout << mr(mr.width()-1,0) << std::endl << mr(mr.width()-1,1) << std::endl;

	/*
	 * Calculate and output Sampson RMS
	 */
		sout << "Calculate Sampson RMS..." << std::endl;
		cimg_library::CImg<T> sampsonMatrix;
		sampsonMatrix.assign(sampson(F,ml,mr));
		sout << "Calculate Average Sampson..." << std::endl;
		double res = sqrt(sampsonMatrix.sum()/(ml.width()));
		sout << "Sampson RMS= " << res  << std::endl;

		//TODO check if IR has same dims. as IL

		// Create data strukt we pass through levmar to the cost Funktion
			costData data;
			data.w = IR.width();
			data.h=  IR.height();
			data.ml = &ml;
			data.mr=  &mr;

		//initialize Params for levmar solver
			sout << "initialize Parameters for levmar..." << std::endl;
			double p[6];

			//Todo:
			//
			//double *x = new double[ml.width()];
			//this->solver()->solve(this->costRectif,p,x,6,ml.width(),lb,ub,maxIterations,
			//						opts,info,NULL,NULL,&data);
			//fails. The solver doesn't find the correct solution.
			//Therefore we pass a NULL pointer for x what
			//leads to the correct solution.
			//On the other way we could also pass x with big size and it will
			//also work correct.
			//Task: Find the reason and a better method than the NULL pointer



			double opts[LM_OPTS_SZ]={0,0,0,0,0}, info[LM_INFO_SZ];
			double lb[6], ub[6];
			const int maxIterations = 1000;
			//opts[0]=LM_INIT_MU; opts[1]=1E-15; opts[2]=1E-15; opts[3]=1E-20; opts[4]= LM_DIFF_DELTA;
			//opts[4]=-LM_DIFF_DELTA; // specifies central differencing to approximate Jacobian; more accurate but more expensive to compute!
		    //set start values for p
			p[0]=0; p[1]=0; p[2]=0; p[3]=0; p[4]=0; p[5]=0;
			//at the Test case (data from RectifKitU)
			// the right solution is
			// p[0]=-0.1884f ; p[1]=0.0299f; p[2]= -0.0039f; p[3]=-0.2295f; p[4]=0.0114f; p[5]= -0.4651f;

			//soutVectorArray(p,6,"start values p");
			//set box constraints
			const double PI=3.1415926535897932384626433832795f;
			lb[0]= lb[1] = lb[2] = lb [3] = lb[4] = -PI/2;
			lb[5]= -1.5f;
			ub[0]= ub[1] = ub[2] = ub [3] = ub[4] = PI/2;
			ub[5]= 1.5f;
			//we need the static callBackObject because we had to make the
			//costRectif method static in order to pass it to the levmar solver
			//With this callBackObject we can use the non-static methods
			//in the costRectif method.
			callBackObject = (void*)this;
			sout << "run levmar..." << std::endl << std::endl;
			int dlevRes  = this->solver()->solve(this->costRectif,p,NULL,6,ml.width(),lb,ub,0,
						opts,info,NULL,NULL,&data);
			sout << "levmar returned " << dlevRes << std::endl;
			soutVectorArray(p,6,"Solution vector: p");

			//check result
			double* x = new double[ml.width()];
			double sumError=0.0f;
			costRectif(p,x,0,0,&data);

			for (int i = 0; i < ml.width(); i++)
			{
				sumError+=x[i]*x[i];
			}
			sumError = sqrt(sumError);
			delete x;

			sout << "Error after rectification: " << sumError << std::endl;
			if (sumError > 20.0f) throw "rectification error: can't minimize give images, result error to high!";



			//soutVectorArray(x,ml.width(),"   m  ");





			int counter=0;
			while(p[5]>1 || p[5]<-1)
			{
				counter++;
				sout<< "focal out of range, trying a random restart(p[5] = " <<
						p[5] << ") try nr. " << counter << std::endl;


				//set new start values for p
					p[0]=0; p[1]=0; p[2]=0; p[3]=0; p[4]=0; p[5]=0.02*(rand()%101)-1.0;
				//	soutVectorArray(p,6,"new start values: p");
					callBackObject = (void*)this;
					sout << "run levmar..."  << std::endl  << std::endl;
					dlevRes  = this->solver()->solve(this->costRectif,p,NULL,6,ml.width(),lb,ub,maxIterations,
								opts,info,NULL,NULL,&data);
					sout << "levmar returned " << dlevRes << std::endl;
								soutVectorArray(p,6,"Solution vector: p");

					if (counter>20)
					{
						sout << "Cannot find a focal in the range, Cancelling" << std::endl;
						return;
					}
			}

			cimg_library::CImg<T> Hl(3,3),Hr(3,3);
			calcHlHr(p,IR.width(),IR.height(),Hl, Hr);
			 this->warp_matrices().assign(2,3,3);
			 this->warp_matrices()[0] = Hl;
			 this->warp_matrices()[1] = Hr;

/* The following lines could be used if you want to warp the image using a flow field instead of the H matrices.
 * The code is not correct!
			// calculate Flow field left image


            double translationsum=0;
			 for (int x=0; x<IL.width();x++)
				 {
					 for (int y=0; y<IL.height();y++)
					 {
						 cimg_library::CImg<T> oldPoint(1,3);
						 oldPoint (0,0)= x; oldPoint (0,1)= y; oldPoint (0,2)= 1;
						 cimg_library::CImg<T> newPoint(1,3);
						 newPoint= Hl* oldPoint;
						 newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
						 newPoint(0,1)= newPoint(0,1)/newPoint(0,2);
						 cimg_library::CImg<T> diff(1,3);
						 // translation for all 3 channels of x
							 flow_sequence_left()[0](x,y,0,0) = oldPoint(0,0)-newPoint(0,0);
							 flow_sequence_left()[0](x,y,0,1) = oldPoint(0,0)-newPoint(0,0);
							 flow_sequence_left()[0](x,y,0,2) = oldPoint(0,0)-newPoint(0,0);
						 // translation for all 3 channels of y
							 flow_sequence_left()[1](x,y,0,0) = oldPoint(0,1)-newPoint(0,1);
							 flow_sequence_left()[1](x,y,0,1) = oldPoint(0,1)-newPoint(0,1);
							 flow_sequence_left()[1](x,y,0,2) = oldPoint(0,1)-newPoint(0,1);
							 translationsum= translationsum + oldPoint(0,0)-newPoint(0,0) +
											 oldPoint(0,1)-newPoint(0,1);
						 }
				 }
				 sout << "sum of translation distances left: " << translationsum << std::endl;
			 // calculate Flow field right image

			             sout<<"calculate Flow field right image" << std::endl;
			             flow_sequence_right().assign(2);
			             flow_sequence_right()[0].assign(IR.width(),
			 										     IR.height(),1,3);
			             flow_sequence_right()[1].assign(IR.width(),
			 										     IR.height(),1,3);

			             translationsum=0;
			 			 for (int x=0; x<IR.width();x++)
			 				 {
			 					 for (int y=0; y<IR.height();y++)
			 					 {
			 						 cimg_library::CImg<T> oldPoint(1,3);
			 						 oldPoint (0,0)= x; oldPoint (0,1)= y; oldPoint (0,2)= 1;
			 						 cimg_library::CImg<T> newPoint(1,3);
			 						 newPoint= Hr* oldPoint;
			 						 newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
									 newPoint(0,1)= newPoint(0,1)/newPoint(0,2);
			 						 cimg_library::CImg<T> diff(1,3);
			 						 // translation for all 3 channels of x
			 							 flow_sequence_right()[0](x,y,0,0) =  oldPoint(0,0)-newPoint(0,0);
			 							 flow_sequence_right()[0](x,y,0,1) =  oldPoint(0,0)-newPoint(0,0);
			 							 flow_sequence_right()[0](x,y,0,2) =  oldPoint(0,0)-newPoint(0,0);
			 						 // translation for all 3 channels of y
			 							 flow_sequence_right()[1](x,y,0,0) =  oldPoint(0,1)-newPoint(0,1);
			 							 flow_sequence_right()[1](x,y,0,1) =  oldPoint(0,1)-newPoint(0,1);
			 							 flow_sequence_right()[1](x,y,0,2) =  oldPoint(0,1)-newPoint(0,1);
										 translationsum= translationsum + oldPoint(0,0)-newPoint(0,0) +
														 oldPoint(0,1)-newPoint(0,1);
			 						  }
			 				 }
						 sout << "sum of translation distances right: " << translationsum << std::endl;
*/
}




#endif /* _rectifyUNCALLIBRATEDIMAGES_HXX_ */



