/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file SolvePnPRansac.h
 *  Declaration of the parameter class SolvePnPRansac.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.09.2012
 */

#ifndef _SOLVEPNPRANSAC_H_
#define _SOLVEPNPRANSAC_H_

#ifdef _MSC_VER
#ifdef solvepnpransac_EXPORTS
/// Visual C++ specific code
#define solvepnpransac_DECLDIR __declspec(dllexport)
#else
#define solvepnpransac_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else
/// Not needed without MSVC
#define solvepnpransac_DECLDIR
#endif

#include <charon-core/ParameteredObject.h>
#include <charon-utils/CImg.h>
#include "hekate/Create3DPointCloud_plugins.h"	
#include <hekate/IndexedMatrix.h>
#include <hekate/Correspondence.hxx>

using namespace cimg_library;
using namespace hekate;

/// Find camera positions for a given set of 2D-3D correspondences
/** Finds external camera parameters for a given set of 2D-3D correspondences
 *  and an internal calibration.
 */
template <typename T>
class solvepnpransac_DECLDIR SolvePnPRansac :
		public TemplatedParameteredObject<T> {
public:
	/// default constructor
	/// \param name          Instance name
	SolvePnPRansac(const std::string& name = "");

	/// 3D point cloud
	InputSlot< CorrespondenceList<T> > points3D;
	/// 2D point cloud
	InputSlot< CorrespondenceList<T> > points2D;
	/// List of calibrations
	InputSlot< cimg_library::CImgList<T> > calibrations;
	/// List of calibrations
	InputSlot< IndexedMatrixList > cameraMatricesDebug;

	/// Camera matrices
	OutputSlot< IndexedMatrixList > cameraMatrices;

	/// Number of iterations.
	Parameter< int > nOIt;
	/// Inlier threshold value used by the RANSAC procedure. The parameter
	/// value is the maximum allowed distance between the observed and
	/// computed point projections to consider it an inlier.
	Parameter< float > reprojError;
	/// Percental Number of inliers. If the algorithm at some stage finds more inliers
	/// than minInliersCount , it finishes.
	Parameter< float > minInlierCount;
	/// 0=CV_ITERATIVE: Iterative method is based on Levenberg-Marquardt
	/// optimization. In this case the function finds such a pose that
	/// minimizes reprojection error, that is the sum of squared distances
	/// between the observed projections imagePoints and the projected (using
	/// projectPoints() ) objectPoints .\n 1=CV_P3P Method is based on the paper
	/// of X.S. Gao, X.-R. Hou, J. Tang, H.-F. Chang Complete Solution
	/// Classification for the Perspective-Three-Point Problem. In this case
	/// the function requires exactly four object and image points.
	/// \n2=CV_EPNP Method has been introduced by F.Moreno-Noguer, V.Lepetit and
	/// P.Fua in the paper EPnP: Efficient Perspective-n-Point Camera Pose
	/// Estimation.
	Parameter< int > flags;
	/// Use SVD method
	Parameter< bool > useSVD;
	/// Remove RANSAC outliers from 3d point cloud
	Parameter< bool > removeOutliers;	

	/// Update object.
	virtual void execute();
};

#endif // _SOLVEPNPRANSAC_H_
