/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file SolvePnPRansac.hxx
 *  Implementation of the parameter class SolvePnPRansac.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.09.2012
 */

#ifndef _SOLVEPNPRANSAC_HXX_
#define _SOLVEPNPRANSAC_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/SolvePnPRansac.h>

template <typename T>
SolvePnPRansac<T>::SolvePnPRansac(const std::string& name) :
		TemplatedParameteredObject<T>(
			"SolvePnPRansac", name,
			"<h2>Find camera positions for a given set of 2D-3D "
			"correspondences</h2><br>"
			"Finds external camera parameters for a given set of 2D-3D "
			"correspondences and an internal calibration."
		),cameraMatricesDebug(true, false)
{
	ParameteredObject::_addInputSlot(
		points3D, "points3D",
		"3D point cloud",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		points2D, "points2D",
		"2D point cloud",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		calibrations, "calibrations",
		"List of calibrations",
		"CImgList<T>");
	ParameteredObject::_addInputSlot(
		cameraMatricesDebug, "cameraMatricesDebug",
		"cameraMatrices for debugging",
		"IndexedMatrixList");

	ParameteredObject::_addOutputSlot(
		cameraMatrices, "cameraMatrices",
		"Camera matrices",
		"IndexedMatrixList");

	ParameteredObject::_addParameter< int >(
		nOIt, "nOIt",
		"Number of iterations.",
		1000, "int");
	ParameteredObject::_addParameter< float >(
		reprojError, "reprojError",
		"Inlier threshold value used by the RANSAC procedure. The parameter"
		"value is the maximum allowed distance between the observed and"
		"computed point projections to consider it an inlier.",
		4, "float");
	ParameteredObject::_addParameter< float >(
		minInlierCount, "minInlierCount",
		"Number of inliers. If the algorithm at some stage finds more inliers"
		"than minInliersCount , it finishes.",
		0.7, "float");
	ParameteredObject::_addParameter< int >(
		flags, "flags",
		"0=CV_ITERATIVE: Iterative method is based on Levenberg-Marquardt"
		"optimization. In this case the function finds such a pose that"
		"minimizes reprojection error, that is the sum of squared distances"
		"between the observed projections imagePoints and the projected"
		"(using projectPoints() ) objectPoints . <br>1=CV_P3P Method is based on"
		"the paper of X.S. Gao, X.-R. Hou, J. Tang, H.-F. Chang Complete"
		"Solution Classification for the Perspective-Three-Point Problem. In"
		"this case the function requires exactly four object and image"
		"points.<br>"
		"2=CV_EPNP Method has been introduced by F.Moreno-Noguer, V.Lepetit"
		"and P.Fua in the paper EPnP: Efficient Perspective-n-Point Camera"
		"Pose Estimation.",
		0, "int");	
	ParameteredObject::_addParameter< bool >(
		useSVD, "useSVD",
		"Use SVD method",
		false, "bool");
	ParameteredObject::_addParameter< bool >(
		removeOutliers, "removeOutliers",
		"Remove RANSAC outliers from 3d point cloud",
		false, "bool");

}

template <typename T>
void SolvePnPRansac<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	sout << "Estimating " << this->points2D().sizeImages() << " camera poses" << std::endl;

	hekate::CorrespondenceList<T> points3D = this->points3D();

	IndexedMatrixList cams;
	if(this->cameraMatricesDebug.connected())
		cams = this->cameraMatricesDebug();

	// Use OpenCV method
	if( !useSVD() ){

		for(size_t i = 0; i < this->points2D().sizeImages(); i++){
		
			try{
				IndexedMatrix P = recoverCamRANSAC(i, this->calibrations(), this->points2D(),
								points3D, nOIt(), reprojError(), minInlierCount(), flags(), 
								cams, removeOutliers() );
				P.setIndex(i);
				this->cameraMatrices().push_back( P );
			}
			catch(...){
				sout << "\tPose estimation by RANSAC failed!" << std::endl;
			}

		}

	}
	// Use SVD method without RANSAC
	else{
	
		for(size_t i = 0; i < this->points2D().sizeImages(); i++){
		
			try{
				IndexedMatrix P = recoverCam(i, this->calibrations(), this->points2D(),
								this->points3D() );
				P.setIndex(i);
				this->cameraMatrices().push_back( P );
			}
			catch(...){
				sout << "\tPose estimation by SVD failed!" << std::endl;
			}

		}
	
	}

}

#endif /* _SOLVEPNPRANSAC_HXX_ */
