/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file SortOutPoints.hxx
 *  Implementation of the parameter class SortOutPoints.
 *  \author Moritz Becker
 *  \date 17.01.2011
 */

#ifndef _SORTOUTPOINTS_HXX_
#define _SORTOUTPOINTS_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/SortOutPoints.h>

template <typename T>
SortOutPoints<T>::SortOutPoints(const std::string& name) :
		TemplatedParameteredObject<T>(
				"SortOutPoints", name,
				"Removes all points with point index i, where at least one point with "
				"this index is an outlier."
		)
			,pointList1(false, false) 
			,pointList2(true, false) 			
			
{	

	ParameteredObject::_addParameter (rejectManualOutliers, "rejectManualOutliers", 
	"Only reject manual outliers given by index list.", false);
	ParameteredObject::_addParameter (manualOutliers, "manualOutliers", 
	"Index list of manual outliers.");
	
	ParameteredObject::_addInputSlot(pointList1, "pointList1", 
	"First point list.", 
	"CorrespondenceList"); 
	ParameteredObject::_addInputSlot(pointList2, "pointList2", 
	"Second point list.", 
	"CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(pointList1out, "pointList1out", 
	"First cleaned point list.", 
	"CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(pointList2out, "pointList2out", 
	"Second cleaned point list.", 
	"CorrespondenceList"); 
	
	ParameteredObject::_setTags("Hekate") ;
	
}

template <typename T>
void SortOutPoints<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	auto it = pointList1().begin() ;
	for(; it != pointList1().end(); it++)
	{
		if(!it->outlier)
			this->pointList1out().push_back(*it);
	}

	/*
	size_t nPoints1=0, nPoints2=0, nImg1=0, nImg2=0;

	if(!pointList1.connected())
		throw(std::runtime_error("SORTOUTPOINTS: No correspondences connected."));
	
	nPoints1 = pointList1().sizePoints();
	nImg1 = pointList1().sizeImages();
	sout << "CorrespondenceList 1    : " << nImg1 << " different image indices" <<
		" with " << nPoints1 << " points per imag index." << std::endl;

	if(this->pointList2.connected()){
		nPoints2 = pointList2().sizePoints();
		nImg2 = pointList2().sizeImages();	
		sout << "CorrespondenceList 2    : " << nImg2 << " different image indices" <<
			" with " << nPoints2 << " points per imag index." << std::endl;
	}


	int j=0;
	for(size_t i=0; i<nPoints1; i++){

		// check 2d corrspondences1
		if(!this->rejectManualOutliers() &&
			!hasOutliers(this->pointList1(),i)
			){}

		// check 2d corrspondences2
		else if(!this->rejectManualOutliers() &&
			this->pointList2.connected() &&
			!hasOutliers(this->pointList2(),i)
			){}
		

		//push back if no outliers
		else{
					
				
			bool foundManual = false;

			if(this->rejectManualOutliers()){		
				// Check, if current correspondence is marked as manual outlier
				for(size_t l=0; l<this->manualOutliers().size(); l++){
					if(this->manualOutliers()[l] == i){
						foundManual = true;
						sout << "Reject manual selected outlier (index " << i << ")" << std::endl;
						break;
					}
				}
			}

			if(!foundManual){
				j++;
				if(this->pointList1out.connected()){
					forwardCorrespondence(this->pointList1out(),this->pointList1(),i);
				}
				else
					throw std::runtime_error("Output slot not connected! Cannot save inliers.");
				if(this->pointList2.connected() && this->pointList2out.connected()){
					forwardCorrespondence(this->pointList2out(),this->pointList2(),i);
				}
			}
		}

	}

	sout << std::endl;
	sout << "------------------------------------------" << std::endl;
	
	sout << nPoints1-j << " outliers of " << nPoints1
		<< " points have been removed from "
		" list 1." << std::endl;

	if(this->pointList2.connected()){
		sout << nPoints2-j << " outliers of " << nPoints2
			<< " points have been removed from "
			" list 2." << std::endl;
	}
	
	sout << this->pointList1out().size() << " points remaining." << std::endl;
	sout << this->pointList1out().sizeImages() << " images remaining." << std::endl;
	sout << this->pointList1out().sizePoints() << " correspondences remaining." << std::endl;
	sout << "------------------------------------------" << std::endl;
	sout << std::endl;
	*/

}

// Check if any featurelist of the correspondence list has an outlier at position "position"
template <typename T>
bool hasOutliers(const CorrespondenceList<T>& corr,
		const size_t position) {

	auto it = corr.get<pointIndex>().lower_bound(position) ;
	for(; it != corr.get<pointIndex>().upper_bound(position) ; it++)
	{
		if(it->outlier)
		{
			return 1;
		}
	}
	
	return 0;

}

template <typename T>
void forwardCorrespondence(CorrespondenceList<T>& corrOut,
		const CorrespondenceList<T>& corr, const size_t position){

	auto it = corr.get<pointIndex>().lower_bound(position) ;
	for(; it != corr.get<pointIndex>().upper_bound(position) ; it++)
	{
		corrOut.push_back(*it);
		sout << "forwarding inlier" << std::endl;
	}

}
					

#endif /* _SORTOUTPOINTS_HXX_ */



