/*  Copyright (C) 2009 Julian Coordts

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file testdriver.h
 *  Declaration of the parameter class testdriver.
 *  \author Julian Coordts
 *  \date 12.03.2010
 */

#ifndef _TESTDRIVER_H_
#define _TESTDRIVER_H_

#if defined(MSVC) && defined(HANDLE_DLL)
#ifdef testdriver_EXPORTS
/// Visual C++ specific code
#define testdriver_DECLDIR __declspec(dllexport)
#else
#define testdriver_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else /* No DLL handling or GCC */
/// Not needed with GCC
#define testdriver_DECLDIR
#endif

#include <charon-core/ParameteredObject.hxx>
#include <charon-utils/CImg.h>
#include <hekate/Correspondence.hxx>
#include <hekate/IndexedMatrix.h>

using namespace hekate;

/// Provides test data in order to support quick testing of new Plugins
/** The module generates a 3D point cloud and two camera matrices. The internal and external parameters 
	of the camera matrices can be modified. While the external parameters (position and orientation) 
	has to be set as parameter list P1 and P2, the internal parameters (focal length, aspect ratio) 
	can be decleared separately with the parameters 'focalLength' and 'aspectRatio'. <br>
	The 3D points X and the defined camera matrices then are used to generate 2D correspondences x1 and x2 
	by backprojection with x1=P1*X and x2=P2*X. To simulate a second time step where the camera pair P1 and P2 
	has moved, the position and orientation of the camera pair can be changed with a translation matrix and 
	three rotation angles. Then a second set of 2D correspondences from the same 3D scene 
	is generated by backprojection again. The 2D correspondences can be modified by adding a Gaussian noise. <br>
	For test purposes, also a 3x3 matrix can be generated by parameter list.

*/
template <typename T>
class testdriver_DECLDIR testdriver :
		public TemplatedParameteredObject<T> {
public:
	
	/// Removes correspondences which wouldn't lie on the image plane when they are backprojected by x=PX.
	Parameter < bool > removePointsNotOnImagePlane;
	/// Width of the image plane.
	Parameter < int > imageWidth;
	/// Height of the image plane.
	Parameter < int > imageHeight;
	/// Number of simulated points.
	Parameter < int > numberOfPoints;
	/// Number of simulated views.
	Parameter < int > numberOfViews;
	/// Angle between subsequent views.
	Parameter < double > angleBetweenViews;
	/// Flag if generated points should have a gaussian distribution.
	Parameter < bool > gaussianNoise;
	/// First camera matrix P1 to generate 2D correspondences by backprojection of 3D points.
	ParameterList < float > pmatrix1;
	/// Second camera matrix P2 to generate 2D correspondences by backprojection of 3D points.
	ParameterList < float > pmatrix2;
	/// This translation matrix moves the cameras P1 and P2 to simulate a second pair of cameras P1' and P2'.
	ParameterList < T > translation;
	/// Rotates the cameras P1 and P2 around the Y-Z axis to simulate a second pair of cameras P1' and P2'.
	Parameter < T > xRot;
	/// Rotates the cameras P1 and P2 around the X-Z axis to simulate a second pair of cameras P1' and P2'.
	Parameter < T > yRot;
	/// Rotates the cameras P1 and P2 around the X-Y axis to simulate a second pair of cameras P1' and P2'.
	Parameter < T > zRot;
	/// Focal length for both cameras.
	ParameterList < T > calibration;
	/// Value of gaussian noise (in pixels) which can be added to the generated correspondences.
	Parameter < T > correspondenceNoise;
	/// Angle between subsequent views.
	Parameter < int > outliers;

	/// Camera matrices P1 and P2
	OutputSlot < IndexedMatrixList > pmatrix;
	/// Camera matrices P1' and P2'
	OutputSlot < IndexedMatrixList > pmatrixSecond;
	/// 2D correspondences x1 and x2
	OutputSlot < typename CorrespondenceList<T> > correspondences;
	/// 2D correspondences x1 and x2
	OutputSlot < typename CorrespondenceList<T> > correspondencesSecond;
	/// Projection images of P1 and P2.
	OutputSlot < cimg_library::CImgList<T> > imageSequence;
	/// 3D point cloud.
	OutputSlot < typename CorrespondenceList<T> > points3Doutput;
	/// 3D point cloud saved as CImgList.
	OutputSlot < cimg_library::CImgList<T> > points3DCImgList;
	/// Calibration matrix which is used to generate the 3D points.
	OutputSlot < cimg_library::CImgList<T> > calibrationOut;
	/// Vector of ints
	OutputSlot < std::vector< int > > vecOfInt;
	/// Vector of vector of pair of ints
	OutputSlot < std::vector< std::vector< std::vector< std::pair<int,int> > > > > vecOfVecOfPair;
	

	/// create a new testdriver object
	/// \param name          Instance name
	testdriver(const std::string& name);

	/// Update object.
	virtual void execute();
};



#endif // _TESTDRIVER_H_



