/*  Copyright (C) 2009 Julian Coordts

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file testdriver.hxx
 *  Implementation of the parameter class testdriver.
 *  \author Julian Coordts
 *  \date 12.03.2010
 */

#define PI	3.14159265358979323846264338327950288419716939937510d

#ifndef _TESTDRIVER_HXX_
#define _TESTDRIVER_HXX_

#include <hekate/Testdriver.h>
#include <charon-utils/ImgTool.hxx>

template <typename T>
testdriver<T>::testdriver(const std::string& name) :
		TemplatedParameteredObject<T>("testDriver", name,
			"Provides test data in order to support quick testing of new Plugins<br>"
			"The module generates a 3D point cloud and two camera matrices. The internal and external parameters "
			"of the camera matrices can be modified. While the external parameters (position and orientation) "
			"has to be set as parameter list P1 and P2, the internal parameters (focal length, aspect ratio) "
			"can be decleared separately with the parameters 'focalLength' and 'aspectRatio'. "
			"The 3D points X and the defined camera matrices then are used to generate 2D correspondences x1 and x2 "
			"by backprojection with x1=P1*X and x2=P2*X. To simulate a second time step where the camera pair P1 and P2 "
			"has moved, the position and orientation of the camera pair can be changed with a translation matrix and "
			"three rotation angles. Then a second set of 2D correspondences from the same 3D scene "
			"is generated by backprojection again. The 2D correspondences can be modified by adding a Gaussian noise. "
			"For test purposes, also a 3x3 matrix can be generated by parameter list."
			)
{
	
	ParameteredObject::_addParameter (imageWidth, "imageWidth", 
		"Width of the image plane.", 1000);
	ParameteredObject::_addParameter (imageHeight, "imageHeight", 
		"Height of the image plane.", 1000);
	ParameteredObject::_addParameter (removePointsNotOnImagePlane, "removePointsNotOnImagePlane", 
		"Removes correspondences which wouldn't lie on the defined image plane "
		"when they are backprojected by x=PX.", true );
	ParameteredObject::_addParameter (numberOfPoints, "numberOfPoints", 
		"Number of simulated points.", 500);
	ParameteredObject::_addParameter (numberOfViews, "numberOfViews", 
		"Number of simulated views.", 8);
	ParameteredObject::_addParameter (angleBetweenViews, "angleBetweenViews", 
		"Angle between subsequent views.", 20.0);
	ParameteredObject::_addParameter (gaussianNoise, "GaussianNoisePoints", 
		"Flag if generated points should have a gaussian distribution.", true, "bool" );
	ParameteredObject::_addParameter (pmatrix1, "P1", 
		"First camera matrix P1 to generate 2D correspondences by backprojection of 3D points." );
	ParameteredObject::_addParameter (pmatrix2, "P2", 
		"Second camera matrix P2 to generate 2D correspondences by backprojection of 3D points." );
	ParameteredObject::_addParameter (translation, "translation", 
		"This translation matrix moves the cameras P1 and P2 to simulate a second pair "
		"of cameras P1' and P2'." );
	ParameteredObject::_addParameter (xRot, "xRot", 
		"Rotates the cameras P1 and P2 around the Y-Z axis to simulate a second pair "
		"of cameras P1' and P2'.", T(0) );
	ParameteredObject::_addParameter (yRot, "yRot", 
		"Rotates the cameras P1 and P2 around the X-Z axis to simulate a second pair "
		"of cameras P1' and P2'.", T(5) );
	ParameteredObject::_addParameter (zRot, "zRot", 
		"Rotates the cameras P1 and P2 around the X-Y axis to simulate a second pair "
		"of cameras P1' and P2'.", T(0) );
	ParameteredObject::_addParameter (calibration, "calibration", 
		"Focal length for both cameras.");
	ParameteredObject::_addParameter (correspondenceNoise, "Noise2D", 
		"Value of gaussian noise (in pixels) which can be added to the generated correspondences.", T(0));
	ParameteredObject::_addParameter (outliers, "outliers", 
		"Number of outliers.", 0);
	 
	ParameteredObject::_addOutputSlot(pmatrix, "cameraMatrix", 
		"Camera matrices P1 and P2", "IndexedMatrixList"); 
	ParameteredObject::_addOutputSlot(pmatrixSecond, "cameraMatrix2", 
		"Camera matrices P1' and P2'", "IndexedMatrixList"); 
	ParameteredObject::_addOutputSlot(correspondences, "correspondences", 
		"2D correspondences x1 and x2", "CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(correspondencesSecond, "correspondencesSecond", 
		"2D correspondences x1 and x2", "CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(points3Doutput, "points3D", "3D point cloud.", "CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(imageSequence, "imageSequence", 
		"Projection images of P1 and P2.", "CImgList<T>"); 
	ParameteredObject::_addOutputSlot(calibrationOut, "calibrationMatrix", 
		"Calibration matrix which was used to generate the points", "CImgList<T>"); 
	ParameteredObject::_addOutputSlot(vecOfInt, "vecOfInt", 
		"Vector of ints which can be used e.g. for a vector of ids", "vector<int>"); 
	ParameteredObject::_addOutputSlot(vecOfVecOfPair, "vecOfVecOfPair", 
		"Vector of ints which can be used e.g. for a vector of ids", "vector<vector<vector<pair>>>"); 


}

template <typename T>
void testdriver<T>::execute() {

	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	CorrespondenceList<T> &points3D = this->points3Doutput();
	int nPoints = this->numberOfPoints();
	
	// Create gaussian points
	if(this->gaussianNoise){
		sout << "Generating gaussian distributed 3D points." << std::endl;
		const double sigma = 2;
		createGaussianPoints(points3D, nPoints, sigma);
	}

	// Create point cube
	else{
		sout << "Generating point cube." << std::endl;
		createPointCube(points3D, nPoints);
	}


	// Create rotation matrix ----------------------------------------
	cimg_library::CImg<T> R1(4,4), R2(4,4), R3(4,4);
	R1.fill(T(0)); R2.fill(T(0)); R3.fill(T(0));

	T rotX = this->xRot;
	T rotY = this->yRot;
	T rotZ = this->zRot;
	sout << "Rotating angles between frames: Alpha(x)=" << rotX
		<< ", Beta(y)=" << rotY << ", Gamma(z)=" << rotZ << std::endl;

	const double PI180 = PI/180.;

	//define x rotation matrix
	R1(0,0)=T(1);
	R1(1,1)=T(cos(rotX*PI180));	R1(2,1)=T(-sin(rotX*PI180)); 
	R1(1,2)=T(sin(rotX*PI180));		R1(2,2)=T(cos(rotX*PI180));
	R1(3,3)=T(1);
	//define y rotation matrix
	R2(0,0)=T(cos(rotY*PI180));	R2(2,0)=T(sin(rotY*PI180)); 
	R2(1,1)=T(1);
	R2(0,2)=T(-sin(rotY*PI180));		R2(2,2)=T(cos(rotY*PI180));
	R2(3,3)=T(1);
	//define z rotation matrix
	R3(0,0)=T(cos(rotZ*PI180));	R3(1,0)=T(-sin(rotZ*PI180)); 
	R3(0,1)=T(sin(rotZ*PI180));		R3(1,1)=T(cos(rotZ*PI180));
	R3(2,2)=T(1);
	R3(3,3)=T(1);
																	

	// Create camera matrices --------------------------------------------------
	cimg_library::CImg<T> leftP(4,3), rightP(4,3), Trans(1,3);
	leftP.fill(T(0)),	
	rightP.fill(T(0));
	Trans.fill(T(0));

	// translation vector 
	if(this->translation().size()==3){
		for(int i=0; i<3; i++)	Trans(0,i)=this->translation()[i];
		sout << "Translation between cameras: " <<
			this->translation()[0] << ", " <<
			this->translation()[1] << ", " <<
			this->translation()[2] << std::endl;
	}
	else{
		 Trans(0,0)=T(0); Trans(0,1)=T(3); Trans(0,2)=T(0);
	}

	// Create left camera
	if(this->pmatrix1().size()==12)
		for(int i=0; i<3; i++)
			for(int j=0; j<4; j++)
				leftP(j,i)=T(this->pmatrix1()[j+4*i]);
	else{
		 leftP(0,0)=T(1); leftP(1,1)=T(1); leftP(2,2)=T(1); leftP(3,2)=T(15);
	}

	// Create right camera
	if(this->pmatrix2().size()==12)
		for(int i=0; i<3; i++)
			for(int j=0; j<4; j++)
				rightP(j,i)=T(this->pmatrix2()[j+4*i]);

	else{
		sout << "test driver: No correct P2 input. Taking default matrix" << std::endl;
		rightP = leftP;
		
		rightP(3,0) += Trans[0];
		rightP(3,1) += Trans[1];
		rightP(3,2) += Trans[2];

		rightP = rightP*R1*R2*R3;

	}	

	// Create calibration matrix --------------------------------------
	cimg_library::CImg<T> calibrationMatrix(3,3);
	if(this->calibration().size() != 9){
		calibrationMatrix.fill(T(0));
		calibrationMatrix(0,0) = T(1000);
		calibrationMatrix(2,0) = T(500);
		calibrationMatrix(1,1) = T(1000);
		calibrationMatrix(2,1) = T(500);
		calibrationMatrix(2,2) = T(1);
	}
	// add calibration by focal length if set
	if(this->calibration().size() == 9){
		calibrationMatrix.fill(T(0));
		calibrationMatrix(0,0)=this->calibration()[0];
		calibrationMatrix(1,0)=this->calibration()[1];
		calibrationMatrix(2,0)=this->calibration()[2];
		calibrationMatrix(0,1)=this->calibration()[3];
		calibrationMatrix(1,1)=this->calibration()[4];
		calibrationMatrix(2,1)=this->calibration()[5];
		calibrationMatrix(0,2)=this->calibration()[6];
		calibrationMatrix(1,2)=this->calibration()[7];
		calibrationMatrix(2,2)=this->calibration()[8];
	}


	// Add calibration to camera matrices ------------------------------
	// and create second set of cameras

	const int numberOfViews = this->numberOfViews();
	const double angleSecondCam = this->angleBetweenViews(); // rotation between cameras in degree
	
	// Define rotation matrix to rotate cameras and points
	// for further correspondences and views
	cimg_library::CImg<T> rotSecondCam(4,4);
	rotSecondCam.fill(T(0));
	rotSecondCam(0,0)=T(cos(angleSecondCam*PI180));	
	rotSecondCam(2,0)=T(sin(angleSecondCam*PI180)); 
	rotSecondCam(1,1)=T(1);
	rotSecondCam(0,2)=T(-sin(angleSecondCam*PI180));		
	rotSecondCam(2,2)=T(cos(angleSecondCam*PI180));
	rotSecondCam(3,3)=T(1);


	// Apply calibration to cameras
	sout << "Adding calibration to camera matrices:" << std::endl;
	ImgTool::printInfo(sout,calibrationMatrix,"K");
	
	//ImgTool::printInfo(sout,leftP,"P1------------");
	//ImgTool::printInfo(sout,rightP,"P2------------");
	//ImgTool::printInfo(sout,R3*R2*R1,"R3*R2*R1------------");
	//ImgTool::printInfo(sout,Trans,"T------------");


	// Create correspondences -----------------------------------------

	cimg_library::CImg<T> cam1, cam2;
	cam1 = leftP;
	cam2 = rightP;

	// Iterate generation for number of views
	for(int i=0; i<numberOfViews; i++){

		// Rotate cameras
		if(i>0){
			cam1 = cam1 * rotSecondCam;
			cam2 = cam2 * rotSecondCam;
		}

		// Multiply with calibration matrix
		cimg_library::CImg<T> cam1K = calibrationMatrix * cam1;
		cimg_library::CImg<T> cam2K = calibrationMatrix * cam2;

		
		ImgTool::printInfo(sout,cam1K.get_pseudoinvert()*cam1K*cam1K.get_pseudoinvert(),"BAB");
		ImgTool::printInfo(sout,cam1K*cam1K.get_pseudoinvert()*cam1K,"ABA");
		ImgTool::printInfo(sout,cam1K.get_pseudoinvert(),"B");
		
		char buffer[10];
		sprintf(buffer,"P%d",i);
		ImgTool::printInfo(sout,cam1K,buffer);

		// Save camera matrices
		IndexedMatrix cam1_ = cam1K;
		cam1_.setIndex(i);
		IndexedMatrix cam2_ = cam2K;
		cam2_.setIndex(i);
		this->pmatrix().push_back(cam1_);
		this->pmatrixSecond().push_back(cam2_);

		// Save calibration matrix
		this->calibrationOut().push_back(calibrationMatrix);
		
		// Iterate for number of points
		int outliers1 = 0;
		int outliers2 = 0;
		int nOutliers = this->outliers();
		for(int j=0; j<points3D.size(); j++){

			// Back project 3D point to image plane
			cimg_library::CImg<T> x1(1,3), x2(1,3);
			x1 = cam1K * points3D(0,j).pt3d() ;
			x1 /= x1[2];
			x2 = cam2K * points3D(0,j).pt3d() ;
			x2 /= x2[2];
			
			// Add noise
			if( this->correspondenceNoise() > 0.00001 ){
				x1.noise(this->correspondenceNoise());
				x1[2] = T(1);
				x2.noise(this->correspondenceNoise());
				x2[2] = T(1);
			}

			// Make first "nOutliers" points an outlier
			if(	nOutliers > 0 && j < nOutliers	){
				x2.noise(double(this->imageWidth())/4.);
				x2[2] = T(1);
			}
			
			// check whether point does lie on image planes
			bool outOfPlane1=false;
			bool outOfPlane2=false;

			if(x1[0]<T(0) || x1[1]<T(0) || x1[0]>T(this->imageWidth()) || x1[1]>T(this->imageHeight())){
				outOfPlane1=true;
				outliers1++;
			}
			if(x2[0]<T(0) || x2[1]<T(0) || x2[0]>T(this->imageWidth()) || x2[1]>T(this->imageHeight())){
				outOfPlane2=true;
				outliers2++;
			}

			// push back correspondence if on plane or removePointsNotOnImagePlane is false
			if(!removePointsNotOnImagePlane() || !outOfPlane1){
				this->correspondences().push_back(Correspondence<T>(x1[0],x1[1],0,x1[2],j,i));
			}
			if(!removePointsNotOnImagePlane() || !outOfPlane2){
				this->correspondencesSecond().push_back(Correspondence<T>(x2[0],x2[1],0,x2[2],j,i));
			}

		}
				
		if(outliers1!=0)
			sout << outliers1 << " reprojected points do not ly on the image plane "
			<< i << " of cam 1." << std::endl;
		if(outliers2!=0)
			sout << outliers2 << " reprojected points do not ly on the image plane "
			<< i << " of STEREO CAM 2." << std::endl;
		

		// Create projection images
		cimg_library::CImg<T> i1(imageWidth(),imageHeight());
		i1.fill(255);
		const unsigned char black[] = { 0,0,0 } ;
		auto it = this->correspondences().begin();
		auto itEnd = this->correspondences().end();
		for (; it != itEnd; it++)
		{
			if( it->imgIndex != i )
				continue;
			char txt[32];
			std::sprintf(txt,"%u",it->pointIndex);
			cimg_library::CImg<T> pt = it->pt2d();
			i1.draw_text((int)pt[0],(int)pt[1], txt, black);
		}		
		this->imageSequence().push_back(i1);

	}
	
	
	// Give out point stats ------------------------------------------------------
	sout << std::endl << points3D.size() << " 3D points" << std::endl;
	sout << this->correspondences().sizePoints() << " 2D correspondences (this->correspondences().sizePoints())" << std::endl;
	sout << this->correspondences().sizeImages() << " different image ids of 2D correspondences" << std::endl;
	sout << this->correspondences().size() << " points overall (this->correspondences().size())" << std::endl;
	sout << this->pmatrix().size() << " cameras" << std::endl;


	// Write vector of ids
	size_t noImages = 10;
	size_t noPatchesPerImage = 10;
	std::pair<int,int> coord(222,333);
	this->vecOfVecOfPair().resize(noImages);
	for( size_t i = 0; i < noImages; i++)
	{
		this->vecOfInt().push_back(i);
		this->vecOfVecOfPair()[i].resize(noPatchesPerImage);
		for( size_t j = 0; j < noPatchesPerImage; j++)
		{
			this->vecOfVecOfPair()[i][j].push_back( coord );
			this->vecOfVecOfPair()[i][j].push_back( coord );
		}
	}


}

/// Creates 3D points positioned as a cube and saved in the structure matrix W.
template <typename T>
void createPointCube(CorrespondenceList<T> &points3D, const int numberOfPoints){
		
	// Create  point
	cimg_library::CImg<double> pt(1,4);

	double pointsPerLength = (int)pow((double)numberOfPoints,(double)1/3);
	double cubeLength = 4.;

	// Define point densitiy of the cube.
	double pointSteps = cubeLength / (double) pointsPerLength;

	// Create translation vector which moves the cube so it lies at origin.
	cimg_library::CImg<double> cubeTranslation(1,4);
	cubeTranslation.fill(-cubeLength/2.);
	cubeTranslation[3] = 0.;

	int numerOfPoints = 0;
	for(int i=0; i<=pointsPerLength; i++){
		for(int j=0; j<=pointsPerLength; j++){
			for(int k=0; k<=pointsPerLength; k++){

				// Create point
				pt(0,0) = pointSteps*double(i);
				pt(1,0) = pointSteps*double(j);
				pt(2,0) = pointSteps*double(k);
				pt(3,0) = 1.;

				pt += cubeTranslation;

				Correspondence<T> corr(pt,numerOfPoints,0);
				points3D.push_back( corr );

				numerOfPoints++;

			}
		}
	}
	sout << "Generated " << numerOfPoints << " points." << std::endl;
}

template <typename T>
void createGaussianPoints(CorrespondenceList<T> &points3D, 
						  const int numberOfPoints, const double sigma){

	sout << "Generating " << numberOfPoints << " gaussian distributed"
		<< " 3D points." << std::endl;
	
	for(int i=0; i<numberOfPoints; i++){

		// Create gaussian distributed point
		cimg_library::CImg<T> pt(1,4);
		pt.fill(T(0)).noise(T(sigma));
		
		// Set homogeneous value to 1
		pt[3] = T(1);

		// Save in CorrespondenceList<T>
		Correspondence<T> corr(pt,i,0);
		points3D.push_back( corr );

	}
}


#endif /* _TESTDRIVER_HXX_ */



