/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file TestdriverEvaluation.hxx
 *  Implementation of the parameter class TestdriverEvaluation.
 *  \author Moritz Becker
 *  \date 08.02.2011
 */

#ifndef _TESTDRIVEREVALUATION_HXX_
#define _TESTDRIVEREVALUATION_HXX_

#define PI	3.14159265358979323846264338327950288419716939937510d

#include <hekate/TestdriverEvaluation.h>

template <typename T>
TestdriverEvaluation<T>::TestdriverEvaluation(const std::string& name) :
		TemplatedParameteredObject<T>(
				"testDriverEvaluation", name,
				"Evaluates 3D reconstruction of given 'testdriver' data. Parallelism and orthogonality is tested."
		)
			,points3D(false, false) 
			
			
{	

	ParameteredObject::_addInputSlot(points3D, "points3D", 
	"points3D", 
	"vector<vector<T> >"); 
	
}

template <typename T>
void TestdriverEvaluation<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	// define 8 corner points
	/*
	p1 front left bottom
	p2 back left bottom
	p3 front right bottom
	p4 back right bottom
	p5 front left top
	p6 back left top
	p7 front right top
	p8 back right top
	*/
	cimg_library::CImg<T> p1(1,3), p2(1,3), p3(1,3), 
		p4(1,3), p5(1,3), p6(1,3), p7(1,3), p8(1,3);

	p1[0]=points3D()[0][0]; 
	p1[1]=points3D()[0][1]; 
	p1[2]=points3D()[0][2]; 
	p2[0]=points3D()[5][0]; 
	p2[1]=points3D()[5][1]; 
	p2[2]=points3D()[5][2]; 
	p3[0]=points3D()[30][0]; 
	p3[1]=points3D()[30][1]; 
	p3[2]=points3D()[30][2]; 
	p4[0]=points3D()[35][0]; 
	p4[1]=points3D()[35][1]; 
	p4[2]=points3D()[35][2]; 
	p5[0]=points3D()[36][0]; 
	p5[1]=points3D()[36][1]; 
	p5[2]=points3D()[36][2]; 
	p6[0]=points3D()[41][0]; 
	p6[1]=points3D()[41][1]; 
	p6[2]=points3D()[41][2]; 
	p7[0]=points3D()[66][0]; 
	p7[1]=points3D()[66][1]; 
	p7[2]=points3D()[66][2];
	p8[0]=points3D()[71][0]; 
	p8[1]=points3D()[71][1]; 
	p8[2]=points3D()[71][2]; 

	sout << "p1" << ": " << p1[0] << ", " << p1[1]	<< ", " <<p1[2] << std::endl;
	sout << "p2" << ": " << p2[0] << ", " << p2[1]	<< ", " <<p2[2] << std::endl;
	sout << "p3" << ": " << p3[0] << ", " << p3[1]	<< ", " <<p3[2] << std::endl;
	sout << "p4" << ": " << p4[0] << ", " << p4[1]	<< ", " <<p4[2] << std::endl;
	sout << "p5" << ": " << p5[0] << ", " << p5[1]	<< ", " <<p5[2]	<< std::endl;
	sout << "p6" << ": " << p6[0] << ", " << p6[1]	<< ", " <<p6[2] << std::endl;
	sout << "p7" << ": " << p7[0] << ", " << p7[1]	<< ", " <<p7[2] << std::endl;
	sout << "p8" << ": " << p8[0] << ", " << p8[1]	<< ", " <<p8[2] << std::endl;

	// define 4 lines for every point plane (normalized to 1)
	cimg_library::CImg<T> l12(1,3), l13(1,3), l24(1,3) ,l34(1,3),
		l15(1,3), l37(1,3), l26(1,3) ,l48(1,3),
		l56(1,3), l57(1,3), l68(1,3) ,l78(1,3);
	l12 = p1-p2;
	l13 = p1-p3;
	l24 = p2-p4;
	l34 = p3-p4;
	l15 = p1-p5;
	l37 = p3-p7;
	l26 = p2-p6;
	l48 = p4-p8;
	l56 = p5-p6;
	l57 = p5-p7;
	l68 = p6-p8;
	l78 = p7-p8;

	// scale ratios
	sout << "Length x: " << getLength(l13) << std::endl;
	sout << "Length y: " << getLength(l12) << std::endl;
	sout << "Length z: " << getLength(l15) << std::endl;

	// normalize lines to length of 1
	normalizeLine(l12);
	normalizeLine(l13);
	normalizeLine(l24);
	normalizeLine(l34);
	normalizeLine(l15);
	normalizeLine(l37);
	normalizeLine(l26);
	normalizeLine(l48);
	normalizeLine(l56);
	normalizeLine(l57);
	normalizeLine(l68);
	normalizeLine(l78);

	//output line coordinates	
	sout << "Line12" << ": " << l12[0] << ", " << l12[1] << ", " <<l12[2] << std::endl;
	sout << "Line13" << ": " << l13[0] << ", " << l13[1] << ", " <<l13[2] << std::endl;
	sout << "Line24" << ": " << l24[0] << ", " << l24[1] << ", " <<l24[2] << std::endl;
	sout << "Line34" << ": " << l34[0] << ", " << l34[1] << ", " <<l34[2] << std::endl;
	sout << "Line15" << ": " << l15[0] << ", " << l15[1] << ", " <<l15[2] << std::endl;
	sout << "Line37" << ": " << l37[0] << ", " << l37[1] << ", " <<l37[2] << std::endl;
	sout << "Line26" << ": " << l26[0] << ", " << l26[1] << ", " <<l26[2] << std::endl;
	sout << "Line48" << ": " << l48[0] << ", " << l48[1] << ", " <<l48[2] << std::endl;
	sout << "Line56" << ": " << l56[0] << ", " << l56[1] << ", " <<l56[2] << std::endl;
	sout << "Line57" << ": " << l57[0] << ", " << l57[1] << ", " <<l57[2] << std::endl;
	sout << "Line68" << ": " << l68[0] << ", " << l68[1] << ", " <<l68[2] << std::endl;
	sout << "Line78" << ": " << l78[0] << ", " << l78[1] << ", " <<l78[2] << std::endl;

	// measure parellelism
	sout << "Measure paralellism" << std::endl;
	sout << "12-13: " << getAngle(l12,l13) << "" << std::endl;
	sout << "24-34: " << getAngle(l24,l34) << "" << std::endl;
	sout << "15-13: " << getAngle(l15,l13) << "" << std::endl;
	sout << "57-37: " << getAngle(l57,l37) << "" << std::endl;
	sout << "37-78: " << getAngle(l37,l78) << "" << std::endl;
	sout << "34-48: " << getAngle(l34,l48) << "" << std::endl;

}

// calculate angle between two lines
template <typename T>
double getAngle(cimg_library::CImg<T> a,cimg_library::CImg<T> b){
	double angle = 180./PI*acos(double(a.dot(b)));
	return angle;
}

// normalize line
template <typename T>
void normalizeLine(cimg_library::CImg<T> &a){
	a /= sqrt(double(a.dot(a)));
}

// length
template <typename T>
double getLength(cimg_library::CImg<T> &a){
	return sqrt(double(a.dot(a)));
}

#endif /* _TESTDRIVEREVALUATION_HXX_ */



