/*  Copyright (C) 2012 Michael Mller

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Triangulation.hxx
 *  Implementation of the parameter class Triangulation.
 *  \author <a href="mailto:michael@fds-team.de">
 *	  Michael Mller</a>
 *  \date 12.04.2012
 */

#ifndef _TRIANGULATION_HXX_
#define _TRIANGULATION_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "hekate/Triangulation.h"
#include "hekate/totalLstsqMinimize.h"
#include <vector>
#include "CImg.h"
#include <charon-utils/ImgTool.hxx>

using namespace cimg_library;

template <typename T>
Triangulation<T>::Triangulation(const std::string& name) :
		TemplatedParameteredObject<T>(
			"Triangulation", name,
			"<h2>Triangulates 2D points and camera matrices to 3D "
			"points</h2><br>"
			"Triangulates 2D points and camera matrices to 3D points"
		)
{
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		cameras, "cameras",
		"cameras",
		"IndexedMatrixList");

	ParameteredObject::_addOutputSlot(
		pointcloud, "pointcloud",
		"pointcloud",
		"CorrespondenceList");

	ParameteredObject::_addParameter< bool >(
		useSVD, "useSVD",
		"use CImg SVD",
		false, "bool");

}



template <typename T>
void Triangulation<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	
	const hekate::CorrespondenceList<T> &correspondenceList = this->correspondences();
	const IndexedMatrixList	&cameraList	= this->cameras();

	sout << "Triangulating:\n\t" << correspondenceList.sizeImages() << " different image indices\n\t"
		<< correspondenceList.sizePoints() << " different point indices\n\t" << cameraList.size()
		<< " cameras\n\t" << correspondenceList.size() << " feature points overall" << std::endl;

	if (cameraList.size() < 2) return;

	size_t maxIndex = correspondenceList.getMaxPointIndex();
	for(size_t index = 0; index <= maxIndex; index++){
	
		std::vector< const hekate::Correspondence<T>*> points = correspondenceList.getListByPointId(index);

		if (points.size() >= 2){
			
			typename std::vector< const hekate::Correspondence<T>*>::iterator pointIt;

			try{
				T *A = new T[4 * points.size() * 2];

				int rowsA = 0;
				
				for(pointIt = points.begin(); pointIt != points.end(); pointIt++){
			
					try{
						const hekate::Correspondence<T> &point2d = **pointIt;
						const IndexedMatrix &camera = cameraList(point2d.imgIndex);

						if (camera.degenerated || point2d.outlier){
							//sout << "camera.degenerated: " << camera.degenerated 
							//	<< "point2d.outlier: " << point2d.outlier << std::endl;
							continue;
						}
						
						//first row of A
						A[rowsA*4+0] = point2d.point[0]*camera(0,2) - camera(0,0);
						A[rowsA*4+1] = point2d.point[0]*camera(1,2) - camera(1,0);
						A[rowsA*4+2] = point2d.point[0]*camera(2,2) - camera(2,0);
						A[rowsA*4+3] = point2d.point[0]*camera(3,2) - camera(3,0);
						rowsA++;

						//second row of A
						A[rowsA*4+0] = point2d.point[1]*camera(0,2) - camera(0,1);
						A[rowsA*4+1] = point2d.point[1]*camera(1,2) - camera(1,1);
						A[rowsA*4+2] = point2d.point[1]*camera(2,2) - camera(2,1);
						A[rowsA*4+3] = point2d.point[1]*camera(3,2) - camera(3,1);
						rowsA++;

					}
					catch(...){
					}
				
				}

				if(rowsA < 2){
					continue;
				}

				CImg<T> X(1,4);

				// If useSVD is set, use CIMG SVD to solve equation
				if(this->useSVD()){
					cimg_library::CImg<T> AA(4,rowsA);
					for(size_t i = 0; i < rowsA; i++){
						AA(0,i) = A[i*4];
						AA(1,i) = A[i*4+1];
						AA(2,i) = A[i*4+2];
						AA(3,i) = A[i*4+3];
					}
					CImg<T> U, S, V;
					AA.SVD(U,S,V);
					V.crop(3,0,3,3);
					X(0, 0) = V[0] / V[3]; 				
					X(0, 1) = V[1] / V[3]; 
					X(0, 2) = V[2] / V[3]; 
					X(0, 3) = 1; 
				}
				else{
					T result[4];
					totalLstsqMinimize(A, rowsA, (T*)&result, false);	

					CImg<T> X(1,4);
					X(0, 0) = result[0] / result[3]; 				
					X(0, 1) = result[1] / result[3]; 
					X(0, 2) = result[2] / result[3]; 
					X(0, 3) = 1; 
				}

				hekate::Correspondence<T> point3D( X, index, 0, false, points[0]->R, points[0]->G, points[0]->B);
				this->pointcloud().push_back(point3D);
				
				delete[] A;
			}
			catch(...){

			}		

		}

	}
	
}

#endif /* _TRIANGULATION_HXX_ */
