/*  Copyright (C) 2012 Michael Mller

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file TriangulationGPU.hxx
 *  Implementation of the parameter class TriangulationGPU.
 *  \author <a href="mailto:michael@fds-team.de">
 *	  Michael Mller</a>
 *  \date 12.04.2012
 */

#ifndef _TriangulationGPU_HXX_
#define _TriangulationGPU_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "hekate/TriangulationGPU.h"
#include "hekate/totalLstsqMinimize_gpu.h"
#include <vector>
#include "CImg.h"
//#include <new>


using namespace cimg_library;

template <typename T>
TriangulationGPU<T>::TriangulationGPU(const std::string& name) :
		TemplatedParameteredObject<T>(
			"TriangulationGPU", name,
			"<h2>Triangulates 2D points and camera matrices to 3D "
			"points</h2><br>"
			"Triangulates 2D points and camera matrices to 3D points"
		)
{
	ParameteredObject::_addInputSlot(
		correspondences, "correspondences",
		"correspondences",
		"CorrespondenceList");
	ParameteredObject::_addInputSlot(
		cameras, "cameras",
		"cameras",
		"IndexedMatrixList");

	ParameteredObject::_addOutputSlot(
		pointcloud, "pointcloud",
		"pointcloud",
		"CorrespondenceList");

}

template <typename T>
class OutputFunctor{

	private:
		hekate::CorrespondenceList<T> *output;

	public:
		OutputFunctor(hekate::CorrespondenceList<T> &output){
			this->output = &output;
		}


		void operator()(int pointIndex, float result[4]){
			CImg<T> X(1,4);				

			X(0, 0) = result[0] / result[3]; 				
			X(0, 1) = result[1] / result[3]; 
			X(0, 2) = result[2] / result[3]; 
			X(0, 3) = 1; 
			this->output->push_back(hekate::Correspondence<T>( X, pointIndex, 0, false));

		}
	
};


template <typename T>
void TriangulationGPU<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	
	const hekate::CorrespondenceList<T> &correspondenceList = this->correspondences();
	const IndexedMatrixList		&cameraList			= this->cameras();

	if (cameraList.size() <= 2) return;

	OutputFunctor<T> outputf(this->pointcloud());

	GPUEigenvectors<OutputFunctor<T> > gpuVec(outputf);

	size_t maxIndex = correspondenceList.getMaxPointIndex();
	for(size_t index = 0; index <= maxIndex; index++){
	
		std::vector< const hekate::Correspondence<T>*> points = correspondenceList.getListByPointId(index);

		if (points.size() >= 2){
			
			typename std::vector< const hekate::Correspondence<T>*>::iterator pointIt;

			try{
				float *A = new float[4 * points.size() * 2];

				int rowsA = 0;

				for(pointIt = points.begin(); pointIt != points.end(); pointIt++){
			
					try{
						const hekate::Correspondence<T> &point2d = **pointIt;
						const IndexedMatrix &camera = cameraList(point2d.imgIndex);

						if (camera.degenerated || point2d.outlier) continue;

						//first row of A
						A[rowsA*4+0] = point2d.point[0]*camera(0,2) - camera(0,0);
						A[rowsA*4+1] = point2d.point[0]*camera(1,2) - camera(1,0);
						A[rowsA*4+2] = point2d.point[0]*camera(2,2) - camera(2,0);
						A[rowsA*4+3] = point2d.point[0]*camera(3,2) - camera(3,0);
						rowsA++;

						//second row of A
						A[rowsA*4+0] = point2d.point[1]*camera(0,2) - camera(0,1);
						A[rowsA*4+1] = point2d.point[1]*camera(1,2) - camera(1,1);
						A[rowsA*4+2] = point2d.point[1]*camera(2,2) - camera(2,1);
						A[rowsA*4+3] = point2d.point[1]*camera(3,2) - camera(3,1);
						rowsA++;

					}catch(...){

					}

				
				}
				
				gpuVec.addMatrix(A, rowsA, index);
				
				delete[] A;

			}catch(...){

			}		

		}
	}	
	gpuVec.executeRemainingJobs();

}

#endif /* _TriangulationGPU_HXX_ */
