/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file UpgradeMetric.hxx
 *  Implementation of the parameter class UpgradeMetric.
 *  \author Moritz Becker
 *  \date 05.08.2010
 */

#ifndef _UPGRADEMETRIC_HXX_
#define _UPGRADEMETRIC_HXX_

#include "UpgradeMetric.h"
#include "charon-utils/ImgTool.hxx"

template <typename T>
UpgradeMetric<T>::UpgradeMetric(const std::string& name) :
		TemplatedParameteredObject<T>(
				"UpgradeMetric", name,
				"Calculates H_euclidean. Needs camera matrix and plane at infinity of affinity."
		)
			,cameraMatrix1(false, true) 
			,planeAtInfinity(false, false) 
			
			
{
	
	ParameteredObject::_addParameter (zeroSkew, "zeroSkew", "Camera has zero skew", true);
	ParameteredObject::_addParameter (principalPoint, "principalPoint", "Principal point at origin", false);
	ParameteredObject::_addParameter (ratioConstraint, "constraintRatio", "aspRatio known", false);
	ParameteredObject::_addParameter (ratio, "ratio", "aspect ratio", 1.);
	ParameteredObject::_addParameter (constraintWeight, "constraintWeight", "Weight of additional constraints", 5.);

	ParameteredObject::_addInputSlot(cameraMatrix1, "cameraMatrix1", 
	"camera matrix 1", 
	"CImgList<T>"); 
	ParameteredObject::_addInputSlot(planeAtInfinity, "planeAtInfinity", 
	"plane at infinity", 
	"std::vector<T>"); 
	ParameteredObject::_addOutputSlot(transformationMatrix, "transformationMatrix", 
	"matrix that upgrades 3D points to metric scale", 
	"CImgList<T>"); 
	
	ParameteredObject::_addInputSlot(input_3d_point_cloud, "input_3d_point_cloud", 
	"The input 3D coordinates accurate up to a projective transformation", 
	"vector<vector<T> >"); 
	ParameteredObject::_addOutputSlot(output_3d_point_cloud, "output_3d_point_cloud", 
	"The output of the affine 3d point coordinates", 
	"vector<vector<T> >"); 
	
}

template <typename T>
void UpgradeMetric<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	//construction of matrix H_infinity (see Hartley p.475 a stratified solution)
	if(this->cameraMatrix1().size() != 0 && this->planeAtInfinity().size()!=0){

		bool zSkew = this->zeroSkew();
		bool pPoint = this->principalPoint();
		double cWeight = this->constraintWeight();

		cimg_library::CImgList<T> H, cameraMatrixLeft, cameraMatrixRight, cameraMatrix ;
		cimg_library::CImg<T> H_temp(3,3), p(1,3) ;

		// cameraMatrix provides all the camera matrices
		for(size_t i=0; i<cameraMatrix1.size(); i++){
			cimg_library::CImg<T> camUnity(4,3);
			camUnity.fill(0); camUnity(0,0)=1;camUnity(1,1)=1;camUnity(2,2)=1;
			if(this->cameraMatrix1[i][0]!=camUnity) cameraMatrix.push_back(this->cameraMatrix1[i][0]);			
			cameraMatrix.push_back(this->cameraMatrix1[i][1]);
		}

		// extract 3x3 and 1x3 matrix of camera matrix
		for(size_t i=0; i<cameraMatrix.size(); i++){
			cameraMatrixLeft.push_back(cameraMatrix[i].get_crop(0,0,2,2));
			cameraMatrixRight.push_back(cameraMatrix[i].get_crop(3,0,3,2));
		}
		
		// initialize plane at infinity
		p(0,0) = this->planeAtInfinity()[0] ;
		p(0,1) = this->planeAtInfinity()[1] ;
		p(0,2) = this->planeAtInfinity()[2] ;

		ImgTool::printInfo(sout,p,"p");

		// Calculate H_infinity (transposed and inverted) for every view pair
		for(size_t i=0; i<cameraMatrix.size(); i++){
			H_temp = cameraMatrixLeft[i] - cameraMatrixRight[i] * p.get_transpose();
			H_temp /= pow(H_temp.det(),1/3) ;

			ImgTool::printInfo(sout,H_temp,"H_temp");

			H_temp.invert();
			H.push_back(H_temp);

		}

		//rearanging equation Ht.w.Hi-w=0 to A*w=0

		// calculate number of cols of A
		cimg_library::CImg<T> A(6,(cameraMatrix.size())*6+4);
		A.fill(0.);

		for(size_t i=0; i<H.size(); i++){

			int j=i*6;

			cimg_library::CImg<T> J;
			J = H[i];

			A(0,0+j) = -1 + J[0] * J[0];		
			A(1,0+j) = 2 * J[0] * J[3];		
			A(2,0+j) = 2 * J[0] * J[6];
			A(3,0+j) = J[3] * J[3];		
			A(4,0+j) = 2 * J[3] * J[6];		
			A(5,0+j) = J[6] * J[6];	

			//second line of A
			A(0,1+j) = J[0] * J[1];		
			A(1,1+j) = -1 + J[1] * J[3] + J[0] * J[4];		
			A(2,1+j) = J[1] * J[6] + J[0] * J[7];
			A(3,1+j) = J[3] * J[4];		
			A(4,1+j) = J[4] * J[6] + J[3] * J[7];		
			A(5,1+j) = J[6] * J[7];

			//tHrd line of A
			A(0,2+j) = J[0] * J[2];		
			A(1,2+j) = J[2] * J[3] + J[0] * J[5];		
			A(2,2+j) = -1 + J[2] * J[6] + J[0] * J[8];
			A(3,2+j) = J[3] * J[5];		
			A(4,2+j) = J[5] * J[6] + J[3] * J[8];		
			A(5,2+j) = J[6] * J[8];
			
			//fourth line of A
			A(0,3+j) = J[1] * J[1];		
			A(1,3+j) = 2 * J[1] * J[4];		
			A(2,3+j) = 2 * J[1] * J[7];
			A(3,3+j) = -1 + J[4] * J[4];		
			A(4,3+j) = 2 * J[4] * J[7];		
			A(5,3+j) = J[7] * J[7];	

			//fifth line of A
			A(0,4+j) = J[1] * J[2];		
			A(1,4+j) =  J[2] * J[4] + J[1] * J[5];		
			A(2,4+j) = J[2] * J[7] + J[1] * J[8];
			A(3,4+j) = J[4] * J[5];		
			A(4,4+j) = -1 + J[5] * J[7] + J[4] * J[8];		
			A(5,4+j) = J[7] * J[8];

			//sixth line of A
			A(0,5+j) = J[2] * J[2];		
			A(1,5+j) = 2 * J[2] * J[5];		
			A(2,5+j) = 2 * J[2] * J[8];
			A(3,5+j) = J[5] * J[5];		
			A(4,5+j) = 2 * J[5] * J[8];		
			A(5,5+j) = -1 + J[8] * J[8];	

			// additional constraints

			// zero skew	
			if(zSkew){
				A(0,cameraMatrix.size()*6) = 0;
				A(1,cameraMatrix.size()*6) = cWeight;
				A(2,cameraMatrix.size()*6) = 0;
				A(3,cameraMatrix.size()*6) = 0;
				A(4,cameraMatrix.size()*6) = 0;
				A(5,cameraMatrix.size()*6) = 0;
			}

			// principal point at origin
			if(pPoint){
				A(0,cameraMatrix.size()*6+1) = 0;
				A(1,cameraMatrix.size()*6+1) = 0;
				A(2,cameraMatrix.size()*6+1) = cWeight;
				A(3,cameraMatrix.size()*6+1) = 0;
				A(4,cameraMatrix.size()*6+1) = 0;
				A(5,cameraMatrix.size()*6+1) = 0;

				// zero skew	
				A(0,cameraMatrix.size()*6+2) = 0;
				A(1,cameraMatrix.size()*6+2) = 0;
				A(2,cameraMatrix.size()*6+2) = 0;
				A(3,cameraMatrix.size()*6+2) = 0;
				A(4,cameraMatrix.size()*6+2) = cWeight;
				A(5,cameraMatrix.size()*6+2) = 0;
			}

			// aspect ratio	
			if(this->ratioConstraint()){
				A(0,cameraMatrix.size()*6+3) = -1*cWeight;
				A(1,cameraMatrix.size()*6+3) = 0;
				A(2,cameraMatrix.size()*6+3) = 0;
				A(3,cameraMatrix.size()*6+3) = pow(this->ratio(),2.)*cWeight;
				A(4,cameraMatrix.size()*6+3) = 0;
				A(5,cameraMatrix.size()*6+3) = 0;
			}
		
		}

		// output A matrix
		for(size_t i=0; i < cameraMatrix.size()*6+4; i++){
			sout << "A[" << i << "] = " ;
			for(int j=0; j < 6; j++){
				sout << A(j,i) << "     ";
			}
			sout << std::endl;
		}	

		//ImgTool::printInfo(sout,A,"A");

		cimg_library::CImg<T> U, S, V, w(1,6);

		A.SVD(U,S,V);

		//DIAC is last column of V (see Hartley p.593)
		w = V.get_crop(5,0,5,5);

		for(int i=0; i<A.height(); i++){
			sout << "A*w[" << i <<"] = " << (A*w)[i] << std::endl;
		}
		ImgTool::printInfo(sout,w,"w");
		ImgTool::printInfo(sout,S,"S");

		//calculate K with cholesky factorization
		cimg_library::CImg<T> K(3,3);
		K=w.get_tensor().invert();
		//set upper triangular matrix to zero
		K(1,0) = 0; K(2,0) = 0; K(2,1) = 0; 		

		double sumK;

		for(int i=0; i<3; i++){		
			//calc entries under the diag
			for(int j=0; j<i; j++){			
				sumK = K(j,i);

				for(int k=0; k<j; k++){			
					sumK -= K(k,i) * K(k,j);			
				}

				K(j,i) = sumK / K(j,j);			
			}

			//calc diag entries
			sumK = K(i,i);

			for(int k=0; k<i; k++){		
				sumK -= K(k,i)*K(k,i);		
			}
						
			if(sumK<=0){
				sout << "ATTENTION: K is not positive-definite!" << std::endl;
				sout << "ATTENTION: Cannot calculate metric transformation" << std::endl;
				break;
			}
			
			K(i,i)=sqrt(sumK);
		
		}
		
		if(sumK>0){

			ImgTool::printInfo(sout,K,"K");

			ImgTool::printInfo(sout,(K*K.get_transpose()).invert(),"(KK^T)^-1");

			sout << "....................................................................................................................." 
				<< std::endl;
			sout << ":   K is positive-definite, calculating metric points" 
				<< std::endl;
			sout << ":...................................................................................................................." 
				<< std::endl;
			sout << std::endl;
		
			cimg_library::CImg<T> H_metric(4,4),K_crop(1,3), H_affin(4,4);
			H_affin.identity_matrix();
			H_affin(0,3) = - p[0];
			H_affin(1,3) = - p[1];
			H_affin(2,3) = - p[2];

			ImgTool::printInfo(sout,H_affin,"H_affin");

			//H_metric (see Hartley p.479)
			H_metric.fill(0.);
			for(int i=0; i<3; i++){
				for(int j=0; j<3; j++){
					H_metric(i,j)=K(i,j);
				}
			}
			H_metric(3,3)=1;

			ImgTool::printInfo(sout,H_metric,"H_metric");		
			ImgTool::printInfo(sout,H_affin*H_metric,"H_a*H_m");		

			this->transformationMatrix().push_back(H_affin*H_metric);

			cimg_library::CImg<T> oldVector(1,4),newVector(1,4);

			int pointSize = this->input_3d_point_cloud().size();
			this->output_3d_point_cloud().resize(pointSize);
			sout << pointSize << " points available" << std::endl;

			double maxX=0., maxY=0., maxZ=0.;

			for(int i=0; i<pointSize; i++){
			
				oldVector(0,0)=this->input_3d_point_cloud()[i][0];
				oldVector(0,1)=this->input_3d_point_cloud()[i][1];
				oldVector(0,2)=this->input_3d_point_cloud()[i][2];
				oldVector(0,3)=1.;

				newVector = (H_affin*H_metric).get_invert()*oldVector;
				newVector /= newVector(0,3);
			
				this->output_3d_point_cloud()[i].push_back(newVector(0,0));
				this->output_3d_point_cloud()[i].push_back(newVector(0,1));
				this->output_3d_point_cloud()[i].push_back(newVector(0,2));

				if(std::abs(maxX) < std::abs(newVector(0,0))) maxX=newVector(0,0);
				if(std::abs(maxY) < std::abs(newVector(0,1))) maxY=newVector(0,1);
				if(std::abs(maxZ) < std::abs(newVector(0,2))) maxZ=newVector(0,2);
			
			}

			double maxxx = maxX;
			if(std::abs(maxxx) < std::abs(maxY)) maxxx=maxY;
			if(std::abs(maxxx) < std::abs(maxZ)) maxxx=maxZ;
			
			for(int i=0; i < pointSize ; i++)
			{				
				output_3d_point_cloud()[i][0] /= maxxx *0.1;
				output_3d_point_cloud()[i][1] /= maxxx *0.1;
				output_3d_point_cloud()[i][2] /= maxxx *0.1;
			}	

		}

	}

}

#endif /* _UPGRADEMETRIC_HXX_ */



