/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file UpgradeMetricADQ.hxx
 *  Implementation of the parameter class UpgradeMetricADQ.
 *  \author Moritz Becker
 *  \date 05.08.2010
 */


#ifndef _UPGRADEMETRICADQ_HXX_
#define _UPGRADEMETRICADQ_HXX_

#include <hekate/UpgradeMetricADQ.h>
#include <charon-utils/ImgTool.hxx>
/*
#include <cmath>
#include <gsl/gsl_math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
*/
#include <opencv/cv.h>
#include <vigra/linear_solve.hxx>
#include <vigra/multi_array.hxx>
#include <vigra/windows.h>


template <typename T>
UpgradeMetricADQ<T>::UpgradeMetricADQ(const std::string& name) :
		TemplatedParameteredObject<T>(
				"UpgradeMetricADQ", name,
				"Calculates the metric upgrade for at least three (projective) camera matrices. "
				"This plugin is based on Thormhlen et al., Robust Linear Auto-Calibration of a "
				"Moving Camera from Image Sequences."
		),cameraMatrices(false,false)
			
{
	
	ParameteredObject::_addParameter (ratio, "ratio", "Aspect ratio", 1.0);

	ParameteredObject::_addInputSlot(cameraMatrices, "cameraMatrices", 
	"Camera matrices.", 
	"IndexedMatrixList"); 
	ParameteredObject::_addInputSlot(input3dPointCloud, "input3dPointCloud", 
	"The projective input 3D points.", 
	"CorrespondenceList"); 
	ParameteredObject::_addInputSlot(calibrationMatrix, "calibrationMatrix", 
	"Calibration matrix to normalize the camera matrices.", 
	"CImgList<T>"); 

	ParameteredObject::_addOutputSlot(transformationMatrix, "transformationMatrix", 
	"Matrix that upgrades 3D points to metric scale.", 
	"CImgList<T>"); 	
	ParameteredObject::_addOutputSlot(output3dPointCloud, "output3dPointCloud", 
	"Transformed 3D points.", 
	"CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(cameraMatricesOut, "cameraMatricesOut", 
	"Output camera matrices.", 
	"IndexedMatrixList"); 
	
}

template <typename T>
void UpgradeMetricADQ<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();


		const IndexedMatrixList cameras = cameraMatrices().getNonDegeneratedList() ;
		const CorrespondenceList<T> &corrs = this->input3dPointCloud();

		// read out aspect ratio from input parameter (normally should be 1)
		double aspectRatio = this->ratio();
		sout << "Aspect ratio: " << aspectRatio << std::endl;

		//estimated calibration matrix
		//ATTENTION: In spite of Pollefeys/Thormhlen, 
		//K_N can also contain x_0, y_0 for principal point.
		CImgList<double> H;
		CImg<double> K_N(3,3), H_temp(3,3);
		K_N = this->calibrationMatrix()[0];
		ImgTool::printInfo(sout,K_N,"K_N");
		//ImgTool::printInfo(sout,K_N*K_N.get_transpose(),"K_N*K_N^T");
		
		// Create normalized camera matrices
		IndexedMatrixList camerasNorm; ;
		for(size_t i=0; i<cameras.size(); i++){
			try
			{
				IndexedMatrix cam = ( K_N.get_invert() * (CImg<double>)cameras[i] );
				size_t index = cameras[i].getIndex1() ;
				cam.setIndex(index);
				camerasNorm.push_back(cam);
				//ImgTool::printInfo(sout,cam,"cam");
			}
			catch(...){
				sout << "Could not read in camera cameras[" << i << "]." << std::endl;
			}
		}

		sout << "Read in " << camerasNorm.size() << " camera matrices." << std::endl;
		

		// Declear matrices T_metric, R_metric and Q_metric where the results
		// with the lowest cost sum will be saved.
		cimg_library::CImg<double> T_metric(4,4);
		cimg_library::CImg<double> R_metric(3,3);
		cimg_library::CImg<double> Q_metric(4,3);

		double minCostSum = 0;
		double k_minCostSum = 0;
		bool firstRun = true;
		
		// Iterate algorithm with different parameters and take the solution with minimal cost sum
		for(int k=0; k<50; k++){

			//sout << "Calculating metric transformation: loop " << k << std::endl;

			// Define parameters for the linear system
			double c1, c2, c3, c4, c5, c6;
			c1 = 100; 
			c2 = 10; 
			c3 = 10;
			c4 = 5; 
			c5 = 10*exp(-0.3*k) ; 
			c6 = 10*exp(-0.3*k) ;
			//sout << "Taking parameters c6 = c5 = " << c5 << std::endl;

			// Create linear system A  (Thormhlen et al. equation 14-19)
			CImg<double> A = createLinearSystem(camerasNorm, c1, c2, c3, c4, c5, c6);

			// Write matrix a to log
			/*
			if(firstRun)
				for(size_t i=0; i < camerasNorm.size()*6; i++){
					sout << "A[" << i << "] = " ;
					for(int j=0; j < 10; j++){
						sout << A(j,i) << "     ";
					}
					sout << std::endl;
				}	
				*/

			cimg_library::CImg<double>  U, S, V, U2, S2, V2, Q(1,10), Q2(4,4), U3, diag, T3, T_M(4,4);

			// Calculate the SVD of A
			A.SVD(U,S,V);
			
			//DIAC is last column of V (see Hartley p.593)
			Q = V.get_crop(9,0,9,9);
			//ImgTool::printInfo(sout,Q,"Q");

			/*
			double sumAQ = 0;
			for(int i=0; i<A.height(); i++){
				sumAQ += std::abs((A*Q)[i]);
				sout << "AQ" << i << ": " << (A*Q)[i] << std::endl;
			}
			sout << "Sum A*Q = " << sumAQ << std::endl;
			*/

			// Transform vector Q to symmetric matrix Q2
			Q2.fill(0.);
			Q2(0,0) = Q[0];		Q2(1,0) = Q[1];		Q2(2,0) = Q[2];		Q2(3,0) = Q[3];
			Q2(0,1) = Q[1];		Q2(1,1) = Q[4];		Q2(2,1) = Q[5];		Q2(3,1) = Q[6];		
			Q2(0,2) = Q[2];		Q2(1,2) = Q[5];		Q2(2,2) = Q[7];		Q2(3,2) = Q[8];		
			Q2(0,3) = Q[3];		Q2(1,3) = Q[6];		Q2(2,3) = Q[8];		Q2(3,3) = Q[9];

			//ImgTool::printInfo(sout,Q2,"Q2");

			// Create SVD of Q2
			Q2.SVD(U2,S2,V2);

			// Crop the left 3 columns of U2 and save them to U3
			U3 = U2.get_crop(0,0,2,3);

			// Crop the first three tallest values of the diagonal matrix S
			// and take the square root of them
			diag = S2.get_crop(0,0,0,2);
			diag[0] = sqrt((double)diag[0]);
			diag[1] = sqrt((double)diag[1]);
			diag[2] = sqrt((double)diag[2]);
			diag.diagonal();

			T3 = U3*diag;

			cimg_library::CImg<double> T4(1,4);
			T4.fill(0.); T4[3]=1;

			// T_M = [ U3*diag | (0,0,0,1)^T ]
			T_M = T3.get_append(T4);			

			//ImgTool::printInfo(sout,T_M,"T_M");
			//ImgTool::printInfo(sout,T_M*T_M.get_transpose(),"T_M");

			// calculate cost function to rate the current result
			//sout << "Check cost function" << std::endl;
			double costSum = 0;
			cimg_library::CImg<double> R_cost, Q_cost;
			for(size_t i=0; i < camerasNorm.size(); i++){

				cimg_library::CImg<double> A_M;
				R_cost.clear();
				Q_cost.clear();

				//ImgTool::printInfo(sout,cameraMatrix[i]*Q2*cameraMatrix[i].get_transpose(),"P*Q2*P^T");

				A_M = camerasNorm[i] * T_M;

				// get RQ factorization (A_M = R * Q)
				getRQ(A_M, R_cost, Q_cost);

				//ImgTool::printInfo(sout,R_cost,"R_cost");
				//ImgTool::printInfo(sout,Q_cost.get_crop(0,0,2,2)*Q_cost.get_crop(0,0,2,2).transpose(),"Q_cost*Q_cost");

				// read out single parameters from matrix R to calculate the cost function
				double s, cx, cy, r, f;
				s = R_cost(1,0);
				cx = R_cost(2,0);
				cy = R_cost(2,1);
				r = R_cost(1,1)/R_cost(0,0);
				f = R_cost(0,0);

				// calculate cost function
				costSum += (s*s + cx*cx + cy*cy
					+ (r-aspectRatio)*(r-aspectRatio))/(f*f);

			}	

			//sout << "[" << k << "] " << "cost sum " << ": " << costSum << std::endl;

			// overwrite minimal cost value if new value is smaller than the former value
			if(costSum < minCostSum || firstRun){ 
				k_minCostSum = k;
				minCostSum = costSum;
				T_metric = T_M;
				R_metric = R_cost;
				Q_metric = Q_cost;
			}

			firstRun = false;

		}

		sout << "minimum cost sum found after " << k_minCostSum << " iterations: s = " << minCostSum << std::endl;

		// push back upgrade matrix
		this->transformationMatrix().push_back(T_metric);

		// Write the calculated matrices to log
		//ImgTool::printInfo(sout,T_metric,"T_metric");
		//ImgTool::printInfo(sout,R_metric,"R_metric");
		//ImgTool::printInfo(sout,Q_metric,"Q_metric");
		//ImgTool::printInfo(sout,Q_metric.get_crop(0,0,2,2)*Q_metric.get_crop(0,0,2,2).transpose(),"QQ^T");

		// Transform input camera matrices by the calculated transformation matrix
		for(size_t i=0; i<cameras.size(); i++){
			cimg_library::CImg<double> cam1 = cameras[i];
			cimg_library::CImg<double> cam1new;
			cam1new = cam1 * T_metric;
			IndexedMatrix camOut = cam1new;
			camOut.setIndex( cameras[i].getIndex1() );
			this->cameraMatricesOut().push_back(camOut);
			//ImgTool::printInfo(sout,cam1new,"cam1new");

		}		

		// Transform input 3D points by the calculated transformation matrix.
		
		int pointSize = this->input3dPointCloud().size();
		sout << pointSize << " points available" << std::endl;
		

		cimg_library::CImg<double> TInvert;
		TInvert = T_metric.get_invert();
		//TInvert /= TInvert.det();
		ImgTool::printInfo(sout,TInvert,"TInvert");

		
		auto it = corrs.begin();
		auto itEnd = corrs.end();
		for(; it < itEnd; it++){
			
			CImg<double> pt = (CImg<double>) it->pt3d();

			pt = TInvert*pt;
			pt /= pt[3];
			//newVector /= newVector(0,3);

			/*
			if(i%100 == 0){
				ImgTool::printInfo(sout,oldVector,"oldVector");
				ImgTool::printInfo(sout,newVector,"newVector");
			}
			*/
			
			Correspondence<T> ptOut(pt, it->pointIndex, it->imgIndex, false, it->R, it->G, it->B);
			this->output3dPointCloud().push_back(ptOut) ;
			
		}
		
		
}

/*	
	Creates linear system based on Thormhlen et al. equation (14)-(19) to calculate the ADQ.
	For every camera matrix, additional six rows are added to the matrix A.
	At least three camera matrices are needed to overdetermine the linear system.
*/
template <typename T>
CImg<double> UpgradeMetricADQ<T>::createLinearSystem(IndexedMatrixList &cameraMatrix,
					double c1, double c2, double c3, double c4, double c5, double c6){

	if(cameraMatrix.size() == 0)
		throw(std::runtime_error("UpgradeMetricADQ<T>::createLinearSystem: Empty IndexedMatrixList"));
						
	CImg<double> A( 10, cameraMatrix.size() * 6 );

	for(size_t i=0; i < cameraMatrix.size(); i++){

		int j=i*6;

		

		cimg_library::CImg<double> J;
		J = cameraMatrix[i].get_vector();

		A(0,0+j) = J[0] * J[4] *c1;		
		A(1,0+j) = (J[1] * J[4] + J[0] * J[5]) *c1;		
		A(2,0+j) = (J[2] * J[4] + J[0] * J[6]) *c1;
		A(3,0+j) = (J[3] * J[4] + J[0] * J[7]) *c1;		
		A(4,0+j) = J[1] * J[5] *c1;		
		A(5,0+j) = (J[2] * J[5] + J[1] * J[6]) *c1;	
		A(6,0+j) = (J[3] * J[5] + J[1] * J[7]) *c1;	
		A(7,0+j) = J[2] * J[6] *c1;	
		A(8,0+j) = (J[3] * J[6] + J[2] * J[7]) *c1;	
		A(9,0+j) = J[3] * J[7] *c1;

		A(0,1+j) = J[0] * J[8] *c2;		
		A(1,1+j) = (J[1] * J[8] + J[0] * J[9]) *c2;		
		A(2,1+j) = (J[0] * J[10] + J[2] * J[8]) *c2;
		A(3,1+j) = (J[0] * J[11] + J[3] * J[8]) *c2;		
		A(4,1+j) = J[1] * J[9] *c2;		
		A(5,1+j) = (J[1] * J[10] + J[2] * J[9]) *c2;	
		A(6,1+j) = (J[1] * J[11] + J[3] * J[9]) *c2;	
		A(7,1+j) = J[10] * J[2] *c2;	
		A(8,1+j) = (J[11] * J[2] + J[10] * J[3]) *c2;	
		A(9,1+j) = J[11] * J[3] *c2;

		A(0,2+j) = J[4] * J[8] *c3;		
		A(1,2+j) = (J[5] * J[8] + J[4] * J[9]) *c3;		
		A(2,2+j) = (J[10] * J[4] + J[6] * J[8]) *c3;
		A(3,2+j) = (J[11] * J[4] + J[7] * J[8]) *c3;		
		A(4,2+j) = J[5] * J[9] *c3;		
		A(5,2+j) = (J[10] * J[5] + J[6] * J[9]) *c3;	
		A(6,2+j) = (J[11] * J[5] + J[7] * J[9]) *c3;	
		A(7,2+j) = J[10] * J[6] *c3;	
		A(8,2+j) = (J[11] * J[6] + J[10] * J[7]) *c3;	
		A(9,2+j) = J[11] * J[7] *c3;	

		A(0,3+j) = (J[0] * J[0] - J[4] * J[4]) *c4;		
		A(1,3+j) = (2* J[0] * J[1] - 2* J[4] * J[5]) *c4;		
		A(2,3+j) = (2* J[0] * J[2] - 2* J[4] * J[6]) *c4;
		A(3,3+j) = (2* J[0] * J[3] - 2* J[4] * J[7]) *c4;		
		A(4,3+j) = (J[1] * J[1] - J[5] * J[5]) *c4;		
		A(5,3+j) = (2* J[1] * J[2] - 2* J[5] * J[6]) *c4;	
		A(6,3+j) = (2* J[1] * J[3] - 2* J[5] * J[7]) *c4;	
		A(7,3+j) = (J[2] * J[2] - J[6] * J[6]) *c4;	
		A(8,3+j) = (2* J[2] * J[3] - 2* J[6] * J[7]) *c4;	
		A(9,3+j) = (J[3] * J[3] - J[7] * J[7]) *c4;	

		A(0,4+j) = (J[0] * J[0] - J[8] * J[8]) *c5;		
		A(1,4+j) = (2* J[0] * J[1] - 2* J[8] * J[9]) *c5;		
		A(2,4+j) = (2* J[0] * J[2] - 2* J[8] * J[10]) *c5;
		A(3,4+j) = (2* J[0] * J[3] - 2* J[8] * J[11]) *c5;		
		A(4,4+j) = (J[1] * J[1] - J[9] * J[9]) *c5;		
		A(5,4+j) = (2* J[1] * J[2] - 2* J[9] * J[10]) *c5;	
		A(6,4+j) = (2* J[1] * J[3] - 2* J[9] * J[11]) *c5;	
		A(7,4+j) = (J[2] * J[2] - J[10] * J[10]) *c5;	
		A(8,4+j) = (2* J[2] * J[3] - 2* J[10] * J[11]) *c5;	
		A(9,4+j) = (J[3] * J[3] - J[11] * J[11]) *c5;	

		A(0,5+j) = (J[4] * J[4] - J[8] * J[8]) *c6;		
		A(1,5+j) = (2* J[4] * J[5] - 2* J[8] * J[9]) *c6;		
		A(2,5+j) = (2* J[4] * J[6] - 2* J[8] * J[10]) *c6;
		A(3,5+j) = (2* J[4] * J[7] - 2* J[8] * J[11]) *c6;		
		A(4,5+j) = (J[5] * J[5] - J[9] * J[9]) *c6;		
		A(5,5+j) = (2* J[5] * J[6] - 2* J[9] * J[10]) *c6;	
		A(6,5+j) = (2* J[5] * J[7] - 2* J[9] * J[11]) *c6;	
		A(7,5+j) = (J[6] * J[6] - J[10] * J[10]) *c6;	
		A(8,5+j) = (2* J[6] * J[7] - 2* J[10] * J[11]) *c6;	
		A(9,5+j) = (J[7] * J[7] - J[11] * J[11]) *c6;
	}

	return A;

}

// Calculates the RQ transformation for matrix A = R*Q.
template <typename T>
void UpgradeMetricADQ<T>::getRQ(CImg<double> A, CImg<double> &R_out, CImg<double> &Q_out){

	//ImgTool::printInfo(sout,A,"A");

	cimg_library::CImg<double> Reverse(3,3), Reverse2(4,4);
	cimg_library::CImg<double> R(3,4), Q(4,4), R2(4,3);

	// gsl vectors and matrices
	//gsl_matrix *QR, *Q_gsl, *R_gsl;
	//gsl_vector *QR_vec;
	//QR = gsl_matrix_alloc(4,3);
	//Q_gsl = gsl_matrix_alloc(4, 4);
	//R_gsl = gsl_matrix_alloc(4, 3);
    //QR_vec = gsl_vector_alloc(3);
	
	// reverse matrices
	Reverse.fill(0);
	Reverse(2,0)=1;
	Reverse(1,1)=1;
	Reverse(0,2)=1;
	Reverse2.fill(0);
	Reverse2(3,0)=1;
	Reverse2(2,1)=1;
	Reverse2(1,2)=1;
	Reverse2(0,3)=1;

	//ImgTool::printInfo(sout,A,"A=QR");
	
	A = Reverse * A;
	A.transpose();
	
	// Convert CImg A to Vigra Multiview
	double aVec[] = {A(0,0), A(0,1), A(0,2), A(0,3), A(1,0), 
		A(1,1), A(1,2), A(1,3), A(2,0), A(2,1), A(2,2), A(2,3)};
	double bVec[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	double cVec[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

	vigra::MultiArrayView< 2, double> a(vigra::MultiArrayShape<2>::type(4,3), aVec);

	// Vigra q, r Arrays
	vigra::MultiArrayView< 2, double> q(vigra::MultiArrayShape<2>::type(4,4), bVec);
	vigra::MultiArrayView< 2, double> r(vigra::MultiArrayShape<2>::type(4,3), cVec);

	//sout << "Calculate VIGRA QR" << std::endl;

	//sout << a << std::endl;

	// Calculate QR
	vigra::linalg::qrDecomposition(a, q, r) ;

	/*
	for(int i=0; i<3; i++) gsl_vector_set(QR_vec, i, 0);

	for(int i = 0; i<4; i++)
        for(int j = 0; j < 3; j++)
            gsl_matrix_set(QR, i, j, (double)A(j,i));
	
	gsl_linalg_QR_decomp (QR, QR_vec);

	gsl_linalg_QR_unpack (QR, QR_vec, Q_gsl, R_gsl);
			
	for(int i=0; i< 4; i++){
		for(int j=0; j< 4; j++){
			Q(j,i) = T(gsl_matrix_get(Q_gsl, i, j));
		}
	}
	for(int i=0; i< 4; i++){
		for(int j=0; j< 3; j++){
			R(j,i) = T(gsl_matrix_get(R_gsl, i, j));
		}
	}

	*/

	for(unsigned int i=0; i< 4; i++){
		for(unsigned int j=0; j< 4; j++){
			Q(j,i) = q(i,j);
		}
	}
	for(int i=0; i< 4; i++){
		for(int j=0; j< 3; j++){
			R(j,i) = r(i,j);
		}
	}
	
	R.transpose();
	
	for(int i=0; i<4; i++){
		for(int j=0; j<3; j++){
			R2(i,j) = R(3-i,2-j);
		}
	}

	R2.crop(1,0,3,2);

	Q.transpose();
	Q = Reverse2 * Q;
	Q.crop(0,1,3,3);

	//ImgTool::printInfo(sout,R2,"QR_R2");
	//ImgTool::printInfo(sout,R2*Q,"R*Q");

	R_out = R2;
	Q_out = Q;
	      
	//gsl_matrix_free(QR);
	//gsl_matrix_free(R_gsl);
	//gsl_matrix_free(Q_gsl);
	//gsl_vector_free(QR_vec);

}


#endif /* _UPGRADEMETRICADQ_HXX_ */



