/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file upgradeMetricTrackingMatrix.hxx
 *  Implementation of the parameter class upgradeMetricTrackingMatrix.
 *  \author Moritz Becker
 *  \date 19.08.2011
 */

#ifndef _UPGRADEMETRICTRACKINGMATRIX_HXX_
#define _UPGRADEMETRICTRACKINGMATRIX_HXX_

#include <hekate/UpgradeMetricTrackingMatrix.h>
#include <charon-utils/ImgTool.hxx>

template <typename T>
upgradeMetricTrackingMatrix<T>::upgradeMetricTrackingMatrix(const std::string& name) :
		TemplatedParameteredObject<T>(
				"upgradeMetricTrackingMatrix", name,
				"This plugin upgrades a tracking matrix with a given calibration."
			"Then the metric points in space and projective cameras can be retrieved."
		)
			,trackingMatrix(false, false) 
			,calibrationMatrix(false, false) 
			
			
{
	
	ParameteredObject::_addInputSlot(trackingMatrix, "trackingMatrix", 
	"trackingMatrix", 
	"CImgList<T>"); 
	ParameteredObject::_addInputSlot(calibrationMatrix, "calibrationMatrix", 
	"Calibration matrix which is used to perform an euclidean update.", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(metricTrackingMatrix, "metricTrackingMatrix", 
	"metricTrackingMatrix", 
	"CImg<T>"); 
	ParameteredObject::_addOutputSlot(cameraMatrices, "cameraMatrices", 
	"cameraMatrices", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(points3D, "points3D", 
	"points3D", 
	"vector<vector<T> >"); 
	
}

template <typename T>
void upgradeMetricTrackingMatrix<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	// Calculate euclidean upgrade ------------------------------

	cimg_library::CImg<T> upgradeMatrix(4,4);
	const cimg_library::CImg<T> & calibrationMatrix = this->calibrationMatrix()[0];
	ImgTool::printInfo(sout,calibrationMatrix,"CalibrationMatrix");

	const cimg_library::CImgList<T> & trackMatrix = this->trackingMatrix();

	/* 
	Read in tracking matrix ------------------------------------------------
	*/

	// Write uunormalized W to log
	sout << "Read in tracking matrix:" << std::endl;

	// Read in dimensions
	int dimM = trackMatrix.size();
	int dimN = trackMatrix[0].width();

	sout << "Number of different points = " << dimN << std::endl;
	sout << "Number of different frames =" << dimM << std::endl;

	// Create first row of the tracking matrix
	cimg_library::CImg<T> W = trackMatrix[0];

	// Append the missing rows to tracking matrix
	const char axis = 'y';
	for(int j=1; j<dimM; j++){
		W.append(trackMatrix[j],axis);
	}

	ImgTool::printInfo(sout,W,"W");

	// Normalize M*S so the homogeneous coordinate is equal 1 ------------------
	// to check whether the original correspondences are still correct
	cimg_library::CImg<T> Wnorm;
	Wnorm = W;
	for(int i=0; i<dimM; i++){
		for(int j=0; j<dimN; j++){
			Wnorm(j,i*3) /= Wnorm(j,i*3+2);
			Wnorm(j,i*3+1) /= Wnorm(j,i*3+2);
			Wnorm(j,i*3+2) /= Wnorm(j,i*3+2);		
		}
	}
	ImgTool::printInfo(sout,Wnorm,"Wnorm");
	
	// Get metric structure and motion matrices
	cimg_library::CImg<T> metricS, metricM;
	getEuclideanUpgrade(W, metricM, metricS, calibrationMatrix, upgradeMatrix);

	ImgTool::printInfo(sout,metricM,"metricM");
	ImgTool::printInfo(sout,metricS,"metricS");

	// Extract camera matrices ----------------------------------
	for(int i=0; i<dimM; i++){
		cimg_library::CImg<T> pm;
		pm = metricM.get_crop(0,i*3,3,i*3+2);
		// Push back transformed projective matrix to output slot
		this->cameraMatrices().push_back(pm);
	}
	
	// Extract 3D points ----------------------------------------
	std::vector<std::vector<T> > vectorPoints;
	for(int i=0; i<dimN; i++){
		
		// Back transform 3D points
		cimg_library::CImg<T> pt;
		pt = metricS.get_crop(i,0,i,3);
		pt /= pt[3];
		
		// Push back transformed 3D points
		std::vector<T> point;
		point.push_back(pt[0]);
		point.push_back(pt[1]);
		point.push_back(pt[2]);
		point.push_back(pt[3]);
		vectorPoints.push_back(point);
		
	}
	this->points3D() = vectorPoints;


}


/// Perform a metrix update, if calibration is already known and given by the calibration input slot.
template <typename T>
void getEuclideanUpgrade(cimg_library::CImg<T> W, cimg_library::CImg<T> & M, cimg_library::CImg<T> & S, const cimg_library::CImg<T> calibration, cimg_library::CImg<T> upgradeMatrix){


	// Normalize tracking matrix by calibration
	sout << "Normalize tracking matrix by calibration matrix." << std::endl;
	normalizeTrackingMatrixByCalibration(W, calibration);

	// Retrieve projective motion and structure matrices.
	cimg_library::CImg<T> U, D, V;
	W.SVD(U,D,V);
	D.crop(0,0,0,3).diagonal().sqrt();
	M = U.get_crop(0,0,3,U.height()-1) * D;
	S = D * V.get_crop(0,0,3,V.height()-1).transpose();

	// Calculate upgrade matrix from tracking matrix
	sout << "Calculate metric upgrade matrix from tracking matrix." << std::endl;
	getUpgradeMatrix(M, upgradeMatrix);

	M = M*upgradeMatrix;
	S = upgradeMatrix.get_invert()*S;

}


/// Normalizes tracking matrix with calibration matrix
template <typename T>
void normalizeTrackingMatrixByCalibration(cimg_library::CImg<T> & W, const cimg_library::CImg<T> & calibrationMatrix){

	// Read in dimensions
	int dimN = W.width();
	int dimM = W.height();

	cimg_library::CImg<T> W2;

	const char direction = 'y';

	// Iterate the following code for every projective matrix
	for(int i=0; i<dimM; i=i+3){

		// Perform normalization image wise
		W2.append(calibrationMatrix.get_invert()*W.get_crop(0,i,dimN-1,i+2),direction);

	}

	// Save results in tracking matrix
	W = W2;

}



/// Calculate metric upgrade matrix from tracking matrix
template <typename T>
void getUpgradeMatrix(const cimg_library::CImg<T> & M, cimg_library::CImg<T> & upgradeMatrix){


	// Create A*Q=0 where Q = HH^T an H the euclidean update
	cimg_library::CImg<T> A;
	const char direction = 'y';
	for(int i = 0; i < M.height(); i=i+3){

		cimg_library::CImg<T> Mi, Ai(10,5);
		Mi = M.get_crop(0,i,3,i+2);
		
		// N(1,2)==0
		Ai(0,0) = Mi[0]*Mi[4];
		Ai(1,0) = Mi[0]*Mi[5]+Mi[1]*Mi[4];
		Ai(2,0) = Mi[0]*Mi[6]+Mi[2]*Mi[4];
		Ai(3,0) = Mi[0]*Mi[7]+Mi[3]*Mi[4];
		Ai(4,0) = Mi[1]*Mi[5];
		Ai(5,0) = Mi[1]*Mi[6]+Mi[2]*Mi[5];
		Ai(6,0) = Mi[1]*Mi[7]+Mi[3]*Mi[5];
		Ai(7,0) = Mi[2]*Mi[6];
		Ai(8,0) = Mi[2]*Mi[7]+Mi[3]*Mi[6];
		Ai(9,0) = Mi[3]*Mi[7];
		
		// N(1,3)==0
		Ai(0,1) = Mi[0]*Mi[8];
		Ai(1,1) = Mi[1]*Mi[8]+Mi[0]*Mi[9];
		Ai(2,1) = Mi[2]*Mi[8]+Mi[0]*Mi[10];
		Ai(3,1) = Mi[3]*Mi[8]+Mi[0]*Mi[11];
		Ai(4,1) = Mi[1]*Mi[9];
		Ai(5,1) = Mi[1]*Mi[10]+Mi[2]*Mi[9];
		Ai(6,1) = Mi[1]*Mi[11]+Mi[3]*Mi[9];
		Ai(7,1) = Mi[2]*Mi[10];
		Ai(8,1) = Mi[2]*Mi[11]+Mi[3]*Mi[10];
		Ai(9,1) = Mi[3]*Mi[11];
		
		// N(2,3)==0
		Ai(0,2) = Mi[4]*Mi[8];
		Ai(1,2) = Mi[4]*Mi[9]+Mi[5]*Mi[8];
		Ai(2,2) = Mi[4]*Mi[10]+Mi[6]*Mi[8];
		Ai(3,2) = Mi[4]*Mi[11]+Mi[7]*Mi[8];
		Ai(4,2) = Mi[5]*Mi[9];
		Ai(5,2) = Mi[6]*Mi[9]+Mi[5]*Mi[10];
		Ai(6,2) = Mi[7]*Mi[9]+Mi[5]*Mi[11];
		Ai(7,2) = Mi[6]*Mi[10];
		Ai(8,2) = Mi[7]*Mi[10]+Mi[6]*Mi[11];
		Ai(9,2) = Mi[7]*Mi[11];
		
		// N(1,1)-N(2,2)==0
		Ai(0,3) = Mi[0]*Mi[0]-Mi[4]*Mi[4];
		Ai(1,3) = 2*Mi[0]*Mi[1]-2*Mi[4]*Mi[5];
		Ai(2,3) = 2*Mi[0]*Mi[2]-2*Mi[4]*Mi[6];
		Ai(3,3) = 2*Mi[0]*Mi[3]-2*Mi[4]*Mi[7];
		Ai(4,3) = Mi[1]*Mi[1]-Mi[5]*Mi[5];
		Ai(5,3) = 2*Mi[1]*Mi[2]-2*Mi[5]*Mi[6];
		Ai(6,3) = 2*Mi[1]*Mi[3]-2*Mi[5]*Mi[7];
		Ai(7,3) = Mi[2]*Mi[2]-Mi[6]*Mi[6];
		Ai(8,3) = 2*Mi[2]*Mi[3]-2*Mi[6]*Mi[7];
		Ai(9,3) = Mi[3]*Mi[3]-Mi[7]*Mi[7];
		
		// N(1,1)-N(3,3)==0
		Ai(0,4) = Mi[0]*Mi[0]-Mi[8]*Mi[8];
		Ai(1,4) = 2*Mi[0]*Mi[1]-2*Mi[8]*Mi[9];
		Ai(2,4) = 2*Mi[0]*Mi[2]-2*Mi[8]*Mi[10];
		Ai(3,4) = 2*Mi[0]*Mi[3]-2*Mi[8]*Mi[11];
		Ai(4,4) = Mi[1]*Mi[1]-Mi[9]*Mi[9];
		Ai(5,4) = 2*Mi[1]*Mi[2]-2*Mi[9]*Mi[10];
		Ai(6,4) = 2*Mi[1]*Mi[3]-2*Mi[9]*Mi[11];
		Ai(7,4) = Mi[2]*Mi[2]-Mi[10]*Mi[10];
		Ai(8,4) = 2*Mi[2]*Mi[3]-2*Mi[10]*Mi[11];
		Ai(9,4) = Mi[3]*Mi[3]-Mi[11]*Mi[11];

		A.append(Ai,direction);
		
	}

	// Calculate Q where A*Q = 0
	cimg_library::CImg<T> U2, D2, V2;
	A.SVD(U2,D2,V2);
	ImgTool::printInfo(sout,A,"A");
	cimg_library::CImg<T> Q, Q2(4,4);
	Q = V2.get_crop(9,0,9,9);
	
	// Write A*Q to log
	for(int i=0; i<A.height(); i++){
		sout << (A.get_crop(0,i,9,i)*Q)[0] << std::endl;
	}

	// Realign Q as a symmetric matrix
	Q2[0]=Q[0];		Q2[1]=Q[1];		Q2[2]=Q[2];		Q2[3]=Q[3];
	Q2[4]=Q[1];		Q2[5]=Q[4];		Q2[6]=Q[5];		Q2[7]=Q[6];
	Q2[8]=Q[2];		Q2[9]=Q[5];		Q2[10]=Q[7];	Q2[11]=Q[8];
	Q2[12]=Q[3];	Q2[13]=Q[6];	Q2[14]=Q[8];	Q2[15]=Q[9];

	ImgTool::printInfo(sout,Q2/Q2[0],"Q2");

	// Calculate H where Q = H*H^T ---------------------------------------

	cimg_library::CImg<T> U3, D3, V3, H3(3,4), H4(1,4), H;
	H4.fill(T(0));
	H4[3]=T(1);
	Q2.SVD(U3,D3,V3);
	D3 = D3.crop(0,0,0,2).diagonal().sqrt();
	H3 = U3.get_crop(0,0,2,3) * D3;

	// Write calculated upgrade matrix to log
	ImgTool::printInfo(sout,H3,"H3");
	ImgTool::printInfo(sout,H3*H3.get_transpose()/(H3*H3.get_transpose())[0],"H3H3^T");

	H.append(H3).append(H4);

	ImgTool::printInfo(sout,H,"H");

	for(int i = 0; i < M.height(); i=i+3){
		cimg_library::CImg<T> Mi;
		Mi = M.get_crop(0,i,3,i+2);
		ImgTool::printInfo(sout,Mi*Q2*Mi.get_transpose()/(Mi*Q2*Mi.get_transpose())[0],"MQM^T");
	}

	// Save upgrade matrix
	upgradeMatrix.clear();
	upgradeMatrix = H;

}


#endif /* _UPGRADEMETRICTRACKINGMATRIX_HXX_ */



