/*  Copyright (C) 2010 Julian Coordts

 This file is part of Charon.

 Charon is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Charon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with Charon.  If not, see <http://www.gnu.org/licenses/>.
 */
/** \file WarpBB.hxx
 *  Implementation of the parameter class WarpBB.
 *  \author Julian Coordts
 *  \date 08.03.2010
 */

#ifndef _WarpBB_HXX_
#define _WarpBB_HXX_

#include <hekate/WarpBB.h>

template<typename T>
WarpBB<T>::WarpBB(const std::string& name) :
	TemplatedParameteredObject<T> ("WarpBB", name, "Warps an image") {
	ParameteredObject::_addInputSlot(input_images, "input_images",
			"input_images (first left, second right)", "CImgList<T>");
	ParameteredObject::_addInputSlot(interpolator, "interpolator",
			"Interpolator", "Interpolator<T>*");
	ParameteredObject::_addInputSlot(warp_matrices, "warp_matrices",
			"warp matrices (first left, second right)", "CImgList<T>");
	ParameteredObject::_addInputSlot(input_feature_corres, "input_feature_corres",
			"unwarped feature correspondences", "std::vector<std::vector<std::pair<T, T> > >");

	ParameteredObject::_addParameter (quick_warp, "quick_warp", "use nearest neighbor interpolation", true);

	ParameteredObject::_addOutputSlot( warped_images, "warped_images",
			"Warped Images (first left, second right)", "CImgList<T>");
	ParameteredObject::_addOutputSlot(warped_feature_corres, "warped_feature_corres",
			"warped feature correspondences", "std::vector<std::vector<std::pair<T, T> > >");
called= false;
}


template<typename T>
void WarpBB<T>::execute() {
	ParameteredObject::execute();

	if (called==true)
	{
		sout << "\tWarpBB already called, skip calculating..." << std::endl;
		return;
	}
	else called = true;




//Todo: Could handle just 2 dimensions atm.

	int depthIm = this->input_images()[0].depth();
	int spectrumIm = this->input_images()[0].spectrum();

    cimg_library::CImg<T> Hl = this->warp_matrices()[0];
    cimg_library::CImg<T> Hr = this->warp_matrices()[1];

    cimg_library::CImg<T> imageInputLeft= this->input_images()[0];
    cimg_library::CImg<T> imageInputRight= this->input_images()[1];


	 //Calculate BBs and offset
	 cimg_library::CImg<T> BBl(2,2);
	 cimg_library::CImg<T> BBr(2,2);
	 BBl =calcBB(Hl,imageInputLeft);
	 BBr =calcBB(Hr,imageInputRight);
	 cimg_library::CImg<T> boundingBox(2,2);
	 std::pair<T,T> offsetMax;
	 boundingBox(0,0)= std::min(BBl(0,0),BBr(0,0));
	 boundingBox(0,1)= std::min(BBl(0,1),BBr(0,1));
	 boundingBox(1,0)= std::max(BBl(1,0),BBr(1,0));
	 boundingBox(1,1)= std::max(BBl(1,1),BBr(1,1));



	int widthBB =  boundingBox(1,0) - boundingBox(0,0);
	int heightBB = boundingBox(1,1) - boundingBox(0,1);


    this->warped_images().assign(2, widthBB, heightBB, depthIm, spectrumIm);


    cimg_library::CImg<T> imageOutputLeft(widthBB, heightBB,depthIm,spectrumIm);
    cimg_library::CImg<T> imageOutputRight(widthBB, heightBB,depthIm,spectrumIm);

    //Warp left image
    warp(imageInputLeft,
    	imageOutputLeft,
    	 Hl,boundingBox);
    //Warp right image
    warp(imageInputRight,
    	imageOutputRight,
    	 Hr,boundingBox);
    this->warped_images()[0]= imageOutputLeft;
    this->warped_images()[1]= imageOutputRight;


    warpCorrespondences(boundingBox);




}

template<typename T>
void WarpBB<T>::warpCorrespondences(cimg_library::CImg<T>& boundingBox)
{
    double offsetX = boundingBox(0,0);
    double offsetY = boundingBox(0,1);
    cimg_library::CImg<T> Hl = this->warp_matrices()[0];
    cimg_library::CImg<T> Hr = this->warp_matrices()[1];

		sout << this->input_feature_corres().size() << " feature vectors" << std::endl;
		sout << "create separated matrices of correspondences for each image...." << std::endl;
		int correspondences_count = std::min(this->input_feature_corres()[0].size(),
											 this->input_feature_corres()[1].size());

		sout << correspondences_count << " pairs of correspondences found" << std::endl;

		cimg_library::CImg<T> ml(correspondences_count,2,1,1,0);
		cimg_library::CImg<T> mr(correspondences_count,2,1,1,0);
	//	sout << "spreading correspondences..." << std::endl;

		for (int i=0; i< correspondences_count;i++)
		{

				ml(i,0)=this->input_feature_corres()[0][i].first;
				ml(i,1)=this->input_feature_corres()[0][i].second;
				mr(i,0)=this->input_feature_corres()[1][i].first;
				mr(i,1)=this->input_feature_corres()[1][i].second;

		}
//		soutCImgMatrix(ml,"Matchpoints left(x,y)");
//		soutCImgMatrix(mr,"Matchpoints right(x,y)");

    //Warp correspondence points:


    ml.resize(-100, 3, -100,-100,-1);
    mr.resize(-100, 3, -100,-100,-1);


    for (int i=0; i< ml.width(); ++i)
    {
    	ml(i,2)=1;
    	mr(i,2)=1;
    }

//    soutCImgMatrix(ml,"ml");
//    soutCImgMatrix(mr,"mr");

    //warped_feature_correspondences
    cimg_library::CImg<T> wml(ml.width(),ml.height());
    cimg_library::CImg<T> wmr(mr.width(),mr.height());

    wml=Hl*ml;
    wmr=Hr*mr;



    this->warped_feature_corres().resize(2);
    for (int i=0; i< wml.width(); ++i)
    {
    	wml(i,0)=wml(i,0)/wml(i,2);
    	wml(i,1)=wml(i,1)/wml(i,2);
    	wmr(i,0)=wmr(i,0)/wmr(i,2);
    	wmr(i,1)=wmr(i,1)/wmr(i,2);
    	std::pair<T,T> pl,pr;
    	pl.first=wml(i,0)- offsetX;
    	pl.second=wml(i,1)- offsetY;
    	pr.first=wmr(i,0)- offsetX;
    	pr.second=wmr(i,1)- offsetY;
    	  this->warped_feature_corres()[0].push_back(pl);
    	  this->warped_feature_corres()[1].push_back(pr);
    }

//    soutCImgMatrix(wml,"wml");
//    soutCImgMatrix(wmr,"wmr");

}


template<typename T>
void WarpBB<T>::warp(cimg_library::CImg<T>& inputImage,
					 cimg_library::CImg<T>& outputImage,
					 cimg_library::CImg<T>& warpMatrix,
					 cimg_library::CImg<T>& boundingBox)
{
	cimg_library::CImg<T> warpMatrixInv(3,3);
    warpMatrixInv= warpMatrix.get_invert();

	double offsetX = boundingBox(0,0);
    double offsetY = boundingBox(0,1);
    int heightIn = inputImage.height();

	cimg_library::CImg<T> oldPoint(1,3);
	cimg_library::CImg<T> newPoint(1,3);

	T x_new;
	T y_new;
	T z_new;

	//int x, y, z, c;
	int xpos, ypos;
	int colors = outputImage.spectrum();

	cimg_forXYZ(outputImage, x, y, z) {
		oldPoint (0,0)= x + offsetX;
		oldPoint (0,1)= y + offsetY;
		oldPoint (0,2) = 1;
		newPoint = warpMatrixInv * oldPoint;
		newPoint(0,0) = newPoint(0,0)/newPoint(0,2);
		newPoint(0,1) = newPoint(0,1)/newPoint(0,2);

		x_new = newPoint(0,0);
		y_new = newPoint(0,1);
		z_new = z;
		int widthIn = inputImage.width();
		int heigthIn = inputImage.height();
		
		if ((0<x_new) && (x_new<widthIn) && (0<y_new) && (y_new<heightIn)) // if in input image
		{
			if (quick_warp)
			{
				xpos=static_cast<int>(x_new+0.5f);
				ypos=static_cast<int>(y_new+0.5f);
				//Check if casting damaged boundaries
				if (xpos >= widthIn) xpos = widthIn - 1;
				if (xpos < 0) xpos = 0;
				if (ypos >= heigthIn) ypos =  heigthIn - 1;
				if (ypos < 0) ypos =  0;
				for (int i = 0; i < colors; i++)
				{
					outputImage(x, y, z, i) = inputImage(xpos,ypos, (int) z_new, i);
				}
			}
			else
			{
				for (int i = 0; i < colors; i++)
				{
					outputImage(x, y, z, i) = this->interpolator()->interpolate(inputImage, (float) x_new, (float) y_new, (float) z_new, i);
				}
			}
		}
		else 
		{
			for (int i = 0; i < colors; i++)
			{
				outputImage(x, y, z, i) = 0;
			}
		}

	}	// cimg_forXYZ


}



template <typename T>
cimg_library::CImg<T> WarpBB<T>::calcBB(cimg_library::CImg<T>&H,
										cimg_library::CImg<T>&Image)
{
//Calc Bounding Boxes
//Notice: 	For the calculation of H*v the upper left point of v is(1,1) not (0,0)
//			and the right bottom (width,heigth)!!!!

	cimg_library::CImg<T> oldPoint(1,3);
	cimg_library::CImg<T> newPoint(1,3);

	//Calc Bounding Boxes left
	int xmax=0;
	int ymax=0;
	int xmin=0;
	int ymin=0;

	//Transform Point(0,0)
	oldPoint (0,0)= 0; oldPoint (0,1)= 0; oldPoint (0,2)= 1;
	newPoint= H* oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
	if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
	if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
	if (newPoint(0,1) > ymax) ymax = newPoint(0,1);


	//Transform Point(0,Image.height()-1)
	oldPoint (0,0)= 0; oldPoint (0,1)= Image.height(); oldPoint (0,2)= 1;
	newPoint= H* oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
	if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
	if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
	if (newPoint(0,1) > ymax) ymax = newPoint(0,1);

	//Transform Point(Image.width()-1,0)
	oldPoint (0,0)= Image.width(); oldPoint (0,1)= 0; oldPoint (0,2)= 1;
	newPoint= H* oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
	if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
	if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
	if (newPoint(0,1) > ymax) ymax = newPoint(0,1);

	//Transform Point(Image.width()-1,Image.height()-1)
	oldPoint (0,0)= Image.width(); oldPoint (0,1)= Image.height(); oldPoint (0,2)= 1;
	newPoint= H* oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	if (newPoint(0,0) < xmin) xmin = newPoint(0,0);
	if (newPoint(0,1) < ymin) ymin = newPoint(0,1);
	if (newPoint(0,0) > xmax) xmax = newPoint(0,0);
	if (newPoint(0,1) > ymax) ymax = newPoint(0,1);


	sout << "BB: [" << xmin  <<", "   << ymin << "][" << xmax << ", " << ymax << "]" << std::endl;

	cimg_library::CImg<T> res(2,2);
	res(0,0)=xmin;res(1,0)=xmax;
	res(0,1)=ymin;res(1,1)=ymax;
	return res;
}



template<typename T>
void WarpBB<T>::soutCImgMatrix(cimg_library::CImg<T> &m,const char* text = "")
{
	/*
	 * outputs a CImg in human readable matrix format through sout
	 */

	sout << "Output Matrix: " << text << " ("  << m.height() <<  "x" << m.width() << ")" << std::endl;
	for (int y=0;y<m.height();y++)
	{
		sout << "\t";
		for (int x=0;x<m.width();x++)
		{
			sout.width(5);
			sout << m(x,y) << "  ";
		}
		sout << std::endl;
	}
}



#endif /* _WarpBB_HXX_ */

