/*  Copyright (C) 2010 Julian Coordts

 This file is part of Charon.

 Charon is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Charon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with Charon.  If not, see <http://www.gnu.org/licenses/>.

 This File is based on WarpBB.hxx by Julian Coordts.

 */
/** \file WarpBBinv.hxx
 *  Implementation of the parameter class WarpBBinv.
 *  \author Mirko Schmidt
 *  \date 08.03.2010
 */

#ifndef _WarpBBinv_HXX_
#define _WarpBBinv_HXX_

#include <hekate/WarpBBinv.h>
#include <algorithm>

template<typename T>
WarpBBinv<T>::WarpBBinv(const std::string& name) :
	TemplatedParameteredObject<T> ("WarpBBinv", name, "Warps an image") {
	ParameteredObject::_addInputSlot(input_images, "input_images",
			"input_images (first left, second right)", "CImgList<T>");
	ParameteredObject::_addInputSlot(interpolator, "interpolator",
			"Interpolator", "Interpolator<T>*");
	ParameteredObject::_addInputSlot(warp_matrices, "warp_matrices",
			"warp matrices (first left, second right)", "CImgList<T>");

	ParameteredObject::_addParameter (quick_warp, "quick_warp", "use nearest neighbor interpolation", true);

	ParameteredObject::_addOutputSlot( warpedinv_images, "warpedinv_images",
			"Warped Images (first left, second right)", "CImgList<T>");

	called= false;
}


template<typename T>
void WarpBBinv<T>::execute() {
	ParameteredObject::execute();

	if (called==true)
	{
		sout << "\tWarpBBinv already called, skipping..." << std::endl;
		return;
	}
	else called = true;


	int depthIm = this->input_images()[0].depth();
	int spectrumIm = this->input_images()[0].spectrum();

    cimg_library::CImg<T> Hl = this->warp_matrices()[0];
    cimg_library::CImg<T> Hr = this->warp_matrices()[1];

    cimg_library::CImg<T> imageInputLeft= this->input_images()[0];
    cimg_library::CImg<T> imageInputRight= this->input_images()[1];


	 //Calculate BBs and offset
	 cimg_library::CImg<T> BBl(2,2);
	 cimg_library::CImg<T> BBr(2,2);
	 BBl =calcBBinv(Hl,imageInputLeft);
	 BBr =calcBBinv(Hr,imageInputRight);
	 cimg_library::CImg<T> boundingBox(2,2);
	 std::pair<T,T> offsetMax;
	 boundingBox(0,0)= std::min(BBl(0,0),BBr(0,0));
	 boundingBox(0,1)= std::min(BBl(0,1),BBr(0,1));
	 boundingBox(1,0)= std::max(BBl(1,0),BBr(1,0));
	 boundingBox(1,1)= std::max(BBl(1,1),BBr(1,1));



	int widthBB =  boundingBox(1,0) - boundingBox(0,0);
	int heightBB = boundingBox(1,1)- boundingBox(0,1);


    this->warpedinv_images().assign(2, widthBB, heightBB,depthIm,spectrumIm);


    cimg_library::CImg<T> imageOutputLeft(widthBB, heightBB,depthIm,spectrumIm);
    cimg_library::CImg<T> imageOutputRight(widthBB, heightBB,depthIm,spectrumIm);

    //Warp left image
    warp(imageInputLeft,
    	imageOutputLeft,
    	 Hl,boundingBox);
    //Warp right image
    warp(imageInputRight,
    	imageOutputRight,
    	 Hr,boundingBox);
    this->warpedinv_images()[0]= imageOutputLeft;
    this->warpedinv_images()[1]= imageOutputRight;


//    warpCorrespondences(boundingBox);




}



template<typename T>
void WarpBBinv<T>::warp(cimg_library::CImg<T>& inputImage,
					 cimg_library::CImg<T>& outputImage,
					 cimg_library::CImg<T>& warpMatrix,
					 cimg_library::CImg<T>& boundingBox)
{
    double offsetX = boundingBox(0,0);
    double offsetY = boundingBox(0,1);
    int heightIn = inputImage.height();

	cimg_library::CImg<T> oldPoint(1,3);
	cimg_library::CImg<T> newPoint(1,3);

	T x_new;
	T y_new;
	T z_new;

	//int x, y, z, c;
	int xpos, ypos;

	//#define cimg_forX(img,x) cimg_for1((img)._width,x)
	//#define cimg_forY(img,y) cimg_for1((img)._height,y)
	//#define cimg_forZ(img,z) cimg_for1((img)._depth,z)
	//#define cimg_forC(img,c) cimg_for1((img)._spectrum,c)

	int colors = outputImage.spectrum();

	cimg_forXYZ(outputImage, x, y, z)
	{
		oldPoint (0,0)= x + offsetX; oldPoint (0,1)= y + offsetY; oldPoint (0,2) = 1;
		newPoint = warpMatrix * oldPoint;
		newPoint(0,0) = newPoint(0,0)/newPoint(0,2);
		newPoint(0,1) = newPoint(0,1)/newPoint(0,2);


		x_new = newPoint(0,0);
		y_new = newPoint(0,1);
		z_new = z;
		int widthIn = inputImage.width();
		int heigthIn = inputImage.height();
		
		if ( (0<x_new) && (x_new<widthIn) && (0<y_new) && (y_new<heightIn)) // if in input image
		{
			if (quick_warp)
			{
				xpos=static_cast<int>(x_new+0.5f);
				ypos=static_cast<int>(y_new+0.5f);
				//Check if casting damaged boundaries
				if (xpos >= widthIn) xpos = widthIn - 1;
				if (xpos < 0) xpos = 0;
				if (ypos >= heigthIn) ypos =  heigthIn - 1;
				if (ypos < 0) ypos =  0;
				for (int i = 0; i < colors; i++)
				{
					outputImage(x, y, z, i) = inputImage(xpos,ypos, (int) z_new, i);
				}
			}
			else
			{
				for (int i = 0; i < colors; i++)
				{
					outputImage(x, y, z, i) = this->interpolator()->interpolate(inputImage, (float) x_new, (float) y_new, (float) z_new, i);
				}
			}
		}
		else 
		{
			for (int i = 0; i < colors; i++)
			{
				outputImage(x, y, z, i) = 0;
			}
		}

	}	// cimg_forXYZ


}



template <typename T>
cimg_library::CImg<T> WarpBBinv<T>::calcBBinv(cimg_library::CImg<T>&H, cimg_library::CImg<T>&Image)
{

	// invert old H matrix
	cimg_library::CImg<T> Hinv(3,3);
    Hinv = H.get_invert();

	// calculate "inner box"
	int x_coords[4];
	int y_coords[4];
	int x_min, x_max, y_min, y_max; // corner points of inner box

	//Notice: 	For the calculation of H*v the upper left point of v is(1,1) not (0,0)
	//			and the right bottom (width,heigth)!!!!

	cimg_library::CImg<T> oldPoint(1,3);
	cimg_library::CImg<T> newPoint(1,3);

	//Transform Point(0,0)
	oldPoint (0,0)= 0; oldPoint (0,1)= 0; oldPoint (0,2)= 1;
	newPoint= Hinv * oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	x_coords[0] = newPoint(0,0);
	y_coords[0] = newPoint(0,1);
		
	//Transform Point(0,Image.height()-1)
	oldPoint (0,0)= 0; oldPoint (0,1)= Image.height(); oldPoint (0,2)= 1;
	newPoint= Hinv * oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	x_coords[1] = newPoint(0,0);
	y_coords[1] = newPoint(0,1);

	//Transform Point(Image.width()-1,0)
	oldPoint (0,0)= Image.width(); oldPoint (0,1)= 0; oldPoint (0,2)= 1;
	newPoint= Hinv * oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	x_coords[2] = newPoint(0,0);
	y_coords[2] = newPoint(0,1);

	//Transform Point(Image.width()-1,Image.height()-1)
	oldPoint (0,0)= Image.width(); oldPoint (0,1)= Image.height(); oldPoint (0,2)= 1;
	newPoint= Hinv * oldPoint;
	newPoint(0,0)= newPoint(0,0)/newPoint(0,2);
	newPoint(0,1)= newPoint(0,1)/newPoint(0,2);

	x_coords[3] = newPoint(0,0);
	y_coords[3] = newPoint(0,1);

	std::sort(x_coords, x_coords + 4);
	std::sort(y_coords, y_coords + 4);

	// core is inner box; outer box is bounding box
	x_min = x_coords[1];
	x_max = x_coords[2];
	y_min = y_coords[1];
	y_max = y_coords[2];

	sout << "BBinv: [" << x_min  <<", "   << y_min << "][" << x_max << ", " << y_max << "]" << std::endl;

	cimg_library::CImg<T> res(2,2);
	res(0,0)=x_min; res(1,0)=x_max;
	res(0,1)=y_min; res(1,1)=y_max;
	return res;
}



template<typename T>
void WarpBBinv<T>::soutCImgMatrix(cimg_library::CImg<T> &m,const char* text = "")
{
	/*
	 * outputs a CImg in human readable matrix format through sout
	 */

	sout << "Output Matrix: " << text << " ("  << m.height() <<  "x" << m.width() << ")" << std::endl;
	for (int y=0;y<m.height();y++)
	{
		sout << "\t";
		for (int x=0;x<m.width();x++)
		{
			sout.width(5);
			sout << m(x,y) << "  ";
		}
		sout << std::endl;
	}
}



#endif /* _WarpBBinv_HXX_ */

