/*  Copyright (C) 2009 Julian Coordts

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file correspondencepainter.hxx
 *  Implementation of the parameter class correspondencepainter.
 *  \author Julian Coordts
 *  \date 30.03.2010
 */

#ifndef _CORRESPONDENCEPAINTER_HXX_
#define _CORRESPONDENCEPAINTER_HXX_

#include <hekate/features/CorrespondencePainter.h>
#include <hekate/Correspondence.hxx>

using namespace hekate;

template <typename T>
correspondencepainter<T>::correspondencepainter(const std::string& name) :
		TemplatedParameteredObject<T>("correspondencepainter", name,
			"Combines two images into one and draws lines between their correspondence points")
			,imagePath(true, false)
			,imageSequence(true, false)
{
	
	
	ParameteredObject::_addInputSlot(imagePath, "imagePath", "Vector list of paths to the images", "vector<string>");
	ParameteredObject::_addInputSlot(imageSequence, "imageSequence", "CImgList", "CImgList<T>");
	ParameteredObject::_addInputSlot(correspondences, "correspondences", "correspondences", "CorrespondenceList");
	ParameteredObject::_addOutputSlot(image_result, "image_result", "the drawn image", "CImgList<T>");
	
	ParameteredObject::_addParameter (withIndices, "withIndices", 
	"Draw index to every feature correspondence.", false);
	ParameteredObject::_addParameter (withOutliers, "withOutliers", 
	"Also draw correspondences known as outliers.", false);
	ParameteredObject::_addParameter (drawAll, "drawAll", 
	"Draw all correspondences into one image", true);
	ParameteredObject::_addParameter (numberOfImages, "numberOfImages", 
	"Number of maximal images (0 = all)", 0);	
	ParameteredObject::_addParameter (stepSize, "stepSize", 
	"step size", 1);	
	ParameteredObject::_addParameter (width, "width", 
	"image width", 1920);	
	ParameteredObject::_addParameter (height, "height", 
	"image height", 1080);	
	ParameteredObject::_addParameter (drawPointColors, "drawPointColors", 
	"Draw colors of correspondences into one image", false);

	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void correspondencepainter<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	
	const CorrespondenceList<T> &list = this->correspondences();
	size_t nImg = list.sizeImages();
	const size_t nPointsTotal = list.size();	

	size_t nImgPath = 0;
	if(this->imagePath.connected())
		nImgPath = this->imagePath().size();
	else if(this->imageSequence.connected())
		nImgPath = this->imageSequence().size();
	else
		sout << "No images connected!" << std::endl;
	
	sout << "Drawing " << nPointsTotal << " features to " << nImg << " output images..." << std::endl;


	if(nImgPath != nImg){
		sout << "CORRESPONDENCE_PAINTER: " << nImgPath << 
			" images found." << std::endl;
		sout << "CORRESPONDENCE_PAINTER: " << nImg << 
			" imgIndices found." << std::endl;	
	}


	int stepSize = 1;
	if( this->stepSize() > 1 && this->stepSize() < nImg )
		stepSize = this->stepSize() ;
	
	for(size_t j=0; j<nImg; j=j+stepSize){

		time_t tstart, tend;	
		tstart = time(0);	

		sout << "Draw image " << j << std::endl;

		// Create image to paint the correspondences to ---------------------------------------
		// Declare image
		cimg_library::CImg<T> image(width(), height(), 1, 1, T(0));
		if(this->imagePath.connected())
		{
			try
			{
				image.assign( this->imagePath()[j].c_str() );
				//out().load(filename().c_str());
			}
			catch(const cimg_library::CImgException& err)
			{
				throw std::runtime_error(
					this->getClassName() + " instance \"" +
					this->getName() + "\" Could not read file\n\t" + err.what());
			}
			
		}
		else if( this->imageSequence.connected() )
		{
			image = this->imageSequence()[j];
			sout << "Took image from image sequence" << std::endl;
		}
	
		image.resize(image.width(),image.height(),1,3);
		if(image.max() > 255.){image = image >> 4;}

		const unsigned char red[] = { 255,0,0 } ;
		const unsigned char blue[] = { 70,100,255 } ;
		const unsigned char yellow[] = { 230,230,0 } ;
		const unsigned char white[] = { 255,255,255 } ;

		sout << "Painting " << this->correspondences().sizePoints(j) << " features"  << std::endl;

		auto it = list.get<imgIndex>().lower_bound(j) ;
		auto itEnd = list.get<imgIndex>().upper_bound(j) ;
		for(; it != itEnd ; it++)
		{	
						
			if(drawPointColors()){
				const unsigned char color[] = { it->R, it->G, it->B } ;
				image.draw_point(it->x(),it->y(),color);
				continue;
			}

			char txt[32];
			std::sprintf(txt,"%lu",it->pointIndex);

			// Get list of all points with that point id
			std::vector<const Correspondence<T>*> vec = 
				list.getListByPointId(it->pointIndex);

			for(size_t i=0; i<vec.size(); i++)
			{

				// If point belongs to current image
				if(vec[i]->imgIndex == j)
				{
					// draw triangle to feature coordinates
					if(!vec[i]->outlier)
						image.draw_triangle(vec[i]->x(),vec[i]->y(),vec[i]->x()+3,
							vec[i]->y()+3,vec[i]->x()-3,vec[i]->y()+3,blue);
					else if(this->withOutliers())
						image.draw_triangle(vec[i]->x(),vec[i]->y(),vec[i]->x()+3,
							vec[i]->y()+3,vec[i]->x()-3,vec[i]->y()+3,red);
					// Write index next to feature triangle
					if(this->withIndices())
						image.draw_text(vec[i]->x(),vec[i]->y()+4, txt,red,white);
				}
				else if(this->drawAll()) 
				{
					if(!vec[i]->outlier)
					{
						image.draw_triangle(vec[i]->x(),vec[i]->y(),vec[i]->x()+3,
							vec[i]->y()+3,vec[i]->x()-3,vec[i]->y()+3,yellow);
					}
					else if(this->withOutliers())
					{
						image.draw_triangle(vec[i]->x(),vec[i]->y(),vec[i]->x()+3,
							vec[i]->y()+3,vec[i]->x()-3,vec[i]->y()+3,red);
					}
				}
				// If point is the subsequent point to the prior one in this loop
				if( i > 0  &&  vec[i]->imgIndex == vec[i-1]->imgIndex+1 && this->drawAll())
				{
					if(!vec[i]->outlier && !vec[i-1]->outlier)
						image.draw_line(vec[i-1]->x(),vec[i-1]->y(),vec[i]->x(),vec[i]->y(),yellow);
					else if(this->withOutliers())
						image.draw_line(vec[i-1]->x(),vec[i-1]->y(),vec[i]->x(),vec[i]->y(),red);					
				}
			}

		}
		
		this->image_result().push_back(image);

		tend = time(0);
		sout << "	Keypoints of image " << j << " painted in " 
			<< difftime(tend, tstart) << " seconds." << std::endl;

	}

}

#endif /* _CORRESPONDENCEPAINTER_HXX_ */



