/*  Copyright (C) 2010 Moritz Becker

    This source code is distributed under the GNU GPL.
*/
/** \file FeatureDetectionSURF.hxx
 *  Implementation of the parameter class FeatureDetectionSURF.
 *  \author Moritz Becker
 *  \date 01.04.2010
 */

#ifndef _FEATUREDETECTIONSURF_HXX_
#define _FEATUREDETECTIONSURF_HXX_

#include <ctime>

#include <opencv/cv.h>
#include <hekate/openCvUtil.hxx>

#include <hekate/features/FeatureDetectionSURF.h>
#include <OpenSURF/surflib.h>

using namespace cv ;
using namespace openCvUtil ;

template <typename T>
FeatureDetectionSURF<T>::FeatureDetectionSURF(const std::string& name) :
		TemplatedParameteredObject<T>("FeatureDetectionSURF", name,
			"Feature detection based on openSURF")
{
	
	ParameteredObject::_addParameter (thres, "threshold", "blob response threshold", 0.0004f);
	ParameteredObject::_addParameter (interval, "interval", "number of intervals per octave", 8);
	ParameteredObject::_addParameter (octaves, "octaves", "number of octaves to calculate", 4);
	ParameteredObject::_addParameter (init_sample, "init_sample", "initial sampling step" ,2);	

	ParameteredObject::_addInputSlot(image_sequence, "InputImages", "input sequence", "CImgList<T>"); 
	ParameteredObject::_addOutputSlot(correspondence, "correspondence", 
		"matches between features", "vector<vector<vector<T> > >"); 
	ParameteredObject::_addOutputSlot(features1, "features1", "features1", "vector<vector<T> >");
	ParameteredObject::_addOutputSlot(features2, "features2", "features2", "vector<vector<T> >");

	
}

template <typename T>
void FeatureDetectionSURF<T>::execute() {

	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	if (this->image_sequence().size() != 2) {
		sout << "ErrorSURF: Invalud number of images in the sequence ("
				<< this->image_sequence().size() << "), must be 2" << std::endl;
		return;
	}
	this->correspondence().resize(2);

	//calculate order of inhomogen coordinate
	//if order of inhomogen coordinate is much more less than of homogen
	//then triangulation will result in a small range for depth of points.

	IpVec ipts1;
	IpVec ipts2;

	Mat mat1 = CImg2Mat(this->image_sequence()[0].get_normalize(0.0, 255.0), CV_8U) ;
	Mat mat2 = CImg2Mat(this->image_sequence()[1].get_normalize(0.0, 255.0), CV_8U) ;

	IplImage ipl_img1 = mat1;
	IplImage ipl_img2 = mat2;

	surfDetDes(&ipl_img1, ipts1, false, this->octaves(), this->interval(), this->init_sample(), this->thres()); 
 	surfDetDes(&ipl_img2, ipts2, false, this->octaves(), this->interval(), this->init_sample(), this->thres()); 

	// Detect and describe interest points in the image
	clock_t start = clock();

	IpPairVec matches;
	getMatches(ipts1,ipts2,matches);

	sout << std::endl ;

	if(!ipts1.size()){
		sout << "no features detected" << std::endl ;
	}
	else{	

		for(IpPairVec::size_type i = 0; i < matches.size(); i++){
			
			std::vector<T> match1, match2;
			match1.push_back(matches[i].first.x);
			match1.push_back(matches[i].first.y);
			match1.push_back(1);
			match2.push_back(matches[i].second.x);
			match2.push_back(matches[i].second.y);
			match2.push_back(1);

			this->correspondence()[0].push_back(match1); 
			this->correspondence()[1].push_back(match2);

			this->features1().push_back(match1);
			this->features2().push_back(match2);

			sout << "left: " << match1[0] << ", " << match1[1] << std::endl;
			
		}
	}

	clock_t cend = clock();

	sout << std::endl ;

	sout<< "OpenSURF found: " << ipts1.size() << " interesting points in img1" << std::endl;
	sout<< "OpenSURF found: " << ipts2.size() << " interesting points in img2" << std::endl;
	sout<< "OpenSURF found: " << matches.size() << " matches." << std::endl;
	sout<< "OpenSURF took: " << float(cend - start) / CLOCKS_PER_SEC  << " seconds" << std::endl;



}


#endif /* _FEATUREDETECTIONSURF_HXX_ */



