/*  Copyright (C) 2011 Stephan Meister

    This source code is distributed under the GNU GPL.
*/
/** \file FeatureKLT.h
 *  Declaration of the parameter class FeatureKLT.
 *  \author Stephan Meister
 *  \date 06.05.2010
 */

#ifndef _FEATUREKLT_H_
#define _FEATUREKLT_H_

#if defined(MSVC) && defined(HANDLE_DLL)
#ifdef featureklt_EXPORTS
/// Visual C++ specific code
#define featureklt_DECLDIR __declspec(dllexport)
#else
#define featureklt_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else /* No DLL handling or GCC */
/// Not needed with GCC
#define featureklt_DECLDIR
#endif

#include <charon-core/ParameteredObject.hxx>
#include <charon-utils/CImg.h>
#include <hekate/Correspondence.hxx>

using namespace hekate;

/// Detector and matcher based on OpenCV KLT.
/** Uses a structure tensor based corner detector to find features and calculates correspondences by means of Lucas-Kanade optical flow 
	When disparity images are provided, then they are used for the left-to-right correspondences <br>
	Output: <br>
	correspondences[0]			= list of features for first image of imageSequence1<br> 
	correspondences[1]			= list of features for second image of imageSequence1<br> 
	correspondences[0][n]		= n-th feature of first image of imageSequence1<br> 
	correspondences[0][0][0]	= x coordinate of first feature of first image of imageSequence1<br> 
	correspondences[0][0][1]	= y coordinate of first feature of first image of imageSequence1<br> 
	correspondences[0][0][2]	= w (homogenous) coordinate of first feature of first image of imageSequence1
	the second dimension (n) is the same for all vectors and ordered to point to the same feature in all images <br>
*/
template <typename T>
class featureklt_DECLDIR FeatureKLT :
		public TemplatedParameteredObject<T> {
public:
	/// Maximal numbers of features that should be found in every image.
	Parameter <int > _maxFeatures ;
	Parameter <double> _featureQualityLevel ;
	/// Minimal distance between features.
	Parameter <double> _featureMinDistance ;
	Parameter <int> _featureBlockSize ;
	Parameter <bool> _useHarrisDetector ;
	Parameter <double> _harrisFreeParameter ;
	Parameter <bool> _useSubPixRefine ;
	
	Parameter <int> _lkWindowSize ;
	Parameter <int> _lkLevels ;

	/// Image sequence 1.
	InputSlot < cimg_library::CImgList<T> > imageSequenceLR1;
	/// Image sequence 2.
	InputSlot < cimg_library::CImgList<T> > imageSequenceLR2;


	/// Correspondences of image sequence 1.
	OutputSlot < CorrespondenceList<T> > correspondencesLR1;
	/// Correspondences of image sequence 2.
	OutputSlot < CorrespondenceList<T> > correspondencesLR2;
	/// features
	OutputSlot < CorrespondenceList<T> > features;

	/// Disparity for image sequence 1.
	InputSlot< cimg_library::CImgList<T> > dispLR1 ;
	/// Disparity for image sequence 2.
	InputSlot< cimg_library::CImgList<T> > dispLR2 ;


	/// create a new FeatureKLT object
	/// \param name          Instance name
	FeatureKLT(const std::string& name);

	/// Update object.
	virtual void execute();
};

#endif // _FEATUREKLT_H_



