/*  Copyright (C) 2010 Stephan Meister

    This source code is distributed under the GNU GPL.
*/
/** \file FeatureKLT.hxx
 *  Implementation of the parameter class FeatureKLT.
 *  \author Stephan Meister
 *  \date 10.0.2010
 */

#ifndef _FEATUREKLT_HXX_
#define _FEATUREKLT_HXX_

#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cxcore.h>

#include <hekate/openCvUtil.hxx>

#include <hekate/features/FeatureKLT.h>

using namespace cv ;
using namespace openCvUtil ;

#include <stdio.h>

template <typename T>
FeatureKLT<T>::FeatureKLT(const std::string& name) :
		TemplatedParameteredObject<T>("FeatureKLT", name,
			"Feature Detector and matcher based on OpenCV KLT for two image pairs.<br>"
			"Uses a structure tensor based corner detector to find features and calculates correspondences by means of Lucas-Kanade optical flow"
			"When disparity images are provided, then they are used for the left-to-right correspondences"
			"Output:"
			"correspondences[0]			= list of features for first image of imageSequence1<br>"
			"correspondences[1]			= list of features for second image of imageSequence2<br>"
			"correspondences[0][n]		= n-th feature of first image of imageSequence1<br>"
			"correspondences[0][0][0]	= x coordinate of first feature of first image of imageSequence1<br>"
			"correspondences[0][0][1]	= y coordinate of first feature of first image of imageSequence1<br>"
			"correspondences[0][0][2]	= w (homogenous) coordinate of first feature of first image of imageSequence1<br>"
			"<br>the second dimension (n) is the same for all vectors and ordered to point to the same feature in all images"),
			dispLR1(true, false),
			dispLR2(true, false)
{
	
	ParameteredObject::_addInputSlot(imageSequenceLR1, "imageSequenceLR1", "input sequence frame 1 (Left-Right), exspects images ordered along 5. dimension (one CImg each in the CImgList)", "CImgList<T>"); 
	ParameteredObject::_addInputSlot(imageSequenceLR2, "imageSequenceLR2", "input sequence frame 2 (Left-Right), exspects images ordered along 5. dimension (one CImg each in the CImgList)", "CImgList<T>"); 
	
	ParameteredObject::_addInputSlot(dispLR1, "dispLR1", "left-right disparity frame 1", "CImgList<T>") ;
	ParameteredObject::_addInputSlot(dispLR2, "dispLR2", "left-right disparity frame 2", "CImgList<T>") ;

	ParameteredObject::_addOutputSlot(correspondencesLR1, "correspondencesLR1", "correspondences for imageSequence 1", "CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(correspondencesLR2, "correspondencesLR2", "correspondences for imageSequence 2", "CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(features, "features", "features", "CorrespondenceList"); 

	

	ParameteredObject::_addParameter(_maxFeatures, "maxFeatures",
		"maximum number of features to detect<br>"
		"if more get detected, the n best will be selected", 100) ;
	ParameteredObject::_addParameter(_featureQualityLevel, "featureQualityLevel",
		"minimal accepted feature quality level [0.0-1.0]<br>"
		"relative to quality of the best feature, good values should be well below 0.01<br>"
		"higher gives less but usually more robust features", 1e-3) ;
	ParameteredObject::_addParameter(_featureMinDistance, "featureMinDistance",
		"minimum possible euclidean distance between features", 8.0) ;
	ParameteredObject::_addParameter(_featureBlockSize, "featureBlockSize",
		"block size for calculating feature properties", 7) ;
	ParameteredObject::_addParameter(_useHarrisDetector, "useHarrisDetector",
		"use the Harris corner detector instead of minimal eigenvalue of gradient matrices", false) ;
	ParameteredObject::_addParameter(_harrisFreeParameter, "harrisFreeParameter",
		"free Parameter of Harris feature detector, ignored if \"useHarrisDetector\" is false<br>"
		"(more information on the influence of this parameter is needed)", 0.04) ;
	ParameteredObject::_addParameter(_useSubPixRefine, "useSubPixRefine",
		"try to find corners with sub pixel accuracy", false) ;

	ParameteredObject::_addParameter(_lkWindowSize, "lkWindowSize",
		"Size of search window for LK method", 15) ;
	ParameteredObject::_addParameter(_lkLevels, "lkLevels",
		"LK Pyramid levels", 3) ;

}

template <typename T>
void FeatureKLT<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	//name of this plugin for printing info
	const std::string name = this->ParameteredObject::getName() ;

	//copy pixel data to OpenCV Mat
	cimg_library::CImg<float> img = this->imageSequenceLR1()[0] ;
	Mat matL1 = CImg2Mat(img, CV_32FC1) ;
	Mat matL1_8 = CImg2Mat(img, CV_8U) ;
	img = this->imageSequenceLR1()[1] ;
	Mat matR1 = CImg2Mat(img, CV_32FC1) ;
	Mat matR1_8 = CImg2Mat(img, CV_8U) ;
	img = this->imageSequenceLR2()[0] ;
	Mat matL2 = CImg2Mat(img, CV_32FC1) ;
	Mat matL2_8 = CImg2Mat(img, CV_8U) ;
	img = this->imageSequenceLR2()[1] ;
	Mat matR2 = CImg2Mat(img, CV_32FC1) ;
	Mat matR2_8 = CImg2Mat(img, CV_8U) ;
		
	Mat empty ;

	cv::vector<Point2f> featsL1, featsR1 ;
	
	goodFeaturesToTrack(matL1, 
		featsL1,
		_maxFeatures(), //int maxCorners
		_featureQualityLevel(), //double qualityLevel
		_featureMinDistance(), //double minDistance
		empty, //const Mat& mask=Mat(), mask to exclude regions from computation, not used here
		_featureBlockSize(), //int blockSize=3 
		_useHarrisDetector(), //bool useHarrisDetector=false
		_harrisFreeParameter // double k=0.04
	) ;
	sout << "FeatureKLT::" << name << ": found " << featsL1.size() << " features in L1" << std::endl ;

	/*
	goodFeaturesToTrack(mat1, 
		featsR1,
		_maxFeatures(), //int maxCorners
		_featureQualityLevel(), //double qualityLevel
		_featureMinDistance(), //double minDistance
		empty, //const Mat& mask=Mat(), mask to exclude regions from computation, not used here
		_featureBlockSize(), //int blockSize=3 
		_useHarrisDetector(), //bool useHarrisDetector=false
		_harrisFreeParameter // double k=0.04
	) ;
	sout << "FeatureKLT::" << name << ": found " << featsR1.size() << " features in R1" << std::endl ;
	*/
	
	//try to find subpixel exact corner coordinates
	if(_useSubPixRefine())
	{
		//Mat mat = CImg2Mat(this->imageSequenceLR1()[0].get_normalize(T(0.0), T(255.0)), CV_8U) ;
		cornerSubPix(matL1_8, featsL1, 
			Size(5,5), //half size of search window
			Size(-1,-1), //zeroZone, dead region in the center which is not searched
			TermCriteria(TermCriteria::COUNT+TermCriteria::EPS, 32, 0.01)) ;//abort criterium, here after 32 iterations 
	}


	//copy normalized pixel data to OpenCV Mat (optical flow function can only use 8bit images)
	//use same normalization for both images
	//T maxPixVal1 = this->imageSequence1()[0].max() ;
	//T maxPixVal2 = this->imageSequence1()[1].max() ;
	//maxPixVal1 = (maxPixVal1 >= maxPixVal2 ? maxPixVal1 : maxPixVal2) ;
	//mat1 = CImg2Mat(this->imageSequence1()[0] / maxPixVal1 * T(255.0) , CV_8U) ;
	//Mat mat2 = CImg2Mat(this->imageSequence1()[1] / maxPixVal1 * T(255.0) , CV_8U) ;
	
	cv::vector<Point2f> featsL2 ;
	cv::vector<uchar> featsL2status ;
	cv::vector<float> err ;

	calcOpticalFlowPyrLK(matL1_8, matL2_8, featsL1, featsL2, 
			featsL2status, // vector, ==0 if flow could not be computed, 1 otherwise
			err, // error values for each corner
			Size(_lkWindowSize(), _lkWindowSize()), //window Size
			_lkLevels(), //int maxLevel=3
			TermCriteria( TermCriteria::COUNT+TermCriteria::EPS, 50, 0.01), //TermCriteria
			0.5, //double derivLambda
			0//int flags
	) ;
	
	CorrespondenceList<T>& cors1 = correspondencesLR1() ;
	CorrespondenceList<T>& cors2 = correspondencesLR2() ;
	
	for(size_t ii = 0 ; ii < featsL1.size() ; ii++)
	{
		this->features().push_back( Correspondence<T>(featsL1[ii].x, featsL1[ii].y, 0, 1, ii, 0) );
	}
	
	for(size_t ii = 0 ; ii < featsL2.size() ; ii++)
	{
		this->features().push_back( Correspondence<T>(featsL2[ii].x, featsL2[ii].y, 0, 1, ii, 1) );
	}

	T vals[3] = {T(0.0),T(0.0),T(1.0)} ;
	T* valsEnd = vals + sizeof(vals) / sizeof(T) ;

	if(dispLR1.connected() && dispLR1.size() > 0 && dispLR2.connected() && dispLR2.size() > 0)
	{
		const cimg_library::CImg<T>& disp1 = dispLR1()[0] ;
		const cimg_library::CImg<T>& disp2 = dispLR2()[0] ;
	
		for(size_t ii = 0 ; ii < featsL1.size() ; ii++)
		{
			T d1, d2 ;
			if(featsL2status[ii] != 1) //could corner be tracked to L1 ?
			{	continue ;	}
			int x1 = int(floor(featsL1[ii].x)) ;
			int y1 = int(floor(featsL1[ii].y)) ;
			if((d1 = disp1.atXY(x1,y1)) == 0) //can corner be found in R1
			{	continue ;	}

			//go from L1->L2->R2
			int x2 = int(floor(featsL2[ii].x)) ;
			int y2 = int(floor(featsL2[ii].y)) ;
			if((d2 = disp2.atXY(x2,y2)) == 0) //can corner be found in R2
			{ continue ;	}

			//L1		
			vals[0] = T(featsL1[ii].x) ; vals[1] = T(featsL1[ii].y) ;
			cors1.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 0) ) ;
			//R1
			vals[0] = T(featsL1[ii].x - d1) ; vals[1] = T(featsL1[ii].y) ;
			cors1.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 1) ) ;
			//L2
			vals[0] = T(featsL2[ii].x) ; vals[1] = T(featsL2[ii].y) ;
			cors2.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 0) ) ;
			//R2
			vals[0] = T(featsL2[ii].x - d2) ; vals[1] = T(featsL2[ii].y) ;
			cors2.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 1) ) ;
		}
	}
	else //calculate LR correspondences with LK-Flow
	{
		cv::vector<Point2f> featsR1, featsR2 ;
		cv::vector<uchar> featsR1status, featsR2status ;
		cv::vector<float> err ;


		calcOpticalFlowPyrLK(matL1_8, matR1_8, featsL1, featsR1, 
				featsR1status, // vector, ==0 if flow could not be computed, 1 otherwise
				err, // error values for each corner
				Size(_lkWindowSize(), _lkWindowSize()), //window Size
				_lkLevels(), //int maxLevel=3
				TermCriteria( TermCriteria::COUNT+TermCriteria::EPS, 50, 0.01), //TermCriteria
				0.5, //double derivLambda
				0//int flags
		) ;

		calcOpticalFlowPyrLK(matL1_8, matR2_8, featsL1, featsR2, 
				featsR2status, // vector, ==0 if flow could not be computed, 1 otherwise
				err, // error values for each corner
				Size(_lkWindowSize(), _lkWindowSize()), //window Size
				_lkLevels(), //int maxLevel=3
				TermCriteria( TermCriteria::COUNT+TermCriteria::EPS, 50, 0.01), //TermCriteria
				0.5, //double derivLambda
				0//int flags
		) ;

		for(size_t ii = 0 ; ii < featsL1.size() ; ii++)
		{
			if(featsL2status[ii] != 1 || featsR1status[ii] != 1 || featsR2status[ii] != 1) //could corner be tracked 
			{	continue ;	}

			//L1		
			vals[0] = T(featsL1[ii].x) ; vals[1] = T(featsL1[ii].y) ;
			cors1.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 0) ) ;
			//R1
			vals[0] = T(featsR1[ii].x) ; vals[1] = T(featsR1[ii].y) ;
			cors1.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 1) ) ;
			//L2
			vals[0] = T(featsL2[ii].x) ; vals[1] = T(featsL2[ii].y) ;
			cors2.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 0) ) ;
			//R2
			vals[0] = T(featsR2[ii].x) ; vals[1] = T(featsR2[ii].y) ;
			cors2.push_back( Correspondence<T>(std::vector<T>(vals, valsEnd), 0, 1) ) ;

			/*sout << "PtImg1Left " << featsL1[ii].x << ", " << T(featsL1[ii].y) << std::endl;
			sout << "PtImg2Right " << featsR1[ii].x << ", " << T(featsR1[ii].y) << std::endl;
			sout << "PtImg1Left " << featsL2[ii].x << ", " << T(featsL2[ii].y) << std::endl;
			sout << "PtImg2Right " << featsR2[ii].x << ", " << T(featsR2[ii].y) << std::endl;
			sout << "-----------" << std::endl;*/
		}

	}
	sout << "found " <<cors1.size() << " feature tuples" << std::endl ;
}

#endif /* _FEATUREKLT_HXX_ */



