/*  Copyright (C) 2011 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeatureKLTSequence.hxx
 *  Implementation of the parameter class FeatureKLTSequence.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 05.09.2011
 */

#ifndef _FEATUREKLTSEQUENCE_HXX_
#define _FEATUREKLTSEQUENCE_HXX_

#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cxcore.h>

#include <hekate/openCvUtil.hxx>

#include <charon-core/ParameteredObject.hxx>
#include <hekate/features/FeatureKLTSequence.h>

template <typename T>
FeatureKLTSequence<T>::FeatureKLTSequence(const std::string& name) :
		TemplatedParameteredObject<T>(
			"FeatureKLTSequence", name,
			"<h2>Calculates KLT features and correspondences from an image "
			"sequence</h2><br>"
			"Calculates KLT features and correspondences from an image "
			"sequence. The Sequence is given by the path of the first image "
			"file.<br>"
			"The correspondences are outputed as a list of tracking "
			"matrices."
		)
{

	ParameteredObject::_addParameter(useBaseImage, "useBaseImage",
		"Takes the first image in folder as base image which will be taken "
		"as one of two images for correspondence estimation", true) ;

	ParameteredObject::_addParameter(
		imageSequence, "imageSequence",
		"Path to the first image file of the sequence.",
		"fileopen");

	ParameteredObject::_addParameter (maxImages, "maxImages", 
	"Maximal number of images per sequence.", 100
	);

	ParameteredObject::_addParameter(
		trackingMatricesFile, "trackingMatricesFile",
		"Path to the output tracking matrices.",
		"filewrite");

	ParameteredObject::_addParameter(_minNumberOfCorrs, "minNumberOfCorrs",
		"Minimum number of correspondences of an<br>"
		"image sequence.", 50) ;

	ParameteredObject::_addParameter(_maxFeatures, "maxFeatures",
		"maximum number of features to detect<br>"
		"if more get detected, the n best will be selected", 100) ;
	ParameteredObject::_addParameter(_featureQualityLevel, "featureQualityLevel",
		"minimal accepted feature quality level [0.0-1.0]<br>"
		"relative to quality of the best feature, good values should be well below 0.01<br>"
		"higher gives less but usually more robust features", 1e-3) ;
	ParameteredObject::_addParameter(_featureMinDistance, "featureMinDistance",
		"minimum possible euclidean distance between features", 8.0) ;
	ParameteredObject::_addParameter(_featureBlockSize, "featureBlockSize",
		"block size for calculating feature properties", 7) ;
	ParameteredObject::_addParameter(_useHarrisDetector, "useHarrisDetector",
		"use the Harris corner detector instead of minimal eigenvalue of gradient matrices", false) ;
	ParameteredObject::_addParameter(_harrisFreeParameter, "harrisFreeParameter",
		"free Parameter of Harris feature detector, ignored if \"useHarrisDetector\" is false<br>"
		"(more information on the influence of this parameter is needed)", 0.04) ;
	ParameteredObject::_addParameter(_useSubPixRefine, "useSubPixRefine",
		"try to find corners with sub pixel accuracy", false) ;

	ParameteredObject::_addParameter(_lkWindowSize, "lkWindowSize",
		"Size of search window for LK method", 15) ;
	ParameteredObject::_addParameter(_lkLevels, "lkLevels",
		"LK Pyramid levels", 3) ;

	ParameteredObject::_addOutputSlot(
		trackingMatrices, "trackingMatrices",
		"trackingMatrices",
		"CImgList<T>");

	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"Correspodences.",
		"vector<vector<vector<T> > >");

	ParameteredObject::_addOutputSlot(
		images, "images",
		"images.",
		"CImgList<T>");
}

template <typename T>
void FeatureKLTSequence<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	using namespace cv ;
	using namespace openCvUtil ;

	// Read in specified dir
	std::string dir = this->imageSequence() ;
	FileTool::slashConvert(dir);
	if(dir==""){
		throw(std::runtime_error("Not file given."));
	}
	else
		sout << "Opening " << dir << std::endl;


	// Get file path
	string filepath = dir.substr(0,
				dir.find_last_of(FileTool::slash)+1);
	sout << "File path = " << filepath << std::endl;

	// Get file name
	string filename = dir.substr(dir.find_last_of(FileTool::slash)+1,
				dir.length());
	sout << "File name = " << filename << std::endl;
	
	// Get file extension
	string ext = dir.substr(dir.find_last_of('.'), dir.length());
	sout << "Looking for *" << ext << " files" << std::endl;

	// Change working directory
	FileTool::changeDir(filepath);

	//sout << "Working Dir: " << FileTool::getCurrentDir() << std::endl;

	// Read in files with specified extension
	std::vector<std::string> myVec;
	myVec = FileTool::getFilesWithSuffix(ext);

	sout << myVec.size() << " files found" 
		<< " -------------------------" << std::endl;

	// Write found files to log
	for( size_t i=0; i<myVec.size(); i++ ){
		sout << myVec[i] << std::endl;
	}


	/* Get Feature correspondences between images ---------------------------- */

	std::vector<cv::vector<Point2f> > feats;
	cv::vector<uchar> featsstatus ;
	for(int i=0; i<_maxFeatures(); i++){
		featsstatus.push_back(1);
	}

	//copy pixel data to OpenCV Mat
	cimg_library::CImg<float> img((filepath + myVec[0]).c_str());
	Mat mat1 = CImg2Mat(img.get_normalize(0,255), CV_8U) ;
	Mat mat2;
	Mat empty ;

	if(this->useBaseImage())
		sout << "Choose base image: " << std::endl;
	else
		sout << "Choose first image: " << std::endl;
	sout << filepath + myVec[0] << std::endl; 
	sout << "Image width: " <<  img.width() << std::endl;
	sout << "Image height: " <<  img.height() << std::endl;

	cv::vector<Point2f> feats1 ;

	goodFeaturesToTrack(mat1, 
		feats1,
		_maxFeatures(), //int maxCorners
		_featureQualityLevel(), //double qualityLevel
		_featureMinDistance(), //double minDistance
		empty, //const Mat& mask=Mat(), mask to exclude regions from computation, not used here
		_featureBlockSize(), //int blockSize=3 
		_useHarrisDetector(), //bool useHarrisDetector=false
		_harrisFreeParameter // double k=0.04
	) ;

	feats.push_back(feats1);
	
	if(this->useBaseImage())
		sout << "FeatureKLTSequence found " << feats1.size() << " features in base image." << std::endl ;
	else
		sout << "FeatureKLTSequence found " << feats1.size() << " features in first image." << std::endl ;
	sout << "--------------------------------" << std::endl;

	// Push back reference image
	this->images().push_back(img);

	for( size_t i=1; i<myVec.size(); i++ ){

		cv::vector<Point2f> featsNew, featsLast;
		cv::vector<uchar> feats2status, featsstatusCopy=featsstatus ;
		cv::vector<float> err ;


		if(int(i) >= this->maxImages()){
			sout << "Maximal number of images added to sequence ("
				<< this->maxImages() << ")" << std::endl;
			break;
		}

		//copy pixel data to OpenCV Mat
		img.load((filepath + myVec[i]).c_str());

		if(!this->useBaseImage() && i>1){
			// Choose the last image as reference image of all i > 1
			if(i>1)
				mat1 = mat2;
			sout << "Calculate correspondences between image " << myVec[i-1] << 
				" and: " << myVec[i] << std::endl;
			// Choose last frame features as reference features
			featsLast = feats[i-1];
		}
		else{
			sout << "Calculate correspondences between base image and " << myVec[i] << std::endl;
			featsLast = feats1;
		}

		mat2 = CImg2Mat(img.get_normalize(0,255), CV_8U) ;
					

		calcOpticalFlowPyrLK(mat1, mat2, featsLast, featsNew, 
				feats2status, // vector, ==0 if flow could not be computed, 1 otherwise
				err, // error values for each corner
				Size(_lkWindowSize(), _lkWindowSize()), //window Size
				_lkLevels(), //int maxLevel=3
				TermCriteria( TermCriteria::COUNT+TermCriteria::EPS, 50, 0.01), //TermCriteria
				0.5, //double derivLambda
				0//int flags
		) ;

		int noOfCorrs = 0;
		int noOfLastCorrs = 0;
		for(size_t ii = 0 ; ii < featsLast.size() ; ii++)
		{
			// Check if correspondence could be found
			// and if all former images also contain this
			// correspondence.
			if(featsstatus[ii] == 1)
			{	
				noOfLastCorrs++;	
			}
			if(feats2status[ii] == 1 && featsstatus[ii] == 1)
			{	
				noOfCorrs++;	
			}
			else
				featsstatusCopy[ii] = 0;
			//sout << featsstatus[ii] << std::endl;
		}
		if(noOfCorrs >= _minNumberOfCorrs){
			featsstatus = featsstatusCopy;
		}
		else if(i == 1){
			throw(std::runtime_error("No correspondences could be found."));
		}
		else{			
			sout << "Could not find enough correspondences ("
				<< noOfCorrs << ")." << std::endl;
			sout << "At least " << _minNumberOfCorrs 
				<< " correspondences needed." << std::endl;
			sout << "Stop searching for further correspondences "
				<< "in this sequence." << std::endl;
			if(this->useBaseImage()){
				sout << noOfLastCorrs << " correspondences between images 0-" << i-1 << std::endl;
				sout << noOfCorrs << " correspondences between images 0-" << i << std::endl;
			}
			else{
				sout << noOfLastCorrs << " correspondences between images "  
					<< i-1 << " and " << i << std::endl;
				if(i>1)
					sout << noOfCorrs << " correspondences between images "
						<< i-2 << " and " << i-1 << std::endl;
			}

			break;
		}

		if(this->useBaseImage()){
			if(i>1)
				sout << noOfLastCorrs << " correspondences between images 0-" << i-1 << std::endl;
			else
				sout << noOfLastCorrs << " features in image " << i-1 << std::endl;
			sout << noOfCorrs << " correspondences between images 0-" << i << std::endl;
		}
		else{
			sout << noOfLastCorrs << " correspondences between images "  
				<< i-1 << " and " << i << std::endl;
			if(i>1)
				sout << noOfCorrs << " correspondences between images "
					<< i-2 << " and " << i-1 << std::endl;
		}
		
		feats.push_back(featsNew);
		this->images().push_back(img);

	}

	sout << "Extracting correspondences." << std::endl;
	sout << "Numer of images in this image sequence: " 
		<< feats.size() << std::endl;

	std::vector<std::vector<std::vector<T> > > cors ;
	for(size_t i=0; i<feats.size(); i++){

		T vals[3] = {T(0.0),T(0.0),T(1.0)} ;
		T* valsEnd = vals + sizeof(vals) / sizeof(T) ;
		std::vector<std::vector<T> > cors1 ;

		for(size_t ii = 0 ; ii < feats[i].size() ; ii++)
		{

			if(featsstatus[ii] == 1) //could corner be tracked 
			{
		
				vals[0] = T(feats[i][ii].x) ; vals[1] = T(feats[i][ii].y) ;
				cors1.push_back( std::vector<T>(vals, valsEnd)) ;

				/*
				sout << "Image[" << i << "], Point[" << ii << "]: "
					<< feats[i][ii].x << ", " << T(feats[i][ii].y) << std::endl;
				*/

			}

		}

		cors.push_back(cors1);

	}

	this->correspondences() = cors;

}


#endif /* _FEATUREKLTSEQUENCE_HXX_ */
