/*  Copyright (C) 2009 Moritz Becker

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file featureMatcher3dSURF.hxx
 *  Implementation of the parameter class featureMatcher3dSURF.
 *  \author Moritz Becker
 *  \date 12.07.2010
 */

#ifndef _FEATUREMATCHER3DSURF_HXX_
#define _FEATUREMATCHER3DSURF_HXX_

#include <opencv/cv.h>
#include <OpenSURF/surflib.h>
#include <hekate/openCvUtil.hxx>

#include <hekate/features/FeatureMatcher3dSURF.h>

template <typename T>
featureMatcher3dSURF<T>::featureMatcher3dSURF(const std::string& name) :
		TemplatedParameteredObject<T>(
				"featureSURF", name,
				"Matches features between two image pairs and between the images of each pair."
		)
			,imageSequence1(false, false) 
			,imageSequence2(false, false) 
			
			
{
	ParameteredObject::_addParameter (thres, "threshold", "blob response threshold", 0.0004f);
	ParameteredObject::_addParameter (interval, "interval", "number of intervals per octave", 8);
	ParameteredObject::_addParameter (octaves, "octaves", "number of octaves to calculate", 4);
	ParameteredObject::_addParameter (init_sample, "init_sample", "initial sampling step" ,2);	

	ParameteredObject::_addInputSlot(imageSequence1, "imageSequence1", 
	"First image pair", 
	"CImgList<T>"); 
	ParameteredObject::_addInputSlot(imageSequence2, "imageSequence2", 
	"Second image pair", 
	"CImgList<T>"); 
	ParameteredObject::_addOutputSlot(imageCorrespondence1, "imageCorrespondence1", 
	"Matches between first image pair", 
	"vector<vector<vector<T> > >"); 
	ParameteredObject::_addOutputSlot(imageCorrespondence2, "imageCorrespondence2", 
	"Matches between second image pair", 
	"vector<vector<vector<T> > >"); 
	ParameteredObject::_addOutputSlot(imageFeatures11, "imageFeatures11", 
	"Features of the first image", 
	"std::vector<std::vector<T> >"); 
	ParameteredObject::_addOutputSlot(imageFeatures12, "imageFeatures12", 
	"Features of the second image", 
	"std::vector<std::vector<T> >"); 
	
}

template <typename T>
void featureMatcher3dSURF<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	//calculate image correspondences
	getFeatureCorrespondences();

}

template <typename T>
void featureMatcher3dSURF<T>::getFeatureCorrespondences(){
	using namespace cv ;
	using namespace openCvUtil ;

	if (this->imageSequence1().size() != 2) {
		sout << "ErrorSURF: Invalud number of images in sequence 1 ("
				<< this->imageSequence1().size() << "), must be 2" << std::endl;
		return;
	}
	else if (this->imageSequence2().size() != 2) {
		sout << "ErrorSURF: Invalud number of images in sequence 2 ("
				<< this->imageSequence2().size() << "), must be 2" << std::endl;
		return;
	}
	this->imageCorrespondence1().resize(2);
	this->imageCorrespondence2().resize(2);

	//calculate order of inhomogen coordinate
	//if order of inhomogen coordinate is much more less than of homogen
	//then triangulation will result in a small range for depth of points.
	IpVec ipts1;
	IpVec ipts2;
	IpVec ipts3;
	IpVec ipts4;

	Mat mat1 = CImg2Mat(this->imageSequence1()[0].get_normalize(0.0, 255.0), CV_8U) ;
	Mat mat2 = CImg2Mat(this->imageSequence1()[1].get_normalize(0.0, 255.0), CV_8U) ;
	Mat mat3 = CImg2Mat(this->imageSequence2()[0].get_normalize(0.0, 255.0), CV_8U) ;
	Mat mat4 = CImg2Mat(this->imageSequence2()[1].get_normalize(0.0, 255.0), CV_8U) ;

	IplImage ipl_img1 = mat1;
	IplImage ipl_img2 = mat2;
	IplImage ipl_img3 = mat3;
	IplImage ipl_img4 = mat4;

	surfDetDes(&ipl_img1, ipts1, false, this->octaves(), this->interval(), this->init_sample(), this->thres()); 
	surfDetDes(&ipl_img2, ipts2, false, this->octaves(), this->interval(), this->init_sample(), this->thres()); 
	surfDetDes(&ipl_img3, ipts3, false, this->octaves(), this->interval(), this->init_sample(), this->thres()); 
	surfDetDes(&ipl_img4, ipts4, false, this->octaves(), this->interval(), this->init_sample(), this->thres()); 

	sout << std::endl ;

	if(!ipts1.size() || !ipts2.size() || !ipts3.size() || !ipts4.size()){
		sout << "no features detected" << std::endl ;
	}
	else{

		IpPairVec matches1, matches2, matches3, matches4;
		std::vector<std::pair<int, int> > pointOrder ;
		pointOrder.clear();

		getMatches(ipts1,ipts2,matches1);
		getMatches(ipts3,ipts4,matches2);
		getMatches(ipts1,ipts3,matches3);
		getMatches(ipts2,ipts4,matches4);


		// compare correspondences and keep the ones appear in all 3 matches
		std::vector<T> match1, match2, match3, match4;
		for(size_t i=0; i< matches1.size(); i++){
			for(size_t j=0; j< matches2.size(); j++){

				//find intersection of matches 1, 2 and 3
				for(size_t k=0; k< matches3.size(); k++){

					if(std::abs((double)matches1[i].first.x-(double)matches3[k].first.x) < 1. &&  
						std::abs((double)matches1[i].first.y-(double)matches3[k].first.y) < 1. &&
						std::abs((double)matches2[j].first.x-(double)matches3[k].second.x) < 1. &&
						std::abs((double)matches2[j].first.y-(double)matches3[k].second.y) < 1.
						){
							match1.clear();
							match2.clear();
							match3.clear();
							match4.clear();

							pointOrder.push_back(std::make_pair(i,j));

							match1.push_back(matches1[i].first.x);
							match1.push_back(matches1[i].first.y);
							match1.push_back(1);
							match2.push_back(matches1[i].second.x);
							match2.push_back(matches1[i].second.y);
							match2.push_back(1);

							this->imageCorrespondence1()[0].push_back(match1); 
							this->imageCorrespondence1()[1].push_back(match2);

							this->imageFeatures11().push_back(match1);
							this->imageFeatures12().push_back(match2);

							match3.push_back(matches2[j].first.x);
							match3.push_back(matches2[j].first.y);
							match3.push_back(1);
							match4.push_back(matches2[j].second.x);
							match4.push_back(matches2[j].second.y);
							match4.push_back(1);

							this->imageCorrespondence2()[0].push_back(match3); 
							this->imageCorrespondence2()[1].push_back(match4);

							sout << "SURF(" << this->imageCorrespondence1()[0].size() << "): " << "(" 
								<< match1[0] << "," << match1[1] << ") ("
								<< match2[0] << "," << match2[1] << ") ("
								<< match3[0] << "," << match3[1] << ") ("
								<< match4[0] << "," << match4[1] << ")" << std::endl;

					}
				}
				//find intersection of matches 1, 2 and 4
				for(size_t k=0; k< matches4.size(); k++){

					if(	(match1.size()>0 ? matches1[i].second.x!=match2[0] : true) &&  
						(match1.size()>0 ? matches1[i].second.y!=match2[1] : true) &&  
						std::abs((double)matches1[i].second.x-(double)matches4[k].first.x) < 1. &&  
						std::abs((double)matches1[i].second.y-(double)matches4[k].first.y) < 1. &&
						std::abs((double)matches2[j].second.x-(double)matches4[k].second.x) < 1. &&
						std::abs((double)matches2[j].second.y-(double)matches4[k].second.y) < 1.
						){

							match1.clear();
							match2.clear();
							match3.clear();
							match4.clear();

							pointOrder.push_back(std::make_pair(i,j));

							match1.push_back(matches1[i].first.x);
							match1.push_back(matches1[i].first.y);
							match1.push_back(1);
							match2.push_back(matches1[i].second.x);
							match2.push_back(matches1[i].second.y);
							match2.push_back(1);

							this->imageCorrespondence1()[0].push_back(match1); 
							this->imageCorrespondence1()[1].push_back(match2);

							this->imageFeatures11().push_back(match1);
							this->imageFeatures12().push_back(match2); 

							match3.push_back(matches2[j].first.x);
							match3.push_back(matches2[j].first.y);
							match3.push_back(1);
							match4.push_back(matches2[j].second.x);
							match4.push_back(matches2[j].second.y);
							match4.push_back(1);

							this->imageCorrespondence2()[0].push_back(match3); 
							this->imageCorrespondence2()[1].push_back(match4);

							sout << "SURF(" << this->imageCorrespondence1()[0].size() << "): " << "(" 
								<< match1[0] << "," << match1[1] << ") ("
								<< match2[0] << "," << match2[1] << ") ("
								<< match3[0] << "," << match3[1] << ") ("
								<< match4[0] << "," << match4[1] << ") ADDITIONAL" << std::endl;
					}
				}
				
			}
		}

		sout<< "OpenSURF found: " << matches1.size() << " correspondences between img1 and img2" << std::endl;
		sout<< "OpenSURF found: " << matches2.size() << " correspondences between img3 and img4" << std::endl;
		sout<< "OpenSURF found: " << matches3.size() << " correspondences between img1 and img3" << std::endl;
		sout<< "OpenSURF: " << this->imageCorrespondence1()[0].size() << " correspondences could be matched" << std::endl;
		sout<< "OpenSURF: Refusing correspondences that could not be matched." << std::endl;
	}
}

#endif /* _FEATUREMATCHER3DSURF_HXX_ */



