/*  Copyright (C) 2011 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeatureSequence.hxx
 *  Implementation of the parameter class FeatureSequence.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 20.10.2011
 */

#ifndef _FEATURESEQUENCE_HXX_
#define _FEATURESEQUENCE_HXX_

#include <charon-core/ParameteredObject.hxx>
#include <hekate/features/FeatureSequence.h>
#include <hekate/Correspondence.hxx>

template <typename T>
FeatureSequence<T>::FeatureSequence(const std::string & pluginName, const std::string& name,
			const std::string & doc) :
		TemplatedParameteredObject<T>(pluginName, name, doc),
			imgPath(true,false)
{
	ParameteredObject::_addInputSlot(
		imgPath, "imgPath",
		"imgPath",
		"vector<string>");
	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"Estimated correspondences",
		"CorrespondenceList");
	ParameteredObject::_addOutputSlot(
		imgPaths, "imgPaths",
		"imgPaths",
		"vector<string>");
	
	
	ParameteredObject::_addParameter(
		imageDir, "imageDirectory",
		"Directory of image sequence.", "fileopen");
	ParameteredObject::_addParameter< size_t >(
		maximalFrames, "maximalFrames",
		"Maximal number of frames per sliding window from which correspondences are calculated.",
		100);
	ParameteredObject::_addParameter< size_t >(
		minimalNumberOfCorrs, "minimalNumberOfCorrs",
		"Minimal number of correspondences between images. If the number of"
		"correspondences between two images is less than this threshold, the"
		"second image won't be applied to the image/correspondence list.",
		50);
	ParameteredObject::_addParameter< size_t >(
		maxFWC, "maximalNumberOfFramesWithoutCorrs",
		"Maximal number of subsequent frames, which do not exhibit "
		"the minimal number of correspondences (minimalNumberOfCorrs) "
		"between those frames.",
		3);
	ParameteredObject::_addParameter< size_t >(
		matchRange, "matchRange",
		"How many sub sequent frames before and after, a single frame gets matched with",
		5, "size_t");

	

}

template <typename T>
void FeatureSequence<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	const std::string pluginName = this->ParameteredObject::getName();

	this->filenames.clear();
	_numberOfCorrespondences = 0;
	numberOfFiles = 0;
	calcCurrent = 0;
	calcMax = 0;
	
	// Read in images -------------------------------------------
	sout << "Reading in folder:" << std::endl;

	//create a new list
	hekate::CorrespondenceList<T> list ;

	// Read in specified dir
	std::string dir;
	if(this->imgPath.connected()){
		sout << "Path slot connected, taking slot data." << std::endl;
		dir = this->imgPath()[0] ;
	}
	else
		dir = this->imageDir() ;

	FileTool::slashConvert(dir);
	if(dir==""){
		throw(std::runtime_error("No file of directory given."));
	}
	else
		sout << "	Choose " << dir 
			<< " as sequence directory:" << std::endl;


	// Get file path
	this->filepath = dir.substr(0,
				dir.find_last_of(FileTool::slash)+1);
	sout << "	File path = " << filepath << std::endl;


	// Get file name
	std::string filename = dir.substr(dir.find_last_of(FileTool::slash)+1,
				dir.length());
	sout << "	Reference file name = " << filename << std::endl;
	

	// Get file extension
	if(!this->imgPath.connected()){
		std::string ext = dir.substr(dir.find_last_of('.'), dir.length());
		sout << "	Filter for *" << ext << " files" << std::endl;

		// Change working directory
		FileTool::changeDir(filepath);
		//sout << "Working Dir: " << FileTool::getCurrentDir() << std::endl;


		// Read in files with specified extension
		this->filenames = FileTool::getFilesWithSuffix(ext);
		this->numberOfFiles = filenames.size();
		sout << "	" << numberOfFiles << " files found:" << std::endl;

		// Write found files to log
		for( size_t i=0; i<this->numberOfFiles; i++ ){
			sout << "	img" << i << ": " << filenames[i] << std::endl;
		}
	
	}
	else{
		this->numberOfFiles = this->imgPath().size();
		sout << "Found " << this->numberOfFiles << std::endl;
		for( size_t i=0; i<this->numberOfFiles; i++ ){
			std::string ext = this->imgPath()[i].substr(dir.find_last_of(FileTool::slash)+1,
				dir.length());
			sout << "	img" << i << ": " << ext << std::endl;
			this->filenames.push_back( ext );
		}
	}
	

	// Calculate correspondences
	// This function differs in every child class
	_calculate();

		
}


/// Erase correspondence column i
template <typename T>
void FeatureSequence<T>::removeOutliers(std::vector<std::vector<std::vector<T> > > & corr){

	const size_t dimN = corr.size();

	// Search outliers
	for(size_t i=0; i<dimN; i++){
		
		size_t dimM = corr[0].size();
		if(dimM != corr[i].size())
			throw(std::runtime_error("Inconsistent sizes of correspondences"));

		size_t j=0;
		while( j < dimM ){
			
			// If outlier detected, remove complete row from correspondence matrix
			if(corr[i][j][0] == T(0) && corr[i][j][1] == T(0) && corr[i][j][2] == T(0)){
				
				// Erase the row in which the outlier was found
				for(size_t k=0; k<dimN; k++){

					if(dimM != corr[k].size())
						throw(std::runtime_error("Inconsistent sizes of correspondences"));

					const std::vector< std::vector<T> >::const_iterator it = corr[k].begin();
					corr[k].erase(it+j);

				}

				// Apply new correspondence dimension and restart search in list i
				dimM = corr[0].size();
				j--;

			}

			j++;

		}

	}

}

#endif /* _FEATURESEQUENCE_HXX_ */
