/*  Copyright (C) 2012 Martin Treiber

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeatureSequenceASIFT.hxx
 *  Implementation of the parameter class FeatureSequenceGPUASIFT.
 *  \author <a href="mailto:martin.treiber.hd@gmx.de">
 *      Martin Treiber</a>
 *  \date 14.03.2012
 */

//This is basically the same as FeatureSequenceasift.hxx by Moritz Becker


#ifndef _FEATURESEQUENCEGPUASIFT_HXX_
#define _FEATURESEQUENCEGPUASIFT_HXX_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <vector>
using namespace std;

#ifdef _OPENMP
#include <omp.h>
#endif

#include <charon-core/ParameteredObject.hxx>
#include "hekate/features/FeatureSequenceGPUASIFT.h"
#include <hekate/features/FeatureSequence.hxx>

using namespace hekate;


template <typename T>
FeatureSequenceGPUASIFT<T>::FeatureSequenceGPUASIFT(const std::string& name) :
		FeatureSequence<T>(
			"FeatureSequenceGPUASIFT", name,
			"<h2>Affine GpuSIFT version of the sequence feature "
			"extractor</h2><br>"
			"Calculates feature correspondences of an image sequence. For "
			"this, the affine SIFT version ( http://www.ipol.im/pub/algo/my_affine_sift ) is used, "
			"changed in order to use SiftGPU ( http://cs.unc.edu/~ccwu/siftgpu/ ) and CUDA "
			"(http://www.nvidia.com/object/cuda_home_new.html) for additional performance.<br>"
			"For the imgageDirectory, specifiy one of the images of your sequence and make sure "
			"the rest of the images reside in the same directory and have the same ending.<br>"
			"The Parameters minimalNumberOfCorrs and MaximalNumberOfFrames "
			"are deprecated and are only there for compatibility reasons"
		)
{
	ParameteredObject::_addParameter< bool > (scaleFrames, "scaleFrames", "Scale Frames", false);
	ParameteredObject::_addParameter< int > (scaleX, "scaleX", "Width of scaled image", 800, "int");
	ParameteredObject::_addParameter< int > (scaleY, "scaleY", "Height of scaled image", 600, "int");
	ParameteredObject::_addParameter< int > (nTilts, "numberOfTilts", "Number of tilts per image", 7, "int");
	ParameteredObject::_addParameter< int > (nMinCorr, "nMinCorr", "minimal number of correspondences "
		"which are used to define a match between two frames", 50, "int");
	
	ParameteredObject::_setTags("hekate") ;

}

template <typename T>
void FeatureSequenceGPUASIFT<T>::_calculate() {
	
	widthList.clear();
	heightList.clear();
	widthList2.clear();
	heightList2.clear();
	num_of_tilts = 0;
	imgPaths().clear();


	hekate::CorrespondenceList<T> &corrOut = correspondences();
	// list of all keypoints of the selected sequence
	vector< vector< vector< keypointslist > > > keypoints;
	vector< vector< vector< keypointslist > > > keypoints2;
	vector<float> zoom;		
	vector<float> zoom2;		
	
	// Define image scaling
	float wS = scaleX();
	float hS = scaleY();
	if(this->scaleFrames())
		sout << std::endl << "WARNING: The input images are resized to " << wS << "x" << hS << " for ASIFT. " << endl 
		<< "         But the results will be normalized to the original image size." << endl << endl;
	
	
	// Define the SIFT parameters
	// number N of tilts to simulate t = 1, \sqrt{2}, (\sqrt{2})^2, ..., {\sqrt{2}}^(N-1)
	verb = 0;
	default_sift_parameters(siftparameters);
	num_of_tilts = this->nTilts();
	

	// Check iteration boundary
	if(numberOfFiles > maximalFrames())
	{
		numberOfFiles = maximalFrames();
		sout << "maximalFrames: " << maximalFrames() << std::endl;
	}
	sout << "Number of files to calculate: " << numberOfFiles << std::endl;
	
	calcMax = 2* numberOfFiles;
	calcCurrent = 0;
					
	// Calculate keypoints for every image
	for(size_t i=0; i < this->numberOfFiles; i++)
	{

		calcCurrent++;

		sout << "-----------------------------------" << std::endl;
		sout << "Computing keypoints on image " << i << ": " << filenames[i].c_str() << endl;
		
		// Read image i
		float * iarr1;
		cimg_library::CImg<float> img1((filepath + filenames[i]).c_str());
		img1.normalize(0.0,255.0);
		
		imgPaths().push_back((filepath + filenames[i]).c_str()); 
		
		int thx = 2000;
		int thy = 2000;
		int iw = img1.width(), ih = img1.height();

		// Scale images if taller than threshold
		if(iw > thx || ih > thy){
			sout << "ATTENTION: IMAGE IS TOO BIG FOR ASIFT! Scaling..." << std::endl;
			if(iw>ih){
				float c = float(thx)/float(iw);
				img1.resize(2000,int(c*ih));
			}
			else{
				float c = float(thy)/float(ih);
				img1.resize(int(c*iw),2000);
			}
		}

		size_t w1=img1.width(), h1=img1.height();
		iarr1 = img1.data();
		std::vector<float> ipixels1(iarr1, iarr1 + w1 * h1);

		///// Resize the images to area wS*hS in remaining the apsect-ratio	
		///// Resize if the resize flag is not set or if the flag is set unequal to 0		
		std::vector<float> ipixels1_zoom1;
		zoomImage(ipixels1, ipixels1_zoom1, w1, h1, widthList, heightList, zoom);

		vector< vector< keypointslist > > keys1;	
		int num_keys1=0;
		
		time_t tstart, tend;	
		tstart = time(0);	
		num_keys1 = compute_asift_keypoints(ipixels1_zoom1, widthList[i], heightList[i], num_of_tilts, verb, keys1, siftparameters);
		tend = time(0);
		sout << "	Keypoints computation accomplished in " << difftime(tend, tstart) << " seconds." << endl;
		
		keypoints.push_back(keys1);
		// ----------------------------------------------------------
		/*
		// Read image i
		float * iarr2;
		cimg_library::CImg<float> img2((filepath + filenames[i]).c_str());
		if(img2.max() <= 255.)
			img2 = img2 << 4;

		size_t w2=img2.width(), h2=img2.height();
		iarr2 = img2.data();
		std::vector<float> ipixels2(iarr2, iarr2 + w2 * h2);

		///// Resize the images to area wS*hS in remaining the apsect-ratio	
		///// Resize if the resize flag is not set or if the flag is set unequal to 0		
		std::vector<float> ipixels1_zoom2;
		zoomImage(ipixels2, ipixels1_zoom2, w2, h2, widthList2, heightList2, zoom2);

		vector< vector< keypointslist > > keys2;	
		int num_keys2=0;
	
		
		keypoints2.push_back(keys2);
		*/
	}

	
	// Create vectors to keep relation between Correspondence<T>
	// inside the correspondence list, and the keypoints list.
	vector< vector < keypoint > > assignedKeypoints ;
	vector< vector< size_t > > assignedKeypointIndices ;
	assignedKeypoints.resize(this->numberOfFiles);
	assignedKeypointIndices.resize(this->numberOfFiles);

	
	sout << "Matching keypoints..." << std::endl;

	// Calculate ASIFT matches between all keypoints
	for(size_t i = 0; i < this->numberOfFiles; i++)
	{	

		sout << "Choosing reference frame " << i << std::endl;

		calcCurrent++;

		// Calculate continue condition for loop 2
		size_t rangeBorder = i + this->matchRange();
		if(rangeBorder > this->numberOfFiles)
			rangeBorder = this->numberOfFiles;
		
		sout << "Matching up till frame " << rangeBorder << std::endl;
		sout << "matchRange() = " << this->matchRange() << std::endl;

		for(size_t j = i+1; j < rangeBorder; j++)
		{

			sout << "-----------------------------------" << std::endl;
			sout << "Matching keypoints between image "
				<< i << " and " << j << std::endl;

			// Match keys i and j
			matchingslist matchings = matchFrames(keypoints,i,j);
			sout << "match1" << std::endl;
			//matchingslist matchings2 = matchFrames(keypoints2,i,j);
			//sout << "match2" << std::endl;

			//matchings.insert(matchings.end(),matchings2.begin(),matchings2.end());

			// Continue with next iteration, if not enough matches between both
			// frames were found
			if(nMinCorr() > matchings.size()){continue;}

			// Read in according images
			cimg_library::CImg<float> img1 = getImg(i) ;
			cimg_library::CImg<float> img2 = getImg(j) ;

			// If no matches have been calculated before, just save them
			// to the output slot
			if( corrOut.size() == 0 )
			{

				sout << "Save matchings between image 0 and 1 in "
					<< "correspondence list." << std::endl;
				matchingslist::iterator ptrA = matchings.begin();
				size_t numberOfMatches = matchings.size();
				for( size_t k = 0; k < numberOfMatches; k++, ptrA++ )
				{
					float rgb_r, rgb_g, rgb_b;
					T px = zoom[i]*ptrA->first.x, py = zoom[i]*ptrA->first.y;
					getRGB(rgb_r, rgb_g, rgb_b, px, py, img1) ;
					hekate::Correspondence<T> pt1(px, py, 0, 1, k, i, false, rgb_r, rgb_g, rgb_b);
					T px2 = zoom[j]*ptrA->second.x, py2 = zoom[j]*ptrA->second.y;
					getRGB(rgb_r, rgb_g, rgb_b, px2, py2, img2) ;
					hekate::Correspondence<T> pt2(px2, py2, 0, 1, k, j, false, rgb_r, rgb_g, rgb_b);
					// Push back points the output correspondence list
					corrOut.push_back(pt1);
					corrOut.push_back(pt2);
					// Save keypoint and generated index
					assignedKeypointIndices[i].push_back(k);
					assignedKeypointIndices[j].push_back(k);
					assignedKeypoints[i].push_back(ptrA->first);
					assignedKeypoints[j].push_back(ptrA->second);

				}
			}
			// If not the first iteration, check if keypoints already exist
			// in the correspondence list. If yes, use this indices, otherwise
			// create new points in correspondence list
			else
			{

				// Take calculation time for match comparison
				time_t tstart, tend;
				tstart = time(0);
				
				// Iterate over every new match
				int newPoints = 0, removedPoints = 0;
				matchingslist::iterator ptrA = matchings.begin();
				size_t numberOfMatches = matchings.size();
				for(size_t k=0; k < numberOfMatches; k++, ptrA++)		
				{

					bool exist_i = false, exist_j = false ;
					size_t index_i, index_j ;


					if(assignedKeypoints[i].size() != assignedKeypointIndices[i].size()
						  || assignedKeypoints[j].size() != assignedKeypointIndices[j].size())
						throw(std::runtime_error("assignedKeypoints inconsistent size"));
					
					//sout << "assignedKeypoints[i].size() == " << assignedKeypoints[i].size() << std::endl;
					//sout << "assignedKeypoints[j].size() == " << assignedKeypoints[j].size() << std::endl;

					// Check if keypoint i is already in correspondence list
					for( size_t keyIt = 0; keyIt < assignedKeypoints[i].size(); keyIt++)
					{
						if(ptrA->first.x == assignedKeypoints[i][keyIt].x
							&& ptrA->first.y == assignedKeypoints[i][keyIt].y 
							&& ptrA->first.scale == assignedKeypoints[i][keyIt].scale)
						{
							exist_i = true;
							index_i = assignedKeypointIndices[i][keyIt];
							//sout << "consistence in assignedKeypointIndices[" << i << "][" << keyIt << "]" << std::endl;
							break;
						}	
					}

					// Check if keypoint j is already in correspondence list
					for( size_t keyIt = 0; keyIt < assignedKeypoints[j].size(); keyIt++)
					{
						if(ptrA->second.x == assignedKeypoints[j][keyIt].x
							&& ptrA->second.y == assignedKeypoints[j][keyIt].y
							&& ptrA->second.scale == assignedKeypoints[j][keyIt].scale)
						{
							exist_j = true;
							index_j = assignedKeypointIndices[j][keyIt];
							//sout << "consistence in assignedKeypointIndices[" << j << "][" << keyIt << "]" << std::endl;
							break;
						}	
					}


					// If keypoint i is already in correspondence list
					// but keypoint j is not
					if( exist_i && !exist_j )
					{
						// Add keypoint j with point index of existing
						// keypoint i

						float rgb_r, rgb_g, rgb_b;
						T px2 = zoom[j]*ptrA->second.x, py2 = zoom[j]*ptrA->second.y;
						getRGB(rgb_r, rgb_g, rgb_b, px2, py2, img2) ;
						hekate::Correspondence<T> pt2(px2, py2, 0, 1, index_i, j, false, rgb_r, rgb_g, rgb_b);

						// Push back point j the output correspondence list
						corrOut.push_back(pt2);
						//sout << pt2 << std::endl;

						// Save keypoint j and generated index
						assignedKeypointIndices[j].push_back(index_i);
						assignedKeypoints[j].push_back(ptrA->second);

					}
					// If keypoint i is not in correspondence list
					// but keypoint j is
					else if( !exist_i && exist_j )
					{
						// Add keypoint i with point index of existing
						// keypoint j

						float rgb_r, rgb_g, rgb_b;
						T px = zoom[i]*ptrA->first.x, py = zoom[i]*ptrA->first.y;
						getRGB(rgb_r, rgb_g, rgb_b, px, py, img1) ;
						hekate::Correspondence<T> pt1(px, py, 0, 1, index_j, i, false, rgb_r, rgb_g, rgb_b);

						// Push back points the output correspondence list
						corrOut.push_back(pt1);
						//sout << pt1 << std::endl;

						// Save keypoint and generated index
						assignedKeypointIndices[i].push_back(index_j);
						assignedKeypoints[i].push_back(ptrA->first);

					}
					// If both keypoints already are in correspondence list
					else if( exist_i && exist_j)
					{
						// Remove all points with indices index_i and index_j
						// to avoid outliers
						if( index_i != index_j )
						{

							removedPoints++;

							/*
							sout << "WARNING: Matched keypoints already "
								<< "have been attached to correspondence list "
								<< "but with different point indices!" << std::endl;
							sout << "Removing all points with one of these indices to avoid outliers" << std::endl;
							*/
							
							for(size_t iter1=0; iter1<assignedKeypointIndices.size(); iter1++)
							{
								for(size_t iter2=0; iter2<assignedKeypointIndices[iter1].size(); iter2++)
								{
									if(assignedKeypointIndices[iter1][iter2] == index_i
										|| assignedKeypointIndices[iter1][iter2] == index_j)
									{
										assignedKeypoints[iter1][iter2].x = -1;
										assignedKeypoints[iter1][iter2].y = -1;
									}
								}
							}
							auto it2 = corrOut.begin() ;
							for(; it2 != corrOut.end(); it2++)
							{
								if(it2->pointIndex == index_i)
									const_cast<Correspondence<T>& >(*it2).outlier = true;
								else if(it2->pointIndex == index_j)
									const_cast<Correspondence<T>& >(*it2).outlier = true;
							}
							
						}
					}
					// If both keypoints do not exist in 
					// correspondence list
					else
					{

						// Add both keypoints with new point index
						newPoints++;

						size_t newIndex = corrOut.size();
						float rgb_r, rgb_g, rgb_b;
						T px = zoom[i]*ptrA->first.x, py = zoom[i]*ptrA->first.y;
						getRGB(rgb_r, rgb_g, rgb_b, px, py, img1) ;
						hekate::Correspondence<T> pt1(px, py, 0, 1, newIndex, i, false, rgb_r, rgb_g, rgb_b);

						T px2 = zoom[j]*ptrA->second.x, py2 = zoom[j]*ptrA->second.y;
						getRGB(rgb_r, rgb_g, rgb_b, px2, py2, img2) ;
						hekate::Correspondence<T> pt2(px2, py2, 0, 1, newIndex, j, false, rgb_r, rgb_g, rgb_b);

						// Push back points the output correspondence list
						corrOut.push_back(pt1);
						corrOut.push_back(pt2);
						//sout << pt1 << std::endl;
						//sout << pt2 << std::endl;

						// Save keypoint and generated index
						assignedKeypointIndices[i].push_back(newIndex);
						assignedKeypointIndices[j].push_back(newIndex);
						assignedKeypoints[i].push_back( ptrA->first );
						assignedKeypoints[j].push_back( ptrA->second );

					}

				}

				sout << "	" << numberOfMatches-newPoints << " point correspondences consistent with old points "
					<< ". " << newPoints << " new points have been found." << std::endl;
				sout << "\t" << removedPoints << " points have been removed, due to inconsistent matches." << std::endl;

				tend = time(0);
				sout << "	Comparing new matchings with existing ones took " << difftime(tend, tstart) 
					<< " seconds." << endl;

			}
				
		}
		
	}

	// Remove doubles
	_numberOfCorrespondences = corrOut.size();
	sout << "Removed " << corrOut.removeDoubles() << " point doubles." << std::endl;
	sout << "Write " << _numberOfCorrespondences << " points to output slot." << std::endl;

}


/// match keypoints between frame 1 and frame 2
template <typename T>
matchingslist FeatureSequenceGPUASIFT<T>::matchFrames(vector< vector< vector< keypointslist > > > &keypoints,
		int frame1, int frame2)
{

	// Calculate matches
	time_t tstart, tend;
	tstart = time(0);
	matchingslist match;	
	int num_matchings;

	num_matchings = compute_asift_matches(num_of_tilts, num_of_tilts, widthList[frame1], heightList[frame1], widthList[frame2], 
											heightList[frame2], verb, keypoints[frame1], keypoints[frame2], match, siftparameters);
	
	sout << "compute_asift_matches ended" << std::endl;
	tend = time(0);
	sout << "	" << num_matchings << " Keypoints matched in " << difftime(tend, tstart) << " seconds." << endl;

	return match;

}


/// Scale image for faster calculation
template <typename T>
void FeatureSequenceGPUASIFT<T>::zoomImage(std::vector<float> &ipixels1, 
			std::vector<float> &ipixels1_zoom, int w1, int h1, 
			vector<int> &wSList, vector<int> &hSList, vector<float> &zoom){

	float zoom1=0;	
	int wS1=0, hS1=0;
			
	if(this->scaleFrames())
	{
	
		float InitSigma_aa = 1.6;
		
		float fproj_p, fproj_bg;
		char fproj_i;
		float *fproj_x4, *fproj_y4;
		int fproj_o;

		fproj_o = 3;
		fproj_p = 0;
		fproj_i = 0;
		fproj_bg = 0;
		fproj_x4 = 0;
		fproj_y4 = 0;
	
		float areaS =  scaleX() * scaleY();

		// Resize image 1 
		float area1 = w1 * h1;
		zoom1 = sqrt(area1/areaS);
		
		wS1 = (int) (w1 / zoom1);
		hS1 = (int) (h1 / zoom1);
		
		int fproj_sx = wS1;
		int fproj_sy = hS1;     
		
		float fproj_x1 = 0;
		float fproj_y1 = 0;
		float fproj_x2 = wS1;
		float fproj_y2 = 0;
		float fproj_x3 = 0;	     
		float fproj_y3 = hS1;
		
		// Anti-aliasing filtering along vertical direction 
		if ( zoom1 > 1 )
		{
			float sigma_aa = InitSigma_aa * zoom1 / 2;
			GaussianBlur1D(ipixels1,w1,h1,sigma_aa,1);
			GaussianBlur1D(ipixels1,w1,h1,sigma_aa,0);
		}
			
		// simulate a tilt: subsample the image along the vertical axis by a factor of t.
		ipixels1_zoom.resize(wS1*hS1);
		fproj (ipixels1, ipixels1_zoom, w1, h1, &fproj_sx, &fproj_sy, &fproj_bg, &fproj_o, &fproj_p, 
			   &fproj_i , fproj_x1 , fproj_y1 , fproj_x2 , fproj_y2 , fproj_x3 , fproj_y3, fproj_x4, fproj_y4); 

	}
	else 
	{
		ipixels1_zoom.resize(w1*h1);	
		ipixels1_zoom = ipixels1;
		wS1 = w1;
		hS1 = h1;
		zoom1 = 1;

	}
	
	wSList.push_back(wS1);
	hSList.push_back(hS1);
	zoom.push_back(zoom1);

}

float round(float i, int decPlace){
	float fac = std::pow(float(10), float(decPlace));
	float out = i * fac;
	out = floor(out);
	out /= fac;
	return out;
}

template<typename T>
cimg_library::CImg<float> FeatureSequenceGPUASIFT<T>::getImg(size_t i)
{
	cimg_library::CImg<float> img1((filepath + filenames[i]).c_str());
	img1.resize(img1.width(),img1.height(),1,3);
	if(img1.max() > 255.){img1 = img1 >> 4;}

	int thx = 2000;
	int thy = 2000;
	int iw = img1.width(), ih = img1.height();

	// Scale images if taller than threshold
	if(iw > thx || ih > thy){
		sout << "ATTENTION: IMAGE IS TOO BIG FOR ASIFT! Scaling..." << std::endl;
		if(iw>ih){
			float c = float(thx)/float(iw);
			img1.resize(2000,int(c*ih));
		}
		else{
			float c = float(thy)/float(ih);
			img1.resize(int(c*iw),2000);
		}
	}

	return img1;
}

template<typename T>
void getRGB(float &r, float &g, float &b, 
	const float px, const float py, const cimg_library::CImg<T> &img)
{
	r = img((int)px,(int)py,0,0);
	g = img((int)px,(int)py,0,1);
	b = img((int)px,(int)py,0,2);
}

#endif /* _FEATURESEQUENCEASIFT_HXX_ */
