/*  Copyright (C) 2012 Martin Treiber

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeatureSequenceGPUSift.hxx
 *  Implementation of the parameter class FeatureSequenceGPUSift.
 *  \author <a href="mailto:martin.treiber.hd@gmx.de">
 *      Martin Treiberr</a>
 *  \date 08.03.2012
 */

#ifndef _FEATURESEQUENCEGPUSIFT_HXX_
#define _FEATURESEQUENCEGPUSIFT_HXX_

#ifdef _OPENMP
#include <omp.h>
#endif

#include <charon-core/ParameteredObject.hxx>
#include "hekate/features/FeatureSequenceGPUSift.h"
#include <hekate/features/FeatureSequence.hxx>

using namespace hekate;


template <typename T>
FeatureSequenceGPUSift<T>::FeatureSequenceGPUSift(const std::string& name) :
		FeatureSequence<T>(
			"FeatureSequenceGPUSift", name,
			"<h2>GPUSIFT version of the sequence feature "
			"extractor</h2><br>"
			"Calculates feature correspondences of an image sequence. For "
			"this, the GPUSIFT version is used."
		)
{
	ParameteredObject::_addParameter< bool > (useCUDA, "useCUDA", "use CUDA (or GLSL)", false);
	ParameteredObject::_addParameter< bool > (useDI, "useDI", "use Dynamic Array Indexing", false);
	ParameteredObject::_addParameter< bool > (setTight, "setTight", "use tight storage control (not recommended)", false);
	ParameteredObject::_addParameter< bool > (useSubPixel, "useSubPixel", "use Sub-Pixel Localization", true);
	ParameteredObject::_addParameter< int > (startingOctave, "startingOctave", "Octave to start with", -1, "int");
	ParameteredObject::_addParameter< int > (DOGOctave, "DOGOctave", "DOG Levels in an Octave", 5, "int");
	ParameteredObject::_addParameter< float > (DOGTreshold, "DOGTreshold", "DOG Treshold", 0.1f, "float");
	ParameteredObject::_addParameter< float > (EdgeTreshold, "EdgeTreshold", "Edge Treshold", 5.0f, "float");
	ParameteredObject::_addParameter< int > (MaxMatches, "MaxMatches", "Maximal Number of Matches to search", 4096, "int");
	
	ParameteredObject::_setTags("hekate") ;

}

template <typename T>
void FeatureSequenceGPUSift<T>::_calculate() {
	//Initializing SIFTGPU
	SiftGPU sift;
	//Check for some arguments that must be positive:
	if(DOGTreshold()<0)DOGTreshold()=0.02f;
	if(EdgeTreshold()<0)EdgeTreshold()=10.0f;
	sout << "Start Parsing arguments" <<std::endl;
	char* args[15];
	args[0]=new char[4];
	sprintf(args[0],"-fo");
	args[1]=new char[3+(int)log10((float)abs(startingOctave()))];
	sprintf(args[1],"%d",startingOctave());
	args[2]=new char[3];
	sprintf(args[2],"-v");
	args[3]=new char[2];
	sprintf(args[3],"1");
	if(useCUDA())
	{
		args[4]=new char[6];
		sprintf(args[4],"-cuda");
	}else
	{
		args[4]=new char[1];
		sprintf(args[4],"");
	}
	args[5]=new char[3];
	sprintf(args[5],"-d");
	args[6]=new char[2+(int)log10((float)abs(DOGOctave()))];
	sprintf(args[6],"%d",abs(DOGOctave));
	args[7]=new char[3];
	sprintf(args[7],"-t");
	args[8]=new char[3+abs((int)abs(log10(DOGTreshold())))+2*sizeof(float)];
	sprintf(args[8],"%f",DOGTreshold());
	for(int i=0;args[8][i]!= 0 ;++i){if(args[8][i] == '.')args[8][i]=',';}
	args[9]=new char[3];
	sprintf(args[9],"-s");
	args[10]=new char[2];
	if(useSubPixel())
	{
		sprintf(args[10],"1");
	}else
	{
		sprintf(args[10],"0");
	}
	args[11]=new char[7];
	if(setTight())
	{
		sprintf(args[11],"-tight");
	}else
	{
		sprintf(args[11],"");
	}
	args[12]=new char[4];
	if(useDI())
	{
		sprintf(args[12],"-di");
	}else
	{
		sprintf(args[12],"");
	}
	args[13]=new char[3];
	sprintf(args[13],"-e");
	args[14]=new char[3+abs((int)abs(log10(EdgeTreshold())))+2*sizeof(float)];
	sprintf(args[14],"%f",EdgeTreshold());
	for(int i=0;args[14][i]!= 0 ;++i){if(args[14][i] == '.')args[14][i]=',';}
	sout << "Passing Arguments to SiftGPU" << std::endl;
	sift.ParseParam(15,args);
	sout << "Arguments";
	delete[] args[0]; delete[] args[1]; delete[] args[2]; delete[] args[3];
	delete[] args[4]; delete[] args[5]; delete[] args[6]; delete[] args[7];
	delete[] args[8]; delete[] args[9]; delete[] args[10]; delete[] args[11];
	delete[] args[12]; delete[] args[13]; delete[] args[14];
	sout << " Parsed" << std::endl;
	int support = sift.CreateContextGL();
	sout << "Checking for GL-Support"<<std::endl;
	if(support != SiftGPU::SIFTGPU_FULL_SUPPORTED) return;
	sout<< "Fully compatible"<<std::endl;
	//get the vectors for keypoints and descriptors:
	std::vector<std::vector<float> > descriptors(filenames.size());
	std::vector<std::vector<SiftGPU::SiftKeypoint> > keys(filenames.size());
	//Get Features for every image
	unsigned int NumCalculateFrames=(filenames.size()<maximalFrames()?filenames.size():maximalFrames()); 
	filenames.resize(NumCalculateFrames);
	for(unsigned int i=0;i<NumCalculateFrames;++i)
	{
		time_t tstart, tend;
		sout<<"Running Sift on Image "<<i<<":\t"<<filenames[i]<<std::endl;
		filenames[i]=filepath+filenames[i];
		tstart = time(0);
		sift.RunSIFT(filenames[i].c_str());
		int num = sift.GetFeatureNum();
		descriptors[i].resize(128*num);
		keys[i].resize(num);
		//Save Feature Vector
		sift.GetFeatureVector(&keys[i][0],&descriptors[i][0]);
		tend=time(0);
		sout<<"Keypoints calculated in "<<difftime(tend,tstart)<<" seconds\n"<<std::endl;
	}
	//map for index storing purpose
	map<int, map<int, map<int, int> > > Index;
	hekate::CorrespondenceList<T> list;
	SiftMatchGPU matcher(MaxMatches());
	int useIndex, currentPointIndex=0;
	if(matcher.VerifyContextGL() == 0) return;
	//Matching for every pair of images:
	for(unsigned int i=0;i<filenames.size()-1;++i)
		{
			cimg_library::CImg<T> img1(filenames[i].c_str());
			img1.resize(img1.width(),img1.height(),1,3);

			matcher.SetDescriptors(0, keys[i].size(), &descriptors[i][0]);
			for(unsigned int j = i+1; j<NumCalculateFrames;++j)
			{
				cimg_library::CImg<T> img2(filenames[j].c_str());
				img2.resize(img2.width(),img2.height(),1,3);

				time_t tstart, tend;
				tstart=time(0);
				matcher.SetDescriptors(1, keys[j].size(), &descriptors[j][0]);
				//at this Point, write the Matches into the correspondences
				//Save Coresspondences:
				int (*matchBuf)[2]=new int[keys[i].size()][2];
				unsigned int numMatch = matcher.GetSiftMatch(MaxMatches(),matchBuf);
				unsigned int numMatchInlier = 0;
				if(numMatch>=minimalNumberOfCorrs())
				for(unsigned int k=0;k<numMatch;++k)
				{
					//Checking for the existence of found matches
					if(Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[k][0]].y]
						&& Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y]
						&& Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[i][0]].y]!=Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y])
					{
						list(i,Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[k][0]].y]).outlier = true;
						list(j,Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y]).outlier = true;
					}
					else if(Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[k][0]].y]
						&& Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y])
					{
					}
					else if(Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[k][0]].y])
					{
						numMatchInlier++;
						float rgb_r, rgb_g, rgb_b;
						T px = keys[j][matchBuf[k][1]].x, py = keys[j][matchBuf[k][1]].y;
						getRGB(rgb_r, rgb_g, rgb_b, px, py, img2) ;

						useIndex=Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[k][0]].y];
						Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y]=useIndex;
						list.push_back(hekate::Correspondence<T>(keys[j][matchBuf[k][1]].x,keys[j][matchBuf[k][1]].y,0,1,useIndex,j,false,rgb_r, rgb_g, rgb_b));
					}
					else if(Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y])
					{
						numMatchInlier++;
						float rgb_r, rgb_g, rgb_b;
						T px = keys[i][matchBuf[k][0]].x, py = keys[i][matchBuf[k][0]].y;
						getRGB(rgb_r, rgb_g, rgb_b, px, py, img1) ;

						useIndex=Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y];
						Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[k][0]].y]=useIndex;
						list.push_back(hekate::Correspondence<T>(keys[i][matchBuf[k][0]].x,keys[i][matchBuf[k][0]].y,0,1,useIndex,i,false,rgb_r, rgb_g, rgb_b));
					}
					else
					{
						numMatchInlier++;
						float rgb_r, rgb_g, rgb_b;
						T px = keys[i][matchBuf[k][0]].x, py = keys[i][matchBuf[k][0]].y;
						getRGB(rgb_r, rgb_g, rgb_b, px, py, img1) ;

						++currentPointIndex;
						useIndex=currentPointIndex;
						Index[i][keys[i][matchBuf[k][0]].x][keys[i][matchBuf[k][0]].y]=useIndex;
						Index[j][keys[j][matchBuf[k][1]].x][keys[j][matchBuf[k][1]].y]=useIndex;

						list.push_back(hekate::Correspondence<T>(keys[i][matchBuf[k][0]].x,keys[i][matchBuf[k][0]].y,0,1,useIndex,i,false,rgb_r, rgb_g, rgb_b));
						
						px = keys[j][matchBuf[k][1]].x, py = keys[j][matchBuf[k][1]].y;
						getRGB(rgb_r, rgb_g, rgb_b, px, py, img2) ;
						list.push_back(hekate::Correspondence<T>(keys[j][matchBuf[k][1]].x,keys[j][matchBuf[k][1]].y,0,1,useIndex,j,false,rgb_r, rgb_g, rgb_b));
					}
				}
				delete[] matchBuf;
				tend=time(0);
				sout<<numMatchInlier << "/" << numMatch <<" Inlier Matches found for image "<<i<<" and "<<j<<" in "<<difftime(tend,tstart)<<" seconds\n"<<std::endl;
			}
	}
	list.removeDoubles();
	correspondences() = list;
	sout << list.size() << " points saved to output slot." << std::endl;
	sout << "Finishing FeatureSequenceGPUSift"<<std::endl;
	imgPaths()=filenames;
		
}

template<typename T>
void getRGB(float &r, float &g, float &b, 
	const float px, const float py, const cimg_library::CImg<T> &img)
{
	r = img((int)px,(int)py,0,0);
	g = img((int)px,(int)py,0,1);
	b = img((int)px,(int)py,0,2);
}




#endif /* _FEATURESEQUENCEGPUSIFT_HXX_ */
