/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeatureSequenceGPUSiftFM.hxx
 *  Implementation of the parameter class FeatureSequenceGPUSiftFM.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 18.04.2013
 */

#ifndef _FEATURESEQUENCEGPUSIFTFM_HXX_
#define _FEATURESEQUENCEGPUSIFTFM_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "FeatureSequenceGPUSiftFM.h"

template <typename T>
FeatureSequenceGPUSiftFM<T>::FeatureSequenceGPUSiftFM(const std::string& name) :
		TemplatedParameteredObject<T>(
			"FeatureSequenceGPUSiftFM", name,
			"<h2>Determines SIFT features and matches with epipolar "
			"constraint</h2><br>"
			"Determines SIFT features using SIFTGPU "
			"(http://www.cs.unc.edu/~ccwu/siftgpu/). The matching process "
			"uses a calculated fundamental matrix (OpenCV) to prevent "
			"outliers."
		)
{
	ParameteredObject::_addInputSlot(
		imagePaths, "imagePaths",
		"Path to images files",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		correspondences, "correspondences",
		"Calculated correspondences",
		"CorrespondenceList");
	ParameteredObject::_addOutputSlot(
		fundamentalMatrices, "fundamentalMatrices",
		"Calculated fundamental matrices",
		"IndexedMatrixList");
	
	ParameteredObject::_addParameter< int >(
		stepWidth, "stepWidth",
		"Step width of chosen image sequence to process",
		1, "int");
	ParameteredObject::_addParameter< float >(
		startingOctave, "startingOctave",
		"Octave to start with",
		-1, "float");
	ParameteredObject::_addParameter< float >(
		dogOctave, "dogOctave",
		"DOG Levels in an Octave",
		5, "float");
	ParameteredObject::_addParameter< float >(
		dogThreshold, "dogThreshold",
		"DOG Treshold",
		0.1, "float");
	ParameteredObject::_addParameter< float >(
		edgeThreshold, "edgeThreshold",
		"Edge Treshold",
		5, "float");
	ParameteredObject::_addParameter< int >(
		maxMatches, "maxMatches",
		"Maximal Nmber of Matches to search for",
		4096, "int");
	ParameteredObject::_addParameter< int >(
		matchRange, "matchRange",
		"Number of subsequent images to match",
		-1, "int");
	
	ParameteredObject::_setTags("Hekate") ;

}

template <typename T>
void FeatureSequenceGPUSiftFM<T>::execute() {	
	
	const unsigned int steps = this->stepWidth();
	std::vector< std::string > imgSequence;
	for(size_t i = 0; i < this->imagePaths().size(); i += this->stepWidth() )
	{
		imgSequence.push_back( this->imagePaths()[i] );
	}

	// number of frames
	const unsigned int numFrames = imgSequence.size();	

    SiftGPU sift;								// sift instance
    SiftMatchGPU matcher(4096);					// matcher instance

    //Create a context for computation, and SiftGPU will be initialized automatically 
    //The same context can be used by SiftMatchGPU
	sift.CreateContextGL();

    //Verify current OpenGL Context and initialize the Matcher;
    //If you don't have an OpenGL Context, call matcher->CreateContextGL instead;
    matcher.VerifyContextGL(); //must call once
	matcher.SetMaxSift( this->maxMatches() );

	std::vector< std::vector<float > > descriptors(numFrames);				// feature descriptors
    std::vector< std::vector< SiftGPU::SiftKeypoint > > keys(numFrames);	// feature points
    std::vector< int > numKeys;												// number of feature points
	/// vector of SiftGPU matches
	/**
		Dimension1: number of images
		Dimension2: number of key in acc img
		Dimension3: number of ids key is assigned to
		Example:	keyIds[0][5] = (1,6,2)
					--> point 5 of img 0 is assigned to corr 1, 6 and 2
					--> (inliers should only have one assigned corr id)
	*/
	std::vector< std::vector< std::vector< int> > > keyIds(numFrames);

	// hekate correspondence list used to store resulting points
	hekate::CorrespondenceList<T> corr;

	// ----------------------------------  KEY POINT DETECTION -------------------------------------------

	sout << "Run SIFT Key Point Detection" << std::endl;

	for( size_t i = 0; i < numFrames; i++ )
	{
		const std::string imgpath = imgSequence[i];
		//sout << "Running Sift on Image[" << i << "]:\t"<< imgpath << std::endl;

		// Run SIFT and find feature points on image i
		sift.RunSIFT( imgpath.c_str() );

		// Set up descriptor and keypoint lists
		unsigned int num = sift.GetFeatureNum();
		numKeys.push_back( num );
		descriptors[i].resize(128*num);
		keys[i].resize(num);
		
		// Save feature points
		sift.GetFeatureVector(&keys[i][0],&descriptors[i][0]);

		// resize keyIds of img(i) to number of points of img(i)
		keyIds[i].resize(num);

	}

	
	// ----------------------------------  MATCHING PROCESS -------------------------------------------

	sout << "Run matching algorithm" << std::endl;
	
	// point index iterator
	int ptIndex = 0;

	for( size_t i = 0; i < numFrames; i++ )
	{
		// number of frames to match
		int matchRange = std::min( std::max( int(i+1), int( i+this->matchRange() ) ), int (numFrames) );

		if( this->matchRange() == -1 )
			matchRange = numFrames;

		for( size_t j = i+1; j < matchRange; j++ )
		{

			const cimg_library::CImg<float> img1(imgSequence[i].c_str());
			const cimg_library::CImg<float> img2(imgSequence[j].c_str());

			sout << "Match features between frames (" << i << "/" << j << "):" << std::endl;

			// Set up matcher
			matcher.SetDescriptors(0, numKeys[i], &descriptors[i][0]); //image i
			matcher.SetDescriptors(1, numKeys[j], &descriptors[j][0]); //image i+1

			//match and get result.    
			int (*match_buf)[2] = new int[ numKeys[i] ][2];
			int num_match = matcher.GetSiftMatch( numKeys[i], match_buf);
			int num_match_inlier = 0;
    
			std::vector<cv::Point2d> feat1, feat2;

			//enumerate all the feature matches
			for(int k  = 0; k < num_match; ++k)
			{				
				feat1.push_back( cv::Point2d( keys[i][match_buf[k][0]].x, keys[i][match_buf[k][0]].y ) );
				feat2.push_back( cv::Point2d( keys[j][match_buf[k][1]].x, keys[j][match_buf[k][1]].y ) );
			}
			
			cv::Mat mask;
			cv::Mat fmCv = cv::findFundamentalMat(feat1,feat2,mask,cv::RANSAC);
				
			// Convert outlier mask to CImg (1 == inlier, 0 == outlier, width == 2, height == numberOfMatches)
			cimg_library::CImg<double> maskCImg = openCvUtil::Mat2CImg<double>(mask);

			cimg_library::CImg<double> fm = openCvUtil::Mat2CImg<double>(fmCv);
			IndexedMatrix fm2( fm );
			fm2.setIndex( i, j );
			this->fundamentalMatrices().push_back( fm2 );
			

			for( size_t k = 0; k < num_match; k++ )
			{
				// If point pair is epipolar inlier
				if( maskCImg(0,k) == 1 )
				{
					num_match_inlier++;
					
					// Get ids and coordinates
					int id1 = match_buf[k][0], id2 = match_buf[k][1];
					float x1 =  keys[i][id1].x, y1 = keys[i][id1].y;
					float x2 =  keys[j][id2].x, y2 = keys[j][id2].y;

					// Check if points have been already matched							
					bool exist1 = keyIds[i][id1].size();
					bool exist2 = keyIds[j][id2].size();

					// if none of both points already exists, create new correspondence
					if( !exist1 && !exist2 )
					{
						// Get RGB values
						float R1, G1, B1, R2, G2, B2;
						getRGB(img1, x1, y1, R1, G1, B1);
						getRGB(img2, x2, y2, R2, G2, B2);

						// Save correspondence
						keyIds[i][id1].push_back( ptIndex );
						keyIds[j][id2].push_back( ptIndex );

						// Save correspondences
						corr.push_back( hekate::Correspondence<T>(x1, y1, 0, 1, ptIndex, i, 0, R1, G1, B1) );
						corr.push_back( hekate::Correspondence<T>(x2, y2, 0, 1, ptIndex, j, 0, R2, G2, B2) );
						ptIndex++;
					}

					else if( exist1 && !exist2 )
					{
						// Get RGB values
						float R, G, B;
						getRGB(img2, x2, y2, R, G, B);

						// Add key2 to all correspondences of key1
						for( size_t matchid = 0; matchid < keyIds[i][id1].size(); matchid++)
						{
							int key1index = keyIds[i][id1][matchid];
							keyIds[j][id2].push_back( key1index );
							corr.push_back( hekate::Correspondence<T>(x2, y2, 0, 1, key1index, j, 0, R, G, B) );
						}
					}

					else if( !exist1 && exist2 )
					{
						// Get RGB values
						float R, G, B;
						getRGB(img1, x1, y1, R, G, B);

						// Add key1 to all correspondences of key2
						for( size_t matchid = 0; matchid < keyIds[j][id2].size(); matchid++)
						{
							int key2index = keyIds[j][id2][matchid];
							keyIds[i][id1].push_back( key2index );
							corr.push_back( hekate::Correspondence<T>(x1, y1, 0, 1, key2index, i, 0, R, G, B) );
						}
					}

					// if both points already exist, check if they belong to the same correspondence
					else if( exist1 && exist2 )
					{
						// SKIP SINCE BOTH POINTS ALREADY EXISTS


						// Skip, if points already uniquely matched
						//if(corr1.size() == 1 && corr2.size() == 1 && corr1[0] == corr2[0]) continue;
					}

				}
			}
			
			sout << "\t" << num_match << " matches found;\n";
			sout << "\t" << num_match - num_match_inlier << " outlier matches removed\n";
			sout << "\t" << num_match_inlier << " matches remaining\n";
						
			delete[] match_buf;
		
		}
	}

	this->correspondences() = corr;


}


/// Determine RGB values at pixel loation (x,y)
void getRGB(const cimg_library::CImg<float> &img, const float x, const float y, float &R, float &G, float &B)
{
	R = img.linear_atXYZC(x,y,0,0);
	G = img.linear_atXYZC(x,y,0,1);
	B = img.linear_atXYZC(x,y,0,2);
}

#endif /* _FEATURESEQUENCEGPUSIFTFM_HXX_ */

