/*  Copyright (C) 2011 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeatureSequenceSIFT.hxx
 *  Implementation of the parameter class FeatureSequenceSIFT.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 20.10.2011
 */

#ifndef _FEATURESEQUENCESIFT_HXX_
#define _FEATURESEQUENCESIFT_HXX_

#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cxcore.h>
#include <hekate/openCvUtil.hxx>

#include <charon-core/ParameteredObject.hxx>
#include <hekate/features/FeatureSequenceSIFT.h>
#include <hekate/features/FeatureSequence.hxx>
#include <hekate/Correspondence.hxx>

using namespace cv ;
using namespace openCvUtil ;
using namespace hekate;

template <typename T>
FeatureSequenceSIFT<T>::FeatureSequenceSIFT(const std::string& name) :
		FeatureSequence<T>(
			"FeatureSequenceSIFT", name,
			"<h2>Calculates correspondences between images of an image list "
			"by using the OpenCV implementation of SIFT</h2><br>"
			"Calculates correspondences between images of an image list by "
			"using the OpenCV implementation of SIFT"
		)
{
	ParameteredObject::_addParameter< int >(
		maxDistance, "paramMaxDistance",
		"Threshold for maximal distance between descriptors. (Matching parameter)",
		2, "int");
	ParameteredObject::_addParameter< double >(
		peakThreshold, "paramPeakThreshold",
		"The peak threshold filters peaks of the DoG scale space that are "
		"too small (in absolute value).",
		0.0, "double");
	ParameteredObject::_addParameter< double >(
		edgeThreshold, "paramEdgeThreshold",
		"The edge threshold eliminates peaks of the DoG scale space whose "
		"curvature is too small (such peaks yield badly localized frames).",
		90.0, "double");
	ParameteredObject::_addParameter< int >(
		octaves, "paramOctaves",
		"The number of a gaussian pyramid octaves that the detector uses. "
		"It is set to 4 by default. If you want to get very large features, "
		"use the larger value. If you want just small features, decrease it.",
		4, "int");
	ParameteredObject::_addParameter< int >(
		octaveLayers, "paramOctaveLayers",
		"The number of images within each octave of a gaussian pyramid. It is set to 2 by default.",
		3, "int");
	ParameteredObject::_addParameter< int >(
		firstOctave, "paramFirstOctave",
		"paramFirstOctave.",
		-1, "int");
	ParameteredObject::_addParameter< int >(
		angleMode, "paramAngleMode",
		"paramAngleMode.",
		0, "int");
}
		
template <typename T>
void FeatureSequenceSIFT<T>::_calculate() {

	int refFrame = 0;
	
	// Create SIFT detector
	cv::SiftFeatureDetector detector(peakThreshold,edgeThreshold,octaves,octaveLayers,firstOctave,angleMode);

	// Create SIFT feature extrator (converts SIFT features to universal features)
	cv::SiftDescriptorExtractor extractor;
	
	// Create Matcher
	//FlannBasedMatcher matcher;
	BruteForceMatcher< L2<float> > matcher;


	// Find features in reference frame ------------------------------------------
	cimg_library::CImg<T> img1((filepath + filenames[refFrame]).c_str());
	if( refFrame == 0 )
		imgPaths().push_back((filepath + filenames[refFrame]).c_str());

	// Convert CImg to OpenCV Mat
	Mat img1Mat = CImg2Mat(img1.get_normalize(0,255), CV_8U) ;
	
	// Find keypoints in reference frame
	std::vector<KeyPoint> keypoints_1;
	detector.detect( img1Mat, keypoints_1 );	
	sout << "	Found " << keypoints_1.size() << " keypoints in reference frame." << std::endl;

	// Gernerate universal OpenCV descriptor for SIFT keypoints
	Mat descriptors_1;
	extractor.compute( img1Mat, keypoints_1, descriptors_1 );


	// Outlier handling ----------------------------------------------------------

	// Create outlier mask
	typename std::vector<std::vector<std::vector<T> > > outlierMask, corr;

	// Define outlier points
	std::vector<T> outlier(3);
	outlier[0] = T(0); 
	outlier[1] = T(0);
	outlier[2] = T(0);

	// Reference list (first entry of outlierMask)
	std::vector< std::vector<T> > refPoints, outlierPoints;
	
	if(descriptors_1.rows != keypoints_1.size())
		throw(std::runtime_error("Size of keypoints and its decriptors are not equal!"));

	// Apply all features of reference frame to the reference list
	for(int i = 0; i<descriptors_1.rows; i++ ){
		std::vector<T> pt(3);
		pt[0] = T(keypoints_1[i].pt.x); 
		pt[1] = T(keypoints_1[i].pt.y);
		pt[2] = 1;
		refPoints.push_back(pt);
		outlierPoints.push_back(outlier);
	}
	outlierMask.push_back(refPoints);


	// Iterate over all subsequent frames j which were taken after ---------------
	// the current reference frame "refFrame"
	int noCorrs = 0;

	size_t numberOfFiles = filenames.size();

	for( size_t j=refFrame+1; j<numberOfFiles; j++ ){

		// Break iteration if number of frames reached limit.
		if(j-refFrame > maximalFrames)
			break;
		
		sout << "	Searching for correspondences between frame" << refFrame 
			<< " and frame" << j << ":" << std::endl;

		// Find features in corresponding frame ------------------
		cimg_library::CImg<T> img2((filepath + filenames[j]).c_str());
		Mat img2Mat = CImg2Mat(img2.get_normalize(0,255), CV_8U) ;
		std::vector<KeyPoint> keypoints_2;
		detector.detect( img2Mat, keypoints_2 );
		sout << "	Found " << keypoints_2.size() << " keypoints in frame " << j << std::endl;
		Mat descriptors_2;		
		extractor.compute( img2Mat, keypoints_2, descriptors_2 );
			
		// Find matches between reference frame and current corresponding frame j
		std::vector< DMatch > matches, goodMatches;
//		matcher.knnMatch( descriptors_1, descriptors_2, matches, 1 );
		matcher.match( descriptors_1, descriptors_2, matches );

		double max_dist = 0; double min_dist = 100;

		//-- Quick calculation of max and min distances between keypoints
		for( int i = 0; i < descriptors_1.rows; i++ ){ 
			double dist = matches[i].distance;
			if( dist < min_dist ) min_dist = dist;
			if( dist > max_dist ) max_dist = dist;
		}
	
		// Push good matches to outlierMask -----------------------------------
		std::vector<std::vector<T> > good_matches;
		size_t nGoodMatches = 0;

		// Find good matches
		good_matches = outlierPoints;
		for( size_t jj = 0; jj<matches.size(); jj++ ){
			if( matches[jj].distance < maxDistance() * min_dist ){ 

				// Increase number of good matches
				nGoodMatches++;	

				std::vector<T> pt(3);
				pt[0] = T(keypoints_2[matches[jj].trainIdx].pt.x); 
				pt[1] = T(keypoints_2[matches[jj].trainIdx].pt.y);
				pt[2] = 1;
				good_matches[matches[jj].queryIdx] = pt; 
				goodMatches.push_back(matches[jj]);

			}
		}
		// Push outliers to overall mask
		outlierMask.push_back(good_matches);

		sout << "	" << nGoodMatches << " correspondences found." << std::endl;
		

		// Increase noCorrs if not enough correspondences could be found ---------
		if(nGoodMatches < minimalNumberOfCorrs || nGoodMatches == 0){			
			noCorrs++;
			sout << "	Not enough correspondences found." << std::endl;

			// Break if number of frames which do not exhibit a minimal
			// numer of correspondences reached the limit maxFWC
			// and there was no single frame which has more correspondences
			// than the limit. In this case, the reference image will be removed
			// from the complete correspondence data since it is not useable.
			if(noCorrs > maxFWC && noCorrs == j-refFrame){
				sout << "	I could not be found a frame which exhibits"
					<< " the minimal number of correspondences regarding the"
					<< " the reference frame " << refFrame << "." << std::endl;
				sout << "	! Reference frame " << refFrame << " seems to be unusable"
					<< " for tracking purposes." << std::endl;
				//return 0;
			}
			// Break if number of frames which do not exhibit a minimal
			// numer of correspondences reached the limit maxFWC.
			// In this case, there is at least one additional frame which has
			// more than the minimal number of correspondences regarding
			// the reference image.
			else if(noCorrs > maxFWC){
				sout <<	"	Stop searching for corresponding frames regarding the reference image "
					<< refFrame << " since the number of frames which have less correspondences "
					<< " as the definced limit (" << maxFWC << ") was reached." << std::endl;
				break;
			}

		}
		else{
			
			// Check if there are enough common correspondences
			std::vector<std::vector<std::vector<T> > > outlierMaskTemp = outlierMask;
			removeOutliers(outlierMaskTemp);
			sout << "	Checking matches between new features and existing correspondences..." << std::endl;
			sout << "		" << outlierMaskTemp[0].size() << " common correspondences found." << std::endl;
			if(outlierMaskTemp[0].size() >= minimalNumberOfCorrs){
				corr = outlierMask;
				imgPaths().push_back((filepath + filenames[j]).c_str()); 
				/*
				Mat img_matches;
				drawMatches( img1Mat, keypoints_1, img2Mat, keypoints_2, goodMatches, img_matches );
				  //-- Show detected matches
				  imshow( "Good Matches", img_matches );
				*/
			}
			else{ 
				sout << "	Not enough common correspondences found ("
				<< outlierMaskTemp[0].size() << "/" << minimalNumberOfCorrs << ")." << std::endl;
			}
			/*
			else{
				sout << "	Not enough common correspondences. No more frames "
					<< "will be added to the sequence." << std::endl;
				break;
			}
			*/

		}
			
	}	

	// Delete outliers from outlier mask
	if(corr.size() > 1){
		sout << "Remove outliers" << std::endl;
		removeOutliers(corr);
	}

	sout << corr.size() << " frames could be matched." << std::endl;

	sout << corr[0].size() << " common correspondences could be found"
		<< " between all appropriate frames." << std::endl;
	
	sout << "Apply correspondences to output slot." << std::endl;
	
	hekate::CorrespondenceList<T> corrOut;
	for(size_t i=0; i<corr.size(); i++){
		for(size_t j=0; j<corr[i].size(); j++){
			corrOut.push_back(hekate::Correspondence<T>(corr[i][j], j, i)) ;
		}
	}
		
	correspondences() = corrOut;
	
	/*
	CorrespondenceList<T>::ListAsVec::iterator it = corrOut.begin() ;
	for(; it != corrOut.end(); it++)
	{
		sout << it->x() << ", " << it->y() << ", " << it->pointIndex << ", " << it->imgIndex << std::endl;
	}
	*/

	/*
			// Save matches in vectors
			std::vector<std::vector<T> > corrs1;
			std::vector<std::vector<T> > corrs2;
	
			for( int i=0; i < good_matches.size(); i++ ){ 

				/*
				sout << "	(" << i << ") Good Match " << " Keypoint 1: " << good_matches[i].queryIdx << " -- Keypoint 2: " 
					<< good_matches[i].trainIdx << "	" << keypoints_1[good_matches[i].queryIdx].pt.x << "/" 
					<< keypoints_1[good_matches[i].queryIdx].pt.y	<< ", " << keypoints_2[good_matches[i].trainIdx].pt.x 
					<< "/" << keypoints_2[good_matches[i].trainIdx].pt.y << std::endl; 
					*/
				/*
				std::vector<T> pt1(3), pt2(3);

				// Add reference points if this is the first calculation
				if(j=refFrame+1){
					pt1[0] = T( keypoints_1[good_matches[i].queryIdx].pt.x );
					pt1[1] = T( keypoints_1[good_matches[i].queryIdx].pt.y );
					pt1[2] = T( 1 );
					corrs1.push_back(pt1);
				}

				pt2[0] = T( keypoints_2[good_matches[i].trainIdx].pt.x );
				pt2[1] = T( keypoints_2[good_matches[i].trainIdx].pt.y );
				pt2[2] = T( 1 );

				corrs2.push_back(pt2);

			}
	
			if(j=refFrame+1){
				this->correspondences().push_back(corrs1);
			}
	
			this->correspondences().push_back(corrs2);
			*/
			
	
}

#endif /* _FEATURESEQUENCESIFT_HXX_ */
