/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeatureSequenceSURF.hxx
 *  Implementation of the parameter class FeatureSequenceSURF.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 16.01.2014
 */

#ifndef _FEATURESEQUENCESURF_HXX_
#define _FEATURESEQUENCESURF_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "FeatureSequenceSURF.h"

template <typename T>
FeatureSequenceSURF<T>::FeatureSequenceSURF(const std::string& name) :
		TemplatedParameteredObject<T>(
			"FeatureSequenceSURF", name,
			"<h2>Calculates SURF features and matches them</h2><br>"
			"Calculates SURF features and matches them"
		)
{

	ParameteredObject::_addInputSlot(
		imagePaths, "imagePaths",
		"Path to images files",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		correspondence, "correspondence",
		"Correspondence list",
		"CorrespondenceList");
	ParameteredObject::_addOutputSlot(
		fundamentalmatrices, "fundamentalmatrices",
		"Fundamental matrices",
		"IndexedMatrixList");

	ParameteredObject::_addParameter< int >(
		matchRange, "matchingWindow",
		"The time window for with a point is matched with subsequent"
		"correspondences. (0 means match all images)",
		0, "int");
	
	ParameteredObject::_addParameter< int >(
		stepWidth, "stepWidth",
		"Step width of chosen image sequence to process",
		1, "int");
	
	ParameteredObject::_addParameter< int >(
		knn, "knn",
		"How many points shall be compared to find the right correspondence.",
		2, "int");
	
	ParameteredObject::_addParameter< float >(
		ransacDistance, "ransacDistance",
		"Parameter 1 used for RANSAC."
		"It is the maximum distance from a point to an epipolar line in pixels, beyond which "
		"the point is considered an outlier and is not used for computing the final fundamental "
		"matrix. It can be set to something like 1-3, depending on the accuracy of the point "
		"localization, image resolution, and the image noise. See the OpenCV documentation for more info.",
		3, "float");
	
	ParameteredObject::_addParameter< float >(
		ransacConfidence, "ransacConfidence",
		"It specifies a desirable level of confidence (probability) that the estimated matrix is correct.",
		0.9899999999, "float");
	
	ParameteredObject::_addParameter< float >(
		threshold, "threshold",
		"A threshold to define inliers. By decreasing this value, less correspondences will be found.",
		0.9, "float");
	
	ParameteredObject::_addParameter< float >(
		maxFlow, "maxFlow",
		"Maximal allowed pixel distance between correspondences."
		"The value should be near the maximal flow value of this sequence.",
		100, "float");
	
	ParameteredObject::_addParameter< bool >(
		useEqualizedImg, "useEqualizedImg",
		"Additionally use equalized images. This will double the computation time but "
		"also increases the detection probability.",
		true, "bool");



}

template <typename T>
void FeatureSequenceSURF<T>::execute() {

	if( this->imagePaths().size() < 2 ){
		throw(std::range_error("FeatureSequenceSURF<T>::execute(): At least two frames have to be given"));
	}

	const unsigned int steps = this->stepWidth();
	std::vector< std::string > imgSequence;
	for(size_t i = 0; i < this->imagePaths().size(); i += this->stepWidth() )
	{
		imgSequence.push_back( this->imagePaths()[i] );
	}

	// number of frames
	const int numFrames = imgSequence.size();	

	const int knn = std::max( int(1), this->knn() );

	//-- Step 1: Detect the keypoints using SURF Detector
	SurfFeatureDetector detector(400);
	//SurfFeatureDetector detector( minHessian );	// detector
	SurfDescriptorExtractor extractor;		// descriptor generator
		
	
	// point index iterator
	int ptIndex = 0;

	int iterations = 1;
	if( this->useEqualizedImg() )
		iterations = 2;
	for( int mode = 0; mode < iterations; mode++ ){

		if( mode == 1 )
			sout << "\nRun SURF on equalized images..." << std::endl;

		
		std::vector< int > numKeys(numFrames);					// number of feature points
		/// vector of SiftGPU matches
		/**
			Dimension1: number of images
			Dimension2: number of key in according img
			Dimension3: number of ids to which key is assigned
			Example:	keyIds[0][5] = (1,6,2)
						--> point 5 of img 0 is assigned to corr 1, 6 and 2
						--> (inliers should only have one assigned corr id)
		*/
		std::vector< std::vector< std::vector< int> > > keyIds(numFrames);

		std::vector<std::vector<KeyPoint>> keypoints(numFrames);
		std::vector<Mat> descriptors(numFrames);

		CorrespondenceList<T> &corr = this->correspondence();

		sout << "Run detection algorithm" << std::endl;
		
		#pragma omp parallel for
		for( int i = 0; i < numFrames; i++ )
		{

			// get feature points
			//Mat img1 = imread(argv[1], CV_LOAD_IMAGE_GRAYSCALE);
			//Mat img = imread(imgSequence[i].c_str(), CV_LOAD_IMAGE_ANYCOLOR);
			//Mat img = imread(imgpath.c_str(), CV_LOAD_IMAGE_GRAYSCALE);

			Mat img = imread(imgSequence[i].c_str(), CV_LOAD_IMAGE_GRAYSCALE);

			if( mode == 1 ){
				cimg_library::CImg<float> eq( imgSequence[i].c_str() );
				img = CImg2Mat(eq.equalize(255,0,255), CV_8U) ;
			}
			
			detector.detect( img, keypoints[i] );
			// get descriptors
			extractor.compute( img, keypoints[i], descriptors[i] );
			
			int num = keypoints[i].size();
			numKeys[i] = num ;
			// resize keyIds of img(i) to number of points of img(i)
			keyIds[i].resize(num);

		}

		sout << "Run matching algorithm" << std::endl;
		
		for( int i = 0; i < numFrames; i++ )
		{
			// number of frames to match
			int matchRange = std::min( std::max( int(i+1), int( i+this->matchRange() ) ), int (numFrames) );

			if( this->matchRange() < 1 )
				matchRange = numFrames;
			
#pragma omp parallel for
			for( int j = i+1; j < matchRange; j++ )
			{

				const cimg_library::CImg<float> img1(imgSequence[i].c_str());
				const cimg_library::CImg<float> img2(imgSequence[j].c_str());

				sout << "Match features between frames (" << i << "/" << j << "):" << std::endl;
		
				//FlannBasedMatcher matcher;
				FlannBasedMatcher matcher;
				
				// MATCHING AND CROSSCHECKING
				vector<DMatch> matches;	// final matches
				vector<vector<DMatch> > matches12, matches21; // matches to be compared
				matcher.knnMatch( descriptors[i], descriptors[j], matches12, knn );
				matcher.knnMatch( descriptors[j], descriptors[i], matches21, knn );
				int good_check = 0;
			    for( size_t m = 0; m < matches12.size(); m++ )
			    {
			        bool findCrossCheck = false;
			        for( size_t fk = 0; fk < matches12[m].size(); fk++ )
			        {
			            DMatch forward = matches12[m][fk];
			
			            for( size_t bk = 0; bk < matches21[forward.trainIdx].size(); bk++ )
			            {
			                DMatch backward = matches21[forward.trainIdx][bk];
			                if( backward.trainIdx == forward.queryIdx )
			                {
			                    matches.push_back(forward);
			                    findCrossCheck = true;
								good_check++;
			                    break;
			                }
			            }
			            if( findCrossCheck ) break;
			        }
			    }

				sout << "Found " << matches12.size() - good_check << " outliers by crosschecking" << std::endl;

				std::vector< DMatch > good_matches;
				getGoodMatches( good_matches, matches, keypoints[i], keypoints[j] );

				int num_match = good_matches.size();
				int num_match_inlier = 0;
    
				std::vector<cv::Point2f> feat1, feat2;

				//enumerate all the feature matches
				for(int k  = 0; k < num_match; ++k)
				{						
					int id1 = good_matches[k].queryIdx, id2 = good_matches[k].trainIdx;
					// Get the keypoints from the good matches
					feat1.push_back(keypoints[i][id1].pt );
					feat2.push_back(keypoints[j][id2].pt );
				}
			
				cv::Mat mask;
				cv::Mat fmCv = cv::findFundamentalMat(feat1,feat2,mask,cv::RANSAC,this->ransacDistance(),this->ransacConfidence());
				
				// Convert outlier mask to CImg (1 == inlier, 0 == outlier, width == 2, height == numberOfMatches)
				cimg_library::CImg<double> maskCImg = openCvUtil::Mat2CImg<double>(mask);

				cimg_library::CImg<double> fm = openCvUtil::Mat2CImg<double>(fmCv);
				IndexedMatrix fm2( fm, i, j );		

#pragma omp critical (readwrite)
{
				this->fundamentalmatrices().push_back( fm2 );
	
				for( size_t k = 0; k < num_match; k++ )
				{
					// If point pair is epipolar inlier
					if( maskCImg(0,k) == 1 )
					{
						num_match_inlier++;

						int id1 = good_matches[k].queryIdx, id2 = good_matches[k].trainIdx;
					
						// Get ids and coordinates
						float x1 =  keypoints[i][id1].pt.x, y1 = keypoints[i][id1].pt.y;
						float x2 =  keypoints[j][id2].pt.x, y2 = keypoints[j][id2].pt.y;
						
						// Check if points have been already matched
						bool exist1 = keyIds[i][id1].size();
						bool exist2 = keyIds[j][id2].size();

						// if none of both points already exists, create new correspondence
						if( !exist1 && !exist2 )
						{
							// Get RGB values
							float R1, G1, B1, R2, G2, B2;
							getRGB(img1, x1, y1, R1, G1, B1);
							getRGB(img2, x2, y2, R2, G2, B2);

							// Save correspondence
							keyIds[i][id1].push_back( ptIndex );
							keyIds[j][id2].push_back( ptIndex );

							// Save correspondences
							corr.push_back( hekate::Correspondence<T>(x1, y1, 0, 1, ptIndex, i, 0, R1, G1, B1) );
							corr.push_back( hekate::Correspondence<T>(x2, y2, 0, 1, ptIndex, j, 0, R2, G2, B2) );
							ptIndex++;

						}

						else if( exist1 && !exist2 )
						{
							// Get RGB values
							float R, G, B;
							getRGB(img2, x2, y2, R, G, B);

							// Add key2 to all correspondences of key1
							for( size_t matchid = 0; matchid < keyIds[i][id1].size(); matchid++)
							{
								int key1index = keyIds[i][id1][matchid];
								keyIds[j][id2].push_back( key1index );
								corr.push_back( hekate::Correspondence<T>(x2, y2, 0, 1, key1index, j, 0, R, G, B) );

							}
						}

						else if( !exist1 && exist2 )
						{
							// Get RGB values
							float R, G, B;
							getRGB(img1, x1, y1, R, G, B);

							// Add key1 to all correspondences of key2
							for( size_t matchid = 0; matchid < keyIds[j][id2].size(); matchid++)
							{
								int key2index = keyIds[j][id2][matchid];
								keyIds[i][id1].push_back( key2index );
								corr.push_back( hekate::Correspondence<T>(x1, y1, 0, 1, key2index, i, 0, R, G, B) );
							}
						}

						// if both points already exist, check if they belong to the same correspondence
						else if( exist1 && exist2 )
						{
							// SKIP SINCE BOTH POINTS ALREADY EXISTS


							// Skip, if points already uniquely matched
							//if(corr1.size() == 1 && corr2.size() == 1 && corr1[0] == corr2[0]) continue;
						}

					}
				}

}
			
				sout << "\t" << num_match << " matches found;\n";
				sout << "\t" << num_match - num_match_inlier << " outlier matches removed by epipolar constraint\n";
				sout << "\t" << num_match_inlier << " matches remaining\n";
								
			}
		}
	}

}

/// Extract good matches by comparing distances
template< typename T >
void FeatureSequenceSURF<T>::getGoodMatches( std::vector< DMatch > &good_matches, 
		const std::vector< DMatch > &matches,
		const std::vector<KeyPoint> &keypoints1,
		const std::vector<KeyPoint> &keypoints2 ){

	double max_dist = 0; double min_dist = 100;

	int good = 0;

	float c = std::max(float(0.1),float(this->threshold()));

	//-- Quick calculation of max and min distances between keypoints
	for( int i = 0; i < matches.size(); i++ )
	{ 
		double dist = matches[i].distance;
		if( dist < min_dist ) min_dist = dist;
		if( dist > max_dist ) max_dist = dist;
	}
	
	for( int i = 0; i < matches.size(); i++ )
	{ 
		if( matches[i].distance < c*(max_dist+min_dist) )
		{ 
			float x1 = keypoints1[matches[i].queryIdx].pt.x;
			float y1 = keypoints1[matches[i].queryIdx].pt.y;
			float x2 = keypoints2[matches[i].trainIdx].pt.x;
			float y2 = keypoints2[matches[i].trainIdx].pt.y;
			float dist = std::sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) );
			if(dist < this->maxFlow()){
				good++;
				good_matches.push_back( matches[i]); 
			}
		}
	}

	sout << "Removed " << matches.size() - good << " matches by given treshold ("
		<< c << ")" << std::endl;

}

/// Determine RGB values at pixel loation (x,y)
void getRGB(const cimg_library::CImg<float> &img, const float x, const float y, float &R, float &G, float &B)
{
	R = img.linear_atXYZC(x,y,0,0);
	G = img.linear_atXYZC(x,y,0,1);
	B = img.linear_atXYZC(x,y,0,2);
}

#endif /* _FEATURESEQUENCESURF_HXX_ */
