/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FeaturesSIFT.hxx
 *  Implementation of the parameter class FeaturesSIFT.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 23.01.2012
 */

#ifndef _FEATURESSIFT_HXX_
#define _FEATURESSIFT_HXX_


#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cxcore.h>
#include <hekate/openCvUtil.hxx>

#include <charon-core/ParameteredObject.hxx>
#include "FeaturesSIFT.h"

using namespace cv ;
using namespace openCvUtil ;
using namespace hekate;

template <typename T>
FeaturesSIFT<T>::FeaturesSIFT(const std::string& name) :
		TemplatedParameteredObject<T>(
			"FeaturesSIFT", name,
			"<h2>Extracts SIFT features from an image or image "
			"sequence</h2><br>"
			"Extracts the OpenCV version of SIFT features from an image or "
			"image sequence."
		)
{

	ParameteredObject::_addOutputSlot(
		features, "features",
		"extracted features put into an CorrespondenceList<T> struct",
		"CorrespondenceList");
	
	ParameteredObject::_addOutputSlot(
		imgPaths, "imgPaths",
		"imgPaths",
		"vector<string>");

	ParameteredObject::_addParameter(
		imageDir, "imagePath",
		"path to image",
		"fileopen");

	ParameteredObject::_addParameter(
		useSequence, "useSequence",
		"Choose all images of the selected type inside the according folder.",
		"bool");
	ParameteredObject::_addParameter< int >(
		maxDistance, "paramMaxDistance",
		"Threshold for maximal distance between descriptors. (Matching parameter)",
		2, "int");
	ParameteredObject::_addParameter< double >(
		peakThreshold, "paramPeakThreshold",
		"The peak threshold filters peaks of the DoG scale space that are "
		"too small (in absolute value).",
		0.0, "double");
	ParameteredObject::_addParameter< double >(
		edgeThreshold, "paramEdgeThreshold",
		"The edge threshold eliminates peaks of the DoG scale space whose "
		"curvature is too small (such peaks yield badly localized frames).",
		90.0, "double");
	ParameteredObject::_addParameter< int >(
		octaves, "paramOctaves",
		"The number of a gaussian pyramid octaves that the detector uses. "
		"It is set to 4 by default. If you want to get very large features, "
		"use the larger value. If you want just small features, decrease it.",
		4, "int");
	ParameteredObject::_addParameter< int >(
		octaveLayers, "paramOctaveLayers",
		"The number of images within each octave of a gaussian pyramid. It is set to 2 by default.",
		3, "int");
	ParameteredObject::_addParameter< int >(
		firstOctave, "paramFirstOctave",
		"paramFirstOctave.",
		-1, "int");
	ParameteredObject::_addParameter< int >(
		angleMode, "paramAngleMode",
		"paramAngleMode.",
		0, "int");
}
		 
template <typename T>
void FeaturesSIFT<T>::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	
	// Read in image(s) -------------------------------------------
	if(!this->useSequence())
		sout << "Read in image:" << std::endl;
	else
		sout << "Reading in images in folder:" << std::endl;


	//create a new list
	hekate::CorrespondenceList<T> list ;

	// Read in specified dir
	std::string dir = this->imageDir() ;
	FileTool::slashConvert(dir);
	if(dir==""){
		throw(std::runtime_error("No file of directory given."));
	}
	else
		sout << "	Choosing " << dir 
			<< " as directory:" << std::endl;

	
	// Get file path
	std::string filepath = dir.substr(0,
				dir.find_last_of(FileTool::slash)+1);
	sout << "	File path = " << filepath << std::endl;


	// Get file name
	std::string filename = dir.substr(dir.find_last_of(FileTool::slash)+1,
				dir.length());
	sout << "	File name = " << filename << std::endl;
	

	// Get file extension
	std::string ext = dir.substr(dir.find_last_of('.'), dir.length());
	if(this->useSequence())
		sout << "	Filter for *" << ext << " files" << std::endl;


	// Change working directory
	FileTool::changeDir(filepath);
	//sout << "Working Dir: " << FileTool::getCurrentDir() << std::endl;


	// Read in files with specified extension
	std::vector<std::string> filenames = FileTool::getFilesWithSuffix(ext);
	size_t numberOfFiles = 1;
	if(this->useSequence())
		numberOfFiles = filenames.size();
	sout << "	" << numberOfFiles << " *." << ext << " files found:" << std::endl;
	// Write found files to log
	for( size_t i=0; i<numberOfFiles; i++ ){
		sout << "	img" << i << ": " << filenames[i] << std::endl;
	}

	
	// Create SIFT detector	
	sout << "	Creating SIFT detector" << std::endl;
	cv::SiftFeatureDetector detector(peakThreshold,edgeThreshold,octaves,octaveLayers,firstOctave,angleMode);

	// read in images to CImg
	for(size_t i=0; i<numberOfFiles; i++){

		cimg_library::CImg<T> img((filepath + filenames[i]).c_str());
		Mat img1Mat = CImg2Mat(img.get_normalize(0,255), CV_8U) ;

		// Write image path to output slot
		this->imgPaths().push_back( (filepath + filenames[i]).c_str() );
		
		// Find keypoints in reference frame
		std::vector<KeyPoint> keypoints_1;

		sout << "	Searching for SIFT features..." << std::endl;
		detector.detect( img1Mat, keypoints_1 );	
		sout << "	Found " << keypoints_1.size() << " keypoints in image " << i << std::endl;

		for(size_t j=0; j<keypoints_1.size(); j++){

			list.push_back(Correspondence<T>(keypoints_1[j].pt.x,
				keypoints_1[j].pt.y, 0, 0, j, i));

		}
		
	}

	this->features() = list;

}

#endif /* _FEATURESSIFT_HXX_ */
