/*  Copyright (C) 2009 Julian Coordts

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file densification.hxx
 *  Implementation of the parameter class densification.
 *  \author Julian Coordts, Mirko Schmidt
 *  \date 09.04.2010
 */

#ifndef _DENSIFICATION_HXX_
#define _DENSIFICATION_HXX_

#include <hekate/stereo/Densification.h>
#include <cmath> 

template <typename T>
densification<T>::densification(const std::string& name) :
		TemplatedParameteredObject<T>("densification", name,
			"Fills pixel in an image, where confidence is zero. <br> This plugin is necessary to get a dense disparity map.")
{
	ParameteredObject::_addInputSlot(input_image, "input_image", "sparse image", "CImgList<T>"); 
	ParameteredObject::_addInputSlot(confidence_map, "confidence_map", "confidence map (0 = no info, 100 = full info)", "CImgList<T>"); 
	ParameteredObject::_addOutputSlot(output_image, "output_image", "dense image", "CImgList<T>"); 
	ParameteredObject::_addOutputSlot(new_confidence_map, "new_confidence_map", "updated output confidence map", "CImgList<T>"); 
	ParameteredObject::_addParameter (double_interpolation, "double_interpolation", "interpolates top-down+left-right", true);

	ParameteredObject::_setTags("hekate") ;
}

template <typename T>
void densification<T>::execute() {

	std::cout << "Densification called" << std::endl;

	cimg_library::CImg<T> sparseImage, denseImage, confidenceImage, confidenceImageNew;

	sparseImage.assign(input_image()[0]);
	denseImage.assign(sparseImage.width(),sparseImage.height());
	confidenceImage.assign(confidence_map()[0]);
	confidenceImageNew.assign(sparseImage.width(),sparseImage.height());
	confidenceImageNew.assign(confidenceImage);

	if (   sparseImage.width() != confidenceImage.width() || sparseImage.height() != confidenceImage.height())	
	{
		throw "Densification: Input images (sparseImage, confidence_map) are not of equal size!";
	}


	int lastNonZeroX=-1;

	for(int y=0; y<sparseImage.height();y++)
	{
		for(int x=0;x< sparseImage.width();x++)
		{
			if (confidenceImage(x,y)==0)
			{
				int nx=x;
				//Search for next not empty Pixel at line y
				while ((nx < sparseImage.width()) && (confidenceImage(nx, y)==0)) ++nx;

				//calculate the value of the next not empty pixel, check if upper bound reached
				T nextPixelVal = 0;
				T nextPixelConf = 0;
				if (nx==sparseImage.width())	// upper x bound reached
				{
					nextPixelVal = 0;
					nextPixelConf = 0;
				}
				else 
				{
					nextPixelVal = sparseImage(nx,y);
					nextPixelConf = confidenceImage(nx,y);
				}

				////calculate the value of the previous not empty pixel, check if lower bound reached
				T previousPixelVal = 0;
				T previousPixelConf = 0;
				if (lastNonZeroX == -1) // lower x bound reached
				{
					previousPixelVal = 0;
					previousPixelConf = 0;
				}
				else 
				{
					previousPixelVal = sparseImage(lastNonZeroX,y);
					previousPixelConf = confidenceImage(lastNonZeroX,y);
				}

				float m = ((float)nextPixelVal - (float)previousPixelVal) / ((float)nx - (float)lastNonZeroX);
				float c =  (float)previousPixelVal - m * (float)lastNonZeroX;
				for (;x<nx;x++)
				{
					T newConfValue;
					denseImage(x,y)= m * (float)x + c;
					newConfValue = std::max(previousPixelConf *	pow(0.5, (x - lastNonZeroX)), nextPixelConf * pow(0.5, (nx - x)));
					if (newConfValue != 0) 
						confidenceImageNew(x,y) = std::max(newConfValue, (T)1.0); 
					else 
						confidenceImageNew(x,y) = 0;

					// example: old confidence map:		25		0		0		0		0		0		0		0		0		0		10		0		80
					//			new confidence map:		25		12.5	6.25	3.125	1.561)	(1.0	1.0)	(1.25	2.5		5		10		(40		80
					//				explanation:		old		from left	-	-	-	)	(minimum 1.0)		(from right -	-	-	old		f.rght.	old
					// new confidence value is half of neighbors value
					// maximum dominates (i.e. left neighbor is strong: left dominates)
					// minimum is 1 (if derived from some measured value) (not at borders!)
					// confidence of borders is zero
				}
				if (nx < denseImage.width()) denseImage(x,y) = sparseImage(x,y);
			}
			else
			{
				denseImage(x,y) = sparseImage(x,y);
			}
			lastNonZeroX = x;
		}
		lastNonZeroX = -1;
	}

//the same in y direction
if (this->double_interpolation)
{
	sparseImage.assign(denseImage);					// use results from x-densification	
	confidenceImage.assign(confidenceImageNew);		// .... and new confidence map
	
	int lastNonZeroY = -1;

	for(int x=0;x< sparseImage.width();x++)
	{
		for(int y=0; y<sparseImage.height();y++)
		{
			if (confidenceImage(x,y)==0.0)
			{
				
				int ny=y;
				//Search for next not empty Pixel at column x
				while ((ny < sparseImage.height()) && (confidenceImage(x, ny)==0)) ++ny;
				
				//calculate the value of the next not empty pixel, check if upper bound reached
				T nextPixelVal = 0;
				T nextPixelConf = 0;
				if (ny==sparseImage.height())	// upper y bound reached
				{
					nextPixelVal = 0;
					nextPixelConf = 0;
				}
				else 
				{
					nextPixelVal = sparseImage(x,ny);
					nextPixelConf = confidenceImage(x,ny);
				}
				//calculate the value of the previous not empty pixel, check if lower bound reached
				T previousPixelVal = 0;
				T previousPixelConf = 0;
				if (lastNonZeroY == -1) // lower y bound reached
				{
					previousPixelVal = 0;
					previousPixelConf = 0;
				}
				else 
				{
					previousPixelVal = sparseImage(x,lastNonZeroY);
					previousPixelConf = confidenceImage(x,lastNonZeroY);
				}



				float m = ((float)nextPixelVal - (float)previousPixelVal) / ((float)ny - (float)lastNonZeroY);
				float c =  (float)previousPixelVal - m * (float)lastNonZeroY;

				for (;y<ny;y++)
				{
					T newConfValue;
					denseImage(x,y)= m * (float)y + c;
					newConfValue = std::max(previousPixelConf *	pow(0.5, (y - lastNonZeroY)), nextPixelConf * pow(0.5, (ny - y)));
					if (newConfValue != 0) 
						confidenceImageNew(x,y) = std::max(newConfValue, (T)1.0); 
					else 
						confidenceImageNew(x,y) = 0;
				}

				if (ny < denseImage.height()) denseImage(x,y) = sparseImage(x,y);
			}
			else
			{
				denseImage(x,y) = sparseImage(x,y);
			}
			lastNonZeroY = y;
		}
		lastNonZeroY = -1;
	}
}

	output_image().assign(1,denseImage.width(),denseImage.height());
	output_image()[0] = denseImage;
	new_confidence_map().assign(1, confidenceImageNew.width(),confidenceImageNew.height());
	new_confidence_map()[0] = confidenceImageNew;
}


#endif /* _DENSIFICATION_HXX_ */



