/*  Copyright (C) 2009 Cornelius Ratsch

 This file is part of Charon.

 Charon is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Charon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with Charon.  If not, see <http://www.gnu.org/licenses/>.
 */
/** \file StereoEarlyVision.hxx
 *  Implementation of the parameter class StereoEarlyVision.
 *  \author Cornelius Ratsch
 *  \date 02.03.2010
 */

#ifndef _STEREOEARLYVISION_HXX_
#define _STEREOEARLYVISION_HXX_

#include <hekate/stereo/StereoEarlyVision.h>
#include <hekate/stereo/Stereo.hxx>
#include <EarlyVision/apply_stereo.h>

template<typename T>
StereoEarlyVision<T>::StereoEarlyVision(const std::string& name) :
	Stereo<T> ("StereoEarlyVision", name,
			"Takes two stereo images and calculates the depth "
				"information using the Early Vision method") {
	this->_addParameter(iterations, "iterations",
			"number of BP iterations at each scale", 5);
	this->_addParameter(scales, "scales", "number of scales", 5);

	this->_addParameter(trunk_disc_cost, "trunk_disc_cost",
			"truncation of discontinuity cost", 1.7f);
	this->_addParameter(trunk_data_cost, "trunk_data_cost",
			"truncation of data cost", 15.0f);
	this->_addParameter(weight_data_cost, "weight_data_cost",
			"weighting of data cost", 0.07f);

	this->_addParameter(large_cost, "large_cost", "large cost", 1E20f);
	this->_addParameter(poss_disparities, "poss_disparities",
			"number of possible disparities", 16);
	this->_addParameter(scaling, "scaling",
			"scaling from disparity to graylevel in output", 16);

	this->_addParameter(smoothing, "smoothing",
			"amount to smooth the input images", 0.7);

	ParameteredObject::_setTags("Hekate;Stereo") ;
}

template<typename T>
void StereoEarlyVision<T>::_calculate() {
	if (this->image_sequence().size() != 2) {
		sout << "Error: Invalud number of images in the image sequence ("
				<< this->image_sequence.size() << "), must be 2" << std::endl;
		return;
	}
	/*
	 * @Todo The stereo code doesn't use the CImg library, therefore we need to
	 * convert the input images. This should be improved.
	 */
	image<unsigned char> * img1, *img2, *out;

	{
		img1 = new image<unsigned char> (this->image_sequence()[0].width(),
				this->image_sequence()[0].height(), false);
		// Temporary char-type image
		cimg_library::CImg<unsigned char> temp =
				this->image_sequence()[0].get_normalize(0, 255);
		// copy the CImg image data into img1
		memmove(img1->data, temp.data(), temp.width() * temp.height());

		img2 = new image<unsigned char> (this->image_sequence()[1].width(),
				this->image_sequence()[1].height(), false);
		// Temporary char-type image
		temp = this->image_sequence()[1].get_normalize(0, 255);
		// copy the CImg image data into img2
		memmove(img2->data, temp.data(), temp.width() * temp.height());
	}

	// compute disparities
	out = stereo_ms(img1, img2, this->iterations(), this-> scales(),
			this-> trunk_disc_cost(), this-> trunk_data_cost(),
			this-> weight_data_cost(), this-> large_cost(),
			this-> poss_disparities(), this-> scaling(), this-> smoothing());

	// save output into the CImg output slot
	this->result().assign(1, out->width(), out->height());
	for (int x = 0; x < out->width(); ++x) {
		for (int y = 0; y < out->height(); ++y) {
			this->result()[0](x, y) = out->data[out->width() * y + x];
		}
	}

	delete img1;
	delete img2;
	delete out;
}

#endif /* _STEREOEARLYVISION_HXX_ */

