/*  Copyright (C) 2010 Stephan Meister, template by Cornelius Ratsch

 This file is part of Charon.

 Charon is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Charon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with Charon.  If not, see <http://www.gnu.org/licenses/>.
 */
/** \file StereoOCV.h
 *  Implementation of the parameter class StereoOCV.
 *  \author Stephan Meister
 *  \date 11.03.2010
 */

#ifndef _STEREOOCV_HXX_
#define _STEREOOCV_HXX_

#include <opencv/cv.h>
#include <opencv/cv.hpp>

#include <hekate/stereo/StereoOCV.h>
#include <hekate/stereo/Stereo.hxx>

#include <hekate/openCvUtil.hxx>

using namespace cv ;
using namespace openCvUtil ;

template<typename T>
StereoOCV<T>::StereoOCV(const std::string& name) :
	Stereo<T> ("StereoOCV", name,
			"Takes two stereo images and calculates the depth "
				"information using the OpenCV 2.0 method")
{
	this->_addParameter(_nDisparities, "Disparities",
			"optimal disparity range in pixels", 64);
	this->_addParameter(_bmWindowSize, "BmWindowSize",
			"BlockMatching Window size, larger is more robust but tends to blur disparity map", 15);

	ParameteredObject::_setTags("Hekate;Stereo") ;
}

template<typename T>
void StereoOCV<T>::_calculate()
{
	if (this->image_sequence().size() != 2)
	{
		sout << "Error: Invalid number of images in the image sequence ("
				<< this->image_sequence.size() << "), must be 2" << std::endl;
		return;
	}
	//load CImgList into a OpenCV Mat instance, Mat constructor makes no copy!
	Mat mat1 = CImg2Mat(this->image_sequence()[0].get_normalize(T(0.0), T(255.0)), CV_8U) ;
	Mat mat2 = CImg2Mat(this->image_sequence()[1].get_normalize(T(0.0), T(255.0)), CV_8U) ;
	
	/*
	cimg_library::CImg<unsigned char> img = this->image_sequence()[0].get_normalize(0.0, 255.0) ;
	Mat mat1(img.height(), img.width(), CV_8UC1, reinterpret_cast<void*>(img.data())) ;
	cimg_library::CImg<unsigned char> img2 = this->image_sequence()[1].get_normalize(0.0, 255.0) ;
	Mat mat2(img2.height(), img2.width(), CV_8UC1, reinterpret_cast<void*>(img2.data())) ;
	*/

	
	Mat disparity ;
	StereoBM stereoBM(StereoBM::BASIC_PRESET, _nDisparities(), _bmWindowSize()) ;
	//stereoBM.state.obj->SADWindowSize = 10; // protected :-(
	try
	{	stereoBM(mat1, mat2, disparity) ;	}
	catch(cv::Exception&)
	{	
		;
	}
	//write back Mat data into CImgList
	/*
	signed short* ptr = reinterpret_cast<signed short*>(disparity.data) ;
	cimg_library::CImg<signed short> res(ptr, disparity.cols, disparity.rows) ;
	*/

	this->result().assign(1, disparity.cols, disparity.rows);
	//this->result()[0] = res ;
	this->result()[0] = Mat2CImg<T>(disparity) ;	// raw data

        this->disparity_map().assign(1, this->result()[0].width(), this->result()[0].height());
        this->confidence_map().assign(1, this->result()[0].width(), this->result()[0].height());

        T* result_data = this->result()[0].data();
        T* disp_map_data = this->disparity_map()[0].data();
        T* conf_map_data = this->confidence_map()[0].data();

	double norm_factor = 1 / 16.0; // = SAD_size + 1 ????
	
	//outputImages()[0] = image_ol;
        for (long i = 0; i < this->result()[0].width() * this->result()[0].height() * this->result()[0].depth() * this->result()[0].spectrum(); i ++)
	{
		if (result_data[i] > 0.0)		// good disparity estimation	 
		{
			disp_map_data[i] = T(floor(result_data[i] * norm_factor + 0.5));
			conf_map_data[i] = T(100.0); 
		}
		else						// no usefull result of disparity estimation
		{
			disp_map_data[i] = T(0.0);
			conf_map_data[i] = T(0.0); 
		}
	}

}

#endif /* _STEREOOCV_HXX_ */
