/*  Copyright (C) 2010 Stephan Meister
 
 This file is part of Charon.

 Charon is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Charon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with Charon.  If not, see <http://www.gnu.org/licenses/>.
 */
/** \file StereoSGBM.h
 *  Declaration of the parameter class StereoSGBM.
 *  \author Stephan Meister
 *  \date 04.06.2010
 */

#ifndef _STEREOSGBM_H_
#define _STEREOSGBM_H_

#if defined(MSVC) && defined(HANDLE_DLL)
#ifdef stereosgbm_EXPORTS
/// Visual C++ specific code
#define StereoSGBM_DECLDIR __declspec(dllexport)
#else
#define StereoSGBM_DECLDIR __declspec(dllimport)
#endif /*Export or import*/
#else /* No DLL handling or GCC */
/// Not needed with GCC
#define StereoSGBM_DECLDIR
#endif

#include <charon-core/ParameteredObject.hxx>
#include <charon-utils/CImg.h>
#include <hekate/stereo/Stereo.h>

/// Takes two rectified stereo images and calculates the disparity map using a semi-global block matching alogorithm by  H. Hirschmuller
/// Parameter descriptions are taken from the OpenCV 2.1 C++ documentation (Section "Camera Calibration and 3D Reconstruction")
template<typename T>
class StereoSGBM_DECLDIR StereoSGBM: public Stereo<T> {
public:

	/// create a new StereoSGBM object
	/// \param name          Instance name
	StereoSGBM(const std::string& name);

	/// Performs the calculation .
	virtual void _calculate();

public:

	/// minimum possible disparity value, 0 for non-shifted images
	Parameter<int> _minDisparities;
	
	/// This is maximum disparity minus minimum disparity. Always greater than 0. In the current implementation this parameter must be divisible by 16.
	Parameter<int> _nDisparities ;
	
	/// SADWindowSize  The matched block size. Must be an odd number >=1 . Normally, it should be somewhere in 3..11 range
	Parameter<int> _SADWindowSize;

	///Parameters that control disparity smoothness. The larger the values, the smoother the disparity. 
	///P1 is the penalty on the disparity change by plus or minus 1 between neighbor pixels. 
	///P2  is the penalty on the disparity change by more than 1 between neighbor pixels. 
	///The algorithm requires P2 > P1. 
	///Leave both at zero for automatic calculation
	///(P1 = 8 * SADWindowSize^2)
	///(P2 = 32 * SADWindowSize^2)
	Parameter<int> _p1 ;
	
	/// See P1
	Parameter<int> _p2 ;

	///Maximum allowed difference (in integer pixel units) in the left-right disparity check. Set it to <= 0 value to disable the check
	Parameter<int> _disp12MaxDiff ;

	/// Truncation value for the prefiltered image pixels. 
	///The algorithm first computes x-derivative at each pixel and clips its value by [-preFilterCap, preFilterCap] interval. The result values are passed to the Birchfield-Tomasi pixel cost function
	Parameter<int> _preFilterCap ;

	///The margin in percents by which the best (minimum) computed cost function value should "win" the second best value to consider the found match correct. Normally, some value within 5-15 range is good enough.
	Parameter<int> _uniquenessRatio ;

	///Maximum size of smooth disparity regions to consider them noise speckles and invdalidate. Set it to 0 to disable speckle filtering. Otherwise, set it somewhere in 50-200 range.
	Parameter<int> _speckleWindowSize ;

	///Maximum disparity variation within each connected component. If you do speckle filtering, set it to some positive value, multiple of 16. Normally, 16 or 32 is good enough.
	Parameter<int> _speckleRange ;

	///Set it to true  to run full-scale 2-pass dynamic programming algorithm. It will consume O(W*H*numDisparities) bytes, which is large for 640x480 stereo and huge for HD-size pictures.
	Parameter<int> _fullDP ;
};

#endif // _stereoSGBM_H_

