/*  Copyright (C) 2010 Stephan Meister
 This file is part of Charon.

 Charon is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Charon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with Charon.  If not, see <http://www.gnu.org/licenses/>.
 */
/** \file StereoSGBM.hxx
 *  Implementation of the parameter class StereoSGBM.
 *  \author Stephan Meister
 *  \date 04.06.2010
 */

#ifndef _STEREOSGBM_HXX_
#define _STEREOSGBM_HXX_

#include <opencv/cv.h>
#include <opencv/cv.hpp>

#include <hekate/stereo/StereoSGBM.h>
#include <hekate/stereo/Stereo.hxx>

#include <hekate/openCvUtil.hxx>


//using namespace cv ; //dont use namespace as StereoSGBM and cv::StereoSGBM are ambigious
using namespace openCvUtil ;

template<typename T>
StereoSGBM<T>::StereoSGBM(const std::string& name) :
	Stereo<T> ("StereoSGBM", name,
			"Takes two rectified stereo images and calculates the disparity map using a modfied semi-global block matching algorithm by  H. Hirschmuller "
			"The following parameter and algorithm descriptions are taken from the OpenCV 2.1 C++ documentation (Section Camera Calibration and 3D Reconstruction) "
			""
			"The main differences between the implemented algorithm and the original one are: "
			"* by default the algorithm is single-pass, i.e. instead of 8 directions we only consider 5. Set fullDP=true to run the full variant of the algorithm (which could consume a lot of memory) "
			"* the algorithm matches blocks, not individual pixels (though, by setting SADWindowSize=1 the blocks are reduced to single pixels) "
			"* mutual information cost function is not implemented. Instead, we use a simpler Birchfield-Tomasi sub-pixel metric from BT96 , though the color images are supported as well. "
			"* we include some pre- and post- processing steps from K. Konolige algorithm FindStereoCorrespondceBM , such as pre-filtering ( CV_STEREO_BM_XSOBEL type) and post-filtering (uniqueness check, quadratic interpolation and speckle filtering) "
			)
{
	this->_addParameter(_minDisparities, "minDisparities",
			"minimum possible disparity value, 0 for non-shifted images", 0) ;
	
	this->_addParameter(_nDisparities, "Disparities",
			"This is maximum disparity minus minimum disparity. Always greater than 0. In the current implementation this parameter must be divisible by 16.", 160);
	
	this->_addParameter(_SADWindowSize, "SADWindowSize",
			"SADWindowSize - The matched block size. Must be an odd number >=1 . Normally, it should be somewhere in 3..11 range", 7);

	this->_addParameter(_p1, "P1",
		"Parameters that control disparity smoothness. The larger the values, the smoother the disparity."
	"P1 is the penalty on the disparity change by plus or minus 1 between neighbor pixels."
	"P2  is the penalty on the disparity change by more than 1 between neighbor pixels."
	"The algorithm requires P2 > P1."
	"Leave both at zero for automatic calculation (P1 = 8 * SADWindowSize^2) (P2 = 32 * SADWindowSize^2)",
	0) ;

	this->_addParameter(_p2, "P2", "See P1 for documentation", 0) ;

	this->_addParameter(_disp12MaxDiff, "Disp12MaxDiff", 
		"Maximum allowed difference (in integer pixel units) in the left-right disparity check. Set it to <= 0 value to disable the check",
		0) ;

	this->_addParameter(_preFilterCap, "preFilterCap",
		"Truncation value for the prefiltered image pixels."
		"The algorithm first computes x-derivative at each pixel and clips its value by [-preFilterCap, preFilterCap] interval. The result values are passed to the Birchfield-Tomasi pixel cost function.",
		0) ;

	this->_addParameter(_uniquenessRatio, "UniquenessRatio",
		"The margin in percents by which the best (minimum) computed cost function value should \"win\" the second best value to consider the found match correct. Normally, some value within 5-15 range is good enough.",
		10) ;

	this->_addParameter(_speckleWindowSize, "SpeckleWindowSize",
		"Maximum size of smooth disparity regions to consider them noise speckles and invdalidate. Set it to 0 to disable speckle filtering. Otherwise, set it somewhere in 50-200 range.",
		0) ;

	this->_addParameter(_speckleRange, "SpeckleRange",
		"Maximum disparity variation within each connected component. If you do speckle filtering, set it to some positive value, multiple of 16. Normally, 16 or 32 is good enough.",
		16) ;

	this->_addParameter(_fullDP, "FullDP",
		"Set it to true  to run full-scale 2-pass dynamic programming algorithm. It will consume O(W*H*numDisparities) bytes, which is large for 640x480 stereo and huge for HD-size pictures.",
		0) ;

	ParameteredObject::_setTags("Hekate;Stereo") ;
}

template<typename T>
void StereoSGBM<T>::_calculate()
{
	if (this->image_sequence().size() != 2)
	{
		sout << "Error: Invalid number of images in the image sequence ("
				<< this->image_sequence.size() << "), must be 2" << std::endl;
		return;
	}
	//load CImgList into a OpenCV Mat instance, Mat constructor makes no copy!
	cv::Mat mat1 = CImg2Mat(this->image_sequence()[0].get_normalize(T(0.0), T(255.0)), CV_8U) ;
	cv::Mat mat2 = CImg2Mat(this->image_sequence()[1].get_normalize(T(0.0), T(255.0)), CV_8U) ;
	
	/*
	cimg_library::CImg<unsigned char> img = this->image_sequence()[0].get_normalize(0.0, 255.0) ;
	Mat mat1(img.height(), img.width(), CV_8UC1, reinterpret_cast<void*>(img.data())) ;
	cimg_library::CImg<unsigned char> img2 = this->image_sequence()[1].get_normalize(0.0, 255.0) ;
	Mat mat2(img2.height(), img2.width(), CV_8UC1, reinterpret_cast<void*>(img2.data())) ;
	*/

	
	int p1 = _p1() ;
	int p2 = _p2() ;
	if(p1 == 0 || p2 == 0)
	{	
		p1 = _SADWindowSize() * _SADWindowSize() * 8 ;
		p2 = p1 * 4 ;
	}

	cv::Mat disparity ;
	cv::StereoSGBM stereoSGBM(_minDisparities(), _nDisparities(), _SADWindowSize(), p1, p2, _disp12MaxDiff(),
		_preFilterCap(), _uniquenessRatio(), _speckleWindowSize(), _speckleRange(), _fullDP()) ;
	try
	{	stereoSGBM(mat1, mat2, disparity) ;	}
	catch(cv::Exception& err)
	{	
		sout << "Caught OpenCV Exception: " << err.what() << std::endl ;
	}
	//write back Mat data into CImgList
	/*
	signed short* ptr = reinterpret_cast<signed short*>(disparity.data) ;
	cimg_library::CImg<signed short> res(ptr, disparity.cols, disparity.rows) ;
	*/

	this->result().assign(1, disparity.cols, disparity.rows);
	//this->result()[0] = res ;
	this->result()[0] = Mat2CImg<T>(disparity) ;	// raw data

        this->disparity_map().assign(1, this->result()[0].width(), this->result()[0].height());
        this->confidence_map().assign(1, this->result()[0].width(), this->result()[0].height());

        T* result_data = this->result()[0].data();
        T* disp_map_data = this->disparity_map()[0].data();
        T* conf_map_data = this->confidence_map()[0].data();

	double norm_factor = 1 / 16.0; // = SAD_size + 1 ????
	
	//outputImages()[0] = image_ol;
        for (long i = 0; i < this->result()[0].width() * this->result()[0].height() * this->result()[0].depth() * this->result()[0].spectrum(); i ++)
	{
		if (result_data[i] > 0.0)		// good disparity estimation	 
		{
			disp_map_data[i] = T(floor(result_data[i] * norm_factor + 0.5));
			conf_map_data[i] = T(100.0); 
		}
		else						// no usefull result of disparity estimation
		{
			disp_map_data[i] = T(0.0);
			conf_map_data[i] = T(0.0); 
		}
	}

}

#endif /* _STEREOSGBM_HXX_ */
