#ifndef _totalLstsqMinimize_H_
#define _totalLstsqMinimize_H_

#include <cmath>
#include <limits>
#include <cstring>
#include <iostream>
#ifndef CUDA
#include <charon-utils/CImg.h>
#endif

#ifdef CUDA
	#define CUDA_DEVICE __device__
	typedef float dfloat;
	#define dfloatabs(x) abs(x)

#else
	#define CUDA_DEVICE
	typedef double dfloat;
	#define dfloatabs(x) std::abs(x)

	#include "CImg.h"
	using namespace cimg_library;

#endif

#define PI  3.141592653589793238462643383279

int findRoot4(dfloat solutions[], dfloat t3, dfloat t2, dfloat t1, dfloat t0);
int findRoot3(dfloat solutions[], dfloat t2, dfloat t1, dfloat t0);
int findRoot2(dfloat solutions[], dfloat t1, dfloat t0);

CUDA_DEVICE inline dfloat root3(const dfloat x){ // computes x^(1/3)
	#ifdef CUDA
		return cbrt(x);
	#else
		if( x < 0 ) return -pow(-x, 1.0/3.0);
		return pow(x, 1.0/3.0);
	#endif
}


// http://de.wikipedia.org/wiki/Quartische_Gleichung
CUDA_DEVICE int findRoot4(dfloat solutions[], const dfloat t3, const dfloat t2, const dfloat t1, const dfloat t0){
	if( t0 == 0.0 ){ // one solution is 0, the others can be obtained
		solutions[0] = 0.0;
		return 1 + findRoot3(&solutions[1], t3, t2, t1);
	}

	// algorithm for t^4 + t3 * t^3 + t2 * t^2 + t1 * t + t0 = 0

	dfloat alpha = -3.0*t3*t3/8.0 + t2;
	dfloat beta  = t3*t3*t3/8.0 - t3*t2/2.0 + t1;
	dfloat gamma = -3.0*t3*t3*t3*t3/256.0 + t3*t3*t2/16.0 - t3*t1/4.0 + t0;

	// u^4 + alpha*u^2 + beta*u + gamma = 0

	if( beta == 0.0 ){   // u^4 + alpha*u^2 + gamma = 0
		dfloat z[2];
		int nz = findRoot2(z, alpha, gamma);

		int numSol = 0;

		for(int i = 0; i < nz; i++ ){
			solutions[numSol++] = -t3/4.0 - sqrt(z[i]);
			solutions[numSol++] = -t3/4.0 + sqrt(z[i]);
		}

		return numSol;
	}

	dfloat z[3];
	int nz = findRoot3(z, 2.0*alpha, (alpha*alpha - 4.0*gamma), -beta*beta);

	if( nz != 3 ){
		#ifndef CUDA
		std::cout << "Critical error! There are complex roots! (Not supported by this algorithm)" << std::endl;
		#endif
		return 0;
	}

	// only complex roots
	if( z[0] < 0.0 || z[1] < 0.0 || z[2] < 0.0 ){
		return 0;
	}

	// sign of beta
	dfloat sigma = 1.0;
	if( beta < 0.0 ) sigma = -1.0;

	solutions[0] = -t3/4.0 + sigma/2.0 * ( sqrt(z[0]) + sqrt(z[1]) - sqrt(z[2]) );
	solutions[1] = -t3/4.0 + sigma/2.0 * ( sqrt(z[0]) - sqrt(z[1]) + sqrt(z[2]) );
	solutions[2] = -t3/4.0 + sigma/2.0 * (-sqrt(z[0]) + sqrt(z[1]) + sqrt(z[2]) );
	solutions[3] = -t3/4.0 + sigma/2.0 * (-sqrt(z[0]) - sqrt(z[1]) - sqrt(z[2]) );
	return 4;
}

// http://de.wikipedia.org/wiki/Cardanische_Formeln
CUDA_DEVICE int findRoot3(dfloat solutions[], const dfloat t2, const dfloat t1, const dfloat t0){
	if( t0 == 0.0 ){ // one solution is 0, the others can be obtained
		solutions[0] = 0.0;
		return 1 + findRoot2(&solutions[1], t2, t1);
	}

	// algorithm for t^3 + t2 * t^2 + t1 * t + t0 = 0

	dfloat p = t1 - t2*t2/3.0;
	dfloat q = 2.0*t2*t2*t2/27.0 - t2*t1/3.0 + t0;

	// reduced form: z^3 + p*z + q = 0

	dfloat D = q*q/4 + p*p*p/27.0; // resolvent

	// Please note: we have to transform all solutions back by adding (- t2/3.0)

	if( D > 0 ){
		// one dfloat solution, two imaginary

		dfloat u = root3(-q/2.0 + sqrt(D));
		dfloat v = root3(-q/2.0 - sqrt(D));

		// substitute back: x = z - t2/3
		solutions[0] = u + v - t2/3;
		return 1;

	}else if( D == 0 ){
		solutions[0] = 3.0*q / p	  - t2/3.0;
		solutions[1] = -3.0*q/(2.0*p) - t2/3.0;
		solutions[2] = solutions[1];
		return 3;

	}

	// D < 0
	dfloat alpha = acos( -q/2.0 * sqrt(-27.0/(p*p*p)) ) / 3.0;
	dfloat f	 = sqrt(-4.0/3.0 * p);

	solutions[0] = -f * cos(alpha + PI/3.0)	- t2/3.0;
	solutions[1] =  f * cos(alpha)			- t2/3.0;
	solutions[2] = -f * cos(alpha - PI/3.0)	- t2/3.0;
	return 3;
}

CUDA_DEVICE int findRoot2(dfloat solutions[], const dfloat t1, const dfloat t0){
	if( t0 == 0.0 ){ // one solution is 0, the others can be obtained
		solutions[0] = 0.0;
		solutions[1] = -t1;
		return 2;
	}

	// algorithm for t^2 + t1 * t + t0 = 0
	dfloat p2 = -t1 / 2.0;
	dfloat r  = p2 * p2 - t0;

	if( r < 0 ) return 0;

	r = sqrt(r);

	solutions[0] = p2 - r;
	solutions[1] = p2 + r;
	return 2;
}

#ifndef CUDA
void debug4x4(dfloat matrix[]){
	const int width  = 4;
	const int height = 4;


	for(int y = 0; y < height; y++){
		for(int x = 0; x < width; x++){
			std::cout << matrix[y * width + x] << ((x < width - 1) ? "\t" : "");
		}
		std::cout << std::endl;
	}

}
#endif

CUDA_DEVICE inline void swap(dfloat &a, dfloat &b){
	dfloat temp = a;
	a = b;
	b = temp;
}

// Swaps two rows of a 4x4 matrix
CUDA_DEVICE inline void swapRows4x4(dfloat matrix[], const int y1, const int y2){
	const int width  = 4;
	const int height = 4;

	if( 0 <= y1 && y1 < height && 0 <= y2 && y2 < height && y1 != y2){
		for(int x = 0; x < width; x++){
			swap( matrix[ y1 * width + x ], matrix[ y2 * width + x ] );
		}
	}

}

// Row echelon form of a 4x4 matrix
CUDA_DEVICE void rowEchelonForm4x4(dfloat matrix[], const dfloat accuracy){
	const int width  = 4;
	const int height = 4;

	int currentRow = 0; // Current row

	int pivot[width];
	int numPivots = 0;

	// Check each column
	for(int x = 0; x < width; x++){

		// Go through each row
		for(int y = currentRow; y < height; y++){

			if( dfloatabs( matrix[ y * width + x ] ) >= accuracy ){

				// Swap this to the current row
				if( y != currentRow ){
					swapRows4x4(matrix, y, currentRow);
				}

				// Normalize the current row
				dfloat temp = 1.0;
				swap( matrix[ currentRow * width + x ], temp );

				for(int i = x + 1; i < width; i++){
					matrix[ currentRow * width + i ] /= temp;
				}

				// Subtract the current row from other rows
				for(int y2 = currentRow + 1; y2 < height; y2++){
					dfloat temp = 0.0;
					swap( matrix[ y2 * width + x ], temp);
					if( temp != 0.0 ){
						for(int i = x + 1; i < width; i++){
							matrix[ y2 * width + i ] -= matrix[ currentRow * width + i ] * temp;
						}
					}
				}

				currentRow++;

				pivot[numPivots++] = x; // Save pivot element
				break;

			}else if( matrix[ y * width + x ] != 0.0 ){
				matrix[ y * width + x ] = 0.0;

			}

		}

	}

	for(int currentRow = numPivots; currentRow-- > 0; ){
		int x = pivot[currentRow];

		// Subtract the current row from other rows
		for(int y2 = 0; y2 < currentRow; y2++){
			dfloat temp = 0.0;
			swap(matrix[ y2 * width + x ], temp);
			if( temp != 0.0 ){
				for(int i = x + 1; i < width; i++){
					matrix[ y2 * width + i ] -= matrix[ currentRow * width + i] * temp;
				}
			}
		}

	}


	return;
}

#ifdef CUDA
CUDA_DEVICE void caculateAtmulA(const float matrix[], const int n, dfloat result[]){
#else
template <typename T>
void caculateAtmulA(const T matrix[], const int n, dfloat result[]){
#endif

	// Calculate B = A^t * A (symmetrical matrix)
	for(int i = 0; i < 4; i++){
		for(int j = 0; j <= i; j++){

		// B_ij = At_ik * A_kj = A_ki * A_kj

		dfloat temp = 0.0;
		for(int k = 0; k < n; k++){
			temp += matrix[k*4+i] * matrix[k*4+j];
		}

		result[i*4+j] = temp;
		if( i != j ) result[j*4+i] = temp;

	  }
	}	

}

#ifndef CUDA
//template <typename T>
void caculateAtmulA(const CImg<double> &matrix, dfloat result[]){

	int n = matrix.height();

	// Calculate B = A^t * A (symmetrical matrix)
	for(int i = 0; i < 4; i++){
		for(int j = 0; j <= i; j++){

		// B_ij = At_ik * A_kj = A_ki * A_kj

		dfloat temp = 0.0;
		for(int k = 0; k < n; k++){
			temp += matrix[k*4+i] * matrix[k*4+j];
		}

		result[i*4+j] = temp;
		if( i != j ) result[j*4+i] = temp;

	  }
	}	

}
#endif

#ifdef CUDA
CUDA_DEVICE int minEigenvector4x4(dfloat B[], float result[], const bool normalize){
#else
template <typename T>
bool minEigenvector4x4(dfloat B[], T result[], const bool normalize){
#endif

	// Resize matrix for CUDA
	#ifdef CUDA
	dfloat maxVal = (B[0] + B[10] + B[15] + B[5]);
	
	if( maxVal > 0.0 ){
		for(int i = 0; i < 16; i++){
			B[i] /= maxVal;
		}
	}
	#endif

	// Find the characteristical polynominal (requires symmetric matrix)
	dfloat squares[6];
	squares[0] = B[4]*B[4];
	squares[1] = B[8]*B[8];
	squares[2] = B[9]*B[9];
	squares[3] = B[12]*B[12];
	squares[4] = B[13]*B[13];
	squares[5] = B[14]*B[14];

	dfloat t3 = -(B[0] + B[10] + B[15] + B[5]);
	dfloat t2 = -(squares[0] + squares[1] + squares[2] + squares[3] + squares[4] + squares[5]) + B[0]*(B[5] + B[10] + B[15]) + 
			B[5]*(B[10] + B[15]) + B[10]*B[15];
	dfloat t1 = squares[0]*(B[10] + B[15]) + squares[1]*(B[5] + B[15]) + squares[2]*(B[0] + B[15]) + squares[3]*(B[10] + B[5]) + 
			squares[4]*(B[0] + B[10]) + squares[5]*(B[0] + B[5]) - B[0]*(B[5]*B[15] + B[10]*(B[5] + B[15])) - B[5]*B[10]*B[15] - 
			2.0*B[12]*(B[4]*B[13] + B[8]*B[14]) - 2.0*B[9]*(B[13]*B[14] + B[4]*B[8]);
	dfloat t0 = squares[0]*(squares[5] - B[10]*B[15]) + squares[1]*(squares[4] - B[5]*B[15]) + squares[2]*(squares[3] - B[0]*B[15]) - 
			squares[3]*B[5]*B[10] - squares[4]*B[0]*B[10] - squares[5]*B[0]*B[5] + B[0]*(B[5]*B[10]*B[15] + 2*B[9]*B[13]*B[14]) + 
			2*(B[4]*B[10]*B[12]*B[13] - B[12]*(B[8]*B[9]*B[13] + B[14]*(B[4]*B[9] - B[5]*B[8])) - B[4]*B[8]*(B[13]*B[14] - B[9]*B[15]));

	// Determine the eigenvalues:
	dfloat eig[4];
	int numEig = findRoot4(eig, t3, t2, t1, t0);

	if( numEig == 0 ){
		#ifndef CUDA
		std::cout << "Something went wrong" << std::endl;
		#endif
		return false;
	}

	//Select minimum and maximum eigenvalue
	dfloat minEig = eig[0]; //std::numeric_limits<double>::infinity();
	dfloat maxEig = 0;

	for(int i = 0; i < numEig; i++){
		if( eig[i] < minEig ) minEig = eig[i];
		if( eig[i] > maxEig ) maxEig = eig[i];
	}

	//subtract the eigenvalue on the diagonal
	B[0]  -= minEig;
	B[5]  -= minEig;
	B[10] -= minEig;
	B[15] -= minEig;

	// Accuracy
	#ifdef CUDA
	dfloat 	accuracy = maxEig * 1E-07f;
	#else
	dfloat 	accuracy = maxEig * 1E-13; //1E-15;
	#endif
	int 	numIt = 10;

	dfloat  C[4*4];
	int		resultColumn = -1;

	do{
		#ifdef CUDA
		for(int i=0; i < 4*4; i++){
			C[i] = B[i];
		}

		#else
		memcpy(C, B, sizeof(C));
		#endif

		// row reduce
		rowEchelonForm4x4(C, accuracy);

		// Perform the -1 trick
		for(int y = 0; y < 4; y++){
			if( C[y * 4 + y] == 0.0 ){

				// Create some space here by moving all other elements one down
				for(int y2 = 4 - 1; y2 > y; y2--){
					swapRows4x4(C, y2, y2-1);
				}

				// Insert this row with (-1) in the correct column
				for(int x = 0; x < 4; x++){
					C[y * 4 + x] = (x == y) ? (-1.0) : 0.0;
				}

				// Save this solution
				resultColumn = y;

				// Note: If there are several solutions, we choose the last one		
			}
		}
	
		/*
		#ifndef CUDA
		if(resultColumn < 0)
			std::cout << "one more iteration to find solution" << std::endl;
		#endif
		*/

		// Decrease accuracy, continue until we have got a solution or numIt == 0
		accuracy *= 10;

	}while(resultColumn < 0 && numIt-- > 0);

	if( resultColumn < 0 ){
		#ifndef CUDA
		std::cout << "decrease rowechelonform accuracy" << std::endl;
		#endif
		return false;
	}

	// read the result to a vector
	for(int i = 0; i < 4; i++){
		result[i] = C[i*4 + resultColumn];
	}
	
	// normalize this vector
	if(normalize){	
		dfloat norm = 0.0;
		for(int i = 0; i < 4; i++){
			norm += result[i] * result[i];
		}
	
		norm = 1.0 / sqrt(norm);
		for(int i = 0; i < 4; i++){
			result[i] *= norm;
		}
	}

	return true;
}

#ifdef CUDA
CUDA_DEVICE int totalLstsqMinimize(const float matrix[], const int n, float result[], const bool normalize){
#else
template <typename T>
bool totalLstsqMinimize(const T matrix[], const int n, T result[], const bool normalize){
#endif
	
	dfloat AtA[4*4];
	caculateAtmulA(matrix, n, AtA);

	return minEigenvector4x4(AtA, result, normalize);
}

#ifndef CUDA
template <typename T>
bool totalLstsqMinimizeImg(const CImg<double> &matrix, T result[], const bool normalize){

	if (matrix.width() != 4){
		std::cout << "ERROR: SVD only supports matrices with width 4" << std::endl;
		return false;
	}

	dfloat AtA[4*4];
	caculateAtmulA(matrix, AtA);

	return minEigenvector4x4(AtA, result, normalize);
}
#endif

#endif //_totalLstsqMinimize_H_

