#ifndef _totalLstsqMinimize_gpu_H_
#define _totalLstsqMinimize_gpu_H_


#include "hekate/TriangulationGPU_cu.h"
#include "hekate/totalLstsqMinimize.h"
#include <vector>
#include <list>
#include <cstring>
#include <limits>
#include <new>
#include <math.h>

class GPUMemoryManager{

	private:
		bool 	*Memory;
		int 	blockCount;
		int 	freeBlocks;
		int 	nextIndex;

	public:
		GPUMemoryManager(int blockCount){
			this->Memory = new (std::nothrow) bool [blockCount];

			this->blockCount = blockCount;
			this->freeBlocks = blockCount;

			if(this->Memory){
				for(int i = 0; i < blockCount; i++){
					this->Memory[i] = false;
				}
			}

			this->nextIndex = 0;
		};

		//GPUMemoryManager(const GPUMemoryManager& other) = delete;
		//GPUMemoryManager& operator=(const GPUMemoryManager& other) = delete;

		bool good(){
			return Memory;
		}

		int allocate(){
			if(!Memory)			
				return MATRIX_OUTINDEX_INVALID; //-1;

			// Instead of searching from the beginning, we search from the last position

			for(int i = this->nextIndex; i < blockCount; i++){
				if( Memory[i] == false ){
					Memory[i] = true;
					freeBlocks--;

					this->nextIndex = i+1;
					if( this->nextIndex >= blockCount ) this->nextIndex = 0;

					return i;
				}
			}

			for(int i = 0; i < this->nextIndex; i++){
				if( Memory[i] == false ){
					Memory[i] = true;
					freeBlocks--;

					this->nextIndex = i+1;
					if( this->nextIndex >= blockCount ) this->nextIndex = 0;

					return i;
				}
			}

			
			return MATRIX_OUTINDEX_INVALID;
		}

		void free(int index){
			
			if(Memory && index >= 0 && index < blockCount){
				if( Memory[index] ){
					Memory[index] = false;
					freeBlocks++;
				}
			}
		
		}

		int countFree(){
			return freeBlocks;
		}

		~GPUMemoryManager(){
			if(Memory)	
				delete[] Memory;	
		}

};

template <class F>
class GPUEigenvectors{
	private:

		class GPUJob{
			public:

				int pointIndex;
				std::vector<int> outIndices; // calculation parts
				bool complete;

				GPUJob(int pointIndex){
					this->pointIndex = pointIndex;
					this->complete	 = false;
				}

				//GPUJob(const GPUJob& other) = delete;
				//GPUJob& operator=(const GPUJob& other) = delete;

				void setComplete( bool complete ){
					this->complete = complete;
				}

				void addOutIndex(int outIndex){
					this->outIndices.push_back( outIndex );
				}

		};

		// Temporary data
		float 				*Matrices;
		ControlStep 		*Control;
		int					matrixIndex;
		int					controlIndex;
	
		// GPU
		GPUTriangulate 		gpu;

		// GPU Memory
		GPUMemoryManager	gpuMemory;	
		std::list<GPUJob>	gpuJobs;

		// Initialization successful
		bool 				initSuccess;

		// Output 3D Points
		F*					functor;

	public:

		GPUEigenvectors(F &functor) : gpuMemory(MATRIX_OUTPUT_COUNT) {	
			this->Matrices 		= new (std::nothrow) float [MATRIX_MAX_ROWS * 4];
			this->Control 		= new (std::nothrow) ControlStep [MATRIX_CONTROL_COUNT];

			this->matrixIndex 	= 0;
			this->controlIndex 	= 0;

			this->initSuccess 	= gpu.init();
			this->functor		= &functor;
		}
	
		//GPUEigenvectors(const GPUEigenvectors& other) = delete;
		//GPUEigenvectors& operator=(const GPUEigenvectors& other) = delete;

		bool execute(){

			// Step 1: At * A
			if( this->controlIndex != 0){ // this->matrixIndex != 0

				if( !this->gpu.executeStep1(this->Matrices, this->matrixIndex, this->Control, this->controlIndex) ){
					std::cout << "ERROR: Failed to call CUDA kernel Step1. Check BLOCK_SIZE!" << std::endl;
					this->initSuccess = false;
					return false;
				}
				
			}

			// Step 2: Addition and eigenvectors
			if( this->gpuJobs.size() != 0 ){
				
				std::vector< std::pair<int, int> > completedJobs;

				this->controlIndex = 0;

				// Iterate through all jobs
				typename std::list<GPUJob>::iterator currentJob = this->gpuJobs.begin();
				while( currentJob != this->gpuJobs.end() ){

					// We are (almost) ready with this job?
					if( currentJob->complete && currentJob->outIndices.size() <= MATRIX_MAX_MERGE){ 

						// Add all remaining jobs
						for(int i=0; i < MATRIX_MAX_MERGE; i++){
							if( currentJob->outIndices.empty() ){ //currentJob == --this->gpuJobs.end() ){
								this->Control[controlIndex].step2.outIndex[i] = MATRIX_OUTINDEX_INVALID;

							}else{
								this->gpuMemory.free( currentJob->outIndices.back() );

								this->Control[controlIndex].step2.outIndex[i] = currentJob->outIndices.back();
								currentJob->outIndices.pop_back();
							}
						}

						completedJobs.push_back( std::pair<int, int>( currentJob->pointIndex, controlIndex ) );

						this->Control[controlIndex].step2.complete = true;
						controlIndex++; // Next index

						if( this->controlIndex == MATRIX_CONTROL_COUNT )
							break;

						currentJob = this->gpuJobs.erase( currentJob ); // iterate
					
					// Not ready
					}else{

						for( std::vector<int>::iterator outIndexFirst = currentJob->outIndices.begin(); outIndexFirst != currentJob->outIndices.end(); outIndexFirst++){
				
							int numOutIndices = 0;
							this->Control[controlIndex].step2.outIndex[0] = *outIndexFirst;

							for(int i=1; i < MATRIX_MAX_MERGE; i++){
								if( outIndexFirst == --currentJob->outIndices.end() ){ // last element ... nothing to delete anymore
									this->Control[controlIndex].step2.outIndex[i] = MATRIX_OUTINDEX_INVALID;

								}else{
									this->gpuMemory.free( currentJob->outIndices.back() );

									this->Control[controlIndex].step2.outIndex[i] = currentJob->outIndices.back();
									currentJob->outIndices.pop_back();

									numOutIndices++;
								}
							}

							if( numOutIndices ){
								this->Control[controlIndex].step2.complete = false;
								controlIndex++; // Next index

								if( this->controlIndex == MATRIX_CONTROL_COUNT )
									break;

							}

						}

						if( this->controlIndex == MATRIX_CONTROL_COUNT )
							break;

						currentJob++; // iterate
					}			
				}

				// Everything okay for the final computation :)				
				if( !this->gpu.executeStep2(this->Control, this->controlIndex) ){
					std::cout << "ERROR: Failed to call CUDA kernel Step2. Check BLOCK_SIZE!" << std::endl;
					this->initSuccess = false;
					return false;
				}

				// Read out the results
				for(std::vector< std::pair<int, int> >::iterator it = completedJobs.begin(); it != completedJobs.end(); it++ ){
					int pointIndex		= it->first;
					int controlIndex	= it->second;
					
					//if(!isnan(this->Control[controlIndex].result[0]))
						(*this->functor)(pointIndex, this->Control[controlIndex].result);
				}

			}
			
			// Restore everything
			this->matrixIndex 	= 0;
			this->controlIndex 	= 0;

			return true;
		}

		bool executeRemainingJobs(){
			while( this->gpuJobs.size() != 0 ){
				if( ! this->execute() )
					return false;
			}


			return true;
		}

		bool addMatrix(const float Matrix[], int rows, int pointIndex){
			if(!this->Matrices || !this->Control || !this->initSuccess || !this->gpuMemory.good())
				return false;			

			// Nothing to compute ?
			if( rows <= 0 )
				return false;

			// Add this job to the list
			typename std::list<GPUJob>::iterator currentJob = this->gpuJobs.insert( this->gpuJobs.end(), GPUJob(pointIndex) );

			for(int row = 0; row < rows; row++){

				// Write control commands or out of matrix memory
				if(row % MATRIX_SPLIT_COUNT == 0 || this->matrixIndex == MATRIX_MAX_ROWS){

					// Write end command at this position
					if(row != 0){
						this->Control[this->controlIndex].step1.numRows 	= (this->matrixIndex - this->Control[this->controlIndex].step1.startRow);
						this->controlIndex++;
					}

					// Execute everything on the stack
					if( this->matrixIndex == MATRIX_MAX_ROWS || this->controlIndex == MATRIX_CONTROL_COUNT || this->gpuMemory.countFree() == 0){
						if( ! this->execute() ){
							this->initSuccess = false;
							return false;
						}
					}

					// Restart the matrix system
					this->Control[controlIndex].step1.startRow 		= this->matrixIndex;
					this->Control[controlIndex].step1.outIndex		= this->gpuMemory.allocate();

					if(this->Control[controlIndex].step1.outIndex == MATRIX_OUTINDEX_INVALID){
						this->initSuccess = false;
						return false;
					}

					// Add this outIndex to our current job
					currentJob->addOutIndex( this->Control[controlIndex].step1.outIndex );					

				}
				
				memcpy(&this->Matrices[matrixIndex * 4], &Matrix[row * 4], sizeof(float) * 4);
				this->matrixIndex++;

				/* same as:
					this->Matrices[matrixIndex * 4 + 0] = Matrix [row * 4 + 0];  
					this->Matrices[matrixIndex * 4 + 1] = Matrix [row * 4 + 1]; 
					this->Matrices[matrixIndex * 4 + 2] = Matrix [row * 4 + 2]; 
					this->Matrices[matrixIndex * 4 + 3] = Matrix [row * 4 + 3]; 
					this->matrixIndex++;
				*/

			}

			this->Control[this->controlIndex].step1.numRows 	= (this->matrixIndex - this->Control[this->controlIndex].step1.startRow);
			this->controlIndex++;

			// This job is ready (theoretically)
			currentJob->setComplete( true );

			return true;

		}

		~GPUEigenvectors(){
			if(this->Matrices)
				delete[] this->Matrices;

			if(this->Control)
				delete[] this->Control;
		}
};
#endif //_totalLstsqMinimize_gpu_H_
