/*  Copyright (C) 2010 Stephan Meister

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file TriangulationFromDisparity.hxx
 *  Implementation of the parameter class TriangulationFromDisparity.
 *  \author Stephan Meister
 *  \date 06.06.2010
 */

#ifndef _TRIANGULATIONFROMDISPARITY_HXX_
#define _TRIANGULATIONFROMDISPARITY_HXX_

#include <hekate/triangulation/TriangulationFromDisparity.h>
#include <charon-utils/ImgTool.hxx> 

using namespace cimg_library ;

template <typename T>
TriangulationFromDisparity<T>::TriangulationFromDisparity(
			const std::string& name) :
		TemplatedParameteredObject<T>("TriangulationFromDisparity", name,
			"Triangulation of 3D points from disparity map and "
			"camera parameters"),
			_pmatrix(true, false)
{
	ParameteredObject::_addParameter<double>(
			_stereoBaseline, "stereoBaseline",
			"Distance between the focal points of the stereo camera "
			"system [meter]",
			0.3) ;
	ParameteredObject::_addParameter<double>(
			_focalLength, "focalLength",
			"Focal length of the lenses [meter]", 0.025) ;
	ParameteredObject::_addParameter<double>(
			_pixelWidth, "pixelWidth",
			"Width of a single sensor pixel [meter]", 8.0e-6) ;
	ParameteredObject::_addParameter<double>(
			_pixelHeight, "pixelHeight",
			"Height of a single sensor pixel [meter]", 8.0e-6) ;
	ParameteredObject::_addInputSlot(
			_pmatrix, "p_matrices",
			"Camera matrices P1, P2(currently not used)", "CImgList<T>");
	ParameteredObject::_addInputSlot(
			_disparities, "disparities",
			"disparity map", "CImgList<T>");
//	ParameteredObject::_addOutputSlot(
//			_triangulatedPoints, "triangulatedPoints",
//			"triangulatedPoints", "CImgList<T> ") ;
	ParameteredObject::_addOutputSlot(
			triangulatedPoints, "triangulatedPoints",
			"triangulatedPoints", "std::vector<std::vector<T> >");
}

template <typename T>
void TriangulationFromDisparity<T>::execute() {
	
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	cimg_library::CImg<T> disp = this->_disparities()[0] ;

	double f = _focalLength() ;

	//calculates the viewing-ray through the focal point for each pixel
	//the actual 3D-point is then calculated from the length given by the
	//disparity at each pixel
	
	//proportional scaling factor for distsparity->distance calculation
	//keep rounding errors in mind when f and pixel sizes are small
	double distProportional = f * _stereoBaseline() / _pixelWidth() / 2.0 ;
	cimg_forXY(disp, x, y)
	{
		//set the principal point into the display center
		double X = double((x - disp.width() / 2)) * _pixelWidth() ;
		double Y = double((y - disp.height() / 2)) * _pixelHeight() ;
		double Z = f ;
		double dist = 0 ;
		if(disp(x,y) != 0) 
		{	dist = distProportional / disp(x,y) ;	}
		//here dist is the distance from the camera plane,
		//not the camera center

		
		//The points lie on a sphere cap, so the distances are xy-dependant
		double scale = dist / sqrt(f*f + X*X + Y*Y) ;
		if(scale != 0.0)
		{
			X *= scale ;
			Y *= scale ;
			Z *= scale ;
			
			std::vector<T> vecX;
			vecX.push_back(X);
			vecX.push_back(Z);
			vecX.push_back(Y);
			this->triangulatedPoints().push_back(vecX);
		}

	}
}

#endif /* _TRIANGULATIONFROMDISPARITY_HXX_ */



