/*  Copyright (C) 2010 Moritz Becker

    This source code is distributed under the GNU GPL.
*/
/** \file TriangulationHartleySturm.hxx
 *  Implementation of the parameter class TriangulationHartleySturm.
 *  \author Moritz Becker
 *  \date 26.05.2010
 */

#ifndef _TRIANGULATIONHARTLEYSTURM_HXX_
#define _TRIANGULATIONHARTLEYSTURM_HXX_

#include <hekate/triangulation/TriangulationHartleySturm.h>
#include <charon-utils/ImgTool.hxx> 
#include <hekate/Correspondence.hxx>

using namespace hekate;

template <typename T>
TriangulationHartleySturm<T>::TriangulationHartleySturm(const std::string& name) :
		TemplatedParameteredObject<T>("TriangulationHartleySturm", name,
			"Triangulation based on Hartley and Sturm."
			"This plugin triangulates two corresponding 2D points from two different cameras "
			"to receive the according 3D point. Therefore the camera matrices have to be known."
			)
{
	ParameteredObject::_addParameter (rejectNegativeDepths, "rejectNegativeDepths", 
		"If this flag is set, all reconstructed 3D points that ly behind the image "
		"planes will not be saved.", true);
	ParameteredObject::_addParameter (calculateAll, "calculateAll", 
		"Calculates all correspondences instead of only the first correspondence pair", true);
	ParameteredObject::_addParameter (useReferenceCam, "useReferenceCam", 
		"Triangulate between P_i and reference cam K(I|0)", true);

	ParameteredObject::_addInputSlot(pmatrix, "p_matrices", "Camera matrices P1, P2", "CImgList<T>"); 
	ParameteredObject::_addInputSlot(correspondences, "correspondences", "2D correspondences.", "CorrespondenceList"); 
	ParameteredObject::_addOutputSlot(triangulatedPoints, "triangulatedPoints", "Reconstructed 3D points.", "CorrespondenceList"); 
	
}

template <typename T>
void TriangulationHartleySturm<T>::execute() {

	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	
	const CorrespondenceList<T> &corrs = this->correspondences();

	size_t maxIt = 1;
	size_t maxItCam = this->pmatrix().size();
	if(this->calculateAll())
	{
		maxIt = corrs.sizeImages() -1 ;
		if( maxIt != maxItCam -1 )
			throw std::range_error("Inconsistent size of CorrespondenceList and camera matrices.");
	}


	// Iterate over all correspondence pairs
	for(size_t j=0; j<maxIt; j++)
	{

		cimg_library::CImg<T> P1, P2;
		if(this->useReferenceCam())
			P1 = this->pmatrix()[0];
		else
			P1 = this->pmatrix()[j];
		P2 = this->pmatrix()[j+1];
		
		//ImgTool::printInfo(sout,P1,"	Cam1");
		//ImgTool::printInfo(sout,P2,"	Cam2");

		int outputPointSize = 0;
		int positiveDepth = 0;

		auto it = corrs.get<imgIndex>().lower_bound(j) ;
		auto itEnd = corrs.get<imgIndex>().upper_bound(j);
		for(; it != itEnd ; it++ )
		{
			// Continue with next iteration, if point is an outlier
			if(it->outlier)
				continue;

			// Try getting a correspondence in frame j+1
			try
			{
				Correspondence<T> point2 = corrs(j+1,it->pointIndex);
			
				//sout << *it << std::endl;
				//sout << point2 << std::endl;

				// Continue with next iteration, if correspondence
				// is an outlier
				if(point2.outlier)
					continue;			

				// Triangulate points
				cimg_library::CImg<T> pt1(1,3), pt2(1,3);
				pt1(0,0) = it->x();
				pt1(0,1) = it->y();
				pt1(0,2) = it->w();
				pt2(0,0) = point2.x();
				pt2(0,1) = point2.y();
				pt2(0,2) = point2.w();

				pt1/=pt1(0,2);
				pt2/=pt2(0,2);

				//ImgTool::printInfo(sout,pt1,"pt1------------");
				//ImgTool::printInfo(sout,pt2,"pt2------------");

				// define matrix A = x cross X and x' cross X 
				cimg_library::CImg<T> A(4,4);
				//first row of A
				A(0,0) = pt1(0,0)*P1(0,2) - P1(0,0);
				A(1,0) = pt1(0,0)*P1(1,2) - P1(1,0);
				A(2,0) = pt1(0,0)*P1(2,2) - P1(2,0);
				A(3,0) = pt1(0,0)*P1(3,2) - P1(3,0);
				//second row of A
				A(0,1) = pt1(0,1)*P1(0,2) - P1(0,1);
				A(1,1) = pt1(0,1)*P1(1,2) - P1(1,1);
				A(2,1) = pt1(0,1)*P1(2,2) - P1(2,1);
				A(3,1) = pt1(0,1)*P1(3,2) - P1(3,1);		
				//third row of A
				A(0,2) = pt2(0,0)*P2(0,2) - P2(0,0);
				A(1,2) = pt2(0,0)*P2(1,2) - P2(1,0);
				A(2,2) = pt2(0,0)*P2(2,2) - P2(2,0);
				A(3,2) = pt2(0,0)*P2(3,2) - P2(3,0);
				//fourth row of A
				A(0,3) = pt2(0,1)*P2(0,2) - P2(0,1);
				A(1,3) = pt2(0,1)*P2(1,2) - P2(1,1);
				A(2,3) = pt2(0,1)*P2(2,2) - P2(2,1);
				A(3,3) = pt2(0,1)*P2(3,2) - P2(3,1);

				// determine 3D point via SVD of A (Hartley p.592)
				cimg_library::CImg<T> X(1,4), S, V, U;
				A.SVD(U, S, V);
			
				//sout << "Triangulation SVD: " << std::endl;
				//ImgTool::printInfo(sout,U,"U");
				//ImgTool::printInfo(sout,S,"S");
				//ImgTool::printInfo(sout,V,"V");

				X = V.get_crop(3,0,3,3);
				X /= X(0,3);
				
				//ImgTool::printInfo(sout,X,"Triangulated Point");

				//check if amount of triangulated points is equal amount of input
				//this can differ if the correspondences were corrected before
				outputPointSize ++;
				
				// check whether the triangulated point lies before or behind the priciple planes
				// the principle plane is the plane the camera center lies on and which is
				// intersected of by the priciple ray perpendicularly.

				double depth1 = getDepth(P1, X);
				double depth2 = getDepth(P2, X);
				
				//sout << "Depth check (" << i << ") cam1: " << depth1 << std::endl;
				//sout << "Depth check (" << i << ") cam2: " << depth2 << std::endl;
			
				//push back 3D corrdinates
				if(this->rejectNegativeDepths() && (depth1 < 0 || depth2 < 0)){
					//sout << "Negative depths: Rejecting triangulated point." << std::endl;
					//this->triangulatedPoints().push_back(vecX);
				}
				else{
					Correspondence<T> point3D(X,it->pointIndex, j, it->R, it->G, it->B);
					this->triangulatedPoints().push_back(point3D);
					positiveDepth ++;
					//sout << point3D << std::endl;
				}

			}
			catch(...)
			{
				// Continue with next iteration, if no correspondence
				// could be found
				continue;
			}

	

		}
		
		sout << "-------------------------------------------" << std::endl;
		sout << "" << outputPointSize-positiveDepth << " of " <<
			outputPointSize << " triangulated points" <<
			" with negative depths found." << std::endl;

	}
}

/// Calculates depth of point X in relation to camera P1
template <typename T>
double getDepth(const cimg_library::CImg<T> &P1, const cimg_library::CImg<T> &X){

	int signP1;
	(P1.get_crop(0,0,2,2).det() < 0 ? signP1 = -1 : signP1 = 1);
		
	double m3 = pow((double)(P1.get_crop(0,2,2,2)*P1.get_crop(0,2,2,2).transpose())[0], 0.5);
		
	double depth = signP1 * (P1*X)[2] / (m3 * X[3]);

	return depth;
}


#endif /* _TRIANGULATIONHARTLEYSTURM_HXX_ */



