/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgeFeaturesWrite.cpp
 *  Implementation of parameter class DepthEdgeFeaturesWrite.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 14.10.2013
 */

#include <charon-core/ParameteredObject.hxx>
#include <hekate/DepthEdgeFeaturesWrite.h>
#include <string>

DepthEdgeFeaturesWrite::DepthEdgeFeaturesWrite(const std::string& name) :
		ParameteredObject(
			"DepthEdgeFeaturesWrite", name,
			"<h2>Saves feature images in several CImgLists</h2><br>"
			"Saves feature images in CImgLists sperately for every frame. "
			"Every frame has one corresponding *.cimg file containing "
			"several CImg objects."
		)
{

	ParameteredObject::_addInputSlot(
		depthEdgesFeatures, "depthEdgesFeatures",
		"CImgLists of depth edge features",
		"vector<CImgList>");

	ParameteredObject::_addParameter(filename, "filename", "Path and filename pattern",
			"filewrite");

	ParameteredObject::_setTags("Hekate;Edges") ;

}

void DepthEdgeFeaturesWrite::execute() {

	sout << "Write features" << std::endl;

	try {

		// Get working directory (should be the one containing the *.wrp file)
		std::string workingPath = FileTool::getCurrentDir();

		const std::string path = this->filename();
		
		if( path == "" )
			throw(std::runtime_error("DepthEdgeFeaturesWrite: empty path"));
	
		const std::vector<cimg_library::CImgList<float>> &files = this->depthEdgesFeatures();

		for( size_t i = 0; i < files.size(); i++ ){

			char ss[10];
			sprintf( ss, "%05d.cimg", i);
			std::string filename = path + ss;

			sout << "\tSave file " << filename << std::endl;

			files[i].save( filename.c_str() );

		}
	
		// Reset working directory to that one containing the *.wrp file
		FileTool::changeDir( workingPath );

	}
	catch (const cimg_library::CImgException& err) {
		std::string error = "\tCould not write file \""
			+ this->filename() + "\":\n\t" + err.what();
		sout << error << std::endl;
	}

}

// the following functions are needed
// for class DepthEdgeFeaturesWrite to work as a charon plugin.
extern "C" depthedgefeatureswrite_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new DepthEdgeFeaturesWrite(name);
}

extern "C" depthedgefeatureswrite_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" depthedgefeatureswrite_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
