/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgesRFReader.cpp
 *  Implementation of parameter class DepthEdgesRFReader.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 16.07.2013
 */

#include <charon-core/ParameteredObject.hxx>
#include <hekate/DepthEdgesRFReader.h>


DepthEdgesRFReader::DepthEdgesRFReader(const std::string& name) :
		ParameteredObject(
			"DepthEdgesRFReader", name,
			"<h2>Reads in random forests saved by "
			"DepthEdgesRFWriter</h2><br>"
			"Reads in a depth edge random forest project including a "
			"index.txt file and hdf5 random forest files.<br>"
			"The index file thereby contains the hdf5 file names, its "
			"corresponding frame ids and also the coordinates of "
			"corresponding image patches."
		)
{
	
	ParameteredObject::_addOutputSlot(
		depthEdgeModel, "depthEdgeModel",
		"The depth edge model which keeps random forests, its patch coordinates"
		" and its ids. It also contains the feature setup which was chosen"
		" at learning runtime.");

	ParameteredObject::_addParameter(
		filename, "filename",
		"Path to the index file of the random forests",
		"fileOpen");

}

void DepthEdgesRFReader::execute() {

	// Get working directory (should be the one containing the *.wrp file)
	std::string workingPath = FileTool::getCurrentDir();

	hekate::DepthEdgeModel model;
	RF_list rfList;
	RF_ids idList;

	std::ifstream file(this->filename()) ;
	if(filename().empty())
	{	raise("Filename is empty!") ;	}
	if(!file || !file.good())
	{
		raise(std::string("Could not open \"")+ filename() + "\" for reading!") ;
	}

	// Get directory of files
	sout << "	Reading files of path " << this->filename() << std::endl;
	
	// Skip first two lines commented lines (by #)
	for( size_t i=0; i<4; i++)
	{
		std::string line ; 
		getline(file, line) ;
		sout << line << std::endl;
	}
			
	// Get number of random forests to read in
	int nid;
	file >> nid ;

	sout << "nid " << nid << std::endl;

	// Resize arrays
	model.imageIds.resize(nid);

	sout << FileTool::getCurrentDir() << std::endl;

	// Read hdf file
	std::string filename;
	file >> filename;
	sout << filename << std::endl;

	sout << this->filename() << std::endl;
	std::string filepath = this->filename().substr(0,
				this->filename().find_last_of("/")+1);

	filepath += filename;

//	filepath = "C:\\Users\\mobecker\\Desktop\\TRASH-learn-model2\\rf.hdf5";

	//filepath = "C:/Users/mobecker/Desktop/rf.hdf5";

	sout << "Read from hdf5 file: " << filepath << std::endl;
	
	sout << "Read in " << nid << " image ids" << std::endl;
	// Iterate over lines containing image ids
	for( size_t k=0; k<nid; k++)
	{
		file >> model.imageIds[k] ;
	}
	
	sout << "Read in " << nid << " random forest sets" << std::endl;

	// Iterate over lines containing rf names, ids and coordinates
	for( size_t k=0; k<nid; k++)
	{

		if(!file.good())
			break;

		if(!file.good())
			sout<< "!file.good1" << std::endl;
				
		// get current line
		std::string rf_name;
		if(!file.good())
			sout<< "!file.good2" << std::endl;
		file >> rf_name;

		if(!file.good())
			sout<< "!file.good2" << std::endl;

		sout << rf_name << std::endl;

		// Generate coordinate vector
		
		//std::string currentFilename = filename + FileTool::slash + rf_name ;
		//sout << "Read " << currentFilename << std::endl;

		vigra::RandomForest<int> rf;
		vigra::rf_import_HDF5(rf, filepath, rf_name);
		//vigra::rf_import_HDF5(rf, "C:/Users/mobecker/Desktop/rf.hdf5", "/rf_frame_0");
		model.randomForests.push_back( rf );

		sout << "Read group: " << rf_name << std::endl;
		sout << "\tFeature count: " << rf.feature_count() << std::endl;
		sout << "\tClass count: " << rf.class_count() << std::endl;
		
		if(!file.good())
			sout<< "!file.good3" << std::endl;

	}
	
	this->depthEdgeModel() = model;

	sout << "Read in:\n\t" << model.randomForests.size() << " random forests of frames:\n";
	for( int i = 0; i < model.imageIds.size(); i++ ){
		sout << "\t" << model.imageIds[i] << ", " << std::endl;
		sout << "RFclass_count[" << i << "] = " << model.randomForests[0].class_count() << std::endl;
		sout << "RFfeature_count[" << i << "] = " << model.randomForests[0].feature_count() << std::endl;
	}

	// Reset working directory to that one containing the *.wrp file
	FileTool::changeDir( workingPath );

}

// the following functions are needed
// for class DepthEdgesRFReader to work as a charon plugin.
extern "C" depthedgesrfreader_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new DepthEdgesRFReader(name);
}

extern "C" depthedgesrfreader_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" depthedgesrfreader_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
