/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DepthEdgesRFWriter.cpp
 *  Implementation of parameter class DepthEdgesRFWriter.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 08.07.2013
 */

#include <charon-core/ParameteredObject.hxx>
#include <hekate/DepthEdgesRFWriter.h>

#include <iostream>
#include <fstream>

#include <vigra/windows.h>
#include <vigra/hdf5impex.hxx>
#include <vigra/random_forest.hxx>
#include <vigra/random_forest_hdf5_impex.hxx>

DepthEdgesRFWriter::DepthEdgesRFWriter(const std::string& name) :
		ParameteredObject(
			"DepthEdgesRFWriter", name,
			"<h2>Writes random forests learned by DepthEdgesLearn to a local "
			"file</h2><br>"
			"Saves random forests and corresponding image ids to a seperate "
			"file"
		)
{

	ParameteredObject::_addInputSlot(
		depthEdgeModel, "depthEdgeModel",
		"The depth edge model which keeps random forests, its patch coordinates"
		" and its ids. It also contains the feature setup which was chosen"
		" at learning runtime.");
	
	ParameteredObject::_addParameter(
		filepath, "filepath",
		"Path to write files to.",
		"Path");
	ParameteredObject::_addParameter(
		deleteHDF5, "deleteHDF5",
		"Delete old HDF5 file",
		true);

}

void DepthEdgesRFWriter::execute() {

	// Get working directory (should be the one containing the *.wrp file)
	std::string workingPath = FileTool::getCurrentDir();

	setlocale(LC_NUMERIC, "C");
	
	// Check input consistency
	size_t nIds = this->depthEdgeModel().imageIds.size();
	size_t nRF = this->depthEdgeModel().randomForests.size();
	if( nRF != nIds )
	{
		sout << "number of random forests and ids not equal." << std::endl;
		sout << "	depthEdgeModel().imageIds.size()      = " << depthEdgeModel().imageIds.size() << std::endl;
		sout << "	depthEdgeModel().randomForests.size() = " << depthEdgeModel().randomForests.size() << std::endl;
		throw( std::range_error("DepthEdgesRFWriter: number of random forests and ids not equal.") );
	}

	// Create binary file containing ids, names of randomforest hdf5 files and its coordinates
	std::string indexFileName( this->filepath()+"/rf_index.txt" ) ;
	if(this->filepath() == "")
	{	raise("Filepath is empty!") ;	}	

	std::string file_hdf5( "rf.hdf5" ) ;

	std::ofstream indexFile;
	indexFile.open (indexFileName, std::ios::trunc | std::ios::out );
	
	if( !indexFile.is_open()) {
		throw std::runtime_error(std::string("Could not open file ") + indexFileName);
	}

	FILE* fp2; 
	fp2 = fopen(file_hdf5.c_str(),"w");
	if(fp2 == NULL)
	{	throw std::runtime_error(std::string("Could not open file ") + file_hdf5) ;	}

	// Remove hdf5 file to write to new one
	if( this->deleteHDF5() )
	{
		try{
			std::string fname = this->filepath()+FileTool::slash+"rf.hdf5";
			if( remove( fname.c_str() ) != 0 )
				sout << "Deleting existing file not possible" << std::endl;
			sout << checksum << " Replaced " << file_hdf5 << " to write a new file" << std::endl;
		}
		catch(...){
			sout << file_hdf5 << " does not exist. Write to a new file." << std::endl;
		}
	}
	else
	{
		sout << file_hdf5 << " will be updated by new values" << std::endl;
	}	

	sout << "Save " << nIds << " random forests in total" << std::endl;

	// Write meta data
	indexFile << "# This file was generated by the charon module FileWriterIndexedMatrix\n";
	indexFile << "# number_of_unique_ids\n";
	indexFile << "# path_to_hdf5_file\n";
	indexFile << "# img_id, randomforest_filename\n";
	
	indexFile << nIds << "\n";
	indexFile << file_hdf5 << "\n";
	
	// Iterate over number of different images
	for( size_t i = 0; i < nIds; i++ ) {
		indexFile << this->depthEdgeModel().imageIds[i]<< "\n";
	}

	// Iterate over number of different images
	for( size_t i = 0; i < nIds; i++ )
	{
		std::stringstream s ;
		s << this->filepath() << FileTool::slash << "rf.hdf5" ;
		std::string filename = s.str();

		std::stringstream s2 ;
		s2 << "rf_frame_" << i;
		std::string groupname = s2.str() ;

		indexFile << "rf_frame_" << i <<"\n";

		sout << "Write hdf5 file to " << filename << "\n\tgroup: " << groupname ;
		sout << "\n\tFeature count: " << this->depthEdgeModel().randomForests[i].feature_count() ;
		sout << "\n\tClass count: " << this->depthEdgeModel().randomForests[i].class_count() << std::endl; ;
		
		// Save random forest as hdf 5			
		vigra::rf_export_HDF5(	this->depthEdgeModel().randomForests[i],
								filename, 
								groupname);
			
	}
	
	indexFile.close();
	
	// Reset working directory to that one containing the *.wrp file
	FileTool::changeDir( workingPath );

}

// the following functions are needed
// for class DepthEdgesRFWriter to work as a charon plugin.
extern "C" depthedgesrfwriter_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new DepthEdgesRFWriter(name);
}

extern "C" depthedgesrfwriter_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" depthedgesrfwriter_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
