/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file ExifReader.cpp
 *  Implementation of parameter class ExifReader.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 13.07.2012
 */

#include <charon-core/ParameteredObject.hxx>
#include <hekate/ExifReader.h>
#include <iomanip>
#include <exiv2/exiv2.hpp>

ExifReader::ExifReader(const std::string& name) :
		ParameteredObject(
			"ExifReader", name,
			"<h2>Reads EXIF data from images</h2><br>"
			"The plugin takes a list of file paths and reads out the "
			"according exif data."
		)
{

	ParameteredObject::_addInputSlot(
		imagePaths, "imagePaths",
		"Paths to the images",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		exifData, "exifData",
		"Exif data:",
		"exifData");

	ParameteredObject::_addOutputSlot(
		calibration, "calibration",
		"calibration matrices",
		"CImgList<double>");

}

void ExifReader::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();

	size_t nImages = this->imagePaths().size();
	if(nImages == 0)
		throw(std::range_error("ExifReader::No input images available"));

	for(size_t i=0; i<nImages; i++){

		std::string filepath = this->imagePaths()[i];
		
		sout << "-----------------------------" << std::endl;
		sout << "Image[" << i << "]:" << filepath << std::endl;

		float exifFocalLength = 0;
		float exifXRes = 0, exifYRes = 0;
		int exifWidth = 0, exifHeight = 0;
		int exifResUnit = 0;
		float focalLengthX = 0, focalLengthY = 0;

		try {
			Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open(filepath);
			assert(image.get() != 0);
			image->readMetadata();

			exifWidth = image->pixelWidth();
			exifHeight = image->pixelHeight();
			sout << "Width(px): " << exifWidth << std::endl;
			sout << "Height(px): " << exifHeight << std::endl;

			Exiv2::ExifData &exifData = image->exifData();
			
			if (exifData.empty()) {

				sout << "No calibration could be found in EXIF data." << std::endl;
				cimg_library::CImg<double> cali(3,3);
				cali.fill(0.);
				cali[0] = std::max(exifWidth,exifHeight) * 1.2; // focal length in x direction
				cali[4] = std::max(exifWidth,exifHeight) * 1.2; // in y direction
				cali[8] = 1.;
				cali[2] = exifWidth / 2.; // principal point x
				cali[5] = exifHeight / 2.; // principal point x

				// Save calibration matrix in output slot
				this->calibration().push_back(cali);
				
				std::string error(filepath);
				error += ": No Exif data found in the file";
				throw Exiv2::Error(1, error);
			}
			
			Exiv2::ExifData::const_iterator end = exifData.end();
			for (Exiv2::ExifData::const_iterator i = exifData.begin(); i != end; ++i) {
				const char* tn = i->typeName();
				/*std::cout << std::setw(44) << std::setfill(' ') << std::left
						  << i->key() << " "
						  << "0x" << std::setw(4) << std::setfill('0') << std::right
						  << std::hex << i->tag() << " "
						  << std::setw(9) << std::setfill(' ') << std::left
						  << (tn ? tn : "Unknown") << " "
						  << std::dec << std::setw(3)
						  << std::setfill(' ') << std::right
						  << i->count() << "  "
						  << std::dec << i->value()
						  << "\n";
				*/
				// Read out focal length
				cimg_library::CImg<float> K;
				if(i->key() == "Exif.Photo.FocalLength"){
					exifFocalLength = i->value().toFloat();
					sout << "Focal Length: " << exifFocalLength << std::endl;
				}
				else if(i->key() == "Exif.Photo.FocalPlaneXResolution"){
					exifXRes = i->value().toFloat();
					sout << "XRes (px/length): " << exifXRes << std::endl;
				}
				else if(i->key() == "Exif.Photo.FocalPlaneYResolution"){
					exifYRes = i->value().toFloat();
					sout << "YRes (px/length): " << exifYRes << std::endl;
				}
				else if(i->key() == "Exif.Photo.FocalPlaneResolutionUnit"){
					exifResUnit = i->value().toLong();
					sout << "FocalPlaneResolutionUnit: ";
					if(exifResUnit == 1) sout << "undefined";
					else if(exifResUnit == 2) sout << "inch";
					else if(exifResUnit == 3) sout << "cm";
					sout << std::endl;
				}	
				else if(i->key() == "Exif.Photo.ExifImageWidth"){
					sout << "Width (px): " << i->value() << std::endl;
				}
				else if(i->key() == "Exif.Photo.ExifImageHeight"){
					sout << "Height (px): " << i->value() << std::endl;
				}	

			}
				
			// Calculate focal length in pixels
			if(exifFocalLength!=0 && exifXRes!=0 && exifYRes!=0 && exifResUnit != 0){

				float xRes, yRes;

				if(exifResUnit == 3) xRes = exifXRes / 10;
				else if(exifResUnit == 2) xRes = exifXRes / 25.4;
				focalLengthX = exifFocalLength * xRes;

				if(exifResUnit == 3) yRes = exifYRes / 10;
				else if(exifResUnit == 2) yRes = exifYRes / 25.4;
				focalLengthY = exifFocalLength * yRes;

				sout << "focalLengthX: " << focalLengthX << std::endl;
				sout << "focalLengthY: " << focalLengthY << std::endl;

				std::vector<float> exifVector;
				exifVector.push_back(exifWidth);
				exifVector.push_back(exifHeight);
				exifVector.push_back(exifFocalLength);
				exifVector.push_back(xRes);
				exifVector.push_back(yRes);
				
				cimg_library::CImg<double> cali(3,3);
				cali.fill(0.);
				cali[0] = focalLengthX; // focal length in x direction
				cali[4] = focalLengthY; // in y direction
				cali[8] = 1.;
				cali[2] = exifWidth / 2.; // principal point x
				cali[5] = exifHeight / 2.; // principal point x

				// Save calibration matrix in output slot
				this->calibration().push_back(cali);

			}
			// If no EXIF sensor information is available, guess calibration
			else{
				sout << "No calibration could be found in EXIF data." << std::endl;
				cimg_library::CImg<double> cali(3,3);
				cali.fill(0.);
				cali[0] = std::max(exifWidth,exifHeight) * 1.2; // focal length in x direction
				cali[4] = std::max(exifWidth,exifHeight) * 1.2; // in y direction
				cali[8] = 1.;
				cali[2] = exifWidth / 2.; // principal point x
				cali[5] = exifHeight / 2.; // principal point x

				// Save calibration matrix in output slot
				this->calibration().push_back(cali);

			}

		}
		//catch (std::exception& e) {
		//catch (Exiv2::AnyError& e) {
		catch (Exiv2::Error& e) {
			std::cout << "Caught Exiv2 exception '" << e.what() << "'\n";
		}

	}


}



// the following functions are needed
// for class ExifReader to work as a charon plugin.
extern "C" exifreader_DECLDIR ParameteredObject*
		create(const std::string& name, template_type) {
	return new ExifReader(name);
}

extern "C" exifreader_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" exifreader_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
