/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FileReaderIndexedMatrix.cpp
 *  Implementation of parameter class FileReaderIndexedMatrix.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 20.02.2013
 */

#include <charon-core/ParameteredObject.hxx>
#include <hekate/FileReaderIndexedMatrix.h>

FileReaderIndexedMatrix::FileReaderIndexedMatrix(const std::string& name) :
		ParameteredObject(
			"FileReaderIndexedMatrix", name,
			"<h2>Reads class IndexedMatrixList from txt file</h2><br>"
			"Reads class IndexedMatrixList from txt file"
		)
{

	ParameteredObject::_addOutputSlot(
		indexedMatrices, "indexedMatrices",
		"indexedMatrices",
		"IndexedMatrixList");
	
	ParameteredObject::_addParameter(
		filename, "filename",
		"File to read from",
		"FileOpen");
	ParameteredObject::_addParameter(
		nMatrices, "nMatrices",
		"Number of matrices to read in", -1,
		"int");

	ParameteredObject::_setTags("Hekate") ;
}

void FileReaderIndexedMatrix::execute() {
			
	std::ifstream file(filename()) ;
	if(filename().empty())
	{	raise("Filename is empty!") ;	}
	if(!file || !file.good())
	{
		raise(std::string("Could not open \"")+ filename() + "\" for reading!") ;
	}

	size_t linenumber = 1 ;

	// Skip first two lines commented lines (by #)
	for( size_t i=0; i<2; i++)
	{
		std::string line ; 
		getline(file, line) ;
		sout << line << std::endl;
	}
			
	// Get matrix dimensions
	unsigned int nMat, w, h;
	file >> nMat >> w >> h ;

	sout << "Read in " << nMat << " matrices" << std::endl;
	sout << "Matrix dimensions: " << w << " columns / " << h << " rows" << std::endl;

	for( size_t k=0; k<nMat; k++)
	{

		if(!file.good())
			break;
		
		bool outlier;

		// Get matrix indices
		int i1, i2;
		file >> i1 >> i2;

		// Get matrix degenerated property
		file >> outlier ;
		
		// Fill CImg matrix
		CImg<double> m(w,h);
		for(size_t i=0; i<h; i++)
			for(size_t j=0; j<w; j++)
				file >> m(j,i);

		// Copy CImg matrix to indexed matrix list
		IndexedMatrix mat(m);

		// Set indices of indexed matrix list
		mat.setIndex(i1,i2);

		// Set outlier property
		mat.degenerated = outlier;

		// Read out calibration properties
		file >> mat.f >> mat.radial >> mat.distortion_type;

		// Save indexed matrix to output list
		this->indexedMatrices().push_back( mat );
		
		linenumber++ ;

		if(!file.good())
			sout<< "!file.good" << std::endl;

	}
	
	file.close();

	
}

// the following functions are needed
// for class FileReaderIndexedMatrix to work as a charon plugin.
extern "C" filereaderindexedmatrix_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new FileReaderIndexedMatrix(name);
}

extern "C" filereaderindexedmatrix_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" filereaderindexedmatrix_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
