/*  Copyright (C) 2010 Julian Coordts

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
//#include "PluginGui.h"
#include <stdio.h>
#include <iostream>
#include <string>
#include <charon-core/ParameteredObject.hxx>
#include <iomanip>

#include "ui_FileSequenceReader.h"

PluginGui::PluginGui(FileSequenceReaderInterface* parent,
			  std::string rightPath,
			  std::string leftPath,
			  std::string rightName,
			  std::string leftName,
			  int numbersCount) :
		_parent(parent) {

	_window = new QWidget();
	_gui.setupUi(_window);
	_gui.leftView->setScene(new QGraphicsScene(this));
	_gui.rightView->setScene(new QGraphicsScene(this));

	_gui.editLeftNames->setText(QString(leftName.c_str()));
	_gui.editRightNames->setText(QString(rightName.c_str()));
	_gui.editLeftDir->setText(QString(leftPath.c_str()));
	_gui.editRightDir->setText(QString(rightPath.c_str()));
	_gui.signsSpinBox->setValue(numbersCount);

	connect(_gui.rangeSlider,SIGNAL(valueChanged(int)),
			this,SLOT(sliderChanged()));
	connect(_gui.searchFilesButton, SIGNAL(clicked()),
			this, SLOT(loadPaths()));
	connect(_gui.runButton, SIGNAL(clicked()),
			this,SLOT(runWorkflow()));

	_window->show();
	loadPaths(false);
}

void PluginGui::runWorkflow() {
	this->_parent->sendNewImagePair(
			curLeftName().toAscii().constData(),
			curRightName().toAscii().constData());
}

void PluginGui::loadPaths(bool message) {
	if (!inputValid()) {
		if (message)
			QMessageBox::warning(
					_window, tr("invalid input"),
					tr("Please make sure that the content<br>"
					   "of the four input fields is valid!"));
		return;
	}

	// pretty paths (without ...bla/../bla/... etc.)
	_gui.editLeftDir->setText(QDir::cleanPath(_gui.editLeftDir->text()));
	_gui.editRightDir->setText(QDir::cleanPath(_gui.editRightDir->text()));

	int seqMin = -1;
	int seqMax = -1;

	//Check sequence index range
	for (int i=0;i<=int(pow(10.,_gui.signsSpinBox->value())) ;i++) {
		if (valid(i)) {
			seqMax=i;
			if(seqMin < 0)
				seqMin=i;
		}
		// stop search at first invalid number
		// if some valid items have been already been found
		else if (seqMin >= 0)
			break;
	}
	if (seqMin < 0) {
		if (message)
			QMessageBox::warning(
					_window, tr("no images found"),
					tr("no images matching the given pattern found!"));
		_gui.resultFrame->setPalette(Qt::red);
		_gui.rangeBox->setEnabled(false);
		_gui.displayBox->setEnabled(false);
		return;
	}

	_gui.rangeSlider->setRange(seqMin,seqMax);
	_gui.rangeSpinBox->setMinimum(seqMin);
	_gui.rangeSpinBox->setMaximum(seqMax);
	_gui.rangeSlider->setValue(seqMin);
	_gui.rangeSpinBox->setValue(seqMin);
	_gui.resultFrame->setPalette(Qt::green);
	_gui.rangeBox->setEnabled(true);
	_gui.displayBox->setEnabled(true);
	sliderChanged();
}

void PluginGui::sliderChanged() {
	const int maxLen = 25;

	// truncate filename strings to display
	QString leftName = curLeftName();
	if (leftName.length() > maxLen) {
		leftName.remove(0, leftName.length()-maxLen+3);
		leftName = QString("...%1").arg(leftName);
	}
	QString rightName = curRightName();
	if (rightName.length() > maxLen) {
		rightName.remove(0, rightName.length()-maxLen+3);
		rightName = QString("...%1").arg(rightName);
	}
	_gui.labelLeftView->setText(leftName);
	_gui.labelRightView->setText(rightName);

	// clear items
	QList<QGraphicsItem*> items = _gui.leftView->scene()->items();
	for(int i=0; i < items.size(); i++) {
		_gui.leftView->scene()->removeItem(items[i]);
		delete items[i];
	}
	items = _gui.rightView->scene()->items();
	for(int i=0; i < items.size(); i++) {
		_gui.rightView->scene()->removeItem(items[i]);
		delete items[i];
	}

	// show new items
	QGraphicsItem* item;
	item = _gui.leftView->scene()->addPixmap(QPixmap(curLeftName()));
	_gui.leftView->fitInView(item,Qt::KeepAspectRatio);
	item = _gui.rightView->scene()->addPixmap(QPixmap(curRightName()));
	_gui.rightView->fitInView(item,Qt::KeepAspectRatio);
}

PluginGui::~PluginGui() {
	_window->close();
	delete _window;
}

QString PluginGui::curLeftName() const {
	return leftName(_gui.rangeSlider->value());
}

QString PluginGui::curRightName() const {
	return rightName(_gui.rangeSlider->value());
}

QString PluginGui::leftName(int i) const {
	if (!inputValid())
		return "";

	return QString("%1/%2")
			.arg(_gui.editLeftDir->text())
			.arg(_gui.editLeftNames->text()
				 .arg(i, _gui.signsSpinBox->value(), 10, QChar('0')));
}

QString PluginGui::rightName(int i) const {
	if (!inputValid())
		return "";

	return QString("%1/%2")
			.arg(_gui.editRightDir->text())
			.arg(_gui.editRightNames->text()
				 .arg(i, _gui.signsSpinBox->value(), 10, QChar('0')));
}

bool PluginGui::inputValid() const {
	if (_gui.editLeftDir->text().isEmpty() ||
			_gui.editRightDir->text().isEmpty() ||
			_gui.editLeftNames->text().isEmpty() ||
			_gui.editRightNames->text().isEmpty() )
		return false;
	if (_gui.editLeftNames->text().indexOf("%1") < 0)
		return false;
	if (_gui.editRightNames->text().indexOf("%1") < 0)
		return false;
	return true;
}

bool PluginGui::valid(int i) const {
	return QFileInfo(leftName(i)).isFile() &&
			QFileInfo(rightName(i)).isFile();
}

bool PluginGui::currentExists() const {
	return valid(_gui.rangeSlider->value());
}

#include "PluginGui.moc"
