/*  Copyright (C) 2013 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file FileWriterIndexedMatrix.cpp
 *  Implementation of parameter class FileWriterIndexedMatrix.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 20.02.2013
 */

#include <charon-core/ParameteredObject.hxx>
#include <hekate/FileWriterIndexedMatrix.h>

FileWriterIndexedMatrix::FileWriterIndexedMatrix(const std::string& name) :
		ParameteredObject(
			"FileWriterIndexedMatrix", name,
			"<h2>Writes class IndexedMatrixList to txt file</h2><br>"
			"Writes IndexedMatrixList to txt file with lines as follows:<br>"
			""
		)
{

	ParameteredObject::_addInputSlot(
		indexedMatrices, "indexedMatrices",
		"indexedMatrices",
		"IndexedMatrixList");

	ParameteredObject::_addParameter(
		filename, "filename",
		"File to write IndexedMatrixList",
		"FileWrite");

	ParameteredObject::_setTags("Hekate") ;
}

void FileWriterIndexedMatrix::execute() {
	
	setlocale(LC_NUMERIC, "C");

	if( this->indexedMatrices().size() == 0)
		throw( std::range_error("IndexedMatrixList is empty!"));

	// Get file path
    std::string file = this->filename() ;

	// Check if input string already contains file type .wrl
	if(file.empty())
	{
		std::string msg = this->ParameteredObject::getClassName() + " : " + this->ParameteredObject::getName() + " : filename is empty!";
		throw std::invalid_argument(msg) ;
	}
	if(file.length() > 4 && file.substr(file.size()-4,4)!=".txt") 
		file.append(".txt");

	FILE* fp; 
	fp = fopen(file.c_str(),"w");
	if(fp == NULL)
	{	throw std::runtime_error(std::string("Could not open file ") + std::string(file)) ;	}
	
	// Write meta data
	fprintf(fp, "# IndexedMatrixList generated by FileWriterIndexedMatrix\n# width, height, ");
	fprintf(fp, "# numberOfMatrices, width, height");
	fprintf(fp, "index1, index2, outlier, ");
	const IndexedMatrix &m = this->indexedMatrices()[0];
	for(size_t i=0; i<m.height(); i++) for(size_t j=0; j<m.width(); j++) fprintf(fp, "(%u,%u), ", i, j);
	fprintf(fp, "focal_length, radial, distortion_type\n");
	
	fprintf(fp, "%u %u %u\n", this->indexedMatrices().size(), m.width(), m.height() );

	// Write camera parameters to file
	for( size_t i=0; i<this->indexedMatrices().size(); i++)
	{
		const IndexedMatrix &mat = this->indexedMatrices()[i] ;
		// index1 index2 outlier
		fprintf(fp, "%u %u %u ",
			mat.getIndex1(), mat.getIndex2(), mat.degenerated
		);
		// m(0,0) ... m(i,j)
		for(size_t i=0; i<m.height(); i++)
			for(size_t j=0; j<m.width(); j++) 
				fprintf(fp, "%f ", mat(j,i));
		// focal_length, radial, distortion_type
		fprintf(fp, "%f %f %u",
			mat.f, mat.radial, mat.distortion_type
		);
		fprintf(fp, "\n");

	}

	fclose(fp);

}

// the following functions are needed
// for class FileWriterIndexedMatrix to work as a charon plugin.
extern "C" filewriterindexedmatrix_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new FileWriterIndexedMatrix(name);
}

extern "C" filewriterindexedmatrix_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" filewriterindexedmatrix_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
