/*  Copyright (C) 2012 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file InitialCameraMatrices.cpp
 *  Implementation of parameter class InitialCameraMatrices.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 22.08.2012
 */

#include <charon-core/ParameteredObject.hxx>
#include <hekate/InitialCameraMatrices.h>

InitialCameraMatrices::InitialCameraMatrices(const std::string& name) :
		ParameteredObject(
			"InitialCameraMatrices", name,
			"<h2>Generates two initial camera matrices</h2><br>"
			"Generates two initial camera matrices"
		),
		camera1("1000;0;0;0;0;1000;0;0;0;0;1;0"),
		camera2("1000;0;0;100;0;1000;0;0;0;0;1;0")
{

	ParameteredObject::_addOutputSlot(
		cameraMatrices, "cameraMatrices",
		"Camera matrices",
		"IndexedMatrixList");
	
	ParameteredObject::_addParameter(
		camera1, "camera1",
		"Camera 1: K*(R|t)");
	ParameteredObject::_addParameter(
		camera2, "camera2",
		"Camera 2: K2*(R2|t2)");
	ParameteredObject::_addParameter(
		index1, "index1",
		"Initial index 1" ,0,
		"int");
	ParameteredObject::_addParameter(
		index2, "index2",
		"Initial index 2", 1,
		"int");

}

void InitialCameraMatrices::execute() {
	PARAMETEREDOBJECT_AVOID_REEXECUTION;
	ParameteredObject::execute();
	
	if(this->camera1().size() != 12)
		throw(std::range_error("InitialCameraMatrices: Invalid number of parameters"));

	if(this->camera2().size() != 12)
		throw(std::range_error("InitialCameraMatrices: Invalid number of parameters"));

	cimg_library::CImg<float> cam1(4,3), cam2(4,3);
	
	// Read in camera 1
	cam1(0,0) = this->camera1()[0];
	cam1(1,0) = this->camera1()[1];
	cam1(2,0) = this->camera1()[2];
	cam1(3,0) = this->camera1()[3];
	cam1(0,1) = this->camera1()[4];
	cam1(1,1) = this->camera1()[5];
	cam1(2,1) = this->camera1()[6];
	cam1(3,1) = this->camera1()[7];
	cam1(0,2) = this->camera1()[8];
	cam1(1,2) = this->camera1()[9];
	cam1(2,2) = this->camera1()[10];
	cam1(3,2) = this->camera1()[11];
	
	// Read in camera 2
	cam1(0,0) = this->camera2()[0];
	cam2(1,0) = this->camera2()[1];
	cam2(2,0) = this->camera2()[2];
	cam2(3,0) = this->camera2()[3];
	cam2(0,1) = this->camera2()[4];
	cam2(1,1) = this->camera2()[5];
	cam2(2,1) = this->camera2()[6];
	cam2(3,1) = this->camera2()[7];
	cam2(0,2) = this->camera2()[8];
	cam2(1,2) = this->camera2()[9];
	cam2(2,2) = this->camera2()[10];
	cam2(3,2) = this->camera2()[11];

	IndexedMatrix cam1i(cam1), cam2i(cam2);
	cam1i.setIndex( this->index1() );
	cam2i.setIndex( this->index2() );
	
	this->cameraMatrices().push_back( cam1i );
	this->cameraMatrices().push_back( cam2i );

}

// the following functions are needed
// for class InitialCameraMatrices to work as a charon plugin.
extern "C" initialcameramatrices_DECLDIR ParameteredObject*
		create(const std::string& name, template_type) {
	return new InitialCameraMatrices(name);
}

extern "C" initialcameramatrices_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" initialcameramatrices_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
