/*  Copyright (C) 2011 Julian Wintermayr

    This file is part of Charon.

    Charon is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Charon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "mainwindow.h"
#include "drawingarea.h"


// Hauptfenster mit Zeichenbereich und Toolbar erstellen und Bild laden
MainWindow::MainWindow(QImage img, std::vector<DrawingArea::drawnLines>* outputPoints,
                                   std::vector<DrawingArea::drawnLines>* outputParallelLines,
                                   std::vector<DrawingArea::drawnLines>* outputOrthographicLines,
                                   QWidget* parent) : QDialog(parent) {

    drawingArea = new DrawingArea;
    scrollArea = new QScrollArea;
    dock = new QDockWidget(tr("Choose a mode to draw"), this);

    dock->setWidget(drawingArea->getTree());
    dock->setFeatures(QDockWidget::NoDockWidgetFeatures);

    scrollArea->setFocusPolicy(Qt::StrongFocus);
    scrollArea->viewport()->installEventFilter(this);

    createActions();
    createToolBar();

    setWindowTitle(tr("ManualUpgrade"));
    resize(1000, 600);

    imageLoaded = false;
    scaleFactor = 1.0;

    pOutputPoints = outputPoints;
    pOutputParallelLines = outputParallelLines;
    pOutputOrthographicLines = outputOrthographicLines;

    layoutV = new QVBoxLayout();
    layoutV->addWidget(mainToolBar);
    layoutV->addWidget(scrollArea, 3);
    layoutV->addWidget(dock);

    setLayout(layoutV);

    // Bild laden
    drawingArea->openImage(img);
    drawingArea->setPixmap(QPixmap::fromImage(*(drawingArea->getImage())));
 
    scrollArea->setWidget(drawingArea);
    scrollArea->setFocus();

    imageLoaded = true;
}


// Rechstklick-Menü
void MainWindow::contextMenuEvent(QContextMenuEvent* event) {

    // Notwendig, damit mit einem Rechtsklick kein Listeneintrag markiert
    // werden kann, wenn schon ein anderer markiert ist

    if(drawingArea->getLineSelected()) {

        QList<QTreeWidgetItem*> selectedItem = drawingArea->getTree()->selectedItems();

        for (int i = 0; i < selectedItem.size(); ++i) {

            selectedItem.at(i)->setSelected(false);
        }

        drawingArea->markListElement(true);
    }

    // Notwendig, damit man auch mit einem Rechtsklick einen Listeneintrag markieren kann
    // und die Linie dazu markiert wird

    if(drawingArea->getMode() != 0 && !drawingArea->getLineSelected()) {

        QList<QTreeWidgetItem*> selectedItem = drawingArea->getTree()->selectedItems();

        if(selectedItem.size() == 1) {

            drawingArea->markLineFromList(selectedItem.at(0), selectedItem.at(0));
        }
    }

    QMenu menu(this);

    if(drawingArea->getMode() == 2) {

        menu.addAction(editLineGroupAct);
    }

    menu.addAction(deleteLineAct);
    menu.exec(event->globalPos());
}


void MainWindow::points() {

    // IF damit man den Modus nicht wechseln kann, solange der Computer noch rechnet und zeichnet

    if(!(drawingArea->getLock1()) && !(drawingArea->getLock2()) && !(drawingArea->getLock3())) {

        dock->setWindowTitle(tr("Points on the plane at infinity"));

        pointsAct->setChecked(true);
        parallelLinesAct->setChecked(false);
        orthographicLinesAct->setChecked(false);

        drawingArea->points();

        dock->setWidget(drawingArea->getTree());

        connect(drawingArea->getTree(), SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)),
                drawingArea, SLOT(markLineFromList(QTreeWidgetItem*, QTreeWidgetItem*)));
    }
}


void MainWindow::parallelLines() {

    if(!(drawingArea->getLock1()) && !(drawingArea->getLock2()) && !(drawingArea->getLock3())) {

        dock->setWindowTitle(tr("Parallel lines"));

        pointsAct->setChecked(false);
        parallelLinesAct->setChecked(true);
        orthographicLinesAct->setChecked(false);

        drawingArea->parallelLines();

        dock->setWidget(drawingArea->getTree());

        connect(drawingArea->getTree(), SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)),
                drawingArea, SLOT(markLineFromList(QTreeWidgetItem*, QTreeWidgetItem*)));
    }
}


void MainWindow::orthographicLines() {

    if(!(drawingArea->getLock1()) && !(drawingArea->getLock2()) && !(drawingArea->getLock3())) {

        dock->setWindowTitle(tr("Orthographic lines"));

        pointsAct->setChecked(false);
        parallelLinesAct->setChecked(false);
        orthographicLinesAct->setChecked(true);

        drawingArea->orthographicLines();

        dock->setWidget(drawingArea->getTree());

        connect(drawingArea->getTree(), SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)),
                drawingArea, SLOT(markLineFromList(QTreeWidgetItem*, QTreeWidgetItem*)));
    }
}


// Gezeichnete Linien als Vektor an die Ausgabe weiterleiten
void MainWindow::calculate() {

    drawingArea->saveVector();

    std::vector<DrawingArea::drawnLines>& points = *pOutputPoints;
    std::vector<DrawingArea::drawnLines>& parallelLines = *pOutputParallelLines;
    std::vector<DrawingArea::drawnLines>& orthographicLines = *pOutputOrthographicLines;

    points.clear();
    parallelLines.clear();
    orthographicLines.clear();

    for(size_t i = 0; i < drawingArea->getPoints().size(); ++i) {

        points.push_back(drawingArea->getPoints()[i]);
    }

    for(size_t i = 0; i < drawingArea->getParallelLines().size(); ++i) {

        parallelLines.push_back(drawingArea->getParallelLines()[i]);
    }

    for(size_t i = 0; i < drawingArea->getOrthographicLines().size(); ++i) {

        orthographicLines.push_back(drawingArea->getOrthographicLines()[i]);
    }

    this->close();
}


// Damit man mit der Leertaste das Bild verschieben kann
void MainWindow::keyPressEvent(QKeyEvent* event) {

    if(event->key() == Qt::Key_Space) {

        drawingArea->createCursor(Qt::ClosedHandCursor);
        QPoint scrollTo = drawingArea->mapFromGlobal(drawingArea->getCursor()->pos());
        scrollArea->ensureVisible(scrollTo.x(), scrollTo.y(), 300, 300);
    }

    if(event->key() == Qt::Key_Control) {

        drawingArea->setSelectionMode(true);
    }

    if(event->key() == Qt::Key_Shift) {

        drawingArea->setMoveMode(true);
    }
}


// Damit sich beim Loslassen der Leertaste der der Cursor wieder verändert
void MainWindow::keyReleaseEvent(QKeyEvent* event) {

    if(event->key() == Qt::Key_Space) {

        drawingArea->createCursor(Qt::CrossCursor);
    }

    if(event->key() == Qt::Key_Control) {

        drawingArea->setSelectionMode(false);
    }

    if(event->key() == Qt::Key_Shift) {

        drawingArea->setMoveMode(false);
    }

    if(event->key() == Qt::Key_Delete) {

        drawingArea->deleteLine();
    }
}


// Damit man mit dem Mausrad nur zoomen und nicht mehr scrollen kann
bool MainWindow::eventFilter(QObject* /*obj*/, QEvent* evt) {

    if(evt->type() == QEvent::Wheel) {

        evt->ignore();
    }

    return true;
}


// Scrollen ist deaktiviert
// Zoomen wird hier definiert
void MainWindow::wheelEvent(QWheelEvent *event) {

    // Man kann nur scrollen, wenn ein Bild geladen ist und sich der Cursor
    // innerhalb der Pixmap befindet

    if(imageLoaded &&
       drawingArea->pixmap()->size().rwidth() >
       drawingArea->mapFromGlobal(drawingArea->getCursor()->pos()).x() &&
       drawingArea->pixmap()->size().rheight() >
       drawingArea->mapFromGlobal(drawingArea->getCursor()->pos()).y()) {

        if(event->delta() > 0) {

            scaleImage(1.25);

        } else if(event->delta() < 0) {

            scaleImage(0.75);
        }

        scrollArea->setFocus();
        event->accept();
    }
}


void MainWindow::scaleImage(float factor) {

    scaleFactor *= factor;

    drawingArea->scale(scaleFactor);

    // Scroll Bar so anpassen, dass das Bild um die Mausposition mittig ist (geht noch nicht richtig --> TODO)
    scrollArea->horizontalScrollBar()->setValue(drawingArea->mapFromGlobal(drawingArea->getCursor()->pos()).x());
    scrollArea->horizontalScrollBar()->setValue(drawingArea->mapFromGlobal(drawingArea->getCursor()->pos()).x());
    scrollArea->verticalScrollBar()->setValue(drawingArea->mapFromGlobal(drawingArea->getCursor()->pos()).y());
    scrollArea->verticalScrollBar()->setValue(drawingArea->mapFromGlobal(drawingArea->getCursor()->pos()).y());

    //scrollArea->horizontalScrollBar()->setValue(drawingArea->getCursor()->pos().x());
    //scrollArea->verticalScrollBar()->setValue(drawingArea->getCursor()->pos().y());
}


// Actions
void MainWindow::createActions() {

    // MainToolBar Menü
    pointsAct = new QAction(QIcon(":/icons/points.png"),
                            tr("Draw points on the plane at infinity"), this);
    connect(pointsAct, SIGNAL(triggered()), this, SLOT(points()));

    parallelLinesAct = new QAction(QIcon(":/icons/parallelLines.png"),
                                   tr("Draw parallel lines"), this);
    connect(parallelLinesAct, SIGNAL(triggered()), this, SLOT(parallelLines()));

    orthographicLinesAct = new QAction(QIcon(":/icons/orthographicLines.png"),
                                       tr("Draw orthographic lines"), this);
    connect(orthographicLinesAct, SIGNAL(triggered()), this, SLOT(orthographicLines()));

    calculateAct = new QAction(QIcon(":/icons/calculate.png"), tr("Calculate"), this);
    calculateAct->setStatusTip(tr("Start calculating"));
    connect(calculateAct, SIGNAL(triggered()), this, SLOT(calculate()));

    // Rechtsklick Menü
    editLineGroupAct = new QAction(tr("Change line group"), this);
    connect(editLineGroupAct, SIGNAL(triggered()), drawingArea, SLOT(editLineGroup()));

    deleteLineAct = new QAction(tr("Delete"), this);
    connect(deleteLineAct, SIGNAL(triggered()), drawingArea, SLOT(deleteLine()));


    pointsAct->setCheckable(true);
    parallelLinesAct->setCheckable(true);
    orthographicLinesAct->setCheckable(true);
}


void MainWindow::createToolBar() {

    mainToolBar = new QToolBar(tr("Main"));
    mainToolBar->addAction(pointsAct);
    mainToolBar->addAction(parallelLinesAct);
    mainToolBar->addAction(orthographicLinesAct);
    mainToolBar->addSeparator();
    mainToolBar->addAction(calculateAct);
    mainToolBar->setMovable(false);

    // PNGs liegen in 200x200px vor
    QSize mainToolBarIconSize(50, 50);
    mainToolBar->setIconSize(mainToolBarIconSize);
}
