
#include "QtOGPCV.h"
#include <QtGui/QMouseEvent>
#include <iostream>
#include <algorithm>
#include <QFontMetrics>
#include <QString>
#include <cmath>
#include "mathfunctions.h"

//Allows easy comparison of position and rotations
bool operator==(const Matrix3fT& lhs, const Matrix3fT& rhs)
{
	for(int i = 0; i < 9; i++){
		if(lhs.M[i] != rhs.M[i])
			return false;
	}
	return true;
}

bool operator==(const Tuple3fT& lhs, const Tuple3fT& rhs)
{
	for(int i = 0; i < 3; i++){
		if(lhs.T[i] != rhs.T[i])
			return false;
	}
	return true;
}

QtOGPCV::QtOGPCV(QWidget *parent) : QGLWidget(parent), ArcBall(640.0f, 480.0f) {

	setMouseTracking(true);

	position.s.X = 0.0;
	position.s.Y = 0.0;
	position.s.Z = 0.0;

	currentCamera = -1;

	pointSize 	= 4.0;
	mode 		= VBO;
	updateList  = false;

	LoadIdentyMatrix(Transform, 4, 4);
	LoadIdentyMatrix(LastRot, 	3, 3);
	LoadIdentyMatrix(ThisRot, 	3, 3);

	isClicked  = false;							
	isRClicked = false;							
	isDragging = false;	

	invertBGColor 	= false;
	drawPoints 		= false;
	useCalibration	= false;
	drawCamBG		= false;
	moveSpeed		= 0.005;

	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(timerTimeout()));
	timer->setInterval(20);

}

void QtOGPCV::initializeGL() {

	// Set required OpenGL Parameters
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glEnable(GL_POINT_SMOOTH);
	glClearColor(0, 0, 0, 0);

	// Init VBOs
	if(!vertexVBO.create() || !colorVBO.create() || !vertexVBO.isCreated() || !colorVBO.isCreated()){

		// Fallback to displaylists
		mode = DISPLAYLIST; 
		indexList = glGenLists(1);

		// Fallback to slow mode with glbegin/glend
		if(!indexList){
			mode = SLOW; 
		}

	}

	timer->start();
}

void QtOGPCV::resizeGL(int w, int h) {
	glViewport (0, 0,  w, h); 
	width = w;
	height = h;
	ArcBall.setBounds((GLfloat)w, (GLfloat)h);	
	updateGL();
}

void QtOGPCV::paintGL() {

	// Clear screen
	glClear(GL_COLOR_BUFFER_BIT);

	if (points.size() <= 0)
		return;

	// Process ArcBall
	if (isRClicked)	{

		// Reset view - Rightclick
		if(currentCamera != -1){

			// Reset view to a camera
			LastRot = ThisRot = this->cameras[currentCamera].rotation;
			this->position = this->cameras[currentCamera].position;
		}else{

			// Reset to default view
			Matrix3fSetIdentity(&LastRot);
			Matrix3fSetIdentity(&ThisRot);
			position.s.X = 0.0;
			position.s.Y = 0.0;
			position.s.Z = 0.0;
		}
		Matrix4fSetRotationFromMatrix3f(&Transform, &ThisRot);
	}

	if (!isDragging){

		if (isClicked){
			isDragging = true;							
			LastRot = ThisRot;							
			ArcBall.click(&MousePt);						
		}

	}else{

		if (isClicked){

			Quat4fT ThisQuat;

			ArcBall.drag(&MousePt, &ThisQuat);					
			Matrix3fSetRotationFromQuat4f(&ThisRot, &ThisQuat);			
			Matrix3fMulMatrix3f(&ThisRot, &LastRot);				
			Matrix4fSetRotationFromMatrix3f(&Transform, &ThisRot);	
		
		}else{										
			isDragging = false;
		}
	}

	// Process Keys
	GLfloat f;

	for (std::map<Qt::Key, QElapsedTimer>::iterator it = keys.begin(); it!=keys.end(); it++) {

		Qt::Key key = (*it).first;    	

		switch(key){
			case Qt::Key_R:
			case Qt::Key_F:
				f = ( key == Qt::Key_R ) ? -moveSpeed : moveSpeed;
				f *= (*it).second.restart();
				

				this->position.s.X += this->Transform.s.M10 * f;
				this->position.s.Y += this->Transform.s.M11 * f;
				this->position.s.Z += this->Transform.s.M12 * f;
				break;

			case Qt::Key_W:
			case Qt::Key_S:
				f = ( key == Qt::Key_S ) ? -moveSpeed : moveSpeed;
				f *= (*it).second.restart();

				this->position.s.X += this->Transform.s.M20 * f;
				this->position.s.Y += this->Transform.s.M21 * f;
				this->position.s.Z += this->Transform.s.M22 * f;
				break;
				
			case Qt::Key_D:
			case Qt::Key_A:
				f = ( key == Qt::Key_D ) ? -moveSpeed : moveSpeed;
				f *= (*it).second.restart();

				this->position.s.X += this->Transform.s.M00 * f;
				this->position.s.Y += this->Transform.s.M01 * f;
				this->position.s.Z += this->Transform.s.M02 * f;
				break;
			default:
				break;
		}
	}

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	// load calibration matrix / standard view
	if (useCalibration && currentCamera != -1){
		glMultMatrixf(this->cameras[currentCamera].calibration.M); 
	}else{
		gluPerspective(45, (double)width/height, 1, 50);
	}

	// apply rotation matrix
	glMultMatrixf(Transform.M);	

	// move camera to it's position
	glTranslatef(position.s.X, position.s.Y, position.s.Z);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity();

	// set pointsize
	glPointSize(pointSize);

	// draw points
	if(mode == VBO){

		// Allocate memory for a colour and vertex array
		// Transfer the array into the VRAM and free it

		if(updateList){

			updateList = false;

			int size = points.size();

			if(vertexVBO.isCreated() && colorVBO.isCreated() && size > 0){

				float *colors = NULL;
				float *vetrex = NULL;
		
				try{	
					vetrex = new float[size*3];
					colors = new float[size*3];
				}catch(...){
					if(vetrex) delete[] vetrex;
					if(colors) delete[] colors;
					return;
				} 

				std::vector<point>::iterator it;

				int currentColor 	= 0;
				int currentVetrex 	= 0;

				for ( it=points.begin(); it != points.end(); it++ ){

					colors[currentColor] = it->r; currentColor++;
					colors[currentColor] = it->g; currentColor++;
					colors[currentColor] = it->b; currentColor++;
				
					vetrex[currentVetrex] = it->x; currentVetrex++;
					vetrex[currentVetrex] = it->y; currentVetrex++; 
					vetrex[currentVetrex] = it->z; currentVetrex++;
					
				}

				vertexVBO.bind();
				vertexVBO.allocate(vetrex, size * 3 * sizeof(float));
				vertexVBO.release();
		 
				colorVBO.bind();
				colorVBO.allocate(colors, 	size * 3 * sizeof(float));	
				colorVBO.release();	

				delete[] vetrex;
				delete[] colors;

			}

		}

		// Draw points
		vertexVBO.bind();
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(3, GL_FLOAT, 0, 0);

		colorVBO.bind();
		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(3, GL_FLOAT, 0, 0);
		glDrawArrays(GL_POINTS, 0, points.size());		

		glDisableClientState(GL_COLOR_ARRAY);
		colorVBO.release();
		glDisableClientState(GL_VERTEX_ARRAY);
		vertexVBO.release();

		if(drawPoints) drawPointsNr();
	
		
	}else if(mode == DISPLAYLIST){

		// Update displaylist
		if(updateList){

			updateList = false;

			glNewList(indexList, GL_COMPILE_AND_EXECUTE);

			this->paintPoints();
	
			glEndList();			

		}else{
			// Draw displaylist
			glCallList(indexList);
		}

	}else{
		// Paint points
		this->paintPoints();
	}


	if (currentCamera != -1){

		//Try to load image, if it is not loaded yet
		if (!this->cameras[currentCamera].textur){
			this->cameras[currentCamera].textur = this->bindTexture(this->cameras[currentCamera].image);
		}

		if (drawCamBG && this->cameras[currentCamera].textur &&
			LastRot == ThisRot && ThisRot == this->cameras[currentCamera].rotation &&	 
			this->position == this->cameras[currentCamera].position){

			/*
				Disable perspective mode and switch to orthogonal 2D rendering mode.
				Change alpha blending and draw the image on a rectangle with opacity 0.7.
				This allows us to view the points behind the image.
			*/

			glMatrixMode (GL_PROJECTION);
			glLoadIdentity();
			glOrtho(-1,1,-1,1, 1, 100);


			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, this->cameras[currentCamera].textur);	

			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

			glMatrixMode(GL_TEXTURE);
			glLoadIdentity();

			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

			glColor4f(1.0, 1.0, 1.0, 0.7);

			glBegin(GL_QUADS);
				glTexCoord2i(0,	0); glVertex3f(-1, -1, -2);
				glTexCoord2i(1, 0); glVertex3f( 1, -1, -2);
				glTexCoord2i(1, 1); glVertex3f( 1,  1, -2);
				glTexCoord2i(0,	1); glVertex3f(-1,  1, -2);
			glEnd();		

			glBlendFunc(GL_ONE, GL_ZERO);

			glDisable(GL_TEXTURE_2D);

		}
	}

	// Draw text about current view, invert colour if necessary

	if(invertBGColor){
		qglColor(QColor("#000000"));
	}else{
		qglColor(QColor("#ffffff"));
	}

	QFontMetrics metrics(this->font());
	if (currentCamera == -1){
		renderText (5, metrics.height(), "No Camera selected");
	}else{
		if(useCalibration){
			renderText (5, metrics.height(), QString("Camera: ") + QString::number(this->cameras[currentCamera].originalNr) + QString(" (using calibration)"));
		}else{
			renderText (5, metrics.height(), QString("Camera: ") + QString::number(this->cameras[currentCamera].originalNr));
		}
	}
}

void QtOGPCV::paintPoints(){

	// draw points with glbegin/glEnd();
	std::vector<point>::iterator it;

	for ( it=points.begin(); it != points.end(); it++ ){

		glColor3f(it->r,it->g,it->b); 

		glBegin(GL_POINTS);
			glVertex3f(it->x, it->y, it->z);
		glEnd();	
	}

	if(drawPoints) drawPointsNr();
}

void QtOGPCV::drawPointsNr(){

	// draw points - very slow, because the data is not cached, just for debugging
	std::vector<point>::iterator it;

	if(invertBGColor){
		qglColor(QColor("#000000"));
	}else{
		qglColor(QColor("#ffffff"));
	}

	for ( it=points.begin(); it != points.end(); it++ ){
		this->renderText(it->x, it->y, it->z, QString::number(it->number));
	}

}

void QtOGPCV::mousePressEvent(QMouseEvent *event) {
	
	if( event->button () & Qt::LeftButton )  isClicked  = true;
	if( event->button () & Qt::RightButton ) isRClicked = true;
	updateGL();

}
void QtOGPCV::mouseMoveEvent(QMouseEvent *event) {

	MousePt.s.X = event->x();
	MousePt.s.Y = event->y();
	updateGL();

}

void QtOGPCV::mouseReleaseEvent(QMouseEvent *event){
	if( !(event->buttons () & Qt::LeftButton) )  isClicked  = false;
	if( !(event->buttons () & Qt::RightButton) ) isRClicked = false;
	updateGL();	
}

void QtOGPCV::keyPressEvent(QKeyEvent* event) {

	// ignore auto repeated key presses (for smooth movement)
	if (event->isAutoRepeat())
		return;

	switch(event->key()) {

		// Keys for movement
		case Qt::Key_R:
		case Qt::Key_F:
		case Qt::Key_W:
		case Qt::Key_S:
		case Qt::Key_D:
		case Qt::Key_A:

			if ( keys.find((Qt::Key)event->key()) == keys.end()){
				QElapsedTimer timer;
				timer.start();
				keys[(Qt::Key)event->key()] = timer;
			}

			event->accept();
			break;
		
		// Keys to switch cameras
		case Qt::Key_Right:
		case Qt::Key_Left:

			if (this->cameras.size() > 0 ){

				if (event->key() == Qt::Key_Right)
					if (currentCamera + 1 > this->cameras.size() -1){
						currentCamera = -1;
					}else{
						currentCamera = currentCamera + 1;
					}
				else
					currentCamera = (currentCamera - 1) < -1 ? this->cameras.size() -1 : currentCamera - 1;
			
				if (currentCamera == -1){

					Matrix3fSetIdentity(&LastRot);
					Matrix3fSetIdentity(&ThisRot);
					position.s.X = 0.0;
					position.s.Y = 0.0;
					position.s.Z = 0.0;

				}else{
					this->position = this->cameras[currentCamera].position;
					LastRot = ThisRot = this->cameras[currentCamera].rotation;
				}

				Matrix4fSetRotationFromMatrix3f(&Transform, &ThisRot);

				updateGL();
			}

			event->accept();
			break;

		// Keys to control movement speed
		case Qt::Key_Up:
		case Qt::Key_Down:

			if (event->key() == Qt::Key_Up)
				moveSpeed += 0.001;
			else
				moveSpeed -= 0.001;

			if (moveSpeed < 0.001)
				moveSpeed = 0.001;

			event->accept();
			break;

		// Invert background colour
		case Qt::Key_I:

			if (!invertBGColor){
				glClearColor(1, 1, 1, 0);
				invertBGColor = true;
			}else{
				glClearColor(0, 0, 0, 0);
				invertBGColor = false;
			}
			updateGL();

			event->accept();
			break;

		// Show points
		case Qt::Key_P:

			drawPoints = !drawPoints;
			updateGL();

			event->accept();
			break;

		// Show background image for cameras
		case Qt::Key_B:

			drawCamBG = !drawCamBG;
			updateGL();

			event->accept();
			break;

		// Toggle calibration matrix
		case Qt::Key_C:

			useCalibration = !useCalibration;
			updateGL();

			event->accept();
			break;

		default:
			event->ignore();
			break;
	}
}

void QtOGPCV::keyReleaseEvent(QKeyEvent *event){

	if (event->isAutoRepeat())
		return;

	switch(event->key()) {				
		case Qt::Key_R:
		case Qt::Key_F:
		case Qt::Key_W:
		case Qt::Key_S:
		case Qt::Key_D:
		case Qt::Key_A:
			keys.erase((Qt::Key)event->key());
			event->accept();
			break;
	}

}

// This timer is only used to update the image while at least one key is pressed
void QtOGPCV::timerTimeout(){

	if (keys.size() == 0)
		return;

	updateGL();
}

// Load points
void QtOGPCV::setPoints(const hekate::CorrespondenceList<T> &pointCloud){

	this->points.clear();

	hekate::CorrespondenceList<T>::ListAsVec::iterator it = pointCloud.begin();
	
	for(; it != pointCloud.end(); it++){
		point newPoint;

		// Invert Y and Z axis, because the opengl coordinate system differs
		newPoint.x =  it->point(0);
		newPoint.y = -it->point(1);
		newPoint.z = -it->point(2);

		newPoint.r = (double)it->R/255.0;
		newPoint.g = (double)it->G/255.0;
		newPoint.b = (double)it->B/255.0;

		newPoint.number = it->pointIndex;

		this->points.push_back(newPoint);
	};
	
	this->points = points;
	updateList = true;

	updateGL();
}

// Change the size of a displayed point
void QtOGPCV::setPointSize(float size){
	this->pointSize = size;
	updateList = true;
	updateGL();
}

void QtOGPCV::freeCameras(){

	// Check for valid rendering context and free camera textures if loaded
	if(this->isValid ()){

		this->makeCurrent();

		for(std::vector<camera>::iterator it = this->cameras.begin(); it != this->cameras.end(); it++){
			if (it->textur)
				this->deleteTexture(it->textur);
		}

	}

	this->cameras.clear();
}

void QtOGPCV::setCameras(const IndexedMatrixList &cameras, const std::vector<QImage> &images){

	// free old cameras and textures
	this->freeCameras();

	for(unsigned int nr = 0; nr < cameras.size(); nr++){

		// if we do not have an image for this camera, ignore it
		if (nr >= images.size())
			break;

		double matrix[4*4];
		double inverse[4*4];

		// load input matrix into the top left corner of a new 4x4 matrix
		for(int i = 0; i < 12; i++)
			matrix[i] = cameras[nr][i];
		for(int i = 12; i < 15; i++)
			matrix[i] = 0;
		matrix[15] = 1;

		// invert our matrix
		invertMatrix4x4(matrix, inverse);

		// The position is given by: -inverse[i*4+3]. 
		// Y and Z have to be swapped because of the different coordinate system
		double X = -inverse[0*4+3];
		double Y =  inverse[1*4+3];
		double Z =  inverse[2*4+3];

		// create a new camera and set position & number
		camera newCam;
		newCam.position.s.X = X;
		newCam.position.s.Y = Y;
		newCam.position.s.Z = Z;
		newCam.originalNr	= nr;

		// copy values to temp3
		double temp3[9];
		for (int i = 0; i < 3; i++){
			for(int j = 0; j < 3; j++){	
				temp3[i*3+j] = inverse[i*4+j]; 
			}
		}

		// split up the calibration in temp3 and rotation in q
		double q[9];
		decomposeQR3x3(temp3, q);

		// Rotations always have detQ = 1 - swap sign if it isnt a valid rotation yet
		double detQ = q[0]*(q[4]*q[8] - q[5]*q[7]) + q[1]*(q[5]*q[6] - q[3]*q[8]) + q[2]*(q[3]*q[7] - q[4]*q[6]);		
		if( detQ < 0 ){	
			for(int i=0; i < 3; i++){
				for(int j=0; j < 3; j++){
					q[i*3+j] *= -1;
				}
			}
		}

		// Convert to other coordinate system
		for(int i = 0; i < 3; i++){
			q[0*3+i] *= -1.0;
			q[i*3+0] *= -1.0;
		}

		// invert temp3 to get calibration matrix
		double K[9];
		invertMatrix3x3(temp3, K);

		// save rotation matrix
		// Normally we would have to calculate the inverse here, which is given by q[j*3+i] 
		// but as OpenGL matrices are always transposed, we can just copy
		for (int i = 0; i < 3; i++){
			for(int j = 0; j < 3; j++){
				newCam.rotation.M[i*3+j] = q[i*3+j];
			}
		}

		// Normalize K
		for(int i = 0; i < 3; i++){
			for(int j = 0; j < 3; j++){
				K[i*3+j] /= K[2*3+2];
			}
		}

		// Copy image and get height / width
		newCam.image	= images[nr];
		newCam.textur	= 0;
		newCam.width 	= newCam.image.width();
		newCam.height 	= newCam.image.height();

		double znear = 1;
		double zfar  = 100;

		// Convert calibration matrix to OpenGL projection matrix
		// See http://strawlab.org/2011/11/05/augmented-reality-with-OpenGL/
		newCam.calibration.M[0*4+0] = 2*K[0*3+0] / newCam.width;
		newCam.calibration.M[1*4+0] = -2*K[0*3+1] / newCam.width;
		newCam.calibration.M[2*4+0] = -2*K[0*3+2] / newCam.width + 1.0;
		newCam.calibration.M[3*4+0] = 0;

		newCam.calibration.M[0*4+1] = 0;
		newCam.calibration.M[1*4+1] = 2*K[1*3+1] / newCam.height;
		newCam.calibration.M[2*4+1] = 2*K[1*3+2] / newCam.height - 1.0; 
		newCam.calibration.M[3*4+1] = 0;

		newCam.calibration.M[0*4+2] = 0;
		newCam.calibration.M[1*4+2] = 0;
		newCam.calibration.M[2*4+2] = -(zfar + znear)/(zfar - znear);
		newCam.calibration.M[3*4+2] = -2*zfar*znear/(zfar - znear);

		newCam.calibration.M[0*4+3] = 0;
		newCam.calibration.M[1*4+3] = 0;
		newCam.calibration.M[2*4+3] = -1;
		newCam.calibration.M[3*4+3] = 0;

		this->cameras.push_back(newCam);
	}

	// reset view
	currentCamera = -1;
	Matrix3fSetIdentity(&LastRot);
	Matrix3fSetIdentity(&ThisRot);
	position.s.X = 0.0;
	position.s.Y = 0.0;
	position.s.Z = 0.0;

	updateGL();

	return;

}

QtOGPCV::~QtOGPCV(){
	this->freeCameras();	
}