#include "hekate/TriangulationGPU_cu.h"
#define CUDA
#include "hekate/totalLstsqMinimize.h"

#define BLOCK_SIZE 16

__global__ void mulAtA(float *Matrices, ControlStep *Control, int controlCount, float *Output){
	
	int x = blockDim.x * blockIdx.x + threadIdx.x;
  	int y = blockDim.y * blockIdx.y + threadIdx.y;

	int index = y * gridDim.x * blockDim.x + x;

	if(index >= controlCount)
		return;

	ControlStep1 &current = Control[index].step1;
	caculateAtmulA(&Matrices[current.startRow * 4], current.numRows, &Output[current.outIndex * 16]);	
}

__global__ void calcEigenvector(float *Matrices, ControlStep *Control, int controlCount){
	
	int x = blockDim.x * blockIdx.x + threadIdx.x;
  	int y = blockDim.y * blockIdx.y + threadIdx.y;

	int index = y * gridDim.x * blockDim.x + x;

	if(index >= controlCount)
		return;

	ControlStep2 &current = Control[index].step2;

	if (current.outIndex[0] == MATRIX_OUTINDEX_INVALID)
		return;

	float *Matrix = &Matrices[current.outIndex[0] * 16];
	
	for (int addition = 1; addition < MATRIX_MAX_MERGE; addition++){

		if (current.outIndex[addition] == MATRIX_OUTINDEX_INVALID)
			continue;
		
		float *AddMatrix = &Matrices[current.outIndex[addition] *16];

		for(int i = 0; i < 4; i++)
			for(int j = 0; j < 4; j++)
				Matrix[i * 4 + j] += AddMatrix[i * 4 + j];
	} 
	
	if (current.complete){
		if(!minEigenvector4x4(Matrix, &Control[index].result[0], false)){

			// Show that we dont have any result
			Control[index].result[0] = __int_as_float(0xFFE00000); //NaN
			Control[index].result[1] = __int_as_float(0xFFE00000); //NaN
			Control[index].result[2] = __int_as_float(0xFFE00000); //NaN
			Control[index].result[3] = __int_as_float(0xFFE00000); //NaN
	
		}
	}
}



GPUTriangulate::GPUTriangulate(){
	
}

int GPUTriangulate::init(){

	initSuccess = false;

	int deviceCount;

	if(cudaGetDeviceCount(&deviceCount) != cudaSuccess || deviceCount <= 0)
		return false;

	cudaSetDevice(0);

	if (cudaMalloc(&MatricesDevice, MATRIX_MAX_ROWS * 4 * sizeof(float)) == cudaSuccess){

		if (cudaMalloc(&ControlDevice, MATRIX_CONTROL_COUNT  * sizeof(ControlStep)) == cudaSuccess){
			
			if (cudaMalloc(&MulOutputDevice, MATRIX_OUTPUT_COUNT * 4 * 4 * sizeof(float)) == cudaSuccess){

				initSuccess = true;
				return true;

			}

			cudaFree(ControlDevice);
		}

		cudaFree(MatricesDevice);
	}

	return false;
}

int GPUTriangulate::executeStep1(float *Matrices, int matrixCount, ControlStep *Control, int controlCount){


	if(initSuccess){

		if (cudaMemcpy(MatricesDevice, Matrices, matrixCount * 4 * sizeof(float),  cudaMemcpyHostToDevice) != cudaSuccess)
			return false;

		if (cudaMemcpy(ControlDevice, Control, controlCount * sizeof(ControlStep),  cudaMemcpyHostToDevice) != cudaSuccess)
			return false;
		
		int dimCount 	= ceil(std::sqrt((double)controlCount));
		int gridSizeDim = (dimCount + BLOCK_SIZE - 1) / BLOCK_SIZE; 				

		dim3 dimGrid	(gridSizeDim, gridSizeDim);	
		dim3 dimBlock	(BLOCK_SIZE, BLOCK_SIZE);
		
		mulAtA<<< dimGrid, dimBlock >>> (MatricesDevice, ControlDevice, controlCount, MulOutputDevice);
		
		if(cudaGetLastError() == cudaSuccess)
			return true;
	}

	return false;	
}

int GPUTriangulate::executeStep2(ControlStep *Control, int controlCount){


	if(initSuccess){

		if (cudaMemcpy(ControlDevice, Control, controlCount * sizeof(ControlStep),  cudaMemcpyHostToDevice) != cudaSuccess)
			return false;
		

		int dimCount 	= ceil(std::sqrt((double)controlCount));
		int gridSizeDim = (dimCount + BLOCK_SIZE - 1) / BLOCK_SIZE; 				

		dim3 dimGrid	(gridSizeDim, gridSizeDim);	
		dim3 dimBlock	(BLOCK_SIZE, BLOCK_SIZE);
		
		calcEigenvector<<< dimGrid, dimBlock >>> (MulOutputDevice, ControlDevice, controlCount);
		if(cudaGetLastError() != cudaSuccess)
			return false;

		if (cudaMemcpy(Control, ControlDevice, controlCount * sizeof(ControlStep),  cudaMemcpyDeviceToHost) != cudaSuccess)
			return false;		

		return true;
	}

	return false;	
}



GPUTriangulate::~GPUTriangulate(){
	
	if(initSuccess){

		cudaFree(MulOutputDevice);
		cudaFree(ControlDevice);
		cudaFree(MatricesDevice);		
	}

}
