#include <QtTest>

#include "openCvUtil.h"
#include <openCvUtil.hxx>
#include <charon-utils/CImg.h>

#include <iostream>
#include <climits>

using namespace cimg_library ;

template <typename T>
bool isMatValid(const cv::Mat& mat, const cimg_library::CImg<T>& img)
{
	if (mat.empty() || mat.cols != img.width() || mat.rows != img.height())
	{	return false ;	}
	
	// matrix data should not be user-allocated but be managed by Mat itself
	if (mat.refcount == 0 || mat.refcount == NULL)
	{	return false ; }

	return true ;
}

template <typename T>
bool areMatEqu(const cv::Mat& mat, const cimg_library::CImg<T>& img)
{
	cimg_forXY(img, x, y)
	{
		//mind the switched xy-dimensions for Mat objects!
		if(mat.at<T>(y, x) != img(x, y))
		{	return false ;	}
	}
	return true ;
}

template <typename T, typename T2>
bool areMatEqu(const cv::Mat& mat, T, const cimg_library::CImg<T2>& img)
{
	cimg_forXY(img, x, y)
	{
		//mind the switched xy-dimensions for Mat objects!
		if(mat.at<T>(y, x) != T(img(x, y)))
		{	return false ;	}
	}
	return true ;
}

void testCimg2Mat::init() {
	width = 13;
	height = 19;
}

void testCimg2Mat::testCImgFloat() {
	CImg<float> img(width, height, 1, 1, 0.0) ;
	mat1 = openCvUtil::CImg2Mat(img, CV_32F) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1,img));

	img.rand(0.0, 1e6) ;
	mat1 = openCvUtil::CImg2Mat(img, CV_32F) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1,img));

	mat1 = openCvUtil::CImg2Mat(img, CV_8U) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1, (unsigned char)(0u), img));
}

void testCimg2Mat::testCImgDouble() {
	CImg<double> img(width, height, 1, 1, 0.0) ;
	img.rand(-1e6, 1e6) ;
	mat1 = openCvUtil::CImg2Mat(img, CV_64F) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1,img));

	mat1 = openCvUtil::CImg2Mat(img, CV_16S) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1, (signed short)(0), img));
}

void testCimg2Mat::testCImgUChar() {
	CImg<unsigned char> img(width, height, 1, 1, 0) ;
	img.rand(0, UCHAR_MAX) ;
	mat1 = openCvUtil::CImg2Mat(img, CV_8U) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1,img));
}

void testCimg2Mat::testCImgChar() {
	CImg<signed char> img(width, height, 1, 1, 0) ;
	img.rand(CHAR_MIN, CHAR_MAX) ;
	mat1 = openCvUtil::CImg2Mat(img, CV_8S) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1,img));
}

void testCimg2Mat::testCImgUShort() {
	CImg<unsigned short> img(width, height, 1, 1, 0) ;
	img.rand(0, USHRT_MAX) ;
	mat1 = openCvUtil::CImg2Mat(img, CV_16U) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1,img));
}

void testCimg2Mat::testCImgShort() {
	CImg<signed short> img(width, height, 1, 1, 0) ;
	img.rand(SHRT_MIN, SHRT_MAX) ;
	mat1 = openCvUtil::CImg2Mat(img, CV_16S) ;
	QVERIFY(isMatValid(mat1, img));
	QVERIFY(areMatEqu(mat1,img));
}


void testMat2CImg::init() {
	width = 13;
	height = 19;
}

void testMat2CImg::test32FtoFloat() {
	cv::Mat mat(height, width, CV_32F, cv::Scalar(0)) ;
	cimg_library::CImg<float> img =
			openCvUtil::Mat2CImg<float>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QVERIFY(areMatEqu(mat,img));
}

void testMat2CImg::test32FtoDouble() {
	cv::Mat mat(height, width, CV_32F, cv::Scalar(0)) ;
	cimg_library::CImg<double> img =
			openCvUtil::Mat2CImg<double>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QVERIFY(areMatEqu(mat,float(0),img));
}

void testMat2CImg::test64FtoFloat() {
	cv::Mat mat(height, width, CV_64F, cv::Scalar(0)) ;
	cimg_library::CImg<float> img =
			openCvUtil::Mat2CImg<float>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QVERIFY(areMatEqu(mat,double(0),img));
}

void testMat2CImg::test64FtoDouble() {
	cv::Mat mat(height, width, CV_64F, cv::Scalar(0)) ;
	cimg_library::CImg<double> img =
			openCvUtil::Mat2CImg<double>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QVERIFY(areMatEqu(mat,double(0),img));
}

void testMat2CImg::test64FtoChar() {
	cv::Mat mat(height, width, CV_64F, cv::Scalar(0)) ;
	cimg_library::CImg<char> img =
			openCvUtil::Mat2CImg<char>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QVERIFY(areMatEqu(mat,double(0),img));
}

void testMat2CImg::testCasting64FtoChar() {
	cv::Mat mat(height, width, CV_64F, cv::Scalar(1e6)) ;
	cimg_library::CImg<char> img =
			openCvUtil::Mat2CImg<char>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QCOMPARE((int)img.max(), (int)(char)int(1e6));
	QCOMPARE((int)img.min(), (int)(char)int(1e6));
}

void testMat2CImg::testCasting64FtoUChar() {
	cv::Mat mat(height, width, CV_64F, cv::Scalar(1e6)) ;
	cimg_library::CImg<unsigned short> img =
			openCvUtil::Mat2CImg<unsigned short>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QCOMPARE((int)img.max(), (int)(unsigned short)int(1e6));
	QCOMPARE((int)img.min(), (int)(unsigned short)int(1e6));
}

void testMat2CImg::test8UtoUchar() {
	cv::Mat mat(height, width, CV_8U, cv::Scalar(0)) ;
	cimg_library::CImg<unsigned char> img =
			openCvUtil::Mat2CImg<unsigned char>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QVERIFY(areMatEqu(mat,(unsigned char)(0),img));
}

void testMat2CImg::test8UtoFloat() {
	cv::Mat mat(height, width, CV_8U, cv::Scalar(0)) ;
	cimg_library::CImg<float> img =
			openCvUtil::Mat2CImg<float>(mat) ;
	QCOMPARE(img.width(), width);
	QCOMPARE(img.height(), height);
	QCOMPARE(img._is_shared, false);
	QVERIFY(areMatEqu(mat,(unsigned char)(0),img));
}

int main(int argc, char *argv[]) {
	testCimg2Mat t1;
	testMat2CImg t2;
	QTest::qExec(&t1, argc, argv);
	QTest::qExec(&t2, argc, argv);
}

#include "openCvUtil.moc"
